
#ifndef Askme_includeD
#define Askme_includeD yes

typedef int (*Cleanup_app_handler_T)();


struct Bitmap2D {
 int width;
 int size;
 char *map;
};


struct LreadeR { /* A reader context for scdbackup's content list format
                    see: cd_backup_planer -complete_list_adr */
 char adr[4096];
 FILE *fp;
 char line[3*4096];
 off_t linecount;
 char linebuffer[3*4096]; /* temporary changeable copy of parts of .line */
 int errcode; /* use Lreader_strerror() to obtain message string */

 char list_type[81];
 int has_wd;
 char initial_wd[4096];
 char id[81];
 char date[81];
 int volumes_start_at_0;
 int layout; /*
      bit0= type
      bit1= chmod
      bit2= owner
      bit3= group
      bit4= date
      bit5= size
      bit6= checksum
    */
 int highest_volume;

 int level;
 char younger_than[81];
 int volume;

 int valid; /*
      bit0= type
      bit1= chmod
      bit2= owner
      bit3= group
      bit4= date
      bit5= size
      bit6= checksum

      bit12= md5 rather than crc
      bit13= inode number
      bit14= device number
      bit15= mtime
      bit16= ctime
    */
 char *line_source;
 char absolute_source[4096];
 char *source_cursor;
 char *line_target;
 char absolute_target[4096];
 char *target_cursor;

 char type_string[4096];
 int file_type; 
   /* 0=unknown
      1=regular file
      2=directory
      3=symbolic link
      4=named pipe
      5=multiple hardlink (if requested to be marked as such)  
      6=block device
    */
 int part_status;
   /* 0=no partial files involved
      1=file part ("p")
      2=full original of part ("P")
    */
 off_t part_start;
 off_t part_size;
 off_t part_full_size;
 ino_t inode;
 dev_t device;
 off_t file_size;
 unsigned long crc32;
 char md5[16];
 int permissions; /* see struct FsinfO.permissions */
 char owner[4096];
 char group[4096];
 char timestamp[81]; /* from directory list */
 time_t mtime;       /* from content list ,t... */
 time_t ctime;       /* from content list ,c... */

 /* Convenience payload. Else those would be frequent function parameters.*/
 struct FsnodE *source_root;
 struct FsnodE *target_root;

};


struct FsrptmaP { /* a list used to keep track of repeated matches */
 char *name;
 struct Bitmap2D *bitmap;
 struct FsrptmaP *next;
};


/* translates FsinfO.permissions (classic layout of .st_mode)
   to stat.st_mode (local layout of .st_mode) */
#define Fsinfo_mode_masK (S_ISUID|S_ISGID|S_ISVTX|S_IRWXU|S_IRWXG|S_IRWXO)
#define Fsinfo_modebit_counT 12
static int Fsinfo_modebitS[Fsinfo_modebit_counT]= 
                        {S_IXOTH,S_IWOTH,S_IROTH,
                         S_IXGRP,S_IWGRP,S_IRGRP,
                         S_IXUSR,S_IWUSR,S_IRUSR,
                         S_ISVTX,S_ISGID,S_ISUID};


struct FsinfO { /* registers additional properties of a FsnodE */

 int flag;
    /* bit0-bit6: type
              0= type not detected yet (or refuses to tell)
            compatible with return of Type_and_size(,16) :
              1= data file
              2= directory
              3= link
              4= ( named pipe , not used yet)
              5= ( multiple hardlink , probably never used )
              6= block device
              7= block device
            127= other type (already evaluated)
       bit7: deleted entry
       bit10: size is valid
       bit11: crc is valid
       bit12: md5 is valid
       bit13: inode is valid
   */

#ifndef Askme_fsnode_with_sizE
 off_t size;
#endif

 unsigned long crc32;
 char *md5;
 ino_t inode;
 int permissions; /* analogous to traditional stat(2) :
                     bit0-bit2: others: xwr
                     bit3-bit5: group : xwr
                     bit6-bit8: owner : xwr
                     bit9     : sticky
                     bit10    : setgid
                     bit11    : setuid
                     bit31: permissions are valid
                   */ 
};


#define Fsdirinfo_initial_hashsizE 4
#define Fsdirinfo_avg_hash_lengtH 25
#define Fsdirinfo_hash_expansioN 2.0
#define Fsdirinfo_hash_maX 4096

struct FsdirinfO { /* registers specific properties of directory nodes */

 int flag;
    /*
       bit14= device is valid
    */
 dev_t device;
 int hashsize;
 struct FsnodE **sublist;
 off_t subcount;
 int permissions; /* meaning : see struct FsinfO 
                     this is a economy redundancy because dirs will rarely use
                     more than  .permissions  within struct FsinfO
                   */
};


struct FspartreC {
 int level;
 int volume;

 int partno;
 int partcount;
 
 off_t start_byte;
 off_t byte_count;
 
 int tempno;
 char *tempadr;
};


/*
#define Fstlink_with_partreC
*/

struct FstlinK { /* links a node in the target tree with a node in the
                    source tree */
 int status;  /* bit0=source side is outdated */
 struct FsnodE *target;
 char *target_path; /* needed as persistent storage location */
 struct FsnodE *source;
 struct FsnodE *source_dir;
 struct FspartreC *partrec;
 struct FstlinK *next_in_source;
 struct FstlinK *next_in_target;
};


struct FsextrA { /* consolidates any properties which are optional to a plain
                   file node (i.e. FsextrA is mandatory for directory nodes) */
 struct FsnodE *up;
 struct FstlinK *tlink;
 struct FsdirinfO *dirinfo;
 struct FsinfO *nodeinfo;
 struct FsnodE *outdated;

};


struct FsnodE { /* A file object in a filesystem tree model */

 char *name;
 int level;
 int volume;               /* -1 : file object deleted on the given level 
                               0 : only implicitely given */

#ifdef Askme_fsnode_with_timeS
 time_t mtime;
 time_t ctime;
#endif /* Askme_fsnode_with_timeS */

#ifdef Askme_fsnode_with_sizE
 off_t size;
#endif

 struct FsextrA *extra;
 struct FsnodE *next;

};


struct FprotectoR { /* A list element for individual overwrite protection */

 int valid; /*
   bit0= adr
   bit1= inode+device
   bit2= redirection
   bit3= shadow (application defined meaning:
                 for AskmE: a match shall not be considered as error)
   bit4= is a automatically created link protector, do not list
   bit5= is an emergency end of a link protector subchain 
         (inode obtained by stat rather than lstat)
 */
 char *adr;
 ino_t inode;
 dev_t device;
 char *redirection;
 char *last_ok;
 struct FprotectoR *prev,*next;

};

/*
 ts A71119
 Rev 2 emulates outdate back links by stacking one FstoutdatE record per
       outdate step.
*/
#define Fstoutdate_rev_2 1

#ifdef Fstoutdate_rev_2

struct FstoutdatE {/* stack element for open outdate chains during traversal */
 struct FsnodE *outdater; /* fresher node, target for back jump */
 struct FsnodE *bossdir;  /* The *dirnode of outdater */
 struct FsnodE *old;      /* the outdated node */
 struct FstoutdatE *next; 
};

#else /* Fstoutdate_rev_2 */

struct FstoutdatE {/* stack element for open outdate chains during traversal */
 struct FsnodE *root; /* outdate list start node */
 struct FsnodE *tip;  /* the outdate list end end */
 struct FstoutdatE *next; 
};

#endif /* ! Fstoutdate_rev_2 */


struct FstraversaL { /* keeps track of position in nearly-tree-graph */
 int cursor_state; /*
        bit0= currently in target tree
        bit1= node is outdated
        bit2= node is deleted
        bit5= target tree controlled run
        bit6= permission to follow outdate links
                    */
 struct FsnodE *node;
 struct FsnodE *dirnode;

 struct FstoutdatE *outdate_stack; /* the stack of open outdate lists */

 struct FstlinK *tlink;
 struct FsnodE *tlink_node;
 struct FsnodE *tlink_dirnode;

};
 

/* registers what options are already preset during Askme_prescan_args() */
struct Askme_prestatE {
 int language_set;
 int pkt_set;
 int picky_set;
 int logfile_set[4];
 int logfile_pkt_set;
 int media_mode_set;
 int info_source_enabled;
 int execute_first_argument;
 int is_resume_from_file;
};


/* bundles the local state of Askme_restore_copy() and its subordinates */
struct Askme_rescP {
 struct Askme_rescP *boss;
 struct AskmE *askme;
 struct FstraversaL *trav;
 int level;
 int volume;
 char *backup_adr;     /* external: must kept existing during whole lifetime */
 char *restore_adr;    /* external: must kept existing during whole lifetime */

 int simulate_action;
 int use_model_tree;
 int verbous_action;
 int verbous_file;
 int overwrite;                          /* see askme->restore_mode:bit11-14 */
 int move_to_asylum;
 int use_xorriso;                      /* 0= no , 1= per file , 2 = per tree */
 char *xorriso_dev;    /* external: must kept existing during whole lifetime */

 int is_moved_to_asylum;
 int is_second_try;

 struct FsnodE *node;
 struct FsnodE *dirnode;
 int cstate;
 int is_dir;
 int file_type;
 int times_valid;
 int owner_valid;
 int perms_valid;
 int perms;
 time_t atime;
 time_t mtime;
 uid_t owner_id;
 gid_t group_id;

 char source[4096];
 char target[4096];

 int old_type;
 int truncate_old;
 
 int is_surely_clean;
 int original_altered;  /* properties of original may be altered */
 int file_is_new;       /* eventually existing file object is not the original
                           but was created by restore */
 int dir_has_valid_file; /* a file was restored successfully by recursion */
 int error_is_handled;

 int *change_list;     /* bit0=finally restore perms , bit1=newly created */
 int *permission_list;
 int plcount;
 int path_is_prepared;

 int write_fd;
 double restored_bytes;

 struct Askme_rescP *prev,*next;
};


/* external commands to handle removable media */
#define Askme_media_load_cmD "eject -t"
#define Askme_media_mount_cmD "mount"
#define Askme_media_umount_cmD "umount"
#define Askme_media_eject_cmD "eject"

/* The base value for absolute -restore_mode */
#define Askme_restore_mode_defaulT (1<<7)
/* start up with restore disabled */
#define Askme_restore_mode_startuP ((1<<7)|(1<<30))

/* maximum number of history lines to be reported with -status:long_history */
#define Askme_status_history_maX 100

/* try to catch signals and ignore them during abort handling */
#define Askme_abort_handler_defaulT (1|(2<<4))


struct AskmE { /* the global context of askme */

 /* source */
 char progname[4096];
 char filename[4096];
 int all_with_fsinfo;
 char initial_wd[4096];
 char wd[4096];
 int did_something_useful;

 /* database */
 char backup_id[81];
 int backup_id_level;
 char recent_backup_date[81];
 int num_level_info;
 int count_level_info;
 char **level_info;
 int num_sublevels;
 struct AskmE **sublevels;
 int num_volumes;
 int num_highest_level_volumes;
 int *files_in_volume;
 int *count_in_volume;
 char **volume_info;
 char ***volumes;

 /* Fileobject tree model */
 char conf_dir[4096];
 struct FsnodE *root_node;
 struct FsnodE *target_root;
 int highest_level;
 int highest_volno;
 struct Bitmap2D *lev_vol_map;

 /* ASKME options */
 int sub;
 int suball;
 int levelall;
 int nolevel;
 int dialog;
 int fully_compatible;
 int search_mode;
 /* 0= start text
    1= fgrep , 
    2= regular expression
    3= (eventually structured) shell parser expression
    4= shell parser expression for leaf name
 */
 int structured_search;
 /* 0= flat text search
    1= '/' is a significant separator that cannot be matched by wildcards
    2= like 1 : but report only occurence in tree, no payload, no location
    3= like 2 : but report first content level of matching directories
    4= actually not structured but unique find mode (with search_mode 4)
 */
 int search_target;
 int result_format;
 /* bit0= print level header lines           "levhdr"
    bit1= print level number in result lines "levno"
    bit2= print volume number                "volno"
    bit3= print target address               "target"
    bit4= print source address               "source"
    bit5= print expression                   "expr"
    ifdef Askme_fsnode_with_timeS :
      bit6= print mtime                      "mtime"
      bit7= print ctime                      "ctime"

   bit19= apply mkisofs-escaping to target   "target_esc"
   bit20= mark directories by slash          -only settable from inside-
 */
 int use_stdin; /* use raw stdin even if redline support is compiled */
 int result_page_length;
 int result_page_width;
 int german_dialog;
 char mark_text[4096]; /* ( stdout+stderr, M: ) */
 int packet_output;
 int read_pickiness; /* behavior with problems while reading tree model:
        0=accept faulty data as good as possible
        1=accept irregularities as long as the outcome is the same as if
          all was neat (e.g. do not abort on attempt to delete a deleted node)
        2=default behavior: abort on irregularities but allow silent overwrite
        3=do not allow overwrite of existing data nodes (will fail with 
          scdbackup content lists because of per-volume ASKME copies)
 */
 int show_outdated;
 int volumewise; /* bit0= separate volume searches , bit1= level ascending */
 int no_dir_list;
 int status_history_max; /* for -status:long_history */
 int abort_handler; /* how to use the signal handler
   bit0-3: during normal operation
     0= no signal handling (this leads to exit on signals)
     1= try to handle eventual dirty restore states and exit
     2= try to ignore all signals
   bit4-7: during signal handling
     0= set signals to default handling (this leads to exit on signals)
     1= leave handling as is
     2= try to ignore all signals
   bit8= extra verbous abort
 */

 /* temporary search facilities */
#ifdef Askme_with_regeX
 regex_t *re;
 regmatch_t match[1]; 
#endif /* Askme_with_regeX */
 char **re_constants;
 int re_count;
 int re_fill;
 char reg_expr[2*4096];
 struct Bitmap2D *search_lv_map;

 /* resume facility */
 int resume_at_level;
 int resume_at_volume;
 char resume_at_source[4096];
 int resume_last_level;
 int resume_last_volume;
 char resume_last_source[4096];
 char resume_last_dirty[4096];
 char resume_search_text[2*4096];
 int resume_was_restore;
 char resume_state_file[4096];

 /* media handling */
 int media_mode; /*
    bit0= allow bit2 and bit4 in abort handling
    bit2= try to mount, umount
    bit4= try to eject after umount
 */
 char media_load_cmd[4096];
 char media_mount_cmd[4096];
 char media_umount_cmd[4096];
 char media_eject_cmd[4096];

 /* restore facilities */
 int restore_mode; /*
    bit0= restore mode in effect for search
    bit1= use target address to determine restore target

    bit3= set restore_mode:bit0 on ordinary searches

    bit5= simulation mode: do not insist on existing source, 
    bit6= simulation mode: do not create or change files
    bit7+8= verbosity:  0=silent  1=subfile  2=volume  3=action
    bit9= simulation mode: do not load, mount, umount, eject
    bit10= give owner rwx-permission to restored directories in any case
    bit11-14= overwrite mode: 
           0= overwrite_off
           1= overwrite_by_unlink
           2= overwrite_by_truncation
           3= overwrite_by_move
    bit15= do not copy ownership (which applies to superuser only)
    bit16-18= behavior when errors are encountered:
           0= end operation on first error
           1= continue operation after errors
           2= exit program after first error
           3= continue operation after errors but exit when done
           4= prompt user [continue, retry, skip volume, abort]
    bit19= do not copy atime and mtime from backup file object
    bit20= record -resume_last_source only before dialog
           (else with each new source) 
    bit21= use truncate and fsync to write -resume_last_source"_dirty" file 
    bit22= do not remove incompletely restored file objects
    bit23= do not lift existing access permissions
    bit30= do not allow restore mode (revokable)
    bit31= do not allow restore mode (may not be revoked !)
 */
 char restore_from[4096];
 int restore_from_compcount; 
 char restore_to[4096];
 char restore_skip[4096];
 char restore_asylum[4096];
 char restore_pre_prompt[4096];
 char restore_post_prompt[4096];
 struct FprotectoR *protections;
 int default_dir_perms;

 /* future interface to xorriso for file extraction */
 int use_xorriso;
 char xorriso_dev[4096];
 pid_t xorriso_pid;
 int xorriso_stdin;
 int xorriso_stdout;
 int xorriso_stderr;

 /* run state */
 struct Askme_prestatE *prescan_state;
 int is_dialog;
 int bar_is_fresh;
 char pending_option[4096]; /* eventual option entered at page prompt */
 int request_to_abort;
 int request_to_skip_volume;
 int suppress_result;
 int look_for_dir; /* 1=searching a directory,abort on first match,no print */ 
 double idle_time;
 int re_failed_at; /* mismatch position with structured_search */
 int prepended_wd;
 int mounted_level;
 int mounted_volume;
 double restore_count;
 double restore_bytes;
 double error_count; /* double will not roll over */
 struct Askme_rescP *restore_ctl_stack;

 /* result (stdout, R: ) */
 int found;
 int search_failed;
 char result_line[5*4096];
 int result_line_counter;
 int result_page_counter;
 int result_open_line_len;
 struct FstraversaL *found_trav;

 /* info (stderr, I:) */
 char info_text[10*4096];

};


/* some function which are used before they are defined: makes -Wall happier */

int Sfile_destroy_argv();

int Dirseq_new();
int Dirseq_destroy();
int Dirseq_next_adr();
int Dirseq_next_adrblock();

int Lreader_destroy();
char *Lreader_strerror();

int Fsrptmap_destroy();

int Fsdirinfo_destroy();

int Fsinfo_destroy();

int Fstlink_destroy();

int Fsextra_new();
int Fsextra_destroy();
int Fsextra_set_target();

int Fsnode_destroy();
int Fsnode_graft_branch();
int Fsnode_get_adr();

int Bitmap2d_new();
int Bitmap2d_destroy();

int Fprotector_destroy();
int Fprotector_determine_inode();

int Askme_destroy();
int Askme_prescan_args();
int Askme_interpreter();
int Askme_execute_option();
int Askme_info();
int Askme_umount_eject();
int Askme_load_mount();
int Askme_identify_volume();
int Askme_status();
int Askme_prepare_regex();
int Askme_make_restore_adr();
int Askme_rm_dirty_dir();
int Askme_prepare_path();
int Askme_asylum_adr();
int Askme_announce_copy();
int Askme_restore_copy();
int Askme_error_on_restore();
int Askme_record_resume_state();
int Askme_record_resume_source();
int Askme_record_resume_dirty();
int Askme_handle_picky0();
int Askme_get_level_info();


#endif /* Askme_includeD */


/*
                          The File System Model Graph

The file system model generated from content-vanish lists is actually not
a tree but a two-rooted bush graph. It simultaneously represents the content
and the level history of the backup area (i.e. the "source tree") as well
as the content of the backup volumes of each level ("target tree").

A file object node is represented by FsnodE which is very sparse in order to
keep the model lean. It merely contains the node address (level, volume, 
leafname), a pointer to an eventual node extension container FsextrA, and a 
pointer to the next node in the node's hosting directory.

Any further property of file objects has to be encoded in an extension
object and linked to the node via the node's FsextrA object. Obviously
an FsextrA is needed as soon as an extension object is wanted.
Extension objects may be:
  FsinfO     for file system attributes of the object
  FsdirinfO  for directory nodes
  FstlinK    connects target tree with source tree
Other FsextrA information constitutes optional graph linkage to FsnodEs
  an upward link to node of hosting directory
  a list of outdated nodes

FsdirinfO maintains a hashed list which dynamically changes its hash width
as it grows. Finally the hashed list may be transformed into a sorted list
(i.e. hash width 1).

A node in a target tree can have several FstlinKs while a node in the source
tree at most has one of them. Each FstlinK is owned by one source and one
target node. If either of them vanishes, the FstlinK vanishes too.

Outdated nodes only occur in the source tree. They are file objects which
have been deleted between two backup levels or non-directories which have
been updated by a younger backup level. (Directories do not get outdated
by re-appearence on a younger level.)
The target tree consists entirely of directory nodes so it cannot produce
outdated nodes but rather accumulates FstlinKs. (Even the target node of
a regular file is technically a directory. This might be ingenous or stupid.)

This quite complicated bush model can be traversed by help of a state
object FstraversaL and its method Fstraversal_next().

The model is built in Askme_load_old_tree() by one or more calls to
Fsnode_read_tree(). This function opens the list file and reads it by help of
a LreadeR object. The returned file objects are brought into the tree by
function Fsnode_graft_branch().

*/
