/*
  cc -g -D_FILE_OFFSET_BITS=64 -D_LARGEFILE_SOURCE -DCdbackup_planer_developpeR -o ../bin/cd_backup_planer ../src/cd_backup_planer.c
*/

#define PROG_VERSION "0.9.2"
static int verbous=1;

/* Disable old file checker code and leave only new Crawler active */
#define Crawler_ruleS yes

/* Enable pseudo split_dir -xorriso_cut_out */
#define Cdbackup_planer_xorriso_cut_ouT yes


#include <ctype.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include <sys/wait.h>
#include <errno.h>
#include <time.h>
#include <utime.h>
#include <sys/time.h>
#include <pwd.h>
#include <grp.h>
#include <dirent.h>

#ifdef Cd_backup_planer_with_regeX
#include <regex.h>
#endif /* Cd_backup_planer_with_regeX */


/* Types which may not become 64 bit */
/* depending on compiler and operating system this might have to be adjusted */
typedef short Stypes_int16;
typedef unsigned short Stypes_uint16;
typedef int Stypes_int32;
typedef unsigned int Stypes_uint32;


/* Macro for creation of arrays of objects (or single objects) */
#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));


/* set by -agetest_debug */
static int Cdbackup_planer_agetest_debuG= 0;

#define Return_with_debug_msG(x,y) { \
if(Cdbackup_planer_agetest_debuG) \
  fprintf (stderr,"agetest_debug: File_is_changed = %d , because %s\n", x, y);\
return(x);\
} 


/* set by -test_command list_search_debug: 1=debug, 2=show single tests */ 
static int Cdbackup_planer_listsearch_debuG= 0;

/* set by -test_command tree_build_debug : 1=debug, 2=use listsearch_debuG */
static int Cdbackup_planer_treebuild_debuG= 0;


/* if 1 : do not abort on bad ".." in Fileliste_read_serialized_tree() */
/* (set by -test_command tolerate_bad_dirpop) */
static int Cdbackup_planer_tolerate_bad_dirpoP= 0;


/* 0520 */
/* if 1 : remove directories from model tree although they are not empty */
static int Cdbackup_planer_tolerate_bad_rmdiR= 1;


/* trust bit6 of return value of Fileliste_precheck_sub() which will spare
   a filelist search in a potentially fat intermediate result list.
   bit6 never failed but as soon as i trust it i will not learn about failures
   any more. So this will stay 0 for a while */
static int Cd_backup_planer_trust_fret64= 0; 
static int Cd_backup_planer_false_young= 0;
static int Cd_backup_planer_false_zu_alt= 0;

/* a questionable trick to spare a filelist search when reading the -level 0
   content file list. It risks to produce double entries in old_tree and thus
   needs double search on delete. This will probably be disabled in stable
   versions but by now i want to know wether it is suitable. */
/* A50929 : it produced double entries . Disabled now. */
static int Cd_backup_planer_risk_d_enabled= 0;
static int Cd_backup_planer_double_deleted= 0;
static int Cd_backup_planer_not_empty_on_delete= 0;


/* re-use composition size information when splitting */
/* to be set by : -composition_memory \
                     Cdbp_compose_keep_sublevels \
                     Cdbp_compose_sublevel_maxsize \
                     Cdbp_compose_sublevel_maxmem \
*/
static int Cdbp_compose_keep_sublevels= 3;
static int Cdbp_compose_sublevel_count= 0;
static int Cdbp_compose_sublevel_maxsize= 1000;
static double Cdbp_compose_sublevel_maxmem= 4*1024*1024;
static int Cdbp_compose_do_not_reuse= 0;

/* this is used to watch composition memory consumption */
static int Cdbp_count_subs= 0;
static double Cdbp_subs= 0;
static double Cdbp_subs_lines= 0;
static double Cdbp_subs_mem= 0.0;

/* a constant dev_t to be used as placeholder */
static dev_t dev_dummy= 0;


/* global pacifier display time stamp */
time_t Cd_backup_planer_pacifier_timestamP= 0;
int Cd_backup_planer_pacifier_intervaL= 1;
/* buffered pacifier parameters for a wake-up call to Print_pacifier() */
int Cd_backup_planer_pacifier_wannabe_counT= 0;
char Cd_backup_planer_pacifier_wannabe_texT[4096+160]= {""};
int Cd_backup_planer_pacifier_wannabe_flaG= 0; /* like flag in Print_pacifier()
                                                  but: bit7=wannabe waiting */
int Cd_backup_planer_pacifier_show_timE= 0; /* show timestamp with pacifier */
int Cd_backup_planer_pacifier_new_linE= 0;  /* show full lines + newline */
int Cd_backup_planer_start_timE;            /* to be set at begin of main() */

/* for entertaining or confusing messages during compose run */
static double Cd_backup_planer_compose_size= 0.0;

/* with entertaining messages: display composition_memory bufferstate */
static int Cd_backup_planer_compose_pacifier= 0;


/* global shortcut to avoid frequent call of getcwd() */
char Cd_backup_planer_cwD[4096]= {""};


/* global setting wether to follow symbolic links */
int Cd_backup_planer_follow_linkS= 0;


/* buffer for header lines of a loaded list */
char Cd_backup_planer_list_heaD[3*4096];


/* Experimental : check -not files by device and inode number */
/* Will be enabled by first occurence of option -not_inode */
int Cdbackup_planer_sperrliste_inodE= 0;


double Cdbackup_planer_l0_basE= 0.0;
double Cdbackup_planer_l0_treE= 0.0;
double Cdbackup_planer_l0_changeD= 0.0;
double Cdbackup_planer_l0_deleteD= 0.0;
double Cdbackup_planer_l0_addeD= 0.0;


/* The size of the extra long integer vectors in 30 bit steps */
/* CdbplivS = Cd_backup_planer_ivec_sizE */
#define CdbplivS 3


/* The maximum number of valid checksum tags to be recorded during scan.
   This size is used up twice : once for the whole scan and once for
   the scan interval up to the next listed checksum record's size.
*/
static int Cd_backup_planer_tag_list_maX= 1000;


/* current id headers of several file formats appended to backup volumes */
#define Cdbpl_permutation_starT "scdbackup block permutation header v0.1\n"
#define Cdbpl_multiplication_marK "scdbackup volume multiplication mark v0.1 "
#define Checkrecord_blocklist_starT "scdbackup checksum block list v0.1\n"
/* WARNING : First two letters may appear in this text only once at the start */
#define Checkrecord_tag_starT "scdbackup_checksum_tag_v0.1 "


/* The maximum depth for function recursions which are caused by
   the examined directory trees.
*/
int Cd_backup_planer_tree_max_deptH= 128;


/* usually Large File Support is gained by 
     cc ... -D_FILE_OFFSET_BITS=64 -D_LARGEFILE_SOURCE ...
*/

#ifdef _LARGEFILE_SOURCE
#define Cd_backup_planer_fseekO fseeko
#endif /* _LARGEFILE_SOURCE */


/* better leave Cd_backup_planer_use_stat64 and Cd_backup_planer_define_stat64
   undefined unless you do not see any other way.
*/
#ifdef Cd_backup_planer_use_stat64

#define Cd_backup_planer_struct_staT stat64
#define Cd_backup_planer_call_staT stat64
#define Cd_backup_planer_call_lstaT lstat64

#else /* Cd_backup_planer_use_stat64 */

#define Cd_backup_planer_struct_staT stat
#define Cd_backup_planer_call_staT stat
#define Cd_backup_planer_call_lstaT lstat

#endif /* ! Cd_backup_planer_use_stat64 */


#ifdef Cd_backup_planer_define_stat64

/* if you got the functions stat64(), lstat64() but no header then try this */

struct stat64 {
        unsigned short  st_dev;
        unsigned char   __pad0[10];

        unsigned long   __st_ino;

        unsigned int    st_mode;
        unsigned int    st_nlink;

        unsigned long   st_uid;
        unsigned long   st_gid;

        unsigned short  st_rdev;
        unsigned char   __pad3[10];

        long long       st_size;
        unsigned long   st_blksize;

        unsigned long   st_blocks;      /* Number 512-byte blocks allocated. */
         unsigned long   __pad4;         /* future possible st_blocks high bits */
        unsigned long   st_atime;
        unsigned long   __pad5;

        unsigned long   st_mtime;
        unsigned long   __pad6;

        unsigned long   st_ctime;
        unsigned long   __pad7;         /* will be high 32 bits of ctime someday */

        unsigned long long      st_ino;
};

#else /* Cd_backup_planer_define_stat64 */

#ifdef Cd_backup_planer_use_stat64
/* >>> possible need to include the header with stat64.h */
#endif /* Cd_backup_planer_use_stat64 */

#endif /* ! Cd_backup_planer_define_stat64 */



/* --------------------------- nonportable --------------------------*/

/* Wrappers for code not usable on all target systems */



int Sfile_is_local_fs(name,flag)
char *name;
int flag;
{
 /* A00907_0060_dave_erikson , code and ideas contributed by Dave Erikson */
 /* Do not place more executable statements outside the #ifdef __xxx */
 int ret= 1;

#ifdef __sgi
 struct stat stbuf;
 if(Cd_backup_planer_call_lstaT(name,&stbuf)==-1) {
   ret= 0;
 } else {
   if(strncasecmp ("nfs", stbuf.st_fstype, strlen ("nfs")) == 0)
     ret= 0;
 }

#endif /* __sgi */


#ifdef __linux
#ifdef Cd_backup_planer_localdev_linuX

/* #include <sys/statfs.h> */ /* as alternative to <sys/vfs.h> */
#include <sys/vfs.h>

/* #include <linux/nfs_fs.h> */ /* this is a complete mess in SuSE 6.4 */
/* so i have to define it by hand */
#define NFS_SUPER_MAGIC       0x6969
#define SMB_SUPER_MAGIC       0x517B

 struct statfs fstbuf;

 if(statfs (name,&fstbuf)==-1) {
   ret= 0;
 } else {
   switch (fstbuf.f_type){
   case NFS_SUPER_MAGIC: case SMB_SUPER_MAGIC:
     ret= 0;
     break;
   default:
     break;
   }
 }

#endif /* Cd_backup_planer_localdev_linuX */
#endif /* __linux */


 return(ret);
}



/* ----------------- Misc --------------- */


int Lstat(path,stbuf,flag)
char *path;
struct stat *stbuf;
int flag;
/*
 bit0= print error message if file object is missing
*/
{
 int ret;
 char *callname;

 if(Cd_backup_planer_follow_linkS) {
   callname= "stat";
   ret= Cd_backup_planer_call_staT(path,stbuf);
 } else {
   callname= "lstat";
   ret= Cd_backup_planer_call_lstaT(path,stbuf);
 }
 if(ret==-1 && (flag&1))
   fprintf(stderr,"%s(%s) failed. %s\n",callname,path,strerror(errno));
 return(ret);
}


char *Text_shellsafe(in_text,out_text,flag)
char *in_text,*out_text;
int flag;
{
 int l,i,w=0;

 /* enclose everything by hard quotes */
 l= strlen(in_text);
 out_text[w++]= '\'';
 for(i=0;i<l;i++){
   if(in_text[i]=='\''){
     /* escape hard quote within the text */
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
     out_text[w++]= '"';
     out_text[w++]= '\'';
   } else {
     out_text[w++]= in_text[i];
   }
 }
 out_text[w++]= '\'';
 out_text[w++]= 0;
 return(out_text);
}


int Decode_date_input_format(erg,text,flag)
/* MMDDhhmm[[CC]YY][.ss]] */
struct tm *erg;
char *text;
int flag;
{
 int i,l,year;
 time_t current_time;
 struct tm *now;

 current_time= time(0);
 now= localtime(&current_time);
 for(i=0;i<sizeof(struct tm);i++)
   ((char *) erg)[i]= ((char *) now)[i];

 l= strlen(text);
 for(i=0;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=8 && i!=10 && i!=12)
   return(0);
 if(text[i]==0)
   goto decode;
 if(text[i]!='.' || l!=15)
   return(0);
 i++;
 if(text[i]<'0'||text[i]>'9')
   return(0);  
 i++;
 if(text[i]<'0'||text[i]>'9')
   return(0);

decode:; 
 /* MMDDhhmm[[CC]YY][.ss]] */
 i= 0;
 erg->tm_mon=  10*(text[0]-'0')+text[1]-'0'-1;
 erg->tm_mday= 10*(text[2]-'0')+text[3]-'0';
 erg->tm_hour= 10*(text[4]-'0')+text[5]-'0';
 erg->tm_min=  10*(text[6]-'0')+text[7]-'0';
 erg->tm_sec= 0;
 if(l==8)
   return(1);
 if(l>10){
   year= 1000*(text[8]-'0')+100*(text[9]-'0')+10*(text[10]-'0')+(text[11]-'0');
 }else{
   year= 1900+10*(text[8]-'0')+(text[9]-'0');
   if(year<1970)
     year+= 100;
 }
 erg->tm_year= year-1900;
 if(l<=12)
   return(1);
 erg->tm_sec=  10*(text[13]-'0')+text[14]-'0';
 return(1);
}


int String_to_time(code,test_ok,date,flag)
char *code;
int *test_ok; /* out: bit0= usable for age test (else disable age_test) */ 
time_t *date;
int flag;
{
 char *cpt,scale_chr;
 double value,seconds;
 struct tm result_tm;

 (*test_ok)&= ~1;
 *date= 0;
 cpt= code;
 if(code[0]=='-' || code[0]=='+'){
   if(code[1]==0)
     return(1);
   if(code[1]>='0' && code[1]<='9') {
     seconds= time(0);
     value= -1;
     sscanf(code,"%lf",&value);
     scale_chr= code[strlen(code)-1];
     if(isalpha(scale_chr))
       scale_chr= tolower(scale_chr);
     if     (scale_chr=='s') seconds+= 1.0*value;
     else if(scale_chr=='h') seconds+= 3600.0*value;
     else if(scale_chr=='d') seconds+= 86400.0*value;
     else if(scale_chr=='w') seconds+= 86400.0*7.0*value;
     else if(scale_chr=='m') seconds+= 86400.0*31.0*value;
     else if(scale_chr=='y') seconds+= 86400.0*(365.25*value+1.0);
     else                    seconds+= 1.0*value;
     goto completed;
   }
   cpt= code+1;
 } else if(Sfile_decode_datestr(&result_tm,code,0)) {
   /* YYMMDD[.hhmm[ss]] */
   result_tm.tm_isdst= -1;
   seconds= mktime(&result_tm);
   goto completed;
 } else if(Decode_date_input_format(&result_tm,code,0)) {
   /* MMDDhhmm[[CC]YY][.ss]] */
   result_tm.tm_isdst= -1;
   seconds= mktime(&result_tm);
   goto completed;
 } else if(0) { /* ??? Mth dd hh:mm:ss Tzn YYYY */
   /* >>> */;
 }
 if(strcmp(cpt,"time")==0 || strcmp(cpt,"bigbang")==0 || strcmp(cpt,"off")==0)
   return(1);
 return(0);
completed:;
 if(seconds<0 || seconds>=0x7FFFFFFF)
   return(0);
 *date= seconds;
 (*test_ok)|= 1;
 return(1);
}


/* >>> anglification mark : hic sunt germani <<< */


int Open_listfile(fppt,adr,mode,flag)
FILE **fppt;
char *adr;
char *mode;
int flag;
/*
 bit0= pipe through gzip/gunzip if adr indicates so (*.gz and *.GZ) 
 bit1= mode w : unlink but do not create new file nor return file pointer
return:
 <=0 = error
 1 = Sfile_fopen() successful
 2 = popen() successful
*/
{
 int ret= 0,adr_has_dotgz= 0,l;
 char cmd[4096+80],shell_adr[3*4096],dotgz[4];
 FILE *Sfile_fopen();
 struct stat stbuf;

 /* Remove dead children if needed */
 wait3(NULL,WNOHANG,NULL);

 *fppt= NULL;
 l= strlen(adr);
 if(l>3)
   if(strcmp(adr+(l-3),".gz")==0 || strcmp(adr+(l-3),".GZ")==0)
     adr_has_dotgz= 1;
 if(strcmp(mode,"w")==0) {
   unlink(adr);
   if(Cd_backup_planer_call_staT(adr,&stbuf)!=-1) {
     fprintf(stderr,"--- Open_listfile( %s ,w) : obstinate file\n",adr);
     ret= 0; goto ex;
   }
   if(flag&2)
     return(1);
   if((flag&1) && adr_has_dotgz) {
     sprintf(cmd,"gzip >%s",Text_shellsafe(adr,shell_adr,0));
     *fppt= popen(cmd,"w");
     if(*fppt==NULL) {
       fprintf(stderr,"--- popen(\"%s\",\"w\") failed\n",cmd);
       fprintf(stderr,"--- errno= %d , '%s'\n",errno,strerror(errno));
     }
     ret= 2;
   } else {
     *fppt= Sfile_fopen(adr,"wb",0);
     ret= 1;
   }
 } else if(strcmp(mode,"r")==0) {
   dotgz[0]= 0;
   if((flag&1) && adr_has_dotgz)
     goto try_popen_gunzip;
   *fppt= Sfile_fopen(adr,"rb",1);
   if(*fppt==NULL) {
     /* temporarily store the error message of fopen in case there is no .gz */
     sprintf(shell_adr,"--- errno= %d , '%s'\n",errno,strerror(errno));
     /* try to find old_complete_list_adr.gz */
     strcpy(dotgz,".gz");
try_popen_gunzip:;
     sprintf(cmd,"%s%s",adr,dotgz);
     *fppt=  Sfile_fopen(cmd,"rb",1);
     if(*fppt!=NULL) {
       Sfile_fclose(*fppt);
       sprintf(cmd,"gunzip <%s%s",Text_shellsafe(adr,shell_adr,0),dotgz);
       *fppt=  popen(cmd,"r");
       if(*fppt==NULL) {
         fprintf(stderr,"--- popen(\"%s\",\"r\") failed\n",cmd);
         fprintf(stderr,"--- errno= %d , '%s'\n",errno,strerror(errno));
       }
       ret= 2;
     } else {
       fprintf(stderr,"failed to open %s in read mode\n",adr);
       fprintf(stderr,"%s",shell_adr);  /* the stored error message of fopen */
       ret= 0;
     }
   } else 
     ret= 1;
 } else {
   fprintf(stderr,"--- Open_listfile : mode %s not supported\n",mode);
 }
 if(*fppt==NULL)
   ret= 0;
ex:;
 return(ret);
}


int Is_forbidden_mountpoint(name,parent_dev,return_dev,flag)
char *name;
dev_t parent_dev;
dev_t *return_dev;
int flag;
/*
 bit0-3= mode
    0= anything is acceptable
    1= reject any mount point
    2= reject net mount points
    3= mode 1 (for compatibility reasons)
*/
{
 int mode;
 dev_t name_dev;
 struct stat stbuf;

 /* check for mountpoints */
 /* A00907_0060_dave_erikson , code and ideas contributed by Dave Erikson */
 mode= flag&15;
 if(mode==0)
   return(0);
 if(Cd_backup_planer_call_lstaT(name,&stbuf)==-1)
   return(-1);
 name_dev= stbuf.st_dev;
 if(return_dev!=NULL)
   *return_dev= name_dev;
 if(name_dev!=parent_dev) {
   if(mode==1 || mode==3) {
     return(1);
   } else if(mode==2) {
     if(!Sfile_is_local_fs(name,0))
       return(1);
   }
 }
 return(0);
}


int Split_fileaddress(line,target,source,flag)
char line[4096],target[4096],source[4096];
int flag;
{
 char *cpt;

 cpt= strchr(line,'=');
 if(cpt!=NULL) {
   strncpy(target,line,cpt-line);
   target[cpt-line]= 0;
   strcpy(source,cpt+1);
 } else {
   target[0]= 0;
   strcpy(source,line);
   if(strcmp(line,"-")==0 || strcmp(line,".")==0)
     source[0]= 0;
 }
 return(1);
}


int Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                        device,inode,mtime,ctime,
                        valid_flag,flag)
char *typetext,*lengthtext,*crctext;
int typ;
double size;
unsigned long crc;
unsigned char md5[16];
dev_t device;
ino_t inode;
time_t mtime;
time_t ctime;
int valid_flag; /*
  bit1=typ
  bit2=size
  bit3=crc
  bit4=md5 rather than crc
  bit5=inode
  bit6=device
  bit7=mtime
  bit8=ctime
*/
int flag;
/*
 bit0= do not append blank to typetext
 bit9= prepend type
bit10= prepend length
bit11= preprend crc
bit13= prepend inode number and device (as extension to type)
bit16=prepend mtime
bit17=prepend ctime

*/
{
 int i;

 typetext[0]= lengthtext[0]= crctext[0]= 0;
 if(flag&512) {
   if(valid_flag&2)
     strcpy(typetext,typ==1?"-":typ==2?"d":typ==3?"l":
                     typ==4?"p":typ==5?"m":typ==6?"b":"0");
   else
     strcpy(typetext,"0");
 }
 if(flag&1024) {
   if(valid_flag&4)
     sprintf(lengthtext,"%.lf ",size);
   else
     strcpy(lengthtext,"-1 ");
 }
 if(flag&2048) {
   if(valid_flag&8) {
     if(valid_flag&16) {
       for(i=0;i<16;i++) 
         sprintf(crctext+2*i,"%2.2x ",md5[i]);
     } else
       sprintf(crctext,"%lX ",crc);
   } else
     strcpy(crctext,"-1 ");
 }

 /* Note to myself: Reserved type extensions:
     \(obsoleted separator count) , i(Inode) , d(Device) , 
     t(mtime in seconds since 1970) , c(ctime in seconds since 1970) ,
     p(split file part $start:$size) , P(split file original $start:$count) ,
 */
 if(flag&8192) {
   if(valid_flag&32) {
     int ino_vec[CdbplivS],dev_vec[CdbplivS];
     char numtext[81];

     Intvec_from_dev_ino(dev_vec,CdbplivS,ino_vec,device,inode,0);
     if(valid_flag&64) {
       Intvec_to_dec(dev_vec,CdbplivS,numtext,sizeof(numtext),0);
       sprintf(typetext+strlen(typetext),",d%s",numtext);
     }
     Intvec_to_dec(ino_vec,CdbplivS,numtext,sizeof(numtext),0);
     sprintf(typetext+strlen(typetext),",i%s",numtext);
   }
 }
 if(flag&(1<<16))
   if(valid_flag&128)
     sprintf(typetext+strlen(typetext),",t%.lf",(double) mtime);
 if(flag&(1<<17))
   if(valid_flag&256)
     sprintf(typetext+strlen(typetext),",c%.lf",(double) ctime);

 if(flag&512) 
   if(!(flag&1))
     strcat(typetext," ");
 
 return(1);
}


int Get_user_group_name(uid,gid,username,groupname,flag)
uid_t uid;
uid_t gid;
char username[256];
char groupname[256];
int flag;
{
 struct passwd *pwd;
 struct group *grp;

 strcpy(username,"?");
 strcpy(groupname,"?");
 pwd= getpwuid(uid);
 if(pwd==NULL)
   return(0);
 if(strlen(pwd->pw_name)>255)
   return(0);
 strcpy(username,pwd->pw_name);
 grp= getgrgid(gid);
 if(grp==NULL)
   return(0);
 if(strlen(grp->gr_name)>255)
   return(0);
 strcpy(groupname,grp->gr_name);
 return(1);
}


int Get_user_group_id(username,groupname,uid,gid,flag)
char username[256];
char groupname[256];
uid_t *uid;
uid_t *gid;
int flag;
{
 struct passwd *pwd;
 struct group *grp;

 pwd= getpwnam(username);
 if(pwd==NULL)
   return(0);
 *uid= pwd->pw_uid;
 grp= getgrnam(groupname);
 if(grp==NULL)
   return(0);
 *gid= grp->gr_gid;
 return(1);
}


int Record_dir_properties(fp,fp_aux,name,adr,adr_aux,lengthtext,crctext,flag)
FILE *fp,*fp_aux;
char *name;                       /* address where to find file now */
char *adr,*adr_aux;               /* address text to be recorded */
char *lengthtext,*crctext;        /* to be used w/o bit0. Trailing blanks */
int flag;
/*
 bit0= report as shell commands which set the properties
       else as formated line with additionaly length and checksum
 bit1= report both forms : adr as formated line on fp, s
                           adr_aux as shell commands on fp_aux
*/
{
 char shell_adr[3*4096],username[256],groupname[256],*npt,*scr_adr;
 struct stat stbuf;
 int octomode= 0,i;
 char *Sfile_datestr();
 FILE *scr_fp;
 
 if(Lstat(name,&stbuf,0)==-1)
   return(0);
 /*  seems to be covered by  man 2 stat  and  man 1 chmod */
 octomode= stbuf.st_mode&0xfff;
 Get_user_group_name(stbuf.st_uid,stbuf.st_gid,username,groupname,0);
 if(flag&(1|2)) {
   /* report as some shell commands */;
   if(flag&2) {
     scr_fp= fp_aux;
     scr_adr= adr_aux;
   } else {
     scr_fp= fp;
     scr_adr= adr;
   }
   for(npt= scr_adr;*npt=='/';npt++);
   if(*npt==0)
     fprintf(scr_fp,"x='.'\n");
   else
     fprintf(scr_fp,"x=%s\n",Text_shellsafe(npt,shell_adr,0));
   fprintf(scr_fp,"chmod %4.4o \"$x\"\n",octomode);
   fprintf(scr_fp,"chown %s \"$x\"\n",Text_shellsafe(username,shell_adr,0));
   fprintf(scr_fp,"chgrp %s \"$x\"\n",Text_shellsafe(groupname,shell_adr,0));
   fprintf(scr_fp,"\n");
 }
 if((!(flag&1)) || (flag&2)) {
   /* report as : */
   /* D permissionbits owner group time lengthtext crctext adr */
   for(i=strlen(username);i>=0;i--)
     if(isspace(username[i]))
       username[i]= '!';
   for(i=strlen(groupname);i>=0;i--)
     if(isspace(groupname[i]))
       groupname[i]= '!';
   /* capital D indicates presence of chmod,chown,chgrp,date */
   fprintf(fp,"D %4.4o %s %s %s %s%s%s\n",
              octomode,username,groupname,Sfile_datestr(stbuf.st_mtime,1|2),
              lengthtext,crctext,adr);
 }
 return(1);
}
 

int Type_and_size(filename,size,flag)
char *filename;
double *size;
int flag;
/*
 bit0= return -1 if file is missing 
 bit1= return a hardlink with siblings as type 5
 bit2= eventually return properties of link target
 bit3= determine size of block device by reading its whole content
 bit4= return unknown type as 127 rather than 0
*/
/*
 return:  0=unknown  1=regular  2=directory  3=symbolic link  4=named pipe
          5=multiple hardlink (with bit1)  6=block device
*/
{
 int ret;
 struct Cd_backup_planer_struct_staT stbuf;
 double Count_file_bytes();

 if(flag&4) {
   if(Cd_backup_planer_call_staT(filename,&stbuf)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 } else {
   if(Lstat(filename,&stbuf,0)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 }
 if(size!=NULL)
   *size= stbuf.st_size;
 ret= Stbuf_type(&stbuf,0);
 if(ret==6) {
   if(flag&8) {
     *size= Count_file_bytes(filename,0);
     if(*size<0) {
       if(flag&1) return(-1);
       else       return(0);
     }
   }
 }
 return(ret);
}


int Stbuf_type(stbuf,flag)
struct stat *stbuf;
int flag;
{
 if(S_ISREG(stbuf->st_mode)) {
   if(flag&2)
     if(stbuf->st_nlink>1)
       return(5);
   return(1);
 }
 if(S_ISDIR(stbuf->st_mode))
   return(2);
 if(S_ISLNK(stbuf->st_mode))
   return(3);
 if(S_ISFIFO(stbuf->st_mode))
   return(4);
 if(S_ISBLK(stbuf->st_mode)) {
   return(6);
 }
 if(flag&16)
   return(127);
 return(0);
}


int File_inode(adr,device,inode,flag)
char *adr;
dev_t *device;
ino_t *inode;
int flag;
{
 struct stat stbuf;

 if(Lstat(adr,&stbuf,0)==-1)
   return(0);
 (*device)= stbuf.st_dev;
 (*inode)= stbuf.st_ino;
 return(1);
}


int Is_older_file(candidate,reference,flag)
char *candidate;
char *reference;
int flag;
{
 struct stat stbuf_c,stbuf_r;
 time_t age_c,age_r;

 if(Cd_backup_planer_call_staT(candidate,&stbuf_c)==-1)
   return(-1);
 if(Cd_backup_planer_call_staT(reference,&stbuf_r)==-1)
   return(-1);
 if(stbuf_c.st_mtime>stbuf_c.st_ctime)
   age_c= stbuf_c.st_mtime;
 else
   age_c= stbuf_c.st_ctime;
 if(stbuf_r.st_mtime>stbuf_c.st_ctime)
   age_r= stbuf_c.st_mtime;
 else
   age_r= stbuf_c.st_ctime;
 if(age_r>age_c)
   return(1);
 return(0);
}


int Parent_is_trustfully_old(adr,age_limit,flag)
char *adr;
time_t age_limit;
int flag;
/*
 bit0= is recursion
 bit1= do only check direct parent and not all directories in path
 bit14= ignore ctime
*/
{
 int ret;
 char pd[4096],*spt;
 struct stat stbuf;

 if(strcmp(adr,"/")==0)
   return(1); /* virtual parent of rootdir is considered old */
 else if(strcmp(adr,".")==0) {
   if(getcwd(pd,sizeof(pd))!=NULL)
     if(strcmp(pd,"/")==0)
       return(1); /* another name of rootdir */
 } else
   strcpy(pd,adr);
 spt= strrchr(pd,'/');
 if(spt==NULL)
   strcpy(pd,"..");
 else if(spt==pd) {
   if(pd[1]==0)
     return(1);
   else
     strcpy(pd,"/");
 } else
   (*spt)= 0;
 if(Cd_backup_planer_call_staT(pd,&stbuf)==-1)
   return(0);
 if(!(flag&(1<<14)))
   if(stbuf.st_ctime>=age_limit)
     return(0);
 if(stbuf.st_mtime>=age_limit)
   return(0);
 if((flag&2) || strcmp(pd,"/")!=0 || strcmp(pd,"..")!=0)
   ret= 1;  
 else
   ret= Parent_is_trustfully_old(pd,age_limit,flag|1);
 return(ret);
}


int Dir_is_empty(dir,flag)
char *dir;
int flag;
/*
 return:
  1= ok
  0= directory with content
 -1= not an accessible directory
*/
{
 int ret;
 char *Sfile_fgets();
 char **filev= NULL;
 int filec= 0;

 ret= Dir_ls(dir,&filec,&filev,1,1|4); /* skip "." and ".." , no stderr */
 if(ret<=0)
   return(-1);
 if(filec<1)
   return(1);
 return(0);
}


int Remove_dir(dir,flag)
char *dir;
int flag;
/*
 bit0= report dir at stdout if it exists rather than deleting it
 bit1= with bit0==0 : 
       print rmdir command to stdout rather than execute rmdir()
 bit2= no error messages
*/
/*
 return:
  -1 = directory missing or not accessible
   0 = could not remove directory 
   1 = ok
*/
{
 int ret;
 char shell_adr[3*4096];

 ret= Dir_is_empty(dir,0);
 if(ret==1) {
   if(flag&1) {
     fprintf(stdout,"%s/\n",dir);
   } else {
     if(flag&2)
       fprintf(stdout,"rmdir %s\n",Text_shellsafe(dir,shell_adr,0));
     else {
       if(rmdir(dir)==-1) {
         if(!(flag&4))
           fprintf(stderr,"--- Cannot remove directory : %s \n    %s\n",
                          dir,strerror(errno));
         return(0);
       }
     }
   }
 } else if(ret==0) {
   if(flag&1) {
     fprintf(stdout,"%s/\n",dir);
   } else {
     if(!(flag&4))
       fprintf(stderr,"--- directory not empty : %s\n",dir);
     if(flag&2)
       fprintf(stdout,"# rmdir %s\n",Text_shellsafe(dir,shell_adr,0));
     return(0);
   }
 } else if(ret==-1) {
   if(!(flag&4))
     fprintf(stderr,"--- directory missing or not accessible : %s\n",dir);
   return(-1);
 }
 return(1);
}


int Remove_file(victim,remove_target,flag)
char *victim;
char *remove_target;
int flag;
{
 char line[2*4096],dirpath[4096],*cpt,*npt;
 int ret,errn,dtype;
 double size;
 
 if(remove_target[0]==0) {
   ret= unlink(victim);
   if(ret==-1) {
     fprintf(stderr,"Remove_file : unlink failed :\n");
     fprintf(stderr,"victim=%s\n",victim);
     fprintf(stderr,"%s\n",strerror(errno));
     {ret= 0; goto ex;}
   }
 } else {
   if(Sfile_type(remove_target,1)>=0) {
     fprintf(stderr,"Remove_file : remove_target already exists\n");
     fprintf(stderr,"victim=%s\n",victim);
     fprintf(stderr,"remove_target=%s\n",remove_target);
     {ret= 0; goto ex;}
   }

   /* create necessary directories */
   cpt= remove_target;
   dirpath[0]= 0;
   if(*cpt=='/') {
     strcpy(dirpath,"/");
     cpt++;
   }
   while(1) {
     npt= strchr(cpt,'/');
     if(npt==NULL)
   break;
     if(npt-cpt>0) {
       strncat(dirpath,cpt,npt-cpt);
       dtype= Type_and_size(dirpath,&size,1);
       if(dtype==-1) {
         ret= mkdir(dirpath,0700);
         if(ret==-1) {
           fprintf(stderr,
                  "Remove_file : creation of directory failed :\n");
           fprintf(stderr,"dirpath=%s\n",dirpath);
           fprintf(stderr,"%s\n",strerror(errno));
           {ret= 0; goto ex;}
         }
         chmod(dirpath,0700);
       }
     } 
     strcat(dirpath,"/");
     cpt= npt+1;
   }

   ret= Sfile_move_file(victim,remove_target,&errn,line,0);
   if(ret<=0) {
     fprintf(stderr,"Remove_file : removal failed :\n");
     fprintf(stderr,"victim=%s\n",victim);
     fprintf(stderr,"remove_target=%s\n",remove_target);
     fprintf(stderr,"%s\n",line);
     goto ex;
   }
 }
 ret= 1;
ex:
 return(ret);
}


int Stdout_pipe_broker(data,datacount,flag)
char *data;
int datacount;
int flag;
/*
 bit0= set command (else write output)
 bit15= end output
*/
{
 static char cmd[4096];
 static FILE *fp= NULL;
 static int banned= 0;
 int ret;

 if(flag&1) {
   if(fp!=NULL && fp!=stdout)
     pclose(fp);
   fp= NULL;
   strcpy(cmd,data);
   banned= 0;
   return(1);
 }
 if(flag&(1<<15)) {
   if(fp!=NULL && fp!=stdout) {
     ret= pclose(fp);
     if(ret!=0)
       return(0);
   }
   fp= NULL;
   strcpy(cmd,"");
   return(1);
 }
 if(banned)
   return(0);
 if(fp==NULL) {
   if(strlen(cmd)>0) {
     fp= popen(cmd,"w");
     if(fp==NULL){
       fprintf(stderr,"popen(\"%s\",\"w\") failed. errno=%d",cmd,errno);
       banned= 1;
       return(0);
     }
   } else
     fp= stdout;
 }
 ret= fwrite(data,1,datacount,fp);
 if(ret<datacount) {
   if(strlen(cmd)>0) {
     fprintf(stderr,"write to pipe failed. errno=%d",errno);
     fprintf(stderr,"pipe command was : %s\n",cmd);
     banned= 1;
     return(0);
   }
 }
 return(1);
}


int File_is_changed(f_adr,f_flag,f_type,f_size,f_crc,f_md5,
                    f_device,f_inode,
                    o_flag,o_type,o_size,o_crc,o_md5,
                    o_device,o_inode,
                    flag)
/* recent file properties */
char *f_adr;
int f_flag;  /* subset of FilelistE->flags :
                bit3= f_crc valid
                bit4= use f_md5 rather than f_crc
                bit5= compute block device crc from content and not from stat
                bit6= f_device and f_inode valid
              */
int f_type;
double f_size;
unsigned long f_crc;
unsigned char f_md5[16];
dev_t f_device;
ino_t f_inode;
/* old file properties */
int o_flag;  /* subset of FilelistE->flags :
                bit3= o_crc valid ,
                bit4= use o_md5 rather than o_crc 
                bit6= o_device & o_inode valid
              */
int o_type;
double o_size;
unsigned long o_crc;
unsigned char o_md5[16];
dev_t o_device;
ino_t o_inode;
int flag;
/*
 bit0= determine f_type and f_size from from f_adr, set f_flags=0
 bit1= count missing information as file change
 bit2= read properties of eventual link target

 bit6= Determine_file_checksum(): eventually pacify by size count

 bit9= check type 
 bit10= check size
 bit11= check crc
 bit13= check device and inode
 bit14= permission to return 0 after inode test reveiled no change

 bit17= with bit13: ignore device

return:
 <0 = error
  0 = no change detected
  1 = change detected
  2 = not found in filesystem (with bit0)
*/
{
 int ret,i;
 unsigned char md5[16];

 if(flag&1) {
   f_type= Type_and_size(f_adr,&f_size,1|(flag&4));
   if(f_type<0)
     Return_with_debug_msG(2, "cannot determine f_type");
 } 
 if((flag&512)) {
   if(o_type<=0) {
     if(flag&2)
       Return_with_debug_msG(1, "o_type<=0");
   } else if(o_type!=f_type)
       Return_with_debug_msG(1, "o_type!=ftype");
 }
 if(flag&1024) {
   if(o_size<0 || f_size<0) {
     if(flag&2)
       Return_with_debug_msG(1, "o_size<0 || f_size<0");
   } else if(f_size>=0 && o_size!=f_size)
     Return_with_debug_msG(1, "f_size>=0 && o_size!=f_size");
 }
 if((flag&(1<<13))) {
   if(o_flag&64) {
     if(!(f_flag&64)) {
       if(File_inode(f_adr,&f_device,&f_inode,0)>0)
         f_flag|= 64;
     }
     if(f_flag&64) {
       if(f_device!=o_device && !(flag&(1<<17)))
         Return_with_debug_msG(1, "f_device!=o_device");
       if(f_inode!=o_inode)
         Return_with_debug_msG(1, "f_inode!=o_inode");
       if(flag&(1<<14))              /* rely on inode test and skip CRC test */
         Return_with_debug_msG(0, "inode test suffices");
     } else if(flag&2)
       Return_with_debug_msG(1, "f_flag indicates no inode+device info");
   } else if(flag&2)
     Return_with_debug_msG(1, "o_flag indicates no inode+device info");
 }
 if((flag&2048) && (o_flag&8)) {
   if(o_flag&16) {
     if((f_flag&(8|16))==(8|16)) {
       memcpy(md5,f_md5,16);
       ret= 1;
     } else {

       /* ??? does ((f_flag&32)^32) make sense ?  
              do we ever get such items here ? if ever, is it worth ?
              A50830: yes we get here  */
       ret= Determine_file_checksum(f_adr,NULL,NULL,md5,md5,&f_size,NULL,
                                    NULL,((f_flag&32)^32)|(1<<12)|(flag&64));
     }

     if(ret<=0 && (flag&2))
       Return_with_debug_msG(1, "failure to obtain file MD5");
     if(ret>0) {
       for(i=0;i<16;i++)
         if(md5[i]!=o_md5[i])
       break;
       if(i<16)
         Return_with_debug_msG(1, "MD5 mismatch");
     }
   } else {
     if((f_flag&(8|16))==8) {
       ret= 1;
     } else
       ret= Determine_file_checksum(f_adr,NULL,NULL,&f_crc,&f_crc,&f_size,
                                    NULL,NULL,(flag&64)|((f_flag&32)^32));
     if(ret<=0 && (flag&2))
       Return_with_debug_msG(1, "failure to obtain file CRC");
     if(ret>0 && o_crc!=f_crc)
       Return_with_debug_msG(1, "CRC mismatch");
   }
 }
 Return_with_debug_msG(0, "all tests passed");
}


int Print_pacifier(count,in_text,flag)
int count;
char *in_text;
int flag;
/*
 bit0= print regardless of time (if not totally disabled)
 bit1= print count
 bit2= print text
 bit3= end of pacifying, print newline
 bit4= make line empty
 bit5= permission to truncate text if necessary
 bit6= make half line empty
 bit7= try wether there is a wannabe waiting and wether it may be shown now 
       caution: restores flag as it was with wannabe (plus bit7 set)
 bit8= prepend text to wannabe text and print even if wannabe invalidated 
 bit9= use short text format for count
bit10= do not become wannabe
bit11= bit8 with invalidated wannabe only if timestamp with pacifier
*/
{
 int maxlen= 79;
 char endchar,*text,timestamp[20];
 static char wannabe_text[2*80+1],truncated_text[81];

 if(Cd_backup_planer_pacifier_intervaL<0)
   return(2);

 if(Cd_backup_planer_pacifier_new_linE) {
   if((flag&(16|64)) && (flag&(2|4))==0)         /* skip mere line cleaners */
     return(2);
   flag&= ~(16|32|64);              /* disable line cleaning and truncation */
   flag|= 8;                                              /* enable newline */
 }

 text= in_text;
 if(Cd_backup_planer_pacifier_show_timE && (flag&(1<<11)) &&
    !(Cd_backup_planer_pacifier_wannabe_flaG&128)) /* no wannabe waiting */
   flag&= ~256;

 if((!(flag&1)) && time(0)<Cd_backup_planer_pacifier_timestamP+
                           Cd_backup_planer_pacifier_intervaL) {
   if(!(flag&(128|1024))) {
     /* register attempt as current wannabe */
     Cd_backup_planer_pacifier_wannabe_counT= count;
     strncpy(Cd_backup_planer_pacifier_wannabe_texT,text,
             sizeof(Cd_backup_planer_pacifier_wannabe_texT)-1);
     Cd_backup_planer_pacifier_wannabe_texT[
                          sizeof(Cd_backup_planer_pacifier_wannabe_texT)-1]= 0;
     Cd_backup_planer_pacifier_wannabe_flaG= flag|128;
   }
   return(2);
 }
 if(flag&(128|256)) {
   if((Cd_backup_planer_pacifier_wannabe_flaG&128)||(flag&256)) { 
     /* reinstate wannabe as pacifier call */
     count= Cd_backup_planer_pacifier_wannabe_counT;
     if(flag&256) {
       if(Cd_backup_planer_pacifier_new_linE) 
         sprintf(wannabe_text,"%s%s",
                              text,Cd_backup_planer_pacifier_wannabe_texT);
       else
         sprintf(wannabe_text,"%.80s%.80s",
                              text,Cd_backup_planer_pacifier_wannabe_texT);
       text= wannabe_text;
     } else
       text= Cd_backup_planer_pacifier_wannabe_texT;
     flag= Cd_backup_planer_pacifier_wannabe_flaG|128|(flag&256);
   } else
     return(2); /* no wannabe waiting */
 }
 if(flag&16)
   fprintf(stderr,"\r%78.78s\r","");
 else if(flag&64)
   fprintf(stderr,"\r%40.40s\r","");
 if(flag&8)
   endchar= '\n';
 else
   endchar= '\r';

 timestamp[0]= 0;
 if(Cd_backup_planer_pacifier_show_timE) {
   sprintf(timestamp,"%d",(int) time(0)-Cd_backup_planer_start_timE);
   strcat(timestamp," ");
   maxlen-= strlen(timestamp);
 }

 if((flag&6)==2)
   fprintf(stderr,"%s  %d%c",timestamp,count,endchar);
 else if((flag&6)==4) {
   if((flag&32) && strlen(text)>maxlen-2) {
     strncpy(truncated_text,text,maxlen-5);
     strcpy(truncated_text+maxlen-5,"...");
     text= truncated_text;
   }
   fprintf(stderr,"%s%s%c",timestamp,text,endchar);
 } else if((flag&6)==6) {
   if((flag&32) && strlen(text)>maxlen-14) {
     strncpy(truncated_text,text,maxlen-17);
     strcpy(truncated_text+maxlen-17,"...");
     text= truncated_text;
   }
   if(flag&512)
     fprintf(stderr,"%s  %d  %s%c",timestamp,count,text,endchar);
   else
     fprintf(stderr,"%s  %-10d %s%c",timestamp,count,text,endchar);
 }
 Cd_backup_planer_pacifier_timestamP= time(0);
 if(!(flag&(128|256))) {
   /* register new pacifier as invalidated wannabe (ready for bit8) */
   Cd_backup_planer_pacifier_wannabe_counT= count;
   strncpy(Cd_backup_planer_pacifier_wannabe_texT,text,
           sizeof(Cd_backup_planer_pacifier_wannabe_texT)-1);
   Cd_backup_planer_pacifier_wannabe_texT[
                          sizeof(Cd_backup_planer_pacifier_wannabe_texT)-1]= 0;
   Cd_backup_planer_pacifier_wannabe_flaG= flag&~128;
 }
 return(1);
}


double Scanf_io_size(text,flag)
char *text;
int flag;
/*
 bit0= default value -1 rather than 0 
*/
{
 int c;
 double ret= 0.0;

 if(flag&1)
   ret= -1.0;
 if(text[0]==0)
   return(ret);
 sscanf(text,"%lf",&ret);
 c= text[strlen(text)-1];
 if(c=='k' || c=='K') ret*= 1024.0;
 if(c=='m' || c=='M') ret*= 1024.0*1024.0;
 if(c=='g' || c=='G') ret*= 1024.0*1024.0*1024.0;
 if(c=='t' || c=='T') ret*= 1024.0*1024.0*1024.0*1024.0;
 if(c=='p' || c=='P') ret*= 1024.0*1024.0*1024.0*1024.0*1024.0;
 if(c=='e' || c=='E') ret*= 1024.0*1024.0*1024.0*1024.0*1024.0*1024.0;
 return(ret);
}

int Fseek_double(fp,adr,flag)
FILE *fp;
double adr;
int flag;
{
 int ret;

#ifdef Cd_backup_planer_fseekO
 ret= Cd_backup_planer_fseekO(fp,(off_t) adr,SEEK_SET);
#else
 if(adr>=2.0*1024.0*1024 && sizeof(long)<6) {
   Print_pacifier(0,"",1|8|16);
   fprintf(stderr,"Byteaddress too large : %.lf\n",adr);
   return(0);
 }
 ret= fseek(fp,(long) adr,SEEK_SET);
#endif /* Cd_backup_planer_fseekO */
 return(ret);
}


int Write_broker_or_file(fp,buffer,bufcount,flag)
FILE *fp;
char *buffer;
int bufcount;
int flag;
/*
 bit0= in case of error message clear pacifier line first
*/
{
 int ret;

 if(bufcount<=0)
   return(1);
 if(fp==stdout)
   ret= Stdout_pipe_broker(buffer,bufcount,0);
 else
   ret= fwrite(buffer,bufcount,1,fp);
 if(ret<=0) {
   if(ferror(fp) && (flag&1)) {
     Print_pacifier(0,"",1|16);
     perror("");
   }  
   return(ret);
 }
 return(1);
}


double Permutation_index(x,n,permutation_key,flag)
double x;
double n;
int permutation_key; 
           /* 0= id
             1= rotate by half ( m<=n/2 )
             2= mirror
             3= swap quarters ( q1<=m/2 , q2<=(m+n)/2 )
            +4= interlace resulting first and second half
           */
int flag;
/*
 bit0= invert
*/
{
 int mode,interlace;
 int hvec[CdbplivS];
 double idx,m,q1,q2;  /*  0--q1--m--q2--n */
 double Intvec_to_double();

 Intvec_from_double(hvec,CdbplivS,n/2.0,0);
 m= Intvec_to_double(hvec,CdbplivS,0);                 /* instead of floor() */
 idx= x;
 mode= permutation_key&3;
 interlace= !!(permutation_key&4);
 if(interlace && !(flag&1)) {
   Intvec_from_double(hvec,CdbplivS,idx/2.0,0);     /* instead of % operator */
   if(Intvec_to_double(hvec,CdbplivS,0)*2.0!=idx)
     idx= (idx-1.0)/2.0;
   else
     idx= idx/2.0+m;
 }
 if(mode==0) { /* identical mapping */
   ;
 } else if(mode==1) { /* swap halves */
   if(flag&1) {
     if(idx<m)
       idx= idx+n-m;
     else
       idx= idx-m;
   } else {
     if(idx>=n-m)
       idx= idx-(n-m);
     else
       idx= idx+m; 
   }
 } else if(mode==2) { /* mirror */
   idx= n-1.0-idx;
 } else if(mode==3) { /* swap quarters */;
   Intvec_from_double(hvec,CdbplivS,m/2.0,0);
   q1= Intvec_to_double(hvec,CdbplivS,0);
   Intvec_from_double(hvec,CdbplivS,(m+n)/2.0,0);
   q2= Intvec_to_double(hvec,CdbplivS,0);
   if(flag&1) {
     if(idx<q1)
       idx= idx+m-q1;
     else if(idx<m)
       idx= idx-q1;
     else if(idx<q2)
       idx= idx+n-q2;
     else
       idx= idx-q2+m;
   } else {
     if(m-q1<=idx && idx<=m-1)
       idx= idx-m+q1;
     else if(0<=idx && idx<=m-1-q1)
       idx= idx+q1;
     else if(m+n-q2<=idx && idx<=n-1)
       idx= idx-n+q2;
     else
       idx= idx+q2-m;
   }
 }
 if(interlace && (flag&1)) {
   if(idx<m)
     idx= idx*2.0+1.0;
   else
     idx= 2.0*(idx-m); 
 }
 return(idx);
}


int Decode_numvar(line,keytext,result,flag)
char *line;
char *keytext;
double *result;
int flag;
{
 int l,ret;
 int hvec[CdbplivS];
 double Intvec_to_double();

 (*result)= 0;
 l= strlen(keytext);
 if(strncmp(line,keytext,l)!=0)
   return(0);
 ret= Intvec_from_dec(hvec,CdbplivS,line+l,1);
 if(ret<=0)
   return(0);
 (*result)= Intvec_to_double(hvec,CdbplivS,0);
 return(1);
}


int Pad_broker_or_file(fpo,pad_len,flag)
FILE *fpo;
double pad_len;
int flag;
/*
 bit0= in case of error message clear pacifier line first
*/
{
 int ret;
 char buffer[81];

 sprintf(buffer,"%79.79s\n","");
 while(pad_len>0.0) {
   if(pad_len>80.0) {
     ret= Write_broker_or_file(fpo,buffer,80,flag&1);
     pad_len-= 80.0;
   } else {
     buffer[((int) pad_len)-1]= '\n';
     ret= Write_broker_or_file(fpo,buffer,(int) pad_len,flag&1);
     pad_len= 0.0;
   }
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Strcmp_pt(pt1,pt2)
char ***pt1,***pt2;
{
 return(strcmp(**pt1,**pt2));
}


int Strcmp(pt1,pt2)
char **pt1,**pt2;
{
 return(strcmp(*pt1,*pt2));
}



int Filter_fp(fp_in,fp_out,from_text,to_text,textcount,writecount,flag)
FILE *fp_in,*fp_out;
char *from_text,*to_text;
unsigned *textcount,*writecount;
int flag;
/* 
 bit0= do not replace from_text by to_text
 bit1= demand exact match (without line end marker)
*/
{
 int from_len,to_len;
 char line[4096];
 char *Sfile_fgets();

 from_len= strlen(from_text);
 to_len= strlen(to_text);
 while(1) {
   if(Sfile_fgets(line,sizeof(line)-1,fp_in)==NULL)
 break;
   if( (from_len>0 && strncmp(line,from_text,from_len)==0)
       && ((!(flag&2)) || strlen(line)==from_len) ) {
     if(flag&1) {
       fprintf(fp_out,"%s\n",line);
     } else {
       fprintf(fp_out,"%s",to_text);
       fprintf(fp_out,"%s\n",line+from_len);
     }
     (*textcount)++;
   } else {
     fprintf(fp_out,"%s\n",line);
   }
   (*writecount)+= strlen(line)+1;
 }
 return(1);
}


int Filter_command(cmd,from_text,to_text,fp,textcount,flag)
char *cmd,*from_text,*to_text;
FILE *fp;
unsigned *textcount;
int flag;
/* 
 bit0= do not replace from_text by to_text
 bit1= demand exact match (without line end marker)
*/
{
 int ret;
 unsigned writecount= 0;
 FILE *fp_in= NULL;

 wait3(NULL,WNOHANG,NULL);
 fp_in= popen(cmd,"r");
 if(fp_in==NULL)
   {ret= 0; goto ex;}
 ret= Filter_fp(fp_in,fp,from_text,to_text,textcount,&writecount,flag&3);
ex:;
 if(fp_in!=NULL)
   pclose(fp_in);
 return(ret);
}


int Filter_file(file_in,file_out,from_text,to_text,textcount,flag)
char *file_in,*file_out;
char *from_text,*to_text;
unsigned *textcount;
int flag;
/* 
 bit0= do not replace from_text by to_text
 bit1= demand exact match (without line end marker)
*/
{
 int ret;
 unsigned writecount= 0;
 FILE *fp_in= NULL,*fp_out= NULL;

 fp_in= fopen(file_in,"r");
 if(fp_in==NULL) {
   perror(file_in);
   {ret= 0; goto ex;}
 }
 fp_out= fopen(file_out,"r+");
 if(fp_out==NULL) {
   perror(file_out);
   {ret= 0; goto ex;}
 }
 ret= Filter_fp(fp_in,fp_out,from_text,to_text,textcount,&writecount,flag&3);
ex:;
 if(fp_in!=NULL)
   fclose(fp_in);
 if(fp_out!=NULL) {
   fclose(fp_out);
 }
 return(ret);
}


int Sort_argv(argc,argv,flag)
int argc;
char **argv;
int flag;
{
 if(argc<=0)
   return(2);
 qsort(argv,(size_t) argc,sizeof(char *),Strcmp);
 return(1);
}


/* ------------------------------------------------------------------------- */
/* counting mechanism for up to 900 bit (usually : iv_size=3 : 90 bit ) */

#define Intvec_maxsizE 32
#define Intvec_b30 (1<<30)

static double Intvec_bscalE[Intvec_maxsizE]= {0.0};


int Intvec__init(flag)
int flag;
{
 int i;

 Intvec_bscalE[0]= 1.0;
 for(i=1;i<Intvec_maxsizE;i++) {
   Intvec_bscalE[i]= Intvec_bscalE[i-1]*(double) Intvec_b30;
 }
 return(1);
} 


int Intvec_sign(ivec,ivec_size,flag)
int ivec[];
int ivec_size;
int flag;
{
 int i;

 for(i=ivec_size-1;i>=0;i--) {
   if(ivec[i]>0)
     return(1);
   if(ivec[i]<0)
     return(-1);
 }
 return(0);
}


int Intvec_cmp(vec1,ivec_size,vec2,flag)
int vec1[];
int ivec_size;
int vec2[];
int flag;
/*
 bit0= invert sign of vec2
*/
{
 int i;

 if(flag&1) {
   for(i= ivec_size-1;i>=0;i--) {
     if(vec1[i]>-vec2[i])
       return(1);
     if(vec1[i]<-vec2[i])
       return(-1);
   }
 } else {
   for(i= ivec_size-1-1;i>=0;i--) {
     if(vec1[i]>vec2[i])
       return(1);
     if(vec1[i]<vec2[i])
       return(-1);
   }
 }
 return(0);
}


int Intvec_invert(ivec,ivec_size,flag)
int ivec[];
int ivec_size;
int flag;
{
 int i;

 for(i=0;i<ivec_size;i++)
   ivec[i]= -ivec[i];
 return(1);
}


int Intvec_copy(source,ivec_size,target,flag)
int source[];
int ivec_size;
int target[];
int flag;
{
 int i;

 for(i=0;i<ivec_size;i++)
   target[i]= source[i];
 return(1);
}


#ifdef Cdbackup_planer_iso_C
int Intvec_add(int res[], int ivec_size, int op1[], int op2[], int flag)
#else /* Cdbackup_planer_iso_C */
int Intvec_add(res,ivec_size,op1,op2,flag)
int res[];
int ivec_size;
int op1[],op2[];
int flag;
#endif /* ! Cdbackup_planer_iso_C */
/*
 bit0= subtract
*/
{
 int i,j,k,a,c,s;
 int ivec[Intvec_maxsizE];

 /* perform the preliminary operation */
 c= 0;
 s= 1-2*(flag&1);
 for(i=0;i<ivec_size;i++) {
   a= op1[i]+s*op2[i]+c;
   if(a>=Intvec_b30) {
     c= 1;
     a-= Intvec_b30;
   } else if(a<= -Intvec_b30) {
     c= -1;
     a+= Intvec_b30;
   } else
     c= 0;
   ivec[i]= a;
 }

 /* make signs coherent */
 s= 0;
 for(i=ivec_size-1;i>=0;i--) {
   if(s==0 && ivec[i]==0)
 continue;
   if(s==0)
     s= (ivec[i]>0?1:-1);
   /* is next non zero component of different sign ? */
   for(j= i-1;j>=0;j--) 
     if(ivec[j]!=0)
   break;
   if(j<0)
 break;
   if(ivec[j]*s>=0) {
     i= j+1;
 continue;
   }
   /* borrow s and the eventual overflow carry to the lower ranks */
   ivec[i]-= s-c;
   c= 0;
   for(k=i-1;k>j;k--)  
     ivec[k]= s*(Intvec_b30-1);
   ivec[j]+= s*Intvec_b30;
   i= j+1;
 }
 if(c!=0) {
   /* overflow */
   return(0);
 }

 for(i=ivec_size-1;i>=0;i--)
   res[i]= ivec[i];

 return(1);
}


#ifdef Cdbackup_planer_iso_C 
int Intvec_add_b30(int *ivec,int ivec_size,int addon,int flag)
#else
int Intvec_add_b30(ivec,ivec_size,addon,flag)
int *ivec;
int ivec_size;
int addon;
int flag;
#endif
/*
 bit0= subtract
*/
{
 int i,ret;
 int hvec[Intvec_maxsizE];

 if(ivec==NULL || ivec_size==0)
   return(0);
 if(addon>=Intvec_b30 || addon<=-Intvec_b30)
   return(-2);
 for(i=1;i<ivec_size;i++)
   hvec[i]= 0;
 hvec[0]= addon;
 ret= Intvec_add(ivec,ivec_size,ivec,hvec,flag&1);
 return(ret);
}


int Intvec_mult10(ivec,ivec_size,flag)
int ivec[];
int ivec_size;
int flag;
{
 int i,c;
 double a;

 c= 0;
 for(i=0;i<ivec_size;i++) {
   a= ((double) ivec[i])*10.0+((double) c);
   if(a>=Intvec_b30) {
     c= a/((double) Intvec_b30);
     a-= ((double) c)*((double) Intvec_b30);
   } else if(a<= -Intvec_b30) {
     c= -a/((double) Intvec_b30);
     a+= ((double) c)*((double) Intvec_b30);
   } else
     c= 0;
   ivec[i]= a;
 }
 if(c!=0) {
   /* overflow */
   return(0);
 }

 return(1);
}


int Intvec_div10(ivec,ivec_size,flag)
int ivec[];
int ivec_size;
int flag;
/*
 return:
  <0 = error
   0 - 9 : remainder of division
*/
{
 int i,c;
 double a;

 c= 0;
 a= 0.0;
 for(i=ivec_size-1;i>=0;i--) {
   a= a*Intvec_b30+ivec[i];
   if(a>-10.0 && a<10.0) {
     ivec[i]= 0;
 continue;
   }
   c= a/10.0;
   a-= 10.0*((double) c);
   ivec[i]= c;
 } 
 return((int) a);
}


int Intvec_div(numerator,ivec_size,denominator,result,remainder,flag)
int *numerator;
int ivec_size;
int *denominator;
int *result;
int *remainder;
int flag;
{
 Intvec_from_double(result,ivec_size,0.0,0);
 Intvec_copy(numerator,ivec_size,remainder,0);
 if(Intvec_sign(denominator,ivec_size,0)==0)
   return(0);
 while(Intvec_sign(remainder,ivec_size,0)>0) {
   Intvec_add(remainder,ivec_size,remainder,denominator,1);
   Intvec_add_b30(result,ivec_size,1,0);
 }
 if(Intvec_sign(remainder,ivec_size,0)!=0) {
   Intvec_add_b30(result,ivec_size,1,1);
   Intvec_add(remainder,ivec_size,remainder,denominator,0);
 }
 return(1);
}

   
int Intvec_from_double(ivec,ivec_size,value,flag)
int *ivec;
int ivec_size; 
double value;
int flag;
/*
 return:
   <=0 = error
    >0 = minimum ivec_size which fits the number
*/
{
 double pvalue;
 int i,min_size;

 if(ivec==NULL || ivec_size==0)
   return(0);
 if(Intvec_bscalE[0]==0.0)
   Intvec__init(0);
 for(i=0;i<ivec_size;i++)
   ivec[i]= 0;
 if(value>0)
   pvalue= value;
 else
   pvalue= -value;
 for(i=1;i<Intvec_maxsizE;i++)
   if(Intvec_bscalE[i]>pvalue)
 break;
 min_size= i;
 if(min_size>=Intvec_maxsizE || min_size-1>=ivec_size) {  /* ??? -1 with min_size ??? */
   /* more than 30*Intvec_maxsizE bits or more than ivec can take */
   return(-1);
 }
 for(i--;i>=0;i--) {
   ivec[i]= pvalue/Intvec_bscalE[i];
   pvalue-= ivec[i]*Intvec_bscalE[i];
   if(value<0)
     ivec[i]= -ivec[i];
 }
 return(min_size);
}


double Intvec_to_double(ivec,ivec_size,flag)
int *ivec;
int ivec_size;
int flag;
{
 double r;
 int i;

 if(ivec==NULL || ivec_size==0)
   return(0);
 if(Intvec_bscalE[0]==0.0)
   Intvec__init(0);
 r= 0.0;
 for(i=0;i<ivec_size;i++)
   r+= Intvec_bscalE[i]*(double) ivec[i];
 return(r);
}


int Intvec_to_b30(ivec,ivec_size,flag)
int *ivec;
int ivec_size;
int flag;
{
 int i;

 if(ivec==NULL || ivec_size==0)
   return(0x40000000);
 for(i=1; i<ivec_size;i++)
   if(ivec[i]>0)
     return(0x40000000);
 return(ivec[0]);
}


int Intvec_from_dec(ivec,ivec_size,dec,flag)
int *ivec;
int ivec_size;
char *dec;
int flag;
/*
  bit0= return nicely when encountering railing non-numericals 
 return:
  -1= overflow
   0= not a decimal number
  >0= minimum ivec_size which fits the number
*/
{
 int i,l,c,ret,s= 1;
 
 if(ivec==NULL || ivec_size==0)
   return(0);
 Intvec_from_double(ivec,ivec_size,0.0,0);
 l= strlen(dec);
 for(i=0;i<l;i++) {
   if(dec[i]=='-'&&i==0) {
     s= -1;
 continue;
   }
   c= dec[i]-'0';
   if(c<0 || c>9) {
     if(flag&1)
       goto finish;
     return(0);
   }
   ret= Intvec_mult10(ivec,ivec_size,0);
   if(ret<=0)
     return(-1);
   ret= Intvec_add_b30(ivec,ivec_size,c,0);
   if(ret<=0)
     return(-1);
 }
finish:;
 if(s<0)
   Intvec_invert(ivec,ivec_size,0);
 for(i=ivec_size-1;i>0;i--)
   if(ivec[i]!=0)
 break;
 return(i+1);
}


int Intvec_to_dec(ivec,ivec_size,dec,dec_size,flag)
int *ivec;
int ivec_size;
char *dec;
int dec_size;
int flag;
{
 int i,ret,s,l= 0,tr;
 int hvec[Intvec_maxsizE];

 if(ivec==NULL || ivec_size==0)
   return(0);
 if(dec_size<=0)
   return(-2);
 dec[0]= 0;
 s= Intvec_sign(ivec,ivec_size,0);
 Intvec_copy(ivec,ivec_size,hvec,0);
 if(s<0)
   Intvec_invert(hvec,ivec_size,0);
 while(l+1+1<dec_size) {
   ret= Intvec_div10(hvec,ivec_size,0);
   if(ret<0)
     return(0);
   dec[l++]= ret+'0';
   dec[l]= 0;
   if(Intvec_sign(hvec,ivec_size,0)==0)
 break;
 }
 if(l+1>=dec_size || Intvec_sign(hvec,ivec_size,0)!=0)
   return(-1);
 if(s<0)
   dec[l++]= '-';
 dec[l]= 0;
 for(i=0;i<l/2;i++) {
   tr= dec[i];
   dec[i]= dec[l-i-1];
   dec[l-i-1]= tr;
 }  
 return(1);
}


int Intvec_count_bits(ivec,ivec_size,flag)
/* counts the magnitude bits of ivec but not the sign bits */
int *ivec;
int ivec_size;
int flag;
{
 int i,j;

 for(i= ivec_size-1;i>0;i--)
   if(ivec[i]&0x3fffffff)
 break;
 for(j=29;j>=0;j--)
    if(ivec[i]&(1<<j))
 break;
 return(i*30+j+1);
} 


/* the following routines assume dev_t and ino_t to be unsigned
   integers of some unknown length */

int Intvec_from_dev_ino(dev_vec,ivec_size,ino_vec,dev,ino,flag)
int *dev_vec;
int ivec_size;
int *ino_vec;
dev_t dev;
ino_t ino;
int flag;
{
 int i;
 dev_t dev_a;
 ino_t ino_a;

 dev_a= dev;
 Intvec_from_double(dev_vec,ivec_size,0.0,0);
 for(i= 0; dev_a>0 && i<ivec_size-1; i++) {
    dev_vec[i]= dev_a&0x3fffffff;
    dev_a= dev_a>>30;
 }
 if(dev_a>0)
   return(0); /* overflow */
 ino_a= ino;
 Intvec_from_double(ino_vec,ivec_size,0.0,0);
 for(i= 0; ino_a>0 && i<ivec_size-1; i++) {
    ino_vec[i]= ino_a&0x3fffffff;
    ino_a= ino_a>>30;
 }
 if(ino_a>0)
   return(0); /* overflow */
 return(1);
}


int Intvec_to_dev_ino(dev_vec,ivec_size,ino_vec,dev,ino,flag)
int *dev_vec;
int ivec_size;
int *ino_vec;
dev_t *dev;
ino_t *ino;
int flag;
{
 int i;
 dev_t dev_a;
 ino_t ino_a;

 if(Intvec_sign(dev_vec,ivec_size,0)<0)
   return(0);
 if(Intvec_count_bits(dev_vec,ivec_size,0)>sizeof(dev_t)*8)
   return(0);
 if(Intvec_sign(ino_vec,ivec_size,0)<0)
   return(0);
 if(Intvec_count_bits(ino_vec,ivec_size,0)>sizeof(ino_t)*8)
   return(0);

 dev_a= 0;
 for(i= ivec_size-1; i>=0; i--) {
   dev_a= dev_a<<30;
   dev_a|= dev_vec[i]&0x3fffffff;
 }
 *dev= dev_a;
 ino_a= 0;
 for(i= ivec_size-1; i>=0; i--) {
   ino_a= ino_a<<30;
   ino_a|= ino_vec[i]&0x3fffffff;
 }
 *ino= ino_a;
 return(1);
}


off_t Intvec_to_off_t(ivec,ivec_size,flag)
int *ivec;
int ivec_size;
int flag;
{
 int i;
 off_t off_a;

 if(Intvec_sign(ivec,ivec_size,0)<0)
   return(0);
 if(Intvec_count_bits(ivec,ivec_size,0)>sizeof(off_t)*8)
   return(0);
 off_a= 0;
 for(i= ivec_size-1; i>=0; i--) {
   off_a= off_a<<30;
   off_a|= ivec[i]&0x3fffffff;
 }
 return(off_a);
}
 

/* ----------------------------------------------------------------- */


double Very_large_chunk_manager(duty_par,indexvec,indexcount,
                                blocksize,blockvec,flag)
double *duty_par;
int indexvec[];
int *indexcount;
double blocksize;
int blockvec[];
int flag;
/*
 bit0= initialize indexvec by *duty_par (*duty_par is set to an
       estimation of duty if not flag&2)
 bit1= do not advance a block after initialization. return 1
 bit2= allow automatically adjusted *indexcount
 bit3= no need to compute blockvec from blocksize
*/
{
 int ret,count1,count2;
 double duty,r;

 if(*indexcount<1)
   return(0);
 duty= *duty_par;
 if(flag&1) {
   ret= Intvec_from_double(indexvec,*indexcount,duty,0);
   if(ret<=0)
     return(ret);
   count1= ret;
   ret= Intvec_from_double(blockvec,*indexcount,blocksize,0);
   if(ret<=0)
     return(ret);
   count2= ret;
   if(flag&4) {
     if(count1>count2)
       *indexcount= count1;
     else
       *indexcount= count2;
   }
   if(flag&2)
     return(1);
 }

 if(!(flag&8)) {
   /* convert blocksize to blockvec */
   ret= Intvec_from_double(blockvec,*indexcount,blocksize,0);
   if(ret<=0)
     return(0);
 }
 if(blocksize<0.0)
   if(Intvec_cmp(indexvec,*indexcount,blockvec,1)<=0) {
     /* we are done : convert indexvec to double, zeroize and return */;
     r= Intvec_to_double(indexvec,*indexcount,0);
     Intvec_from_double(indexvec,*indexcount,0.0,0);
     return(r);
   }

 ret= Intvec_add(indexvec,*indexcount,indexvec,blockvec,0);
 if(ret<=0)
   return(ret);
 (*duty_par)= Intvec_to_double(indexvec,*indexcount,0);
 if(blocksize<0.0)
   return(-blocksize);
 return(blocksize);
}


double Count_file_bytes(adr,flag)
char *adr;
int flag;
/*
 bit0= do not issue pacifier messages
*/
{
 FILE *fp= NULL;
 double r;
 int countvec[CdbplivS],blockvec[CdbplivS],indexcount= CdbplivS;
 int blocksize,last_ret,ret,block_count= 0;
 char buffer[65536],msg[4096+100],size_text[8],dec[80];
 struct stat stbuf;
 
 r= -1.0;
 Intvec_from_double(countvec,indexcount,-1.0,0);
 if(!(flag&1)) {
   sprintf(msg,"beginning to count bytes from %s",adr);
   Print_pacifier(0,msg,1|4|16|32);
 }
 blocksize= sizeof(buffer);
 if(strcmp(adr,"-")==0) {
   fp= stdin;
 } else
   if(Cd_backup_planer_call_staT(adr,&stbuf)!=-1) {
     if(S_ISBLK(stbuf.st_mode))
       if(stbuf.st_blksize<blocksize)
         blocksize= stbuf.st_blksize;
     fp= fopen(adr,"rb");
   } 
 if(fp==NULL)
   goto ex;

 r= 0.0;
 Intvec_from_double(countvec,indexcount,0.0,0);
 Intvec_from_double(blockvec,indexcount,((double) blocksize),0);
 while(1) {
   ret= fread(buffer,1,blocksize,fp);
   if(ret<=0)
 break;
   last_ret= ret;
   ret= Intvec_add_b30(countvec,indexcount,ret,0);
   if(ret<=0) {
     /* >>> overflow of counter array (3*30 bit) */;
     ret= -1; goto ex;
   }
   if(!(flag&1)) {
     if(++block_count>=256) {
       block_count= 0;
       r= Intvec_to_double(countvec,indexcount,0);
       Sfile_scale(r,size_text,7,1e5,1);
       sprintf(msg,"  counting %s bytes from %s",size_text,adr);
       Print_pacifier(0,msg,4|16|32);
     }  
   }
 }
ex:;
 if(!(flag&1)) {
   r= Intvec_to_double(countvec,indexcount,0);
   Intvec_to_dec(countvec,indexcount,dec,sizeof(dec),0);
   if(r<0.0) {
     sprintf(msg,"--- could not count data bytes from %s\n",adr);
   } else if(r<1e5) {
     sprintf(msg,"counted from %s : %s bytes\n",adr,dec);
   } else {
     Sfile_scale(r,size_text,7,1e5,1);
     sprintf(msg,"counted from %s : %s bytes (%s)\n",adr,dec,size_text);
   }
   Print_pacifier(0,msg,1|4|8|16|32);
 }
 if(fp!=NULL)
   fclose(fp);
 return(r);
}


int Determine_file_checksum(adr,fp_in,handle,crc,pre_crc,size,ivec,skipvec,
                            flag)
char *adr;
FILE *fp_in;
char **handle;
char *crc;        /* should be (void *) */
char *pre_crc;    /* should be (void *) */
double *size;
int ivec[];
int skipvec[];
int flag;
/*
  bit0= assume datafile at fp_in 
  bit1= filtermode with bit0 : write data to stdout
  bit2= read only the first *size bytes of input
  bit3= with bit0 but not bit2: compute a pre_crc for the first *size bytes
  bit4= read eventual link target
  bit5= compute block device from stat information and not from content
  bit6= prepend size info to pacifier update
  bit7= with fseekable file objects: skip skipvec[] bytes
  bit8= with bit0 : do not initialize but use handle
  bit9= with bit0 : do not finalize but return handle

 bit12= produce MD5 in crc.md5[] rather than CRC32 crc.crc32
*/
{
 FILE *fp= NULL,*popen_fp= NULL;
 char buf[4*4096],*md5,*pre_md5,size_text[80];
 unsigned long *crc32,*pre_crc32;
 char *ctx= NULL, *pre_ctx= NULL;                     /* should be (void *) */
 int ret,first= 1,typ,make_pre_crc= 0,cut,i,read_bytes,bufsiz,extra_ret= 0;
 int pacifier_cycles= 0,pacifier_cycle_threshhold= 256;
 int countvec[CdbplivS],limitvec[CdbplivS];
 int hvec[CdbplivS],vec_size= CdbplivS;
 double todo,tomuch,r;
 struct stat stbuf;
 char *Sfile_fgets();
 struct DirseQ *dirseq= NULL;

 if(flag&(1<<12)) {
   md5= crc;
   pre_md5= pre_crc;
   crc32= pre_crc32= NULL;
 } else {
   crc32= (unsigned long *) crc;
   pre_crc32= (unsigned long *) pre_crc;
   md5= pre_md5= NULL;
 }
 if(flag&256)
   first= 0;
 make_pre_crc= !!(flag&8);
 if((flag&4) || make_pre_crc) {
   if(ivec!=NULL && vec_size>0)
     Intvec_copy(ivec,vec_size,limitvec,0);
   else
     ret= Intvec_from_double(limitvec,vec_size,*size,0);
 } else
   ret= Intvec_from_double(limitvec,vec_size,0.0,0);
 
 if(flag&256) {
   if(flag&(1<<12))
     ctx= *handle;
   *handle= NULL;
 } else {
   if(flag&(1<<12)) {
     ret= Sregex_md5(&ctx,buf,0,md5,1);
     if(ret<=0)
       {ret= 0; goto ex;}
   } else
     *(crc32)= 0;
 }
 if(flag&1) {
   typ= 1;
 } else {
   typ= Type_and_size(adr,size,(flag&16)>>2);
   if(typ<=0)
     {ret= 0; goto ex;}
   Intvec_from_double(ivec,vec_size,*size,0);/* this is safe for ivec==NULL */
 }
 if(typ==3) {                                              /* symbolic link */
#ifdef Stypes_on_ms_windowS
   {ret= 0; goto ex;}
#else /* Stypes_on_ms_windowS */
   ret= readlink(adr,buf,sizeof(buf)-1);
   if(ret<0)
     {ret= 0; goto ex;}
   buf[ret]= 0; 
   if(flag&(1<<12))
     Sregex_md5(&ctx,buf,ret,md5,0);
   else 
     Sregex_crc32_yenc(buf,ret,crc32,1);
#endif /* ! Stypes_on_ms_windowS */
 } if(typ==2) {                                                /* directory */
   ret= Dirseq_new(&dirseq,adr,0);
   if(ret<=0)
     {ret= 0; goto ex;}
   buf[0]= '\n'; /* line break at begin of each directory item */
   while(1) {
     ret= Dirseq_next_adr(dirseq,buf+1,0);
     if(ret<0)
       goto ex;
     if(ret==0)
   break;   
     extra_ret= 1;
     if(flag&(1<<12))
       Sregex_md5(&ctx,buf,ret,md5,0);
     else 
       Sregex_crc32_yenc(buf,strlen(buf),crc32,first);
     first= 0;
   }
 } else if(typ==1 || typ==5 || (typ==6 && !(flag&32)) ) {
                        /* regular file , hardlink , block device content */
   Intvec_from_double(countvec,vec_size,0.0,0);
   if(flag&1)
     fp= fp_in;
   else
     fp= fopen(adr,"rb");
   if(fp==NULL)
     {ret= 0; goto ex;}

   if(flag&128) {
#ifdef Cd_backup_planer_fseekO
     off_t Intvec_to_off_t();
     ret= Cd_backup_planer_fseekO(fp,Intvec_to_off_t(skipvec,vec_size,0),
                                  SEEK_SET);
#else
     double Intvec_to_double();
     ret= fseek(fp,(long) Intvec_to_double(skipvec,vec_size,0),SEEK_SET);
#endif /* Cd_backup_planer_fseekO */
   }

   bufsiz= sizeof(buf);
   while(1) {
     if((flag&4) || make_pre_crc) {
       Intvec_add(hvec,vec_size,limitvec,countvec,1); /* limitvec - countvec */
       todo= Intvec_to_double(hvec,vec_size,0);
       if(todo<=0)
   break;
       if(todo<bufsiz)
         bufsiz= todo; 
     }
     ret= fread(buf,1,bufsiz,fp);
     if(ret<=0)
   break;

     if((flag&3)==3) {
       /* write to the pipe broker (idea : A30407_0208_marcus_rocha) */
       Stdout_pipe_broker(buf,ret,0);
     }

     Intvec_add_b30(countvec,vec_size,ret,0);
     if(make_pre_crc && Intvec_cmp(countvec,vec_size,limitvec,0)>=0) {
       /* preview checksum for a fixed size start piece */
       read_bytes= ret;
       Intvec_add(hvec,vec_size,countvec,limitvec,1); /* countvec - limitvec */
       tomuch= Intvec_to_double(hvec,vec_size,0);
       cut= ret-tomuch;
       if(flag&(1<<12)) {
         /* compute */
         Sregex_md5(&ctx,buf,cut,md5,0);
         /* obtain intermediate value */
         ret= Sregex_md5(&pre_ctx,ctx,0,pre_md5,1|4);
         if(ret<=0)
           goto ex;
         Sregex_md5(&pre_ctx,buf,0,pre_md5,2|(1<<15));
       } else {
         Sregex_crc32_yenc(buf,ret,crc32,first);
         /* obtain intermediate value */
         *(pre_crc32)= *(crc32);
       }
       first= 0;
       /* shift remaining buffer bytes */
       ret= read_bytes;
       for(i=cut;i<ret;i++)
         buf[i-cut]= buf[i];
       ret-= cut;
       make_pre_crc= 0;
       bufsiz= sizeof(buf);
   continue;
     }
     if(flag&(1<<12))
       Sregex_md5(&ctx,buf,ret,md5,0);
     else
       Sregex_crc32_yenc(buf,ret,crc32,first);
     first= 0;
     pacifier_cycles++;
     if(pacifier_cycles>=pacifier_cycle_threshhold) {
       /* try to make visible or update a waiting pacifier */
       if(flag&64) {
         r= Intvec_to_double(countvec,vec_size,0);
         Sfile_scale(r,size_text,4,1.0e4,1);
         strcat(size_text," ");
       }
       Print_pacifier(0,size_text,128|((flag&64)<<2));
       pacifier_cycles= 0;
     }
   }
   if(!(flag&1))
     fclose(fp);
   *size= Intvec_to_double(countvec,vec_size,0);
   if(ivec!=NULL && vec_size>0) {
     Intvec_from_double(ivec,vec_size,0.0,0);
     Intvec_copy(countvec,vec_size,ivec,0);
   }
  
 } else if (typ==6) { /* && !(flag&32) */        /* block device properties */
   if(flag&1)
     ret= fstat(fileno(fp_in),&stbuf);
   else
     ret= stat(adr,&stbuf);
   if(ret==-1)
     {ret= 0; goto ex;}
   sprintf(buf,"%lu %lu %lu %lu %lu %lu ",
           (unsigned long) stbuf.st_dev,(unsigned long) stbuf.st_ino,
           (unsigned long) stbuf.st_mode,(unsigned long) stbuf.st_nlink,
           (unsigned long) stbuf.st_uid,(unsigned long) stbuf.st_gid);
   sprintf(buf+strlen(buf),"%lu %lu %lu %lu %lu %lu",
           (unsigned long) stbuf.st_rdev,(unsigned long) stbuf.st_size,
           (unsigned long) stbuf.st_blksize,(unsigned long) stbuf.st_blocks,
           (unsigned long) stbuf.st_mtime,(unsigned long) stbuf.st_ctime);
   if(flag&(1<<12))
     Sregex_md5(&ctx,buf,strlen(buf),md5,0);
   else
     Sregex_crc32_yenc(buf,strlen(buf),crc32,1);
 } else if (typ==4) {                                         /* named pipe */
   /* >>> what to do rather than checksum 0 ? */;
 } else {                                                     /* unknown */
   /* >>> what to do rather than checksum 0 ? */;
 }
 ret= 1;
ex:;
 if(flag&512) {
   /* do not destroy ctx but hand it over to caller */
   if(flag&(1<<12))
     *handle= ctx;
 } else {
   if(flag&(1<<12)) 
     Sregex_md5(&ctx,buf,0,md5,2|(1<<15));
 }
 Dirseq_destroy(&dirseq,0);
 if(popen_fp!=NULL)
   pclose(popen_fp);

 if(ret!=1)
   return(ret);
 return(1+extra_ret);
}


/* ------------------------------------------------------------------------ */
/* a very compact list format for stream block checksums */
struct BlksuM {
 unsigned char md5[16];
 struct BlksuM *next;
};


int Blksum_new(objpt,md5,prev,next,flag)
struct BlksuM **objpt;
unsigned char md5[16];
struct BlksuM *prev;
struct BlksuM *next;
int flag;
{
 struct BlksuM *o;
 int i;

 *objpt= o= TSOB_FELD(struct BlksuM,1);
 if(o==NULL)
   return(-1);
 for(i=0;i<16;i++)
   o->md5[i]= md5[i];
 if(prev!=NULL)
   prev->next= o;
 o->next= next;
 return(1);
}


int Blksum_destroy_all(objpt,flag)
struct BlksuM **objpt;
int flag;
{
 struct BlksuM *o,*next;

 if((*objpt)==NULL)
   return(0);
 for(o= *objpt;o!=NULL;o= next) {
   next= o->next;
   free((char *) o);
 }
 return(1);
}


int Blksum_of_stream(blockmd5,fp,blocksize,countvec,md5,flag)
struct BlksuM **blockmd5;   /* must point to NULL or freeable memory */
FILE *fp;
double blocksize;
int countvec[CdbplivS];
char md5[16];
int flag;
/*
 bit1= write data to stdout pipe broker
*/
{
 char buf[4*4096],local_md5[16];
 char *ctx= NULL, *blockctx= NULL;              /* should rather be (void *) */
 int ret,bufsiz,readsiz;
 double todo;
 struct BlksuM *blksm= NULL;

 Intvec_from_double(countvec,CdbplivS,0.0,0);
 Blksum_destroy_all(blockmd5,0);
 bufsiz= sizeof(buf);
 todo= blocksize;
 ret= Sregex_md5(&ctx,buf,0,md5,1);
 if(ret<=0)
   goto ex;
 ret= Sregex_md5(&blockctx,buf,0,local_md5,1);
 if(ret<=0)
   goto ex;

 while(1) {

   if(todo>bufsiz)
     readsiz= bufsiz;
   else
     readsiz= todo;

   ret= fread(buf,1,readsiz,fp);
   if(ret<=0)
 break;
   if(flag&2)
     Stdout_pipe_broker(buf,ret,0);
   Intvec_add_b30(countvec,CdbplivS,ret,0);
   todo-= ret;
   Sregex_md5(&ctx,buf,ret,md5,0);
   Sregex_md5(&blockctx,buf,ret,local_md5,0);
   if(todo<=0) {
     /* close and evaluate block */;
     Sregex_md5(&blockctx,buf,0,local_md5,2|(1<<15));
     /* create and add BlksuM */;
     ret= Blksum_new(&blksm,local_md5,blksm,NULL,0);
     if(ret<=0)
       goto ex;
     if((*blockmd5)==NULL)
       (*blockmd5)= blksm;
     /* open new block */
     todo= blocksize;
     ret= Sregex_md5(&blockctx,buf,0,local_md5,1);
     if(ret<=0)
       goto ex;
   }
 }

 ret= 1;
ex:;
 Sregex_md5(&ctx,buf,0,md5,2|(1<<15));
 if(todo!=blocksize) {
   /* record last incomplete block */
   Sregex_md5(&blockctx,buf,0,local_md5,2|(1<<15));
   ret= Blksum_new(&blksm,local_md5,blksm,NULL,0);
   if(ret>0 && (*blockmd5)==NULL)
     (*blockmd5)= blksm;
 }
 if(ret<=0) {
   Blksum_destroy_all(blockmd5,0);
 }
 return(ret);
}


/* ------------------------------------------------------------------------ */
/* DirseQ : crawl along a directory's content list */

static int Dirseq_buffer_sizE= 1000;

struct DirseQ {
  DIR *dirpt;
  int count;  
  char **buffer;
  int buffer_size;
  int buffer_fill;
  int buffer_rpt;
};


int Dirseq_new(o,dir,flag)
struct DirseQ **o;
char *dir;
int flag;
/*
 bit0= do not complain about failed opendir()
*/
{
 int ret,i;
 struct DirseQ *m;

 m= *o= TSOB_FELD(struct DirseQ,1);
 if(m==NULL)
   return(-1);
 m->dirpt= NULL;
 m->count= 0;
 m->buffer= NULL;
 m->buffer_size= 0;
 m->buffer_fill= 0;
 m->buffer_rpt= 0;
 m->buffer= TSOB_FELD(char *,Dirseq_buffer_sizE);
 if(m->buffer==NULL)
   {ret= -1; goto failed;}
 m->buffer_size= Dirseq_buffer_sizE;
 for(i= 0;i<m->buffer_size;i++)
   m->buffer[i]= NULL;
 if(dir[0]==0)
   m->dirpt= opendir(".");
 else
   m->dirpt= opendir(dir);
 if(m->dirpt==NULL) {
   if(!(flag&1))
     fprintf(stderr,"opendir(%s) failed : %s\n",dir,strerror(errno));
   ret= 0; goto failed;
 }
 return(1);
failed:;
 Dirseq_destroy(o,0);
 return(ret);
}


int Dirseq_destroy(o,flag)
struct DirseQ **o;
int flag;
{
 int i;

 if(*o==NULL)
   return(0);
 if((*o)->dirpt!=NULL)
   closedir((*o)->dirpt);
 if((*o)->buffer!=NULL) {
   for(i=0;i<(*o)->buffer_size;i++)
     if((*o)->buffer[i]!=NULL)
       free((*o)->buffer[i]);
   free((char *) (*o)->buffer);
 }
 free((char *) *o);
 (*o)= NULL;
 return(1);
}


int Dirseq_rewind(o,flag)
struct DirseQ *o;
int flag;
{
 rewinddir(o->dirpt);
 return(1);
}


int Dirseq_next_adr(o,reply,flag)
struct DirseQ *o;
char reply[4096];
int flag;
/*
flag:
 bit0= permission to use buffer
 bit1= do not increment counter
 bit2= ignore buffer in any case
 bit3= do not exclude '.' and '..'
 bit4= sort buffer
 bit5= sort only incomplete last buffer
return:
 <0 error
 0= no more entries available
 1= ok, reply is valid
*/
{
 int ret;
 struct dirent *entry;
 char *name;

 static int override_flag_0= 0,override_flag_1= 32;
 flag= (flag&~override_flag_0)|override_flag_1;

 if((flag&1) && o->buffer_rpt>=o->buffer_fill) {
   /* permission to buffer and buffer empty : load a buffer */
   ret= Dirseq_next_adrblock(o,o->buffer,&(o->buffer_fill),
                             o->buffer_size,2|4|(flag&16));
   if(ret<=0)
     return(ret);
   o->buffer_rpt= 0;
   if((flag&32) && o->buffer_fill<o->buffer_size && o->buffer_fill>0) 
     Sort_argv(o->buffer_fill,o->buffer,0);
 }
 if(o->buffer_rpt<o->buffer_fill && !(flag&4)) {
   strcpy(reply,o->buffer[o->buffer_rpt]);
   Sregex_string(&(o->buffer[o->buffer_rpt]),NULL,0);
   (o->buffer_rpt)++;
   if(!(flag&2))
     o->count++;
   return(1);
 }
 do {
   entry= readdir(o->dirpt);
   if(entry==NULL) {
     /* >>> how to distinguish error from EOF , do i need a (FILE *) ? */
     return(0);
   }
   if(strlen(entry->d_name)>4095) {
     fprintf(stderr,"--- oversized directory entry (number %d) :\n    %s\n",
                    o->count+1,entry->d_name);
     return(-1);
   }
   name= entry->d_name;
   if(flag&8)
 break;
   /* skip "." and ".." */
 } while(name[0]=='.' && ((name[1]=='.' && name[2]==0) || name[1]==0));
 strcpy(reply,name);
 if(!(flag&2))
   o->count++;
 return(1);
}


int Dirseq_next_adrblock(o,replies,reply_count,max_replies,flag)
struct DirseQ *o;
char *replies[];                       /* a vector of Sregex_string pointers */
int *reply_count;
int max_replies;
int flag;
/*
flag:
 bit0= permission to use buffer
 bit1= do not increment counter
 bit2= ignore buffer in any case
 bit4= sort replies
return:
 <0 error
 0= no more entries available
 1= ok, reply is valid
*/
{
 int i,ret;
 char reply[4096];

 *reply_count= 0;
 for(i=0;i<max_replies;i++) {
   ret= Dirseq_next_adr(o,reply,flag&(1|2|4));
   if(ret<0)
     return(ret);
   if(ret==0)
 break;
   if(Sregex_string(&(replies[i]),reply,0)<=0)
     return(-1);
   (*reply_count)++;
 }
 if((*reply_count)==0)
   return(0);
 if(flag&16) 
   Sort_argv(*reply_count,replies,0);
 return(1);
}


int Dir_ls(dir,filec,filev,maxv,flag)
char *dir;
int *filec;
char ***filev;
int maxv;
int flag;
/*
 bit0= skip '.' and '..'
 bit1= just release list filev and return
 bit2= redirect stderr to /dev/null
 bit3= try to display pending pacifier text from time to time
 bit4= sort result
*/
{
 int ret,count,i,pass,overflown= 0;
 char name[4096];
 struct DirseQ *o= NULL;

 Sfile_destroy_argv(filec,filev,0);
 if(flag&2)
   return(1);

 ret= Dirseq_new(&o,dir,0); 
 if(ret<=0)
   goto ex;
 for(pass=0; pass<2;pass++) {
   count= 0;
   while(1) {
     ret= Dirseq_next_adr(o,name,0);
     if(ret<0)
       goto ex;
     if(ret==0)
   break;
     if(count>=maxv && maxv>0) {
       overflown= 1;
   break;
     }
     if(flag&1)
       if(name[0]=='.' && ((name[1]=='.' && name[2]==0) || name[1]==0))
   continue;
     count++;
     if(pass==1) {
       if(count>*filec)
   break;
       ret= Sregex_string(&(*filev)[count-1],name,0);
       if(ret<=0)
         {ret= -1; goto ex;}
     }
   }
   if((flag&8) && (count%200)==0 && count>0)
     Print_pacifier(0,"",128|256|(1<<11));
   if(pass==0) {
     if(count==0)
 break; 
     *filev= TSOB_FELD(char *,count);
     if(*filev==NULL)
       {ret= -1; goto ex;}
     for(i=0;i<count;i++)
       (*filev)[i]= NULL;
     *filec= count;
   }
   Dirseq_rewind(o,0);
 }
 *filec= count;
 ret= 1+overflown;
ex:;
 if(ret>0) { 
   if((flag&16)&&(*filec)>0)
     Sort_argv(*filec,*filev,0);
 } else
   Sfile_destroy_argv(filec,filev,0);
 if(o!=NULL)
   Dirseq_destroy(&o,0);
 return(ret);
}


/* ------------------------------------------------------------------------ */

struct Blksum_lisT {
 struct CheckrecorD *rec;

 /* block list description and content */
 int blocksize[CdbplivS];
 int complete_blocks[CdbplivS];
 int remainder[CdbplivS];
 struct BlksuM *blockmd5;
 struct BlksuM *last_checksum;
 char md5_of_list[36];
 int blocklist_is_reverted;

 /* context of list generation */
 char *total_ctx;
 char *blockctx;
 int todo_in_block[CdbplivS];

 /* context of check and recovery processing */
 int readpos[CdbplivS];
 int read_sequential;
 int writepos[CdbplivS];

 struct CheckrecorD *verifier;
};


int Blksum_list_new(objpt,rec,flag)
struct Blksum_lisT **objpt;
struct CheckrecorD *rec;
int flag;
{
 struct Blksum_lisT *o;
 int i;

 *objpt= o= TSOB_FELD(struct Blksum_lisT,1);
 if(o==NULL)
   return(-1);
 o->rec= rec;
 Intvec_from_double(o->blocksize,CdbplivS,0.0,0);
 Intvec_from_double(o->complete_blocks,CdbplivS,0.0,0);
 Intvec_from_double(o->remainder,CdbplivS,0.0,0);
 o->blockmd5= NULL;
 o->last_checksum= NULL;
 o->total_ctx= NULL;
 o->blockctx= NULL;
 Intvec_from_double(o->todo_in_block,CdbplivS,0.0,0);
 for(i=0;i<sizeof(o->md5_of_list);i++)
   o->md5_of_list[i]= 0;
 o->blocklist_is_reverted= 0;
 Intvec_from_double(o->readpos,CdbplivS,0.0,0);
 Intvec_from_double(o->writepos,CdbplivS,0.0,0);
 o->read_sequential= 0;
 return(1);
}


int Blksum_list_destroy(objpt,flag)
struct Blksum_lisT **objpt;
int flag;
{
 struct Blksum_lisT *o;

 o= *objpt;
 if(o==NULL)
   return(0);

 Blksum_destroy_all(&(o->blockmd5),0);
 o->last_checksum= NULL;
 Sregex_md5(&(o->total_ctx),NULL,0,NULL,1<<15);
 Sregex_md5(&(o->blockctx),NULL,0,NULL,1<<15);

 free((char *) o);
 *objpt= NULL;
 return(1);
}


int Blksum_list_set_rec(o,rec,flag)
struct Blksum_lisT *o;
struct CheckrecorD *rec;
int flag;
{
 o->rec= rec;
 return(1);
}


int Blksum_list_set_sizes(o,blocksize,complete_blocks,remainder,flag)
struct Blksum_lisT *o;
int blocksize[CdbplivS];
int complete_blocks[CdbplivS];
int remainder[CdbplivS];
int flag;
/*
 bit0= set blocksize
 bit1= set complete_blocks
 bit2= set remainder
*/
{
 if(flag&1)
   Intvec_copy(blocksize,CdbplivS,o->blocksize,0);
 if(flag&2)
   Intvec_copy(complete_blocks,CdbplivS,o->complete_blocks,0);
 if(flag&4)
   Intvec_copy(remainder,CdbplivS,o->remainder,0);
 return(1);
}


int Blksum_list_get_sizes(o,blocksize,complete_blocks,remainder,flag)
struct Blksum_lisT *o;
int blocksize[CdbplivS];
int complete_blocks[CdbplivS];
int remainder[CdbplivS];
int flag;
/*
 bit0= get blocksize
 bit1= get complete_blocks
 bit2= get remainder
*/
{
 if(flag&1)
   Intvec_copy(o->blocksize,CdbplivS,blocksize,0);
 if(flag&2)
   Intvec_copy(o->complete_blocks,CdbplivS,complete_blocks,0);
 if(flag&4)
   Intvec_copy(o->remainder,CdbplivS,remainder,0);
 return(1);
}


int Blksum_list_add(o,md5,flag)
struct Blksum_lisT *o;
char md5[16];
int flag;
/*
 bit0= add new item
*/
{
 int ret;

 ret= Blksum_new(&(o->last_checksum),md5,o->last_checksum,0);
 if(ret<=0)
   return(ret);
 if(o->blockmd5==NULL)
   o->blockmd5= o->last_checksum;
 return(1);
}


int Blksum_list_set_listmd5(o,md5,flag)
struct Blksum_lisT *o;
char md5[36];
int flag;
{
 int i;

 for(i=0;i<32;i++) {
   if(!isxdigit(md5[i]))
     return(0);
   o->md5_of_list[i]= md5[i];
 }
 return(1);
}


int Blksum_list_from_data(objpt,data,datacount,blocksize,countvec,flag)
struct Blksum_lisT **objpt;   /* must point to NULL or freeable memory */
char *data;
int datacount;
int blocksize[CdbplivS];
int countvec[CdbplivS];
int flag;
/*
 bit0= init 
 bit1= finish 
 bit2= do not destroy *objpt on failure
 bit3= return 1 if *objpt==NULL at data processing time
*/
{
 unsigned char local_md5[16];
 int ret,readsize,datadone= 0,hret;
 int hvec[CdbplivS];
 double todo;
 struct BlksuM *blksm= NULL;
 struct Blksum_lisT *o;

 if(flag&1) {
   Intvec_from_double(countvec,CdbplivS,0.0,0);
   Blksum_list_destroy(objpt,0);
   ret= Blksum_list_new(objpt,NULL,0);
   if(ret<=0)
     goto ex;
   o= *objpt;
   Intvec_copy(blocksize,CdbplivS,o->blocksize,0);
   Intvec_copy(o->blocksize,CdbplivS,o->todo_in_block,0);
   ret= Sregex_md5(&(o->total_ctx),NULL,0,NULL,1);
   if(ret<=0)
     goto ex;
   ret= Sregex_md5(&(o->blockctx),NULL,0,NULL,1);
   if(ret<=0)
     goto ex;
 }
 o= *objpt;
 if(o==NULL)
   {ret= !!(flag&8); goto ex;}

 while(datacount-datadone>0) {
   /* determine number of bytes to process */
   Intvec_from_double(hvec,CdbplivS,(double) (datacount-datadone),0);
   if(Intvec_cmp(o->todo_in_block,CdbplivS,hvec,0)>0)
     todo= datacount-datadone;
   else 
     todo= (int) Intvec_to_double(o->todo_in_block,CdbplivS,0);
   if(todo>=Intvec_b30)
     todo= Intvec_b30-1;
   readsize= todo;
   /* process */
   Sregex_md5(&(o->total_ctx),data+datadone,readsize,NULL,0);
   Sregex_md5(&(o->blockctx),data+datadone,readsize,NULL,0);
   /* count data as processed */
   datadone+= readsize;
   Intvec_add_b30(countvec,CdbplivS,readsize,0);
   Intvec_add_b30(o->todo_in_block,CdbplivS,readsize,1);
   if(Intvec_sign(o->todo_in_block,CdbplivS,0)<=0) {
     /* close and evaluate block */;
     Sregex_md5(&(o->blockctx),NULL,0,local_md5,2|(1<<15));
     /* create and add BlksuM */;
     ret= Blksum_new(&blksm,local_md5,o->last_checksum,NULL,0);
     if(ret<=0)
       goto ex;
     if((o->blockmd5)==NULL)
       (o->blockmd5)= blksm;
     o->last_checksum= blksm;
     Intvec_add_b30(o->complete_blocks,CdbplivS,1,0);
     /* open new block */
     Intvec_copy(o->blocksize,CdbplivS,o->todo_in_block,0);
     ret= Sregex_md5(&(o->blockctx),NULL,0,NULL,1);
     if(ret<=0)
       goto ex;
   }
 }

 ret= 1;
ex:;
 if(ret>0 && (flag&2)) {
   /* finalize : compute remainder and record last incomplete block */
   Intvec_add(o->remainder,CdbplivS,o->blocksize,o->todo_in_block,1);
   Sregex_md5(&(o->blockctx),NULL,0,local_md5,2|(1<<15));
   if(Intvec_sign(o->remainder,CdbplivS,0)>0) {
     hret= Blksum_new(&blksm,local_md5,o->last_checksum,NULL,0);
     if(hret<=0)
       {ret= -1; goto ex;}
     if((o->blockmd5)==NULL)
       (o->blockmd5)= blksm;
     o->last_checksum= blksm;
   }
 }
 if(ret<=0) 
   if(!(flag&4))
     Blksum_list_destroy(objpt,0);
 return(ret);
}


int Blksum_list_of_stream(objpt,fp,blocksize,countvec,md5,flag)
struct Blksum_lisT **objpt;   /* must point to NULL or freeable memory */
FILE *fp;
double blocksize;
int countvec[CdbplivS];
char md5[16];
int flag;
/*
 bit1= write data to stdout pipe broker
*/
{
 int ret;
 struct BlksuM *blockmd5= NULL;
 char buf[4*4096];
 int blocksizevec[CdbplivS];
 char *dummy_ctx= NULL;

 Intvec_from_double(blocksizevec,CdbplivS,blocksize,0);
 Intvec_from_double(countvec,CdbplivS,0.0,0);
 ret= Blksum_list_from_data(objpt,buf,0,blocksizevec,countvec,1); /* init */
 if(ret<=0)
   goto ex;
 while(1) {
   ret= fread(buf,1,sizeof(buf),fp);
   if(ret<=0)
 break;
   if(flag&2)
     Stdout_pipe_broker(buf,ret,0);
   ret= Blksum_list_from_data(objpt,buf,ret,blocksizevec,countvec,0);
   if(ret<=0)
     goto ex;
 }
 ret= Blksum_list_from_data(objpt,buf,0,blocksizevec,countvec,2);
 if(ret<=0)
   goto ex;
 ret= Sregex_md5(&dummy_ctx,(*objpt)->total_ctx,0,md5,1|2|4|(1<<15));
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 Blksum_destroy_all(&blockmd5,0);
 return(ret);
}


int Blksum_list_write(o,fp,countvec,verifier_bl,flag)
struct Blksum_lisT *o;
FILE *fp;
int countvec[CdbplivS];
struct Blksum_lisT **verifier_bl;
int flag;
/*
 bit0= do not produce verifier_bl checksum list
 bit1= transfer countvec to local_countvec before writing
*/
{
 char text[4096],size_text[161];
 unsigned char md5[16];
 int hvec[CdbplivS];
 int local_countvec[CdbplivS],posvec[CdbplivS];
 int ret,i,l,append_local_list= 0;
 struct BlksuM *blksm;
 char *ctx= NULL;                               /* should rather be (void *) */
 struct CheckrecorD *rec;

 if(verifier_bl==NULL)
   flag|= 1;
 ret= Sregex_md5(&ctx,md5,0,md5,1);
 if(ret<=0)
   goto ex;
 Intvec_from_double(local_countvec,CdbplivS,0.0,0);
 if(flag&2) {
   Intvec_add(local_countvec,CdbplivS,local_countvec,countvec,0);
   Intvec_from_double(countvec,CdbplivS,0.0,0);
 } 
 sprintf(text,"%s",Checkrecord_blocklist_starT);
 Intvec_to_dec(o->blocksize,CdbplivS,size_text,sizeof(size_text),0);
 sprintf(text+strlen(text),"blocksize=%s\n",size_text);
 Intvec_to_dec(o->complete_blocks,CdbplivS,size_text,sizeof(size_text),0);
 sprintf(text+strlen(text),"complete_blocks=%s\n",size_text);
 Intvec_to_dec(o->remainder,CdbplivS,size_text,sizeof(size_text),0);
 sprintf(text+strlen(text),"remainder=%s\n",size_text);
 ret= Write_broker_or_file(fp,text,strlen(text),0);
 if(ret<=0)
   goto ex;

 Sregex_md5(&ctx,text,strlen(text),NULL,0);
 if(flag&1)
   Intvec_add_b30(countvec,CdbplivS,strlen(text),0);
 else {
   ret= Blksum_list_from_data(verifier_bl,text,strlen(text),o->blocksize,
                              local_countvec,8);
   if(ret<=0)
     goto ex;
 }

 for(blksm= o->blockmd5;blksm!=NULL;blksm= blksm->next) {
 ret= Write_broker_or_file(fp,blksm->md5,16,0);
 if(ret<=0)
   goto ex;
   Sregex_md5(&ctx,blksm->md5,16,NULL,0);
   if(flag&1)
     Intvec_add_b30(countvec,CdbplivS,16,0);
   else {
     ret= Blksum_list_from_data(verifier_bl,blksm->md5,16,o->blocksize,
                                local_countvec,8);
     if(ret<=0)
       goto ex;
   }
 }
 Sregex_md5(&ctx,md5,0,md5,2);
 sprintf(text,"\nmd5_of_list=");
 for(i=0;i<16;i++)
   sprintf(text+strlen(text),"%2.2x",md5[i]);
 Intvec_from_double(hvec,CdbplivS,1024.0,0);
 if(Intvec_cmp((*verifier_bl)->blocksize,CdbplivS,hvec,0)>=0)
   if((*verifier_bl)!=NULL) 
     if((*verifier_bl)->blockmd5!=NULL) 
       /* blocksize >= 1k and at least one complete block emerged up to now */
       append_local_list= 1;
 l= strlen(text);
 if(append_local_list) {
   sprintf(text+l,"\nchained tag follows\n");
 } else {
   sprintf(text+l,"\nend of list\n");
 }
 ret= Write_broker_or_file(fp,text,strlen(text),0);
 if(ret<=0)
   goto ex;

 if(append_local_list) {
   ret= Blksum_list_from_data(verifier_bl,text,strlen(text),
                              o->blocksize,local_countvec,2);
   if(ret<=0)
     goto ex;
 } else {
   if(flag&1)
     Intvec_add_b30(countvec,CdbplivS,strlen(text),0);
   else
     Intvec_add_b30(local_countvec,CdbplivS,strlen(text),0);
 }

 if(append_local_list) {
   Sregex_md5(&((*verifier_bl)->total_ctx),NULL,0,md5,2|(1<<15));
   ret= Checkrecord_new(&rec,"blocklist",time(0),0.0,local_countvec,
                        1,(unsigned long) 0,md5,0);
   if(ret<=0)
     goto ex;
   /* write tag and blocklist */;
   Intvec_copy(local_countvec,CdbplivS,posvec,0);
   Intvec_from_double(hvec,CdbplivS,0.0,0);
   ret= Blksum_list_to_fp((*verifier_bl),fp,rec,posvec,hvec,0);
   if(ret<=0)
     goto ex;
   Intvec_add(local_countvec,CdbplivS,local_countvec,hvec,0);
 }
 Intvec_add(countvec,CdbplivS,countvec,local_countvec,0);
 ret= 1;
ex:;
 Sregex_md5(&ctx,md5,0,md5,(1<<15));
 return(ret); 
}


int Blksum_list_to_fp(o,fp,rec,posvec,countvec,flag)
struct Blksum_lisT *o;
FILE *fp;
struct CheckrecorD *rec;
int posvec[CdbplivS];
int countvec[CdbplivS];
int flag;
{
 int ret;
 int min_dutyvec[CdbplivS];

 Intvec_from_double(min_dutyvec,CdbplivS,1.0,0);
 Checkrecord_set_blocklist(rec,o,0);
 ret= Checkrecord_write_tags(rec,fp,posvec,min_dutyvec,countvec,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Blksum_list_to_file(o,adr,rec,flag)
struct Blksum_lisT *o;
char *adr;
struct CheckrecorD *rec;
int flag;
/*
 bit0= do not append record information to adr
 bit1= report file address on stderr
*/
{
 int ret,i;
 int posvec[CdbplivS],countvec[CdbplivS];
 char fileadr[4096],record_text[161],*cpt;
 FILE *fp= NULL;
 FILE *Sfile_fopen();

 if(flag&1)
   strcpy(fileadr,adr);
 else {
   Checkrecord_to_text(rec,record_text,0);
   cpt= record_text;
   for(i= 0;i<3;i++) {
     cpt= strchr(cpt,' ');
     if(cpt==NULL)
   break;
     if(i<2)
       *cpt= '_';
     else
       *cpt= 0;
   }
   sprintf(fileadr,"%s%s",adr,record_text);
 } 
 fp=  Sfile_fopen(fileadr,"wb",0);
 if(fp==NULL)
   {ret= 0; goto ex;}
 Intvec_from_double(posvec,CdbplivS,0.0,0);
 Intvec_from_double(countvec,CdbplivS,0.0,0);
 ret= Blksum_list_to_fp(o,fp,rec,posvec,countvec,0);
 if(ret<=0)
   goto ex;
 if(flag&2)
   fprintf(stderr,"\nNote: blocklist file written : %s\n",fileadr);
 ret= 1;
ex:;
 Sfile_fclose(fp);
 return(ret);
}


int Blksum_list_read_block(o,blockno,fp,buf,flag)
/* works only for blocksizes up to 2 GB -1 */
struct Blksum_lisT *o;
int blockno[CdbplivS];
FILE *fp;
char *buf;
int flag;
/*
 bit0= random access : do not use or change o->readpos
 bit1= expect read error, do not complain
 bit2= sequential read : do not use or change o->readpos
 bit3= in case of error provide a zeroed block and return success
*/
/*
 return:  <=0 : error , >0 : number of bytes in block (successfully read)
*/
{
 int todo,ret,i;
 char size_text[161];
 double r,blksize,start_byte;

 /* determine number of bytes to read */;
 ret= Intvec_cmp(o->complete_blocks,CdbplivS,blockno,0);
 if(ret>0)
   r= Intvec_to_double(o->blocksize,CdbplivS,0);
 else if(ret==0)
   r= Intvec_to_double(o->remainder,CdbplivS,0);
 todo= r;

 if(!(flag&4))
   if((flag&1) || Intvec_cmp(blockno,CdbplivS,o->readpos,0)!=0) {
     blksize= Intvec_to_double(o->blocksize,CdbplivS,0);
     start_byte= Intvec_to_double(blockno,CdbplivS,0)*(double) blksize;
     ret= Fseek_double(fp,start_byte,0);
     if(ret<0)
       {ret= 0; goto ex;}
   }
 ret= fread(buf,todo,1,fp);
 if(ret!=1) {
   if(flag&2)
     {ret= 0; goto ex;}
   Intvec_to_dec(blockno,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"While reading block %s ",size_text);
   Intvec_to_dec(o->complete_blocks,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"of %s :\n",size_text);
   if(feof(fp)) {
     fprintf(stderr,"Premature EOF\n");
   } else {
     fprintf(stderr,"Error %d : %s\n",errno,strerror(errno));
   }  
   ret= 0; goto ex;
 }

 if(!(flag&(1|4)))
   Intvec_add_b30(o->readpos,CdbplivS,1,0);
 ret= todo;
ex:;
 if(ret<=0 && (flag&8)) {
   for(i=0;i<todo;i++)
     buf[i]= 0;
   ret= todo;
 }
 return(ret);
}


int Blksum_list_write_block(o,blockno,fp,target_adr,buf,bufcount,flag)
/* works only for blocksizes up to 2 GB -1 */
struct Blksum_lisT *o;
int blockno[CdbplivS];
FILE *fp;
char *target_adr;
char *buf;
int bufcount;
int flag;
/*
 bit0= position write pointer in any case 
*/
{
 int ret;
 double start_byte,blksize;

 if((flag&1) || Intvec_cmp(blockno,CdbplivS,o->writepos,0)!=0) {
   blksize= Intvec_to_double(o->blocksize,CdbplivS,0);
   start_byte= Intvec_to_double(blockno,CdbplivS,0)*(double) blksize;
   ret= Fseek_double(fp,start_byte,0);
   if(ret<0)
     return(0);
 }
 ret= fwrite(buf,bufcount,1,fp);
 if(ret<=0) {
   Print_pacifier(0,"",1|64);
   fprintf(stderr,
           "While trying to write data for -process_blocklist to : %s\n",
           target_adr);
   perror("");
   return(0);
 }
 return(1);
}


int Blksum_list_revert_chain(o,flag)
struct Blksum_lisT *o;
int flag;
{
 struct BlksuM *block,*nblock,*pblock= NULL;

 block= o->blockmd5;
 if(block!=NULL) {
   while(block->next!=NULL) {
     nblock= block->next;
     block->next= pblock;
     pblock= block;
     block= nblock;
   }
   block->next= pblock;
 }
 o->blockmd5= block;
 o->blocklist_is_reverted= !o->blocklist_is_reverted;
 return(1+o->blocklist_is_reverted);
}


int Blksum_list_process(o,source,target,startblock,endblock,flag)
struct Blksum_lisT *o;
char *source;
char *target;
int startblock[CdbplivS];
int endblock[CdbplivS];
int flag;
/*
 bit0-7: mode
   0= report  : compare blocklist with source and print mismatches to stdout
   1= collect : read source in single pass and write matching blocks to target
   2= mend    : read target and replace mismatching blocks by matching ones
                from source
   3= merge   : read source and target in a single pass and try to mend
                errors in target by blocks from source. Output is stdout.
 bit8= process interval [startblock,endblock] (inlusive endblock)
 bit9= return number of next block to read in endblock
 bit10= return number of mismatches in startblock
*/
{
 int ret,mode= -1,mismatch,blksize,i,pacifier_count= 0,actually_read;
 int last_block_was_match,demand_random_target= 0,dummy,backward= 0;
 int source_is_dead= 0, reserve_is_dead= 0,reserve_had_error= 0;
 int real_interval= 0;
 int blockno[CdbplivS],stopmark[CdbplivS],mismatch_count[CdbplivS];
 int hvec[CdbplivS],block_count[CdbplivS],mend_count[CdbplivS];
 int min_dutyvec[CdbplivS],countvec[CdbplivS],missing[CdbplivS];
 char *blkbuf= NULL,size_text[160],pacifier_msg[160+20];
 double size,start_byte;
 FILE *fps= NULL,*fpt= NULL,*fp_out;
 FILE *Sfile_fopen();
 struct BlksuM *block;
 char *ctx= NULL;
 unsigned char md5[16];

 pacifier_msg[0]= 0;
 Intvec_from_double(blockno,CdbplivS,0.0,0);
 Intvec_from_double(mismatch_count,CdbplivS,0.0,0);
 Intvec_from_double(mend_count,CdbplivS,0.0,0);
 Intvec_from_double(block_count,CdbplivS,0.0,0);
 fprintf(stderr,"Processing block checksum list :\n");
 Intvec_to_dec(o->blocksize,CdbplivS,size_text,sizeof(size_text),0);
 fprintf(stderr," blocksize= %s  ,",size_text);
 Intvec_to_dec(o->complete_blocks,CdbplivS,size_text,sizeof(size_text),0);
 fprintf(stderr," complete_blocks= %s  ,",size_text);
 Intvec_to_dec(o->remainder,CdbplivS,size_text,sizeof(size_text),0);
 fprintf(stderr," remainder= %s\n",size_text);
 block= o->blockmd5;
 Intvec_copy(o->complete_blocks,CdbplivS,stopmark,0);
 if(Intvec_sign(o->remainder,CdbplivS,0)>0)
   Intvec_add_b30(stopmark,CdbplivS,1,0);
 if(flag&256) {
   if(Intvec_sign(startblock,CdbplivS,0)<0) {
     Intvec_copy(stopmark,CdbplivS,startblock,0);
     Intvec_add_b30(startblock,CdbplivS,1,1);
   }
   if(Intvec_sign(endblock,CdbplivS,0)<0) {
     Intvec_copy(stopmark,CdbplivS,endblock,0);
     Intvec_add_b30(endblock,CdbplivS,1,1);
   }
   Intvec_add_b30(stopmark,CdbplivS,1,1);
   if(Intvec_cmp(startblock,CdbplivS,stopmark,0)>0 ||
      Intvec_cmp(endblock,CdbplivS,stopmark,0)>0) {
     fprintf(stderr,"Error: given block interval out of range :\n");
     Intvec_to_dec(startblock,CdbplivS,size_text,sizeof(size_text),0);
     fprintf(stderr," startblock= %s  ,",size_text);
     Intvec_to_dec(endblock,CdbplivS,size_text,sizeof(size_text),0);
     fprintf(stderr," endblock= %s\n",size_text);
     {ret= -1; goto ex;}
   }
   /* check wether this is really not the full interval */;
   real_interval= 0; 
   if(Intvec_sign(startblock,CdbplivS,0)!=0 &&
      Intvec_sign(endblock,CdbplivS,0)!=0)
     real_interval= 1;
   if(Intvec_cmp(startblock,CdbplivS,stopmark,0)!=0 &&
      Intvec_cmp(endblock,CdbplivS,stopmark,0)!=0)  
     real_interval= 1;
   Intvec_copy(startblock,CdbplivS,blockno,0);
   Intvec_copy(endblock,CdbplivS,stopmark,0);
   Intvec_add_b30(stopmark,CdbplivS,1,0);

   if(o->blocklist_is_reverted)
     Blksum_list_revert_chain(o,0);
   for(Intvec_copy(startblock,CdbplivS,hvec,0);
       Intvec_sign(hvec,CdbplivS,0)>0;
       Intvec_add_b30(hvec,CdbplivS,1,1)) {
     block= block->next;
     if(block==NULL) {
no_block:;
       fprintf(stderr,"Data format error : block item NULL\n");
       {ret= -1; goto ex;}
     }
   }
   if(Intvec_cmp(startblock,CdbplivS,endblock,0)>0)
     backward= 1;
   
   fprintf(stderr," note: will process block interval ");
   Intvec_to_dec(startblock,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"%s to ",size_text);
   Intvec_to_dec(endblock,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"%s\n",size_text);
   if(backward)
     fprintf(stderr," note: will process blocks in backward sequence\n");
 }
 if(backward^o->blocklist_is_reverted) 
   Blksum_list_revert_chain(o,0);
 mode= (flag&255);
 if(mode==0) {
   fprintf(stderr,"Numbers of mismatching blocks will be printed to stdout. '@' is end mark.\n");
 } else if(mode==1 || mode==2) {
   if(mode==2) {
     if(strcmp(source,"-")==0) {
       fprintf(stderr,
           "Error: This action mode demands a random access file as source\n");
       {ret= -1; goto ex;}
     }
   }
   demand_random_target= 1;
   if(strcmp(target,"-")==0) {
no_random_target:;
     fprintf(stderr,
           "Error: This action mode demands a random access file as target\n");
     {ret= -1; goto ex;}
   }
   if(mode==1)
     fprintf(stderr,"Matching blocks will be written to file %s\n",target);
   if(mode==2)
     fprintf(stderr,"Nonmatching blocks of %s will be fetched from %s\n",
                    target,source);
 } else if(mode==3) {
   fprintf(stderr,
           "The merged stream of %s and %s will be written to stdout.\n",
           source,target);
 }
 if(mode==1 || mode==2) {
   /* make sure target exists before source and target get opened */
   ret= Sfile_type(target,1);
   if(ret==-1) {
     /* create file since "r" cannot do that */
     fpt= Sfile_fopen(target,"wb");
     if(fpt==NULL)
       {ret= 0; goto ex;}
     Sfile_fclose(fpt);
   } else if(ret!=1 && ret!=3 && ret!=6)
     goto no_random_target;
 }
 fps= Sfile_fopen(source,"rb",0);
 if(fps==NULL)
   {ret= 0; goto ex;}
 o->read_sequential= 0;
 if(strcmp(source,"-")==0)
   o->read_sequential= 1;
 if(mode==1 || mode==2) {
   fpt= Sfile_fopen(target,"r+b",0);
   if(fpt==NULL)
     {ret= 0; goto ex;}
   fp_out= fpt;
 } else if(mode==3) {
   fpt= Sfile_fopen(target,"rb",0);
   if(fpt==NULL)
     {ret= 0; goto ex;}
   fp_out= stdout;
 } else
   fp_out= NULL;

 /* allocate block buffer */
 size= Intvec_to_double(o->blocksize,CdbplivS,0);
 if(size>(1<<30)) {
   fprintf(stderr,"Blocksize over 1 GB refused\n");
   {ret= 0; goto ex;}
 }
 blksize= size;
 blkbuf= TSOB_FELD(char,blksize);
 if(blkbuf==NULL) {
   Intvec_to_dec(o->blocksize,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"Not enough memory for blocksize=%s\n",size_text);
   {ret= -1; goto ex;}
 }
 mismatch= -1;
 for(;Intvec_cmp(blockno,CdbplivS,stopmark,0)!=0;
      Intvec_add_b30(blockno,CdbplivS,1,backward)) {
   last_block_was_match= !mismatch;
   if(block==NULL)
     goto no_block;
   if(mode==2) {
     /* read block from target (always random access) */;
     actually_read= Blksum_list_read_block(o,blockno,fpt,blkbuf,1|2);
     if(actually_read<0)
       {ret= actually_read; goto ex;}
     if(actually_read<=0)
       mismatch= 1;
   } else {
     /* read block from source (sequential if possible) */;
     if(source_is_dead)
       actually_read= 0;
     else
       actually_read= Blksum_list_read_block(o,blockno,fps,blkbuf,
                                             ((mode==3)*(2|4)));
     if(actually_read<=0) {
       if(mode==0 || mode==1 || reserve_is_dead || actually_read<0)
         {ret= actually_read; goto ex;}
       mismatch= 1;
       source_is_dead= 1;
     }
   }

   pacifier_count++;
   if((pacifier_count%10)==0) {
     Intvec_copy(block_count,CdbplivS,hvec,0);
     Intvec_add_b30(hvec,CdbplivS,1,0);
     Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
     sprintf(pacifier_msg,"  %s blocks",size_text);
     Intvec_to_dec(mismatch_count,CdbplivS,size_text,sizeof(size_text),0);
     sprintf(pacifier_msg+strlen(pacifier_msg)," |  %s bad   ",size_text);
     Print_pacifier(0,pacifier_msg,4|32);
     pacifier_count= 0;
   }
   if(actually_read<=0)
     goto process_block;

   ret= Sregex_md5(&ctx,blkbuf,actually_read,md5,1|2|(1<<15));
   if(ret<=0)
     goto ex;
   /* compare with block */;
   mismatch= 0;
   for(i=0;i<16;i++)
     if(md5[i]!=block->md5[i])
       mismatch= 1;
process_block:;
   if(mode==0) {
     if(mismatch) {
       /* remove eventual pacifier line from screen. half line is enough */
       if(pacifier_msg[0]!=0)
         Print_pacifier(0,pacifier_msg,1|64);
       /* report to stdout */;
       Intvec_to_dec(blockno,CdbplivS,size_text,sizeof(size_text),0);
       strcat(size_text,"\n");
       Stdout_pipe_broker(size_text,strlen(size_text),0);
       /* restore pacifier line on screen */
       if(pacifier_msg[0]!=0)
         Print_pacifier(0,pacifier_msg,1|4|32);
     }
   } else if (mode==1) {
     if(!mismatch) {
       ret= Blksum_list_write_block(o,blockno,fpt,target,blkbuf,actually_read,
                                    !last_block_was_match);
       if(ret<=0)
         {ret= 0; goto ex;}
     }
   } else if (mode==2 || mode==3) {
     if(mismatch) {
       Intvec_add_b30(mend_count,CdbplivS,1,0);
       if(mode==3) {
         /* read block from target (sequential if possible) */;
         if(reserve_is_dead)
           actually_read= 0;
         else
           actually_read= Blksum_list_read_block(o,blockno,fpt,blkbuf,2|4);
         if(actually_read<=0) {
           reserve_is_dead= 1;
           if(source_is_dead || actually_read<0)
             {ret= actually_read; goto ex;}
         }
       } else {
         /* read block from source (random access) */;
         actually_read= Blksum_list_read_block(o,blockno,fps,blkbuf,
                                               1|(reserve_had_error<<1));
         if(actually_read<=0)
           reserve_had_error= 1;
         if(actually_read<0)
           {ret= actually_read; goto ex;}
       }
       ret= Sregex_md5(&ctx,blkbuf,actually_read,md5,1|2|(1<<15));
       if(ret<=0)
         goto ex;
       mismatch= 0;
       /* compare with list checksum */;
       for(i=0;i<16;i++)
         if(md5[i]!=block->md5[i])
           mismatch= 1;
       if(mode==3) {
         Stdout_pipe_broker(blkbuf,actually_read,0);
       } else if(!mismatch) {
         ret= Blksum_list_write_block(o,blockno,fp_out,target,
                                      blkbuf,actually_read,1);
         if(ret<=0)
           {ret= 0; goto ex;}
       }
     } else if (mode==3) {
       Stdout_pipe_broker(blkbuf,actually_read,0);
       /* consume unneeded target block */;
       if(reserve_is_dead)
         actually_read= 0;
       else
         actually_read= Blksum_list_read_block(o,blockno,fpt,blkbuf,2|4|8);
       if(actually_read<=0)
         if(source_is_dead || actually_read<0)
         {ret= actually_read; goto ex;}
       ret= Sregex_md5(&ctx,blkbuf,actually_read,md5,1|2|(1<<15));
       if(ret<=0)
         goto ex;
       mismatch= 0;
       /* compare with list checksum */;
       for(i=0;i<16;i++)
         if(md5[i]!=block->md5[i])
           mismatch= 1;
       if(mismatch)
         Intvec_add_b30(mend_count,CdbplivS,1,0);
       mismatch= 0;
     }
   }
   if(mismatch)
     Intvec_add_b30(mismatch_count,CdbplivS,1,0);
   Intvec_add_b30(block_count,CdbplivS,1,0);
   block= block->next;
 }
 if(pacifier_msg[0]!=0)
   Print_pacifier(0,"",1|8|16);
 pacifier_msg[0]= 0;
 if((mode==1 || mode==2 || mode==3) && o->rec!=NULL) {
   /* append checksum tag and block list */;
   Checkrecord_get_sizevec(o->rec,hvec,&dummy,0);
   start_byte= Intvec_to_double(hvec,CdbplivS,0);
   if(fp_out!=stdout) {
     ret= Fseek_double(fp_out,start_byte,0);
     if(ret<0)
       {ret= 0; goto ex;}
   }
   Intvec_from_double(min_dutyvec,CdbplivS,1.0,0);
   ret= Checkrecord_write_tags(o->rec,fp_out,hvec,min_dutyvec,countvec,0);
   if(ret<=0)
     goto ex;
 } else {
   /* >>> */;
 }
 ret= 1;
ex:;
 if(pacifier_msg[0]!=0)
   Print_pacifier(0,"",1|8|16);
 fprintf(stderr,"                                                                             \n"); /* eat up possible foreign pacifiers */
 if(ret<0) {
   fprintf(stderr,"Program run aborted.\n");
   fprintf(stderr,"The data may still need more recovery work.\n");
   goto really_ex;
 }
 if(mode==0) {
   /* report missing blocks to stdout and set end mark */;
   for(Intvec_copy(blockno,CdbplivS,hvec,0);
       Intvec_cmp(hvec,CdbplivS,stopmark,0)<0;
       Intvec_add_b30(hvec,CdbplivS,1,0)) {
     Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
     strcat(size_text,"\n");
     Stdout_pipe_broker(size_text,strlen(size_text),0);
   }
   Stdout_pipe_broker("@\n",strlen("@\n"),0);
 }
 Intvec_to_dec(block_count,CdbplivS,size_text,sizeof(size_text),0);
 fprintf(stderr,"Blocks read : %8.8s\n",size_text);
 Intvec_add(missing,CdbplivS,stopmark,blockno,1);
 Intvec_to_dec(missing,CdbplivS,size_text,sizeof(size_text),0);
 if(!(mode==2))
   fprintf(stderr,"Missing     : %8.8s\n",size_text);
 if(mode==2 || mode==3) {
   Intvec_to_dec(mend_count,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"Mismatches  : %8.8s\n",size_text);
   Intvec_add(hvec,CdbplivS,mismatch_count,missing,0);
   Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"Unresolved  : %8.8s\n",size_text);
 } else {
   Intvec_to_dec(mismatch_count,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"Mismatches  : %8.8s\n",size_text);
 } 

 if(real_interval) {
   fprintf(stderr,"The data may still need more recovery work.\n");
   if(ret>0)
     ret= 0;
 } else if(Intvec_sign(mismatch_count,CdbplivS,0)==0 && 
    Intvec_sign(missing,CdbplivS,0)==0) {
   fprintf(stderr,"The data are supposed to be in perfect condition now.\n");
 } else {
   if(mode==1)
     fprintf(stderr,"The data may still need more recovery work.\n");
   else 
     fprintf(stderr,"The data still need more recovery work.\n");
   if(ret>0)
     ret= 0;
 }
really_ex:;
 if(flag&512)
   Intvec_copy(blockno,CdbplivS,endblock,0);
 if(flag&1024)
   Intvec_copy(mismatch_count,CdbplivS,endblock,0);
 if(blkbuf!=NULL)
   free((char *) blkbuf);
 Sfile_fclose(fps);
 Sfile_fclose(fpt);
 return(ret);
}


/* ----------------------------------------------------------------------- */


int Scan_checksum(scanstate,handle,fp,md5,crc32,
                  basevec,limitvec,countvec,
                  check_rec,flag)
struct Checkrecord_taG **scanstate;
char **handle;
FILE *fp;
unsigned char *md5;
unsigned long *crc32;
int *basevec;
int *limitvec;
int *countvec;
struct CheckrecorD **check_rec; /* out : list of found matching checksums */
int flag;
/*
  bit0= intermediately report checksums
  bit1= return after finding first valid tag
  bit2= last return yielded successful match
  bit3= there already was a successful match. Do not report mismatchs
  bit4= there already was a mismatch. Warn with successful matches
  bit5= permission to set bit3 and bit4 after found tags
  bit6= scan for blocklist
  bit7= set *check_rec to start of matchlist 
  bit8= with bit0 : do not initialize but use handle
  bit9= with bit0 : do not finalize but return handle
 bit10= Just load tags. No data outpout. No need to compare checksums. 
 bit11= write file tail from first tag (not at position 0, up to limitvec)
        to stdout
 bit12= produce MD5 in crc.md5[] rather than CRC32 crc.crc32
 bit13= print 1m pacifier messages
 bit14= with bit11 : accept first tag regardless of position match
*/
/*
 return:
 <=0 : error
   1 : ok
*/
{
 char buf[4*4096],record_text[161],size_text[160],pacifier_msg[160+20];
 char *ctx= NULL,*hctx= NULL;
 int ret,first= 1,bufsiz,bufpos,hflag,dummy;
 int to_write,pacifier_count= 0,head_len;
 int vec_size= CdbplivS,bufcount;
 int hvec[CdbplivS],posvec[CdbplivS],tail_limit[CdbplivS];
 double todo,rec_count= 0,counted;
 struct CheckrecorD *old_rec= NULL,*prev;

 pacifier_msg[0]= 0;

 if(flag&1024)
   flag&= ~1;
 if(flag&(1<<11)) {
   Intvec_copy(limitvec,CdbplivS,tail_limit,0);
   Intvec_from_double(limitvec,CdbplivS,1.0e24,0);
 }
 old_rec= *check_rec;
 if(flag&256) {
   if(flag&(1<<12))
     ctx= *handle;
   *handle= NULL;
 } else {
   if(flag&(1<<12)) {
     ret= Sregex_md5(&ctx,buf,0,md5,1);
     if(ret<=0)
       return(ret);
   } else
     *(crc32)= 0;
 }
 Intvec_copy(basevec,vec_size,posvec,0);
 Intvec_from_double(countvec,vec_size,0.0,0);
 bufsiz= sizeof(buf);
 while(1) {
   Intvec_add(hvec,vec_size,limitvec,countvec,1); /* limitvec - countvec */
   todo= Intvec_to_double(hvec,vec_size,0);
   if(todo<=0)
 break;
   if(todo<bufsiz)
     bufsiz= todo; 
   ret= fread(buf,1,bufsiz,fp);
   if(ret<=0)
 break;
   bufcount= ret;
   if(flag&(1<<13)) {
     pacifier_count+= bufcount;
     if(pacifier_count>=1024*1024) {
       counted= Intvec_to_double(countvec,vec_size,0)+bufcount;
       Sfile_scale(counted,size_text,7,1e5,1);
       sprintf(pacifier_msg,"  %s bytes scanned for checksum tag",size_text);
       Print_pacifier(0,pacifier_msg,4|32);
       pacifier_count= 0;
     }
   }
   if(!(flag&((1<<10)|(1<<11))))
     Stdout_pipe_broker(buf,bufcount,0);

   if(Checkrecord_tag_is_tail(*scanstate,0)) { /* tailcutter mode */;
     Stdout_pipe_broker(buf,bufcount,0);
     Intvec_add_b30(countvec,vec_size,bufcount,0);
     Intvec_add_b30(posvec,vec_size,bufcount,0);
 continue;
   }
   /* scan here */;
   bufpos= 0;
resume_scan:;
   hflag= ((*scanstate)==NULL)|(flag&(1<<12));
   hflag|= 2;
   hflag|= (flag&64)>>4;
   if(flag&(1<<11))
     hflag|= 8;
   hflag|= (flag&(1<<14));
   ret= Checkrecord_scan_for_tags(check_rec,scanstate,posvec,ctx,crc32,
                                  buf,bufcount,&bufpos,
                                  hflag);
   if(ret<0)
 break;
   if(ret>0) {
     /* found valid checksum tag */;
     if(flag&(1<<11)) {
       /* write version_id and position_in_stream to stdout */
       ret= Checkrecord_tag_write_head(*scanstate,stdout,0);
       if(ret<=0)
         goto ex;
       head_len= ret;
       /* reset counter to start of tag */
       Intvec_copy(tail_limit,CdbplivS,limitvec,0);
       Intvec_add_b30(posvec,vec_size,bufcount,0);
       /* write rest of buffer to stdout */
       to_write= (bufcount-bufpos);
       Intvec_from_double(hvec,CdbplivS,(double) to_write,0);
       if(Intvec_cmp(limitvec,CdbplivS,hvec,0)<0)
         Intvec_copy(limitvec,CdbplivS,hvec,0);
       to_write= Intvec_to_double(hvec,CdbplivS,0);
       Stdout_pipe_broker(buf+bufpos,to_write,0);
       Intvec_from_double(countvec,vec_size,(double) (to_write),0);
       Intvec_add_b30(countvec,vec_size,head_len,0);
       bufpos+= to_write;
 continue;
     }
     rec_count++;
     if(flag&1) {
       /* report found tag */;
       Checkrecord_to_text(*check_rec,record_text,0);
       if(Checkrecord_is_tested_match(*check_rec,0)) {
         Checkrecord_get_sizevec(*check_rec,hvec,&dummy,0);
         if((flag&16))
           fprintf(stderr,"CAUTION: There have been mismatches before !\n");
         if((flag&4) && Intvec_cmp(hvec,CdbplivS,posvec,0)==0) {
           fprintf(stderr,
                   "Note: Checksum tag found by scanning confirms match.\n\n");
         } else {
           fprintf(stderr,
                   "\n\nFound matching checksum tag by scanning :\n%s\n\n",
                   record_text);
         }
         if(flag&32)
           flag|= 8;
       } else {
         if(!(flag&8))
           fprintf(stderr,
 "\n\nWARNING : Scanning found valid but non-matching checksum tag :\n%s\n\n",
                  record_text); 
         if(flag&32)
           flag|= 16;
       }
     }
     /* eventually remove a record to prevent memory stress */;
     if(rec_count>Cd_backup_planer_tag_list_maX)
       Checkrecord_del_minor(check_rec,old_rec,0);
     if(flag&2)
 break;
     if(bufpos<bufcount)
       goto resume_scan;
   }

   Intvec_add_b30(countvec,vec_size,bufpos,0);
   Intvec_add_b30(posvec,vec_size,bufpos,0);
   if(!(flag&1024)) {
     if(flag&(1<<12))
       Sregex_md5(&ctx,buf,bufpos,md5,0);
     else
       Sregex_crc32_yenc(buf,bufpos,crc32,first);
     first= 0;
   }
 }

 if(flag&512) {
   /* do not destroy ctx but hand it over to caller */
   if(flag&(1<<12)) {
     /* get md5 from copy of ctx */
     Sregex_md5(&hctx,ctx,0,md5,1|2|4|(1<<15));
     *handle= ctx;
   }
 } else {
   if(flag&(1<<12)) 
     Sregex_md5(&ctx,buf,0,md5,2|(1<<15));
 }
 ret= 1;
ex:;
 if(pacifier_msg[0]!=0)
   Print_pacifier(0,"",1|8|16);

 if(flag&(1<<11)) {

   /* >>> do not return any records */;

   if(ret>0 && !Checkrecord_tag_is_tail(*scanstate,0))
     ret= 0;
 }

 if((flag&128) && (*check_rec)!=NULL) {
   Checkrecord_get_prev(*check_rec,&prev,0);
   for(;prev!=NULL;Checkrecord_get_prev(prev,&prev,0))
     (*check_rec)= prev;
 }
 return(ret);
}


int Filetimestamp(adr,timestamp,flag)
char *adr;
char timestamp[81];
int flag;
{
 time_t age;
 struct stat stbuf;
 char *Sfile_datestr();

 strcpy(timestamp,"000000.000000");
 if(Lstat(adr,&stbuf,0)==-1)
   return(0);
 age= stbuf.st_ctime;
 if(stbuf.st_mtime>stbuf.st_ctime)
   age= stbuf.st_mtime;
 strcpy(timestamp,Sfile_datestr(age,3));
 return(1);
}


int Fileadr_plus_slash(slashadr,adr,subptpt,flag)
char slashadr[4096];
char *adr;
char **subptpt;
int flag;
{
 char *subpt;

 sprintf(slashadr,"%s/",adr);
 subpt= slashadr+strlen(slashadr);
 /* avoid trailing double '/' */
 for(subpt= slashadr+strlen(slashadr);
   subpt-2>=slashadr && *(subpt-2)=='/';*(subpt--)= 0);
 (*subptpt)= subpt;
 return(1);
}


int Get_filetimes(fileadr,atime,mtime,ctime,flag)
char *fileadr;
time_t *atime;
time_t *mtime;
time_t *ctime;
int flag;
/*
 bit0= return 1 if file is missing
*/
{
 struct stat stbuf;

 if(lstat(fileadr,&stbuf)==-1)
   return(flag&1);

/* Why was this ? ts A61018 :
 if(! S_ISREG(stbuf.st_mode))
   return(0);
*/

 *atime= stbuf.st_atime;
 *mtime= stbuf.st_mtime;
 *ctime= stbuf.st_ctime;
 return(1);
}


/* ----------------- Misc - end -------- */


/* ----------------- Sfile extract -------- */

#define Smem_malloC malloc
#define Smem_freE free

#define SfileadrL 4096


char *Sfile_fgets(line,maxl,fp)
char *line;
int maxl;
FILE *fp;
{
int l;
char *ret;
 ret= fgets(line,maxl,fp);
 if(ret==NULL) return(NULL);
 l= strlen(line);
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 if(l>0) if(line[l-1]=='\n') line[--l]= 0;
 if(l>0) if(line[l-1]=='\r') line[--l]= 0;
 return(ret);
}


int Sfile_decode_datestr(reply,text,flag)
/* YYMMDD[.hhmm[ss]] */
struct tm *reply;
char *text;
int flag;
{
 int i,l;
 time_t current_time;
 struct tm *now;

 current_time= time(0);
 now= localtime(&current_time);
 for(i=0;i<sizeof(struct tm);i++)
   ((char *) reply)[i]= ((char *) now)[i];

 if(text[0]<'0'|| (text[0]>'9' && text[0]<'A') || text[0]>'Z')
   return(0);
 l= strlen(text);
 for(i=1;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=6)
   return(0);
 if(text[i]==0)
   goto decode;
 if(text[i]!='.' || (l!=11 && l!=13))
   return(0);
 for(i++;i<l;i++)
   if(text[i]<'0'||text[i]>'9')
     break;
 if(i!=l)
   return(0);

decode:;
 reply->tm_hour= 0;
 reply->tm_min= 0;
 reply->tm_sec= 0;
 i= 0;
 if(text[0]>='A') 
   reply->tm_year= 100+(text[i]-'A')*10+text[1]-'0';
 else
   reply->tm_year= 10*(text[0]-'0')+text[1]-'0';
 reply->tm_mon=  10*(text[2]-'0')+text[3]-'0'-1;
 reply->tm_mday= 10*(text[4]-'0')+text[5]-'0';
 if(l==6)
   return(1);
 reply->tm_hour= 10*(text[7]-'0')+text[8]-'0';
 reply->tm_min=  10*(text[9]-'0')+text[10]-'0';
 if(l==11)
   return(1);
 reply->tm_sec=  10*(text[11]-'0')+text[12]-'0';
 return(1);
}



char *Sfile_datestr(tim,flag)
time_t tim;
short int flag;
/*
 bit0=with hours+minutes  
 bit1=with seconds

 bit8= local time rather than UTC
*/
{
 static char zeitcode[80]={"000000"};
 char puff[80];
 struct tm *azt;

 if(flag&256) 
   azt = localtime(&tim);
 else
   azt = gmtime(&tim);
 
 if(azt->tm_year>99)
   sprintf(zeitcode,"%c%1.1d%2.2d%2.2d",
           'A'+(azt->tm_year-100)/10,azt->tm_year%10,
           azt->tm_mon+1,azt->tm_mday);
 else
   sprintf(zeitcode,"%2.2d%2.2d%2.2d",
           azt->tm_year,azt->tm_mon+1,azt->tm_mday);
 if(flag&1){
   sprintf(puff,".%2.2d%2.2d",azt->tm_hour,azt->tm_min);
   strcat(zeitcode,puff);
 }
 if(flag&2){
   sprintf(puff,"%2.2d",azt->tm_sec);
   strcat(zeitcode,puff);
 }

 return(zeitcode);
}


double Sfile_microtime(flag)
int flag;
{
#ifdef Stypes_on_ms_windowS
 return((double) time(0));
#else /* Stypes_on_ms_windowS */
 struct timeval tv;
 struct timezone tz;
 gettimeofday(&tv,&tz);
 return((double) (tv.tv_sec+1.0e-6*tv.tv_usec));
#endif /* ! Stypes_on_ms_windowS */
}


FILE *Sfile_fopen(filename,mode,flag)
char *filename;
char *mode;
int flag;
/*
 bit0= do not print error message on failure
*/
{
 FILE *fp= NULL;

 if(strcmp(filename,"-")==0) {
   if(mode[0]=='a' || mode[0]=='w' ||
      (mode[0]=='r' && mode[1]=='+') ||
      (mode[0]=='r' && mode[1]=='b' && mode[2]=='+')) 
     fp= stdout;
   else
     fp= stdin;
 } else if(strncmp(filename,"tcp:",4)==0){
   fprintf(stderr,"sorry - TCP/IP service isn't implemented yet.\n");
 } else if(strncmp(filename,"file:",5)==0){
   fp= fopen(filename+5,mode);
 } else {
   fp= fopen(filename,mode);
 }
 if(fp==NULL){
   if(!(flag&1)) {
     fprintf(stderr,"failed to open %s in %s mode\n",filename,mode);
     if(errno>0)
       fprintf(stderr,"   %s\n",strerror(errno));
   }
   return(NULL);
 }
 return(fp);
}


int Sfile_fclose(fp)
FILE *fp;
{
 if(fp!=stdout && fp!=stdin && fp!=NULL)
   fclose(fp);
 return(1);
}


int Sfile_bin_to_file(filename,data,size,flag)
char *filename;
char *data;
int size;
int flag;
/*
 bit0= permission to overwrite existing file
 bit1= append to file
 bit2= add newline character
*/
{
 FILE *fp;
 struct stat stbuf;
 int ret;

 if(!(flag&3))
   if(Cd_backup_planer_call_staT(filename,&stbuf)!=-1)
     return(0);
 if(flag&2)
   fp= Sfile_fopen(filename,"ab",0);
 else
   fp= Sfile_fopen(filename,"wb",0);
 if(fp==NULL)
   return(-1);
 if(size>0)
   ret= fwrite(data,size,1,fp);
 else
   ret= 1;
 if(ret==1 && (flag&4))
   ret= (fputc('\n',fp)!=EOF);
 Sfile_fclose(fp);
 return(ret);
}


int Sfile_string_to_file(filename,text,flag)
char *filename;
char *text;
int flag;
/*
 bit0= permission to overwrite existing file
 bit1= append to file
 bit2= add newline character
*/
{
 int ret;

 ret= Sfile_bin_to_file(filename,text,strlen(text),flag&7);
 return(ret);
}


int Sfile_file_to_text(filename,text,textsize,flag)
char *filename;
char *text;
int textsize;
int flag;
/*
 bit0= return(2) if file is empty (0 else)
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 FILE *fp;
 int ret;

 if(textsize<=0)
   return(-1);
 fp= Sfile_fopen(filename,"rb",(flag&2)>>1);
 if(fp==NULL)
   return(0);
 if(Sfile_fgets(text,textsize,fp)==NULL)
   ret= 2*(flag&1);
 else
   ret= 1;
 Sfile_fclose(fp);
 return(ret);
}


int Sfile_file_to_double(filename,num,flag)
/*
 bit0= return(2) if file is empty (0 else)
 bit1= Sfile_fopen:bit0 : no error messages 
*/
char *filename;
double *num;
int flag;
{
 char text[SfileadrL];
 double dv;
 int ret;

 if(Sfile_file_to_text(filename,text,sizeof(text),flag&2)<=0)
   return(0);
 ret= sscanf(text,"%lf",&dv);
 if(ret<=0)
   return(0);
 *num= dv;
 return(1);
}


int Sfile_file_to_int(filename,num,flag)
char *filename;
int *num;
int flag;
/*
 bit1= Sfile_fopen:bit0 : no error messages 
*/
{
 double dv;

 if(Sfile_file_to_double(filename,&dv,flag&2)<=0)
   return(0);
 if(dv<-0x7FFFFFFF || dv>=0x7FFFFFFF)
   return(0);
 *num= dv;
 return(1);
}


int Sfile_identical(file1,file2,flag)
char *file1,*file2;
int flag;
/*
 bit0= check wether it is the same  i-node 
*/
{
 FILE *fp1,*fp2;
 struct stat stbuf;
 int ret,c1,c2;
 ino_t ino1,ino2;

 ret= 0;

 if(flag&1){
   if(Cd_backup_planer_call_staT(file1,&stbuf)==-1) return(0);
   ino1= stbuf.st_ino;
   if(Cd_backup_planer_call_staT(file2,&stbuf)==-1) return(0);
   ino2= stbuf.st_ino;
   return(ino1==ino2);
 }

 fp1= Sfile_fopen(file1,"rb",0);
 fp2= Sfile_fopen(file2,"rb",0);
 if(fp1==NULL||fp2==NULL) goto fertig;
 while(1){
   c1= fgetc(fp1);
   c2= fgetc(fp2);
   if(c1==EOF && c2==EOF){
     ret= 1;
 break;
   }
   if(c1!=c2)
 break;
 }
fertig:;
 if(fp1!=NULL) fclose(fp1);
 if(fp2!=NULL) fclose(fp2);
 return(ret);
}


int Sfile_copy_file(source,target,errn,errmsg,flag)
char *source;
char *target;
int *errn;
char errmsg[SfileadrL];
int flag;
{
 FILE *fp,*fw;
 char *line[1024];
 int ret;

 if(Sfile_identical(source,target,1)>0)
   return(2);
 fp= fopen(source,"rb");
 if(fp==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %d on opening input file : %s\n%s",
                  errno,source,strerror(errno));
   return(0);
 }
 fw= fopen(target,"wb");
 if(fw==NULL) {
   *errn= errno;
   sprintf(errmsg,"Error %d on opening output file : %s\n%s",
                  errno,target,strerror(errno));
   ret= 0;
   goto ex;
 }
 while((ret= fread(line,1,sizeof(line),fp))>0){
   fwrite(line,1,ret,fw);
   if(ret<sizeof(line))
 break;
 }
 ret= 1;
ex:;
 if(fp!=NULL) fclose(fp);
 if(fw!=NULL) fclose(fw);
 return(ret);
}


int Sfile_move_file(source,target,errn,errmsg,flag)
char *source;
char *target;
int *errn;
char *errmsg;
int flag;
/*
 bit0= do not unlink target if source cannot be unlinked
 bit1= touch ctime
 bit7= use overwrite+delete rather than rename
 bit8= like bit7 but do not unlink source (i.e. copy >>> with properties)
*/
{
 int ret,ovw_del=0;
 struct stat stbuf;

 if(flag&(128|256))
   ovw_del= 1;
 if(!ovw_del)
   if(rename(source,target)!=-1) {
     if(flag&2) 
       if(Cd_backup_planer_call_lstaT(target,&stbuf)!=-1) 
         chmod(target,stbuf.st_mode);
     return(1);
   }
 if(errno==EXDEV || ovw_del) {
   /* try copy and unlink */;
   ret= Sfile_copy_file(source,target,errn,errmsg,0);
   if(ret<=0)
     return(ret);
   if(flag&2) 
     if(Cd_backup_planer_call_lstaT(target,&stbuf)!=-1) 
       chmod(target,stbuf.st_mode);
   if(!(flag&128)) {
     /* transfer as many properties as possible */;
     Sfile_copy_properties(source,target,1|4);
   }
   if(flag&256)
     return(3);
   if(unlink(source)==-1) {
     *errn= errno;
     sprintf(errmsg,
             "Error %d on removing file %s\nafter copying to %s\n\"%s\"",
             errno,source,target,strerror(errno));
     if(Cd_backup_planer_call_lstaT(source,&stbuf)!=-1 && !(flag&1)) {
       strcat(errmsg,"\n(Operation failed. Target file has been removed now)");
       unlink(target);
     }
     return(0);
   }
   return(2);
 }
 *errn= errno;
 sprintf(errmsg,"Error %d on renaming\n   %s\nto %s\n\"%s\"",
                errno,source,target,strerror(errno));
 return(0);
}

 
#ifndef Stypes_on_ms_windowS

int Sfile_set_filetimes(fileadr,mtime,atime,flag)
char *fileadr;
time_t mtime;
time_t atime;
int flag;
{
 struct stat stbuf;
 struct utimbuf buf;
 int ret;
 
 if(Cd_backup_planer_call_staT(fileadr,&stbuf)==-1)
   return(flag&1);
 if(!S_ISREG(stbuf.st_mode))
   return(0);
 buf.actime= atime;
 buf.modtime= mtime;
 ret= utime(fileadr,&buf);
 if(ret<=0)
   return(0);
 return(1);
}

#endif /* ! Stypes_on_ms_windowS */


int Sfile_set_properties(target,ownership,uid,gid,atime,mtime,ctime,flag)
char *target;
int ownership; /* low 12 bit of stat.st_mode */
uid_t uid;
gid_t gid;
time_t atime;
time_t mtime;
time_t ctime;
int flag;
/*
 bit0= set owner access bits
 bit1= set group access bits
 bit2= set others access bits
 bit3= set special access bits (S_ISUID,S_ISGID,S_ISVTX)
 bit4= set uid
 bit5= set gid
 bit6= set atime
 bit7= set mtime
 reserved until i know how to set it: bit8= set ctime
 bit13= with bit0: make target objects writable by owner
*/
{
 int mode_mask= 0,ret;
 mode_t new_mode;
 struct stat stbuf;

 if(Cd_backup_planer_call_staT(target,&stbuf)==-1)
   return(0);
 if(flag&1) {
   mode_mask|= 0700;
   if(flag&(1<<13))
     ownership|= S_IWUSR;     
 }
 if(flag&2)
   mode_mask|= 070;
 if(flag&4)
   mode_mask|= 07;
 if(flag&8)
   mode_mask|= 07000;
 if(mode_mask!=0) {
   new_mode= stbuf.st_mode&~mode_mask;
   new_mode|= mode_mask&ownership;
   new_mode&= 07777;
   if(chmod(target,new_mode)==-1)
     return(0);
 }
 if(!(flag&16))
   uid= stbuf.st_uid;
 if(!(flag&32))
   gid= stbuf.st_gid;
 if(flag&(16|32)) {
   if(chown(target,uid,gid)==-1)
     return(0);
 }

#ifndef Stypes_on_ms_windowS
 if(!(flag&64))
   atime= stbuf.st_atime;
 if(!(flag&128))
   mtime= stbuf.st_mtime;
 if((flag&(64|128)) && S_ISREG(stbuf.st_mode)) {
   ret= Sfile_set_filetimes(target,mtime,atime,0);
   if(ret<=0)
     return(ret);
 }
#endif /* ! Stypes_on_ms_windowS */

 return(1);
}


int Sfile_copy_properties(source,target,flag)
char *source;
char *target;
int flag;
/*
 bit0= copy access control
 bit1= copy ownership
 bit2= copy times
 bit13= with bit0: make target objects writable by owner
*/
{
 struct stat stbuf;
 int ret,sflag= 0;

 if(Cd_backup_planer_call_staT(source,&stbuf)==-1)
   return(0);
 if(flag&1)
   sflag|= (1|2|4|8);
 if(flag&2)
   sflag|= (16|32);
 if(flag&4)
   sflag|= (64|128|256);
 sflag|= flag&(1<<13);
 ret= Sfile_set_properties(target,stbuf.st_mode&0xfff,
                                  stbuf.st_uid,stbuf.st_gid,
                                  stbuf.st_atime,stbuf.st_mtime,stbuf.st_ctime,
                                  sflag);
 return(ret);
}


int Sfile_read_argv(progname,filename,argc,argv,flag)
char *progname;
char *filename;
int *argc;
char ***argv;
int flag;
/*
 bit0= read progname as first argument from line
 bit1= just release argument list argv and return
*/
{
 int ret,i,pass,maxl=0,l,argzaehl=0;
 char buf[SfileadrL];
 FILE *fp= NULL;

 Sfile_destroy_argv(argc,argv,0);
 if(flag&2)
   return(1);

 for(pass=0;pass<2;pass++) {
   if(!(flag&1)){
     argzaehl= 1;
     if(pass==0)
       maxl= strlen(progname);
     else
       strcpy((*argv)[0],progname);
   } else {
     argzaehl= 0;
     if(pass==0) maxl= 0;
   }
   fp= Sfile_fopen(filename,"rb",0);
   if(fp==NULL) 
     {ret= 0; goto ex;}
   while(Sfile_fgets(buf,sizeof(buf)-1,fp)!=NULL) {
     l= strlen(buf);
     if(pass==0){
       if(l>maxl) 
         maxl= l;
     }else{
       strcpy((*argv)[argzaehl],buf);
     }
     argzaehl++;
   }
   if(pass==0){
     *argc= argzaehl;
     if(argzaehl>0) {
       *argv= (char **) Smem_malloC(argzaehl*sizeof(char *));
       if(*argv==NULL)
         {ret= -1; goto ex;}
     }
     for(i=0;i<*argc;i++) {
       (*argv)[i]= (char *) Smem_malloC((maxl+1));
       if((*argv)[i]==NULL)
         {ret= -1; goto ex;}
     }
   }     
   Sfile_fclose(fp,0); fp= NULL;
 }

 ret= 1;
ex:;
 if(fp!=NULL)
   Sfile_fclose(fp,0);
 return(ret);
}


int Sfile_destroy_argv(argc,argv,flag)
int *argc;
char ***argv;
int flag;
{
 int i;

 if(*argc>0 && *argv!=NULL){
   for(i=0;i<*argc;i++){
     if((*argv)[i]!=NULL) 
       Smem_freE((*argv)[i]);
   }
   Smem_freE((char *) *argv);
 }
 *argc= 0;
 *argv= NULL;
 return(1);
}



int Sfile_type(filename,flag)
char *filename;
int flag;
/*
 bit0= return -1 if file is missing
 bit1= return a hardlink with siblings as type 5
 bit2= evaluate eventual link target rather than the link object itself
*/
/*
 return:  0=unknown  1=regular  2=directory  3=symbolic link  4=named pipe
          5=multiple hardlink (with bit1) , 6=block device
*/
{
 struct stat stbuf;

 if(flag&4) {
   if(Cd_backup_planer_call_staT(filename,&stbuf)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 } else {
   if(Cd_backup_planer_call_lstaT(filename,&stbuf)==-1) {
     if(flag&1) return(-1);
     else       return(0);
   }
 }
 if(S_ISREG(stbuf.st_mode)) {
   if(flag&2)
     if(stbuf.st_nlink>1)
       return(5);
   return(1);
 }
 if(S_ISDIR(stbuf.st_mode))
   return(2);
 if(S_ISLNK(stbuf.st_mode))
   return(3);
 if(S_ISFIFO(stbuf.st_mode))
   return(4);
 if(S_ISBLK(stbuf.st_mode))
   return(6);
 return(0);
}


int Sfile_readlink(filename,buff,buff_size,flag)
char *filename;
char *buff;
int buff_size;
int flag;
/*
 bit0= return -1 if file is missing
*/
{
 int l;

#ifdef Stypes_on_ms_windowS

 return(0);

#else /* Stypes_on_ms_windowS */

 if(Sfile_type(filename,0)!=3)
   return(0);
 l= readlink(filename,buff,buff_size-1);
 if(l<=0)
   return(0);
 buff[l]= 0;
 return(1);

#endif /* ! Stypes_on_ms_windowS */
}


int Sfile_scale(value,result,siz,thresh,flag)
double value;
char *result;
int siz;
double thresh;
int flag;
/*
 bit0= eventually ommit 'b'
 bit1= make text as short as possible
 bit2= no fraction (if it would fit at all)
*/
{
 char scale_c,scales[7],form[80];
 int i,dec_siz= 0,avail_siz= 1;

 strcpy(scales,"bkmgtp");
 scale_c= scales[0];
 for(i=1;scales[i]!=0;i++) {
   if(value<thresh-0.5) 
 break;
   value/= 1024.0;
   scale_c= scales[i];
 } 
 if(scale_c!='b' && !(flag&4)) { /* is there room for fractional part ? */
   avail_siz= siz-1;
   sprintf(form,"%%.lf");
   sprintf(result,"%.lf",value);
   if(strlen(result)<=avail_siz-2) 
     dec_siz= 1;                                  /* we are very modest */
 }
 if(scale_c=='b' && (flag&1)) {
   if(flag&2)
     sprintf(form,"%%.lf");
   else
     sprintf(form,"%%%d.lf",siz);
   sprintf(result,form,value);
 } else {
   if(flag&2)
     sprintf(form,"%%.lf%%c");
   else if(dec_siz>0) 
     sprintf(form,"%%%d.%dlf%%c",avail_siz,dec_siz);
   else
     sprintf(form,"%%%d.lf%%c",siz-1);
   sprintf(result,form,value,scale_c);
 }
 return(1);
}


/* ----------------- Sfile extract end -------- */


/* ----------------- Sregex extract -------- */

int Sregex_string_cut(handle,text,len,flag)
char **handle;
char *text;
int len;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int l=0;
 char *old_handle;

 if((flag&1)&&*handle!=NULL)
   l+= strlen(*handle);
 old_handle= *handle;
 if(text!=NULL) {
   l+= len;
   *handle= TSOB_FELD(char,l+1);
   if(*handle==NULL) {
     *handle= old_handle;
     return(0);
   }
   if((flag&1) && old_handle!=NULL)
     strcpy(*handle,old_handle);
   else
     (*handle)[0]= 0;
   if(len>0)
     strncat(*handle,text,len);
 } else {
   *handle= NULL;
 }
 if(old_handle!=NULL)
   Smem_freE(old_handle);
 return(1);
}


int Sregex_string(handle,text,flag)
char **handle;
char *text;
int flag;
/*
 bit0= append (text!=NULL)
*/
{
 int ret,l=0;

 if(text!=NULL)
   l= strlen(text);
 
/* #define Sregex_looking_for_contenT 1 */
#ifdef Sregex_looking_for_contenT
 /* a debugging point if a certain text content has to be caught */
 if(text!=NULL)
   if(strcmp(text,"clear")==0)
     ret= 0;
#endif

 ret= Sregex_string_cut(handle,text,l,flag&1);
 return(ret);
}


/*
  from Juergen Helbing's  yencode-c.txt 
  found at  www.yenc.org  , Feb 4,  2002
  license : "It is public domain - as well as the yEnc specification"
  some alterations made to adapt it to Sregex_crc32() 

  seems to be compatible with SFV/CSV files
*/                       

int Sregex_crc32_yenc(data,len,crc,flag)
char *data;
int len;
unsigned long *crc;
int flag;
/*
 bit0= start of data polynom
*/
{
 /* X^32+X^26+X^23+X^22+X^16+X^12+X^11+X^10+X^8+X^7+X^5+X^4+X^2+X^1+X^0 */
 static unsigned long crc_tab[256] = {
   0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 
   0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
   0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988,
   0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
   0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de,
   0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
   0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec,
   0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
   0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172,
   0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
   0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940,
   0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
   0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116,
   0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
   0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924,
   0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
   0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a,
   0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
   0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818,
   0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
   0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e,
   0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
   0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c,
   0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
   0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2,
   0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
   0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0,
   0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
   0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086,
   0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
   0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4,
   0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
   0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a,
   0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
   0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8,
   0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
   0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe,
   0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
   0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc,
   0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
   0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252,
   0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
   0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60,
   0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
   0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236,
   0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
   0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04,
   0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
   0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a,
   0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
   0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38,
   0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
   0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e,
   0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
   0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c,
   0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
   0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2,
   0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
   0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0,
   0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
   0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6,
   0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
   0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94,
   0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

 int k;
 unsigned long crc_val,ch1,ch2,cc;

 if(flag&1)
   crc_val = ~0;
 else
   crc_val = ~*crc;

 for(k=0;k<len;k++) {
   cc= data[k] & 0xffL;
   ch1= (crc_val ^ cc) & 0xffL;
   ch1= crc_tab[ch1];
   ch2= (crc_val>>8L) & 0xffffffL;
   crc_val= ch1 ^ ch2;
 }
 *crc= ~crc_val;
 return(1);
}


/* This code is derived from RFC 1321 and implements computation of the
    "RSA Data Security, Inc. MD5 Message-Digest Algorithm" */

#define Sregex_S11 7
#define Sregex_S12 12
#define Sregex_S13 17
#define Sregex_S14 22
#define Sregex_S21 5
#define Sregex_S22 9
#define Sregex_S23 14
#define Sregex_S24 20
#define Sregex_S31 4
#define Sregex_S32 11
#define Sregex_S33 16
#define Sregex_S34 23
#define Sregex_S41 6
#define Sregex_S42 10
#define Sregex_S43 15
#define Sregex_S44 21


/* F, G, H and I are basic MD5 functions.
 */
#define Sregex_F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define Sregex_G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define Sregex_H(x, y, z) ((x) ^ (y) ^ (z))
#define Sregex_I(x, y, z) ((y) ^ ((x) | (~z)))

/* ROTATE_LEFT rotates x left n bits.
 */
#define Sregex_ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */
#define Sregex_FF(a, b, c, d, x, s, ac) { \
 (a) += Sregex_F ((b), (c), (d)) + (x) + (Stypes_uint32)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define Sregex_GG(a, b, c, d, x, s, ac) { \
 (a) += Sregex_G ((b), (c), (d)) + (x) + (Stypes_uint32)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define Sregex_HH(a, b, c, d, x, s, ac) { \
 (a) += Sregex_H ((b), (c), (d)) + (x) + (Stypes_uint32)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }
#define Sregex_II(a, b, c, d, x, s, ac) { \
 (a) += Sregex_I ((b), (c), (d)) + (x) + (Stypes_uint32)(ac); \
 (a) = Sregex_ROTATE_LEFT ((a), (s)); \
 (a) += (b); \
  }


/* MD5 context. */
typedef struct {
  Stypes_uint32 state[4];                                   /* state (ABCD) */
  Stypes_uint32 count[2];        /* number of bits, modulo 2^64 (lsb first) */
  unsigned char buffer[64];                         /* input buffer */
} Sregex_md5_ctX;


/* MD5 basic transformation. Transforms state based on block.
 */
int Sregex_md5__transform (state, block)
Stypes_uint32 state[4];
unsigned char block[64];
{
 Stypes_uint32 a = state[0], b = state[1], c = state[2], d = state[3], x[16];
 unsigned int i, j;

 for (i = 0, j = 0; j < 64; i++, j += 4)
   x[i] = ((Stypes_uint32)block[j]) | (((Stypes_uint32)block[j+1]) << 8) |
    (((Stypes_uint32)block[j+2]) << 16) | (((Stypes_uint32)block[j+3]) << 24);

  /* Round 1 */
  Sregex_FF (a, b, c, d, x[ 0], Sregex_S11, 0xd76aa478); /* 1 */
  Sregex_FF (d, a, b, c, x[ 1], Sregex_S12, 0xe8c7b756); /* 2 */
  Sregex_FF (c, d, a, b, x[ 2], Sregex_S13, 0x242070db); /* 3 */
  Sregex_FF (b, c, d, a, x[ 3], Sregex_S14, 0xc1bdceee); /* 4 */
  Sregex_FF (a, b, c, d, x[ 4], Sregex_S11, 0xf57c0faf); /* 5 */
  Sregex_FF (d, a, b, c, x[ 5], Sregex_S12, 0x4787c62a); /* 6 */
  Sregex_FF (c, d, a, b, x[ 6], Sregex_S13, 0xa8304613); /* 7 */
  Sregex_FF (b, c, d, a, x[ 7], Sregex_S14, 0xfd469501); /* 8 */
  Sregex_FF (a, b, c, d, x[ 8], Sregex_S11, 0x698098d8); /* 9 */
  Sregex_FF (d, a, b, c, x[ 9], Sregex_S12, 0x8b44f7af); /* 10 */
  Sregex_FF (c, d, a, b, x[10], Sregex_S13, 0xffff5bb1); /* 11 */
  Sregex_FF (b, c, d, a, x[11], Sregex_S14, 0x895cd7be); /* 12 */
  Sregex_FF (a, b, c, d, x[12], Sregex_S11, 0x6b901122); /* 13 */
  Sregex_FF (d, a, b, c, x[13], Sregex_S12, 0xfd987193); /* 14 */
  Sregex_FF (c, d, a, b, x[14], Sregex_S13, 0xa679438e); /* 15 */
  Sregex_FF (b, c, d, a, x[15], Sregex_S14, 0x49b40821); /* 16 */

  /* Round 2 */
  Sregex_GG (a, b, c, d, x[ 1], Sregex_S21, 0xf61e2562); /* 17 */
  Sregex_GG (d, a, b, c, x[ 6], Sregex_S22, 0xc040b340); /* 18 */
  Sregex_GG (c, d, a, b, x[11], Sregex_S23, 0x265e5a51); /* 19 */
  Sregex_GG (b, c, d, a, x[ 0], Sregex_S24, 0xe9b6c7aa); /* 20 */
  Sregex_GG (a, b, c, d, x[ 5], Sregex_S21, 0xd62f105d); /* 21 */
  Sregex_GG (d, a, b, c, x[10], Sregex_S22,  0x2441453); /* 22 */
  Sregex_GG (c, d, a, b, x[15], Sregex_S23, 0xd8a1e681); /* 23 */
  Sregex_GG (b, c, d, a, x[ 4], Sregex_S24, 0xe7d3fbc8); /* 24 */
  Sregex_GG (a, b, c, d, x[ 9], Sregex_S21, 0x21e1cde6); /* 25 */
  Sregex_GG (d, a, b, c, x[14], Sregex_S22, 0xc33707d6); /* 26 */
  Sregex_GG (c, d, a, b, x[ 3], Sregex_S23, 0xf4d50d87); /* 27 */
  Sregex_GG (b, c, d, a, x[ 8], Sregex_S24, 0x455a14ed); /* 28 */
  Sregex_GG (a, b, c, d, x[13], Sregex_S21, 0xa9e3e905); /* 29 */
  Sregex_GG (d, a, b, c, x[ 2], Sregex_S22, 0xfcefa3f8); /* 30 */
  Sregex_GG (c, d, a, b, x[ 7], Sregex_S23, 0x676f02d9); /* 31 */
  Sregex_GG (b, c, d, a, x[12], Sregex_S24, 0x8d2a4c8a); /* 32 */

  /* Round 3 */
  Sregex_HH (a, b, c, d, x[ 5], Sregex_S31, 0xfffa3942); /* 33 */
  Sregex_HH (d, a, b, c, x[ 8], Sregex_S32, 0x8771f681); /* 34 */
  Sregex_HH (c, d, a, b, x[11], Sregex_S33, 0x6d9d6122); /* 35 */
  Sregex_HH (b, c, d, a, x[14], Sregex_S34, 0xfde5380c); /* 36 */
  Sregex_HH (a, b, c, d, x[ 1], Sregex_S31, 0xa4beea44); /* 37 */
  Sregex_HH (d, a, b, c, x[ 4], Sregex_S32, 0x4bdecfa9); /* 38 */
  Sregex_HH (c, d, a, b, x[ 7], Sregex_S33, 0xf6bb4b60); /* 39 */
  Sregex_HH (b, c, d, a, x[10], Sregex_S34, 0xbebfbc70); /* 40 */
  Sregex_HH (a, b, c, d, x[13], Sregex_S31, 0x289b7ec6); /* 41 */
  Sregex_HH (d, a, b, c, x[ 0], Sregex_S32, 0xeaa127fa); /* 42 */
  Sregex_HH (c, d, a, b, x[ 3], Sregex_S33, 0xd4ef3085); /* 43 */
  Sregex_HH (b, c, d, a, x[ 6], Sregex_S34,  0x4881d05); /* 44 */
  Sregex_HH (a, b, c, d, x[ 9], Sregex_S31, 0xd9d4d039); /* 45 */
  Sregex_HH (d, a, b, c, x[12], Sregex_S32, 0xe6db99e5); /* 46 */
  Sregex_HH (c, d, a, b, x[15], Sregex_S33, 0x1fa27cf8); /* 47 */
  Sregex_HH (b, c, d, a, x[ 2], Sregex_S34, 0xc4ac5665); /* 48 */

  /* Round 4 */
  Sregex_II (a, b, c, d, x[ 0], Sregex_S41, 0xf4292244); /* 49 */
  Sregex_II (d, a, b, c, x[ 7], Sregex_S42, 0x432aff97); /* 50 */
  Sregex_II (c, d, a, b, x[14], Sregex_S43, 0xab9423a7); /* 51 */
  Sregex_II (b, c, d, a, x[ 5], Sregex_S44, 0xfc93a039); /* 52 */
  Sregex_II (a, b, c, d, x[12], Sregex_S41, 0x655b59c3); /* 53 */
  Sregex_II (d, a, b, c, x[ 3], Sregex_S42, 0x8f0ccc92); /* 54 */
  Sregex_II (c, d, a, b, x[10], Sregex_S43, 0xffeff47d); /* 55 */
  Sregex_II (b, c, d, a, x[ 1], Sregex_S44, 0x85845dd1); /* 56 */
  Sregex_II (a, b, c, d, x[ 8], Sregex_S41, 0x6fa87e4f); /* 57 */
  Sregex_II (d, a, b, c, x[15], Sregex_S42, 0xfe2ce6e0); /* 58 */
  Sregex_II (c, d, a, b, x[ 6], Sregex_S43, 0xa3014314); /* 59 */
  Sregex_II (b, c, d, a, x[13], Sregex_S44, 0x4e0811a1); /* 60 */
  Sregex_II (a, b, c, d, x[ 4], Sregex_S41, 0xf7537e82); /* 61 */
  Sregex_II (d, a, b, c, x[11], Sregex_S42, 0xbd3af235); /* 62 */
  Sregex_II (c, d, a, b, x[ 2], Sregex_S43, 0x2ad7d2bb); /* 63 */
  Sregex_II (b, c, d, a, x[ 9], Sregex_S44, 0xeb86d391); /* 64 */

  state[0] += a;
  state[1] += b;
  state[2] += c;
  state[3] += d;

  /* Zeroize sensitive information. */
  memset ((char *) x, 0, sizeof (x));
  return(1);
}


int Sregex_md5__encode(output, input, len)
unsigned char *output;
Stypes_uint32 *input;
unsigned int len;
{
  unsigned int i, j;

  for (i = 0, j = 0; j < len; i++, j += 4) {
    output[j] = (unsigned char)(input[i] & 0xff);
    output[j+1] = (unsigned char)((input[i] >> 8) & 0xff);
    output[j+2] = (unsigned char)((input[i] >> 16) & 0xff);
    output[j+3] = (unsigned char)((input[i] >> 24) & 0xff);
  }
  return(1);
}



int Sregex_md5_init(ctx,flag)
Sregex_md5_ctX *ctx;
int flag;
{
 ctx->count[0] = ctx->count[1] = 0;
 /* Load magic initialization constants. */
 ctx->state[0] = 0x67452301;
 ctx->state[1] = 0xefcdab89;
 ctx->state[2] = 0x98badcfe;
 ctx->state[3] = 0x10325476;
 return(1);
}


/* MD5 block update operation. Continues an MD5 message-digest
  operation, processing another message block, and updating the
  context.
 */
int Sregex_md5_update(ctx,data,datalen,flag)
Sregex_md5_ctX *ctx;
unsigned char *data;
int datalen;
int flag;
{
 unsigned int i, index, partlen;

 /* Compute number of bytes mod 64 */
 index = (unsigned int)((ctx->count[0] >> 3) & 0x3F);
 /* Update number of bits */
 if ((ctx->count[0] += ((Stypes_uint32) datalen << 3)) < 
     ((Stypes_uint32) datalen << 3))
   ctx->count[1]++;
 ctx->count[1] += ((Stypes_uint32) datalen >> 29);
 partlen = 64 - index;

 /* Transform as many times as possible. */
 if (datalen >= partlen) {
   memcpy((char *) &ctx->buffer[index], (char *) data, partlen);
   Sregex_md5__transform(ctx->state, ctx->buffer);
   for (i = partlen; i + 63 < datalen; i += 64)
     Sregex_md5__transform(ctx->state, &data[i]);
   index = 0;
 } else
   i = 0;

 /* Buffer remaining data */
 memcpy((char *) &ctx->buffer[index], (char *) &data[i],datalen-i);

 return(1);
}


int Sregex_md5_final(ctx,result,flag)
Sregex_md5_ctX *ctx;
char result[16];
int flag;
{
 unsigned char bits[8];
 unsigned int index, padlen;
 static unsigned char PADDING[64] = {
   0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 };

 /* Save number of bits */
 Sregex_md5__encode(bits, ctx->count, 8);
 /* Pad out to 56 mod 64.  */
 index = (unsigned int)((ctx->count[0] >> 3) & 0x3f);
 padlen = (index < 56) ? (56 - index) : (120 - index);
 Sregex_md5_update(ctx, PADDING, padlen,0);
 /* Append length (before padding) */
 Sregex_md5_update(ctx, bits, 8,0);
 /* Store state in result */
 Sregex_md5__encode(result, ctx->state, 16);
 /* Zeroize sensitive information.  */
 memset ((char *) ctx, 0, sizeof (*ctx));
 return(1);
}


#ifdef Cdbackup_planer_iso_C
int Sregex_md5(Sregex_md5_ctX **ctx, unsigned char *data, int datalen,
               char result[16],int flag)
#else /* Cdbackup_planer_iso_C */
int Sregex_md5(ctx,data,datalen,result,flag)
Sregex_md5_ctX **ctx;     /* *ctx has to be NULL or point to freeable memory */
unsigned char *data;
int datalen;
char result[16];
int flag;
#endif /* ! Cdbackup_planer_iso_C */
/*
 bit0= allocate and init *ctx
 bit1= transfer ctx to result
 bit2= with bit 0 : clone new *ctx from data
 bit15= free *ctx
*/
{
 if(flag&1) {
   if(*ctx!=NULL)
     free((char *) *ctx);
   *ctx= TSOB_FELD(Sregex_md5_ctX,1);
   if(*ctx==NULL)
     return(-1);
   Sregex_md5_init(*ctx,0);
   if(flag&4)
     memcpy((char *) *ctx,data,sizeof(Sregex_md5_ctX));
 }
 if(*ctx==NULL)
   return(0);
 if(datalen>0)
   Sregex_md5_update(*ctx,data,datalen,0);
 if(flag&2) {
   Sregex_md5_final(*ctx,result,0);
 }
 if(flag&(1<<15)) {
   free((char *) *ctx);
   *ctx= NULL;
 }
 return(1);
}


int Md5_from_text(md5,md5_text,flag)
unsigned char md5[16];
char md5_text[33];
int flag;
{
 int i,ret;
 char h[3];
 
 h[2]= 0;
 for(i=0;i<16;i++) {
   h[0]= md5_text[2*i];
   h[1]= md5_text[2*i+1];
   ret= 255;
   sscanf(h,"%x",&ret);
   md5[i]= ret;
 }
 return(1);
}


/* ----------------- Sregex extract end -------- */


/* ----------------------- Lies_popen --------------------- */


struct Lies_popeN {
 char cmd[4096];
 int anz;          /* count */
 char **lines;
};


int Lies_popen_init(cmd,gelesen,flag)
char *cmd;
struct Lies_popeN *gelesen;
int flag;
{
 FILE *fp;
 struct ZeilE {
   struct ZeilE *prev;
   struct ZeilE *next;
   char *text;
 };
 struct ZeilE *zeile,*neuzeile,*startzeile= NULL;
 char line[4096],*fret;
 int i,zeilz=0,gesl,ret=0,l;

 /* remove dead children if needed */
 wait3(NULL,WNOHANG,NULL);

 strcpy(gelesen->cmd,cmd);
 gelesen->anz= 0;
 gelesen->lines= NULL;

 fp= popen(cmd,"r");
 if(fp==NULL){
   fprintf(stderr,"popen(\"%s\",\"r\") failed. errno=%d",cmd,errno);
   return(0);
 }

 /* create a list of lines  */
 startzeile=  TSOB_FELD(struct ZeilE,1);
 startzeile->prev= startzeile->next= NULL;
 startzeile->text= NULL;
 /* read in */
 zeile= startzeile;
 while(1) {
   fret= Sfile_fgets(line,sizeof(line)-1,fp);
 if(fret==NULL) break;
   zeile->text= malloc(strlen(line)+1);
 if(zeile->text == NULL) break;
   strcpy(zeile->text,line);
   neuzeile= TSOB_FELD(struct ZeilE,1);
 if(neuzeile==NULL) break;
   zeile->next= neuzeile;
   neuzeile->prev= zeile;
   neuzeile->next= NULL;
   neuzeile->text= NULL;
   zeile= neuzeile;
   zeilz++;
 } 
 if(fret!=NULL) {
   /* something went wrong */
speichermangel:;
   fprintf(stderr,"out of memory while reading output of: %s\n",cmd);
   ret= -1;
   goto tmp_freigeben;
 }

 pclose(fp);
 if(zeilz==0) goto tmp_freigeben;

 /* create return data */
 gelesen->anz= zeilz;
 gelesen->lines= TSOB_FELD(char *,gelesen->anz);
 if(gelesen->lines==NULL) goto speichermangel;
 for(i=0;i<gelesen->anz;i++)
   gelesen->lines[i]= NULL;
 gesl= 0;
 zeile= startzeile;
 for(i=0;i<zeilz;i++){
   l= strlen(zeile->text);
   gelesen->lines[i]= TSOB_FELD(char,l+1);
   if(gelesen->lines[i]==NULL) goto speichermangel;
   strcpy(gelesen->lines[i],zeile->text);
   gesl+= l+1;
   zeile= zeile->next;
 }

 ret= 1;
tmp_freigeben:;
 if(ret<=0)
    Lies_popen_cleanup(gelesen,0);
 zeile= startzeile;
 while(zeile!=NULL){
   if(zeile->text!=NULL) free(zeile->text);
   neuzeile= zeile->next;
   free((char *) zeile);
   zeile= neuzeile;
 } 
 return(ret);

}


int Lies_popen_cleanup(gelesen,flag)
struct Lies_popeN *gelesen;
int flag;
{
 int i;

 /* remove dead children if needed */
 wait3(NULL,WNOHANG,NULL);

 if(gelesen==NULL)
   return(0);
 if(gelesen->lines!=NULL){
   for(i= 0;i<gelesen->anz;i++)
     if(gelesen->lines[i]!=NULL)
       free(gelesen->lines[i]);
   free((char *) gelesen->lines);
   gelesen->lines= NULL;
 }
 gelesen->anz= 0;
 return(1);
}


int Lies_popen_transplant(gelesen,target,index,flag)
struct Lies_popeN *gelesen;
char **target;
int index;
int flag;
{
 (*target)= gelesen->lines[index];
 gelesen->lines[index]= NULL;
 return(1);
}


/* ----------------------- Lies_popen - end -------------- */


/* ----------------- Checkrecord_tag ----------------- */

/* A serialized checksum format to be appended to data streams 
   so it may be recognized on the fly when reading that stream */

/* Tag layout :
    version_id position_in_stream record_len checksum_record record_md5
*/

struct Checkrecord_taG {
 int state;  /* looking for :
              0= version_id (tag start)
              1= position_in_stream
              2= record_len
              3= checksum_record
              4= record_md5
              5= newline
              6= tag complete but not yet validated
              7= tag validated , decide wether to look for blocklist

             16= checksum blocklist id
             17= blocksize line
             18= number of blocks line
             19= remainder line
             20= block data (using block_count,block_bytes)
             21= newline
             22= md5_of_list
             23= end of list\n 
             24= blocklist is complete and valid

             32= head of a tag at correct position found in tailcutter mode

     2000000000= nothing (tag is completed)
              */
 int hashtype;
 int hash_valid;
 Sregex_md5_ctX *md5_ctx;
 unsigned long crc32;
 int posvec[CdbplivS];
 int vid_pos;
 char position_in_stream[36];
 char record_len[12];
 char checksum_record[164];
 char record_md5[36];

 char linebuffer[161];

 struct CheckrecorD *pending_record;

 int scan_for_blocklist;

 /* read status of Blksum_lisT */
 int bid_pos;
 int blocks_to_read[CdbplivS];
 int block_count[CdbplivS]; /* up to complete_blocks+!!remainder */
 int block_bytes;           /* up to 16 */
 char md5buffer[16];
 Sregex_md5_ctX *blocklist_md5_ctx;
 struct Blksum_lisT *blocklist;

};


int Checkrecord_tag_new(objpt,flag)
struct Checkrecord_taG **objpt;
int flag;
/*
  bit1 = scan for blocklist
  bit12= hashtype (0=CRC32, 1=MD5)
*/
{
 struct Checkrecord_taG *o;

 *objpt= o= TSOB_FELD(struct Checkrecord_taG,1);
 if(o==NULL)
   return(-1);
 Checkrecord_tag_reset(o,1|(flag&(2|(1<<12))));
 return(1);
}


int Checkrecord_tag_reset(o,flag)
struct Checkrecord_taG *o;
int flag;
/*
  bit0 = init : do not free non-NULL pointers
  bit1 = scan for blocklist
  bit12= hashtype (0=CRC32, 1=MD5)
*/
{
 int i;

 o->state= 0;
 if(flag&1)
   o->hashtype= !!(flag&(1<<12));
 o->hash_valid= 0;
 if(!(flag&1))
   if(o->md5_ctx!=NULL)
     Sregex_md5(&(o->md5_ctx),"",0,NULL,1<<15);
 o->md5_ctx= NULL;
 o->crc32= 0;
 o->vid_pos= 0;
 Intvec_from_double(o->posvec,CdbplivS,0.0,0);
 o->position_in_stream[0]= 0;
 o->record_len[0]= 0;
 o->checksum_record[0]= 0;
 o->record_md5[0]= 0;
 o->linebuffer[0]= 0;
 if(!(flag&1))
   Checkrecord_destroy(&(o->pending_record),0);
 o->pending_record= NULL;
 if(flag&1)
   o->scan_for_blocklist= !!(flag&2);
 o->bid_pos= 0;
 Intvec_from_double(o->blocks_to_read,CdbplivS,0.0,0);
 Intvec_from_double(o->block_count,CdbplivS,0.0,0);
 o->block_bytes= 0;
 for(i=0;i<16;i++)
   o->md5buffer[i]= 0;
 if(!(flag&1))
   if(o->blocklist_md5_ctx!=NULL)
     Sregex_md5(&(o->blocklist_md5_ctx),"",0,NULL,1<<15);
 o->blocklist_md5_ctx= NULL;
 if(!(flag&1))
   Blksum_list_destroy(&(o->blocklist),0);
 o->blocklist= NULL;

 return(1);
}


int Checkrecord_tag_destroy(objpt,flag)
struct Checkrecord_taG **objpt;
int flag;
{
 struct Checkrecord_taG *o;

 o= *objpt;
 if(o==NULL)
   return(0);
 if(o->md5_ctx!=NULL)
   Sregex_md5(&(o->md5_ctx),"",0,NULL,1<<15);
 Checkrecord_destroy(&(o->pending_record),0);
 if(o->blocklist_md5_ctx!=NULL)
   Sregex_md5(&(o->blocklist_md5_ctx),"",0,NULL,1<<15);
 Blksum_list_destroy(&(o->blocklist),0);

 free((char *) o);
 *objpt= NULL;
 return(1);
}


#ifdef Cdbackup_planer_iso_C 
int Checkrecord_tag_clone_ctx(struct Checkrecord_taG *o,
                              Sregex_md5_ctX *ctx, unsigned long crc32,
                              int *posvec, char *buf, int bufpos, int flag)
#else /* Cdbackup_planer_iso_C */
int Checkrecord_tag_clone_ctx(o,ctx,crc32,posvec,buf,bufpos,flag)
struct Checkrecord_taG *o;
Sregex_md5_ctX *ctx;
unsigned long crc32;
int *posvec;
char *buf;
int bufpos;
int flag;
#endif /* ! Cdbackup_planer_iso_C */
/*
 bit0-2= verbosity, see Checkrecord_scan_for_tags()

 bit12= work with MD5 rather than CRC32
*/
{
 int ret,verbous;
 int todovec[CdbplivS];
 char size_text[81];

 verbous= (flag&7);

 if(bufpos<0) {
   if(verbous>=2) {
     Intvec_copy(posvec,CdbplivS,todovec,0);
     Intvec_add_b30(todovec,CdbplivS,bufpos,0);
     Intvec_to_dec(todovec,CdbplivS,size_text,sizeof(size_text),0);
     fprintf(stderr,"Checksum tag with gap to data at position %s\n",
                    size_text);
   }
   ret= 0;
   goto ex;
 }
 if(o->hashtype) {
   if(o->md5_ctx!=NULL)
     Sregex_md5(&(o->md5_ctx),"",0,size_text,1<<15);
   ret= Sregex_md5(&(o->md5_ctx),(unsigned char *) ctx,0,size_text,1|4);
   if(ret<=0)
     {ret= -1; goto ex;}
   Sregex_md5(&(o->md5_ctx),buf,bufpos,size_text,0);
 } else {
   o->crc32= crc32;
   Sregex_crc32_yenc(buf,bufpos,&(o->crc32),0);
 }
 o->hash_valid= 1;
 ret= 1;
ex:
 return(ret);
}


int Checkrecord_tag_new_blocklist(o,flag)
struct Checkrecord_taG *o;
int flag;
{
 int ret;

 if(o==NULL)
   return(0);
 Intvec_from_double(o->block_count,CdbplivS,0.0,0);
 o->block_bytes= 0;
 if(o->blocklist_md5_ctx!=NULL)
   Sregex_md5(&(o->blocklist_md5_ctx),"",0,NULL,1<<15);
 ret= Sregex_md5(&(o->blocklist_md5_ctx),"",0,NULL,1);
 if(ret<=0)
   return(-1);
 Blksum_list_destroy(&(o->blocklist),0);
 ret= Blksum_list_new(&(o->blocklist),NULL,0);
 if(ret<=0)
   return(-1);
 return(1);
}


int Checkrecord_tag_is_tail(s,flag)
struct Checkrecord_taG *s;
int flag;
{
 if(s==NULL)
   return(0);
 return(s->state==32);
}


int Checkrecord_tag_write_head(s,fp,flag)
struct Checkrecord_taG *s;
FILE *fp;
int flag;
{
 int ret;

 ret= Write_broker_or_file(fp,Checkrecord_tag_starT,strlen(Checkrecord_tag_starT),0);
 if(ret<=0)
   goto ex;
 ret= Write_broker_or_file(fp,s->position_in_stream,strlen(s->position_in_stream),0);
 if(ret<=0)
   goto ex;
 ret= Write_broker_or_file(fp," ",1,0);
 if(ret<=0)
   goto ex;

 ret= strlen(Checkrecord_tag_starT)+strlen(s->position_in_stream)+1;
ex:;
 return(ret);
}


int Checkrecord_tag_scan_num(s,target,target_size,buf,bufcount,i,flag)
struct Checkrecord_taG *s;
char *target;
char target_size;
char *buf;
int bufcount;
int *i;
int flag;
/*
 return:
  0= reset state
  1= state persisted
  2= state advanced
*/
{
 int l;

 l= strlen(target);
 for(;*i<bufcount;(*i)++) {
   if(isdigit(buf[*i])) {
     if(l>=target_size-1)
       return(0);
     target[l]= buf[*i];
     l++;
     target[l]= 0;
   } else if(buf[*i]==' ') {
     s->state++;
     (*i)++;
     return(2);
   } else
     return(0);
 }
 return(1);
}


int Checkrecord_tag_scan_line(s,buf,bufcount,i,flag)
struct Checkrecord_taG *s;
char *buf;
int bufcount;
int *i;
int flag;
/*
 return:
  0= reset state
  1= state persisted
  2= state advanced
*/
{
 int l;

 l= strlen(s->linebuffer);
 for(;*i<bufcount;(*i)++) {
   if(buf[*i]=='\n') {
     s->state++;
     (*i)++;
     return(2);
   } else {
     if(l>=sizeof(s->linebuffer)-1)
       return(0);
     s->linebuffer[l]= buf[*i];
     l++;
     s->linebuffer[l]= 0;
   }
 }
 return(1);
}


int Checkrecord_tag_scan_ivecline(s,expected_prefix,ivec,buf,bufcount,i,flag)
struct Checkrecord_taG *s;
char *expected_prefix;
int ivec[CdbplivS];
char *buf;
int bufcount;
int *i;
int flag;
/*
 bit0-3= verbosity
 bit4= raw : do not read ivec, do not reset s->linebuffer
*/
/*
 return:
  0= reset state
  1= state persisted
  2= state advanced
*/
{
 int ret,verbous,l;

 verbous= flag&15;
 ret= Checkrecord_tag_scan_line(s,buf,bufcount,i,0);
 if(ret<=0) {
   if(verbous>=1) {
     fprintf(stderr,"Oversized header line\n");
     if(verbous>=7)
       fprintf(stderr,"%s...\n",s->linebuffer);
     fprintf(stderr,"Expected : %s[0-9]*\n",expected_prefix);
   }
   return(0);
 } else if(ret==2) {
   /* line complete */;
   l= strlen(expected_prefix);
   if(strncmp(s->linebuffer,expected_prefix,l)!=0) {
     if(verbous>=1) {
       fprintf(stderr,"Unexpected header line\n");
       if(verbous>=7)
         fprintf(stderr,"%s\n",s->linebuffer);
       fprintf(stderr,"Expected : %s[0-9]*\n",expected_prefix);
     }
     return(0);
   } 
   if(!(flag&16)) {
     ret= Intvec_from_dec(ivec,CdbplivS,s->linebuffer+l,0);
     if(ret<=0) {
       if(verbous>=1) {
         fprintf(stderr,"Unusable number text in header line\n");
         if(verbous>=7)
           fprintf(stderr,"%s\n",s->linebuffer);
         fprintf(stderr,"Expected : %s[0-9]*\n",expected_prefix);
       }
       return(0);
     }
     s->linebuffer[0]= 0;
   }
   return(2);
 }
 return(1);
}


/* ----------------- Checkrecord ----------------- */


struct CheckrecorD {

 char *name;
 time_t date;
 int sizevec[CdbplivS];
 int hashtype;   /* 0=CRC32 , 1=MD5 */
 union {
  unsigned long crc32;
  unsigned char md5[16];
 } checksum;
 int status;  /* bit0= tested , bit1= successful match */
 struct Blksum_lisT *blocklist;

 struct CheckrecorD *prev,*next;

};


int Checkrecord_new(objpt,name,date,size,sizevec,hashtype,crc32,md5,flag)
struct CheckrecorD **objpt;
char *name;
time_t date;
double size;
int *sizevec;
int hashtype;
unsigned long crc32;
unsigned char md5[16];
int flag;
/*
 bit0= set date automatically
 bit1= do not cut to 80 chars, do not replace obnoxious characters in name 
       beware: such a record may never be printed by Checkrecord_to_text()
*/
{
 struct CheckrecorD *o;
 int i;

 *objpt= o= TSOB_FELD(struct CheckrecorD,1);
 if(o==NULL)
   return(-1);

 o->name= NULL;
 if(flag&1)
   o->date= time(0);
 else
   o->date= date;
 if(sizevec!=NULL) {
   Intvec_copy(sizevec,CdbplivS,o->sizevec,0);
 } else {
   Intvec_from_double(o->sizevec,CdbplivS,size,0);
 }
 o->hashtype= hashtype;
 o->checksum.crc32= 0;
 for(i=0;i<16;i++)
   o->checksum.md5[i]= 0;
 if(hashtype==0)
   o->checksum.crc32= crc32;
 else if(md5!=NULL) 
   for(i=0;i<16;i++)
     o->checksum.md5[i]= md5[i];
 o->status= 0;
 o->blocklist= NULL;
 o->prev= NULL;
 o->next= NULL;

 if(flag&2) {
   if(Sregex_string(&(o->name),name,0)<=0)
     goto failed;
 } else {
   if(Sregex_string_cut(&(o->name),name,80,0)<=0)
     goto failed;
   if(!(flag&2))
     for(i=0;o->name[i]!=0;i++)
       if(isspace(o->name[i]) || o->name[i]=='\n' || o->name[i]==' ')
         o->name[i]= '_';
 }
 return(1);
failed:;
 Checkrecord_destroy(objpt,0);
 return(-1);
}


int Checkrecord_destroy(objpt,flag)
struct CheckrecorD **objpt;
int flag;
{
 struct CheckrecorD *o;

 o= *objpt;
 if(o==NULL)
   return(0);

 Checkrecord_set_blocklist(o,NULL,0);
 Sregex_string(&(o->name),NULL,0);
 Blksum_list_destroy(&(o->blocklist),0);
 if(o->prev!=NULL)
   o->prev->next= o->next;
 if(o->next!=NULL)
   o->next->prev= o->prev;

 Smem_freE((char *) o);
 *objpt= NULL;
 return(1);
}


int Checkrecord_destroy_all(objpt,flag)
struct CheckrecorD **objpt;
int flag;
{
 struct CheckrecorD *o,*n;

 o= *objpt;
 if(o==NULL)
   return(0);
 for(;o->prev!=NULL;o= o->prev);
 for(;o!=NULL;o= n) {
   n= o->next;
   Checkrecord_destroy(&o,0);
 }
 *objpt= NULL;
 return(1);
}


int Checkrecord_get_name(o,pt,flag)
struct CheckrecorD *o;
char **pt;
int flag;
{
 *pt= o->name;
 return(1);
}


int Checkrecord_get_date(o,pt,flag)
struct CheckrecorD *o;
time_t *pt;
int flag;
{
 *pt= o->date;
 return(1);
}


int Checkrecord_get_sizevec(o,ivec,ivec_size,flag)
struct CheckrecorD *o;
int *ivec;
int *ivec_size;
int flag;
/*
 bit0= do not copy ivec, only read ivec_size
*/
{
 if(!(flag&1))
   Intvec_copy(o->sizevec,CdbplivS,ivec,0);
 *ivec_size= CdbplivS;
 return(1);
}


int Checkrecord_get_size(o,pt,flag)
struct CheckrecorD *o;
double *pt;
int flag;
{
 *pt= Intvec_to_double(o->sizevec,CdbplivS,0);
 return(1);
}


int Checkrecord_get_checksum(o,hashtype,crc32,md5,flag)
struct CheckrecorD *o;
int *hashtype;
unsigned long *crc32;
unsigned char md5[16];
int flag;
{
 int i;

 *hashtype= o->hashtype;
 if(o->hashtype==0)
   *crc32= o->checksum.crc32;
 else
   for(i=0;i<16;i++)
     md5[i]= o->checksum.md5[i];
 return(1);
}


int Checkrecord_set_blocklist(o,blocklist,flag)
struct CheckrecorD *o;
struct Blksum_lisT *blocklist;
int flag;
{
 if(o->blocklist!=NULL)
   Blksum_list_set_rec(o->blocklist,NULL,0);
 if(blocklist!=NULL)
   Blksum_list_set_rec(blocklist,o,0);
 o->blocklist= blocklist;
 return(1);
}


int Checkrecord_is_tested_match(o,flag)
struct CheckrecorD *o;
int flag;
{
 return((o->status&3)==3);
}


int Checkrecord_get_prev(o,pt,flag)
struct CheckrecorD *o;
struct CheckrecorD **pt;
int flag;
{
 *pt= o->prev;
 return(1);
}


int Checkrecord_get_next(o,pt,flag)
struct CheckrecorD *o;
struct CheckrecorD **pt;
int flag;
{
 *pt= o->next;
 return(1);
}


int Checkrecord_link(o,link,flag)
struct CheckrecorD *o;
struct CheckrecorD *link;
int flag;
/*
  bit0= insert as link->prev rather than as link->next
  bit1= append to outer end of link's neighbor list
*/
{
 if(o==NULL)
   return(0);
 if(o->prev!=NULL)
   o->prev->next= o->next;
 if(o->next!=NULL)
   o->next->prev= o->prev;
 o->prev= o->next= NULL;
 if(link==NULL)
   return(1);
 if(flag&1) {
   if(flag&2)
     while(link->prev!=NULL)
       link= link->prev;
   o->next= link;
   o->prev= link->prev;
   if(o->prev!=NULL)
     o->prev->next= o;
   link->prev= o;
 } else {
   if(flag&2)
     while(link->next!=NULL)
       link= link->next;
   o->prev= link;
   o->next= link->next;
   if(o->next!=NULL)
     o->next->prev= o;
   link->next= o;
 }
 return(1);
}


int Checkrecord_count(liststart,flag)
struct CheckrecorD *liststart;
int flag;
{
 struct CheckrecorD *o;
 int count= 0;

 for(o= liststart;o!=NULL;o= o->next) {
   if(o->hashtype<0)
 continue;
   count++;
 }
 return(count);
}


int Checkrecord_compare_checksum(f1,f2)
struct CheckrecorD **f1,**f2;
{
 int i;

 if((*f1)->hashtype<(*f2)->hashtype)
   return(-2);
 if((*f1)->hashtype>(*f1)->hashtype)
   return(2);
 if((*f1)->hashtype == 0) {
   if((*f1)->checksum.crc32<(*f2)->checksum.crc32)
     return(-1);
   if((*f1)->checksum.crc32>(*f2)->checksum.crc32)
     return(1);
 } else {
   for(i=0;i<16;i++) {
     if((*f1)->checksum.md5[i]<(*f2)->checksum.md5[i])
       return(-1);
     if((*f1)->checksum.md5[i]>(*f2)->checksum.md5[i])
       return(1);
   }
 }
 return(0);
}


int Checkrecord_compare_size(f1,f2)
struct CheckrecorD **f1,**f2;
{
 return(Intvec_cmp((*f1)->sizevec,CdbplivS,(*f2)->sizevec,0));
}


int Checkrecord_cmp(f1,f2,flag)
struct CheckrecorD *f1,*f2;
int flag;
{
 int ret;

 ret= Checkrecord_compare_size(&f1,&f2);
 if(ret!=0)
   return(ret);
 if(ret!=0)
   return(ret);
 if(f1->date>f2->date)
   return(1);
 else if(f1->date<f2->date)
   return(-1);
 ret= strcmp(f1->name,f2->name);
 if(ret!=0)
   return(ret);
 ret= Checkrecord_compare_checksum(&f1,&f2);
 if(ret==-2)
   ret= -1;
 else if(ret==2)
   ret= 1;
 if(ret!=0)
   return(ret);
 return(0);
}


int Checkrecord_create_sorted_pt(liststart,pt_io,pt_z,Checkrecord_compare,flag)
struct CheckrecorD *liststart;
int (*Checkrecord_compare)();
struct CheckrecorD ***pt_io;
int *pt_z;
int flag;
{
 struct CheckrecorD *o;
 int i,ret,count= 0;

 *pt_io= NULL;
 count= Checkrecord_count(liststart,0);
 *pt_io= TSOB_FELD(struct CheckrecorD *,count+1);
 if(*pt_io==NULL)
   {ret= -1; goto ex;}
 i= 0;
 for(o= liststart;o!=NULL;o= o->next) {
   if(o->hashtype<0)
 continue;
   if(i>=count)
 break;
   (*pt_io)[i]= o;
   i++;
 }
 count= i;
 if(count>0)
   qsort(*pt_io,count,sizeof(struct CheckrecorD *),Checkrecord_compare);
 *pt_z= count;
 ret= 1;
ex:;
 if(ret<=0) {
   if(*pt_io!=NULL)
     Smem_freE((char *) *pt_io);
   *pt_io= NULL;
 }
 return(ret);
}


int Checkrecord_from_text(objpt,line,flag)
struct CheckrecorD **objpt;
char *line;
int flag;
{
 int ret,valid,hashtype= 0;
 char name[256],date_text[256],size_text[256],checksum_text[256];
 unsigned char md5[16];
 time_t date;
 int sizevec[CdbplivS];
 unsigned long crc32;

 if(strlen(line)>=sizeof(name))
   return(0);
 name[0]= date_text[0]= size_text[0]= checksum_text[0]= 0;
 sscanf(line,"%s %s %s %s",name,date_text,size_text,checksum_text);
 ret= String_to_time(date_text,&valid,&date,0);
 if(ret<=0)
   return(0);
 ret= Intvec_from_dec(sizevec,CdbplivS,size_text,0);
 if(ret<=0)
   return(0);
 /* determine hashtype and decode */
 if(strlen(checksum_text)==32) {
   hashtype= 1;
   Md5_from_text(md5,checksum_text,0);
/*
   for(i=0;i<16;i++) {
     buf[0]= checksum_text[2*i];
     buf[1]= checksum_text[2*i+1];
     buf[2]= 0;
     ret= 255;
     sscanf(buf,"%x",&ret);
     md5[i]= ret;
   }
*/
 } else {
   hashtype= 0;
   crc32= 0;
   sscanf(checksum_text,"%lX",&crc32);
 }
 /* create new record */
 ret= Checkrecord_new(objpt,name,date,0.0,sizevec,hashtype,crc32,md5,0);
 if(ret<=0)
   return(-1);
 return(1);
}


int Checkrecord_read_list(objpt,listadr,rec_count,flag)
struct CheckrecorD **objpt;
char *listadr;
int *rec_count;
int flag;
/*
 bit0= sorted by size 
*/
{
 int ret,line_count= 0,i,sortcount= 0;
 char line[256];
 FILE *fp= NULL;
 struct CheckrecorD *o,**sortpt= NULL;

 *rec_count= 0;
 fp= Sfile_fopen(listadr,"rb",0);
 if(fp==NULL)
   {ret= 0; goto ex;}
 while(Sfile_fgets(line,sizeof(line)-1,fp)!=NULL) {
   (line_count)++;
   o= *objpt;
   ret= Checkrecord_from_text(objpt,line,0);
   if(ret<0)
     goto ex;
   if(ret==0)
 continue;
   Checkrecord_link(*objpt,o,1);
   (*rec_count)++;
 }
 if(*objpt==NULL)
   {ret= 2; goto ex;}

 if(flag&1) {
   /* sort to pointer array */
   ret= Checkrecord_create_sorted_pt(*objpt,&sortpt,&sortcount,
                                     Checkrecord_compare_size,0);
   if(ret<=0)
     goto ex;
   if(sortcount!=*rec_count) 
     {ret= -1; goto ex;}          /* that would be very strange */
   /* relink */
   for(i=0;i<sortcount;i++) {
     if(i==0)
       sortpt[i]->prev= NULL;
     else
       sortpt[i]->prev= sortpt[i-1];
     if(i<sortcount-1)
       sortpt[i]->next= sortpt[i+1];
     else
       sortpt[i]->next= NULL;
   }
   *objpt= sortpt[0];
 }
 ret= 1;
ex:;
 if(sortpt!=NULL)
   Smem_freE((char *) sortpt);
 if(fp!=NULL)
   Sfile_fclose(fp);
 return(ret);
}


int Checkrecord_to_text(o,text,flag)
struct CheckrecorD *o;
char text[161];
int flag;
/*
 bit0= add newline
*/
{
 char size_text[32];
 int i,l;

 /* print name date length checksum */
 Intvec_to_dec(o->sizevec,CdbplivS,size_text,sizeof(size_text),0);
 sprintf(text,"%s %s %s ",o->name,Sfile_datestr(o->date,1|2|256),size_text);
 l= strlen(text);
 if(o->hashtype==1) {
   for(i=0;i<16;i++)
     sprintf(text+l+2*i,"%2.2x",o->checksum.md5[i]);
 } else {
   sprintf(text+l,"%lX",o->checksum.crc32);
 }
 if(flag&1)
   strcat(text,"\n");
 return(1);
}


int Checkrecord_write_tags(o,fp,posvec,min_dutyvec,countvec,flag)
struct CheckrecorD *o;
FILE *fp;
int *posvec;
int *min_dutyvec;
int *countvec;
int flag;
/*
 bit1= do not produce local checksum list
*/
{
 int ret,i,l;
 int dvec[CdbplivS],hvec[CdbplivS],dummy_count[CdbplivS],blocksize[CdbplivS];
 char version_text[129],pos_text[33],len_text[9],record_text[161],md5_text[33];
 char complete_text[513];
 unsigned char md5[16];
 Sregex_md5_ctX *ctx= NULL;
 struct Blksum_lisT *local_bl= NULL;

 Intvec_copy(min_dutyvec,CdbplivS,dvec,0);
 Intvec_from_double(countvec,CdbplivS,0.0,0);

 /* make record and compute its MD5 */
 Checkrecord_to_text(o,record_text,0);
 sprintf(len_text,"%d",(int) strlen(record_text));
 Sregex_md5(&ctx,record_text,strlen(record_text),md5,1|2|(1<<15));
 for(i=0;i<16;i++)
   sprintf(md5_text+2*i,"%2.2x",md5[i]);
 strcpy(version_text,Checkrecord_tag_starT);
 while(Intvec_sign(dvec,CdbplivS,0)>0) {
   Intvec_to_dec(posvec,CdbplivS,pos_text,sizeof(pos_text),0);
   sprintf(complete_text,"%s%s %s %s %s\n",
           version_text,pos_text,len_text,record_text,md5_text);
   l= strlen(complete_text);
   ret= Write_broker_or_file(fp,complete_text,l,0);
   if(ret<=0)
     goto ex;

   if(o->blocklist!=NULL && !(flag&2)) {
     /* create block checksum list and register tag data */
     Blksum_list_get_sizes(o->blocklist,blocksize,blocksize,blocksize,1);
     ret= Blksum_list_from_data(&local_bl,complete_text,l,
                                blocksize,dummy_count,1);
     if(ret<=0)
       goto ex; 
   }

   Intvec_add_b30(posvec,CdbplivS,l,0);
   Intvec_add_b30(dvec,CdbplivS,l,1);
   Intvec_add_b30(countvec,CdbplivS,l,0);

   /* eventually write o->blocklist */;
   if(o->blocklist!=NULL) { 
     Intvec_copy(countvec,CdbplivS,hvec,0);
     ret= Blksum_list_write(o->blocklist,fp,hvec,&local_bl,2);
     if(ret<=0)
       goto ex;
     /* compute number of bytes written */
     Intvec_add(hvec,CdbplivS,hvec,countvec,1);
     /* apply to counters and pointers */
     Intvec_add(posvec,CdbplivS,posvec,hvec,0);
     Intvec_add(dvec,CdbplivS,dvec,hvec,1);
     Intvec_add(countvec,CdbplivS,countvec,hvec,0);
   }
 }
 ret= 1;
ex:;
 if(local_bl!=NULL)
   Blksum_list_destroy(&local_bl,0);
 return(ret); 
}


int Checkrecord_scan_for_tags(objpt,scanstate,posvec,ctx,crc32,
                              buf,bufcount,bufpos,flag)
struct CheckrecorD **objpt;
struct Checkrecord_taG **scanstate;
int *posvec;
Sregex_md5_ctX *ctx;
unsigned long *crc32;
char *buf;
int bufcount;
int *bufpos;
int flag;
/*
 flag:
  bit0= init search 
  bit1= return after finding first valid tag
  bit2= scan for blocklist
  bit3= tailcutter mode :
        enter state 32 as soon as position of tag start is confirmed
  bit4= scan for Cdbpl_multiplication_marK :
        only states 0,1,(2),5,2000000000

 bit12= work with MD5 rather than CRC32
 bit14= with bit3: accept tag even at wrong position

 return:
  <0 = error
   0 = no tag found (yet)
   1 = credible tag found, checksum record returned in *o
*/
{
 static int verbous= 1; /* report :
                           0=nothing 
                           1=illness / wrong hashtype
                           2=dislocation
                           3=repetition
                           7=everything
                         */
 int ret,l,i,k,rec_len,i_pos= -1,hret,i_start,look_for_multmark= 0;
 int accept_any_position;
 int hvec[CdbplivS],pis_vec[CdbplivS];
 struct Checkrecord_taG *s= NULL;
 char *tag_start;
 char *blocklist_start= Checkrecord_blocklist_starT;
 unsigned char md5[16];
 char size_text[81],md5_text[33],dummy_buf[1],tagtype[80];
 Sregex_md5_ctX *temp_ctx= NULL;

 if(flag&16) {
   look_for_multmark= 1;
   tag_start= Cdbpl_multiplication_marK;
   strcpy(tagtype,"Multiplication Mark");
 } else {
   tag_start= Checkrecord_tag_starT;
   strcpy(tagtype,"Checksum Tag");
 }
 accept_any_position= (flag&8) && (flag&(1<<14));
 if(flag&1) {
   ret= Checkrecord_tag_new(scanstate,(flag&(1<<12))|((flag&4)>>1));
   if(ret<=0)
     {ret= -1; goto ex;}
   *bufpos= 0;
 }
 s= *scanstate;
 if(s==NULL)
   {ret= 0; goto ex;}

 i= *bufpos;
 if(0) { /* only reachable by Spaghetti Airways */
resume_state_0:;
   Checkrecord_tag_reset(s,0);
 }
 if(s->state==0) {
   for(;i<bufcount;i++) {
     if(buf[i]==tag_start[s->vid_pos]) {
       (s->vid_pos)++;
       if(tag_start[s->vid_pos]==0) {
         /* complete tag start found */
         s->state= 1;
         i++;
   break;
       }
     } else {
       /* begin from start again*/
       if(s->vid_pos>0) {
         s->hash_valid= 0;
         s->vid_pos= (buf[i]==tag_start[0]);
       }
   continue;
     }
   }
 }
 if(s->state<=0)                /* a shortcut for the most likely situation */
   {ret= 0; goto ex;}

 if(s->state==1) {
   i_pos= i-strlen(tag_start);
   if(s->position_in_stream[0]==0) {
     /* determine tag start position */;
     Intvec_copy(posvec,CdbplivS,s->posvec,0);           /* buffer start pos */
     Intvec_add_b30(s->posvec,CdbplivS,i_pos,0);       /* plus tag start pos */
     if(verbous>=7) {
       Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,"%s header encountered at position %s\n",
                      tagtype,size_text);
     }
   }
   ret= Checkrecord_tag_scan_num(s,s->position_in_stream,
                             sizeof(s->position_in_stream),buf,bufcount,
                             &i,0);
   if(ret<=0) {
     if(verbous>=2) {
       Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,"%s header with unusable PIS at position %s\n",
                      tagtype,size_text);
     }
     goto resume_state_0;
   }
   if(s->state==2) {
     /* check position in stream */;
     Intvec_from_dec(pis_vec,CdbplivS,s->position_in_stream,0);
     if((!accept_any_position) && Intvec_cmp(pis_vec,CdbplivS,s->posvec,0)!=0){
       if(verbous>=2) {
         fprintf(stderr,
            "Checkrecord_scan_for_tag : dislocated %s ignored : \n",tagtype);
         Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
         fprintf(stderr,"streampos= %s",size_text);
         Intvec_to_dec(pis_vec,CdbplivS,size_text,sizeof(size_text),0);
         fprintf(stderr,"  ,  tagpos= %s\n",size_text);
       }
       goto resume_state_0;
     }
     if(flag&8) {
       if(Intvec_sign(pis_vec,CdbplivS,0)==0) {
         if(verbous>=2)
           fprintf(stderr,"%s at position 0 skipped.\n",tagtype);
         goto resume_state_0;
       }
       s->state= 32; /* tailcutter mode */
     }
   }
 }
 if(s->state==2 && look_for_multmark) {
   s->state= 5;
 } else if(s->state==2) {
   ret= Checkrecord_tag_scan_num(s,s->record_len,
                                 sizeof(s->record_len),buf,bufcount,
                                 &i,0);
   if(ret<=0) {
     if(verbous>=1) {
       Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,"Checksum tag header with unusable RL at position %s\n",
                      size_text);
     }
     goto resume_state_0;
   }
 }
 if(s->state==3) {
   /* read record_len bytes plus blank */
   rec_len= 0;
   sscanf(s->record_len,"%d",&rec_len);
   if(rec_len>=sizeof(s->checksum_record)) {
     if(verbous>=1) {
       Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,"Checksum tag header with unusable CSR at position %s\n",
                      size_text);
     }
     goto resume_state_0;
   }
   l= strlen(s->checksum_record);
   for(;i<bufcount && l<rec_len+1;i++) {
     if(buf[i]==0) {
       if(verbous>=1) {
         Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
         fprintf(stderr,
                 "Checksum tag header with unusable TMD5 at position %s\n",
                 size_text);
       }
       goto resume_state_0;
     }
     s->checksum_record[l]= buf[i];
     l++;
     s->checksum_record[l]= 0;
   }
   if(l>=rec_len+1) {
     s->checksum_record[l-1]= 0; /* remove trailing blank */
     s->state++;
   }
 }
 if(s->state==4) {
   /* read 32 bytes of MD5 */;
   l= strlen(s->record_md5);
   for(;i<bufcount && l<32;i++) {
     if(!isalnum(buf[i]))
       goto resume_state_0;
     s->record_md5[l]= buf[i];
     l++;
     s->record_md5[l]= 0;
   }
   if(l>=32)
     s->state++;
 }
 if(i>=bufcount)
   {ret=0; goto ex;}
 if(s->state==5) {
   /* final newline */
   if(buf[i]!='\n') {
     if(verbous>=1) {
       Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
       if(look_for_multmark && buf[i]=='X') {
         fprintf(stderr,"Note: %s header of extracted volume encountered.\n",
                        tagtype);
       } else {
         fprintf(stderr,
                 "%s header without final newline at position %s\n",
                 tagtype,size_text);
       }
     }
     goto resume_state_0;
   }
   if(look_for_multmark) {
     i++;
     s->state= 2000000000;
   } else {
     /* check record with MD5 */;
     Sregex_md5(&temp_ctx,s->checksum_record,strlen(s->checksum_record),
                md5,1|2|(1<<15));
     for(k=0;k<16;k++)
       sprintf(md5_text+2*k,"%2.2x",md5[k]);
     if(strcmp(md5_text,s->record_md5)!=0) {
       if(verbous>=1) {
         Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
         fprintf(stderr,
           "Checksum tag header with inner checksum mismatch at position %s\n",
           size_text);
       }
       goto resume_state_0;
     } 
     i++;
     s->state++;
   }
 }
 if(s->state==6) {
   ret= Checkrecord_from_text(&(s->pending_record),s->checksum_record,0);
   if(ret<0)
     {ret= -1; goto ex;}
   if(ret==0) {
     if(verbous>=1) {
       Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,
         "Checksum tag header with malformed checksum record at position %s\n",
         size_text);
     }
     goto resume_state_0;
   }
   if(s->pending_record->hashtype != s->hashtype) {
     if(verbous>=1) {
       Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,
         "Checksum tag header with non matching checksum type at position %s\n",
         size_text);
     }
     goto resume_state_0;
   }
   if((*objpt)!=NULL) {
     if(Checkrecord_cmp(s->pending_record,*objpt,0)==0) {
       if(verbous>=3) {
         Intvec_to_dec(s->posvec,CdbplivS,size_text,sizeof(size_text),0);
         fprintf(stderr,
                 "Checksum tag repeated checksum record at position %s\n",
                 size_text);
       }
       goto resume_state_0;
     }
   }

   /* check wether that matches */
   if(!s->hash_valid) { /* the whole checksum tag was in this buf */
     ret= Checkrecord_tag_clone_ctx(s,ctx,*crc32,posvec,buf,i_pos,verbous&7);
     if(ret<=0) /* eventual message was printed by function */
       goto resume_state_0;
   }
   s->pending_record->status= (s->pending_record->status&~3)|1;
   if(s->hashtype==1) {
     Sregex_md5(&(s->md5_ctx),dummy_buf,0,md5,2);
     for(k=0;k<16;k++)
       if(md5[k]!=s->pending_record->checksum.md5[k])
     break;
     if(k>=16) /* mark as match */
       s->pending_record->status|= 2;
   } else {
     if(*crc32==s->pending_record->checksum.crc32) /* mark as match */
       s->pending_record->status|= 2;
   }
   s->state++;
 }
 if(s->state==7) {
   if(verbous>=7)
     fprintf(stderr,"Tag state %d , scan_for_blocklist= %d\n",
                    s->state,s->scan_for_blocklist);
   /* found tag is valid, decide wether to demand blocklist */
   if(s->scan_for_blocklist==0)
     s->state= 2000000000;
   else {
     ret= Checkrecord_tag_new_blocklist(s,0);
     if(ret<=0)
       {ret= -1; goto ex;}
     s->state= 16;
   }
 }
 /* -------------------- look for block checksum list ----------------- */
 if(s->state==16) {
   /* checksum blocklist id */
   for(;i<bufcount;i++) {
     if(buf[i]==blocklist_start[s->bid_pos]) {
       (s->bid_pos)++;
       if(blocklist_start[s->bid_pos]==0) {
         /* complete blocklist start found */
         s->state++;
         s->linebuffer[0]= 0;
         i++;
   break;
       }
     } else 
       goto resume_state_0;
   }
   ret= Sregex_md5(&(s->blocklist_md5_ctx),buf,0,NULL,1);
   if(ret<=0) 
     {ret= -1; goto ex;}
   Sregex_md5(&(s->blocklist_md5_ctx),blocklist_start,strlen(blocklist_start),
              NULL,0);
 }
 if(s->state==17) {
   i_start= i;
   ret= Checkrecord_tag_scan_ivecline(s,"blocksize=",hvec,
                                      buf,bufcount,&i,verbous&15);
   if(ret<0) 
     {ret= -1; goto ex;}
   if(ret==0)
     goto resume_state_0;
   if(i>i_start)
     Sregex_md5(&(s->blocklist_md5_ctx),buf+i_start,i-i_start,NULL,0);
   if(ret==2)
     Blksum_list_set_sizes(s->blocklist,hvec,hvec,hvec,1);
 }
 if(s->state==18) {
   i_start= i;
   ret= Checkrecord_tag_scan_ivecline(s,"complete_blocks=",hvec,
                                      buf,bufcount,&i,verbous&15);
   if(ret<0) 
     {ret= -1; goto ex;}
   if(ret==0)
     goto resume_state_0;
   if(i>i_start)
     Sregex_md5(&(s->blocklist_md5_ctx),buf+i_start,i-i_start,NULL,0);
   if(ret==2) {
     Blksum_list_set_sizes(s->blocklist,hvec,hvec,hvec,2);
     Intvec_copy(hvec,CdbplivS,s->blocks_to_read,0);
   }
 }
 if(s->state==19) {
   i_start= i;
   ret= Checkrecord_tag_scan_ivecline(s,"remainder=",hvec,
                                      buf,bufcount,&i,verbous&15);
   if(ret<0) 
     {ret= -1; goto ex;}
   if(ret==0)
     goto resume_state_0;
   if(i>i_start)
     Sregex_md5(&(s->blocklist_md5_ctx),buf+i_start,i-i_start,NULL,0);
   if(ret==2) {
     Blksum_list_set_sizes(s->blocklist,hvec,hvec,hvec,4);
     if(Intvec_sign(hvec,CdbplivS,0)>0)
       Intvec_add_b30(s->blocks_to_read,CdbplivS,1,0);
   }
 }
 if(s->state==20) {
   /* block data (using block_count,block_bytes) */
   i_start= i;
   for(;
       Intvec_cmp(s->block_count,CdbplivS,s->blocks_to_read,0)<0;
       Intvec_add_b30(s->block_count,CdbplivS,1,0)) {
     for(;s->block_bytes<16 && i<bufcount;s->block_bytes++)
       s->md5buffer[s->block_bytes%16]= buf[i++];
     if(s->block_bytes>=16) {
       ret= Blksum_list_add(s->blocklist,s->md5buffer,0);
       if(ret<=0) 
         {ret= -1; goto ex;}
       Sregex_md5(&(s->blocklist_md5_ctx),s->md5buffer,16,NULL,0);
       s->block_bytes= 0;
     }
     if(i>=bufcount)
   break;
   }
   if(Intvec_cmp(s->block_count,CdbplivS,s->blocks_to_read,0)>=0)
     s->state++;
 }
 if(s->state==21) {
   if(i<bufcount) {
     if(buf[i]!='\n') {
       if(verbous>=1)
         fprintf(stderr,"Block checksum list without trailing newline\n");
       goto resume_state_0;
     }
   }
   i++;
   s->state++;
 }
 if(s->state==22) {
   if(verbous>=7)
     fprintf(stderr,"Block checksum list stage %d\n",s->state);
   /* md5 of blocklist */
   ret= Checkrecord_tag_scan_ivecline(s,"md5_of_list=",hvec,
                                      buf,bufcount,&i,16|(verbous&15));
   if(ret<0) 
     {ret= -1; goto ex;}
   if(ret==0)
     goto resume_state_0;
   if(ret==2) {
     char *md5_text_pt;

     md5_text_pt= s->linebuffer+strlen("md5_of_list=");
     ret= Blksum_list_set_listmd5(s->blocklist,md5_text_pt,0);
     if(ret<=0) {
       if(verbous>=1) {
         fprintf(stderr,"Unusable MD5 text in header\n");
         if(verbous>=7)
           fprintf(stderr,"%s\n",md5_text_pt);
         fprintf(stderr,"Expected : 32 hex characters\n");
       }
       goto resume_state_0;
     }
     /* compare with computed MD5 of list */;
     Sregex_md5(&(s->blocklist_md5_ctx),NULL,0,s->md5buffer,2|(1<<15));
     Md5_from_text(md5,md5_text_pt,0);
     for(k=0;k<16;k++) {
       if(md5[k]!=((unsigned char *) s->md5buffer)[k]) {
         if(verbous>0) 
           fprintf(stderr,"Block checksum list with inner MD5 mismatch.\n");
         goto resume_state_0;
       }
     }
     s->linebuffer[0]= 0;
   }
 }
 if(s->state==23) {
   static char eol[]= {"end of list"};
   static char ctf[]= {"chained tag follows"};

   /* end of list\n */
   ret= Checkrecord_tag_scan_line(s,buf,bufcount,&i,0);
   if(ret<0) 
     {ret= -1; goto ex;}
   if(ret==2) {
     if(strncmp(s->linebuffer,eol,strlen(eol))==0) {
       ;
     } else if(strncmp(s->linebuffer,ctf,strlen(ctf))==0) {
       /* ??? */;
     } else {
       if(verbous>=1) {
         fprintf(stderr,"Unexpected header line\n");
         if(verbous>=7)
           fprintf(stderr,"%s\n",s->linebuffer);
         fprintf(stderr,"Expected : \"%s\" or \"%s\"\n",eol,ctf);
       }
       goto resume_state_0;
     }
     s->linebuffer[0]= 0;
   }
   if(ret==0)
     goto resume_state_0;
 }

 if(s->state==24) {
   /* blocklist is complete and valid . transfer to checkrecord */
   if(s->pending_record==NULL) {
     fprintf(stderr,"PROGRAMM ERROR : %s\n",
                    "checksum tag state 23 without a valid checksum record");
     {ret= -1; goto ex;}
   }
   Checkrecord_set_blocklist(s->pending_record,s->blocklist,0);
   s->blocklist= NULL;
   if(verbous>=7) {
     Intvec_to_dec(s->blocks_to_read,CdbplivS,size_text,sizeof(size_text),0);
     fprintf(stderr,"Block checksum list read with %s blocks.\n",size_text);
   }
   s->state= 2000000000;
 }

 if(s->state==32) /* tailcutter mode */
   {ret= 1; goto ex;}

 if(s->state==2000000000) {
   /* this record is done */
   Checkrecord_link(s->pending_record,*objpt,0);
   *objpt= s->pending_record;
   s->pending_record= NULL;
   Checkrecord_tag_reset(s,0);
   if(flag&2)
     {ret= 1; goto ex;}
   goto resume_state_0;
 }


 ret= 0;
ex:;
 *bufpos= i;
 if(s!=NULL)
   if((s->state>0 || s->vid_pos>0) && !s->hash_valid && !look_for_multmark) {
     if(s->state==0)
       i_pos= i-s->vid_pos;
     hret= Checkrecord_tag_clone_ctx(s,ctx,*crc32,posvec,buf,i_pos,verbous&7);
     if(hret<0)
       ret= -1;
     else if(hret==0) /* eventual message was printed by function */
       Checkrecord_tag_reset(s,0);
   }
 return(ret);
}


int Checkrecord_del_minor(objpt,stop_rec,flag)
struct CheckrecorD **objpt;
struct CheckrecorD *stop_rec;
int flag;
{
 int ismatch;
 struct CheckrecorD *o_rec;

 ismatch= Checkrecord_is_tested_match(*objpt,0);
 for(o_rec= (*objpt)->prev;o_rec!=NULL && o_rec!=stop_rec;o_rec= o_rec->prev) {
   /* try to find low ranking victim */
   if(ismatch || !Checkrecord_is_tested_match(o_rec,0)) {
     Checkrecord_destroy(&o_rec,0);
     return(1);
   }
 }
 /* no victim : self removal */ 
 o_rec= (*objpt)->prev;
 Checkrecord_destroy(objpt,0);
 (*objpt)= o_rec;
 return(1);
}


int Checkrecord_get_blocklist(o,blocklist,flag)
struct CheckrecorD *o;
struct Blksum_lisT **blocklist;
int flag;
{
 *blocklist= o->blocklist;
 return(1);
}


/* ----------------- Checkrecord - ende -------- */


/* ----------------- Fileliste ------------------ */


/* extension record for checksum */
struct Filelist_crc_reC {
 unsigned long crc;
 unsigned char *md5;
};


int Filelist_crc_rec_new(o,flag)
struct Filelist_crc_reC **o;
int flag;
/*
 bit0= allocate storage for MD5
*/
{
 *o= TSOB_FELD(struct Filelist_crc_reC,1);
 if(o==NULL)
   goto ex;
 (*o)->crc= 0;
 (*o)->md5= NULL;
 if(flag&1) {
   (*o)->md5= TSOB_FELD(unsigned char,16);
   if((*o)->md5==NULL)
     goto ex;
 }
 return(1);
ex:;
 Filelist_crc_rec_destroy(o,0);
 return(-1);
}


int Filelist_crc_rec_destroy(o,flag)
struct Filelist_crc_reC **o;
int flag;
{
 if(*o==NULL)
   return(0);
 if((*o)->md5!=NULL)
   free((char *) (*o)->md5);
 free((char *) *o);
 *o= NULL;
 return(1);
}


/* extension record for file properties */
struct Filelist_xprop_reC {
 int mod;
 char *own; 
 char *grp;
 time_t date;
};


int Filelist_xprop_rec_new(o,flag)
struct Filelist_xprop_reC **o;
int flag;
{
 *o= TSOB_FELD(struct Filelist_xprop_reC,1);
 if((*o)==NULL)
   return(-1);
 (*o)->mod= -1;
 (*o)->own= NULL;
 (*o)->grp= NULL;
 (*o)->date= 0;
 return(1);
}


int Filelist_xprop_rec_destroy(o,flag)
struct Filelist_xprop_reC **o;
int flag;
{
 if(*o==NULL)
   return(0);
 if((*o)->own!=NULL)
   free((char *) (*o)->own);
 if((*o)->grp!=NULL)
   free((char *) (*o)->grp);
 free((char *) *o);
 *o= NULL;
 return(1);
}


/* ------------------------- end Filelist_xprop_reC ----------------- */


struct Filelist_part_reC {
 int partno;
 int partcount;

 double start_byte;
 double byte_count;

 double original_size;

 int tempno;
 char tempadr[4096];
};


int Filelist_part_rec_new(o,partno,partcount,start_byte,byte_count,
                          original_size, flag)
struct Filelist_part_reC **o;
int partno;
int partcount;
double start_byte;
double byte_count;
double original_size;
int flag;
{
 *o= TSOB_FELD(struct Filelist_part_reC,1);
 if((*o)==NULL)
   return(-1);
 (*o)->partno= partno;
 (*o)->partcount= partcount;
 (*o)->start_byte= start_byte;
 (*o)->byte_count= byte_count;
 (*o)->original_size= original_size;
 (*o)->tempno= 0;
 (*o)->tempadr[0]= 0;
 return(1);
}


int Filelist_part_rec_destroy(o,flag)
struct Filelist_part_reC **o;
int flag;
{
 if((*o)==NULL)
   return(0);
 free((char *) *o);
 *o= NULL;
 return(1);
}


int Filelist_part_rec_clone(o,from,flag)
struct Filelist_part_reC **o;
struct Filelist_part_reC *from;
int flag;
{
 int ret;

 if(from==NULL) {
   *o= NULL;
   return(0);
 }
 ret= Filelist_part_rec_new(o,from->partno,from->partcount,
                            from->start_byte,from->byte_count,
                            from->original_size, 0);
 if(ret<=0)
   return(-1);
 (*o)->tempno= from->tempno;
 strcpy((*o)->tempadr,from->tempadr);
 return(1);
}


int Filelist_part_rec_get_byte_count(rec,byte_count,flag)
struct Filelist_part_reC *rec;
double *byte_count;
int flag;
{
 if(rec==NULL)
   return(0);
 *byte_count= rec->byte_count;
 return(1);
}


int Filelist_part_rec_set_tempno(rec,tempno,flag)
struct Filelist_part_reC *rec;
int tempno;
int flag;
/*
 bit0= set only if rec->tempno==0
*/
{
 if(rec==NULL)
   return(0);
 if(flag&1)
   if(rec->tempno>0)
     return(2);
 rec->tempno= tempno;
 return(1);
}


int Filelist_part_rec_get_tempno(rec,tempno,flag)
struct Filelist_part_reC *rec;
int *tempno;
int flag;
{
 if(rec==NULL)
   return(0);
 *tempno= rec->tempno;
 return(1);
}


int Filelist_part_rec_get_rec(rec,partno,partcount,
                                  start_byte,byte_count,flag)
struct Filelist_part_reC *rec;
int *partno;
int *partcount;
double *start_byte;
double *byte_count;
int flag;
{
 if(rec==NULL)
   return(0);
 *partno= rec->partno;
 *partcount= rec->partcount;
 *start_byte= rec->start_byte;
 *byte_count= rec->byte_count;
 return(1);
}


int Filelist_part_rec_is_equal(rec1,rec2,flag)
struct Filelist_part_reC *rec1;
struct Filelist_part_reC *rec2;
int flag;
/*
flag: What to compare
 bit0= partno
 bit1= partcount
 bit2= start_byte
 bit3= byte_count
*/
{
 if(rec1==NULL || rec2==NULL) {
   if(rec1!=rec2)
     return(0);
   return(1);
 }
 if(flag&1) 
   if(rec1->partno!=rec2->partno)
     return(0);
 if(flag&2) 
   if(rec1->partcount!=rec2->partcount)
     return(0);
 if(flag&4) 
   if(rec1->start_byte!=rec2->start_byte)
     return(0);
 if(flag&8) 
   if(rec1->byte_count!=rec2->byte_count)
     return(0);
 return(1);
}


int Filelist_part_rec_make_tempname(rec,prefix,suffix,filename,tempname,
                                    impl_igno, flag)
struct Filelist_part_reC *rec;
char *prefix;
char *suffix;
char *filename;
char tempname[4096];
char *impl_igno;
int flag;
/*
 bit0= append "-"tempno
 bit1= use basename only
 bit2= replace characters of impl_igno by their hexcodes
 bit3= -xorriso_cut_out : v0_part_1_of_3_at_0_with_2047m_of_5753194821
       0588: "v0" will keep xorriso prior to 0.4.4 from mistaking this as
       split file directory, because then the varying partno is mistaken as
       totalno and causes non-recognizing.
*/
{
 int i,limit= 1;
 char *cpt,form[80];

 tempname[0]= 0;
 if(rec==NULL)
   return(0);
 if((flag&2) || prefix[0]!=0) {
   cpt= strrchr(filename,'/');
   if(cpt!=NULL)
     cpt++;
   else
     cpt= filename;
   if(cpt>filename && !(flag&2)) {
     strncpy(tempname,filename,cpt-filename);
     tempname[cpt-filename]= 0;
   }
   strcat(tempname,prefix);
 } else
   cpt= filename;
 strcat(tempname,cpt);
 strcat(tempname,suffix);

 /* 0497: eventually replace evil characters */;
 if(flag&4) {
   char new_name[4096],*rpt,*wpt;

   wpt= new_name;
   for(rpt=tempname;*rpt!=0 && wpt-new_name<sizeof(new_name)-4;rpt++)
      if(strchr(impl_igno,*rpt)!=NULL){
        sprintf(wpt,"+%-2.2X+",*rpt);
        wpt+= strlen(wpt);
      } else
        *(wpt++)= *rpt;
   *wpt= 0;
   strcpy(tempname,new_name);
 }

 for(i=1;i<10;i++) {
   limit*= 10;
   if(rec->partcount<limit)
 break;
 }
 if(flag&8) {
   if(i<10)
     sprintf(form,"/v0_part_%%%d.%dd_of_%%%d.%dd",i,i,i,i);  
   else
     strcpy(form,"/v0_part_%d_of_%d");
   sprintf(tempname+strlen(tempname),form,rec->partno,rec->partcount);
   sprintf(tempname+strlen(tempname),"_at_%.f_with_%.f_of_%.f",
           rec->start_byte, rec->byte_count, rec->original_size);
 } else {
   if(i<10)
     sprintf(form,"_%%%d.%dd_%%%d.%dd",i,i,i,i);  
   else
     strcpy(form,"_%d_%d");
   sprintf(tempname+strlen(tempname),form,rec->partno,rec->partcount);
   if(flag&1)
     sprintf(tempname+strlen(tempname),"-%2.2X",rec->tempno);
 }
 return(1);
}


int Filelist_part_rec_make_tempadr(rec,filename,split_file_dir,
                                   split_file_prefix,split_file_suffix,
                                   impl_igno,flag)
struct Filelist_part_reC *rec;
char *filename;
char *split_file_dir;
char *split_file_prefix;
char *split_file_suffix;
char *impl_igno;
int flag;
/*
 bit2= replace characters of impl_igno by their hexcodes
*/
{
 int ret;
 char tempname[4096];

 ret= Filelist_part_rec_make_tempname(rec,split_file_prefix,split_file_suffix,
                                    filename,tempname,impl_igno, 1|2|(flag&4));
 if(ret<=0)
   return(ret);
 sprintf(rec->tempadr,"%s/%s",split_file_dir,tempname);
 return(1);
}


int Filelist_part_rec_get_tempadr(rec,adr,flag)
struct Filelist_part_reC *rec;
char adr[4096];
int flag;
{
 strcpy(adr,rec->tempadr);
 return(1);
}


/* ------------------------- end Filelist_part_reC ----------------- */


/* This factor speeds up linear search in a list tree for the price of
   sizeof(int)*Fileliste_hash_sizE additional bytes in every directory node */
static int Fileliste_hash_sizE= 20;
static int Fileliste_hash_size_frozeN= 0; /* after first hashed list is
                                             created there may be no change */
typedef char Fileliste_typ_T;
typedef char Fileliste_flag_T;

struct FilelistE {
 int ref_count;
 struct FilelistE *boss;
 int boss_idx;

 char *dir;
 dev_t dir_device;
 ino_t dir_inode;
 int dir_dev_ino_valid;        /* bit0= dir_device valid
                                  bit1= dir_inode valid */

 char *dir_size_cmd;
 int dir_size_internal;
 char *target_dir;
 char *main_dir;

 struct FilelistE *sperrliste;

 char target_inp_trenner;  /* english : target_inp_separator */
 char *impl_igno;          /* filename characters which cause the backup
                              command to ignore files if they are not given
                              explicitely. (For mkisofs-1.12b3 : ~# )
                            */ 
 int maxanz;               /* english : maxcount */
 int anz;                  /* english : count */
 int enlarge_step;

 int use_hashing;
 int *hashstart;

 char **adressen;          /* english : addresses */
 char **targets;
 Fileliste_typ_T  *typen; /* english : types */
                /* <0 deleted entry
                   0= type not detected yet (or refuses to tell)
                 compatible with return of Type_and_size(,16) :
                   1= data file 
                   2= directory 
                   3= link 
                   4= ( named pipe , not used yet)
                   5= ( multiple hardlink , probably never used )
                   6= block device
                 127= other type (already evaluated)
                 */
 Fileliste_flag_T *flags; 
                /* bit0= map to root 
                   bit1= In tree models: contains young files / is young file
                         In exclusion lists: additionally check for dev,inode
                   bit2= contains old file / is old file 
                   bit3= crcs[idx] valid
                   bit4= crcs[idx] is md5 rather than crc32
                   bit5= this may be a device-like file, 
                         determine size only once (i.e. if still <=0)
                   bit6= inodes[idx] is valid
                   bit7= within old_tree : file seen during tree crawling
                         <<< with composition :
                             currently in use for test purposes :
                             Fileliste_precheck_sub(), provisory_young
                             supervising trust_fret64
                */
 double *groessen;         /* english : sizes */

 struct Filelist_crc_reC **crcs;
 ino_t *inodes;
 struct Filelist_xprop_reC **xprops;

 struct Filelist_part_reC **part_recs;

 int *hashlinks;
 struct FilelistE **subdirs;
};

static char Fileliste_dir_size_cmd[4096]={"-internal"};  /* {"du -sk"}; */
static int Fileliste_dir_size_internal= 1;
/* estimation of management information in a backup archives directory per
   file entry . This is a generous estimation to be on the safe side.
   The number may be changed at run time by  -dir_size_cmd -internal+number  */
static double Fileliste_node_overhead= 256.0;
static int Fileliste_node_addrl= 1;

static char Fileliste_target_inp_trenner= '=';
static int Fileliste_target_modern= 0;
static char Fileliste_impl_igno[256]= {""};

double  Fileliste_size();
double  Fileliste_dir_size();
double  Fileliste_estimate_mem_usage();


int Fileliste_new(liste,maxanz,main_dir,flag)
struct FilelistE **liste;
int maxanz;
char *main_dir;
int flag;
/*
 bit0= list shall use hashing
 bit1= make economy class list = bits 8, 9, 10, 11, 13

 bit8= without xprops vector
 bit9= without hashlinks
 bit10= without subdirs
 bit11= without crcs
 bit12= without ...
 bit13= without inodes
 bit14= without ...
 bit15= without ...
*/
{
 if(flag&2)
   flag|= (256|512|1024|(1<<11)|(1<<13));

 *liste= TSOB_FELD(struct FilelistE,1);
 if(*liste==NULL) return(0);
 (*liste)->ref_count= 0;
 (*liste)->boss= NULL;
 (*liste)->boss_idx= 0;
 (*liste)->dir= NULL;
 (*liste)->dir_device= 0;
 (*liste)->dir_inode= 0;
 (*liste)->dir_dev_ino_valid= 0;
 (*liste)->dir_size_cmd= NULL;
 (*liste)->dir_size_internal= 0;
 (*liste)->target_dir= NULL;
 (*liste)->main_dir= NULL;
 (*liste)->sperrliste= NULL;
 (*liste)->target_inp_trenner= Fileliste_target_inp_trenner;
 (*liste)->impl_igno= NULL;
 (*liste)->use_hashing= ((flag&1)&&!(Fileliste_hash_sizE<=0));
 (*liste)->hashstart= NULL;
 (*liste)->adressen= NULL;
 (*liste)->targets= NULL;
 (*liste)->typen= NULL;
 (*liste)->flags= NULL;
 (*liste)->groessen= NULL;
 (*liste)->crcs= NULL;
 (*liste)->inodes= NULL;
 (*liste)->xprops= NULL;
 (*liste)->part_recs= NULL;
 (*liste)->hashlinks= NULL;
 (*liste)->subdirs= NULL;
 (*liste)->maxanz= 0;

 if(Sregex_string(&((*liste)->dir),"",0)<=0)
   goto wiederweg;
 if(Sregex_string(&((*liste)->dir_size_cmd),Fileliste_dir_size_cmd,0)<=0)
   goto wiederweg;
 (*liste)->dir_size_internal= Fileliste_dir_size_internal;
 if(Sregex_string(&((*liste)->target_dir),"",0)<=0)
   goto wiederweg;
 if(Sregex_string(&((*liste)->main_dir),main_dir,0)<=0)
   goto wiederweg;
 (*liste)->impl_igno= TSOB_FELD(char,strlen(Fileliste_impl_igno)+1);
 strcpy((*liste)->impl_igno,Fileliste_impl_igno);
 (*liste)->maxanz= maxanz;
 (*liste)->anz= 0;
 (*liste)->enlarge_step= 10;
 if(Fileliste_new_vectors(*liste,maxanz,flag&0xff00)<=0)
   goto wiederweg;
 return(1); 
wiederweg:;
 Fileliste_destroy(liste,1);
 return(0);
}


int Fileliste_creation_flag(liste,flag)
/* determine creation flag settings of an existing list */
struct FilelistE *liste;
int flag;
{
 int ret= 0;

 if(liste->use_hashing && liste->hashstart!=NULL)
   ret|= 1;
 if(liste->xprops==NULL)
   ret|= 256;
 if(liste->hashlinks==NULL)
   ret|= 512;
 if(liste->subdirs==NULL)
   ret|= (1<<10);
 if(liste->crcs==NULL)
   ret|= (1<<11);
 if(liste->inodes==NULL)
    ret|= (1<<13);
 return(ret);
}


int Fileliste_init_vectors(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i;

 if(liste->hashstart!=NULL)
   for(i=0;i<Fileliste_hash_sizE;i++)
     liste->hashstart[i]= -1;
 for(i=0;i<liste->maxanz;i++) {
   liste->adressen[i]= NULL;
   liste->targets[i]= NULL;
   liste->typen[i]= 0;
   liste->flags[i]= 0;
   liste->groessen[i]= -1;
   if(liste->crcs!=NULL)
     liste->crcs[i]= NULL;
   if(liste->inodes!=NULL)
     liste->inodes[i]= 0;
   if(liste->xprops!=NULL)
     liste->xprops[i]= NULL;
   if(liste->hashlinks!=NULL)
     liste->hashlinks[i]= -1;
   if(liste->subdirs!=NULL)
     liste->subdirs[i]= NULL;
 }
 if(liste->part_recs!=NULL)
   for(i=0;i<liste->maxanz;i++) 
     liste->part_recs[i]= NULL;
 return(1);
}


int Fileliste_new_vectors(liste,maxanz,flag)
struct FilelistE *liste;
int maxanz;
int flag;
/*
 bit8-15= without some vector, see Fileliste_new()
*/
{
 Fileliste_destroy_vectors(liste,0);
 if(liste->use_hashing && Fileliste_hash_sizE>0) {
   liste->hashstart= TSOB_FELD(int, Fileliste_hash_sizE);
   if(liste->hashstart==NULL)
     return(-1);
   Fileliste_hash_size_frozeN= 1;
 } else
   liste->use_hashing= 0;
 if(maxanz<=0) {
/*
   Fileliste_init_vectors(liste,0);
   return(2);
*/
   maxanz= 1;
 }
 liste->adressen= TSOB_FELD(char *,maxanz);
 if(liste->adressen==NULL)
   return(-1);
 liste->targets= TSOB_FELD(char *,maxanz);
 if(liste->targets==NULL)  
   return(-1);
 liste->typen= TSOB_FELD(Fileliste_typ_T,maxanz);
 if(liste->typen==NULL)  
   return(-1);
 liste->flags= TSOB_FELD(Fileliste_flag_T,maxanz);
 if(liste->flags==NULL)  
   return(-1);
 liste->groessen= TSOB_FELD(double,maxanz);
 if(liste->groessen==NULL)  
   return(-1);
 if(!(flag&(1<<11))) {
   liste->crcs= TSOB_FELD(struct Filelist_crc_reC *,maxanz);
   if(liste->crcs==NULL)  
     return(-1);
 }
 if(!(flag&(1<<13))) {
   liste->inodes= TSOB_FELD(ino_t,maxanz);
   if(liste->inodes==NULL)
     return(-1);
 }
 if(!(flag&256)) {
   liste->xprops= TSOB_FELD(struct Filelist_xprop_reC *,maxanz);
   if(liste->xprops==NULL)  
     return(-1);
 }

 /* No new part_recs vector here. This is created on demand only. */

 if(!(flag&512)) {
   liste->hashlinks= TSOB_FELD(int,maxanz);
   if(liste->hashlinks==NULL)  
     return(-1);
 }
 if(!(flag&1024)) {
   liste->subdirs= TSOB_FELD(struct FilelistE *,maxanz);
   if(liste->subdirs==NULL)  
     return(-1);
 }
 liste->maxanz= maxanz;
 Fileliste_init_vectors(liste,0);
 return(1);
}


double Fileliste_estimate_mem_usage(liste,flag)
struct FilelistE *liste;
int flag;
/*
 bit0= whole liste tree : add size of all subdirs
*/
{
 int anz_fak= 0,i;
 double size;

 if(liste==NULL)
   return(0.0);
 size= sizeof(struct FilelistE);
 if(liste->dir!=NULL)
   size+= strlen(liste->dir)+1;
 if(liste->dir_size_cmd!=NULL)
   size+= strlen(liste->dir_size_cmd)+1; 
 if(liste->target_dir!=NULL)
   size+= strlen(liste->target_dir)+1; 
 if(liste->main_dir!=NULL)
   size+= strlen(liste->main_dir)+1; 
 if(liste->impl_igno!=NULL)
   size+= strlen(liste->impl_igno)+1; 
 if(liste->hashstart!=NULL)
   size+= sizeof(int)*Fileliste_hash_sizE; 

 anz_fak= 2*sizeof(char *);
 anz_fak+= sizeof(Fileliste_typ_T)+sizeof(Fileliste_flag_T);
 anz_fak+= sizeof(double);
 if(liste->crcs!=NULL)
   anz_fak+= sizeof(struct Filelist_crc_reC *); 
 if(liste->inodes!=NULL)
   anz_fak+= sizeof(ino_t); 
 if(liste->xprops!=NULL)
   anz_fak+= sizeof(struct Filelist_xprop_reC *); 
 if(liste->part_recs!=NULL)
   anz_fak+= sizeof(struct Filelist_part_reC *); 
 if(liste->hashlinks!=NULL)
   anz_fak+= sizeof(int); 
 if(liste->subdirs!=NULL)
   anz_fak+= sizeof(struct FilelistE *); 
 size+= liste->maxanz*anz_fak;

 for(i=0;i<liste->anz;i++) {
   if(liste->adressen[i]!=NULL)
     size+= strlen(liste->adressen[i])+1;
   if(liste->targets[i]!=NULL)
     size+= strlen(liste->targets[i])+1;
   if(liste->crcs!=NULL)
     if(liste->crcs[i]!=NULL) {
       size+= sizeof(struct Filelist_crc_reC);
       if(liste->crcs[i]->md5!=0)
         size+= 16;
     }
   if(liste->xprops!=NULL)
     if(liste->xprops[i]!=NULL) {
       size+= sizeof(struct Filelist_xprop_reC);
       if(liste->xprops[i]->own!=NULL)
         size+= strlen(liste->xprops[i]->own)+1;
       if(liste->xprops[i]->grp!=NULL)
         size+= strlen(liste->xprops[i]->grp)+1;
     }
   if(liste->part_recs!=NULL)
     if(liste->part_recs[i]!=NULL)
       size+= sizeof(struct Filelist_part_reC);
   if(flag&1)
     if(liste->subdirs!=NULL)
       size+= Fileliste_estimate_mem_usage(liste->subdirs[i],1);
 }
 return(size);
}


int Fileliste_assert_subdirs(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i;

 if(liste==NULL)
   return(0);
 if(liste->subdirs!=NULL)
   return(1);
 if(liste->maxanz<=0)
   return(0);
 liste->subdirs= TSOB_FELD(struct FilelistE *,liste->maxanz);
 if(liste->subdirs==NULL)  
   return(-1);
 for(i=0;i<liste->maxanz;i++)
   liste->subdirs[i]= NULL;
 return(1);
}


int Fileliste_destroy(liste,flag)
struct FilelistE **liste;
int flag;
/*
 bit0= the list is not finished (this bit does not do anything yet)
 bit1= deletion of a dependend list : lower ref_count
 bit2= this is a subdirs[] and possibly subject to subs counting
*/
{
 if(*liste==NULL)
   return(0);
 if(flag&2)
   if((*liste)->ref_count>0)
     ((*liste)->ref_count)-- ;
 if((*liste)->ref_count>0){
   /* This list is still used by another list */
   return(1);
 }
 if((flag&4) && !Cdbp_compose_do_not_reuse)
   Fileliste_count_sub(*liste,0,
                       &Cdbp_subs,&Cdbp_subs_lines,&Cdbp_subs_mem,1|2);
 Fileliste_destroy_vectors(*liste,0);
 if((*liste)->impl_igno!=NULL)
    free((char *) (*liste)->impl_igno);
 /* (*liste)->sperrliste was not created by Fileliste_new() */
 if((*liste)->target_dir!=NULL)
   free((char *) (*liste)->target_dir);
 if((*liste)->main_dir!=NULL)
   free((char *) (*liste)->main_dir);
 if((*liste)->dir_size_cmd!=NULL)
   free((char *) (*liste)->dir_size_cmd);
 if((*liste)->dir!=NULL)
   free((char *) (*liste)->dir);
 if((*liste)!=NULL) 
   free((char *) (*liste));
 *liste= NULL;
 return(1);
}


int Fileliste_destroy_vectors(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i;

 if(liste->subdirs!=NULL) {
   for(i=0;i<liste->anz;i++)
     if(liste->subdirs[i]!=NULL)
       Fileliste_destroy(&(liste->subdirs[i]),2|4);
   free((char *) liste->subdirs);
   liste->subdirs= NULL;
 }
 if(liste->hashlinks!=NULL)
   free((char *) liste->hashlinks);
 liste->hashlinks= NULL;
 if(liste->part_recs!=NULL) {
   for(i=0;i<liste->anz;i++)
     Filelist_part_rec_destroy(&(liste->part_recs[i]),0);
   free((char *) liste->part_recs);
   liste->part_recs= NULL;
 }
 if(liste->crcs!=NULL) {
   for(i=0;i<liste->anz;i++)
     Filelist_crc_rec_destroy(&(liste->crcs[i]),0);
   free((char *) liste->crcs);
   liste->crcs= NULL;
 }
 if(liste->inodes!=NULL)
   free((char *) liste->inodes);
 liste->inodes= NULL;
 if(liste->xprops!=NULL) {
   for(i=0;i<liste->anz;i++)
     Filelist_xprop_rec_destroy(&(liste->xprops[i]),0);
   free((char *) liste->xprops);
   liste->xprops= NULL;
 }
 if(liste->groessen!=NULL)
   free((char *) liste->groessen);
 liste->groessen= NULL;
 if(liste->flags!=NULL)
   free((char *) liste->flags);
 liste->flags= NULL;
 if(liste->typen!=NULL)
   free((char *) liste->typen);
 liste->typen= NULL;
 if(liste->targets!=NULL){
   for(i=0;i<liste->anz;i++)
     if(liste->targets[i]!=NULL)
       free(liste->targets[i]);
   free((char *) liste->targets);
   liste->targets= NULL;
 }
 if(liste->adressen!=NULL) {
   for(i=0;i<liste->anz;i++)
     if(liste->adressen[i]!=NULL)
       free(liste->adressen[i]);
   free((char *) liste->adressen);
   liste->adressen= NULL;
 }
 if(liste->hashstart!=NULL)
   free((char *) liste->hashstart);
 liste->hashstart= NULL;
 liste->anz= 0;
 liste->maxanz= 0;
 return(1);
}


int Fileliste_destroy_item(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
/*
 bit10= do not destroy eventual subdirs[idx] but only detach
*/
{
 if(liste->adressen[idx]!=NULL)
   free(liste->adressen[idx]);
 liste->adressen[idx]= NULL;
 if(liste->targets[idx]!=NULL)
   free(liste->targets[idx]);
 liste->targets[idx]= NULL;
 if(liste->subdirs!=NULL) 
   if(liste->subdirs[idx]!=NULL) {
     if(flag&(1<<10)) {
       if(liste->subdirs[idx]->ref_count>0)
         liste->subdirs[idx]->ref_count--;
       liste->subdirs[idx]= NULL;
     } else
       Fileliste_destroy(&(liste->subdirs[idx]),2);
   }
 if(liste->crcs!=NULL)
   Filelist_crc_rec_destroy(&(liste->crcs[idx]),0);
 if(liste->xprops!=NULL)
   Filelist_xprop_rec_destroy(&(liste->xprops[idx]),0);
 Filelist_destroy_part_rec(liste,idx,0);
 liste->typen[idx]= 0;
 liste->flags[idx]= 0;
 return(1);
}


int Fileliste_enlarge(liste,maxanz,flag)
struct FilelistE *liste;
int maxanz;
int flag;
{
 int i,ti,ret,hflag;
 struct FilelistE *neuliste;
 char *tr; /* should be (void *) */

 static double acceleration= 0.2;
 static int wink= 0;

 if(liste->maxanz*acceleration>liste->enlarge_step)
   liste->enlarge_step= liste->maxanz*acceleration;
 i= (liste->enlarge_step)*(maxanz/(liste->enlarge_step));
 if(i<maxanz) i+= liste->enlarge_step;
 maxanz= i;

 if(liste->maxanz>=maxanz || maxanz<=0) return(1);

 if(wink)
   fprintf(stderr,"%d -> %d",liste->maxanz,maxanz);

 /* create new list  */
 hflag= Fileliste_creation_flag(liste,0);
 if(Fileliste_new(&neuliste,maxanz,liste->main_dir,hflag)<=0)
   {ret= -1; goto ex;}

 /* copy old content */
 for(i=0;i<liste->maxanz;i++)
   Fileliste_copy_item(liste,i,neuliste,i,2);

 /* swap data handles */

 tr= (char *) liste->adressen;
 liste->adressen= neuliste->adressen;
 neuliste->adressen= (char **) tr;

 tr= (char *) liste->targets;
 liste->targets= neuliste->targets;
 neuliste->targets= (char **) tr;

 tr= (char *) liste->typen;
 liste->typen= neuliste->typen;
 neuliste->typen= (Fileliste_typ_T *) tr;

 tr= (char *) liste->flags;
 liste->flags= neuliste->flags;
 neuliste->flags= (Fileliste_flag_T *) tr;

 tr= (char *) liste->groessen;
 liste->groessen= neuliste->groessen;
 neuliste->groessen= (double *) tr;

 tr= (char *) liste->crcs;
 liste->crcs= neuliste->crcs;
 neuliste->crcs= (struct Filelist_crc_reC **) tr;

 tr= (char *) liste->inodes;
 liste->inodes= neuliste->inodes;
 neuliste->inodes= (ino_t *) tr;

 tr= (char *) liste->xprops;
 liste->xprops= neuliste->xprops;
 neuliste->xprops= (struct Filelist_xprop_reC **) tr;

 tr= (char *) liste->part_recs;
 liste->part_recs= neuliste->part_recs;
 neuliste->part_recs= (struct Filelist_part_reC **) tr;

 tr= (char *) liste->hashlinks;
 liste->hashlinks= neuliste->hashlinks;
 neuliste->hashlinks= (int *) tr;

 if(liste->hashstart!=NULL && neuliste->hashstart!=NULL) {
   for(i=0;i<Fileliste_hash_sizE;i++) {
     ti= liste->hashstart[i];
     liste->hashstart[i]= neuliste->hashstart[i];
     neuliste->hashstart[i]= ti;
   }
 }

 tr= (char *) liste->subdirs;
 liste->subdirs= neuliste->subdirs;
 neuliste->subdirs= (struct FilelistE **) tr;

 /* update back pointers of sub lists */
 if(liste->subdirs!=NULL) {
   for(i=0;i<liste->maxanz;i++)
     if(liste->subdirs[i]!=NULL) {
       liste->subdirs[i]->boss= liste;
       liste->subdirs[i]->boss_idx= i;
     }
 }
 neuliste->maxanz= liste->maxanz;
 liste->maxanz= maxanz;  

 /* delete temporary list object */
 Fileliste_destroy(&neuliste,0);

 ret= 1;
ex:;
 if(wink)
   fprintf(stderr,"\r                                    \r");
 return(ret);
}


int Fileliste_resize(liste,maxanz,flag)
struct FilelistE *liste;
int maxanz;
int flag;
{
 int hflag;

 liste->anz= 0;

 if(maxanz>liste->maxanz || liste->adressen==NULL ||
      liste->targets==NULL || liste->flags==NULL ||
      liste->typen==NULL || liste->groessen==NULL) {
   /* liste->maxanz= maxanz; */
   hflag= Fileliste_creation_flag(liste,0);
   if(Fileliste_new_vectors(liste,maxanz,hflag&0xff00)<=0) 
    return(-1);
 } else {
   Fileliste_init_vectors(liste,0);
 }
 return(1);
}


int Fileliste__compute_hashvalue(adr,flag)
char *adr;
int flag;
{
 unsigned long crc;

 if(Fileliste_hash_sizE<=0)
   return(Fileliste_hash_sizE);
 Sregex_crc32_yenc(adr,strlen(adr),&crc,1);
 crc= crc%Fileliste_hash_sizE;
 return((int) crc);
}


int Fileliste_hashlink(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
/*
 bit0= unlink
*/
{
 static int programmer_notified= 0;
 int start_idx,pred_idx,*pred_pt;

 if(liste->use_hashing==0 || liste->hashstart==NULL || Fileliste_hash_sizE<=0)
   return(2);
 if(liste->adressen[idx]==NULL)
   return(2);
 if(liste->hashlinks==NULL) {
   if(programmer_notified<10) {
     fprintf(stderr,
          "??? programmer notification : missing hashlinks vector on link\n");
     programmer_notified++;
   }
 }

 start_idx= Fileliste__compute_hashvalue(liste->adressen[idx],0);
 if(flag&1) {
   /* search predecessor in hash list */
   pred_pt= &liste->hashstart[start_idx];
   pred_idx= *pred_pt;
   while(pred_idx>=0) { 
     if(liste->hashlinks[pred_idx]<0) {
       pred_idx= -1;
   break;
     }
     if(liste->hashlinks[pred_idx]==idx)
   break;
     pred_pt= &(liste->hashlinks[pred_idx]);
     pred_idx= *pred_pt;
   }
   /* set predecessor's hashlink to successor of idx */
   if(pred_idx<0)
     return(0);  
   *pred_pt= liste->hashlinks[idx];
 } else {
   liste->hashlinks[idx]= liste->hashstart[start_idx];
   liste->hashstart[start_idx]= idx;
 }
 return(1);
}


int Fileliste_set_address(liste,idx,address,flag)
struct FilelistE *liste;
int idx;
char *address;
int flag;
{
 if(liste->anz<=idx)
   return(0);
 if(liste->adressen[idx]!=NULL)
   free((char *) liste->adressen[idx]);
 liste->adressen[idx]= NULL;
 if(address!=NULL){
   liste->adressen[idx]=  TSOB_FELD(char,strlen(address)+1);
   if(liste->adressen[idx]==NULL)
     return(-1);
   strcpy(liste->adressen[idx],address);
 }
 return(1);
}


int Fileliste_set_groessen(liste,idx,value,flag)
struct FilelistE *liste;
int idx;                                           /* -1 = last item in list */
double value;
int flag;
{
 if(liste->anz<=idx)
   return(0);
 if(idx<0)
   idx= liste->anz-1;
 if(idx<0)
   return(0);
 liste->groessen[idx]= value;
 return(1);
}


int Fileliste_set_part_rec(liste,idx,rec,flag)
struct FilelistE *liste;
struct Filelist_part_reC *rec;
int idx;
int flag;
/*
 bit0= make a copy rather than just setting the pointer
 bit1= do not free existing part_rec 
*/
{
 int i,ret;

 if(idx<0||idx>=liste->maxanz)
   return(0);
 if(rec==NULL) {
   if(liste->part_recs==NULL)
     return(1);
 } else {
   if(liste->part_recs==NULL) {
     liste->part_recs= TSOB_FELD(struct Filelist_part_reC *,liste->maxanz);
     if(liste->part_recs==NULL)
       return(-1);
     for(i=0;i<liste->maxanz;i++)
       liste->part_recs[i]= NULL;
   }
 }
 if((!(flag&2)) && liste->part_recs[idx]!=NULL)
   Filelist_part_rec_destroy(&(liste->part_recs[idx]),0);
 if(rec!=NULL && (flag&1)) {
   ret= Filelist_part_rec_clone(&(liste->part_recs[idx]),rec,0);
   if(ret<=0)
     return(-1);
 } else {
   liste->part_recs[idx]= rec;
 }
 return(1);
}


int Fileliste__print_split_cleanup(fp,split_file_dir,split_file_prefix,
                                   split_file_suffix,flag)
FILE *fp;
char *split_file_dir;
char *split_file_prefix;
char *split_file_suffix;
int flag;
/*
 bit0 = write as function  split_dir_cleanup
 bit1 = write as call of function split_dir_cleanup
*/
{
 char tempname[4096],txs_buf[4096*3];

#ifdef Cdbackup_planer_xorriso_cut_ouT
/* CUT_OUT: this function prepares cleanup of the split file directory
            in the backup script.  With xorriso -cut_out this can be skipped.
*/
 if(strcmp(split_file_dir, "-xorriso_cut_out")==0)
   return(1);
#endif

 if(flag&2) {
   fprintf(fp,"split_dir_cleanup\n");
   return(1);
 }
 if(flag&1) 
   fprintf(fp,"\nsplit_dir_cleanup() {\n");

 sprintf(tempname,"%s/%s",
                  split_file_dir,"scdbackup_split_file_directory_mark");
 Text_shellsafe(tempname,txs_buf,0);
 fprintf(fp,"if test -e %s ; then\n",txs_buf);
 fprintf(fp,"  dummy=dummy\n");
 fprintf(fp,"else\n");
 fprintf(fp,"  echo >&2\n");
 fprintf(fp,
     "  echo 'Directory for split files not enabled. Cannot clean up.' >&2\n");
 fprintf(fp,"  echo 'Markfile missing : '\"%s\" >&2\n",txs_buf);
 fprintf(fp,"  echo 'CAUTION: Create this file only in a dedicated empty directory' >&2\n");
 fprintf(fp,"  echo 'WARNUNG: Diesen File nur in reserviertem, leeren Directory erzeugen' >&2\n");
 fprintf(fp,"  echo >&2\n");
 fprintf(fp,"  exit 1\n");
 fprintf(fp,"fi\n");
 Text_shellsafe(split_file_dir,txs_buf,0);
 fprintf(fp,"if ls %s/*_[0-9]*_[0-9]*-[0-9A-F]*[0-9A-F] >/dev/null 2>/dev/null ; then\n",txs_buf);
 fprintf(fp,"  rm %s/*_[0-9]*_[0-9]*-[0-9A-F]*[0-9A-F]\n",txs_buf);
 fprintf(fp,"fi\n");
 if(flag&1) 
   fprintf(fp,"return 0\n}\n\n");
 return(1);
}


int Fileliste_count_part_recs(liste,count,flag)
struct FilelistE *liste;
int *count;
int flag;
/*
 bit0= set ->tempno
 bit1= set ->tempno only if 0
*/
{
 int i,split_file_count= 0;

 if(liste->part_recs==NULL)
   return(0);
 for(i=0;i<liste->anz;i++)
   if(liste->typen[i]>=0) 
     if(liste->part_recs[i]!=NULL) {
       split_file_count++;
       if(flag&1)
         Filelist_part_rec_set_tempno(liste->part_recs[i],
                                      (*count)+split_file_count,(flag&2)>>1);
     }
 (*count)+= split_file_count;
 if(split_file_count<=0)
   return(0);
 return(split_file_count);
}


int Fileliste_make_parts(liste,fp,cut_file_cmd,split_file_dir,
                         split_file_prefix,split_file_suffix,count,flag)
struct FilelistE *liste;
FILE *fp;
char *cut_file_cmd;
char *split_file_dir;
char *split_file_prefix;
char *split_file_suffix;
int *count;
int flag;
/*
 bit0 = write function  split_dir_cleanup
 bit1 = write call of function split_dir_cleanup
 bit2= replace characters of liste->impl_igno by their hexcodes
 
return:
 <0 = error
  0 = no part_recs in list
  1 
*/
{
 int i,ret,partno,partcount,split_file_count= 0;
 char tempname[4096],txs_buf[4096*3];
 double start_byte,byte_count;

 split_file_count= Fileliste_count_part_recs(liste,&split_file_count,0);
 if(split_file_count<=0)
   return(0);

#ifdef Cdbackup_planer_xorriso_cut_ouT
/* >>> CUT_OUT: this function writes -cut_file commands into the backup script.
       With xorriso -cut_out this can be skipped.
   ??? Open question: is *count resp. plan->split_count needed for xorriso ?
*/

 if(strcmp(split_file_dir, "-xorriso_cut_out")==0)
   return(split_file_count);
#endif

 if(split_file_count==1)
   fprintf(fp,"echo '-split_large_files : extracting 1 file part'\n");
 else
   fprintf(fp,"echo '-split_large_files : extracting %d file parts'\n",
              split_file_count);
 Fileliste__print_split_cleanup(fp,split_file_dir,split_file_prefix,
                                split_file_suffix,flag&(1|2));
 if((flag&(1|2))==1) /* Function definition written. Now write call : */
   Fileliste__print_split_cleanup(fp,split_file_dir,split_file_prefix,
                                  split_file_suffix,2);

 for(i=0;i<liste->anz;i++) {
   if(liste->typen[i]<0) 
 continue;
   if(liste->part_recs[i]!=NULL) {
     (*count)++;
     Filelist_part_rec_set_tempno(liste->part_recs[i],*count,1);
     /* print command to produce file part */
     Filelist_part_rec_get_rec(liste->part_recs[i],&partno,&partcount,
                               &start_byte,&byte_count,0);
     sprintf(tempname,"%s/",split_file_dir);
     ret= Filelist_part_rec_make_tempname(liste->part_recs[i],
                    split_file_prefix,split_file_suffix,liste->adressen[i],
                    tempname+strlen(tempname), liste->impl_igno, 1|2|(flag&4));
     Text_shellsafe(liste->adressen[i],txs_buf,0);
     fprintf(fp,"%s %s %.lf %.lf",cut_file_cmd,txs_buf,start_byte,byte_count);
     Text_shellsafe(tempname,txs_buf,0);
     fprintf(fp," %s || exit 1\n",txs_buf);
   }
 }
 return(split_file_count);
}


int Filelist_destroy_part_rec(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
{
 int ret;

 if(liste->part_recs==0)
   return(0);
 if(idx<0||idx>=liste->maxanz)
   return(0);
 ret= Filelist_part_rec_destroy(&(liste->part_recs[idx]),0);
 return(ret);
}


int Fileliste_dir(liste,dir,linecount,flag)
struct FilelistE *liste;
char *dir;
int *linecount;
int flag;
/*
 bit0= forwarded to Fileliste_dir_size() (see there)
 bit1= no need to determine sizes
 bit2= do not read properties
 bit3= try to update pacifier from time to time
 bit4= set ->groessen[idx] of unknown filetypes to -1
 bit6= print pacifier with linecount
 bit7= recursion (for Fileliste_dir_size)
*/
{
 int i,l,dirl,ret;
 struct Lies_popeN *gelesen,gelesen_data;
 char **filev= NULL;
 int filec= 0;
 static int sort_threshhold= 1000; /* don't waste time in giant directories */

 gelesen= &gelesen_data;

 for(i=0;i<liste->anz;i++)
   Sregex_string(&(liste->adressen[i]),NULL,0);
 liste->anz= 0;

 if(Sregex_string(&(liste->dir),dir,0)<=0)
   return(-1);

 dirl= strlen(dir);
 ret= Dir_ls(dir,&filec,&filev,2000000000,1|(flag&8));  /* skip "." and ".." */
 if(ret<=0)
   return(ret);
 if(Fileliste_resize(liste,(filec>0?filec:1),0)<=0)
   {ret= -1; goto ex;}

 if(filec<sort_threshhold)
   Sort_argv(filec,filev,0);

 for(i=0;i<filec;i++) {
   /* prepend dir and add to list  */
   if(dir[0]==0) {
     liste->adressen[i]= filev[i];
     filev[i]= NULL;
   } else {
     l= dirl+1+strlen(filev[i]);
     liste->adressen[i]= TSOB_FELD(char,l+1);
     if(liste->adressen[i]==NULL)
       {ret= -1; goto ex;}
     if(dir[dirl-1]!='/') 
       sprintf(liste->adressen[i],"%s/%s",dir,filev[i]);
     else
       sprintf(liste->adressen[i],"%s%s",dir,filev[i]);
     Sregex_string(&(filev[i]),NULL,0);
   }
   if((i%1000)==0 && (flag&8) && i>0)
     Print_pacifier(0,"",128|256|(1<<11));
 }
 liste->anz= filec;
 if(flag&4)
   ret= 1;
 else
   ret= Fileliste_read_properties(liste,linecount,
                                  ((flag&1)<<2)|(flag&(2|16|64|128))|(1<<13));
ex:;
 Sfile_destroy_argv(&filec,&filev,0);
 if(ret<=0)
   return(ret);
 return(liste->anz);
}


int Fileliste_single(liste,name,typ,groesse,flag)
struct FilelistE **liste;
char *name;
int typ;
double groesse;
int flag;
/*
 bit0= map file to root (set liste->flags:bit0 )
 bit1= set target_inp_trenner to '='  
 bit3= do not separate target from source (name is all source)
 bit4= economy class list (bit1 of Fileliste_new())

 bit8-15= see Fileliste_new
*/
{
 int ret,hflag;

 hflag= flag&0xff00;
 if(flag&16)
   hflag|= 2;
 if(Fileliste_new(liste,1,"",hflag)<=0) 
   return(0);
 if(flag&2)
   Fileliste_set_target_inp_trenner(liste,"=",0);
 ret= Fileliste_append(*liste,name,1|(flag&8));
 if(ret<=0)
   return(ret);
 (*liste)->typen[0]= typ;
 (*liste)->flags[0]|= (flag&1);
 (*liste)->groessen[0]= groesse;
 return(1);
}


int Fileliste_set_crc(liste,idx,crc,md5,flag)
struct FilelistE *liste;
int idx;
unsigned long crc;
unsigned char md5[16];
int flag;
/*
 bit0= make crc invalid
 bit1= do not complain about missing crcs vector
 bit2= permission to add a missing crc vector
 bit11= make crc valid
 bit12= set md5 rather than crc
*/
{
 static int programmer_notified= 0;
 int ret,i;

 if(liste==NULL)
   return(0);
 if(liste->crcs==NULL) {
   if((flag&4) && liste->maxanz>0) {
     liste->crcs= TSOB_FELD(struct Filelist_crc_reC *,liste->maxanz);
     if(liste->crcs==NULL)
        return(-1);
     for(i=0;i<liste->maxanz;i++)
       liste->crcs[i]= NULL;
   } else if(programmer_notified<10 && !(flag&2)) {
     fprintf(stderr,"??? programmer notification : missing crcs vector\n");
     programmer_notified++;
     return(0);
   } else
     return(0);
 }
 if(liste->anz<=idx)
   return(0);
 if(flag&1)
   liste->flags[idx]&= ~8;
 else {
   if(liste->crcs[idx]==NULL) {
     ret= Filelist_crc_rec_new(&liste->crcs[idx],!!(flag&(1<<12)));
     if(ret<=0)
       return(ret);
   }
   if(flag&(1<<11)) {
     if(flag&(1<<12)) {
       memcpy(liste->crcs[idx]->md5,md5,16);
       liste->flags[idx]|= 16;
     } else {
       liste->crcs[idx]->crc= crc;
     }
     liste->flags[idx]|= 8;
   }
 }
 return(1);
}


int Fileliste_set_inode(liste,idx,inode,flag)
struct FilelistE *liste;
int idx;
ino_t inode;
int flag;
{
 if(liste->inodes==NULL)
   return(0);
 if(idx<0||idx>liste->anz)
   return(0);
 liste->inodes[idx]= inode;
 liste->flags[idx]|= 64;
 return(1);
}


int Fileliste_set_xprop(liste,idx,mod,own,grp,date,flag)
struct FilelistE *liste;
int idx;
int mod;
char *own;
char *grp;
time_t date;
int flag;
/*
*/
{
 int ret;

 if(liste==NULL)
   return(0);
 if(liste->xprops==NULL)
   return(0);
 if(liste->anz<=idx)
   return(0);
 if(liste->xprops[idx]==NULL) {
   ret= Filelist_xprop_rec_new(&liste->xprops[idx],0);
   if(ret<=0)
     return(ret);
 }
 liste->xprops[idx]->mod= mod;
 if(Sregex_string(&(liste->xprops[idx]->own),own,0)<=0)
   return(-1);
 if(Sregex_string(&(liste->xprops[idx]->grp),grp,0)<=0)
   return(-1);
 liste->xprops[idx]->date= date;
 return(1);
}


double Fileliste_size(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i;
 double groesse;

 groesse= 0.0;
 for(i=0;i<liste->anz;i++)
   if(liste->typen[i]==1||liste->typen[i]==2)
     if(liste->groessen[i]>0)
       groesse+= liste->groessen[i];
 return(groesse);
}


double Fileliste_dir_size(liste,idx,linecount,flag)
struct FilelistE *liste;
int idx;
int *linecount;
int flag;
/*
 bit0= determine size even if directory is excluded (else 0 byte )
 bit2= attach subdir 
 bit6= print pacifier with linecount
 bit7= recursion (usually via Fileliste_dir-Fileliste_read_properties_idx)
*/
{
 struct Lies_popeN *gelesen,gelesen_data;
 struct FilelistE *sub;
 double kbytes,groesse= -1.0,new_mem= 0.0,compose_size_rec;
 char line[4096],txs_buf[4096];
 int ret,gleich=0,echt_drueber=0,hflag,original_sublevel_rec;

 gelesen= &gelesen_data;

 if(idx>liste->anz) return(-1.0);

 if(!(flag&128)) {
   Cd_backup_planer_compose_size= 0;
   Cdbp_compose_sublevel_count= 0;
 }
 original_sublevel_rec= Cdbp_compose_sublevel_count;

 /* attempt to reuse already recorded sizes */
 if(liste->groessen[idx]>0)
   return(liste->groessen[idx]);
 if(liste->subdirs!=NULL)
   if(liste->subdirs[idx]!=NULL){
     groesse= Fileliste_size(liste->subdirs[idx],0);
     /* >>> what about the size of the directory itself ? */
     return(groesse);
   }

 if(Fileliste_search_file(liste->sperrliste,liste->adressen[idx],32)>=0)
   gleich= 1;
 if(Fileliste_search_file(liste->sperrliste,liste->adressen[idx],1|4|32)>=0)
   echt_drueber= 1;
  
 if( gleich  && !(flag&1) ){ /* is the directory itself excluded ? */
   return(-1.0);
 }else if(echt_drueber){ /* is a sub directory excluded ? */
   /* new file list from directory , economize on exotic vectors */;
   if(Cdbp_compose_keep_sublevels>Cdbp_compose_sublevel_count+1
      && Cdbp_compose_sublevel_count>=0)
     hflag= 256|512|2048|(1<<13);                       /* will need subdirs */
   else
     hflag= 2;                                       /* will need no subdirs */
   if(Fileliste_new(&sub,0,liste->main_dir,hflag)<=0) 
     return(-1.0);
   Fileliste_set_sperrliste(sub,liste->sperrliste,0);

   if(flag&64) { 
     Fileliste__compose_pacifier(*linecount,-1,Cd_backup_planer_compose_size,
                                 liste->adressen[idx],0);
     (*linecount)++;
   }

   /* obtain name list only */
   if(Fileliste_dir(sub,liste->adressen[idx],linecount,4|8)<=0) 
     goto sub_ex;

   if(Cdbp_compose_sublevel_count>=0) {
     if(Cdbp_compose_keep_sublevels>Cdbp_compose_sublevel_count
        && liste->subdirs!=NULL) {
       if(sub->anz>Cdbp_compose_sublevel_maxsize) {
         Cdbp_compose_sublevel_count= -2; /* no deeper */
       } else {
         new_mem+= Fileliste_estimate_mem_usage(sub,0);
         if(Cdbp_compose_sublevel_maxmem<=0 ||
            Cdbp_subs_mem+new_mem<Cdbp_compose_sublevel_maxmem)
           Fileliste_set_subdir(liste,idx,sub,0);
       }
     }
     Cdbp_compose_sublevel_count++;
   }
   compose_size_rec= Cd_backup_planer_compose_size;
   if(Fileliste_read_properties(sub,linecount,(flag&64)|128)<=0) 
     goto sub_ex;
   groesse= Fileliste_size(sub,0);

   if(flag&64) { 
     Cd_backup_planer_compose_size= compose_size_rec+groesse;
     Fileliste__compose_pacifier(*linecount,-1,Cd_backup_planer_compose_size,
                                 liste->adressen[idx],0);
   }

sub_ex:;
   /* this deletes sub only if it was not attached as a sub list */
   Fileliste_destroy(&sub,0);
   Cdbp_compose_sublevel_count= original_sublevel_rec;
   return(groesse);
 }else{
   if(liste->dir_size_internal) {
     ret= Fileliste_determine_treesize(liste,idx,0,&groesse,linecount,flag&64);
     if(ret<=0)
       groesse= -1.0;
   } else {
     /* determine size by  dir_size_cmd  */
     Text_shellsafe(liste->adressen[idx],txs_buf,0);
     sprintf(line,"%s %s",liste->dir_size_cmd,txs_buf);
     if(Lies_popen_init(line,gelesen,0)<=0) return(-1.0);
     if(gelesen->anz>0) {
       if(sscanf(gelesen->lines[0],"%lf",&kbytes)==1)
         groesse= kbytes*1024.0;
     }
     Lies_popen_cleanup(gelesen,0);
   }
 }
 return(groesse);
}


int Fileliste_read_properties_idx(liste,i,linecount,flag)
struct FilelistE *liste;
int i;
int *linecount;
int flag;
/*
 bit0= do not complain about missing file
 bit1= no determination of size necessary
 bit2= determine size of excluded directories
 bit3= determine size even if the list item already has a size recorded
 bit4= set groessen[i] of unknown filetypes to -1
 >>> bit5= determine xprop (mod,own,grp,date)
 bit6= print pacifier with linecount
 bit7= recursion (for Fileliste_dir_size)
 bit8= accept block devices (and determine size by reading them)
 >>> bit11= determine checksum
 >>> bit12= checksum MD5 rather than CRC32
 bit13= make list items invalid if they produce errors
 
*/
{
 double groesse;
 char target[2*4096];
 int adrlen,fixed_size;
 struct Lies_popeN *gelesen,gelesen_data;
 struct Cd_backup_planer_struct_staT stbuf;
 static int verbous= 0;

 gelesen= &gelesen_data;
 fixed_size= ((liste->flags[i]&32) && liste->groessen[i]>0.0);
 liste->typen[i]= 0;
 if((flag&8)){
   Fileliste_set_subdir(liste,i,NULL,0);
   if(!(liste->flags[i]&32))
     liste->groessen[i]= -1.0;
 }
 if(Lstat(liste->adressen[i],&stbuf,!(flag&1))==-1) {
   if(flag&(1<<13))
     Fileliste_defunct_item(liste,i,0);
   return(0);
 }
 if(S_ISREG(stbuf.st_mode)) {
   /* Data file */
   liste->typen[i]= 1;

   /* >>> if the file is not readable : complain and block out */

   if(!fixed_size) {
     liste->groessen[i]= stbuf.st_size;
     if(liste->part_recs!=NULL)
       if(liste->part_recs[i]!=NULL)
         Filelist_part_rec_get_byte_count(liste->part_recs[i],
                                          &(liste->groessen[i]),0);
   }
 } else if (S_ISDIR(stbuf.st_mode)) {
   /* Directory */
   if(flag&2){
     liste->typen[i]= 2;
   }else{

     /* >>> if the directory is not readable : complain and block out */

     if(!fixed_size) {
       groesse= Fileliste_dir_size(liste,i,linecount,
                                   (flag&(64|128))|4|((flag>>2)&1));
       if(groesse>=0.0) {
         liste->groessen[i]= groesse;
         liste->typen[i]= 2;
       }
     }
   }
 } else if (S_ISLNK(stbuf.st_mode)) {
   /* Softlink */
   liste->typen[i]= 3;
   fixed_size = (liste->flags[i]&32); /* link will eventually get redirected 
                                         do not set size before */
   if(!fixed_size)
     liste->groessen[i]= stbuf.st_size;
 } else if (S_ISBLK(stbuf.st_mode) ){ /* <<< && (liste->flags[i]&32)) */
   if(!(flag&2))
     Fileliste_type_and_size(liste,i,&(liste->groessen[i]),0,0);
   liste->typen[i]= 6;
 } else {  /* unclassified */
   liste->typen[i]= 127;
   if(!fixed_size) {
     if(flag&16) 
       liste->groessen[i]= -1;
     else
       liste->groessen[i]= stbuf.st_size;
   }
 }

 /* eventually add name length and node_overhead */
 if(liste->groessen[i]>=0 && liste->typen[i]!=2 && !(liste->flags[i]&32)) {
   if(Fileliste_node_addrl) {
     if(Fileliste_target(liste,i,target,2)<=0)
       adrlen= strlen(liste->adressen[i]);
     else
       adrlen= strlen(target);
   } else
     adrlen= 0;
   liste->groessen[i]+= adrlen+Fileliste_node_overhead;
 }

 if(verbous)
   fprintf(stderr,"%d %.lf %s\n",(int) liste->typen[i],liste->groessen[i],
                                 liste->adressen[i]);

 return(1);
}


int Fileliste_read_properties(liste,linecount,flag)
struct FilelistE *liste;
int *linecount;
int flag;
/*
 flag is forwarded to Fileliste_read_properties_idx
 bit6= print pacifier with linecount
*/
{
 int i,ret,ret_one= 0,ret_zero= 0;
 double start_time,current_time;

 /* determine file types and sizes */
 start_time= Sfile_microtime(0);
 for(i=0;i<liste->anz;i++) {
   if(!(liste->flags[i]&32))
     liste->groessen[i]= -1;
   ret= Fileliste_read_properties_idx(liste,i,linecount,flag);
   if(ret<0) 
     return(ret);
   if(ret==0)
     ret_zero++;
   else
     ret_one++;
   if((flag&64) && (i%100)==0 && i>0) {
     current_time= Sfile_microtime(0);
     if(current_time-start_time>=1.0)
       Fileliste__compose_pacifier(*linecount,i,Cd_backup_planer_compose_size,
                                   liste->adressen[i],0);
   }
 }
 if(flag&32)
   if(ret_one)
     ret_zero=0;
 return(!ret_zero);
}


int Fileliste_filetyp(adr,flag)
char *adr;
int flag;
{
 int ret;
 double size;

 ret= Type_and_size(adr,&size,1);
 if(ret<0)
   ret= 0;
 else if(ret==0)
   ret= 127;
 else if(ret==4)
   ret= 127;
 return(ret);
}


int Fileliste_defunct_item(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
/*
 bit0= try to save as much memory as possible without producing an ill item
 bit1= if not bit0 : delete at least sublists[idx]
*/
{
 if(idx>=liste->anz)
   return(0);
 if(liste->typen[idx]>0)
   liste->typen[idx]= -liste->typen[idx];
 else if(liste->typen[idx]==0)
   liste->typen[idx]= -128;
 if(flag&3) {
   if(liste->subdirs!=NULL) 
     Fileliste_destroy(&(liste->subdirs[idx]),2|4);
   if((flag&1) && liste->adressen[idx]!=NULL)
     Sregex_string(&(liste->adressen[idx]),"",0);
   if((flag&1) && liste->targets[idx]!=NULL)
     Sregex_string(&(liste->targets[idx]),"",0);
 }
 return(1);
}


int Fileliste_copy_item(von,von_idx,nach,nach_idx,flag)
struct FilelistE *von;
int von_idx;
struct FilelistE *nach;
int nach_idx;
int flag;
/*
 flag: for Fileliste_copy_item_hd
  blocked is :  bit0 
*/
{
 return(Fileliste_copy_item_hd(von,von_idx,nach,nach_idx,flag&~1));
}


/* was once more different from Fileliste_copy_item(). Now : only bit1 */
int Fileliste_copy_item_hd(von,von_idx,nach,nach_idx,flag)
struct FilelistE *von;
int von_idx;
struct FilelistE *nach;
int nach_idx;
int flag;
/*
 bit0= determine von.target write it explicitely into nach
       ("von" = from , "nach" = to )
 bit1= move allocated data rather than copying them
*/
{
 int ret;
 char target[2*4096],*targetpt;
 struct Filelist_part_reC *vpr;
 targetpt= target;

 if(von_idx>=von->anz) return(0);
 if(nach_idx>=nach->maxanz)
   if(Fileliste_enlarge(nach,nach_idx+1,0)<=0) return(-1);

 /* if nach already hashlinked then unlink */
 if(nach_idx<nach->anz)
   Fileliste_hashlink(nach,nach_idx,1);

 nach->typen[nach_idx]= von->typen[von_idx];
 nach->flags[nach_idx]= von->flags[von_idx];
 nach->groessen[nach_idx]= von->groessen[von_idx];
 if(nach->inodes!=NULL) {
   if(von->inodes!=NULL)
     nach->inodes[nach_idx]= von->inodes[von_idx];
   else {
     nach->inodes[nach_idx]= 0;
     nach->flags[nach_idx]&= ~64;
   }
 } else
   nach->flags[nach_idx]&= ~64;
 if(nach->adressen[nach_idx]!=NULL) 
   free(nach->adressen[nach_idx]);
 if(flag&2) {
   nach->adressen[nach_idx]= von->adressen[von_idx];
   von->adressen[von_idx]= NULL;
   if(nach->targets[nach_idx]!=NULL) 
     free(nach->targets[nach_idx]);
   nach->targets[nach_idx]= von->targets[von_idx];
   von->targets[von_idx]= NULL;

   if(nach->crcs!=NULL) {
     if(nach->crcs[nach_idx]!=NULL) 
       free((char *) nach->crcs[nach_idx]);
     if(von->crcs!=NULL) {
       nach->crcs[nach_idx]= von->crcs[von_idx];
       von->crcs[von_idx]= NULL;
     } else
       nach->crcs[nach_idx]= NULL;
   } else if(von->crcs!=NULL) {
     if(von->crcs[von_idx]!=NULL)
       free((char *) von->crcs[von_idx]);
     von->crcs[von_idx]= NULL;
   }

   if(nach->xprops!=NULL) {
     if(nach->xprops[nach_idx]!=NULL) 
       free((char *) nach->xprops[nach_idx]);
     if(von->xprops!=NULL) {
       nach->xprops[nach_idx]= von->xprops[von_idx];
       von->xprops[von_idx]= NULL;
     } else
       nach->xprops[nach_idx]= NULL;
   } else if(von->xprops!=NULL) {
     if(von->xprops[von_idx]!=NULL)
       free((char *) von->xprops[von_idx]);
     von->xprops[von_idx]= NULL;
   }

   vpr= NULL;
   if(von->part_recs!=NULL)
     vpr= von->part_recs[von_idx];
   ret= Fileliste_set_part_rec(nach,nach_idx,vpr,0);
   if(ret<=0)
     goto ex;
   Fileliste_set_part_rec(von,von_idx,NULL,2);
   von->flags[von_idx]&= ~16;

   if(nach->subdirs!=NULL) {
     Fileliste_destroy(&(nach->subdirs[nach_idx]),2);
     if(von->subdirs!=NULL) {
       nach->subdirs[nach_idx]= von->subdirs[von_idx];
       von->subdirs[von_idx]= NULL;
     } else
       nach->subdirs[nach_idx]= NULL;
   } else if(von->subdirs!=NULL) {
     Fileliste_destroy(&(von->subdirs[von_idx]),2);
   }
   if(nach->subdirs!=NULL) {
     if(nach->subdirs[nach_idx]!=NULL) {
       nach->subdirs[nach_idx]->boss= nach;
       nach->subdirs[nach_idx]->boss_idx= nach_idx;
     }
   }

 } else {
   if(von->adressen[von_idx]==NULL){
     nach->adressen[nach_idx]= NULL;
   }else{
     nach->adressen[nach_idx]= TSOB_FELD(char,strlen(von->adressen[von_idx])+1);
     if(nach->adressen[nach_idx]==NULL) return(0);
     strcpy(nach->adressen[nach_idx],von->adressen[von_idx]);
   }
   if(flag&1){
     Fileliste_target(von,von_idx,target,2);
   }else{
     targetpt= von->targets[von_idx];
   }
   Fileliste_set_target(nach,nach_idx,targetpt,1);

   if(von->crcs!=NULL) {
     if(von->crcs[von_idx]!=NULL) {
       ret= Fileliste_set_crc(nach,nach_idx,
                              von->crcs[von_idx]->crc,von->crcs[von_idx]->md5,
                              (int) ( ((!!(von->flags[von_idx]& 8))<<11) |
                                      ((!!(von->flags[von_idx]&16))<<12) | 
                                      2
                                    )
                             );
       if(ret<0)        /* ret==0 indicates invalid target and is tolerable */
         return(0);
     }
   }


   /* >>> ??? what about xprops here ? Did i leave that out ??? */

   vpr= NULL;
   if(von->part_recs!=NULL)
     vpr= von->part_recs[von_idx];
   ret= Fileliste_set_part_rec(nach,nach_idx,vpr,1);
   if(ret<=0)
     goto ex;
   if(von->subdirs!=NULL) {
     Fileliste_set_subdir(nach,nach_idx,von->subdirs[von_idx],2);
   } else if(nach->subdirs!=NULL) {
     Fileliste_destroy(&(nach->subdirs[nach_idx]),2);
   }
 }
 if(nach->anz<=nach_idx) nach->anz= nach_idx+1;
 Fileliste_hashlink(nach,nach_idx,0);
 ret= 1;
ex:;
 return(ret);
}


int Fileliste_truncate(liste,anz,flag)
struct FilelistE *liste;
int anz;
int flag;
/*
 bit10= do not destroy eventual subdirs[] but only detach
*/
{
 int i;

 for(i=anz;i<liste->anz;i++)
   Fileliste_destroy_item(liste,i,flag&(1<<10));
 liste->anz= anz;
 return(1);
} 


int Fileliste_garbage(liste,flag)
struct FilelistE *liste;
int flag;
{
 int i,w;
 w= 0;
 for(i=0;i<liste->anz;i++){
   if(liste->typen[i]<0 || liste->adressen[i]==NULL)
 continue;
   if(w!=i)
     Fileliste_copy_item(liste,i,liste,w,0);
   w++;
 }
 Fileliste_truncate(liste,w,0);
 return(1);
}


int Fileliste_source_target(liste,filename,source,target,flag)
struct FilelistE *liste;
char *filename;
char source[4096];
char target[2*4096];
int flag;
/*
 bit0= fill target in any case, eventually use copy of source
 bit3= do not separate source from target but copy filename into both
*/
{
 char *cpt;

 if(liste->target_inp_trenner && !(flag&8)) {
   Fileliste__target_source_limit(filename,liste->target_inp_trenner,&cpt,0);
/*
   cpt= strchr(filename,liste->target_inp_trenner);
*/
 } else
   cpt= NULL;
 if(cpt==NULL){
   strcpy(source,filename);
   if(flag&1)
     strcpy(target,source);
   else
     target[0]= 0;
 } else {
   strcpy(source,cpt+1);
   strncpy(target,filename,cpt-filename);
   target[cpt-filename]= 0;
   Fileliste__unescape_target(target,NULL,liste->target_inp_trenner,0);
 }
 return(1);
}


int Fileliste_append(liste,filename,flag)
struct FilelistE *liste;
char *filename;
int flag;
/*
 bit0= no Fileliste_read_properties_idx()
 bit1= later map to root
 bit2= record only if the file exists
 bit3= do not separate target
 bit5= set bit5 in flags
 bit6= do not remove trailing slashes
 bit7= record only if not already recorded
 bit8= set bit1 in flags
 bit18= this is level 0
*/
{
 int i,ret,l,zero= 0;
 char quelle[4096],ziel[2*4096];

 Fileliste_source_target(liste,filename,quelle,ziel,flag&8);

 if(flag&4){
   struct stat stbuf;
   if(Lstat(quelle,&stbuf,0)==-1) 
     return(2);
 }
 if(flag&128) {
   for(i=0;i<liste->anz;i++) {
     if(liste->typen[i]<0)
   continue;
     if(liste->adressen[i]==NULL)
   continue;
     if(strcmp(liste->adressen[i],quelle)!=0)
   continue;
     if(liste->targets[i]==NULL) {
       if(ziel[0]==0)
   break;
       else
   continue;
     }
     if(strcmp(liste->targets[i],ziel)!=0)
   continue;
   break;
   }
   if(i<liste->anz)
     return(3);
 }
 i= liste->anz;
 if(i>=liste->maxanz)
   if(Fileliste_enlarge(liste,i+1,0)<=0) return(-1);
 if(liste->adressen[i]!=NULL)
   free(liste->adressen[i]);
 liste->adressen[i]= TSOB_FELD(char,strlen(quelle)+1);
 if(liste->adressen[i]==NULL)
   return(-1);
 strcpy(liste->adressen[i],quelle);
 if(!(flag&64)) {
   l= strlen(liste->adressen[i]);
   while(l>1) {
     if(liste->adressen[i][l-1]=='/') {
       liste->adressen[i][l-1]= 0;
       l--;
     } else
   break;
   }
 }
 Fileliste_hashlink(liste,i,0);
 if(liste->targets[i]!=NULL)
   free(liste->targets[i]);
 if(ziel[0]){
   liste->targets[i]= TSOB_FELD(char,strlen(ziel)+1);
   if(liste->targets[i]==NULL) return(-1);
   strcpy(liste->targets[i],ziel);
 }else{
   liste->targets[i]= NULL;
 }
 liste->flags[i]= 0;
 if(flag&2)
   liste->flags[i]|= 1;
 liste->flags[i]|= (flag&32);
 if(flag&256)
   liste->flags[i]|= 2;
 liste->anz++;
 if(flag&1){
   liste->groessen[i]= -1;
   liste->typen[i]= 0;
 }else{
   ret= Fileliste_read_properties_idx(liste,i,&zero,0);
   if(ret<0)
     return(ret);
 }
 return(1);
}


int Fileliste_append_prop(liste,adr,target,typ,size,flag)
struct FilelistE *liste;
char *adr,*target;
int typ;
double size;
int flag;
/*
 see Fileliste_append, bit3 is always on
*/
{
 int ret;

 ret= Fileliste_append(liste,adr,flag|8);
 if(ret<=0)
   return(ret);
 liste->typen[liste->anz-1]= typ;
 liste->groessen[liste->anz-1]= size;
 ret= Fileliste_set_target(liste,0,target,2);
 return(ret);
}


int Fileliste_relativ_adr(oberdir,adr,rel_adr,flag)
char *oberdir;
char *adr;
char *rel_adr;
int flag;
{
 int l,slash_ende=0;

 rel_adr[0]= 0;
 l= strlen(oberdir);
 if(l>0) slash_ende= (oberdir[l-1]=='/');
 if(slash_ende) l--;
 
 if(strncmp(oberdir,adr,l)!=0) return(0);
 if(strlen(adr)==l) 
   return(2);
 if(adr[l]=='/'){
   strcpy(rel_adr,adr+l+1);
   return(1);
 }
 return(0);
}


int Fileliste_search_inode(liste,filename,flag)
struct FilelistE *liste;
char *filename;
int flag;
/*
 bit0= test wether filename is a parent directory of a listed file 
 bit1= test wether filename is a sub file of a listed directory
 bit2= do not accept exact equality
 bit3= only accept items with  liste->typen[i]>=0
 bit4= debug: this is a payload search (not a tree creation search)
 bit5= exclusion list mode. Only process items marked by (liste->flags[])&2
*/
{
 int i, ret, original_adr;
 ino_t inode,idx_inode;
 dev_t device,idx_device;
 int ino_vec[CdbplivS],dev_vec[CdbplivS];
 char numtext[81], idx_adr[4096], adr[4096], *spt;

 if(liste==NULL) return(-1);
 if(File_inode(filename,&device,&inode,0)<=0)
   return(-1);
 if(Cdbackup_planer_listsearch_debuG>=2) if(flag&16) {
   Intvec_from_dev_ino(dev_vec,CdbplivS,ino_vec,device,inode,0);
   Intvec_to_dec(dev_vec,CdbplivS,numtext,sizeof(numtext),0);
   fprintf(stderr,"listsearch_debug: dev= %s , ",numtext);
   Intvec_to_dec(ino_vec,CdbplivS,numtext,sizeof(numtext),0);
   fprintf(stderr,"inode= %s\n",numtext);
 }

 for(i=0;i<liste->anz;i++){
   if(liste->adressen[i]==NULL)
 continue;
   if((flag&32) && !(liste->flags[i]&2))
 continue;
   if(flag&8)
     if(liste->typen[i]<0)
 continue;
   strcpy(adr, filename);
   if((flag&2) && i!=0)
     if(File_inode(adr,&device,&inode,0)<=0)
 continue;
   strcpy(idx_adr, liste->adressen[i]);
   original_adr= 1;

   while(idx_adr[0] && adr[0]) {
     if((flag&2) && !original_adr) {
       if(File_inode(adr,&device,&inode,0)<=0)
         goto next_idx;
       Intvec_from_dev_ino(dev_vec,CdbplivS,ino_vec,device,inode,0);
     } else {
       if(File_inode(idx_adr,&idx_device,&idx_inode,0)<=0)
         goto next_idx;
       Intvec_from_dev_ino(dev_vec,CdbplivS,ino_vec,idx_device,idx_inode,0);
     }

     if(Cdbackup_planer_listsearch_debuG>=2) if(flag&16) {
       Intvec_to_dec(dev_vec,CdbplivS,numtext,sizeof(numtext),0);
       fprintf(stderr,"listsearch_debug: idx= %d : dev= %s , ",i,numtext);
       Intvec_to_dec(ino_vec,CdbplivS,numtext,sizeof(numtext),0);
       fprintf(stderr,"inode= %s\n",numtext);
       if((flag&2) && !original_adr)
         fprintf(stderr," : %s",adr);
       else
         fprintf(stderr," : %s",idx_adr);
     }

     if((device==idx_device && inode==idx_inode)&& !((flag&4) && original_adr))
       {ret= i; goto ex;}

     if(flag&2)                      /* cut off another basename of filename */
       spt= strrchr(adr,'/');
     else if(flag&1)                /* cut off another basename of list file */
       spt= strrchr(idx_adr,'/');
     else                              /* no cutting : we only want equality */
       goto next_idx;
     if(spt!=NULL)
       *spt= 0;
     original_adr= 0;
   }
next_idx:;
 }
 ret= -1;
ex:
 return(ret);
}


int Fileliste_search_file(liste,filename,flag)
struct FilelistE *liste;
char *filename;
int flag;
/*
 bit0= test wether filename is a parent directory of a listed file 
 bit1= test wether filename is a sub file of a listed directory
 bit2= do not accept exact equality
 bit3= only accept items with  liste->typen[i]>=0
 bit4= debug: this is a payload search (not a tree creation search)
 bit5= exclusion list mode. If feature enabled: test inode and device number
*/
{
 static int programmer_notified= 0;
 int i, ret;
 char rel_adr[4096];

 if(liste==NULL) return(-1);
 if(flag&2)
   goto unter_suche;
 if(liste->use_hashing && liste->hashstart!=NULL && (flag&7)==0) {
   if(liste->hashlinks==NULL && programmer_notified<10) {
     fprintf(stderr,
         "??? programmer notification : missing hashlinks vector on search\n");
     programmer_notified++;
   } else
     goto hashed_search;
 }

 for(i=0;i<liste->anz;i++){
   if(liste->adressen[i]==NULL)
 continue;
   if(flag&8)
     if(liste->typen[i]<0)
 continue;

   if(Cdbackup_planer_listsearch_debuG>=2) if(flag&16)
     fprintf(stderr,"listsearch_debug: idx= %d : %s\n",i,liste->adressen[i]);

   if(Fileliste_relativ_adr(filename,liste->adressen[i],rel_adr,0)>0) {
     if(rel_adr[0]==0 && !(flag&4))
 break;
     if(rel_adr[0]!=0 && (flag&1))
 break;
   }
 }
 if(i<liste->anz)
   ret= i;
 else
   ret= -1;
 goto ex;

unter_suche:;
 for(i=0;i<liste->anz;i++){
     if(liste->adressen[i]==NULL)
 continue;

   if(Cdbackup_planer_listsearch_debuG>=2) if(flag&16)
     fprintf(stderr,"listsearch_debug: idx= %d : %d  %s\n",
                     i,liste->typen[i],liste->adressen[i]);

   if(flag&8)
     if(liste->typen[i]<0)
 continue;
   if(Fileliste_relativ_adr(liste->adressen[i],filename,rel_adr,0)>0) {
     if(rel_adr[0] || !(flag&4))
 break;
   }
 }
 if(i<liste->anz)
   ret= i;
 else
   ret= -1;
 goto ex;

hashed_search:;
 i= Fileliste__compute_hashvalue(filename,0);

 if(Cdbackup_planer_listsearch_debuG>=2) if(flag&16)
   fprintf(stderr,"listsearch_debug: hash= %d , hashstart= %d\n",
                  i,liste->hashstart[i]);

 for(i= liste->hashstart[i]; i>=0; i= liste->hashlinks[i]) {
   if(flag&8)
     if(liste->typen[i]<0)
 continue;
   if(liste->adressen[i]==NULL)
 continue;

   if(Cdbackup_planer_listsearch_debuG>=2) if(flag&16)
     fprintf(stderr,"listsearch_debug: idx= %d : %s\n",i,liste->adressen[i]);

   if(strcmp(filename,liste->adressen[i])==0)
     {ret= i; goto ex;}
 }
 ret= -1;
ex:
 if(ret<0 && (flag&32) && Cdbackup_planer_sperrliste_inodE)
   ret= Fileliste_search_inode(liste,filename,flag&63);
 return(ret);
}


int Fileliste_set_dir_size_cmd(liste,cmd,flag)
struct FilelistE *liste;
char *cmd;
int flag;
{
 int ret;

 if(liste==NULL)
   return(0);
 ret= Sregex_string(&(liste->dir_size_cmd),cmd,0);
 if(strncmp(cmd,"-internal",9)==0)
   liste->dir_size_internal= 1;
 else
   liste->dir_size_internal= 0;
 return(ret);
}


int Fileliste_default_dir_size_cmd(cmd,flag)
char *cmd;
int flag;
{
 char *cpt;
 double wert= 0.0;
 int separator_minus= 0;

 if(strlen(cmd)>sizeof(Fileliste_dir_size_cmd)-1) return(0);
 strcpy(Fileliste_dir_size_cmd,cmd);
 if(strncmp(cmd,"-internal",9)==0)
   Fileliste_dir_size_internal= 1;
 else
   Fileliste_dir_size_internal= 0;
 if(strncmp(cmd,"-internal",9)==0 && strlen(cmd)>9) {
   cpt= strchr(cmd,'+');
   if(cpt==NULL) {
     cpt= strchr(cmd+1,'-'); /* +1 = skip first '-' */
     separator_minus= 1;
   }
   if(cpt!=NULL) {
     if(*(cpt+1)!=0) {
       sscanf(cpt+1,"%lf",&wert);
       if(wert>=0 && wert<=1000000000)
         Fileliste_node_overhead= wert;
     }
     if(separator_minus)
       Fileliste_node_addrl= 0;
     else
       Fileliste_node_addrl= 1;
   }
 }
 return(1);
}


int Fileliste_set_impl_igno(liste,text,flag)
struct FilelistE *liste;
char *text;
int flag;
{
 if(liste==NULL) return(0);
 if(liste->impl_igno!=NULL) free((char *) liste->impl_igno);
 liste->impl_igno= TSOB_FELD(char,strlen(text)+1);
 strcpy(liste->impl_igno,text);
 return(1);
}


int Fileliste_default_impl_igno(text,flag)
char *text;
int flag;
{
 if(strlen(text)>sizeof(Fileliste_impl_igno)-1) return(0);
 strcpy(Fileliste_impl_igno,text);
 return(1);
}


int Fileliste_set_target_inp_trenner(liste,trenner,flag)
struct FilelistE *liste;
char *trenner;
int flag;
{
 if(liste==NULL) return(0);
 liste->target_inp_trenner= trenner[0];
 return(1);
}


int Fileliste_default_target_inp_trenner(trenner,flag)
char *trenner;
int flag;
{
 Fileliste_target_inp_trenner= trenner[0];
 return(1);
}


int Fileliste_set_sperrliste(liste,sperrliste,flag)
struct FilelistE *liste,*sperrliste;
int flag;
{
 if(liste==NULL) return(0);
 liste->sperrliste= sperrliste;
 return(1);
}


int Fileliste_set_subdir(liste,idx,sub,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *sub;
int flag;
/*
 bit0= do not set boss pointer of sub
 bit1= do not complain about missing subdirs vector
*/
{
 static int programmer_notified= 0;

 if(liste->subdirs==NULL) {
   if(programmer_notified<10 && !((flag&2)||sub==NULL)) {
     fprintf(stderr,
       "??? programmer notification : missing subdirs vector on set_subdir\n");
     if(liste->adressen[idx]!=NULL)
       fprintf(stderr,"???   %s\n",liste->adressen[idx]);
     programmer_notified++; 
   }
   return(0);
 }
 if(liste->subdirs[idx]!=NULL)
   Fileliste_destroy(&(liste->subdirs[idx]),2|4);
 if(sub!=NULL) {
   sub->ref_count++;
   if(!(flag&1)) {
     sub->boss= liste;
     sub->boss_idx= idx;
   }
 }
 liste->subdirs[idx]= sub;
 Fileliste_count_sub(liste,idx,&Cdbp_subs,&Cdbp_subs_lines,&Cdbp_subs_mem,0);
 return(1);
}


int Fileliste_count_sub(liste,idx,sub_count,line_count,mem_count,flag)
struct FilelistE *liste;
int idx;
double *sub_count,*line_count,*mem_count;
int flag;
/*
 bit0= subtract rather than add
 bit1= count liste itself rather than liste->subs[idx]
*/
{
 struct FilelistE *tocount;
 double used_mem= 0.0;
 
 if(Cdbp_count_subs==0)
   return(2);
 if(flag&2) {
   tocount= liste;
 } else {
   if(liste->subdirs==NULL)
     return(2);
   if(liste->subdirs[idx]==NULL)
     return(2);
   tocount= liste->subdirs[idx];
 }
 used_mem= Fileliste_estimate_mem_usage(tocount,0);
 if(flag&1) {
   (*sub_count)--;
   (*line_count)-= tocount->maxanz;
   (*mem_count)-= used_mem;
 } else {
   (*sub_count)++;
   (*line_count)+= tocount->maxanz;
   (*mem_count)+= used_mem;
 }
 return(1);
}

int Fileliste_hauptdir(hauptdir,liste,flag)
char *hauptdir;
struct FilelistE *liste;
int flag;
{
 int i,j,slmerk;

 if(liste==NULL) return(0);
 hauptdir[0]= 0;
 if(liste->anz<=0) return(1);
 strcpy(hauptdir,liste->adressen[0]);
 for(i=1;i<liste->anz;i++){
   slmerk= 0;
   for(j=0;hauptdir[j]!=0;j++){
     if(hauptdir[j]!=liste->adressen[i][j]) {
       if(hauptdir[j]=='/' && liste->adressen[i][j]==0) slmerk= j;
       hauptdir[slmerk]= 0;
   break;
     }
     if(hauptdir[j]=='/') slmerk= j;
   }
 if(hauptdir[0]==0) break;
 }
 return(1);
}


int Fileliste_mapped_to_root(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
{
 if(liste==NULL) return(0);
 if(idx>=liste->anz) return(0);
 return((int) !!(liste->flags[idx]&1));
}


int Fileliste_is_impl_igno(liste,adr,flag)
struct FilelistE *liste;
char *adr;
int flag;
{
 int k;

 for(k=0;k<strlen(liste->impl_igno);k++)
   if(strchr(adr,liste->impl_igno[k])!=NULL) break;
 if(k<strlen(liste->impl_igno)) 
   return(1);
 return(0);
}


int Fileliste_set_target_dir(liste,dir,flag)
struct FilelistE *liste;
char *dir;
int flag;
{
 if(Sregex_string(&(liste->target_dir),dir,0)<=0)
   return(-1);
 return(1);
}


int Fileliste_set_main_dir(liste,dir,flag)
struct FilelistE *liste;
char *dir;
int flag;
{
 if(Sregex_string(&(liste->main_dir),dir,0)<=0)
   return(-1);
 return(1);
}


int Fileliste_target(liste,idx,target,flag)
struct FilelistE *liste;
int idx;
char target[2*4096];
int flag;
/*
 bit0= do not mark directories by a trailing / (eventually remove it)
 bit1= do not escape or replace "="
*/
{
 int ret,l;
 char *cpt,*tpt,*hauptdir;

 hauptdir= liste->main_dir;
 target[0]= 0;
 if(liste->targets[idx]!=NULL)
  if(liste->targets[idx][0]){
   target[0]= 0;
   if(liste->targets[idx][0]!='/')
     strcat(target,"/");
   strcat(target,liste->targets[idx]);
   ret= 2;
   goto fertig;
 }
 if(liste->target_dir[0]!=0) {
   target[0]= 0;
   if(liste->target_dir[0]!='/')
     strcat(target,"/");
   strcat(target,liste->target_dir);
   if(target[strlen(target)-1]!='/') 
     strcat(target,"/");
   Fileliste_leafname(liste->adressen[idx],target+strlen(target),0);
   ret= 3;
   goto fertig;
 }

 cpt= liste->adressen[idx];
 while(1){
   if(cpt[0]=='.' && cpt[1]=='/') cpt+= 2;
   else if(cpt[0]=='.' && cpt[1]=='.' && cpt[2]=='/') cpt+=3;
   else 
 break;
 }

 tpt= cpt;
 if(Fileliste_mapped_to_root(liste,idx,0)) 
   tpt+= strlen(tpt);
 if(strlen(hauptdir))
   if(strncmp(cpt,hauptdir,strlen(hauptdir))==0)
     if(tpt[strlen(hauptdir)]=='/' || tpt[strlen(hauptdir)]==0)
       tpt+= strlen(hauptdir);
 if(liste->typen[idx]==2){
   if(tpt[0]=='.'&&tpt[1]==0) tpt++;
   else if(tpt[0]!='/' && tpt[0]) strcat(target,"/");
 }else{
   if(tpt[0]!='/' && tpt[0]) strcat(target,"/");
   if(tpt[0]==0 && tpt!=cpt){ /* single file and -main_dir -auto */
     /* look for the start of the file name */
     do{
       if(*(tpt-1)=='/') 
     break;
     }while((--tpt)>cpt);
   }
 }
 strcat(target,tpt);
 if(!(flag&1) && liste->typen[idx]==2 &&
    !(strlen(target)>0 && target[strlen(target)-1]=='/'))
   strcat(target,"/");

 ret= 1;

fertig:;
 l= strlen(target);
 if(flag&1){
   if(l>0)
     if(target[l-1]=='/') 
       target[l-1]= 0;
 } else {
   if(liste->typen[idx]==2 && l>0)
     if(target[l-1]!='/')
       strcat(target,"/");
 }

 if(!(flag&2)) {
   if(Fileliste_target_modern) {
     Fileliste__escape_target(target,0);
   } else {
     /* replace any '=' in the target address */
     for(cpt=target;*cpt!=0;cpt++) 
       if(*cpt=='=') *cpt= '_';
   }
 }
 return(ret);
}


int Fileliste_leafname(adr,leafname,flag)
char *adr,*leafname;
int flag;
{
 char *cpt;
 cpt=strrchr(adr,'/');
 if(cpt==NULL) cpt= adr;
 else          cpt++;
 strcpy(leafname,cpt);
 return(1);
}


int Fileliste_set_target(liste,idx,target,flag)
struct FilelistE *liste;
int idx;
char *target;
int flag;
/*
 bit0= do not check index
 bit1= set idx to liste->anz-1
*/
{
 if(flag&2)
   idx= liste->anz-1;
 if(!(flag&1))
   if(idx<0||idx>=liste->anz)
     return(0);
 if(liste->targets[idx]!=NULL)
   free(liste->targets[idx]);
 if(target==NULL){
   liste->targets[idx]= NULL;
 }else{
   liste->targets[idx]= TSOB_FELD(char,strlen(target)+1);
   if(liste->targets[idx]==NULL) return(-1);
   strcpy(liste->targets[idx],target);
 }
 return(1);
}


int Fileliste_inherit_target(liste,idx,sub,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *sub;
int flag;
/*
 bit0= only set target at directory items
*/
{
 int ret,i;
 char target[2*4096],bosstarget[2*4096];

 if(liste->targets[idx]==NULL)
   goto eventually_release;
 if(liste->targets[idx][0]!=0) {
   Fileliste_target(liste,idx,bosstarget,2);
   for(i=0;i<sub->anz;i++){
     if((flag&1) && sub->typen[i]!=2)
   continue;
     strcpy(target,bosstarget);
     Fileliste_leafname(sub->adressen[i],target+strlen(target),0);
     if(sub->typen[i]==2)
       strcat(target,"/");
     ret= Fileliste_set_target(sub,i,target,0);
     if(ret<=0)
       return(ret);
   }
 } else {

eventually_release:;
   /* >>> eventually release existing targets of sub */;

 }
 return(1);
}


int Fileliste_is_collision_sub(liste,ober_idx,sperrliste,
                               coll_idx,coll_rel_adr,flag)
struct FilelistE *liste;
int ober_idx;
struct FilelistE *sperrliste;
int coll_idx;
char *coll_rel_adr;
int flag;
/*
 bit8= directory opened virtually, call Fileliste_is_impl_igno() in any case 
*/
{
 char target[2*4096],target_i[2*4096],line[4096];
 struct FilelistE *sub1=NULL,*sub2=NULL,*sub=NULL;
 int i,ret=0,typ,zero= 0;

 i= coll_idx;
 Fileliste_target(liste,ober_idx,target,1);
 Fileliste_target(liste,i,target_i,1);
 if(Fileliste_relativ_adr(target_i,target,coll_rel_adr,1)<=0)
   return(0);
 if(coll_rel_adr[0]==0) {
   if(flag&256)
     if(Fileliste_is_impl_igno(liste,coll_rel_adr,0)>0)
       {ret= 0; goto ex;}
   /* both target addresses are equal */;
   if(liste->typen[i]!=2 || liste->typen[ober_idx]!=2) {
     /* one of the items is no directory : collision detected */
     ret= 2;
     goto ex;
   }
   strcpy(line,liste->adressen[i]);
 } else {

   /* check wether coll_rel_adr would be rejected when found within a 
      directory */
   if(Fileliste_is_impl_igno(liste,coll_rel_adr,0)>0)
     {ret= 0; goto ex;}

   /* check wether liste->adressen[i]/coll_rel_adr  exists */;
   strcpy(line,liste->adressen[i]);
   if(line[strlen(line)-1]!='/') strcat(line,"/");
   strcat(line,coll_rel_adr);
   if(Fileliste_search_file(sperrliste,line,2)>=0)
     {ret= 0; goto ex;}
   typ= Fileliste_filetyp(line,0);
   if(typ<0)
     goto kaputt;
   if(typ<=0)
     {ret= 0; goto ex;}
   if(typ!=2 || liste->typen[ober_idx]!=2) {
     /* is not a directory : collision detected */
     ret= 3; 
     goto ex;
   }
 }
 if(Fileliste_new(&sub1,0,liste->main_dir,2)<=0) 
   goto kaputt;
 if(Fileliste_dir(sub1,line,&zero,2)<=0) 
   goto kaputt;
 if(Fileliste_new(&sub2,0,liste->main_dir,2)<=0) 
   goto kaputt;
 if(Fileliste_dir(sub2,liste->adressen[ober_idx],&zero,2)<=0) 
   goto kaputt;
 if(Fileliste_new(&sub,sub1->anz+sub2->anz+1,liste->main_dir,0)<=0)
   goto kaputt;
 for(i=0;i<sub1->anz;i++)
   Fileliste_copy_item(sub1,i,sub,i,0);
 for(i=0;i<sub2->anz;i++)
   Fileliste_copy_item(sub2,i,sub,i+sub1->anz,0);
 if(sub1!=NULL) Fileliste_destroy(&sub1,0); sub1= NULL;
 if(sub2!=NULL) Fileliste_destroy(&sub2,0); sub2= NULL;
 ret= Fileliste_collision_test(sub,sperrliste,1|256);
 if(ret==1) 
   goto ex;
 if(ret<0)
   goto kaputt;

 ret= 0; 
ex:;
 if(sub1!=NULL) Fileliste_destroy(&sub1,0);
 if(sub2!=NULL) Fileliste_destroy(&sub2,0);
 if(sub!=NULL) Fileliste_destroy(&sub,0);
 return(ret);
kaputt:
 ret= -1;
 goto ex;
}


int Fileliste_is_collision(liste,target_idx,sperrliste,
                           coll_idx,coll_rel_adr,flag)
struct FilelistE *liste;
int target_idx;
struct FilelistE *sperrliste;
int *coll_idx;
char *coll_rel_adr;
int flag;
/*
 bit0= check for inclusion in both directions
       (else : wether target is included in a list element )
 bit2= check even if excluded by sperrliste 
 bit8= directory opened virtually
*/
{
 char target[2*4096];
 int i,ret=0;

 *coll_idx= -1;
 coll_rel_adr[0]= 0;
 if(Fileliste_search_file(sperrliste,liste->adressen[target_idx],
                          2|(flag&4))>=0)
   return(0);

 Fileliste_target(liste,target_idx,target,1);
 if(target[0]==0 || (target[0]=='/' && target[1]==0)){
   if(liste->typen[target_idx]!=2) 
     /* Pseudo collision: only directories may be mapped to / */
     return(4);
 }

 for(i=0;i<liste->anz;i++){
   *coll_idx= i;
   if(liste->typen[i]<0)
 continue;
   if(i==target_idx) 
 continue;
   if(Fileliste_search_file(sperrliste,liste->adressen[i],2)>=0)
 continue;
   ret= Fileliste_is_collision_sub(liste,target_idx,sperrliste,
                                   *coll_idx,coll_rel_adr,flag&256);
   if(ret!=0) return(ret);
   if(flag&1){
     ret= Fileliste_is_collision_sub(liste,*coll_idx,sperrliste,
                                     target_idx,coll_rel_adr,flag&256);
     if(ret!=0) return(ret);
   }
 }
 return(0);
}


int Fileliste_collision_avoid(liste,idx,sperrliste,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *sperrliste;
int flag;
{
 int l,zaehler,coll_idx;
 char coll_rel_adr[4096],target[4096];
 
 Fileliste_target(liste,idx,target,1|2);
 l= strlen(target);
 zaehler= 0;
 do{
   sprintf(target+l,"_%-3.3d",zaehler++);
   Fileliste_set_target(liste,idx,target,0);
 }while(Fileliste_is_collision(liste,idx,sperrliste,
                               &coll_idx,coll_rel_adr,1)>0);
 
 return(1);
}


int Fileliste_redirect_links(liste,flag)
struct FilelistE *liste;
int flag;
/*
 bit0= do not try to redirect but flatly remove the link from the list
*/
{
 int i,typ,sret,stage,stage_limit= 10;
 char filename[4096],buff[4096],*cpt,target[4096];
 static char typcodes[]={"?fdlphb"};

 for(i=0;i<liste->anz;i++){
   typ= Fileliste_filetyp(liste->adressen[i],0);
   if(typ!=3)
 continue;
   if(flag&1) {
     fprintf(stderr,"--- Symbolic link detected in file list and excluded\n");
     fprintf(stderr,"       :   %s\n",liste->adressen[i]);
     Fileliste_defunct_item(liste,i,0);
 continue;
   } else if(typ==3) {
     fprintf(stderr,"--- Symbolic link detected\n");
     fprintf(stderr,"       :   %s\n",liste->adressen[i]);

     /* try to redirect source redirect to link target */
     strcpy(filename,liste->adressen[i]);
     for(stage= 0; stage<stage_limit && typ==3;stage++) {
       sret= Sfile_readlink(filename,buff,sizeof(buff),0);
       if(sret<=0) {
         fprintf(stderr,"       :   cannot read link target from link\n");
     break;
       }
       if(buff[0]=='/')
         filename[0]= 0;
       else {
         cpt= strrchr(filename,'/');
         if(cpt!=NULL)
           *(cpt+1)= 0;
         else
           filename[0]= 0;
       }
       strcat(filename,buff);
       typ= Sfile_type(filename,1);
       if(typ<=0) {
         fprintf(stderr,"   -> ?:   %s\n",filename);
         if(typ<0)
           fprintf(stderr,"       :   link target not found\n");
         else
           fprintf(stderr,
                      "       :   link target not a recognizable file type\n");
         typ= 3;
     break;
       }
       if(typ<0||(typ>3 && !(typ==6 && (liste->flags[i]&32))))
         typ= 0;
       fprintf(stderr,"   -> %c:   %s\n",typcodes[typ],filename);
     }
     if(stage>=stage_limit) {
       fprintf(stderr,"       :   too many link levels (max. %d)\n",
                      stage_limit);
     }
     if(typ==3 || typ<=0) {
       /*  no valid link target found */
       Fileliste_defunct_item(liste,i,0);
       fprintf(stderr,
                 " SORRY :   could not redirect source to link target\n");
 continue;
     } else {
       /* only change source address but not target.  */
       Fileliste_target(liste,i,target,2);
       if(Fileliste_set_address(liste,i,filename,0)<=0)
         return(-1);
       if(Fileliste_set_target(liste,i,target,0)<=0)
         return(-1);
       liste->typen[i]= typ;
       fprintf(stderr,
                 " OK    :   source successfully redirected to link target\n");
     }
   }
 }
 return(1);
}


int Fileliste_collision_test(liste,sperrliste,flag)
struct FilelistE *liste;
struct FilelistE *sperrliste;
int flag;
/*
 bit0= exit after first collision, do not report collision to stderr
 bit1= change targets to avoid collisions
 bit2= check even if excluded by sperrliste 
 bit3= like bit2 but only with the first item
 bit8= directory opened virtually
*/
{
 int i,ret,coll_idx,swarwas=0,typ,t_flag;
 char coll_rel_adr[4096],target[4096];
 static char typcodes[]={"?fdl"};

 for(i=0;i<liste->anz;i++){
   if((flag&4)||((flag&8)&&i==0)) t_flag= 4;
   else                           t_flag= 0;
   t_flag|= (flag&256);
   ret= Fileliste_is_collision(liste,i,sperrliste,
                               &coll_idx,coll_rel_adr,t_flag);
   if(ret>0) {
     swarwas++;
     if(flag&1) return(1);
     if(coll_rel_adr[0] || i>coll_idx){
       fprintf(stderr,"--- Target collision detected:\n"); 
       typ= Fileliste_filetyp(liste->adressen[i],0);
       if(typ<0||typ>3) typ= 0;
       fprintf(stderr,"    1 %c:   %s\n",typcodes[typ],liste->adressen[i]); 
       if(coll_idx>=0){
         typ= Fileliste_filetyp(liste->adressen[coll_idx],0);
         if(typ<0||typ>2) typ= 0;
         fprintf(stderr,"    2 %c:   %s",typcodes[typ],
                                         liste->adressen[coll_idx]);
         if(coll_rel_adr[0]) 
           fprintf(stderr," / %s\n",coll_rel_adr);
         else
           fprintf(stderr,"\n");
       } else {
         fprintf(stderr,
                 "       :   not a usable target for data files\n"); 
       }
       /* report target address */
       Fileliste_target(liste,i,target,1|2);
       fprintf(stderr,"    CD :   %s\n",target);

       if(flag&2){
         if(Fileliste_collision_avoid(liste,i,sperrliste,0)>0){
           Fileliste_target(liste,i,target,2);
           fprintf(stderr,"    1! :   %s=%s\n",target,liste->adressen[i]);
         } else {
           fprintf(stderr,"    failed to find new target address\n");
         }
       }
     }
   }
   if(ret<0) return(-1); 
 } 
 return(swarwas>0);
}


int Fileliste__compose_pacifier(linecount,subcount,size,text,flag)
int linecount;
int subcount;
double size;
char text[4096];
int flag;
/*
 bit0= print regardless of time (if not totally disabled)
 bit3= end of pacifying, print newline
*/
{
 char line[4096+100],level_text[16],size_text[16],mem_text[8],*mpt;

 Sfile_scale(size,size_text,5,1e4,1);
 if(Cd_backup_planer_compose_pacifier) {
   Sfile_scale(Cdbp_subs_mem,mem_text,4,1e3,1);
   for(mpt= mem_text;(*mpt)==' ';mpt++);
   if(Cdbp_compose_sublevel_count>=0)
     sprintf(level_text,"%d",Cdbp_compose_sublevel_count);
   else
     strcpy(level_text,"X");
   sprintf(line,"%s:%.lf:%s %s %s",level_text,Cdbp_subs,mpt,size_text,text);
 } else
   sprintf(line,"%s %s",size_text,text);
 if(subcount>=0) { /* modest subcount */
   char hline[4096+200];
   sprintf(hline,"%d/%d  %s",linecount,subcount,line);
   Print_pacifier(linecount,hline,4|16|32|(1<<10)|(flag&(1|8)));
 } else  
   Print_pacifier(linecount,line,2|4|16|32|512|(flag&(1|8)));
 return(1);
}


int Fileliste_determine_treesize(liste,idx,dirlen,size,linecount,flag)
/* Note: does not care for exclusion lists. Fileliste_dir_size() does that */
struct FilelistE *liste;
int idx;
int dirlen;    /* length of (target) directory name */
double *size;
int *linecount;
int flag;
/*
 bit0= recursion
 bit6= print pacifier with linecount
*/
{
 int i,typ,ret,adrlen= 0,hflag,original_sublevel_rec;
 double start_time,current_time;
 double sub_size= 0.0,compose_size_rec= 0.0,new_mem= 0.0;
 struct FilelistE *sub= NULL;
 char *name,*cpt,target[4096];

 original_sublevel_rec= Cdbp_compose_sublevel_count;
 *size= 0.0;
 typ= Fileliste_type_and_size(liste,idx,size,1,0);
 if(typ<=0)
   {ret=0; goto ex;}
 liste->typen[idx]= typ;
 if(Fileliste_node_addrl) {
   if(!(flag&1)) { /* top level : set adrlen */
     if(Fileliste_target(liste,idx,target,2)<=0) {
       adrlen= strlen(liste->adressen[idx]);
     } else {
       adrlen= strlen(target);
     }
   } else {
     name= liste->adressen[idx];
     for(cpt= liste->adressen[idx]; *cpt!=0; cpt++)
       if(*cpt=='/')
         if(*(cpt+1)!=0)
           name= cpt+1;
     adrlen= dirlen+strlen(name);
   }
 } else 
   adrlen= 0;
 *size+= adrlen+Fileliste_node_overhead;
 liste->groessen[idx]= *size;

 if(typ!=2)
   {ret= 1; goto ex;}

 if(flag&64) { 
   compose_size_rec= Cd_backup_planer_compose_size;
   if((*linecount%100)==0 && (*linecount)>0) {
     Cd_backup_planer_compose_size+= *size;
     Fileliste__compose_pacifier(*linecount,-1,Cd_backup_planer_compose_size,
                                 liste->adressen[idx],0);
   }
   (*linecount)++;
 }

 /* create file list (more or less economy class) and read directory content */
 if(Cdbp_compose_keep_sublevels>Cdbp_compose_sublevel_count+1
    && Cdbp_compose_sublevel_count>=0)
   hflag= 256|512|2048|(1<<13);                         /* will need subdirs */
 else 
   hflag= 2;                                         /* will need no subdirs */
 if(Fileliste_new(&sub,0,liste->main_dir,hflag)<=0)
   {ret= -1; goto ex;}

 Fileliste_dir(sub,liste->adressen[idx],linecount,4|8|(flag&64));

 if(Cdbp_compose_sublevel_count>=0) {
   if(Cdbp_compose_keep_sublevels>Cdbp_compose_sublevel_count
      && liste->subdirs!=NULL) {
     if(sub->anz>Cdbp_compose_sublevel_maxsize) {
       Cdbp_compose_sublevel_count= -2; /* no deeper */
     } else {
       new_mem+= Fileliste_estimate_mem_usage(sub,0);
       if(Cdbp_compose_sublevel_maxmem<=0 ||
          Cdbp_subs_mem+new_mem<Cdbp_compose_sublevel_maxmem)
         Fileliste_set_subdir(liste,idx,sub,0);
     }
   }
   Cdbp_compose_sublevel_count++;
 }

 start_time= Sfile_microtime(0);
 for(i=0;i<sub->anz;i++) {
   ret= Fileliste_determine_treesize(sub,i,adrlen,&sub_size,linecount,
                                     1|(flag&64));
   if(ret<0)
     goto ex;
   if(ret>0)
     *size+= sub_size;
   if((flag&64) && (i%100)==0 && i>0) {
     current_time= Sfile_microtime(0);
     if(current_time-start_time>=1.0) {
       Cd_backup_planer_compose_size= compose_size_rec+ *size;
       Fileliste__compose_pacifier(*linecount,i,Cd_backup_planer_compose_size,
                                   liste->adressen[idx],0);
       start_time= current_time-0.2;
     }
   }
 }
 if((flag&64)) {
   Cd_backup_planer_compose_size= compose_size_rec+ *size;
   Fileliste__compose_pacifier(*linecount,-1,Cd_backup_planer_compose_size,
                               liste->adressen[idx],0);
 }
 liste->groessen[idx]= *size;
 ret= 1;
ex:;
 Cdbp_compose_sublevel_count= original_sublevel_rec;
 Fileliste_destroy(&sub,0); /* will leave sub alive if attached to subdirs[] */
 return(ret);
}


int Fileliste_leanfind(liste,idx,parent_dev,fp,fp_aux,linecount,depth,flag)
/* replacement for shell command  "find"  */
struct FilelistE *liste;
int idx;
dev_t parent_dev;
FILE *fp,*fp_aux;
int *linecount;
int depth;
int flag;
/*
 bit0= recursion
 bit1= use relative addresses
        /x/x/x/   absolute directory address
        /x/x/x/   absolute file address
        ./x/x/x/  relative directory address
        ./x/x/x   relative file name
        ../../..  one or more directories up
        ../       to start directory
        #=# xxx   additional program information
       top_level addresses get prepended the target
 bit2= do not follow mount points
 bit3= do not follow net mount points
 bit4= print linecount
 bit5= generate dir_list (with property information)
 bit6= with bit 5 : generate dir_script (chown, chmod)

 <<< bit7= ignore names listed in liste->sperrliste

 bit9= prepend type
 bit10= prepend length
 bit11= preprend crc
 bit12= use MD5 rather than CRC32
 bit13= prepend inode information (as extension to type)
 bit14= with bit5 : generate both : dir_list to fp , dir_script to fp_aux 
 bit15= do not use Dirseq (i.e. use ls -a : all at once, sorted )
 bit16= prepend mtime
 bit17= prepend ctime
*/
{
/* A71226 : switched from local function memory to malloc()
            for more recursion depth.
    char *typetext[256],lengthtext[80],crctext[80];
    char *adr_data[4096],combined_adr[4*4096],adr_aux[4096];
*/
 char *name,*npt,slash[2],dotslash[3];
 char *typetext= NULL,*lengthtext= NULL,*crctext= NULL;
 char *adr,*adr_data= NULL,*combined_adr= NULL,*adr_aux= NULL;
 char *target= NULL, *temp_src_adr= NULL,*part_adr= NULL;
 char *count_text= NULL,*part_crc_text= NULL;
 unsigned char md5[16];
 int i,ret,is_dir= 0,typ,read_ret= 0,report_step= 1000,has_part_rec= 0;
 int zero= 0;
 double size;
 unsigned long crc;
 struct FilelistE *sub= NULL;
 dev_t st_dev;
 dev_t device;
 ino_t inode;
 time_t atime,mtime,ctime;
 struct DirseQ *dirseq= NULL;

 ret= Cd_backup_planer_assert_depth(depth, "", 0);
 if(ret<=0)
   {ret= -1; goto ex;}

 adr_data= malloc(4096);
 combined_adr= malloc(4*4096);
 adr_aux= malloc(4096);
 typetext= malloc(256);
 lengthtext= malloc(80);
 crctext= malloc(80);
 if(adr_data==NULL || combined_adr==NULL || adr_aux==NULL ||
    typetext==NULL || lengthtext==NULL || crctext==NULL) {
   fprintf(stderr, "Memory exhaustion in directory traversal\n");
   ret= -1; goto ex;
 }

 adr= adr_data;
 if(flag&2048)
   report_step= 100;
 /* test name */
 name= liste->adressen[idx];
 if(name==NULL)
   {ret= 0; goto ex;}
 if(name[0]==0)
   {ret= 0; goto ex;}
 if(liste->part_recs!=NULL)
   if(liste->part_recs[idx]!=NULL)
     has_part_rec= 1;
 /* obtain requested prepend information or determine yet unknown file type */
 typetext[0]= lengthtext[0]= crctext[0]= 0;
 read_ret= 2|4|8;
 typ= liste->typen[idx];
 if((flag&1024)||typ==0) {
   typ= Type_and_size(name,&size,1|16);
   if(typ<=0)
     read_ret&= ~(2|4);
   if(liste->typen[idx]==0)
     liste->typen[idx]= typ;
 }
 is_dir= (liste->typen[idx]==2);
 if((flag&32) && !is_dir)
   {ret= 2; goto ex;}
 if(is_dir&& (flag&1)) 
   if(Is_forbidden_mountpoint(name,parent_dev,&st_dev,
                              16|((flag>>2)&3)))
     {ret= 2; goto ex;}

 if(flag&16) 
   if((((*linecount)+1)%(report_step/100))==0 || (*linecount)<=2) 
     Print_pacifier((*linecount)+1,name,2|4|16|32);

 if((flag&2048)) {
   if(typ==0)
     ret= 0;
   else if(flag&(1<<12)) {
     read_ret|= 16;
     ret= Determine_file_checksum(name,NULL,NULL,md5,md5,&size,NULL,NULL,
                                  32|64|(1<<12));
   } else {
     ret= Determine_file_checksum(name,NULL,NULL,&crc,&crc,&size,NULL,NULL,
                                  32|64);
   }
   if(ret<=0)
     read_ret&= ~(8|16);
 }

 if(flag&(1<<13)) {
   if(File_inode(name,&device,&inode,0)>0) {
     read_ret|= 32;
     if((typ==2)||!(flag&1)) /* use deviceno if directory or toplevel */
       read_ret|= 64;
   }
 }
 if(flag&(3<<16)) {
   ret= Get_filetimes(name,&atime,&mtime,&ctime,2)  ;
   if(ret>0)
     read_ret|= (128|256);
 }

 if(flag&2) {
   slash[0]= slash[1]= 0;
   if(is_dir)
                              /*  represent root-dir as /=/ and not as //=// */
     if(name[strlen(name)-1]!='/') 
       slash[0]= '/';
   if(flag&1) {
     /* ./basename[/] , ignore any trailing slash */;
     for(npt= name+strlen(name)-1;npt>name;npt--)
       if(*npt=='/')
     break;
     if(*npt=='/')
       npt++;
     sprintf(adr,"./%s%s",npt,slash);
   } else {
     dotslash[0]= dotslash[1]= dotslash[2]= 0;
     if(name[0]!='/')
       strcpy(dotslash,"./");
     if(name[0]=='.')
       if(name[1]=='/')
         dotslash[0]= 0;
     sprintf(adr,"%s%s%s",dotslash,name,slash);
   }
 } else
   strcpy(adr,name);
 strcpy(adr_aux,name);

 Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                     device,inode,mtime,ctime,
                     read_ret&(2|4|8|16|32|64|128|256),
                     1|(flag&(512|1024|2048|8192|(3<<16))));
 if((flag&64)&&!(flag&(1<<14))) {
   /* script format has to record target addresses */
   Fileliste_target(liste,idx,adr,1|2);
   if(adr[0]==0)
     strcpy(adr,".");
 } else {
   if((flag&2)&&!(flag&1)) { /* at top recursion level */
     /* prepend  target  to adr */;
     if(target==NULL)
       target= malloc(4096);
     Fileliste_target(liste,idx,target,2);
     Fileliste__format_combined_adr(combined_adr,target,adr,0);
   } else /* 0520 */
     Fileliste__esc_newline(adr,combined_adr,0);
   adr= combined_adr;
   if(flag&(1<<14)) {
     /* script format has to record target addresses */
     Fileliste_target(liste,idx,adr_aux,1|2);
     if(adr_aux[0]==0)
       strcpy(adr_aux,".");
   } 
 }
 if(typetext[0]!=0)
   strcat(typetext," ");
 if(flag&32) {
   ret= Record_dir_properties(fp,fp_aux,name,adr,adr_aux,lengthtext,crctext,
                              ((flag>>6)&1)|((flag>>13)&2));
   if(ret<=0)
     goto ex;
 } else {
   if(has_part_rec) {
     /* 0497: */
     int type_len,partno,partcount;
     double start_byte,byte_count;
/*   A71226 : switched from local function memory to malloc()
              for more recursion depth.
              Pointer declarations have been moved to function head.
     char count_text[80],part_crc_text[80];
     char target[4096],temp_src_adr[4096],part_adr[3*4096];
*/
     int sizevec[CdbplivS],skipvec[CdbplivS];

     if(count_text==NULL)
       count_text= malloc(80);
     if(part_crc_text==NULL)
       part_crc_text= malloc(80);
     if(target==NULL)
       target= malloc(4096);
     if(temp_src_adr==NULL)
       temp_src_adr= malloc(4096);
     if(part_adr==NULL)
       part_adr= malloc(3*4096);
     if(target==NULL || temp_src_adr==NULL || part_adr==NULL ||
        count_text==NULL || part_crc_text==NULL) {
       fprintf(stderr, "Memory exhaustion in directory traversal\n");
       ret= -1; goto ex;
     }

     part_crc_text[0]= 0;
     type_len= strlen(typetext);
     if(type_len==0) {
       /* ??? what to do if there is no type column ? */;
       fprintf(stderr,
               "cd_backup_planer: missing type column with split file\n");
       {ret= -1; goto ex;}
     }
     type_len--;
     Filelist_part_rec_get_rec(liste->part_recs[idx],&partno,&partcount,
                               &start_byte,&byte_count,0);

     /* write line for file part */;
     sprintf(typetext+type_len,",p%.f:%.f ",start_byte,size);
     sprintf(count_text,"%.f ",byte_count);
     /* determine part checksum */;
     part_crc_text[0]= 0;
     if((flag&2048)) {
       Intvec_from_double(sizevec,CdbplivS,byte_count,0);
       Intvec_from_double(skipvec,CdbplivS,start_byte,0);
       if(typ==0)
         ;
       else if(flag&(1<<12)) {
         ret= Determine_file_checksum(name,NULL,NULL,md5,md5,&size,sizevec,
                                      skipvec,32|64|128|(1<<12));
         for(i=0;i<16;i++)
           sprintf(part_crc_text+2*i,"%2.2x ",md5[i]);
       } else {
         ret= Determine_file_checksum(name,NULL,NULL,&crc,&crc,&size,sizevec,
                                      skipvec,32|64|128);
         sprintf(part_crc_text,"%lX",crc);
       }
     } 
     /* compose part_adr */; 
     Fileliste_target(liste,idx,target,2);

     /* >>> SPLIT: for some reason this retrieves -00 with any adr */
     Filelist_part_rec_get_tempadr(liste->part_recs[idx],temp_src_adr,0);

     Fileliste__format_combined_adr(part_adr,target,temp_src_adr,0);
     fprintf(fp,"%s%s%s%s\n",typetext,count_text,part_crc_text,part_adr);
     /* change typetext for original file line */
     sprintf(typetext+type_len,",P%.f:%.f ",start_byte,byte_count);
   }
   fprintf(fp,"%s%s%s%s\n",typetext,lengthtext,crctext,adr);

 }

 (*linecount)++;
 /* if not a directory : done */
 if(!is_dir)
   {ret= 1; goto ex;}
 if(!(flag&(1<<15))) {
   /* create directory reader and single-line filelist */
   ret= Dirseq_new(&dirseq,name,0);
   if(ret<0)
     goto ex;
   if(ret==0) {
     /* unreadable directory */
     goto mark_end;
   }
   ret= Fileliste_single(&sub,"dummy_leanfind",0,0.0,16);
   if(ret<=0)
     {ret= -1; goto ex;}
   Fileliste_set_sperrliste(sub,liste->sperrliste,0);
   /* examine directory file objects */
   adr= combined_adr;
   strcpy(adr,name);
   if(adr[strlen(adr)-1]!='/')
     strcat(adr,"/");
   adr+= strlen(adr);
   while(1) {
     ret= Dirseq_next_adr(dirseq,adr,1);
     if(ret<0)
   continue;
     if(ret==0)
   break;
     /* do not examine excluded objects */
     if(Fileliste_search_file(liste->sperrliste,combined_adr,32)>=0)
   continue;
     ret= Fileliste_set_address(sub,0,combined_adr,0); 
     if(ret<=0)
       {ret= -1; goto ex;}
     sub->typen[0]= 0;
     sub->flags[0]= 0;
     ret= Fileliste_leanfind(sub,0,st_dev,fp,fp_aux,linecount,depth+1,1|flag);
     if(ret<0)
       goto ex;
   }
 } else {
   /* create file list and read directory content */
   if(Fileliste_new(&sub,0,liste->main_dir,2)<=0) 
     {ret= -1; goto ex;}
   Fileliste_set_sperrliste(sub,liste->sperrliste,0);
   Fileliste_dir(sub,name,&zero,2);
   if(flag&32)
     Fileliste_inherit_target(liste,idx,sub,1);
 
   /* examine found file objects */
   ret= 0;
   for(i=0;i<sub->anz;i++) {
     /* do not examine excluded objects */
     if(Fileliste_search_file(liste->sperrliste,sub->adressen[i],32)>=0)
   continue;
     ret|= Fileliste_leanfind(sub,i,st_dev,fp,fp_aux,linecount,depth+1,1|flag);
   }
 }

mark_end:;
 /* mark end of (sub-)directory */
 if((flag&2)&&!(flag&64)) {
   if(flag&1)
     fprintf(fp,"..\n");
   else
     fprintf(fp,"../\n");
 }

 ret= 1;
ex:;
 if(adr_data!=NULL)
   free(adr_data);
 if(combined_adr!=NULL)
   free(combined_adr);
 if(adr_aux!=NULL)
   free(adr_aux);
 if(typetext!=NULL)
   free(typetext);
 if(lengthtext!=NULL)
   free(lengthtext);
 if(crctext!=NULL)
   free(crctext);
 if(count_text!=NULL)
   free(count_text);
 if(part_crc_text!=NULL)
   free(part_crc_text);
 if(target!=NULL)
   free(target);
 if(temp_src_adr!=NULL)
   free(temp_src_adr);
 if(part_adr!=NULL)
   free(part_adr);
 Dirseq_destroy(&dirseq,0);
 Fileliste_destroy(&sub,0);
 return(ret);
}


int Fileliste_merge_trees(liste,idx,parent_dev,linecount,
                          source_prefix,target_prefix,removed_prefix,old_tree,
                          flag)
/* move one file tree into another one */
struct FilelistE *liste;
int idx;
dev_t parent_dev;
int *linecount;
char *source_prefix;
char *target_prefix;
char *removed_prefix;
struct FilelistE *old_tree;
int flag;
/*
 bit0= recursion
 bit1= precheck mode : do not create or move objects
 bit2= do not follow mount points
 bit3= do not follow net mount points
 bit4= print linecount
 bit5= do not abort at first error
 bit6= copy file objects rather than moving them
 bit8-11= merge mode
       0= never overwrite existing target objects
       1= always overwrite existing target objects
       2= overwrite existing target objects if older than source object
 bit12= initialize rootlist with source_prefix (if not recursion)
 bit13= make target objects writable by owner
*/
{
 char *name,*srcpt;
 char target[4096],link_target[4096],new_link_target[4096],removed[4096];
 char line[2*4096];
 int i,ret,is_dir= 0,is_link= 0,typ,report_step= 1000, merge_mode;
 int target_type,link_limit= 100,link_count= 0,directory_link= 0,zero= 0;
 int new_dir_created= 0,errn,rec_ret,is_dir_link,prop_flag= 0,old_mode=0700;
 double size;
 struct FilelistE *sub= NULL,*rootlist= NULL;
 struct stat stbuf;
 dev_t st_dev;

 uid_t uid;
 gid_t gid;
 time_t date;
 struct FilelistE *found_list= NULL;
 int found_idx;


 merge_mode= (flag>>8)&15;

 if((flag&(1<<12)) && ! (flag&1)) {
   typ= Type_and_size(source_prefix,&size,1);
   if(typ<0) {
no_source_prefix:;
     fprintf(stderr,
             "Fileliste_merge_trees : source_prefix does not exist\n");
     fprintf(stderr,"source_prefix=%s\n",source_prefix);
     ret= 0; goto ex;
   }
   ret= Fileliste_single(&rootlist,source_prefix,typ,size,8);
   if(ret<=0)
     goto ex;
   liste= rootlist;
   idx= 0;
   *linecount= 0;
   if(Cd_backup_planer_call_lstaT(source_prefix,&stbuf)==-1)
     goto no_source_prefix;
   parent_dev= stbuf.st_dev;
 }

 is_dir= (liste->typen[idx]==2);
 is_link= (liste->typen[idx]==3);
 if(is_link)
   is_dir_link= (Type_and_size(liste->adressen[idx],&size,4)==2);
 /* test name */
 name= liste->adressen[idx];
 if(name==NULL)
   {ret= 0; goto ex;}
 if(name[0]==0)
   {ret= 0; goto ex;}
 if(is_dir&& (flag&1)) 
   if(Is_forbidden_mountpoint(name,parent_dev,&st_dev,
                              16|((flag>>2)&3)))
     {ret= 2; goto ex;}

 /* compose target address */;
 srcpt= strstr(name,source_prefix);
 if(srcpt!=name) {
   fprintf(stderr,
           "Fileliste_merge_trees : list item does not match source_prefix\n");
   fprintf(stderr,"item[%7.7d]=%s\n",idx,name);
   fprintf(stderr,"source_prefix=%s\n",source_prefix);
   {ret= 0; goto ex;}
 }
 strcpy(target,target_prefix);
 strcat(target,name+strlen(source_prefix));
 target_type= Type_and_size(target,&size,1);
 removed[0]= 0;
 if(!(strcmp(removed_prefix,"-")==0 || removed_prefix[0]==0)) {
   strcpy(removed,removed_prefix);
   strcat(removed,name+strlen(source_prefix));
 }

 if(is_dir) {

   /* does target address exist ? */
   if(target_type!=-1) { 
     /* is it a directory or a link to a directory ? */;
     if(target_type==3) {
       strcpy(new_link_target,target);
       while(target_type==3 && link_count++<link_limit) {
         strcpy(link_target,new_link_target);
         ret= Sfile_readlink(link_target,
                             new_link_target,sizeof(new_link_target),1);
         if(ret<=0)
       break;
         typ= Type_and_size(new_link_target,&size,1);
         if(typ==2) {
           directory_link= 1;
       break;
         } else if(typ!=3)
       break;
       }
     }
     if(target_type==2 || directory_link) {

       /* ok */;

     } else {
       if(merge_mode==1) {
         /*  replace by directory */;
         if(!(flag&2)) {
           ret= Remove_file(target,removed,0);
           if(ret<=0)
             goto ex;
         }
         goto create_new_directory;
       } else {
         fprintf(stderr,"%s : %s",
                        "Fileliste_merge_trees",
                        "directory target exists but is not a directory\n");
         fprintf(stderr,"source=%s\n",name);
         fprintf(stderr,"target=%s\n",target);
         {ret= 0; goto ex;}
       }
     }
   } else {
create_new_directory:;
     if(!(flag&2)) {
       ret= mkdir(target,0700);
       if(ret==-1) {
         fprintf(stderr,
                "Fileliste_merge_trees : creation of directory failed :\n");
         fprintf(stderr,"target=%s\n",target);
         fprintf(stderr,"%s\n",strerror(errno));
       }
       new_dir_created= 1;
     }
   }

 } else if(is_link) {

   /* >>> how to handle softlinks ? */;
     /* >>> readjust link content */;
 
 } else { /* any other filetype */

   /* does target address exist ? */
   if(target_type!=-1) { 
     /* decide wether to overwrite or not */
     if(merge_mode==0)
       {ret= 2; goto ex;}
     else if(merge_mode==2) {
       if(Is_older_file(target,name,0)<=0)
         {ret= 2; goto ex;}
     }
   } 

   /* move file object */;
   if(!(flag&2)) {
     if(target_type!=-1) {
       ret= Remove_file(target,removed,0);
       if(ret<=0)
         goto ex;
     }
     /* copy or move source to target */
     if(flag&64) {
       ret= Sfile_copy_file(name,target,&errn,line,0);
       if(ret<=0) {
         fprintf(stderr,
              "Fileliste_merge_trees : copy from source to target failed :\n");
         fprintf(stderr,"source=%s\n",name);
         fprintf(stderr,"target=%s\n",target);
         fprintf(stderr,"%s\n",line);
         goto ex;
       }
       Sfile_copy_properties(name,target,1|4|(flag&(1<<13)));
     } else {
       ret= Sfile_move_file(name,target,&errn,line,0);
       if(ret<=0) {
         fprintf(stderr,
              "Fileliste_merge_trees : move from source to target failed :\n");
         fprintf(stderr,"source=%s\n",name);
         fprintf(stderr,"target=%s\n",target);
         fprintf(stderr,"%s\n",line);
         goto ex;
       }
     }
   }
 }

 (*linecount)++;
 if(flag&16) 
   if(((*linecount)%(report_step/100))==0 || (*linecount)==2) {
     sprintf(line,"  %-10d : %s",*linecount,name);
     Print_pacifier(0,line,4|16|32);
   }

 /* if not a directory : done */
 if(!is_dir)  /* ???  allow directory links at recursion level 0 ? */
   {ret= 1; goto ex;}

 /* create file list and read directory content */
 if(Fileliste_new(&sub,0,liste->main_dir,0)<=0) 
   {ret= -1; goto ex;}
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);
 Fileliste_dir(sub,name,&zero,2);

 /* cautious settings while installing files */
 if(!(flag&2))
 if(merge_mode==1 || merge_mode==2 || new_dir_created) {
   if(new_dir_created) {
     Cd_backup_planer_call_staT(name,&stbuf);
     prop_flag= 1|2|4; /* set permissions for user+group+others */
   } else {
     Cd_backup_planer_call_staT(target,&stbuf);
     prop_flag= 1;     /* only set permissions for user */
   }
   old_mode= stbuf.st_mode&0xfff;
   Sfile_set_properties(target,0700,(uid_t) 0,(gid_t) 0,(time_t) 0,(time_t) 0,
                        (time_t) 0,prop_flag); 
 }

 /* examine found file objects */
 ret= 0;
 for(i=0;i<sub->anz;i++) {
   /* do not examine excluded objects */
   if(Fileliste_search_file(liste->sperrliste,sub->adressen[i],32)>=0)
 continue;
   rec_ret= Fileliste_merge_trees(sub,i,st_dev,linecount,
                                  source_prefix,target_prefix,removed_prefix,
                                  old_tree,1|flag);
   ret|= rec_ret;
   if(rec_ret<0)
 break;
   if(rec_ret==0 && !(flag&32))
 break;
 }

 /* eventually try to copy directory properties */
 if(!(flag&2)) {
   if(merge_mode==1 || merge_mode==2 || new_dir_created) {
     ret= Sfile_copy_properties(name,target,1|4|(flag&(1<<13)));
     if(ret<=0) {
       fprintf(stderr,"%s : %s",
                      "Fileliste_merge_trees",
                      "cannot copy directory properties\n");
       fprintf(stderr,"source=%s\n",name);
       fprintf(stderr,"target=%s\n",target);
       {ret= 0; goto ex;}
     }
     prop_flag= 0; /* old_mode is invalid now */

#ifndef Not_yeT
     ret= 0;
     /* try to get source permissions as they should be */;
     if(old_tree!=NULL)
       ret= Fileliste_search_branch(old_tree,&found_list,&found_idx,name,0);
     if(ret<0)
       goto ex;
     if(ret>0) {
      if(found_list->xprops!= NULL) {
       if(found_list->xprops[found_idx]!= NULL) {
         ret= Get_user_group_id(found_list->xprops[found_idx]->own,
                               found_list->xprops[found_idx]->grp,&uid,&gid,0);
         if(ret>0) {
           date= found_list->xprops[found_idx]->date;
           Sfile_set_properties(target,found_list->xprops[found_idx]->mod,
                                uid,gid,date,date,date,1|2|4|8|16|32|128);
         }
     }}}
#endif /* !  Not_yeT */

   }

   /* Try to remove the source directory */
   if(!(flag&64))
     Remove_dir(name,4);
 }

 Fileliste_destroy(&sub,0);
 ret= 1; goto ex;

ex:;
 if(prop_flag!=0) /* directory properties have been altered temporarily */
   Sfile_set_properties(target,old_mode,(uid_t) 0,(gid_t) 0,
                               (time_t) 0,(time_t) 0,(time_t) 0,prop_flag); 
 if(rootlist!=NULL)
    Print_pacifier(*linecount,"file objects processed",1|2|4|8|16);
 Fileliste_destroy(&rootlist,0);
 return(ret);
}


int Fileliste_find_path_list(path_list,flag)
char *path_list;
int flag;
{
 char line[4096],*fret;
 struct FilelistE *find_liste= NULL;
 FILE *fp= NULL; 
 int ret,linecount= 0,zero= 0;

 fp= Sfile_fopen(path_list,"rb",0);
 if(fp==NULL)
   {ret= 0; goto ex;}

 while(1) {
   fret= Sfile_fgets(line,sizeof(line)-1,fp);
   if(fret==NULL)
 break;
   if(line[0]==0)
 continue;
   if(Fileliste_single(&find_liste,line,0,0.0,8|16)>0){
     if(Fileliste_read_properties(find_liste,&zero,2)>0) 
       Fileliste_leanfind(find_liste,0,dev_dummy,stdout,NULL,&linecount,0,0);
     Fileliste_destroy(&find_liste,0);
   }
 }
 ret= 1;
ex:;
 if(fp!=NULL)
   Sfile_fclose(fp);
 return(ret);
}


#ifndef Crawler_ruleS


int Fileliste_detect_change(liste,idx,old_tree,old_tree_idx,size,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *old_tree;
int old_tree_idx;
double size;
int flag;
/*
 bit0= do not search but use  old_tree  and  old_tree_idx  as found
 bit6= for File_is_changed(): eventually pacify by size count
 bit8= check for existence in old_tree (else assume no change)
 bit9= with bit8 : check type in old_tree
 bit10= with bit8 : check size in old_tree
 bit11= with bit8 : check crc in old_tree
 bit13= with bit8 : check inode in old_tree
 bit14= with bit8 : allow inode check to prevent crc check
 bit15= with bit8 : demand inode check only if old item has inode information
 bit17= with bit13: allow to ignore device number

return:
 <0 = error
  0 = no change detected
  1 = change detected
  2 = not found in old_tree
*/
{
 int found_idx,ret,inode_test= 0;
 unsigned long lcrc= 0,fcrc= 0;
 unsigned char *lmd5= NULL,*fmd5= NULL;
 char *adrpt,abs_adr[4096];
 static unsigned char dummy_md5[16]= {"not an md5      "};
 struct FilelistE *found_list;
 dev_t fdevice= 0,ldevice;
 ino_t finode= 0,linode;

 if(!(flag&256))
   return(0);
 if(flag&1) {
   found_list= old_tree;
   found_idx= old_tree_idx;
 } else {
   adrpt= liste->adressen[idx];
   if(strcmp(adrpt,"/")==0)
     return(1);

   /* if relative address : prepend working directory */;
   /* <<< this should happen higher in the function stack */
   if(adrpt[0]!='/') {
     if(adrpt[0]=='.' && adrpt[1]=='/')
       adrpt+=2;
     if(Cd_backup_planer_cwD[0]!=0) {
       if(adrpt[0]==0)
         strcpy(abs_adr,Cd_backup_planer_cwD);
       else
         sprintf(abs_adr,"%s/%s",Cd_backup_planer_cwD,adrpt);
       adrpt= abs_adr;
     }
   }

   if(Fileliste_search_branch(old_tree,&found_list,&found_idx,
                              adrpt,0)<=0)
     return(2);
 }

 if(found_list->crcs!=NULL) {
   if(found_list->crcs[found_idx]!=NULL) {
     fcrc= found_list->crcs[found_idx]->crc;
     fmd5= found_list->crcs[found_idx]->md5;
   }
 } else {
   fcrc= 0;
   fmd5= dummy_md5;
 } 
   
 inode_test= flag&(3<<13);
 if(found_list->inodes!=NULL) {
   if(found_list->flags[idx]&64) {
     finode= found_list->inodes[found_idx];
     /* if directory : obtain device number from subdir list */
     if(found_list->typen[found_idx]==2 && !(flag&(1<<17))) {
       int has_device= 1;
       if(found_list->subdirs==NULL)
         has_device= 0;
       else if(found_list->subdirs[found_idx]==NULL)
         has_device= 0;
       else if(!(found_list->subdirs[found_idx]->dir_dev_ino_valid&1))
         has_device= 0;
       if(has_device)
         fdevice= found_list->subdirs[found_idx]->dir_device;
       else {
no_fdevice:;
         if(inode_test) {
           if(flag&(1<<15)) 
             inode_test= 0;
           else
             return(1); /* pretend change */
         }
       }
     } else if(!(flag&(1<<17))) { /* obtain device number from hosting list */
       if(found_list->dir_dev_ino_valid&1)
         fdevice= found_list->dir_device;
       else
         goto no_fdevice;
     }
   } else if(flag&(1<<15)) {
     inode_test= 0;
   }
 }

 if(liste->crcs!=NULL) {
   if(liste->crcs[idx]!=NULL) {
     lcrc= liste->crcs[idx]->crc;
     lmd5= liste->crcs[idx]->md5;
   }
 } else {
   lcrc= 0;
   lmd5= dummy_md5;
 }

 if(liste->inodes!=NULL) {
   if(liste->flags[idx]&64) {
     linode= liste->inodes[idx];
     if(liste->typen[idx]==2 && !(flag&(1<<17))) {
       int has_device= 1;
       if(liste->subdirs==NULL)
         has_device= 0;
       else if(liste->subdirs[idx]==NULL)
         has_device= 0;
       else if(!(liste->subdirs[idx]->dir_dev_ino_valid&1))
         has_device= 0;
       if(has_device)
         ldevice= liste->subdirs[idx]->dir_device;
       else {
no_ldevice:;
         if(inode_test) {
           if(flag&(1<<15)) 
             inode_test= 0;
           else
             return(1); /* pretend change */
         }
       }
     } else if(!(flag&(1<<17))) { /* obtain device number from hosting list */
       if(liste->dir_dev_ino_valid&1)
         ldevice= liste->dir_device;
       else
         goto no_ldevice;
     }
   } else if(inode_test) {
     ret= File_inode(liste->adressen[idx],&ldevice,&linode,0);
     if(ret>0) {
       liste->inodes[idx]= linode;
       if(liste->typen[idx]==2 && !(flag&(1<<17))) {
         int has_subdir= 1;
         if(liste->subdirs==NULL)
           has_subdir= 0;
         else if(liste->subdirs[idx]==NULL)
           has_subdir= 0;
         if(has_subdir) {
           liste->subdirs[idx]->dir_device= ldevice;
           liste->subdirs[idx]->dir_inode= linode;
           liste->subdirs[idx]->dir_dev_ino_valid|= 3;
         } else {

           /* >>> what in case of missing subdir list ? */;

         }
       }
       liste->flags[idx]|= 64;
     }
   }
 }

 ret= File_is_changed(liste->adressen[idx],
                      (int) liste->flags[idx],
                      (int) liste->typen[idx],
                      liste->groessen[idx],
                      lcrc,lmd5,
                      ldevice,linode,
                      (int) found_list->flags[found_idx],
                      (int) found_list->typen[found_idx],
                      found_list->groessen[found_idx],
                      fcrc,fmd5,
                      fdevice,finode,
                      2|(flag&(64|512|1024|2048|(1<<17)))|inode_test);
 return(ret);
}


int Fileliste_ist_zu_alt(liste,idx,altersgrenze,size,flag)
struct FilelistE *liste;
int idx;
time_t altersgrenze;
double *size;
int flag;
/*
 bit0= also take into respect st_ctime (directory operations)
 bit1= do not record in  liste->groessen
*/ 
{
 struct Cd_backup_planer_struct_staT stbuf;
 time_t ges_datum;

 if(Lstat(liste->adressen[idx],&stbuf,0)==-1)
   return(-1);
 *size= stbuf.st_size;
 if(!(flag&2))
   if(!((liste->flags[idx]&32) && liste->groessen[idx]>=0))
     liste->groessen[idx]= *size;
 ges_datum= stbuf.st_mtime;
 if(flag&1)
   if(stbuf.st_ctime>ges_datum)
     ges_datum= stbuf.st_ctime;
 if(ges_datum<altersgrenze)
   return(1);
 return(0);
}


int Fileliste_lohnender_split(zu_alt,junge_ausnahmen,number_of_files,
                              min_sbpf,flag)
struct FilelistE *zu_alt,*junge_ausnahmen;
int number_of_files;
double min_sbpf;
int flag;
{
 double saved_bytes= 0.0;
 int i;

 if(zu_alt->anz+junge_ausnahmen->anz<=0)
   return(2);
 if(number_of_files<=0)
   return(3);
 for(i=0;i<zu_alt->anz;i++)
   if(zu_alt->groessen[i]>0)
     saved_bytes+= zu_alt->groessen[i];
 for(i=0;i<junge_ausnahmen->anz;i++)
   if(junge_ausnahmen->groessen[i]>0)
     saved_bytes-= junge_ausnahmen->groessen[i];
 if(saved_bytes/((double) number_of_files)<min_sbpf) {
   /* exclusion would not be rewarding enough */
   Fileliste_truncate(zu_alt,0,0);
   Fileliste_truncate(junge_ausnahmen,0,0);
   return(0);
 }
 return(1);
}


int Fileliste_precheck_sub(liste,idx,old_tree,
                           namen_krank,kein_zugriff,
                           mount_punkte,parent_dev,
                           altersgrenze,zu_alt,junge_ausnahmen,
                           max_ausnahmen,min_sbpf,
                           exclude_size,too_big,excluded_for_name,
                           linecount,flag)
struct FilelistE *liste;
int idx;
struct FilelistE *old_tree;
struct FilelistE *namen_krank,*kein_zugriff,*mount_punkte;
dev_t parent_dev;
time_t altersgrenze;
struct FilelistE *zu_alt,*junge_ausnahmen;
int max_ausnahmen;
double min_sbpf;
double exclude_size;
struct FilelistE *too_big;
/* 0520 */
struct FilelistE *excluded_for_name;
int *linecount;
int flag;
/*
 bit0= recursion
 bit1= consider files with st_dev != parent_dev to be unreadable
 bit2= don't check for real readability
 bit3= check for age
 bit4= consider remote (e.g. NFS) mountpoints to be unreadable 
 bit5= with bit3 and bit8 : mistrust directory timestamps and crcs resp. inodes
 bit6= ignore timestamps
 bit7= with bit3 : parent directory has old timestamp
 bit8= with bit3 : check for existence in old_tree
 bit9= with bit8 : check type in old_tree
 bit10= with bit8 : check size in old_tree
 bit11= with bit8 : check crc in old_tree
 bit12= put files larger than exclude_size into too_big 
 bit13= with bit12 : eventually use size of link target 
 bit14= ignore ctime
 bit15= do not evaluate economy of directory splits
 bit16= allow usage of inode information as replacement of crc check
 bit17= with bit16: allow to ignore device number
 bit18= with bit3 : dir_trust : allow bit7 to be set without parent bit7
 bit19-bit21: Backup_plaN.newline_adr_rule
return:
 <0: error, abort run
 bit0= ill name
 bit1= not readable
 bit2= mountpoint of unwanted type
 bit3= too old
 bit4= contains some old items
 bit5= too big
 bit6= item was added to zu_alt list
 bit7= item was excluded
*/
{
 char *name,*npt,link_target[4096],line[4096+100];
 char subadr[2*4096],*subname,target_adr[2*4096],*targetname,cwd[4096];
 int i,swarwas= 0,ret= 0,fret,jung_impl,l_max_ausnahmen= 0,jung_merk,lret;
 int big_file= 0,crc_check= 0,force_show= 0,hflag,read_link_target= 0, zero= 0;
 int consider_ctime= 1,is_empty_dir,is_impl_young,name_l,self_in_l_zu_alt= 0;
 int flag_bit7,newline_adr_rule;
 double size,list_weight;
 FILE *fp;
 struct FilelistE *sub= NULL,*l_zu_alt= NULL,*l_junge_ausnahmen= NULL;
 struct FilelistE *provisory_young= NULL;
 struct DirseQ *dirseq= NULL;

 cwd[0]= 0;

 consider_ctime= !(flag&16384);
 newline_adr_rule= (flag>>19)&7;

 /* test name */
 name= liste->adressen[idx];
 if(name==NULL)
   return(0);
 name_l= strlen(name);

 /* recorded in exclusion list ? */
 if(Fileliste_search_file(liste->sperrliste,name,32)>=0)
   return(0);

 if(flag&1) {
   /* obtain properties here, not by Fileliste_dir() one recursion higher */
   ret= Fileliste_read_properties_idx(liste,idx,&zero,2|(1<<13));
   if(ret<=0)
     return(ret);
 }
 
 /* entertain the user */
 big_file= (liste->typen[idx]==1 && liste->groessen[idx]>5*1024*1024);
 if((flag&(8|256|2048))==(8|256|2048))
   if((flag&32)||!(flag&128))
     crc_check= 1;
 if((*linecount%10)==0 || (*linecount)==2 || crc_check || big_file) {
   if(big_file) 
     if(crc_check) 
       force_show= 1;
   if(crc_check)
     ret= 0;
   sprintf(line,"%s: %s",(crc_check?(flag&(1<<16)?"ino check":"crc check"):
                                                  "checking "),
                         liste->adressen[idx]);
   Print_pacifier(*linecount,line,force_show|2|4|16|32);
 }
 (*linecount)++;

 if(newline_adr_rule<=1 && strchr(name,'\n')!=NULL) {
   if(newline_adr_rule==0) {
     Fileliste__report_newline_adr(name,1);
     ret= -1; goto ex;
   } else if(newline_adr_rule==4) {
     /* tolerate newlines */;
   } else {
     Fileliste_copy_item(liste,idx,excluded_for_name,excluded_for_name->anz,0);
     swarwas|= 128;
   }
 }

 /* check for strange characters which confuse old mkisofs */
 npt= strrchr(name,'/');
 if(npt==NULL) npt= name;
 else          npt++;
 for(i=0;i<strlen(liste->impl_igno);i++)
   if(strchr(npt,liste->impl_igno[i])!=NULL)
     break;
 if(i<strlen(liste->impl_igno)){
   Fileliste_copy_item(liste,idx,namen_krank,namen_krank->anz,0);
   swarwas|= 1;
 }

 if(Is_forbidden_mountpoint(name,parent_dev,NULL,
                            ((flag&2)>>1)|((flag&16)>>3))) {
   Fileliste_copy_item(liste,idx,mount_punkte,mount_punkte->anz,0);
   swarwas|= 4;
 }

 /* check for oversized files */
 if(flag&(1<<12) && liste->typen[idx]!=2) {
   read_link_target= 0;
   if(flag&(1<<13))
     read_link_target= 4;
   Fileliste_type_and_size(liste,idx,&size,1|read_link_target,0);
   if(size>=exclude_size) {
     Fileliste_copy_item(liste,idx,too_big,too_big->anz,0);
     ret= swarwas|32; goto ex;
   }
 }

 /* Test readability */ /* A00728_0054_dave_erickson */
 if((liste->typen[idx]>=1 && liste->typen[idx]<=3) && !(flag&4)){
   if(liste->typen[idx]==3){
     if(readlink(name,link_target,sizeof(link_target)-1)==-1) 
       goto not_readable;
   } else if(liste->typen[idx]==2) {
     DIR *dirpt;
     dirpt= opendir(name);
     if(dirpt==NULL)
       goto not_readable;
     closedir(dirpt);
   } else {
     if((fp= fopen(name,"rb"))==NULL){
not_readable:;
       Fileliste_copy_item(liste,idx,kein_zugriff,kein_zugriff->anz,0);
       swarwas|= 2;
     }else
       fclose(fp);
   }
 }

 if(flag&8){  
   /* age test enabled */

   if(Cdbackup_planer_agetest_debuG) 
     fprintf(stderr,"agetest_debug: %d : %s\n",(int) liste->typen[idx],name);

   /* if not a directory : perform age test */
   if(liste->typen[idx]!=2 && liste->typen[idx]>=0) {
     if((flag&64) || 
        Fileliste_ist_zu_alt(liste,idx,altersgrenze,&size,consider_ctime)>0) {
       if(Cdbackup_planer_agetest_debuG) 
          fprintf(stderr,"agetest_debug: timestamps : old\n");
       if((flag&128) && !(flag&32)) {
         /* parent directory is trustfully old. No moves possible */
         if(Cdbackup_planer_agetest_debuG) 
            fprintf(stderr,"agetest_debug: parent directory trustfully old\n");
         goto really_old;
       }
       hflag= flag&(256|512|1024|2048|(1<<17));
       if(flag&(1<<16))
         hflag|= (7<<13); /* inode check rather than crc check if possible */
       ret= Fileliste_detect_change(liste,idx,old_tree,-1,size,hflag|64);
       if(ret>0) {
         if(Cdbackup_planer_agetest_debuG) {
           if(ret==1)
             fprintf(stderr,"agetest_debug: content change detected\n");
           else
             fprintf(stderr,"agetest_debug: not found in old directory tree\n");
         }
         goto young_enough;
       }

really_old:;
       liste->flags[idx]|= 4;
       /* if large enough or top level : exclude */
       if(liste->groessen[idx]>2.0*min_sbpf || (!(flag&1)) || (flag&(1<<15))) {
         Fileliste_copy_item(liste,idx,zu_alt,zu_alt->anz,0);
         swarwas|= 64;
         if(Cdbackup_planer_agetest_debuG) 
            fprintf(stderr,"agetest_debug: DECISION: exclude\n");
       } else {
         if(Cdbackup_planer_agetest_debuG) 
            fprintf(stderr,"agetest_debug: DECISION: too small to exclude\n");
       }
       swarwas|= 8;
     } else {
       if(Cdbackup_planer_agetest_debuG) 
          fprintf(stderr,"agetest_debug: timestamps : young\n");
young_enough:;
       /* young enough or not verified in old tree */
       liste->flags[idx]|= 2;
       if(Cdbackup_planer_agetest_debuG) 
          fprintf(stderr,"agetest_debug: DECISION: accept\n");
     }
   }
 } else { /* (flag&8) */ 
   if(Cdbackup_planer_agetest_debuG) 
     fprintf(stderr,"agetest_debug: NOT TESTED: %d : %s\n",
                   (int) liste->typen[idx],name);
   liste->flags[idx]|= 2; /* never tested for age. thus young enough */
 }

 /* if designated to be skipped or if not a directory : check is done */
 if(liste->typen[idx]!=2 || (swarwas&(2|4|128)))
   {ret= swarwas; goto ex;}

 if(flag&8) {
   if(Cdbackup_planer_agetest_debuG) 
     fprintf(stderr,"agetest_debug: checking directory with old tree\n");
   /* check wether in old_tree */;
   hflag= flag&(256|512);
   /* currently directory content changes are not reflected by the content
      lists of higher levels. Therefore the directory checksums in old_tree
      may be in the state of level 0. If not absolutely necessary, one should
      not take a mismatch of size or checksum as a sign for recent changes.
      A directory update report will be necessary in the future.
   */
   if(flag&64) {
     hflag|= (1024|2048);
     if(flag&(1<<16))
       hflag|= (7<<13); /* inode check rather than crc check if possible */
   }
   ret= Fileliste_detect_change(liste,idx,old_tree,-1,size,hflag|64);
   if(ret==2) {
     if(Cdbackup_planer_agetest_debuG) 
       fprintf(stderr,"agetest_debug: not found in old tree\n");
     /* not in old_tree : entire subtree is young enough */;
     liste->flags[idx]|= 2;
     /* disable agetest for subsequent directories */
     flag&= ~8;
     /* need to continue for other kinds of tests */
   }
   flag_bit7= (flag&128)|((flag&(1<<18))>>11);
   flag&= ~128;
   if(ret==0 && !(flag&32)) {
     /* no changes detected:
        check wether directory is too old to contain moves
     */
     if((flag&64) || 
        Fileliste_ist_zu_alt(liste,idx,altersgrenze,&size,consider_ctime|2)>0){
       if(Cdbackup_planer_agetest_debuG) 
         fprintf(stderr,"agetest_debug: directory is old and unchanged\n");
       flag|= (128&flag_bit7); /* only if already parent was trustfully old */
     }
   }
 }
 ret= Dirseq_new(&dirseq,name,0);
 if(ret<0)
   goto ex;
 if(ret==0) {
   /* unreadable directory */
   ret= swarwas|2;
   goto ex;
 }
 hflag= 256|512|1024;
 ret= Fileliste_single(&sub,"dummy_precheck",0,0.0,hflag);
 if(ret<=0)
   {ret= -1; goto ex;}
 Fileliste_inherit_target(liste,idx,sub,0);
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);
 Fileliste_set_impl_igno(sub,liste->impl_igno,0);
 if(Cd_backup_planer_trust_fret64)
   hflag= 2;
 else
   hflag= 1|256|1024|(1<<11)|(1<<13);
 if(Fileliste_new(&l_zu_alt,0,liste->main_dir,hflag)<=0)
   {ret= -1; goto ex;}
 if(Fileliste_new(&l_junge_ausnahmen,0,liste->main_dir,2)<=0) 
   {ret= -1; goto ex;}
 if(Fileliste_new(&provisory_young,0,liste->main_dir,2)<=0) 
   {ret= -1; goto ex;}

 /* test all files */
 swarwas= 0;
 jung_impl= 0;

 if(Cdbackup_planer_agetest_debuG) 
   fprintf(stderr,"agetest_debug: === checking directory content : %s\n",name);

 is_empty_dir= 1;
 list_weight= 0.0;
 strcpy(subadr,name);
 if(name[name_l-1]!='/')
   strcat(subadr,"/");
 subname= subadr+strlen(subadr);
 ret= Fileliste_target(liste,idx,target_adr,2);
 if(ret<=0)
   {ret= -1; goto ex;}
 targetname= target_adr+strlen(target_adr);
 while(1) {
   ret= Dirseq_next_adr(dirseq,subname,1);
   if(ret<0)
 continue;
   if(ret==0)
 break;
   if(subname[0]=='.' && ((subname[1]=='.' && subname[2]==0) || subname[1]==0))
 continue;                                              /* skip "." and ".." */
   is_empty_dir= 0;
   ret= Fileliste_set_address(sub,0,subadr,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   strcpy(targetname,subname);
   ret= Fileliste_set_target(sub,0,target_adr,0);
   if(ret<=0)
     {ret= -1; goto ex;}
   sub->flags[0]&= ~(2|4|8|32|64);

   l_max_ausnahmen= max_ausnahmen-junge_ausnahmen->anz;
   jung_merk= l_junge_ausnahmen->anz;
   fret= Fileliste_precheck_sub(sub,0,old_tree,
                                namen_krank,kein_zugriff,
                                mount_punkte,parent_dev,
                                altersgrenze,l_zu_alt,l_junge_ausnahmen,
                                l_max_ausnahmen,min_sbpf,
                                exclude_size,too_big,excluded_for_name,
                                linecount,1|flag);
   if(fret<0)
     {ret= -1; goto ex;}
   is_impl_young= 0;
   if(!(fret&8)) {
     if(l_junge_ausnahmen->anz==jung_merk) {
       jung_impl++; /* objects wich are in none of both lists */
       is_impl_young= 1;
     }
   }

   if(flag&8) {
     if(sub->typen[0]>=0)
       if(sub->groessen[0]>0)
         list_weight+= sub->groessen[0];

     /* <<< this fret64 vodoo is just a patch on the bug that a directory
            may enter zu_alt without having bit3 of return set */

     if(((sub->flags[0]&2)||is_impl_young)
          && !(Cd_backup_planer_trust_fret64&&(fret&64))) {
       ret= Fileliste_set_address(sub,0,subname,0); /* memorize relative adr */
       if(ret<=0)
         {ret= -1; goto ex;}
       ret= Fileliste_copy_item(sub,0,provisory_young,provisory_young->anz,0);
       if(ret<=0)
         {ret= -1; goto ex;}
       if(fret&64) {
         /* <<< inofficial flagbit 7 */
         provisory_young->flags[provisory_young->anz-1]|= 128;
       }

     }
   }
   swarwas|= fret&~(64);                         /* bit6 does not propagate */
 }
 if(is_empty_dir) {
   ret= 0;
   if((flag&8)&&(flag&(1<<8))&&(flag&(1<<9)))  /* it was checked in old_tree */
     ret= swarwas|8;                                     /* so it is too old */
   goto ex;
 }
 ret= swarwas;
 if(Cdbackup_planer_agetest_debuG) 
   fprintf(stderr,"agetest_debug: === directory content checked : %s\n",name);

 if(flag&8){ /* age test */
   liste->groessen[idx]= list_weight;
   if(!(ret&8)) {                          /* if everything was young enough */
     /* count this directory as "young" */;
     liste->flags[idx]|= 2;
     if((ret&16)) {        /* if old files have been accepted for efficiency */
       liste->flags[idx]|= 4;
       if(Cdbackup_planer_agetest_debuG) 
         fprintf(stderr,"agetest_debug: accepting old files for efficiency\n");
     }
     if(Cdbackup_planer_agetest_debuG) 
       fprintf(stderr,"agetest_debug: DECISION: accept whole directory\n");
   } else if (jung_impl+l_junge_ausnahmen->anz<=0) {
                                               /* if everything was too old  */
     liste->flags[idx]|= 4;
     Fileliste_truncate(l_zu_alt,0,0);   /* they will be replaced as a whole */
     Fileliste_copy_item(liste,idx,l_zu_alt,l_zu_alt->anz,0);     /* by this */
     self_in_l_zu_alt= 1;
     if(Cdbackup_planer_agetest_debuG) 
       fprintf(stderr,"agetest_debug: DECISION: exclude whole directory\n");
   } else if(l_junge_ausnahmen->anz+jung_impl<l_max_ausnahmen) {
                              /* if there is still room for local youngsters */

     /* count this directory as "old" */;
     liste->flags[idx]|= 6;
     if(Cdbackup_planer_agetest_debuG) 
       fprintf(stderr,"agetest_debug: DECISION: exclude directory\n");
     /* append youngsters from  sub  to  l_junge_ausnahmen */
     if(jung_impl>0) {

       /* >>> ??? as an alternative to a memorized list :
              rescan for young items */
       swarwas= ret;
       for(i=0;i<provisory_young->anz;i++) {
         if(provisory_young->flags[i]&2) {
           /* re-expand to full name */;
           if(name[name_l-1]=='/')
             sprintf(subadr,"%s%s",name,provisory_young->adressen[i]);
           else
             sprintf(subadr,"%s/%s",name,provisory_young->adressen[i]);

           /* Without search, too fat things had been included.
              Cd_backup_planer_trust_fret64 says that such fatsos are now
              already excluded from provisory_young  */
           if((!Cd_backup_planer_trust_fret64)
               && Fileliste_search_file(l_zu_alt,subadr,0)>=0) {

             if(!(provisory_young->flags[i]&128))
               Cd_backup_planer_false_young++;
       
           } else {

             if((provisory_young->flags[i]&128))
               Cd_backup_planer_false_zu_alt++;

             if(Cdbackup_planer_agetest_debuG) 
               fprintf(stderr,"agetest_debug: DECISION: accept exempt : %s\n",
                              subadr);
             strcpy(targetname,provisory_young->adressen[i]);
             ret= Fileliste_set_address(provisory_young,i,subadr,0);
             if(ret<=0)
               {ret= -1; goto ex;}
             ret= Fileliste_set_target(provisory_young,i,target_adr,0);
             if(ret<=0)
               {ret= -1; goto ex;}
             ret= Fileliste_copy_item(provisory_young,i,
                                  l_junge_ausnahmen,l_junge_ausnahmen->anz,0);
             if(ret<=0)
               {ret= -1; goto ex;}
           }
           Fileliste_set_address(provisory_young,i,NULL,0);
           Fileliste_set_target(provisory_young,i,NULL,0);
         }
       }
       ret= swarwas;
     }
     Fileliste_truncate(l_zu_alt,0,0); /* oldies will be replaced as a whole */
     Fileliste_copy_item(liste,idx,l_zu_alt,l_zu_alt->anz,0);     /* by this */
     self_in_l_zu_alt= 1;
     jung_impl= 0;
   } else { 
     if(Cdbackup_planer_agetest_debuG) 
       fprintf(stderr,
       "agetest_debug: DECISION: accept whole directory despite old files\n");
     liste->flags[idx]|= 6;
     /* count this directory as "young" , mark unclean */;
     ret&= ~8;
     ret|= 16;
   }

   if(!(flag&(1<<15))) {
     /* evaluate effectiveness and eventually prevent split */
     lret= Fileliste_lohnender_split(l_zu_alt,l_junge_ausnahmen,
                                     l_junge_ausnahmen->anz+jung_impl,
                                     min_sbpf,0);
     if(lret==0){
       if(Cdbackup_planer_agetest_debuG) 
         fprintf(stderr,"agetest_debug: OVERRIDE DECISION: uneconomic split, accept whole directory\n");
       ret&= ~8;
       ret|= 16;
     }
   }

   if(ret&8) { /* if directory was classified as too old */
     /* append to the "old" list l_zu_alt 
        and to the "young" list l_junge_ausnahmen
     */
     if(Cdbackup_planer_agetest_debuG) 
       fprintf(stderr,
              "agetest_debug: === lists:  old= %d+%d   young_exempts= %d+%d\n",
              zu_alt->anz,l_zu_alt->anz,
              junge_ausnahmen->anz,l_junge_ausnahmen->anz);
     for(i=0;i<l_zu_alt->anz;i++)
       Fileliste_copy_item(l_zu_alt,i,zu_alt,zu_alt->anz,0);
     if(self_in_l_zu_alt)
       ret|= 64;
     for(i=0;i<l_junge_ausnahmen->anz;i++) 
       Fileliste_copy_item(l_junge_ausnahmen,i,
                           junge_ausnahmen,junge_ausnahmen->anz,0);
   }

 }

ex:;
 Fileliste_destroy(&provisory_young,0);
 Fileliste_destroy(&l_junge_ausnahmen,0);
 Fileliste_destroy(&l_zu_alt,0);
 Fileliste_destroy(&sub,0);
 Dirseq_destroy(&dirseq,0);
 if(cwd[0]!=0)
   chdir(cwd);
 return(ret);
}


int Fileliste_precheck(liste,old_tree,
                       namen_krank,kein_zugriff,mount_punkte,
                       altersgrenze,zu_alt,junge_ausnahmen,
                       max_ausnahmen,min_sbpf,
                       exclude_size,too_big,excluded_for_name,
                       flag)
struct FilelistE *liste;
struct FilelistE *old_tree;
struct FilelistE *namen_krank,*kein_zugriff,*mount_punkte;
time_t altersgrenze;
struct FilelistE *zu_alt,*junge_ausnahmen;
int max_ausnahmen;
double min_sbpf;
double exclude_size;
struct FilelistE *too_big;
/* 0520 */
struct FilelistE *excluded_for_name;
int flag;
/*
 bit1= consider files with st_dev != parent_dev to be unreadable
 bit2= don't check for real readability
 bit3= check for age
 bit4= consider remote (e.g. NFS) mountpoints to be unreadable 
 bit5= with bit8 : mistrust directory timestamps 
 bit6= ignore timestamps

 bit8= check for existence in old_tree
 bit9= check type in old_tree
 bit10= check size in old_tree
 bit11= check crc in old_tree
 bit12= consider files larger than exclude_size unreadable 
 bit13= with bit12 : eventually use size of link target 
 bit14= ignore ctime
 bit15= do not evaluate economy of directory splits
 bit16= allow device,inode check as replacement of crc check
 bit17= with bit16: allow to ignore device number
 bit18= with bit3 : dir_trust : allow parent of parent to be new
 bit19-bit21: Backup_plaN.newline_adr_rule
*/
{
 int ret= 0,i,subflag,trustfully_old,hflag;
 dev_t parent_dev= 0;
 struct stat stbuf;
 int linecount= 0;

 for(i=0;i<liste->anz;i++) {
   subflag= flag&(4|8|32|64|256|512|1024|2048|4096|8192|16384|
                  (1<<15)|(1<<16)|(1<<17)|(1<<18)|(7<<19));
   if(flag&(2|16))
     if(Lstat(liste->adressen[i],&stbuf,0)!=-1) {
       parent_dev= stbuf.st_dev;
       subflag|= (flag&(2|16));
     }
                  /* agecheck shortcut is only worth if CRC check is enabled */
   if((flag&8) && (flag&2048)) {
     hflag= flag&(1<<14);
     if(flag&(1<<18))
       hflag|= 2;
     trustfully_old= Parent_is_trustfully_old(liste->adressen[i],altersgrenze,
                                              hflag);
     if(trustfully_old)
       subflag|=128;
   }
   ret= Fileliste_precheck_sub(liste,i,old_tree,
                               namen_krank,kein_zugriff,
                               mount_punkte,parent_dev,
                               altersgrenze,zu_alt,junge_ausnahmen,
                               max_ausnahmen,min_sbpf,
                               exclude_size,too_big,excluded_for_name,
                               &linecount,subflag);
   if(ret<0)
     goto ex;
 }
 Print_pacifier(linecount,"file objects checked",1|2|4|8|16);
 ret= 1;
ex:;
 fprintf(stderr,"\n");
 return(ret);
}

#endif /* ! Crawler_ruleS */


int Fileliste_count_dir_files(liste,dir,flag)
struct FilelistE *liste;                             /* does not get altered */
char *dir;
int flag;
{
 struct FilelistE *sub;
 int count= -1,zero= 0;

 if(Fileliste_new(&sub,0,liste->main_dir,2)<=0)
   return(-1);
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);
 if(Fileliste_dir(sub,dir,&zero,4)<=0) {
   Fileliste_destroy(&sub,0);
   return(-1);
 }
 count= sub->anz;
 Fileliste_destroy(&sub,0);
 return(count);
}


int Fileliste_attach_sub(liste,idx,flag)
struct FilelistE *liste;
int idx;
int flag;
{
 char sub_target[4096];
 int ret,hflag;
 struct FilelistE *sub= NULL;

 hflag= Fileliste_creation_flag(liste,0);
 ret= Fileliste_new(&sub,0,liste->main_dir,hflag);
 if(ret<=0) {
cannot_create_new_list:;
   fprintf(stderr,"--- cannot create new file list\n");
   goto ex;
 }
 Fileliste_set_subdir(liste,idx,sub,0);
 Fileliste_set_sperrliste(sub,liste->sperrliste,0);

 Fileliste_target(liste,idx,sub_target,2);
 if(Fileliste_set_target_dir(sub,sub_target,0)<=0)
   {ret= -1; goto cannot_create_new_list;}
 ret= 1;
ex:;
 if(ret<=0)
   Fileliste_destroy(&sub,2|4);
 return(ret);
}


int Fileliste_build_branch(liste,new_current,line_in,target,
                           typ,size,crc,md5,
                           device,inode,
                           xprop,flag)
struct FilelistE *liste;
struct FilelistE **new_current;
char *line_in;
char *target;                    /* use only if neither NULL nor empty text */
int typ;
double size;
unsigned long crc;
unsigned char md5[16];
dev_t device;
ino_t inode;
struct Filelist_xprop_reC *xprop;
int flag;
/*
 bit0= handle as directory address not as file object
       set *new_current to last resulting list
 bit1= with bit4==0: do not check for existence in list but simply append
 bit2= with bit4==1: make sure that eventual all doubles are deleted
 bit4= delete depicted object rather than create it
       (do not delete if bit0 is set)
 bit8= line valid
 bit9= typ valid
bit10= size valid
bit11= crc valid
bit12= crc is MD5 rather than CRC32
bit13= inode is valid, with directory: device is valid too
bit14= device is valid even with non-directory
bit18= this is level 0
*/
{
 static int programmer_notified= 0;
 int ret,idx,nodecount= 0,i, diff, do_not_count;
 char *cpt,*npt= NULL,line[4096];
 struct FilelistE *current,*sub;

 if(Cdbackup_planer_treebuild_debuG)
   fprintf(stderr,"treebuild_debug:   %s   : flag= %X\n",line_in,flag);

 if(liste->subdirs==NULL) {
   if(programmer_notified<10 && !(flag&2)) {
     fprintf(stderr,
     "??? programmer notification : missing subdirs vector on build_branch\n");
     programmer_notified++;
   }
   return(0);
 }

 strcpy(line,line_in);
 current= liste;

 npt= cpt= line;

 while(npt!=NULL) {
   if(nodecount)
     cpt= npt+1;
   nodecount++;
   idx= -1;
   npt= strchr(cpt,'/');
   if(npt!=NULL)
     *npt= 0;
   if(cpt[0]==0)
 continue;
   if(cpt[0]=='.'&&cpt[1]==0)
 continue;

   if(Cdbackup_planer_treebuild_debuG)
     fprintf(stderr,"treebuild_debug: current= %lX (%d items): cpt= %s\n",
                    (unsigned long) current,current->anz,cpt);

   if((flag&2) && !((flag&16)||npt!=NULL)) {

     if(Cdbackup_planer_treebuild_debuG)
       fprintf(stderr,"treebuild_debug: appending leaf without double check\n");

     idx= -1;
   } else {
     idx= Fileliste_search_file(current,cpt,
                                (Cdbackup_planer_treebuild_debuG>1)<<4);
                                /* eventually allow search_debuG tracing */
   }
   if(idx>=0) {

     if(Cdbackup_planer_treebuild_debuG)
       fprintf(stderr,"treebuild_debug: found in tree\n");

     if(flag&16) {
       if(npt!=NULL && current->subdirs[idx]==NULL) {
         fprintf(stderr,"--- while trying to delete tree entry '%s' :\n",
                        line_in);
         fprintf(stderr,"    branch node #%d ('%s') is not a directory node\n",
                        idx,cpt);
         fprintf(stderr,"    (better make a level 0 backup next)\n");
         ret= 0; goto ex;
       }
     } else if(current->subdirs[idx]==NULL) { 
       if(npt!=NULL || (flag&1) || typ==2) {

         if(Cdbackup_planer_treebuild_debuG)
           fprintf(stderr,
              "treebuild_debug: non-directory transformed into a directory\n");

         ret= Fileliste_attach_sub(current,idx,0);
         if(ret<=0)
           goto ex;
       }
     } else {
       if(npt==NULL && !(flag&1) && typ!=2) {
         if(Cdbackup_planer_treebuild_debuG)
           fprintf(stderr,
              "treebuild_debug: directory transformed into a non-directory\n");
         Fileliste_destroy(&(current->subdirs[idx]),2|4);
       }
     }
     if(npt!=NULL) {
       current= current->subdirs[idx];
 continue;

     } else {

       if(Cdbackup_planer_treebuild_debuG)
         fprintf(stderr,"treebuild_debug: re-recording as type %d : %s\n",
                        typ,current->adressen[idx]);

       current->flags[idx]&= ~(2|4|8|16|64);
       if(flag&512)
         current->typen[idx]= typ; 
       if(flag&1024)
         current->groessen[idx]= size; 
       if(flag&2048) {
         ret= Fileliste_set_crc(current,idx,crc,md5,4|(flag&(3<<11)));
         if(ret<=0)
           goto ex;
       }
       if(flag&(1<<13)) {
         Fileliste_set_inode(current,idx,inode,0);
         if(flag&(1<<14)) {
           if(!(current->dir_dev_ino_valid&1)) {
             current->dir_device= device;
             current->dir_dev_ino_valid|= 1;
           } else {
             if(device!=current->dir_device) {
               /* >>> Oops ? contradicting device numbers */;
             }
           }
         }
       }
       if(xprop!=NULL) {
         ret= Fileliste_set_xprop(current,idx,xprop->mod,xprop->own,xprop->grp,
                                  xprop->date,0);
         if(ret<=0)
           goto ex;
       }
       if(flag&1) {
         if(current->subdirs[idx]==NULL) {
            /* non-directory transformed into a directory inbetween levels */
            ret= Fileliste_attach_sub(current,idx,0);
            if(ret<=0)
              goto ex;
         }
         current= current->subdirs[idx];
         if(flag&(1<<13)) {
           current->dir_device= device;
           current->dir_inode= inode;
           current->dir_dev_ino_valid|= 3;
         }
       }
     }
   } else {

     if(Cdbackup_planer_treebuild_debuG)
       fprintf(stderr,"treebuild_debug: not found in list\n");

     if((flag&16) && !(flag&1))
       {ret= 2; goto ex;}
     ret= Fileliste_append(current,cpt,1|8);
     if(ret<=0) {
cannot_append:;
       fprintf(stderr,"--- cannot append to file list\n");
       goto ex;
     }
     do_not_count= (ret!=1);

     if(Cdbackup_planer_treebuild_debuG) {
       int hash_idx;
       hash_idx= Fileliste__compute_hashvalue(cpt,0);
       fprintf(stderr,"treebuild_debug: new list item : %lX,%d,%d : %s\n",
                      (unsigned long) current,current->anz-1,hash_idx,cpt);
     }

     if(npt==NULL) {

       if(Cdbackup_planer_treebuild_debuG)
         fprintf(stderr,"treebuild_debug: writing properties of type %d\n",typ);

       if(target!=NULL) if(*target!=0) {
         ret= Fileliste_set_target(current,current->anz-1,target,1);
         if(ret<=0)
           goto cannot_append;
       }

       if(flag&512)
         current->typen[current->anz-1]= typ; 
       if(flag&1024)
         current->groessen[current->anz-1]= size; 
       if(flag&2048)
         Fileliste_set_crc(current,current->anz-1,crc,md5,4|(flag&(3<<11)));
       if(flag&(1<<13))
         Fileliste_set_inode(current,current->anz-1,inode,0);
       if(xprop!=NULL) {
         ret= Fileliste_set_xprop(current,current->anz-1,
                                  xprop->mod,xprop->own,xprop->grp,
                                  xprop->date,0);
         if(ret<=0)
           goto ex;
       }
     } else {
                            /* nothing is known about this directory or link */

       if(Cdbackup_planer_treebuild_debuG)
         fprintf(stderr,
                 "treebuild_debug: branch node was not recorded before\n");

       current->typen[current->anz-1]= 0;
     }

     /* if dir : create new sub list */
     if(npt!=NULL || (flag&1)) {
       ret= Fileliste_attach_sub(current,current->anz-1,0);
       if(ret<=0)
         goto ex;

       if(Cdbackup_planer_treebuild_debuG)
         fprintf(stderr,
            "treebuild_debug: new branch node : %lX : %s : type= %d\n",
            (unsigned long) current->subdirs[current->anz-1],
            current->adressen[current->anz-1],
            current->typen[current->anz-1]);

       current= current->subdirs[current->anz-1];
       if(npt==NULL) { /* tip of branch reached */
         if(flag&(1<<13)) {
           current->dir_device= device;
           current->dir_inode= inode;
           current->dir_dev_ino_valid|= 3;
         }
       }
     } else if(npt==NULL && (flag&(1<<14))) { /* device given with non-dir */
       if(current->dir_dev_ino_valid&1) { /* boss already has device number */
         if(current->dir_device!=device) {

           /* >>> Uh oh. A device jump on non-directory. Abort ? */;

           if(Cdbackup_planer_treebuild_debuG)
             fprintf(stderr,
               "treebuild_debug: detected device number mismatch with boss\n");
         }
       } else {
         current->dir_device= device;
         current->dir_dev_ino_valid|= 1;
         if(Cdbackup_planer_treebuild_debuG)
           fprintf(stderr,
          "treebuild_debug: set previously undefined device number of boss\n");
       }
     }
   }
 }

 if((flag&16) && idx>=0 && !(flag&1)) {
   while(1) {

     if(Cdbackup_planer_treebuild_debuG)
       fprintf(stderr,"treebuild_debug: deleting node %d : %s\n",
                      idx,current->adressen[idx]);

     if(current->subdirs[idx]!=NULL) {
       /* directory node : do only remove if empty */;
       sub= current->subdirs[idx];
       for(i=0;i<sub->anz;i++)
         if(sub->adressen[i]!=NULL)
       break;
       if(i<sub->anz) {

         /* 0520 */
         if(! Cdbackup_planer_tolerate_bad_rmdiR) {

           Cd_backup_planer_not_empty_on_delete++;
           fprintf(stderr,
                   "--- branch node subdirectory not empty on delete\n");
           ret= 0; goto ex;
         }
       }
     }

     Fileliste_destroy_item(current,idx,0);

     if(!(flag&4)) 
   break;
     /* find and delete eventual further entries of the same name */;
     idx= Fileliste_search_file(current,cpt,0);
     if(idx>=0) {

       if(Cdbackup_planer_treebuild_debuG)
         fprintf(stderr,
                 "treebuild_debug: detected another node with same address\n");

       Cd_backup_planer_double_deleted++;
     } else
   break;
   } 

 }

 ret= 1;
ex:;
 if(ret>0 && (flag&1))
   *new_current= current;
 return(ret);
}


int Fileliste_search_branch(liste,found_list,found_idx,adr_in,flag)
struct FilelistE *liste;
struct FilelistE **found_list;
int *found_idx;
char *adr_in;
int flag;
/*
 bit0= in case of liste==NULL return 1 rather than 0
 bit4= debug: this is a payload search (not a tree creation search)
*/
{
 static int programmer_notified= 0;
 int ret,idx,nodecount= 0;
 char *cpt,*npt= NULL,line[4096];
 struct FilelistE *current;

 if(Cdbackup_planer_listsearch_debuG) if(flag&16)
   fprintf(stderr,"listsearch_debug: looking for %s\n",adr_in);

 if(liste==NULL) {

   if(Cdbackup_planer_listsearch_debuG) if(flag&16)
     fprintf(stderr,"listsearch_debug: liste==NULL\n");

   return(flag&1);
 }

 if(liste->subdirs==NULL) {
   if(programmer_notified<10 && !(flag&2)) {
     fprintf(stderr,
    "??? programmer notification : missing subdirs vector on search_branch\n");
     programmer_notified++;
   }
   return(0);
 }

 strcpy(line,adr_in);
 current= liste;

 npt= cpt= line;

 while(npt!=NULL) {
   if(nodecount)
     cpt= npt+1;
   nodecount++;
   idx= -1;
   npt= strchr(cpt,'/');
   if(npt!=NULL)
     *npt= 0;
   if(cpt[0]==0)
 continue;
   if(cpt[0]=='.'&&cpt[1]==0)
 continue;

   if(Cdbackup_planer_listsearch_debuG) if(flag&16)
     fprintf(stderr,"listsearch_debug: node '%s' in %lX (%d items)\n",
                    cpt,(unsigned long) current,current->anz);

   idx= Fileliste_search_file(current,cpt,(flag&16));
   if(idx>=0) {
     /* already existing */
     if(current->subdirs[idx]==NULL) { 
       if(npt!=NULL) {

         if(Cdbackup_planer_listsearch_debuG) if(flag&16)
           fprintf(stderr,
         "listsearch_debug: NOT FOUND (found item is not a directory node)\n");

         { ret= 0; goto ex; }
       }
 break;
     }
     if(npt!=NULL) {
       current= current->subdirs[idx];
 continue;
     }
   } else {

     if(Cdbackup_planer_listsearch_debuG) if(flag&16)
       fprintf(stderr,"listsearch_debug: NOT FOUND\n");
     ret= 0; goto ex;

   }
 }

 ret= (idx>=0);
ex:;
 if(ret>0) {
   *found_list= current;
   *found_idx= idx;

    if(Cdbackup_planer_listsearch_debuG) if(flag&16)
      fprintf(stderr,"listsearch_debug: FOUND\n");

 }
 return(ret);
}


int Fileliste_next_in_tree(liste_pt,idx_pt,current_dir,line,typ,size,crc,md5,
                           device,inode,
                           flag)
struct FilelistE **liste_pt;
int *idx_pt;
char current_dir[4096];
char line[4096];
int *typ;
double *size; 
unsigned long *crc;
unsigned char md5[16];
dev_t *device;
ino_t *inode;
int flag;
/*
 bit0= do not advance read position, do not change liste_pt,idx_pt,current_dir
 bit13=compose line from target and source
*/
/*
 return:
  <=0 = error
    bit0= line valid
    bit1= typ valid
    bit2= size valid
    bit3= crc valid
    bit4= crc is MD5 rather than CRC32
    bit5= inode information valid
    bit6= device information valid
    bit13= target has been prepended
*/
{
 int ret= 0,idx,no_crc,is_dir;
 char *cpt;
 struct FilelistE *liste;

 liste= *liste_pt;
 idx= *idx_pt;
 line[0]= 0;
 if(liste==NULL)
   return(0);
 if(liste->boss==NULL) {
   strcpy(line,"/");
   if(!(flag&1))
     current_dir[0]= 0;
 } else
   strcpy(line,"./");

 for(;idx<liste->anz;idx++) {
   if(liste->typen[idx]<0 || liste->adressen[idx]==NULL) {
     if(flag&1)
       {ret= 0; goto ex;}
 continue;
   }
   if(strncmp(liste->adressen[idx],"./",2)==0)
     strcat(line,liste->adressen[idx]+2);
   else
     strcat(line,liste->adressen[idx]);
   /* is it necessary to prepend a target ? */
   if((flag&(1<<13)) && (liste->boss==NULL || liste->targets[idx]!=NULL)) {
     /* prepend target */
     char target[4096];
     char combined_adr[3*4096];

     Fileliste_target(liste,idx,target,1|2);
     Fileliste__format_combined_adr(combined_adr,target,line,0);
     strcpy(line,combined_adr);
     /* give signal that target has been prepended */;
     ret|= (1<<13);
   }
   ret|= 1;
   *typ= liste->typen[idx];
   if(*typ>0)
     ret|= 2;
   *size= liste->groessen[idx];
   if(*typ>0 && *size>=0)                      /* ??? why was :  && *typ!=2 */
     ret|= 4;
   no_crc= 0;
   if(liste->crcs==NULL)
     no_crc= 1;
   else if(liste->crcs[idx]==NULL)
     no_crc= 1;
   if(no_crc) {
     strcpy((char *) md5,"no md5         ");
     *crc= 0;
   } else  if(liste->flags[idx]&16) {
     memcpy(md5,liste->crcs[idx]->md5,16);
     ret|= 16;
   } else 
     *crc= liste->crcs[idx]->crc;
   if((liste->flags[idx]&8)&&!no_crc)
     ret|= 8;
   if(liste->inodes!=NULL && (liste->flags[idx]&64)) {
     *inode= liste->inodes[idx];
     ret|= 32;
   }
   if(liste->dir_dev_ino_valid&1){
     *device= liste->dir_device;
     ret|= 64;
   }
   is_dir= 1;
   if(liste->subdirs==NULL)
     is_dir= 0;
   else if(liste->subdirs[idx]==NULL)
     is_dir= 0;
   if(is_dir) {
     strcat(line,"/");
     if(!(flag&1))
       sprintf(current_dir+strlen(current_dir),"/%s",liste->adressen[idx]);
     if(!(flag&1)) {
       /* enter sub directory */
       *liste_pt= liste->subdirs[idx];
       *idx_pt= 0;
     }
   } else if(!(flag&1))
     *idx_pt= idx+1;
   goto ex;
 }
 if(!(flag&1)) {
   /* list done , go upwards */
   cpt= strrchr(current_dir,'/');
   if(cpt!=NULL)
     *cpt= 0;
   else
     current_dir[0]= 0;
 }
 if(liste->boss==NULL) 
   {ret= 0; goto ex;}
 if(!(flag&1)) {
   *idx_pt= liste->boss_idx+1;
   *liste_pt= liste->boss;
 }
 *typ= 2;
 strcpy(line,"..");
 ret= 3;
ex:;
 return(ret);
}


int Fileliste__escape_target(target,flag)
char *target;
int flag;
{
 int i,offset= 0;
 /* escape '=' and '\\' in target address */
 for(i= strlen(target)-1;i>=0;i--)
   if(target[i]=='=' || target[i]=='\\')
     offset++;
 target[strlen(target)+offset]= 0;
 for(i= strlen(target)-1;i>=0;i--) {
   if(target[i]=='=' || target[i]=='\\') {
     target[i+offset]= target[i];
     offset--;
     target[i+offset]= '\\';
   } else {
     target[i+offset]= target[i];
   }
 }
 return(1);
}


int Fileliste__target_source_limit(line,sep,limit_pt,flag)
char *line;
char sep;
char **limit_pt;
int flag;
{
 char *npt;

 for(npt= line;*npt!=0;npt++) {
   if(*npt=='\\') {
     if(*(npt+1)!=0)
       npt++;
 continue;
   }
   if(*npt=='=')
 break;
 }
 if(*npt==0)
   npt= NULL;
 (*limit_pt)= npt;
 return(npt!=NULL);
}


int Fileliste__unescape_target(line,endpt,sep,flag)
char *line;
char *endpt;
char sep;
int flag;
/* 
 bit0= also unescape newline
*/
{
 char *wpt,*rpt;

 if(endpt==NULL)
   endpt= line+strlen(line);

 /* unescape sep and '\\' and eventually newline */
 wpt= line;
 for(rpt= line; rpt<endpt; rpt++) {
   if(*rpt=='\\') {
     if(*(rpt+1)==sep || *(rpt+1)=='\\' || (*(rpt+1)=='\n' && (flag & 1)))
       rpt++;
   }
   *(wpt++)= *rpt;
 }
 *wpt= 0;
 return(1);
} 


/* 0520 */
int Fileliste__unescape_source(line,flag)
char *line;
int flag;
{
 int l,bsl_count;

 /* unescape at line end '\\' */
 l= strlen(line);
 if(l==0)
   return(1);
 for(bsl_count= 0;bsl_count<l;bsl_count++)
   if(line[l-1-bsl_count]!='\\')
 break;
 if(bsl_count==0)
   return(1);
 if((bsl_count%2)) /* not an escaped line end but escaped backslashes */
    return(1); 
 line[l-(bsl_count/2)]= 0;
 return(1);
} 


/* 0520 : escape newline and trailing backslashes */
int Fileliste__esc_newline(from,to,flag)
char *from,*to;
int flag;
{
 int bsl_count;
 char *rpt,*wpt;

 wpt= to;
 for(rpt= from;*rpt!=0;rpt++) {
   if(*rpt!='\n') {
     *(wpt++)= *rpt;
     if(*(rpt+1)) /* 0580 : escape backslash at name end too */
 continue;
     rpt++; /* 0580 */
   }
   /* escape backslashes directly before newline or name end */
   for(bsl_count= 0;rpt-1-bsl_count>=from;bsl_count++) {
     if(*(rpt-1-bsl_count)!='\\')
   break;
     *(wpt++)= '\\';
   }
   if(*rpt == '\n') {
     *(wpt++)= '\\';
     *(wpt++)= '\n';
   } else
 break; /* 0580 */
 }
 *wpt= 0;
 return(1);
}


int Fileliste__format_combined_adr(combined_adr,target,source,flag)
char combined_adr[3*4096];
char *target,*source;
int flag;
/*
 bit0= target surely contains '=', no need to look for it
 bit1= append '/' to both address parts
*/
{
 int was_escape= 0;
 char *wpt,*rpt;

 if((!(flag&1)) && strchr(target,'=')==NULL && strchr(target,'\\')==NULL &&
     strchr(target,'\n')==NULL && strchr(source,'\n')==NULL ) {
   if(flag&2)
     sprintf(combined_adr,"%s/=%s/",target,source);
   else
     sprintf(combined_adr,"%s=%s",target,source);
 } else {
   wpt= combined_adr;
   for(rpt= target;*rpt!=0;rpt++) {

     /* 0520 : escape newline */
     /* older: escape = and backslash */
     if(*rpt=='=' || *rpt=='\\' || *rpt=='\n') {
       *(wpt++)= '\\';
       was_escape= 1;
     }
     *(wpt++)= *rpt;
   } 
   if(flag&2)
     *(wpt++)= '/';
   *(wpt++)= '=';

/* <<< 0520: old
   strcpy(wpt,source);
*/
   Fileliste__esc_newline(source,wpt,0);

   if(flag&2) {
     wpt+= strlen(wpt);
     *(wpt++)= '/';
     *(wpt++)= 0;
   }
 }
 return(1);
}


int Fileliste__format_serialized_line(serialized_line,
                                      typ,size,crc,md5,target,source,
                                      device,inode,mtime,ctime,
                                      valid_flag,flag)
char serialized_line[200+3*4096];
int typ;
double size;
unsigned long crc;
unsigned char md5[16];
char target[];
char source[];
dev_t  device;
ino_t inode;
time_t mtime;
time_t ctime;
int valid_flag; /* for : int Format_line_prepend() , see there */
int flag;
/*
 bit0= target surely contains '=', no need to look for it
 (see also : int Format_line_prepend() )
 bit9= prepend type
bit10= prepend length
bit11= preprend crc 
bit13= prepend device, inode information (as extension to type)
bit16= prepend mtime
bit17= prepend ctime
*/
{
 char typetext[1024],lengthtext[80],crctext[80];

 if(0 && (!(flag&1)) && strchr(target,'=')==NULL && strchr(target,'\\')==NULL) {
   /* <<< 0520 : deactivated */

   Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                       device,inode,mtime,ctime,
                       valid_flag,flag&((1|2|4|16|128|256)<<9));
   sprintf(serialized_line,"%s%s%s%s=%s",typetext,lengthtext,crctext,
                                         target,source);
 } else {
   char combined_adr[4*4096];

   Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5,
                       device,inode,mtime,ctime,
                       valid_flag,flag&((1|2|4|16|128|256)<<9));
   Fileliste__format_combined_adr(combined_adr,target,source,1);
   sprintf(serialized_line,"%s%s%s%s",
                           typetext,lengthtext,crctext,combined_adr);
 }
 return(1);
}


int Fileliste_read_serialized_line(liste,idx,current_dir,fpin,
                                   line,typ,size,crc,md5,
                                   device,inode,
                                   xprop,flag)
struct FilelistE **liste;
int *idx;
char current_dir[4096];
FILE *fpin;
char *line;
int *typ;
double *size;
unsigned long *crc;
unsigned char md5[16];
dev_t *device;
ino_t *inode;
struct Filelist_xprop_reC **xprop; /* *xprop has to be NULL or freeable */
int flag;
/*
 bit0= print line as read (only when reading from file pointer)
 bit9= prepended is type
bit10= prepended is length
bit11= prepended is crc
bit12= reserved (MD5 bit is automated)
(a fixed field format is recognized by type "D")
bit13= return raw address even if with '='
bit14= if with '=' choose target addresses rather than source addresses
       change to relative addresses with ./ prepended
bit15= if split file : look for part's info rather than for full file info
return:
  <=0 = error
    bit0= line valid
    bit1= typ valid
    bit2= size valid
    bit3= crc valid
    bit4= crc is MD5 (in md5) rather than CRC32 (in *crc)
    bit5= inode information valid
    bit6= device information valid
    bit13= target has been prepended
    bit15= this file is of unwanted type (e.g. File part without flag:bit15)
*/
{
 int ret=0,i,x,is_big_format= 0,valid= 0,hret,codelen,l,bsl_count;
 char *cpt,*npt,buf[3],*date_str= NULL, escaped_adr[4096*2];
 int ino_vec[CdbplivS],dev_vec[CdbplivS];
 dev_t dev;
 ino_t ino;

 line[0]= 0;
 *typ= 0;
 *size= -1;
 *crc= 0;
 if(*liste!=NULL) {
   ret= Fileliste_next_in_tree(liste,idx,current_dir,line,typ,size,crc,md5,
                               device,inode,flag&(1<<13));
 } else {

   /* 0520: concatenate lines ending on single '\' */
   line[0]= 0;
   l= 0;
   while(1) {
     if(Sfile_fgets(line+l,4096-l,fpin)==NULL) 
       {ret= 0; goto ex;}
     if(flag&1)
       printf("%s\n",line);
     l= strlen(line);
     if(l==0)
   break;
     if(line[l-1]!='\\')
   break;
     for(bsl_count= 0;bsl_count<l;bsl_count++)
        if(line[l-1-bsl_count]!='\\')
     break;
     if((bsl_count%2)==0) /* trailing backslashes are escaped */
   break;
     line[l-1]= '\n'; /* found unescaped trailing backslash */
   }

   ret= 1;
   if(line[0]=='#' && line[1]=='=' && line[2]=='#' && line[3]==' ')
     {ret= 1; goto ex;}
   if(line[0]=='.' && line[1]=='.')
     {ret= 1; goto ex;}
   /* eventually split into : type size crc name */
   /* resp : Type permissionbits owner group time lengthtext crctext adr */
   /* where type is : 
         typecode[",="number][",i"inode-number][",d"device-number] */
   cpt= line;
   if(flag&512) {
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     ret|= 2;
     if(isalpha(*cpt))
       if(toupper(*cpt)==*cpt)
         is_big_format= 1;
     if(*cpt=='0')
       *typ= 0;
     if(*cpt=='1' || *cpt=='f' || *cpt=='F' || *cpt=='-') 
       *typ= 1;
     else if(*cpt=='2' || *cpt=='d' || *cpt=='D')
       *typ= 2;
     else if(*cpt=='3' || *cpt=='l' || *cpt=='L')
       *typ= 3;
     else if(*cpt=='4' || *cpt=='p' || *cpt=='P')
       *typ= 4;
     else if(*cpt=='5' || *cpt=='m' || *cpt=='M')
       *typ= 5;
     else if(*cpt=='6' || *cpt=='b' || *cpt=='B')
       *typ= 6;
     else
       ret&= ~2;
     while(*(cpt+1)==',') {
       /* extra info */;
       if(*(cpt+2)=='=') { 
         if(*(cpt+3)=='\\') {

           /* >>> ??? announce escaped '=' and '\\' ? */;
           /* rather not . ts A50903 (when removing num_eqsep) */

         } else {
           /* formerly : number of '=' used as target=source separator 
                         like  -,=2 ... /ab=c==/x/y/ab=c 
              in usage only for one development upload, removed A50903 */
    
           fprintf(stderr,
           "--- obsoleted intermediate list format detected: \"num_eqsep\"\n");
           fprintf(stderr,"    in:  %s\n",line);
           fprintf(stderr,"    For help contact:  scdbackup@gmx.net\n");
           ret= 0; goto ex;
         }

       } else if(*(cpt+2)=='d') {
         Intvec_from_dec(dev_vec,CdbplivS,cpt+3,1);
         Intvec_from_double(ino_vec,CdbplivS,0.0,0);
         if(Intvec_to_dev_ino(dev_vec,CdbplivS,ino_vec,device,&ino,0)>0)
           ret|= 64;

       } else if(*(cpt+2)=='i') {
         Intvec_from_double(dev_vec,CdbplivS,0.0,0);
         Intvec_from_dec(ino_vec,CdbplivS,cpt+3,1);
         if(Intvec_to_dev_ino(dev_vec,CdbplivS,ino_vec,&dev,inode,0)>0)
           ret|= 32;

       } else if(*(cpt+2)=='P') {
         /* ,P$start:$count : full original of split file */;
         /* for incremental old_tree */
         if(flag&(1<<15))
           ret|= 1<<15;

       } else if(*(cpt+2)=='p') {
         /* ,p$start:$size : split file part */;
         /* for unpacking and backup content search */
         if(!(flag&(1<<15)))
           ret|= 1<<15;

       } else {
         /* unknown extra info type : ignore and advance */;
         /* >>> ??? issue warning ??? */
       }
       cpt= strchr(cpt+2,',');
       if(cpt==NULL || cpt>npt)
     break;
       cpt--;
     }
     cpt= npt+1;
   }
   if(is_big_format) {
     if((*xprop)==NULL) {
       hret= Filelist_xprop_rec_new(xprop,0);
       if(hret<=0)
         goto ex;
     }
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     sscanf(cpt,"%o",&((*xprop)->mod));
     cpt= npt+1;
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     if(Sregex_string_cut(&((*xprop)->own),cpt,npt-cpt,0)<=0)
       {ret= -1; goto ex;}
     cpt= npt+1;
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     if(Sregex_string_cut(&((*xprop)->grp),cpt,npt-cpt,0)<=0)
       {ret= -1; goto ex;}
     cpt= npt+1;
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     if(Sregex_string_cut(&date_str,cpt,npt-cpt,0)<=0)
       {ret= -1; goto ex;}
     hret= String_to_time(date_str,&valid,&((*xprop)->date),0);
     if(hret<=0)
       goto ex;
     cpt= npt+1;
   }
   if(flag&1024) {
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     while(*(npt+1)==' ')
       npt++;
     sscanf(cpt,"%lf",size);
     if(*size>=0)
       ret|= 4;
     cpt= npt+1;
   }
   if(flag&2048) {
     npt= strchr(cpt,' ');
     if(npt==NULL)
       {ret= 0; goto ex;}
     codelen= npt-cpt;
     while(*(npt+1)==' ')
       npt++;
     if(*cpt!='-' && *cpt!=0) {
       if(codelen==32) {
         buf[2]= 0;
         for(i=0;i<16;i++) {
           buf[0]= cpt[2*i];
           buf[1]= cpt[2*i+1];
           sscanf(buf,"%x",&x);
           md5[i]= x;
         }
         ret|= 16;
       } else
         sscanf(cpt,"%lx",crc);
       ret|= 8;
     }
     cpt= npt+1;
   }

   /* handle address part */
   if(flag&(1<<13) || *cpt!='/')
     npt= NULL;                           /* return raw address */
   else {

     Fileliste__target_source_limit(cpt,'=',&npt,0);

   }
   if(npt!=NULL) {
     if(flag&(1<<14)) {
       /* cut off source part */
       *npt= 0;
       /* unescape target */
       Fileliste__unescape_target(cpt,npt,'=',1);

       /* 0580 : escape as source so that it fits source treatment */
       Fileliste__esc_newline(cpt, escaped_adr, 0);
       if(strlen(escaped_adr) < 4096)
         strcpy(cpt, escaped_adr);

       /* make relative addresses */
       if(cpt[0]=='.' && cpt[1]=='/') 
         ;
       else {
         if(cpt-line<2) {
           for(npt= cpt+strlen(cpt); npt>=cpt; npt--)
             *(npt+2)= *npt;
           cpt+= 2;
         }
         if(cpt[0]=='/') {
           /* prepend . */;
           *(--cpt)= '.';
         } else {
           /* prepend ./ */;
           *(--cpt)= '/'; /* last first */
           *(--cpt)= '.';
         }
       }
     } else
       /* skip target part */
       cpt= npt+1;
   }
   if(cpt!=line) {
     for(npt= line;*cpt!=0;cpt++)
       *(npt++)= *cpt;
     *npt= 0;
   }
 }
ex:;
 Sregex_string(&date_str,NULL,0);
 return(ret);
}


int Fileliste_decode_serialized_head(liste,line,waiting_for_wd,
                                     initial_wd,initial_wd_list,current_list,
                                     list_type,
                                     layout_flag,got_wd,at_top_level,flag)
struct FilelistE *liste;
char *line;
int *waiting_for_wd;         /* in/out: status of header interpretation */
char *initial_wd;            /* out|const (see flag) */
struct FilelistE **initial_wd_list; /* out */
struct FilelistE **current_list;    /* out */
int *list_type;              /* out : <0= invalid / unusable
                                       0= raw
                                       1= complete backup area file list
                                       2= backup content list
                                       3= complete backup area directory list
                                       4= vanished file objects list
*/
int *layout_flag;            /* out */
int *got_wd;                 /* out */
int *at_top_level;           /* out */
int flag;
/*
 bit0-3= mode :
          2=graft files into a tree under liste
          other modes only  interesting to Fileliste_read_serialized_tree()
 bit4= in mode 2 : delete files from liste tree rather than to insert them
 bit5= do not process wd line when finally found
 bit6= do not accept content lists but only complete and raw formats
       (eventually return(2) )
 bit14= do not overwrite initial_wd
*/
/*
 return:
  <=0 = error
    1 = ok
    2 = not an acceptable list format
    3 = line is a valid address, do not skip
*/
{
 int ret,ignore_wd,mode,del;

 if(*waiting_for_wd==0)
   return(1);
 del= flag&16;
 mode= flag&15;
 ignore_wd= !!(flag&32);
 if(*waiting_for_wd==-1) {
   *waiting_for_wd= 1;
   *list_type= -1;
   if(strcmp(line,"#=# complete backup area file list")==0)
     *list_type= 1;
   else if( strcmp(line,"#=# backup content list")==0)
     *list_type= 2;
   else if(strcmp(line,"#=# complete backup area directory list")==0)
     *list_type= 3;
   else if(strcmp(line,"#=# vanished file objects list")==0)
     *list_type= 4;
   else if(strncmp(line,"#=# ",4)==0) {
     fprintf(stderr,
               "--- this file seems to be in a list format from the future\n");
     ret= 0; goto ex;
   } else if(line[0]=='#' && line[1]=='!') {
     fprintf(stderr,"--- this file seems to be an interpreter script\n");
     ret= 0; goto ex;
   } else {
     *waiting_for_wd= 0;
     *list_type= 0;
     ret= 3; goto ex;
   }
   if(flag&64) 
     if(*list_type>1)
       {ret= 2; goto ex;}
 } else if(*waiting_for_wd==1) { 
   /* gathering header information until working directory is encountered */
   if(strncmp(line,"#=# line layout :",17)==0) {
     *layout_flag= 0;
     if(strstr(line+17," type ")!=NULL)
       *layout_flag|= (1<<9);
     if(strstr(line+17," size ")!=NULL)
       *layout_flag|= (1<<10);
     if(strstr(line+17," crc ")!=NULL || 
        strstr(line+17," md5 ")!=NULL ||
        strstr(line+17," chk ")!=NULL)
         *layout_flag|= (1<<11);
   } else if(strcmp(line,"#=# next line tells current working directory")==0)
     *waiting_for_wd= 2;
 } else if(*waiting_for_wd==2) {
   line[strlen(line)-1]= 0;
   if(!ignore_wd) {
     if(!(flag&(1<<14)))
       strcpy(initial_wd,line);
     if(mode==2) {
       ret= Fileliste_build_branch(liste,initial_wd_list,initial_wd,"",
                                   0,-1.0,0,NULL,
                                   (dev_t) 0, (ino_t) 0,
                                   NULL,1|del);
       if(ret<=0) {
         fprintf(stderr,"--- cannot build file list branch '%s'\n",initial_wd);
         goto ex;
       }
       *current_list= *initial_wd_list;
     }
   }
   *waiting_for_wd= 0;
   *got_wd= 1;
   *at_top_level= 1;
 }

 ret= 1;
ex:;
 return(ret);
}


int Fileliste_read_serialized_tree(liste,fpin,ftree,fpout,fpout_aux,
                                   initial_wd,initial_wd_list,prepend_dir,flag)
struct FilelistE *liste;
FILE *fpin;                             /* to be read from if ftree == NULL */
struct FilelistE *ftree;                     /* to be read from if not NULL */
FILE *fpout;
FILE *fpout_aux;
char initial_wd[4096];
struct FilelistE **initial_wd_list;
char *prepend_dir;
int flag;
/*
 bit0-3= mode :
          0=check existence and print absolute addresses of vanished files
            to fpout (print target and source )
          1=check existence and print vanished files as shell commands rm and
            rmdir to fpout (use target address only)
          2=graft files into a tree under liste
          3=check existence, compare with checksum and delete if it matches
          4=check existence, compare with checksum and report if it matches
          5=report complete addresses
          6=report input line and complete address
          7=combine mode 0 and 1, for mode 1 use fpout_aux
 bit4= in mode 2 : delete files from liste tree rather than to insert them
 bit5= initial_wd and initial_wd_list are already set
 bit6= do not accept content lists but only complete and raw formats
       (eventually return(2) )
 bit7= expect target=source pairs and process only the source part
       do not expect incremental tree notation
       with mode 4 : print the whole input line
 bit8= ignored, assumed always on : write address part of output line 
 bit9= prepend type
bit10= prepend length
bit11= preprend crc
bit12= use MD5 rather than CRC32
bit13= prepend inode information (as extension to type)
       for directories, prepend device number too
       >>> ??? use inode information in modes 3 and 4 ?
       >>>     inode would need a date comparison to be safe
bit14= choose target addresses rather than source addresses
       Use text given with  initial_wd  as prefix. Ignore prepend_dir
bit15= mode 0 or 1 : do not check type
bit16= allow mode 3 with insecure parameters
bit17= load header lines into  Cd_backup_planer_list_heaD (dirty trick)
bit18= this is level 0

return:
 <=0 = error
   1 = complete list
   2 = content list
   3 = raw list
   4 = old_tree
    | 256 = type change
*/
{
 char dir[4096],line[3*4096],adr[4096],*spt,shell_adr[3*4096],read_dir[4096];
 char *target= "",target_dir[4096],target_adr[4096],*pacifier_text= "";
 char serialized_line[200+3*4096];
 char *cpt,*fapt,*rpt;
 int waiting_for_wd= -1,ret= 0,slashno,absolute= 0,got_wd= 0;
 int mode, complete_list= 0,ignore_wd= 0,linecount= 0,dircount= 0;
 int noprint= 0,read_idx,typ,read_ret,layout_flag= 0,hflag,was_mismatch= 0;
 int is_vanished_file_list= 0,hashtype,rec_fits_dir,i,rec_nam_l;
 int malformed= 0,line_flag= 0,at_top_level= 1,tree_target_prepended= 0;
 int slashno_left= 0;
 int list_type,offset,was_type_change= 0,bdevmode;
 int inode_test,inode_valid,risk_d= 0,is_dir= 0,del= 0;
 double size,filecount= 0.0;
 unsigned long crc;
 unsigned char md5[16];
 struct FilelistE *current= NULL,*read_list;
 struct CheckrecorD *dir_rec_stack= NULL,*dir_rec;
 struct Filelist_xprop_reC *xprop= NULL;
 time_t dummy_time= 0;
 FILE *fp;
 dev_t device;
 ino_t inode;

 static int dummy_delete= 0;

 line[0]= 0;
 if(Cd_backup_planer_risk_d_enabled) {
   if(ftree==NULL)
     risk_d= 2;
   else
     risk_d= 4;
 }
 del= flag&16;
 mode= flag&15;
 dir[0]= 0;
 if((flag&32) && ftree==NULL) {
   ignore_wd= 1;
   strcpy(dir,initial_wd);
 } else {
   if(!(flag&(1<<14)))
     strcpy(initial_wd,"");
   if(mode==2)
     *initial_wd_list= liste;
   if(ftree!=NULL)
     waiting_for_wd= 0;
 }
 if(flag&(1<<17))
   strcpy(Cd_backup_planer_list_heaD,"");

 if(mode==2) {
   if(liste==NULL)
     {ret= -1; goto silent_ex;}
   current= liste;
   if(flag&32)
     if(*initial_wd_list!=NULL)
       current= *initial_wd_list;
 }

 if(ftree!=NULL)
   pacifier_text= "tree nodes read";
 else
   pacifier_text= "file list lines read";
 Print_pacifier(0,pacifier_text,1|2|4);

 read_list= ftree;
 read_idx= 0;
 read_dir[0]= 0;
 if(flag&(1<<14))
   line_flag|= (1<<14);
 else
   line_flag|= (1<<13);
 while(1) {
   line_flag= (line_flag&~layout_flag)|(layout_flag*!!got_wd);
   if(mode==6) {
     fprintf(stdout,"   %.lf  ",filecount);
     line_flag|= 1;
   }
   read_ret= Fileliste_read_serialized_line(&read_list,&read_idx,read_dir,fpin,
                                            line,&typ,&size,&crc,md5,
                                            &device,&inode,
                                            &xprop,line_flag);
   if(read_ret<=0) {
     if(got_wd || waiting_for_wd>0)
       goto unexpected_end;
     goto expected_end;
   }
   tree_target_prepended= !!(read_ret&(1<<13));
   bdevmode= 32;
   /* >>> eventually identify -special_file and set bdevmode= 0 */;

   linecount++;
   if(!(mode==0 || mode==1 || mode==7))
   if(((mode==3 || mode==4) && ((linecount%10)==0 || linecount==2))
      || (linecount%100)==0)
     Print_pacifier(linecount,"",2);
   if(line[0]==0)
 continue;

   /* 0497: skip line if of unwanted type (e.g. a part rec) */
   if(read_ret&(1<<15))
 continue;

   if(waiting_for_wd) {
     if(flag&(1<<17)) {
       strcat(Cd_backup_planer_list_heaD,line);
       strcat(Cd_backup_planer_list_heaD,"\n");
     }
     /* try to interpret line as part of file header */
     ret= Fileliste_decode_serialized_head(liste,line,&waiting_for_wd,
                                     initial_wd,initial_wd_list,&current,
                                     &list_type,
                                     &layout_flag,&got_wd,&at_top_level,
                                     flag&(15|16|32|64|(1<<14)));
     if(ret<=0 || ret==2)
       goto ex;
     complete_list= (list_type==1);
     is_vanished_file_list= (list_type==4);
     if(waiting_for_wd==0) {
       strcpy(dir,initial_wd);
       at_top_level= 1;
     }
     if(ret!=3)
 continue;
   }
   if(got_wd) {
     if(strcmp(line,"#=# end")==0)
 break;
     if(line[0]=='#' && line[1]=='=' && line[2]=='#' && line[3]==' ')
 continue;
   }

   target= "";
   fapt= line;
   /* if a raw address was fetched : split it */
   if((flag&128 || at_top_level || tree_target_prepended) && !(flag&(1<<14))) {
     Fileliste__target_source_limit(line,'=',&cpt,0);
     if(cpt!=NULL && !(mode==4 && (flag&128))) {
       Fileliste__unescape_target(line,cpt,'=',1);

       /* 0520: */
       Fileliste__unescape_source(cpt+1,0);

       offset= (cpt-line)-strlen(line);
       *cpt= '=';
       if(offset>0) {
         for(rpt= cpt; *rpt!=0; rpt++)
           *(rpt-offset)= *rpt;
         *(rpt-offset)= 0;
         cpt-= offset;
       }
     }
     if(cpt!=NULL) {
       fapt= cpt+1;
       if(!(mode==4 && (flag&128)))
         *cpt= 0;
       target= line;
     }
   } else {

     /* 0580: relative addresses go here and need unescaping too */
     Fileliste__unescape_source(line,0);

   }
   if(strcmp(fapt,"./")==0 || fapt[0]==0)
 continue;

   /* >>> this does not work correctly with (flag&128) */
   /* >>> ( but maybe it is better this way ) */
   if(fapt[strlen(fapt)-1]=='/') {
     is_dir= 1;
     fapt[strlen(fapt)-1]= 0;
     if(fapt[0]!='.' || fapt[1]!='.')
       dircount++;
     at_top_level= 0;
   } else
     is_dir= 0;

   if(flag&(1<<14))
     strcpy(adr,"");
   else
     strcpy(adr,prepend_dir);
   strcpy(target_adr,""); /* ??? should this be prepend_dir rather than "" ? */

   if(flag&128) {
     strcat(adr,fapt);
     strcat(target_adr,fapt);
   } else if(fapt[0]=='/') {
     absolute= 1;
     if(mode==2) {
       ret= Fileliste_build_branch(
              liste,&current,fapt,target,typ,size,crc,md5, device,inode,
              xprop,risk_d|is_dir|del|(read_ret<<8)|(flag&(1<<18)));
       if(ret<=0) {
could_not_build_branch:;
         fprintf(stderr,"--- cannot build file list branch '%s'\n",fapt);
         goto ex;
       }
     }
     if(is_dir) {
       strcpy(dir,fapt);
       if(target[0]!=0)
         strcpy(target_dir,target);
       else
         strcpy(target_dir,fapt);

       /* push directory stack */
       if(!(read_ret&4))
         size= -1;
       hashtype= -1;
       if(read_ret&8)
         hashtype= 0;
       if(read_ret&16)
         hashtype= 1; 
       ret= Checkrecord_new(&dir_rec,fapt,dummy_time,size,NULL,hashtype,
                            crc,md5,2);
       if(ret<=0) {
         fprintf(stderr,"--- not enough memory for directory stack\n");
         goto ex;
       }
       Checkrecord_link(dir_rec,dir_rec_stack,0);
       dir_rec_stack= dir_rec;

       /* >>> RELATIV : chdir auf neues Relativlevel */

       if(mode==5 || mode==6) {
         filecount++;
         if(flag&(1<<14))
           fprintf(stdout,"%s/\n",target_dir);
         else
           fprintf(stdout,"%s%s/\n",prepend_dir,dir);
       }
 continue;
     }
     strcat(adr,fapt);
     strcat(target_adr,fapt);
   } else if(fapt[0]=='.' && fapt[1]=='/') {
     if(mode==2) {
       ret= Fileliste_build_branch(
              current,&current,fapt,target, typ,size,crc,md5, device,inode,
              xprop, risk_d|is_dir|del|(read_ret<<8)|(flag&(1<<18)));
       if(ret<=0) 
         goto could_not_build_branch;
     }
     if(is_dir) {
       strcat(dir,fapt+1);
       if(target[0]=='/')
         target_dir[0]= 0;
       if(target[0]!=0)
         strcat(target_dir,target);
       else
         strcat(target_dir,fapt+1);

       /* push directory stack */
       if(!(read_ret&4))
         size= -1;
       hashtype= -1;
       if(read_ret&8)
         hashtype= 0;
       if(read_ret&16)
         hashtype= 1; 
       ret= Checkrecord_new(&dir_rec,fapt+2,dummy_time,size,NULL,hashtype,
                            crc,md5,2);
       if(ret<=0) {
         fprintf(stderr,"--- not enough memory for directory stack\n");
         goto ex;
       }
       Checkrecord_link(dir_rec,dir_rec_stack,0);
       dir_rec_stack= dir_rec;

       /* >>> RELATIV : chdir auf neues Relativlevel */

       if(mode==5 || mode==6) {
         filecount++;
         if(flag&(1<<14))
           fprintf(stdout,"%s/\n",target_dir);
         else
           fprintf(stdout,"%s%s/\n",prepend_dir,dir);
       }
 continue;
     } else {
       strcat(adr,dir);
       strcat(adr,fapt+1);
       strcat(target_adr,target_dir);
       strcat(target_adr,fapt+1);
     }
   } else if(fapt[0]=='.' && fapt[1]=='.') {
     if(is_dir) 
       slashno= 0;                      /* do not check the parent directory */
     else 
       slashno= strlen(fapt)/3;         /* computes "/" in "../../../.." */
     for(;slashno>=0;slashno--) {
       /* obtain size and crc */
       dir_rec= dir_rec_stack;
       if(dir_rec==NULL)
     break;
       dir_rec_stack= dir_rec_stack->prev;      /* pop directory stack */
       Checkrecord_link(dir_rec,NULL,0);

       /* >>> RELATIV : one directory up */

       /* does it fit the end of dir ? */
       rec_fits_dir= 0;
       Checkrecord_get_name(dir_rec,&spt,0);

       /* ??? >>> target_adr ??? (adr looks like a temporary text) */
       strcpy(adr,spt);
       rec_nam_l= strlen(adr);

       if(strlen(dir)>=rec_nam_l && rec_nam_l!=0)
         if(strcmp(dir+(strlen(dir)-strlen(adr)),adr)==0)
           rec_fits_dir= 1;
       if(!rec_fits_dir)
     break;
       for(i=strlen(adr)-1;i>=0;i--)
         if(adr[i]=='/')
           slashno--;
       read_ret= 2;
       typ= 2;
       Checkrecord_get_size(dir_rec,&size,0);
       if(size>=0)
         read_ret|= 4;
       Checkrecord_get_checksum(dir_rec,&hashtype,&crc,md5,0);
       if(hashtype==0) 
         read_ret|= 8;
       else if(hashtype==1)
         read_ret|= (16|8);
       Checkrecord_destroy(&dir_rec,0);
       if(mode==0 || mode==1 || mode==7) {
         /* files done, finally check existence of dir */
         if((linecount%10)==0)
           Print_pacifier(linecount,dir,2|4|16|32);

         /* >>> RELATIV : Dir */

         ret= Type_and_size(dir,NULL,1);
         if(ret==-1) {
handle_vanished_directory:;
           if(mode==1 || mode==7) {
             /* write relative to / */
             if(mode==7)
               fp= fpout_aux;
             else
               fp= fpout;
             fprintf(fp,"rmdir %s\n",
                        Text_shellsafe(target_dir+(target_dir[0]=='/'),
                                       shell_adr,0));
           } 
           if(mode==0 || mode==7) {
             hflag= flag&(512|1024|2048|8192);
             ret= Fileliste__format_serialized_line(serialized_line,
                        typ,size,crc,md5,target_dir,dir,
                        device,inode,(time_t) 0,(time_t) 0,
                        read_ret&(2|4|8|16|32|64),hflag);
             if(ret>0)
               fprintf(fpout,"%s\n",serialized_line);
           }
         } else if((!(flag&(1<<15))) && ret!=2) {
           /* directory changed into something else */
           was_type_change= 1;
           goto handle_vanished_directory;
         }
       } else if(mode==2) {
         /* move current upward */
         if(current->boss!=NULL) {
           current= current->boss;
         } else {
           current= *initial_wd_list;
         }
       } else if(mode==3 || mode==4) {
         if(flag&(1<<14))
           strcpy(adr,"");
         else
           strcpy(adr,prepend_dir);
         strcat(adr,dir);
         ret=  Remove_dir(adr,(mode==4)|((!!dummy_delete)<<1));
         if(ret<=0)
           was_mismatch= 1;
       }
       dir[strlen(dir)-1-rec_nam_l]= 0;
       if(strlen(target_dir)-1-rec_nam_l>0)
         target_dir[strlen(target_dir)-1-rec_nam_l]= 0;
       else
         target_dir[0]= 0;
       if(strcmp(dir,initial_wd)==0 && !absolute)
     break;
     }
     if(slashno>=0) {
       if(!slashno_left) {
         fprintf(stderr,
          "\n--- File list read error ! Program error or corrupted format.\n");
         fprintf(stderr,
                 "--- Could not move up to next directory. Levels open: %d\n",
                 slashno+1);
         fprintf(stderr,
                 "\n--- Please preserve list resp. backup configuration\n");
         fprintf(stderr,"--- and contact scdbackup@gmx.net\n\n");
         slashno_left= 1;
         if(!Cdbackup_planer_tolerate_bad_dirpoP) {
           fprintf(stderr,
 "--- (this abort can be prevented by -test_command tolerate_bad_dirpop)\n\n");
           return(-1);
         }
       }
     }
     if(is_dir) {
       at_top_level= 1;
       strcpy(dir,initial_wd);
       strcpy(target_dir,initial_wd);
       if(mode==2)
         current= *initial_wd_list;
       absolute= 0;
       Checkrecord_destroy_all(&dir_rec_stack,0);

       /* >>> RELATIV : chdir auf toplevel */

     }
     if(mode==6) {
       if(flag&(1<<14))
         fprintf(stdout,"%s/\n",target_dir);
       else
         fprintf(stdout,"%s%s/\n",prepend_dir,dir);
     }
 continue;
   } else {
     if(malformed==0) {
       fprintf(stderr,"--- Malformed line of serialized tree !\n");
       fprintf(stderr,"line read    : %s\n",line);
       fprintf(stderr,"address part : %s\n",fapt);
       fprintf(stderr,"--- ( please contact scdbackup@gmx.net )\n");
       malformed= 1;
     }
     strcat(adr,fapt);
 continue;
   }

   if(mode==0 || mode==1 || mode==7) {
     /* test existence of adr */
     if((linecount%10)==0)
       Print_pacifier(linecount,adr,2|4|16|32);

     /* >>> RELATIV : File */

     ret= Type_and_size(adr,NULL,1);
     if(ret==-1) {
handle_vanished_file:;
       /* Vanished file ! */;
       if(mode==1 || mode==7) {
         /* write with address relative to / */
         if(mode==7)
           fp= fpout_aux;
         else
           fp= fpout;
         if((read_ret&2) && typ==2) {
           /* directory given without trailing slash */
           fprintf(fp,"rmdir %s\n",
                   Text_shellsafe(target_adr+(target_adr[0]=='/'),shell_adr,0));
         } else {
           fprintf(fp,"rm %s\n",
                   Text_shellsafe(target_adr+(target_adr[0]=='/'),shell_adr,0));
         }
       }
       if(mode==0 || mode==7) {
         hflag= flag&(512|1024|2048|8192);
         ret= Fileliste__format_serialized_line(serialized_line,
                        typ,size,crc,md5,target_adr,adr,
                        device,inode,(time_t) 0,(time_t) 0,
                        read_ret&(2|4|8|16|32|64),hflag);
         if(ret>0)
           fprintf(fpout,"%s\n",serialized_line);
       }
     } else if(!(flag&(1<<15))) { 
       if(typ>0 && ret!=typ) {
         was_type_change= 1;
         goto handle_vanished_file;
       }
     } 
   } else if((mode==3 || mode==4) && (read_ret&2) && typ==2) {
     /* directory given without trailing slash */
     ret= Remove_dir(adr,(mode==4)|((!!dummy_delete)<<1));
     if(ret<=0)
       was_mismatch= 1;
   } else if(mode==3 || mode==4) {
     if(mode==3 &&
        (layout_flag&((1<<9)|(1<<10)|(1<<11)))!=((1<<9)|(1<<10)|(1<<11))  && 
        !(flag&(1<<16)) ) {
       fprintf(stderr,
        "--- will not delete files without type,size,checksum verification !\n"
              );
       ret= 0; goto ex;
     }
     if(mode==3 && (!is_vanished_file_list) && !(flag&(1<<13)) ) {
       fprintf(stderr,
         "--- will delete files only if recorded by  -vanished_list_adr !\n"
              );
       ret= 0; goto ex;
     }
     /* test existence and eventually type, size, checksum 
        In non-dangerous mode 3 , count missing information as mismatch */

     inode_test= 0;
     inode_valid= 0;

     /* >>> ??? is this a place where to apply inode comparison ? */
     if(0        && (flag&(1<<13)) && (read_ret&32))
                 inode_test= (3<<13);

     if(read_ret&32)
       inode_valid= 64;

     ret= File_is_changed(adr,bdevmode,0,0.0,0,NULL,
                          (dev_t) 0,(ino_t) 0,
                          (read_ret&(8|16))|inode_valid,typ,size,crc,md5,
                          device,inode,
                          1 | ((mode==3 && !(flag&(1<<16)))<<1)
                            | ((read_ret&14)<<8) | inode_test);
     if(ret<0)
       goto ex;
     if(ret==0 && mode==3 && (read_ret&(4|8))!=(4|8) && !(flag&(1<<16)) ) {
       fprintf(stderr,"--- unclear  : %s\n",adr);
     } else if(ret==0) {
       /* unchanged */
       if(mode==3) {
         /* delete file */
         if(dummy_delete)
           fprintf(stdout,"rm %s\n",Text_shellsafe(adr,shell_adr,0));
         else {
           if(unlink(adr)==-1) {
             fprintf(stderr,"--- Cannot remove : %s \n    %s",
                            adr,strerror(errno));
             ret= 0; goto ex;
           }
         }
       } else {
         if(flag&128)
           fprintf(stdout,"%s\n",line);
         else
           fprintf(stdout,"%s\n",adr);
       }
     } else if(ret==1) {
       fprintf(stderr,"--- mismatch : %s\n",adr);
       was_mismatch= 1;
     } else if(ret==2) {
       fprintf(stderr,"--- missing  : %s\n",adr);
       was_mismatch= 1;
     }
   } else if(mode==5 || mode==6) {
     filecount++;
     fprintf(stdout,"%s\n",adr);
   }
 }
expected_end:;
 ret= 1;
 if(complete_list)
   ret= 2;
 if(!got_wd)
   ret= 3;
 if(ftree!=NULL)
   ret= 4;
 if(mode==3 || mode==4)
   if(was_mismatch)
     ret= 0;
 if(was_type_change)
   ret|= 256;

ex:;
 if((flag&64)&&ret==2&&(mode==2))
   noprint= 1;
 if(noprint)
   Print_pacifier(linecount,"",1|8|16);
 else {
   if(ftree!=NULL)
     Print_pacifier(linecount,"tree nodes read",1|2|4|8|16);
   else
     Print_pacifier(linecount,"",1|2|8);
   if(ftree!=NULL)
     sprintf(line,"directory nodes");
   else
     sprintf(line,"directory lines");
   Print_pacifier(dircount,line,1|2|4|8);
 }
silent_ex:;
 Filelist_xprop_rec_destroy(&xprop,0);
 return(ret);

unexpected_end:;
 fprintf(stderr,"\n--- unexpected end of input\n");
 fprintf(stderr,"\n--- last line: '%s'\n",line);
 ret= 0;
 goto ex;
}


double Fileliste__size_by_internal(adr,size,flag)
char *adr;
double size;
int flag;
{
 double res= 0.0;

 res= size+Fileliste_node_overhead;
 if(Fileliste_node_addrl)
   res+= strlen(adr);
 return(res);
}


int Fileliste_type_and_size(liste,idx,ret_size,ts_flag,flag)
struct FilelistE *liste;
int idx;
double *ret_size;  /* this is allowed to be == &(liste->groessen[idx]) */
int ts_flag;
int flag;
/*
 ts_flag : handed over to Type_and_size() , bit3 gets manipulated before
 flag : unused yet
*/
{
 int ret,no_block_dev_read= 0;
 double size;

 if(idx<0||idx>=liste->anz)
   return(0);
 /* do not count block device content by reading yet */
 ret= Type_and_size(liste->adressen[idx],&size,ts_flag&~8);
 if(liste->flags[idx]&32)
   if(liste->groessen[idx]>0.0) {
     no_block_dev_read= 1;
     size= liste->groessen[idx];
   }
 /* now read block device if really necessary */ 
 if(ret==6 && !no_block_dev_read && liste->flags[idx]&32) {
   ret= Type_and_size(liste->adressen[idx],&size,ts_flag|8);
   liste->groessen[idx]= size;
 }
 *ret_size= size;
 return(ret);
}


int Fileliste_special_part_recs(liste,split_file_dir,split_file_prefix,
                                split_file_suffix,tempno,flag)
struct FilelistE *liste;
char *split_file_dir;
char *split_file_prefix;
char *split_file_suffix;
int *tempno;
int flag;
/*
 bit0= only test wether there is a special file in the list
 bit2= replace characters of liste->impl_igno by their hexcodes
*/
{
 int i,norec,ret;
 struct Filelist_part_reC *part_rec;

 for(i=0;i<liste->anz;i++) {
   if(liste->typen[i]<0)
 continue;
   norec= 0;
   if(liste->flags[i]&32) {
     if(flag&1)
       return(1);
     if(liste->part_recs==NULL)
       norec= 1;
     else if(liste->part_recs[i]==NULL)
       norec= 1;
     if(norec) {
       ret= Filelist_part_rec_new(&part_rec,1,1,0.0,
                                  liste->groessen[i],liste->groessen[i],0);
       if(ret<=0)
         return(-1);
       (*tempno)++;
       Filelist_part_rec_set_tempno(part_rec,*tempno,0);
       ret= Fileliste_set_part_rec(liste,i,part_rec,0);
       if(ret<=0)
         return(-1);
       ret= Filelist_part_rec_make_tempadr(part_rec,liste->adressen[i],
                            split_file_dir,split_file_prefix,split_file_suffix,
                            liste->impl_igno,flag&4);
       if(ret<=0)
         return(-1);
     }
   }
 }
 if(flag&1)
   return(0);
 return(1);
}


int Fileliste_sort_index(liste,index,index_count,flag)
struct FilelistE *liste;
int **index;
int *index_count;
int flag;
/*
 bit0-7: mode
   0=source
   1=target
   2=time
*/
{
 int ret,i,pt_count,mode;
 char target[4096];
 char **crit= NULL,***crit_pt= NULL;

 if((*index)!=NULL)
   free((char *) *index);
 (*index)= NULL;
 (*index_count)= 0;
 if(liste->anz<=0)
   return(2);
 mode= (flag&255);

 crit= TSOB_FELD(char *,liste->anz);
 crit_pt= TSOB_FELD(char **,liste->anz);
 if(crit==NULL || crit_pt==NULL)
   {ret= -1; goto ex;}
 for(i=0;i<liste->anz;i++) {
   crit[i]= NULL;
   crit_pt[i]= NULL;
 }
 pt_count= 0;
 for(i=0;i<liste->anz;i++) {
   if(liste->typen[i]<0)
 continue;
   if(mode==1) {
     ret= Fileliste_target(liste,i,target,1|(2*!!Fileliste_target_modern));
     if(ret<=0)
       goto ex;
     if(Sregex_string(&(crit[i]),target,0)<=0)
       {ret= -1; goto ex;}
   } else if(mode==2) {
     Filetimestamp(liste->adressen[i],target,0);
     if(Sregex_string(&(crit[i]),target,0)<=0)
       {ret= -1; goto ex;}
   } else {
     if(Sregex_string(&(crit[i]),liste->adressen[i],0)<=0)
       {ret= -1; goto ex;}
   }
   crit_pt[pt_count]= &(crit[i]);
   pt_count++;
 }
 if(pt_count<=0)
   {ret= 2; goto ex;}
 (*index)= TSOB_FELD(int,pt_count);
 if((*index)==NULL)
   {ret= -1; goto ex;}
 (*index_count)= pt_count;

 qsort(crit_pt,pt_count,sizeof(char *),Strcmp_pt);
 for(i=0;i<pt_count;i++)
   (*index)[i]= crit_pt[i]-crit;
 
 ret= 1;
ex:;
 if(crit!=NULL) {
   for(i=0;i<liste->anz;i++) 
     if(crit[i]!=NULL)
       free((char *) crit[i]);
   free((char *) crit);
 }
 if(crit_pt!=NULL)
   free((char *) crit_pt);
 return(ret);
}


int Fileliste__report_newline_adr(adr,flag)
char *adr;
int flag;
/*
 bit0= print full warning header
*/
{
 if(flag&1) {
   fprintf(stderr,"\n\n!!! VERY PROBLEMATIC FILE ADDRESS ENCOUNTERED !!!\n");
   fprintf(stderr,
        "    It contains at least one newline character and will therefore\n");
   fprintf(stderr,
        "    be unusable with most backup formatters and other tools.\n");
   fprintf(stderr,
        "    Consider to rename that file or to enable automatic exclusion\n");
   fprintf(stderr,
        "    by option  -filecheck -nl_exclude .\n");
   fprintf(stderr,
         "    The offending address is:\n%s\n",adr);
 } else {
   fprintf(stderr,
         "\nFile address with at least one newline character encountered:\n");
   fprintf(stderr,"%s\n",adr);
 }
 fprintf(stderr,
         "    (this line marks the end of the offending address)\n\n");
 return(1);
}


/* ----------------- Fileliste - end ----------- */


int Cd_backup_planer_assert_depth(depth, adr, flag)
int depth;
char *adr;
int flag;
{
 if(depth<=Cd_backup_planer_tree_max_deptH)
   return(1);
 fprintf(stderr,"\n");
 fprintf(stderr, "Maximum tree depth exceeded%s\n", adr[0] ? " by:" : ".");
 if(adr[0])
   fprintf(stderr, "  %s\n",adr);
 fprintf(stderr,
        "Hint: You may extend the limit of %d by option -max_tree_depth\n",
        Cd_backup_planer_tree_max_deptH);
 fprintf(stderr,
        "Hint: In that case ulimit -s should be at least depth * 20kB\n");
 fprintf(stderr,"\n");
 return(-1);
}


/* ----------------- Crawler ------------------- */

struct CrawleR {

 struct Backup_plaN *boss;

 /* intermediate read_serialized_tree result */
 struct FilelistE *old_tree;

 /* precheck parameters */
 struct FilelistE *start_list;
 struct FilelistE *exclude_list;
 int exclude_not_startlist;
 time_t age_limit;
 int max_exempts;
 double min_saved_bytes_per_file;
 double exclude_size;
 int precheck_flag;
      /*
         bit1= consider files with st_dev != parent_dev to be unreadable
         bit2= do not check for real readability
         bit3= check for age
         bit4= consider remote (e.g. NFS) mountpoints to be unreadable
         bit5= with bit8 : mistrust directory timestamps
         bit6= ignore timestamps

         bit8= check for existence in old_tree
         bit9= check type in old_tree
         bit10= check size in old_tree
         bit11= check crc in old_tree
         bit12= consider files larger than exclude_size unreadable
         bit13= with bit12 : eventually use size of link target
         bit14= ignore ctime
         bit15= do not evaluate economy of directory splits
         bit16= allow device,inode check as replacement of crc check
         bit17= with bit16: allow to ignore device number
         bit18= with bit3 : dir_trust : allow parent of parent to be new
      */
 int problemlist_limit;

 /* leanfind parameters */
 /* dev_t parent_dev;  will be stacked and therefore a function parameter */
 int linecount;
 int what_lists_flag;
      /* 
         bit0= -complete_list_adr
         bit1= -dir_list_adr
         bit2= -dir_script_adr
         bit3= -vanish_list_adr
         bit4= -vanish_script_adr
      */
 int output_format_flag;
      /*
         bit8= prepend address (always on)
         bit9= prepend type
         bit10= prepend length
         bit11= preprend crc
         bit12= use MD5 rather than CRC32
         bit13= prepend inode information (as extension to type)
         bit16= prepend mtime (as extension to type)
         bit17= prepend ctime (as extension to type)
      */

 /* property buffer (established by Crawler_check_file) */
 char *adr;
 char *target;
 int typ;
 double size;
 double weight;
 time_t ctime,mtime;
 ino_t inode;
 dev_t dev;
 unsigned long crc;
 unsigned char md5[16];
 int valid_flag; 
      /*
         bit1=typ 
         bit2=size
         bit3=crc
         bit4=md5 rathern than crc
         bit5=inode
         bit6=device
         bit7=mtime
         bit8=ctime
      */
 int is_old;
 struct FilelistE *old_tree_list;
 int old_tree_list_idx;

 /* precheck result */
 struct FilelistE *ill_names;
 struct FilelistE *no_access;
 struct FilelistE *mount_points;
 struct FilelistE *too_old;
 struct FilelistE *young_exempts;
 struct FilelistE *too_big;
 /* 0520 */
 struct FilelistE *excluded_for_name;

 /* leanfind result */
 FILE *completelist_fp;
 FILE *dirlist_fp;
 FILE *dirscript_fp;
 int completelist_count;
 int dirlist_count;

 /* tree traversal */
 int traverse_count;

 /* read_serialized_tree result */
 FILE *vanishlist_fp;
 FILE *vanishscript_fp;
 int vanishlist_count;
 int type_change_count;

 /* composer */
 int composition_mode;
 double max_size;
 double min_size_fak;
 int max_items;
 int piece_count;
 struct Backup_stuecK *pieces;
 struct Backup_stuecK *current_piece;
 struct FilelistE *py_stack; /* provisory_young lists of Crawler_do_tree */
          /* Position of the first file item that is pending in the py_stack */ 
 int py_stack_piecemark; /* index to py_stack */
 int py_list_piecemark;  /* index to py_stack_piecemark item */
 double py_size;
 int py_count;
 struct FilelistE *split_dirs; 

 /* out : content list */
 FILE *contentlist_fp;
 int contentlist_count;

};


int Crawler_search_old_tree(o,adr,old_tree_list_pt,old_tree_list_idx,flag)
struct CrawleR *o;
char *adr;
struct FilelistE **old_tree_list_pt;
int *old_tree_list_idx;
int flag;
/*
 bit0= recursion
*/
{
 int idx;
 char *npt,*adrpt,abs_adr[4096];
 struct FilelistE *old_tree_list,*found_list;

 old_tree_list= *old_tree_list_pt;
 if(flag&1) {
   npt= strrchr(adr,'/');
   if(npt==NULL)
     npt= adr;
   else
     npt++;

   if(Cdbackup_planer_listsearch_debuG) {
     char *list_dir= NULL;
     if(old_tree_list!=NULL) {
       if(old_tree_list->boss!=NULL) {
         if(old_tree_list->boss_idx>=0 && 
            old_tree_list->boss_idx<old_tree_list->boss->anz) 
           list_dir= old_tree_list->boss->adressen[old_tree_list->boss_idx];
         else
           list_dir= "--bad_idx--";
       } else
         list_dir= "/";
       if(list_dir==NULL)
         list_dir= "--null--";
       fprintf(stderr,
               "listsearch_debug: looking for ./%s in %s  ( %lX , %d items)\n",
               npt,list_dir,(unsigned long) old_tree_list,old_tree_list->anz);
     } else 
       fprintf(stderr,"listsearch_debug: will not find ./%s in new branch\n",
                      npt);
   }

   *old_tree_list_idx= Fileliste_search_file(old_tree_list,npt,16);

   if(Cdbackup_planer_listsearch_debuG) {
     if(*old_tree_list_idx>=0)
       fprintf(stderr,"listsearch_debug: FOUND\n");
     else
       fprintf(stderr,"listsearch_debug: NOT FOUND\n");
   }

 } else { /* at top level */
   /* if relative address : prepend working directory */
   adrpt= adr;
   if(adrpt[0]!='/') {
     if(adrpt[0]=='.' && adrpt[1]=='/')
       adrpt+=2;
     if(Cd_backup_planer_cwD[0]!=0) {
       if(adrpt[0]==0)
         strcpy(abs_adr,Cd_backup_planer_cwD);
       else
         sprintf(abs_adr,"%s/%s",Cd_backup_planer_cwD,adrpt);
       adrpt= abs_adr;
     }
   }
   if(Fileliste_search_branch(old_tree_list,&found_list,&idx,adrpt,16)<=0)
     (*old_tree_list_idx)= -1;
   else {
     o->old_tree_list= old_tree_list= (*old_tree_list_pt)= found_list;
     o->old_tree_list_idx= (*old_tree_list_idx)= idx;
   }
 }
 return(1);
}


int Crawler_file_changecheck(o,adr,target,old_tree_list_pt,old_tree_list_idx,
                             result,flag)
struct CrawleR *o;
char *adr;
char *target;
struct FilelistE **old_tree_list_pt; /* eventual directory model in old_tree */
int *old_tree_list_idx;
int *result;
/* (coordinated with *result of Crawler_do_tree)
 bit3= too old
 bit6= item was added to too_old list
 bit8= file type changed
 bit9= has young timestamp
*/
int flag;
/*
 bit0= recursion
 bit1= do not add to old_
 bit5= compute block device crc from content and not from stat
 bit7= parent_directory has old timestamp
*/
{
 int ret= 0,idx,tree_flag,f_flag,hflag,inode_test;
 int has_crc,has_md5,has_inode,has_device,reason= 0;
 ino_t tree_inode= 0;
 dev_t tree_device= 0;
 unsigned long tree_crc;
 unsigned char *tree_md5;
 static unsigned char dummy_md5[16]= {"not an md5      "};
 struct FilelistE *old_tree_list;

 o->old_tree_list_idx= *old_tree_list_idx= -1;
 o->old_tree_list= old_tree_list= *old_tree_list_pt;
 
 if(o->typ==2 || (o->what_lists_flag&(8|16))) {
   /* look up tree in any case for : directories, vanished lists */
   Crawler_search_old_tree(o,adr,old_tree_list_pt,old_tree_list_idx,flag&1);
   idx= o->old_tree_list_idx= *old_tree_list_idx;
   old_tree_list= o->old_tree_list= *old_tree_list_pt;
 }

 o->is_old= 0;
 (*result)&= ~(8|64);
 if(!((o->precheck_flag)&8)) {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: NOT TESTED: %d : %s\n",o->typ,o->adr);
   {reason= 1; goto accept_as_young;}
 }

 if(Cdbackup_planer_agetest_debuG)
   fprintf(stderr,"agetest_debug: TESTING: %d : %s\n",o->typ,o->adr);
 o->is_old= 1;
 if(o->precheck_flag&64) /* assume old */
   ;
 else if(o->ctime >= o->age_limit && !(o->precheck_flag&(1<<14)))
   o->is_old= 0;
 else if(o->mtime >= o->age_limit)
   o->is_old= 0;
 if(o->is_old) {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: timestamps : old\n");
   if((flag&128)&&!(o->precheck_flag&32)) {
     /* parent directory is trustfully old. No moves possible */
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: parent directory trustfully old\n");
     {reason= -1; goto really_old;}
   }
 } else {
   (*result)|= 512;
   if(o->typ!=2) {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,
               "agetest_debug: young timestamp detected with non-directory\n");
     {reason= 2; goto accept_as_young;}
   }
 }
   
 if(!(o->precheck_flag&(256|512|1024|2048))) {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,
             "agetest_debug: not tested against old directory tree\n");
   if(o->is_old)
     {reason= -2; goto really_old;}
   {reason= 3; goto accept_as_young;}
 }

 /* check for change versus old_tree */;
 if(!(o->what_lists_flag&(8|16))) { /* if not already looked up for vanish */
   Crawler_search_old_tree(o,adr,old_tree_list_pt,old_tree_list_idx,flag&1);
   idx= o->old_tree_list_idx= *old_tree_list_idx;
   old_tree_list= o->old_tree_list= *old_tree_list_pt;
 }
 if(idx<0) {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: not found in old directory tree\n");
   {reason= 4; goto accept_as_young;}
 }
 if(Cdbackup_planer_agetest_debuG)
   fprintf(stderr,"agetest_debug: found in old_tree\n");

 /* obtain properties */
 has_crc= has_md5= 0;
 if(old_tree_list->crcs!=NULL && ((old_tree_list->flags[idx])&8)) {
   if(old_tree_list->crcs[idx]!=NULL) {
     tree_crc= old_tree_list->crcs[idx]->crc;
     tree_md5= old_tree_list->crcs[idx]->md5;
     has_crc= 1;
     if((old_tree_list->flags[idx])&16)
       has_md5= 1;
   }
 }
 if(!has_crc) {
   tree_crc= 0;
   tree_md5= dummy_md5;
 }
 has_inode= 0;
 if(old_tree_list->inodes!=NULL && (old_tree_list->flags[idx]&64)) {
   tree_inode= old_tree_list->inodes[idx];
   has_inode= 1;
 }
 has_device= 1;
 if(old_tree_list->typen[idx]==2) {
   /* if directory : obtain device number from subdir list */;
   if(old_tree_list->subdirs==NULL)
     has_device= 0;
   else if(old_tree_list->subdirs[idx]==NULL)
     has_device= 0;
   else if(!(old_tree_list->subdirs[idx]->dir_dev_ino_valid&1))
     has_device= 0;
   if(has_device)
     tree_device= old_tree_list->subdirs[idx]->dir_device;
 } else {
   /* obtain device number from hosting list */
   if(old_tree_list->dir_dev_ino_valid&1)
     tree_device= old_tree_list->dir_device;
   else
     has_device= 0;
 }

 /* eventually enable inode_test */
 inode_test= 0;
 if((o->precheck_flag)&(1<<16)) 
   inode_test= (3<<13);
 if(!(has_inode && (has_device || ((o->precheck_flag)&(1<<17)))))
   inode_test= 0;

 /* check wether essential information is missing 
    eventually count as changed */
 if(inode_test==0 && (o->precheck_flag)&(1<<16) && !((o->precheck_flag)&2048))
   goto ino_dev_needed;
 if((!has_crc) && ((o->precheck_flag)&2048)) {
   if(inode_test==0) {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: no crc information in old_tree\n");
     {reason= 5; goto accept_as_young;}
   } else {
ino_dev_needed:;
     if(!has_inode) {
       if(Cdbackup_planer_agetest_debuG)
         fprintf(stderr,
                 "agetest_debug: no inode information in old_tree\n");
       {reason= 6; goto accept_as_young;}
     }
     if((!has_device) && !((o->precheck_flag)&(1<<17))) {
       if(Cdbackup_planer_agetest_debuG)
         fprintf(stderr,
                 "agetest_debug: no device information in old_tree\n");
       {reason= 7; goto accept_as_young;}
     }
     if(inode_test==0 && ((o->precheck_flag)&(1<<16))) {
       if(Cdbackup_planer_agetest_debuG)
         fprintf(stderr,
                 "agetest_debug: cannot perform inode+device based agetest\n");
       {reason= 11; goto accept_as_young;}
     }
   }
 }

 f_flag= (flag&32)|((o->valid_flag)&(8|16));
 if(((o->valid_flag)&(32|64))==(32|64))
   f_flag|= 64;
 tree_flag= 0;
 if(has_crc)
   tree_flag|= old_tree_list->flags[idx]&(8|16);

 if(has_inode && (has_device || ((o->precheck_flag)&(1<<17))))
   tree_flag|= old_tree_list->flags[idx]&64;
 hflag= ((o->precheck_flag)&(64|512|1024|2048|(1<<17)))|inode_test;
 if(has_crc && (!inode_test) && ((o->precheck_flag)&2048) 
    && !((o->valid_flag)&3)) {
   if(has_md5) 
     ret= Determine_file_checksum(adr,NULL,NULL,o->md5,o->md5,&(o->size),
                                  NULL,NULL,((f_flag&32)^32)|(1<<12)|64);
   else
     ret= Determine_file_checksum(adr,NULL,NULL,&(o->crc),&(o->crc),
                                  &(o->size),NULL,NULL,((f_flag&32)^32)|64);
   if(ret<=0) {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,
              "agetest_debug: cannot determine file checksum on disk\n");
     {reason= 8; goto accept_as_young;}
   }

#ifndef Crawler_ruleS
 /* <<< currently adr is used to stay compatible with _precheck() */
    o->weight= Fileliste__size_by_internal(adr,o->size,0);
#else
  /* >>> as soon as precheck() is abandoned, target will be more realistic */
    o->weight= Fileliste__size_by_internal(target,o->size,0);
#endif

 }

 ret= File_is_changed(adr,f_flag,o->typ,o->size,o->crc,o->md5,o->dev,o->inode,
                      tree_flag,old_tree_list->typen[idx],
                      old_tree_list->groessen[idx],
                      tree_crc,tree_md5,tree_device,tree_inode,
                      2|hflag);
 if(ret<0)
   goto ex;
 if(ret>0) {
   if(old_tree_list->typen[idx]!=o->typ) {
     (*result)|= 256;
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,
               "agetest_debug: type change detected : old= %d , new= %d\n",
               old_tree_list->typen[idx],o->typ);
     reason= 9;   
   } else {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: content change detected\n");
     reason= 10;   
   }
 } else {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: seems unchanged\n");
   {reason= -3; goto really_old;}
 }

accept_as_young:;
 if(o->typ!=2)
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: DECISION: accept\n");
 ret= 1;
ex:
 return(ret);

really_old:;
 if(o->typ==2) /* directories get their fate decided in Crawler_do_tree */
   {ret= 1; goto ex;}
 (*result)|= 8;
 if(flag&2) {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: SPECIAL MODE: do not add to o->too_old\n");
 } else if(o->weight>2.0*(o->min_saved_bytes_per_file)
    || (!(flag&1)) || ((o->precheck_flag)&(1<<15))) {
   ret= Fileliste_append_prop(o->too_old,adr,target,o->typ,o->weight,1);
   if(ret<=0)
     goto ex;
   (*result)|= 64;
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: DECISION: exclude\n");
 } else {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: DECISION: too small to exclude\n");
 }
 ret= 1; goto ex;
}


int Crawler_write_completelists(o,flag)
struct CrawleR *o;
int flag;
/*
 bit0= recursion (not a top level address)
 bit1= report directory end
*/
{
 int ret,is_dir,hflag;
 char combined_adr[4*4096],slash[2],dotslash[3];
 char typetext[1024],lengthtext[80],crctext[80];
 char *name,adr_data[2*4096+100],*adr,*npt,end_text[8];
 FILE *fp= NULL,*fp_aux= NULL;

 if(!((o->what_lists_flag)&(1|2|4)))
   return(2);
 if(flag&2)
   goto report_directory_end;
 name= o->adr;
 adr= adr_data;
 is_dir= (o->typ==2);
 slash[0]= slash[1]= 0;
 if(is_dir)
                            /*  represent root-dir as /=/ and not as //=// */
   if(name[strlen(name)-1]!='/')
     slash[0]= '/';
 if(flag&1) {
   /* relative address */;
   /* ./basename[/] , ignore any trailing slash */;
   for(npt= name+strlen(name)-1;npt>name;npt--)
     if(*npt=='/')
   break;
   if(*npt=='/')
     npt++;
/* <<< 0520 old
   sprintf(adr,"./%s%s",npt,slash);
*/
   strcpy(adr,"./");
   Fileliste__esc_newline(npt,adr+2,0);
   if(slash[0])
     strcat(adr,slash);

 } else {
   dotslash[0]= dotslash[1]= dotslash[2]= 0;
   if(name[0]!='/')
     strcpy(dotslash,"./");
   if(name[0]=='.')
     if(name[1]=='/')
       dotslash[0]= 0;
   sprintf(adr,"%s%s%s",dotslash,name,slash);
   Fileliste__format_combined_adr(combined_adr,o->target,adr,0);
   adr= combined_adr;
 }
 if((o->what_lists_flag&1) || ((o->what_lists_flag&2)&&o->typ==2)) {
   if(o->output_format_flag&2048) {

     /* >>> LISTBUG */
     if(o->output_format_flag&(1<<12)) {

       (o->valid_flag)|= 8|16;
       ret= Determine_file_checksum(name,NULL,NULL,o->md5,o->md5,
                                    &(o->size),NULL,NULL,32|64|(1<<12));
     } else {
       (o->valid_flag)|= 8;
       ret= Determine_file_checksum(name,NULL,NULL,&(o->crc),&(o->crc),
                                    &(o->size),NULL,NULL,32|64);
     }
     if(ret<=0) 
       (o->valid_flag)&= ~(8|16);
   }
   Format_line_prepend(typetext,lengthtext,crctext,o->typ,o->size,
             o->crc,o->md5,o->dev,o->inode,o->mtime,o->ctime,
             (o->valid_flag)&(2|4|8|16|32|64|128|256),
             1|((o->output_format_flag)&(512|1024|2048|8192|(3<<16))));
 }
 
 if(o->what_lists_flag&1) {
   fprintf(o->completelist_fp,"%s%s%s%s\n",typetext,lengthtext,crctext,adr);
   o->completelist_count++;
 }

 if(o->typ!=2)
   {ret= 1; goto ex;}

 /* write to dirlist_fp and/or dirscript_fp */
 if((o->what_lists_flag)&(2|4)) {
   if(((o->what_lists_flag)&(2|4))==(2|4)) {
     fp= o->dirlist_fp;
     fp_aux= o->dirscript_fp;
     hflag= 3;
   } else if((o->what_lists_flag)&4) {
     fp= o->dirscript_fp;
     hflag= 1;
   } else {
     fp= o->dirlist_fp;
     hflag= 0;
   }
   ret= Record_dir_properties(fp,fp_aux,o->adr,adr,o->target,
                              lengthtext,crctext,hflag);
   if(ret<=0)
     goto ex;
   o->dirlist_count++;
 }

 ret= 1;
ex:;
 return(ret);

report_directory_end:;
 if(flag&1)
   strcpy(end_text,"..\n");
 else
   strcpy(end_text,"../\n");
 if((o->what_lists_flag)&1)
   fprintf(o->completelist_fp,end_text);
 if((o->what_lists_flag)&2)
   fprintf(o->dirlist_fp,end_text);
 ret= 1; goto ex;
}


int Crawler_pacifier(o,flag)
struct CrawleR *o;
int flag;
/*
 bit0= print final message 
 bit7= parent directory has old timestamp (i.e. no content check active)
*/
{
 int big_file= 0,crc_check= 0,force_show= 0;
 char line[4096+100],*textpt;

 if(flag&1) {
   Print_pacifier(o->linecount,"file objects examined",1|2|4|8|16);
   return(1);
 }

 /* entertain the user */
 big_file= (o->typ==1) && (o->size>5*1024*1024);
 if((o->precheck_flag&(8|256|2048))==(8|256|2048)) 
   if((o->precheck_flag&32)||!(flag&128))
     crc_check= 1;
   
 if((o->linecount%10)==0 || (o->linecount)==2 || crc_check || big_file) {
   textpt= "crawling";
   if(o->precheck_flag&8) { 
     if(crc_check) {
       if(o->precheck_flag&(1<<16))
         textpt= "i&c crawl";
       else
         textpt= "crc crawl";
     } else if(o->precheck_flag&(1<<16))
       textpt= "ino crawl";
     else
       textpt= "crawling ";
   }
   sprintf(line,"%s: %s",textpt,o->adr);
   Print_pacifier(o->linecount,line,force_show|2|4|16|32);
 }
 (o->linecount)++;
 return(1);
}


int Crawler_check_file(o,adr,target,parent_dev,result,flag)
struct CrawleR *o;
char *adr;         /* should be persistent for whole handling time of file */
char *target;      /* (see above) */
dev_t parent_dev;
int *result;
/*
 bit0= ill name
 bit1= not readable
 bit2= mountpoint of unwanted type
 bit5= too big
 bit6= item was added to too_old list
 bit7= excluded file
*/
int flag;
/*
 bit0= recursion (not a top level address)
 bit1= do not check for ill names
 bit7= parent directory has old timestamp (i.e. no content check active)
*/
/* return:
 <=0 = error
   1 = acceptable file
   2 = file is to be skipped
*/
{
 int ret,i,shall_be_skipped= 0,newline_adr_rule= 0;
 int is_dead_link= 0;
 char *npt,link_target[4096];
 struct stat stbuf;
 FILE *fp;
 double Fileliste__size_by_internal();

 o->valid_flag= 0;
 if((flag&1) || !o->exclude_not_startlist) {
   if(Fileliste_search_file(o->exclude_list,adr,32)>=0) {
     (*result)|= 128;
     return(2);
   }
#ifdef Cd_backup_planer_with_regeX
   ret= Backup_plan_match_not_pattern(o->boss, adr, 0);
   if(ret>0) {
     (*result)|= 128;
     ret= Backup_plan_exclude(o->boss,adr,0);
     if(ret<=0)
       return(ret);
     return(2);
   }
#endif
 }

 /* establish property buffer of adr */;
 o->adr= adr;
 o->target= target;
 ret= Lstat(adr,&stbuf,0);
 if(ret==-1) {
   (*result)|= 2;
   ret= 2; goto ex;
 }
 o->typ= Stbuf_type(&stbuf,0);
 o->inode= stbuf.st_ino;
 o->dev= stbuf.st_dev;
 o->ctime= stbuf.st_ctime;
 o->mtime= stbuf.st_mtime;

 /* note: stbuf not necessarily original after this point */
 o->size= 0.0;
 if(o->typ==3 && (o->precheck_flag&(1<<13))) {
   /* determine size of link target */
   ret= Type_and_size(adr,&(o->size),1|4|8);
   if(ret==-1)
     is_dead_link= 1;
 } else
   o->size= stbuf.st_size;

#ifndef Crawler_ruleS
 /* <<< currently adr is used to stay compatible with _precheck() */
 o->weight= Fileliste__size_by_internal(adr,o->size,0);
#else
 /* >>> as soon as precheck() is abandoned, target will be more realistic */
 o->weight= Fileliste__size_by_internal(target,o->size,0);
#endif

 o->valid_flag= ((o->typ!=0 && o->typ!=127)<<1)|4|32|64|128|256;

 Crawler_pacifier(o,flag&128);

 if(!(flag&2)) {

   /* 0520: check for newlines */;
   if(strrchr(adr,'\n')!=NULL) {
     newline_adr_rule= Backup_plan_get_newline_adr_rule(o->boss,0);
     if(newline_adr_rule==0) {
       Fileliste__report_newline_adr(adr,1);
       ret= -1; goto ex;
     } else if(newline_adr_rule==1) {
       /* 0520: put into newline exclusion list */;
       if(o->problemlist_limit>=o->excluded_for_name->anz) {
         ret= Fileliste_append_prop(o->excluded_for_name,
                                    adr,target,o->typ,o->size,1);
         if(ret<=0)
           goto ex;
       }
       (*result)|= 128;
       shall_be_skipped= 1;
       ret= 2; goto ex;
     }
   }

   /* check for strange characters which confuse old mkisofs */;
   npt= strrchr(adr,'/');
   if(npt==NULL) npt= adr;
   else          npt++;
   for(i=strlen(o->start_list->impl_igno)-1;i>=0;i--)
     if(strchr(npt,o->start_list->impl_igno[i])!=NULL)
       break;
   if(i>=0) {
     (*result)|= 1;
     if(o->problemlist_limit>=o->ill_names->anz) {
       ret= Fileliste_append_prop(o->ill_names,adr,target,o->typ,o->weight,1);
       if(ret<=0)
         goto ex;
     }
   }
 }

 /* check for -xdev , remote mount points */
 if(o->typ==2 && (flag&1) && 
    ((o->precheck_flag&2)||(o->precheck_flag&16))) {
   (*result)&= ~4;
   if(o->dev!=parent_dev) {
     if(o->precheck_flag&2)
       (*result)|= 4;
     else if(o->precheck_flag&16) {
       if(!Sfile_is_local_fs(adr,0))
         (*result)|= 4;
     }
     if((*result)&4) {
       if(o->problemlist_limit>=o->mount_points->anz) {
         ret= Fileliste_append_prop(o->mount_points,adr,target,o->typ,
                                    o->weight,1);
         if(ret<=0)
           goto ex;
       }
       shall_be_skipped= 1;
     }
   }
 }

 /* check for oversized files */
 if(o->precheck_flag&(1<<12) && o->typ!=2) {
   if(o->size>=o->exclude_size) {
     if(o->problemlist_limit>=o->too_big->anz) {
       /* use o->size rather than o->weight in order to avoid confusion */
       ret= Fileliste_append_prop(o->too_big,adr,target,o->typ,o->size,1);
       if(ret<=0)
         goto ex;
     }
     (*result)|= 32;
     shall_be_skipped= 1;
   }
 }

 if((o->typ!=4)&&!(o->precheck_flag&4)) {
   if(o->typ==3) {
     if(is_dead_link || readlink(adr,link_target,sizeof(link_target)-1)==-1)
       goto not_readable;
   } else if(o->typ==2) {
     DIR *dirpt;
     dirpt= opendir(adr);
     if(dirpt==NULL)
       goto not_readable;
     closedir(dirpt);
   } else if(o->typ==1) {
     if((fp= fopen(adr,"rb"))==NULL){
not_readable:;
       if(o->problemlist_limit>=o->no_access->anz) {
         ret= Fileliste_append_prop(o->no_access,adr,target,o->typ,o->weight,1);
         if(ret<=0)
           goto ex;
       }
       (*result)|= 2;
       shall_be_skipped= 1;
     }else
       fclose(fp);
   }
 }

 ret= 1+shall_be_skipped;
ex:
 return(ret);
}


int Crawler_traverse_old_tree(o,liste,source_base,target_base,flag)
struct CrawleR *o;
struct FilelistE *liste;
char *source_base;
char *target_base;
int flag;
/*
 bit0= recursion
 bit8= initialize existence flag
 bit9= report items without existence flag as -vanish_list_adr
bit10= report items without existence flag as -vanish_script_adr
bit11= try to add missing inode information
bit12= write to o->contentlist_fp
bit13= invalidate non-verificable old_tree items
*/
/*
return:
 <=0 : error
 >0 : ok
       with flag bit11 :
        bit0= ok
        bit3= too old
        bit4= contains some old items
*/
{
 int ret,i,hflag,typ,ret_flags= -1,precheck_flag_mark,valid_flag,format_flag;
 int explicit_target,handle_as_dir;
 double size;
 unsigned long crc= 0;
 unsigned char *md5_pt= NULL;
 char serialized_line[200+3*4096],source_adr[4096],target_adr[4096];
 char line[3*4096],shell_adr[3*4096],*cmd,*spt,combined_adr[2*4096+1];
 char typetext[4096],lengthtext[80],crctext[80];
 ino_t inode= 0;
 dev_t device= 0;
 time_t mtime,ctime,atime;

 if(liste==NULL)
   return(0);
 precheck_flag_mark= o->precheck_flag;
 if(!(flag&1)) {
   if(flag&(512|1024)) {
     if(o->vanishlist_fp!=NULL || o->vanishscript_fp!=NULL)
       Print_pacifier(o->vanishlist_count,"vanished file objects recorded",
                      1|2|4);
   } else if(flag&2048)
     Print_pacifier(o->traverse_count,"tree nodes examined",1|2|4);
   else if(flag&4096)
     Print_pacifier(o->contentlist_count,"content list lines written",1|2|4);
 }

 for(i=0;i<liste->anz;i++) {
   if(liste->adressen[i]==NULL || liste->typen[i]<0)
 continue;
   handle_as_dir= 0;
   if((liste->typen[i]==2 || liste->typen[i]==0) && liste->subdirs[i]!=NULL)
     handle_as_dir= 1;
   explicit_target= 0;
   if(liste->adressen[i][0]=='/') {
     strcpy(source_adr,liste->adressen[i]);
     Fileliste_target(liste,i,target_adr,2);
     explicit_target= 1;
   } else {
     sprintf(source_adr,"%s/%s",source_base,liste->adressen[i]);
     if(liste->targets[i]!=NULL) {
       strcpy(target_adr,liste->targets[i]);
       for(spt= target_adr+strlen(target_adr)-1; 
           spt>target_adr && (*spt)=='/'; spt--)
         *spt= 0;
       explicit_target= 1;
     } else 
       sprintf(target_adr,"%s/%s",target_base,liste->adressen[i]);
   }

   if(flag&4096) {
     /* write content list line */
     o->contentlist_count++;
     Print_pacifier(o->contentlist_count,"",2);
     typ= liste->typen[i];
     if(typ==0)
       goto skip_write;
     valid_flag= 2|4;
     size= liste->groessen[i];
     if(liste->crcs!=NULL && (liste->flags[i]&8))
       if(liste->crcs[i]!=NULL) {
         if(liste->flags[i]&16)
           md5_pt= liste->crcs[i]->md5;
         else
           crc= liste->crcs[i]->crc;
         valid_flag|= (liste->flags[i]&(8|16));
       }
     if(liste->inodes!=NULL && (liste->flags[i]&64)) {
       inode= liste->inodes[i];
       valid_flag|= 32;
       if(liste->subdirs[i]!=NULL)
         if(liste->subdirs[i]->dir_dev_ino_valid&1) {
           device= liste->subdirs[i]->dir_device;
           valid_flag|= 64;
         }
     }
     if(o->output_format_flag&(3<<16)) {
       ret= Get_filetimes(source_adr,&atime,&mtime,&ctime,0);
       if(ret>0)
         valid_flag|= (128|256);
     }
     format_flag= 256|(o->output_format_flag&(512|1024|2048|8192|(3<<16)));
     Format_line_prepend(typetext,lengthtext,crctext,typ,size,crc,md5_pt,
                     device,inode,mtime,ctime,valid_flag,format_flag);
     if((flag&1)&&!explicit_target) {
       /* relative address */

       /* 0520 */
/* <<<
       if(typ==2 || handle_as_dir)
         sprintf(combined_adr,"./%s/",liste->adressen[i]);
       else
         sprintf(combined_adr,"./%s",liste->adressen[i]);
*/
       strcpy(combined_adr,"./");
       Fileliste__esc_newline(liste->adressen[i],combined_adr+2,0);
       if(typ==2 || handle_as_dir)
         strcat(combined_adr,"/");

     } else {
       /* absolute address with target */
       Fileliste__format_combined_adr(combined_adr,target_adr,source_adr,
                                      (typ==2 || handle_as_dir)<<1);
     }
     fprintf(o->contentlist_fp,"%s%s%s%s\n",
             typetext,lengthtext,crctext,combined_adr);
   }
skip_write:;

   if((liste->typen[i]==2 || liste->typen[i]==0) && liste->subdirs[i]!=NULL) {
     ret= Crawler_traverse_old_tree(o,liste->subdirs[i],source_adr,target_adr,
                                flag|1);
     if(ret<=0)
       goto ex;
     if(ret&(8|16)) { /* sort out wether full or partially old */
       if((ret&8) && ret_flags<0)
         ret_flags= 8;
       ret_flags|= 16;
     } else {
       if(ret_flags<0)
         ret_flags= 0;
       ret_flags&= ~8;
     }
   }

   if((flag&4096) && liste->typen[i]==2) {
     if(explicit_target)
       fprintf(o->contentlist_fp,"../\n");
     else
       fprintf(o->contentlist_fp,"..\n");
   }

   if(flag&256) 
     (liste->flags[i])&= ~128;

   if(flag&(512|1024)) {
     /* Make sure the file has really vanished resp. changed its type */
     /* <<< (try to make this test unnecessary) */
     /* >>> obstacles : 
              changes in backup area, 
              automatically added scdbackup files
              tree working directory
     */

     /* <<< 0520 */
     if(1 || liste->typen[i]!=0) {

       if(!((liste->flags[i])&128)) {
         typ= Type_and_size(source_adr,&size,1);
         if(typ!=-1) {
           /* 0520 */
           if(typ==liste->typen[i] || liste->typen[i]==0) {
#ifdef Cdbackup_planer_developpeR
             if(liste->typen[i]!=0) {
               wait3(NULL,WNOHANG,NULL);
               Print_pacifier(0,"",1|16);
               fprintf(stderr,
                  "developer message : existing file unchecked in old tree\n");
               sprintf(line,
                     "ls -aldc %s >&2",Text_shellsafe(source_adr,shell_adr,0));
               system(line);
             }
#endif /* Cdbackup_planer_developpeR */

             liste->flags[i]|= 128;

           } else {
             o->type_change_count++;
#ifdef Cdbackup_planer_developpeR
             wait3(NULL,WNOHANG,NULL);
             Print_pacifier(0,"",1|16);
             fprintf(stderr,
                "developer message : type change detected. From %d to %d\n",
                 liste->typen[i],typ);
             sprintf(line,
                     "ls -aldc %s >&2",Text_shellsafe(source_adr,shell_adr,0));
             system(line);
#endif /* Cdbackup_planer_developpeR */
           }
         }
       }
       if(!((liste->flags[i])&128)) {
         o->vanishlist_count++;
         Print_pacifier(o->vanishlist_count,"",2);
         if(flag&512) {
           /* report as -vanish_list_adr */;
           ret= Fileliste_next_in_tree(&liste,&i,source_base,line,
                                       &(o->typ),&(o->size),&(o->crc),o->md5,
                                       &(o->dev),&(o->inode),1);
           if(ret<0)
             goto ex;
           if(ret==0)
 continue;
           o->valid_flag= ret&(2|4|8|16); /* ??? |32|64 for inode and dev ? */
           hflag= (o->output_format_flag)&(512|1024|2048|8192);
           ret= Fileliste__format_serialized_line(serialized_line,
                            o->typ,o->size,o->crc,o->md5,target_adr,source_adr,
                            o->dev,o->inode,(time_t) 0,(time_t) 0,
                            o->valid_flag,hflag);
           if(ret>0)
             fprintf(o->vanishlist_fp,"%s\n",serialized_line);

         }
         if((flag&1024)) {
           /* report as -vanish_script_adr */
           /* 0520 : distinguish type 0 by handle_as_dir */
           if(liste->typen[i]==2 || handle_as_dir)
             cmd= "rmdir";
           else
             cmd= "rm";
           fprintf(o->vanishscript_fp,
                   "%s %s\n",cmd,
                   Text_shellsafe(target_adr+(target_adr[0]=='/'),shell_adr,0));
         }
       }
     }
   }

   if(flag&(2048|8192)) {
     int result= 0,old_tree_list_idx;
     struct FilelistE *old_tree_list;

#ifdef NIX
     if(liste->flags[i]&64) {

       /* >>> check whether inode entry is valid */;

 continue;
     }
#endif

     /* check existence and unchanged state */
     /* no readbility check, age check, mistrust,
        tree: existence, type, size, crc */
     o->precheck_flag= 4|8|32|256|512|1024|2048;

     /* <<< for quick output test only : no crc test <<< */
     /* o->precheck_flag= 4|8|32|256|512|1024; */

     ret= Crawler_check_file(o,source_adr,target_adr,(dev_t) 0,&result,
                             (flag&1)|2);
     if(ret<=0)
       goto ex;
     if(result&2) { /* file not statable */
       if(flag&8192)
         goto invalidate_list_item;
 continue;
     }
     if(ret==2) { 
       /* reply says it is to be skipped. But here process it anyway */;
     }
     if(o->typ==2) {
       /* extra handling of directories */
       if(ret_flags>0)
         result= ret_flags&(8|16);
       else
         result= 8;
     } else {
       old_tree_list= liste;
       ret= Crawler_file_changecheck(o,source_adr,target_adr,
                                     &old_tree_list,&old_tree_list_idx,
                                     &result,(liste->flags[i]&32)|(flag&1)|2);
       if(ret<=0)
         goto ex;
     }
     if(result&8)  {
       if(flag&2048) {
         /* unchanged : set inode and eventually device number */
         Fileliste_set_inode(liste,i,o->inode,0);
         if((liste->typen[i]==2 || liste->typen[i]==0)
            && liste->subdirs[i]!=NULL) {
           liste->subdirs[i]->dir_device= o->dev;
           liste->subdirs[i]->dir_inode= o->inode;
           liste->subdirs[i]->dir_dev_ino_valid= 3;
         }
       }
     } else {
       if(flag&8192) {
invalidate_list_item:;
         Fileliste_defunct_item(liste, i, 0);
       }
     } 
   }

 }
 ret= 1;
 if(ret_flags>0)
   ret|= ret_flags;
ex:
 o->precheck_flag= precheck_flag_mark;
 if(!(flag&1)) {
   if(flag&(512|1024)) {
     if(o->vanishlist_fp!=NULL || o->vanishscript_fp!=NULL)
       Print_pacifier(o->vanishlist_count,"vanished file objects recorded",
                      1|2|4|8|16);
   } else if(flag&2048) {
     Print_pacifier(o->traverse_count,"tree nodes examined",1|2|4|8|16);
   } else if(flag&4096) {
     Print_pacifier(o->contentlist_count,"content list lines written",
                    1|2|4|8|16);
   }
 }
 return(ret);
}


int Crawler_decide_directory(o,adr,target,dir_size,
                             provisory_young,too_old_mark,young_exempts_mark,
                             collected_subresult,result,flag)
struct CrawleR *o;
char *adr;
char *target;
double dir_size;
struct FilelistE *provisory_young;
int too_old_mark;
int young_exempts_mark;
int collected_subresult;
int *result; /* see Crawler_do_tree */
int flag;
/*
 bit0= is_empty_dir
 bit1= already_in_old_tree
*/
{
 int ret,i,is_empty_dir,already_in_old_tree,self_in_l_too_old= 0;
 int new_young_exempts,number_of_young,l_max_exempts;
 double saved_bytes= 0.0;
 char subadr[4096],subtarget[4096],*subnpt,*subtpt;

 is_empty_dir= flag&1;
 already_in_old_tree= !!(flag&2);

 if(is_empty_dir) {
   if(already_in_old_tree) {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: === directory empty and known : %s\n",
                      adr);
     (*result)|= 8;
   } else {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: === directory empty and new : %s\n",adr);
   }
   ret= 1; goto ex;
 }
 new_young_exempts= o->young_exempts->anz - young_exempts_mark;
 number_of_young= provisory_young->anz + new_young_exempts;
 l_max_exempts= o->max_exempts - o->young_exempts->anz;

 if((o->precheck_flag&256) && !already_in_old_tree) {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,
             "agetest_debug: directory not found in old tree\n");
   collected_subresult&= ~(8|16);
 }
 if(!(collected_subresult&8)) {               /* everything was young enough */
   if((collected_subresult)&16) {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: accepting old files for efficiency\n");
     (*result)|= 16;
   }
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: DECISION: accept whole directory\n");
 } else if(number_of_young<=0) {     /* everything was too old  */
   Fileliste_truncate(o->too_old,too_old_mark,0);
   ret= Fileliste_append_prop(o->too_old,adr,target,2,dir_size,1);
   if(ret<=0)
     goto ex;
   self_in_l_too_old= 1;
   (*result)|= 8;
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,"agetest_debug: DECISION: exclude whole directory\n");
 } else if(number_of_young<l_max_exempts) {
   /* there is still room for local youngsters */
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,
             "agetest_debug: DECISION: exclude directory, add exempts\n");
   Fileadr_plus_slash(subadr,adr,&subnpt,0);
   Fileadr_plus_slash(subtarget,target,&subtpt,0);

   /* transfer items from provisory_young to young_exempts */
   for(i=0;i<provisory_young->anz;i++) {
     strcpy(subnpt,provisory_young->adressen[i]);
     strcpy(subtpt,provisory_young->adressen[i]);
     ret= Fileliste_append_prop(o->young_exempts,subadr,subtarget,
                                provisory_young->typen[i],
                                provisory_young->groessen[i],1);
     if(ret<=0)
       goto ex;
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: exempt added : %s\n",subadr);
   }
   Fileliste_truncate(o->too_old,too_old_mark,0);
   ret= Fileliste_append_prop(o->too_old,adr,target,2,dir_size,1);
   if(ret<=0)
     goto ex;
   (*result)|= 8;
   self_in_l_too_old= 1;
 } else {
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,
     "agetest_debug: DECISION: accept whole directory despite old files\n");
   Fileliste_truncate(o->too_old,too_old_mark,0);
   Fileliste_truncate(o->young_exempts,young_exempts_mark,0);
   (*result)&= ~8;
   (*result)|= 16;
 }
 if(o->too_old->anz-too_old_mark>0 && number_of_young>0 &&
    !((o->precheck_flag)&(1<<15))) {
   /* evaluate effectiveness and eventually prevent split */
   for(i= too_old_mark; i<o->too_old->anz; i++)
     if(o->too_old->groessen[i]>0)
       saved_bytes+= o->too_old->groessen[i];
   for(i=young_exempts_mark;i<o->young_exempts->anz;i++)
     if(o->young_exempts->groessen[i]>0)
       saved_bytes-= o->young_exempts->groessen[i];
   if(saved_bytes/((double) number_of_young)<o->min_saved_bytes_per_file) {
     if(Cdbackup_planer_agetest_debuG)
       fprintf(stderr,"agetest_debug: OVERRIDE DECISION: uneconomic split, accept whole directory\n");
     Fileliste_truncate(o->too_old,too_old_mark,0);
     Fileliste_truncate(o->young_exempts,young_exempts_mark,0);
     (*result)&= ~8;
     (*result)|= 16;
     self_in_l_too_old= 0;
   }
 }
 if((*result)&8) {
   /* directory was classified as too old */
   if(Cdbackup_planer_agetest_debuG)
     fprintf(stderr,
            "agetest_debug: === lists:  old= %d+%d   young_exempts= %d+%d\n",
              too_old_mark,o->too_old->anz-too_old_mark,
              young_exempts_mark,o->young_exempts->anz-young_exempts_mark);
   if(self_in_l_too_old)
     (*result)|= 64;
 }
 ret= 1;
ex:
 return(ret);
}


int Crawler_append_to_piece(o,adr,target,typ,size,flag)
struct CrawleR *o;
char *adr;
char *target;
int typ;
double size;
int flag;
{
 int ret;
 struct FilelistE *liste= NULL;

 ret= Fileliste_single(&liste,adr,typ,size,8);
 if(ret<=0)
   goto ex;
 if(Sregex_string(&(liste->targets[0]),target,0)<=0)
   {ret= -1; goto ex;}
 ret= Backup_stueck_append_file(o->current_piece,liste,0,0);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:;
 Fileliste_destroy(&liste,0);
 return(ret);
}


int Crawler_stack_to_piece(o,flag)
struct CrawleR *o;
int flag;
{
 int ret,i,j,j_start,stack_count= 0;
 struct FilelistE *liste;
 char subadr[4096],subtarget[4096],*subadr_pt,*subtarget_pt;
 double stack_bytes= 0.0;

 j_start= o->py_list_piecemark;
 for(i= o->py_stack_piecemark; i<o->py_stack->anz; i++) {
   liste= o->py_stack->subdirs[i];
   Fileadr_plus_slash(subadr,o->py_stack->adressen[i],&subadr_pt,0);
   Fileadr_plus_slash(subtarget,o->py_stack->targets[i],&subtarget_pt,0);
   for(j= j_start;j<liste->anz;j++) {
     strcpy(subadr_pt,liste->adressen[j]);
     strcpy(subtarget_pt,liste->adressen[j]);
     stack_bytes+= liste->groessen[j];
     stack_count++;
     ret= Crawler_append_to_piece(o,subadr,subtarget,liste->typen[j],
                                  liste->groessen[j],0);
     if(ret<=0)
       goto ex;
   }
   j_start= 0;
 }

#ifdef Cdbackup_planer_developpeR
 if(o->py_size!=stack_bytes) {
   Print_pacifier(0,"",1|16);
   fprintf(stderr,"developer message: byte diff= %.lf  count= %d : %d  ratio= ",
           stack_bytes-o->py_size,stack_count,o->py_count);
   if(stack_count>0)
     fprintf(stderr,"%.4lf\n",
             (stack_bytes-o->py_size)/((double) stack_count));
   else
     fprintf(stderr,"-\n");
 }
#endif /* Cdbackup_planer_developpeR */

 o->py_size= 0.0;
 o->py_count= 0;
 /* memorize stack piece limit */
 o->py_stack_piecemark= o->py_stack->anz-1;
 if(o->py_stack_piecemark>=0)
   o->py_list_piecemark= o->py_stack->subdirs[o->py_stack_piecemark]->anz;
 else
   o->py_list_piecemark= 0;
 ret= 1;
ex:;
 return(ret);
}


int Crawler_start_next_piece(o,flag)
struct CrawleR *o;
int flag;
{
 int ret;

 if(o->contentlist_fp!=NULL) {
   /* write content list of old piece */
   ret= Backup_stueck_content_list(o->current_piece,o->piece_count,
                                   o->contentlist_fp,&(o->contentlist_count),
                                   o->output_format_flag,1);
   if(ret<=0)
     goto ex;
 }

 /* start next piece */
 Print_pacifier(0,"",1|16);
 ret= Backup_plan_next_stueck(o->boss,&(o->current_piece),0);
 if(ret<=0)
   goto ex;
 o->piece_count++;
 ret= 1;
ex:;
 return(ret);
}


int Crawler_accept_provisory_item(o,adr,target,typ,size,flag)
struct CrawleR *o;
char *adr;
char *target;
int typ;
double size;
int flag;
/*
 bit0= this is a sub item, do only use to check wether piece is full
*/
{
 int ret,economic= 1;
 double bytes;
 struct FilelistE *liste;

 Backup_stueck_get_list(o->current_piece,&liste,0);
 Backup_stueck_get_bytes(o->current_piece,&bytes,0);

 if((o->max_size >= bytes+o->weight+o->py_size) 
    && (o->max_items >= liste->anz+1+o->py_count))
   goto append_to_piece;

 /* so this exceeds the size or item limits */

 /* >>> is stacked list economic ? */;

 if(economic) {
   ret= Crawler_stack_to_piece(o,0);
   if(ret<=0)
     goto ex;
 }
 ret= Crawler_start_next_piece(o,0);
 if(ret<=0)
   goto ex;

append_to_piece:;
 if(flag&1) {
   o->py_size+= o->weight;
   o->py_count++;
   {ret= 2; goto ex;}
 }
 ret= Crawler_append_to_piece(o,adr,target,typ,size,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);

       /* >>> FLAT_METHOD :*/
         /* >>> if this exceeds the size or item limits */
           /* >>> start next piece */;
           /* >>> if provisory piece part not economic : */;
             /* >>> transfer provisory part to new piece */;
           /* >>> memorize provisory piece limit */;
           /* >>> append to current piece */;
}


int Crawler_accept_provisory_dir(o,adr,target,size,result,
                                 start_piece,start_piece_idx,start_piece_size,
                                 start_py_stackmark,start_py_listmark,
                                 start_py_size,start_py_count,
                                 flag)
struct CrawleR *o;
char *adr;
char *target;
double size;
int result; /* bits see Crawler_do_tree */
struct Backup_stuecK *start_piece; /* backup piece position when */
int start_piece_idx;               /*  processing of directory was started */
double start_piece_size;           /*  backup piece size at that time */
int start_py_stackmark;            /* py_stack position */
int start_py_listmark;             /*  at that time */
double start_py_size;              /* pending file item in py_stack */
int start_py_count;                /*  at that time */
int flag;
/*
 bit0= this is a sub item
*/
{
 int ret;
 double piece_size;
 struct FilelistE *piece_list;

 Backup_stueck_get_list(o->current_piece,&piece_list,0);
 Backup_stueck_get_bytes(o->current_piece,&piece_size,0);
    
 if(result&(2|4|8|128)) {
   /* directory was not accepted as a whole */

   /* >>> ??? rather copy only own provisory list (last of stack) ? */
   ret= Crawler_stack_to_piece(o,0);
   if(ret<=0)
     goto ex;

 } else {
   /* directory was accepted */
   if(start_piece==o->current_piece) { /* still within same piece */
     /* roll back piece to memorized position */ 
     Fileliste_truncate(piece_list,start_piece_idx,0);
     Backup_stueck_set_bytes(o->current_piece,start_piece_size,0);
     o->py_stack_piecemark= start_py_stackmark;
     o->py_list_piecemark= start_py_listmark;
     if(!(flag&1)) { /* this is a top level directory */
       ret= Crawler_append_to_piece(o,adr,target,2,size,0);
       if(ret<=0)
         goto ex;
       o->py_size= 0.0;
       o->py_count= 0;
     } else {
       o->py_size= start_py_size+size;
       o->py_count= start_py_count+1;
     }
   } else { /* new piece was opened since directory was begun */
     ret= Crawler_stack_to_piece(o,0);
     if(ret<=0)
       goto ex;
     ret= Fileliste_append_prop(o->split_dirs,adr,target,2,size,1|128);
     if(ret<=0)
       goto ex;
   }
 }

#ifdef NIX
 if(!(flag&1)) {
   /* check wether this is a good opportunity to start next piece */
   Backup_stueck_get_bytes(o->current_piece,&piece_size,0);
   if(piece_size > o->max_size*o->min_size_fak) {
     ret= Crawler_start_next_piece(o,0);
     if(ret<=0)
       goto ex;
   }
 }
#endif /* NIX */


 ret= 1;
ex:;
 return(ret);

 /* >>> FLAT_METHOD :*/
 if(result&(2|4|8|128)) /* directory was not accepted as a whole */
   {ret= 1; goto ex;}
   
     /* >>> if still within same piece */
         /* >>> remove all newly appended items */;
         /* >>> append to current_piece */;
     /* >>> if new piece */;
         /* >>> discard directory and keep all appended items */;
       /* >>> ??? record directory in split_dirs */;
         /* ??? how to bring split_dirs onto first volume ? */
}


int Crawler_do_tree(o,adr,target,parent_dev,old_tree_list,result,tree_size,
                    depth,flag)
struct CrawleR *o;
char *adr;
char *target;
dev_t parent_dev;
struct FilelistE *old_tree_list;   /* eventual directory model in old_tree */
int *result;
/*
 bit0= ill name
 bit1= not readable
 bit2= mountpoint of unwanted type
 bit3= too old
 bit4= contains some old items
 bit5= too big
 bit6= item was added to zu_alt list
 bit7= excluded file
 bit8= type change versus old_tree
 bit9= has young timestamp
bit10= has been split by piece border
*/
double *tree_size;
int depth;
int flag;
/*
 bit0= recursion
 bit5= this is a -special_file item, retain *tree_size ,
       compute block device crc from content and not from stat
 bit7= parent_directory has old timestamp and all parents are old
*/
/* return:
 <=0 = error
   1 = acceptable file
   2 = file is to be skipped
*/
{
 int ret,subresult,old_tree_list_idx,is_impl_young,trustfully_old= 0;
 int young_mark,is_empty_dir,already_in_old_tree= 0;
 int young_exempts_mark,collected_subresult,l_max_exempts;
 int too_old_mark,too_big_mark;
 int start_idx= 0,py_stack_idx= -1,start_py_count= 0,is_dir= 0,do_compose;
 int start_py_stackmark= 0,start_py_listmark= 0;
 double dir_size= 0.0,subtree_size,start_size;
 double start_py_size= 0.0;
/* A71226 : switched from local function memory to malloc()
            for more recursion depth.
 char subadr[2*4096],subtarget[2*4096],*subnpt,*subtpt;
*/
 char *subadr= NULL,*subtarget= NULL,*subnpt,*subtpt;
 dev_t own_dev;
 struct DirseQ *dirseq= NULL;
 struct FilelistE *old_tree_sublist;
 struct FilelistE *provisory_young= NULL;
 struct Backup_stuecK *start_piece= NULL;

 *result= 0;

 ret= Cd_backup_planer_assert_depth(depth, adr, 0);
 if(ret<=0)
   {ret= -1; goto ex;}

 subadr= malloc(2*4096);
 subtarget= malloc(2*4096);
 if(subadr==NULL || subtarget==NULL) {
   fprintf(stderr, "Memory exhaustion in tree crawler\n");
   ret= -1; goto ex;
 }

 young_exempts_mark= o->young_exempts->anz;
 too_old_mark= o->too_old->anz;
 too_big_mark= o->too_big->anz;
 do_compose= !!(o->composition_mode&(1<<15));

 /* This must be the first activity with a file. 
    It establishes the property buffer */
 ret= Crawler_check_file(o,adr,target,parent_dev,result,flag&(1|128));
 if(ret<=0)
   goto ex;
 if(ret==2) {
   if(too_big_mark<o->too_big->anz) {
     /* 0497: determine the age validity of the oversized file because it
              might get split and included in parts */;
     subresult= 0;
     ret= Crawler_file_changecheck(o,adr,target,
                                   &old_tree_list,&old_tree_list_idx,
                                   &subresult,2|(flag&(1|32|128)));
     if(ret<=0)
       goto ex;
     if(subresult&8) {
       o->too_big->flags[too_big_mark]|= 4;
       (*result)|= 8;
     }
   }
   {ret= 2; goto ex;}
 }
 own_dev= o->dev;
 if(flag&32) {
   dir_size= o->weight= o->size= *tree_size;
 } else if(o->typ==2)
   dir_size= 0; /* directory size is counted by content */
 else
   dir_size= o->weight;
 
 if(!(flag&1))
   parent_dev= own_dev;

 ret= Crawler_write_completelists(o,flag&1);
 if(ret<=0)
   goto ex;

 if(adr[1]==0 && adr[0]=='/') {
   /* there is no root directory to be found in the tree */
   old_tree_sublist= old_tree_list;
   old_tree_list_idx= -2;
   already_in_old_tree= 1;
   goto examine_directory_content;
 }
 ret= Crawler_file_changecheck(o,adr,target,&old_tree_list,&old_tree_list_idx,
                               result,flag&(1|32|128));
 if(ret<=0)
   goto ex;
 old_tree_sublist= NULL;
 if(old_tree_list_idx>=0) { /* adr was found in o->old_tree */
   if(o->typ==2)
     old_tree_sublist= old_tree_list->subdirs[old_tree_list_idx];
   if(old_tree_list->typen[old_tree_list_idx]==0 ||
      old_tree_list->typen[old_tree_list_idx]==o->typ) {
     /* and type did not change */
     old_tree_list->flags[old_tree_list_idx]|= 128;
     already_in_old_tree= 1;

#ifdef Cdbackup_planer_developpeR
   } else {
     Print_pacifier(0,"",1|16);
     fprintf(stderr,"developer message : type change detected : %d -> %d\n",
             old_tree_list->typen[old_tree_list_idx],o->typ);
     fprintf(stderr,"                    %s\n",adr);
#endif /* Cdbackup_planer_developpeR */

   }
 } else {

   /* >>> ??? disable agetest for subsequent directories */;

 }
  
 if(o->typ!=2 || (flag&32)) {
   if(do_compose && !((*result)&(2|4|8|32|128))) { 
     ret= Crawler_accept_provisory_item(o,adr,target,o->typ,o->size,flag&1);
     if(ret<=0)
       goto ex;
   }
   {ret= 1; goto ex;}
 }

examine_directory_content:;
 is_dir= 1;
 if(Fileliste_new(&provisory_young,0,"",2)<=0)
   {ret= -1; goto ex;}
 if(do_compose) {
   struct FilelistE *list;
   /* memorize position in backup piece and provisory_young stack */;
   start_piece= o->current_piece;
   Backup_stueck_get_list(start_piece,&list,0);
   start_idx= list->anz;
   py_stack_idx= o->py_stack->anz;
   ret= Fileliste_append_prop(o->py_stack,adr,target,2,dir_size,1);
   if(ret<=0)
     goto ex;
   Fileliste_set_subdir(o->py_stack,py_stack_idx,provisory_young,0);
   Backup_stueck_get_bytes(start_piece,&start_size,0);
   start_py_stackmark= o->py_stack_piecemark;
   start_py_listmark= o->py_list_piecemark;
   start_py_size= o->py_size;
   start_py_count= o->py_count;
 }
 ret= Dirseq_new(&dirseq,adr,0);
 if(ret<0)
   goto ex;
 if(ret==0) {
   /* unreadable directory */
   (*result)|= 2;
   goto ex;
 }
 Fileadr_plus_slash(subadr,adr,&subnpt,0);
 Fileadr_plus_slash(subtarget,target,&subtpt,0);
 is_empty_dir= 1;
 collected_subresult= 0;
 /* trustfully_old= !((*result)&512); */
 trustfully_old= ((flag&128)||(o->precheck_flag&(1<<18)))&&!((*result)&512);
 
#ifdef Cdbackup_planer_developpeR
 /* <<< */
 { static char searchfor[4096]= {""};
   if(strcmp(adr,searchfor)==0) 
     fprintf(stderr,"\ndeveloper message : wanted file encountered\n");
 }
#endif /* Cdbackup_planer_developpeR */

 while(1) {
   ret= Dirseq_next_adr(dirseq,subnpt,1|32);
   if(ret<0)
 continue;
   if(ret==0)
 break;
   is_empty_dir= 0;
   strcpy(subtpt,subnpt);
   young_mark= o->young_exempts->anz;
   l_max_exempts= o->max_exempts-young_mark;
   ret= Crawler_do_tree(o,subadr,subtarget,own_dev,old_tree_sublist,
                        &subresult,&subtree_size,depth+1,
                        1|(trustfully_old<<7));
   if(ret<0)
     goto ex;
   if(ret==0)
 continue;
   if(ret==2) /* shall be skipped */
 continue;
   collected_subresult|= subresult;
   if(subtree_size>0.0)
     dir_size+= subtree_size;

   /* interpret subresult */;
   is_impl_young= 0;
   if(!(subresult&(8|(1<<10))))
     if(young_mark==o->young_exempts->anz)
       is_impl_young= 1;
   if(((o->precheck_flag)&8)||do_compose) {
     if(is_impl_young) {
       ret= Fileliste_append(provisory_young,subnpt,1|8);
       if(ret<=0)
         goto ex;
       Fileliste_set_groessen(provisory_young,-1,subtree_size,0);
     }
   }

   /* >>> if first item of directory caused a new piece and if it
          was included entirely in the new piece then re-adjust start_piece */;


#ifdef NIX
   /* <<< why do some directories get left out during composition */
   if(do_compose) {
     if(o->py_stack_piecemark>0) {
       if(o->py_stack->subdirs[o->py_stack_piecemark]==provisory_young &&
          o->py_list_piecemark>provisory_young->anz) {
         o->py_list_piecemark= provisory_young->anz;

#ifdef Cdbackup_planer_developpeR
         Print_pacifier(0,"",1|16);
         fprintf(stderr,
                 "developer message : directory gap with : result=%x\n",
                 subresult);
         fprintf(stderr,"                    %s\n",subadr);
#endif /* Cdbackup_planer_developpeR */

       }
     }
   }
#endif /* NIX */


 }
 ret= Crawler_write_completelists(o,(flag&1)|2);
 if(ret<=0)
   goto ex;

 if(Cdbackup_planer_agetest_debuG)
   fprintf(stderr,"agetest_debug: === directory content checked : %s\n",adr);

 if((o->precheck_flag)&8) {
   ret= Crawler_decide_directory(o,adr,target,dir_size,
                               provisory_young,too_old_mark,young_exempts_mark,
                               collected_subresult,result,
                               (!!is_empty_dir)|((!!already_in_old_tree)<<1));
   if(ret<=0)
     goto ex;
 }
 if(do_compose) {
   ret= Crawler_accept_provisory_dir(o,adr,target,dir_size,*result,
                                     start_piece,start_idx,start_size,
                                     start_py_stackmark,start_py_listmark,
                                     start_py_size,start_py_count,flag&1);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 if(!(flag&2))
   *tree_size= dir_size;
 if(subadr!=NULL)
   free(subadr);
 subadr= NULL;
 if(subtarget!=NULL)
   free(subtarget);
 subtarget= NULL;
 if(!is_dir) /* handle any non-directory cleanup before this line */
   return(ret);
 if(do_compose) {
   if(py_stack_idx>=0) {
     Fileliste_truncate(o->py_stack,py_stack_idx,(1<<10));
     if(o->py_stack_piecemark>=py_stack_idx) {
       if(py_stack_idx>0) {
         o->py_stack_piecemark= py_stack_idx-1;  /* parent's provisory_young */
         o->py_list_piecemark= o->py_stack->subdirs[py_stack_idx-1]->anz;
       } else {
         o->py_stack_piecemark= o->py_list_piecemark= 0;
       }
     }
   }
   Fileliste_truncate(o->too_old,0,0);
   Fileliste_truncate(o->young_exempts,0,0);
   if(start_piece!=o->current_piece)
     (*result)|= 1<<10; /* prevent inclusion in provisory_young */
 }
 Fileliste_destroy(&provisory_young,0);
 Dirseq_destroy(&dirseq,0);
 return(ret);
}


int Crawler_do_wood(o,flag)
struct CrawleR *o;
int flag;
{
 int ret,i,result,hflag,trustfully_old;
 char target[4096];
 double tree_size,size;
 struct FilelistE *list;

 list= o->start_list;
 /* initialize existence flag in o->old_tree */;
 if(o->old_tree!=NULL)
   Crawler_traverse_old_tree(o,o->old_tree,"","",256);

 for(i=0; i<list->anz; i++) {
   ret= Fileliste_target(list,i,target,2);
   if(ret<=0)
     {ret= -1; goto ex;}
   tree_size= 0.0;
   if(list->flags[i]&32) {
     if(list->groessen[i]<=0) {
       o->typ= Type_and_size(list->adressen[i],&size,8);
       if(o->typ<=0 || size<0) {
         fprintf(stderr,"--- Cannot determine size of  -special_file %s\n",
                        list->adressen[i]);
         if(o->problemlist_limit>=o->no_access->anz) {
           ret= Fileliste_append_prop(o->no_access,list->adressen[i],
                                      target,o->typ,0.0,1);
           if(ret<=0)
             goto ex;
         }
 continue;
       } else
         list->groessen[i]= size;
     }
     tree_size= list->groessen[i];
   }
   trustfully_old= 0;
                  /* agecheck shortcut is only worth if CRC check is enabled */
   if((o->precheck_flag&8) && (o->precheck_flag&2048)) {
     hflag= o->precheck_flag&(1<<14);
     if(o->precheck_flag&(1<<18))
       hflag|= 2;
     trustfully_old= Parent_is_trustfully_old(list->adressen[i],o->age_limit,
                                              hflag);
   }
   ret= Crawler_do_tree(o,list->adressen[i],target,
                        (dev_t) 0,o->old_tree,&result,&tree_size,0,
                        (list->flags[i]&32)|(trustfully_old<<7));
   if(ret<0)
     goto ex;
 }
 Crawler_pacifier(o,1|128);
 if(o->completelist_fp!=NULL)
   Print_pacifier(o->completelist_count,"complete list items recorded",
                  1|2|4|8);
 if(o->dirlist_fp!=NULL || o->dirscript_fp!=NULL)
   Print_pacifier(o->dirlist_count,"directory items recorded",1|2|4|8);

 /* report non-confirmed members from o->old_tree to
    vanishlist_fp , vanishscript_fp */
 hflag= 0;
 if((o->what_lists_flag)&8)
   hflag|= 512;
 if((o->what_lists_flag)&16)
   hflag|= 1024;
 if(o->old_tree!=NULL) {
   ret= Crawler_traverse_old_tree(o,o->old_tree,"","",
                                  ((o->what_lists_flag)&(8|16))<<6);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


/* ----------------- Crawler - end ------------- */


/* ----------------- Backup_stueck --------------- */


struct Backup_stuecK {               /* english : Backup_piecE */
 struct Backup_plaN *plan;

 double bytes;
 int max_items;
 double max_size;
 struct FilelistE *liste;            /* english : flist */
 int first_payload_item;
 int last_payload_item;
 struct Backup_stuecK *next,*prev;
};


int Backup_stueck_new(stueck,plan,prev,maxanz,max_size,main_dir,flag)
struct Backup_stuecK **stueck;
struct Backup_plaN *plan;
struct Backup_stuecK *prev;
int maxanz;
double max_size;
char *main_dir;                         /* for liste */
int flag;
{
 int l_maxanz;

 *stueck= TSOB_FELD(struct Backup_stuecK,1);
 if(*stueck==NULL)
   return(-1);
 (*stueck)->plan= plan;
 (*stueck)->bytes= 0;
 (*stueck)->max_items= maxanz;
 (*stueck)->max_size= max_size;
 if(maxanz<1000) l_maxanz= maxanz;
 else            l_maxanz= 1000;
 (*stueck)->first_payload_item= -1;
 (*stueck)->last_payload_item= -1;
 (*stueck)->liste= NULL;
 (*stueck)->prev= prev;
 (*stueck)->next= NULL;
 if(Fileliste_new(&((*stueck)->liste),l_maxanz,main_dir,2)<=0)
   return(-1);
 if(prev!=NULL)
   prev->next= *stueck;
 return(1); 
}


int Backup_stueck_destroy(stueck,flag)
struct Backup_stuecK **stueck;
int flag;
{
 Fileliste_destroy(&((*stueck)->liste),0);
 free((char *) (*stueck));
 *stueck= NULL;
 return(1);
}


int Backup_stueck_report_zeile(ziel,line,flag)
char *ziel;
char *line;
int flag;
/*
 bit0= do only count , do not append
*/
{
 int l;
 l= strlen(line);
 if(flag&1)
   return(l);
 strcat(ziel,line);
 return(l);
}


int Backup_stueck_append_file(stueck,liste,idx,flag)
struct Backup_stuecK *stueck;
struct FilelistE *liste;
int idx;
int flag;
/*
 bit0= do not append , make invalid
 bit1= this is added_by_scdbackup , not a payload item
*/
{
 int ret;
 struct FilelistE *ziel;

 if(!(flag&1)){
   ziel= stueck->liste;
   ret= Fileliste_copy_item(liste,idx,ziel,ziel->anz,0);
   if(ret<=0)
     return(ret);
   if(liste->groessen[idx]>0)
     stueck->bytes+= liste->groessen[idx];
   if(!(flag&2)) {
     if(stueck->first_payload_item<0)
       stueck->first_payload_item= ziel->anz-1;
     stueck->last_payload_item= ziel->anz-1;
   }
 }
 Fileliste_defunct_item(liste,idx,0);
 return(1);
}


int Backup_stueck_report(stueck,fp,cmd,volume,total_volumes,flag)
struct Backup_stuecK *stueck;
FILE *fp;
char *cmd;
int volume,total_volumes;
int flag;
/*
 bit0-7= output form: 0= Backup_plan_skript target=source \
                      not done here : 1= Backup_plan_stdout
                      2= Backup_plan_skript source \
                      3= Backup_plan_skript <<'+' source +
 bit8= only with output form 3 :
       with part_recs file list original address , partial address in original
       directory as well as real source address
*/
{
 int i,expl_replace;
 char *cpt,line[3*4096+100],target[2*4096],txs_buf[4096*6],tempname[4096];
 char src_adr[4096*2],*part_prefix,*part_suffix;
 struct FilelistE *liste;

 liste= stueck->liste;
 expl_replace= Backup_plan_get_precheck_expl_replace(stueck->plan,0);
 Backup_plan_get_split_file_fixes(stueck->plan,&part_prefix,&part_suffix,0);

 sprintf(line,"%s",cmd);
 if (volume >= 0) 
   sprintf(line+strlen(line)," -v %d",volume);
 if (total_volumes >= 0) 
   sprintf(line+strlen(line)," -t %d",total_volumes);
 if((flag&255)==3) {
   strcat(line," <<'+'\n");
 } else {
   strcat(line," \\\n");
 }
 fwrite(line,strlen(line),1,fp);
 
 for(i=0;i<liste->anz;i++){
   line[0]= 0;
   cpt= liste->adressen[i];
   if(cpt[0]=='.'){
     if(cpt[1]=='/') cpt+= 2;
   }
   if(liste->part_recs!=NULL) {
     if((flag&255)==3 && (flag&(1<<8))) {
       sprintf(line,"%s\n",cpt);
       fwrite(line,strlen(line),1,fp);
       if(Filelist_part_rec_make_tempname(liste->part_recs[i],
                           part_prefix,part_suffix,liste->adressen[i],
                           tempname,liste->impl_igno,(!!expl_replace)<<2)>0) {
         sprintf(line,"%s\n",tempname);
         fwrite(line,strlen(line),1,fp);
       }
       cpt= "";
     }
     if(Filelist_part_rec_make_tempname(liste->part_recs[i],
                                  part_prefix,part_suffix,
                                  liste->adressen[i],tempname,liste->impl_igno,
                                  1|2|((!!expl_replace)<<2))>0) {
       Backup_plan_get_split_file_dir(stueck->plan,&cpt,0);
       if(cpt[0]!=0 && cpt[strlen(cpt)-1]=='/')
         sprintf(src_adr,"%s/%s",cpt,tempname);
       else
         sprintf(src_adr,"%s/%s",cpt,tempname);
       cpt= src_adr;
     }
   }
   if((flag&255)==2) {
     sprintf(line,"%s \\\n",Text_shellsafe(cpt,txs_buf,0));
   } else if((flag&255)==3) {
     if(cpt[0]!=0)
       sprintf(line,"%s\n",cpt);
     else
 continue;
   } else {
     if(Fileliste_target(liste,i,target,0)<=0)
 continue;
     sprintf(line+strlen(line),"%s=",Text_shellsafe(target,txs_buf,0));
     sprintf(line+strlen(line),"%s \\\n",Text_shellsafe(cpt,txs_buf,0));
   }
   fwrite(line,strlen(line),1,fp);
 }
 if((flag&255)==3) {
   sprintf(line,"+\n");
   fwrite(line,strlen(line),1,fp);
 } else {
   sprintf(line,"\n");
   fwrite(line,strlen(line),1,fp);
 }
 sprintf(line,"\ncmd_return=$?\n");
 fwrite(line,strlen(line),1,fp);

 return(1);
}


int Backup_stueck_report_pl(stueck,fp,verify_source_filter,
                            cmd,volume,total_volumes,
                            max_items_per_tile,path_list,flag)
struct Backup_stuecK *stueck;
FILE *fp;
char *verify_source_filter,*cmd;
int volume,total_volumes;
int max_items_per_tile;
char *path_list;
int flag;
/*
 bit0-7= output form:
         xxx 0= Backup_plan_skript target=source \
         xxx 1= Backup_plan_stdout
         xxx 2= Backup_plan_skript source \
         xxx 3= Backup_plan_skript <<'+' source +
         xxx 4= Backup_plan_stdout(super_raw)
         5= Backup_plan_skript cat target=source >path_list
                               cmd -path-list path_list target1=source1
         6= Backup_plan_skript cat source >path_list
                               cmd -path-list path_list
         7= like 6 plus -dir-path-list of plan->split_dirs
         8= ( cat <<+ 'target'='source' + ) | \
                        command -quoted-path-list -
*/
{
 int i,first_item= -1,verify_source= 1,pass= 1,expl_replace, output_form;
 char *cpt,target[2*4096],txs_buf[4096],*first_cpt= NULL;
 char tempname[4096],src_adr[4096*2],*pipe_target,dpl_buf[4096];
 char *part_prefix,*part_suffix;
 struct FilelistE *liste;
#ifdef Cdbackup_planer_xorriso_cut_ouT
 /* CUT_OUT */

 int partrecs= 0, partno, partcount;
 char sfe[5*4096];
 /* >>> A80315 : this should better become off_t */
 double start_byte,byte_count;
#endif

 output_form= (flag&255);
 Backup_plan_get_split_file_fixes(stueck->plan,&part_prefix,&part_suffix,0);
 expl_replace= Backup_plan_get_precheck_expl_replace(stueck->plan,0);
 if(strcmp(path_list, "-")==0 && (output_form==5 || output_form==8)) {
   strcpy(txs_buf, path_list);
 } else {
   Text_shellsafe(path_list,txs_buf,0);
   fprintf(fp,"test -e %s && rm %s\n",txs_buf,txs_buf);
 }
 sprintf(tempname,"%s_dirs",path_list);
 Text_shellsafe(tempname,dpl_buf,0);
 if(output_form==7)
   fprintf(fp,"test -e %s && rm %s\n",dpl_buf,dpl_buf);

 for(pass= (output_form!=7); pass<2; pass++) {

   if(pass==0) {
     Backup_plan_get_split_dirs(stueck->plan,&liste,0);
     pipe_target= dpl_buf;
   } else if(pass==1) {
     liste= stueck->liste;
     pipe_target= txs_buf;
   } else 
 break;
   fprintf(fp,"( cat <<'+'\n");

   for(i= ((output_form==7) && pass>0);i<liste->anz;i++) {

     cpt= liste->adressen[i];

     if(pass==0) /* skip if no child of cpt is in the volume */
       if(Fileliste_search_file(stueck->liste,cpt,1|4)<0) 
   continue;

     if(cpt[0]=='.'){
       if(cpt[1]=='/') cpt+= 2;
     }
     if(liste->part_recs!=NULL) {
       if(Filelist_part_rec_make_tempname(liste->part_recs[i],part_prefix,
                      part_suffix,liste->adressen[i],tempname,liste->impl_igno,
                      1|2|((!!expl_replace)<<2))>0) {
         Backup_plan_get_split_file_dir(stueck->plan,&cpt,0);

#ifdef Cdbackup_planer_xorriso_cut_ouT
         /* CUT_OUT */

         if((output_form==5 || output_form==8) &&
            strcmp(cpt,"-xorriso_cut_out")==0) {
           if(pass>0)
             partrecs++;
   continue;
         }
#endif

          sprintf(src_adr,"%s/%s",cpt,tempname);
          cpt= src_adr;
       }
     }
  
     if(output_form==5) {
       if(Fileliste_target(liste,i,target,0)<=0)
   continue;
       if(first_item==-1) {
         first_item= i;
         first_cpt= cpt;
         /* this item will be used as command line seed */
   continue;
       }

       /* >>> 0520 : do not put addresses with newlines into path-lists */;

       fprintf(fp,"%s=%s\n",target,cpt);
     } else if(output_form==8) {
       if(Fileliste_target(liste,i,target,0)<=0)
   continue;
       fprintf(fp,"%s=",Text_shellsafe(target,sfe,0));
       fprintf(fp,"%s\n",Text_shellsafe(cpt, sfe, 0));
     } else {                              /* if(output_form==6 || ...==7 ) */

       /* >>> 0520 : do not put addresses with newlines into path-lists */;

       fprintf(fp,"%s\n",cpt);
     }
     if(((i+1)%max_items_per_tile)==0)
       fprintf(fp,"+\ncat <<'+'\n");
   }

   if(verify_source_filter==NULL) 
     verify_source= 0;
   else if(strlen(verify_source_filter)==0) 
     verify_source= 0;
   if(strcmp(pipe_target,"-")==0 && (output_form==5 || output_form==8)) {
     if(verify_source && output_form==5) {
       fprintf(fp,"+\n) | \\\n");
       fprintf(fp,"%s \\\n",verify_source_filter);
       fprintf(fp,"| \\\n");
     } else { 
       fprintf(fp,"+\n) | \\\n");
     }
   } else {
     if(verify_source && output_form==5) {
       fprintf(fp,"+\n) | \\\n");
       fprintf(fp,"%s \\\n",verify_source_filter);
       fprintf(fp,"> %s\n",pipe_target);
     } else { 
       fprintf(fp,"+\n) >%s\n",pipe_target);
     }
   }

 } /* next pass */

 fprintf(fp,"%s \\\n",cmd);
 if(output_form==7) {
   if(stueck->liste->anz>0)
     fprintf(fp,"%s \\\n",stueck->liste->adressen[0]);
   fprintf(fp,"-dir-path-list %s \\\n",dpl_buf);
 }
 if(output_form==8)
   fprintf(fp,"--quoted_path_list %s \\\n",txs_buf);
 else
   fprintf(fp,"-path-list %s \\\n",txs_buf);
 if(output_form==5 || output_form==8) {
   if(first_item>=0) {
     Fileliste_target(liste,first_item,target,0);
     fprintf(fp,"%s=",Text_shellsafe(target,sfe,0));
     fprintf(fp,"%s \\\n",Text_shellsafe(first_cpt, sfe, 0));
   }

#ifdef Cdbackup_planer_xorriso_cut_ouT
   /* CUT_OUT */

   if((output_form==5 || output_form==8) && partrecs>0) {
     fprintf(fp,"-- \\\n",target,first_cpt); /* end -as mkisofs emulation */

     for(i= 0;i<liste->anz;i++) {
       if(liste->part_recs[i]==NULL)
     continue;
       cpt= liste->adressen[i];
       if(cpt[0]=='.')
         if(cpt[1]=='/')
           cpt+= 2;

       Filelist_part_rec_get_rec(liste->part_recs[i],&partno,&partcount,
                               &start_byte,&byte_count,0);

       /* list target was built by Filelist_part_rec_make_tempname() */
       if(Fileliste_target(liste,i,target,0)<=0)
   continue;

       fprintf(fp,"-cut_out %s %.f %.f ",
               Text_shellsafe(cpt, sfe, 0), start_byte, byte_count);
       fprintf(fp,"%s \\\n",Text_shellsafe(target, sfe, 0));

     }
   }
#endif /* Cdbackup_planer_xorriso_cut_ouT */

 }

 /* >>> 0520: write items with newlines as arguments */;

 fprintf(fp,"\ncmd_return=$?\n");
 fprintf(fp,"\n");
 if(output_form==7)
   fprintf(fp,"rm %s\n",dpl_buf);
 if(strcmp(txs_buf, "-")!=0 || (output_form!=5 && output_form!=8))
   fprintf(fp,"rm %s\n",txs_buf);
 return(1);
}


int Backup_stueck_content_list(stueck,piece_number,
                               fp,counter,list_output_format,flag)
struct Backup_stuecK *stueck;
int piece_number;
FILE *fp;
int *counter;
int list_output_format;
int flag;
/*
 bit0= announce writing of list
*/
{
 int ret,j;

 if(flag&1) {
   Print_pacifier(0,"",1|16);
   fprintf(stderr,"writing content list of volume %d\n",piece_number);
 }
 fprintf(fp,"#=# volume %d\n",piece_number);
 for(j=0;j<stueck->liste->anz;j++) {
   ret= Fileliste_leanfind(stueck->liste,j,dev_dummy,fp,
                           NULL,counter,0,
                           2|16|list_output_format);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Backup_stueck_get_list(o,list_pt,flag)
struct Backup_stuecK *o;
struct FilelistE **list_pt;
int flag;
{
 (*list_pt)= o->liste;
 return(1);
}


int Backup_stueck_set_bytes(o,bytes,flag)
struct Backup_stuecK *o;
double bytes;
int flag;
{
 o->bytes= bytes;
 return(1);
}


int Backup_stueck_get_bytes(o,bytes,flag)
struct Backup_stuecK *o;
double *bytes;
int flag;
{
 *bytes= o->bytes;
 return(1);
}


/* ----------------- Backup_stueck - end -------- */


/* ----------------- Backup_plan ----------------- */


#define Backup_plan_max_size_stackL 10


struct Backup_plaN {

 char backup_id[81];
 Sregex_md5_ctX *backup_id_ctx;

 /* Task */

 int max_items;
 int max_items_per_tile;                         /* for output_form==5 */
 char path_list[4096];
 double max_size;
 double default_max_size;
 int save_stack_media;            /* 1= after trying to fill max_size from
                                        stack : do not use max_size_stack media 
                                        if default_max_size would suffice */
 double max_size_stack[Backup_plan_max_size_stackL];
 int max_size_counts[Backup_plan_max_size_stackL];
 int max_size_stackpt;
 double min_size_fak;
 int scattering;
 int composition_mode; /*
     bit0-14:  0=largest item first
               1=alphabetically lowest target name first
               2=alphabetically lowest source name first
               3=oldest item first
     bit15= crawler builtin composition
                        */ 
 char cmd[4096];
 char verify_source_filter[4096];
 struct FilelistE *startliste;            /* english : startlist */
 struct FilelistE *auslassen;             /* english : excludelist */
 int auslassen_nicht_startliste;          /* english : exclude_not_startlist */
 char dir_size_cmd[4096];
 char target_inp_trenner;                 /* english : target_inp_separator */
 char shell_to_use[4096];
 char bdir[4096];
 int result_filetypes;   /* bit0= issue files of any  type
                        bit1-24 : particular types (see also FilelistE)
                            bit1= data files
                            bit2= directories
                            bit3= symbolic links
                            bit6= block devices 

                            bit30= unclassified types (e.g. AF_UNIX sockets)
                            bit31= redirect explicitely given symlinks
                          */

#ifdef Cd_backup_planer_with_regeX
 char **not_pattern_expr;
 regex_t **not_pattern_re;
 int not_pattern_count;
 regmatch_t not_pattern_match[1];
#endif

 char hauptdir[4096];                     /* english : maindir */
 int hauptdir_auto;                       /* english : maindir_auto */

 int auskunft_an;                         /* english : info_enabled */
 int info_short;
 int info_short_on_level;
 struct FilelistE *auskunft_name;         /* english : info_name */
 double auskunft_groesse;                 /* english : info_size */
 int auskunft_include;                    /* english : info_include */
 int auskunft_individuell;                /* english : info_individual */
 int auskunft_skript_root;                /* english : info_skript_root */

 char skript_status_file[4096];           /* english : script_status_file */
 char prompt_skript[4096];                /* english : prompt_script */

 char conf_dir[4096];
 int level;         /* -2:current -1:next >=0:level number */
 int configuration; /* 0=ignore , 1=create , 2=use , 3=update */

 int list_output_format;
    /* bit0=adr
       bit1=type
       bit2=size
       bit3=crc 
       bit4=MD5 rather than CRC32
       bit5=inode
       bit7=mtime
       bit8=ctime
 <<<   bit9=was: dirscript_size  (never published)
      bit10=on_level_crc
      bit11=on_level_md5
      bit12=on_level_inode
    */
 char tmp_dir[4096];

 int use_crawler;

 char complete_list_adr[4096];
 char append_complete_target[4096];
 char dir_list_adr[4096];
 char dir_list_target[4096];
 char dir_script_adr[4096];
 char dir_script_target[4096];
 char content_list_adr[4096];
 char append_content_target[4096];
 char facl_list_adr[4096];
 char facl_list_target[4096];
 char fattr_list_adr[4096];
 char fattr_list_target[4096];
 char faclsrc_list_adr[4096];
 char faclsrc_list_target[4096];
 char fattrsrc_list_adr[4096];
 char fattrsrc_list_target[4096];

 struct FilelistE *old_tree;
 char old_complete_list_adr[4096];
 char old_vanish_list_adr[4096];
 char vanish_list_adr[4096];
 char vanish_list_target[4096];
 char vanish_script_adr[4096];
 char vanish_script_target[4096];
 time_t start_time;

 double pre_checksum_size;
 double checksum_pad_size;
 double block_checksum_size;

 int checksum_scan_mode; /*  bit0= scan on 
                             bit1= end on first match
                             bit2= do not load recordlist
                             bit3= print list of found valid tags
                             bit4= only load tags with block checksum list 
                             bit5= print pacifier messages
                          */
 char checksum_scan_source[4096];
 double checksum_scan_skip; /* if >0 skip number of bytes and count as 0s */
 char blocklist_fileadr[4096];

 int split_large_files;
 char split_file_dir[4096];
 char split_file_prefix[4096];
 char split_file_suffix[4096];
 double split_file_size;
 int split_count;
 char cut_file_cmd[4096];
 char split_list_adr[4096];
 char split_list_target[4096];

 /* A number to be added to all volume numbers which get exposed
    to the user or put in by the user.
 */
 int volume_number_offset;

 /* pacifieres and dialog */
 int compact_num;             /* wether to display form "512.8k" or "525107" */
 int compose_linecount;

 /* Result */

 int anz_stuecke;                         /* english : piece_count */
 struct Backup_stuecK *stuecke;           /* english : pieces */
 struct Backup_stuecK *aktstueck;         /* english : current_piece */

 int output_form; /* 0= Backup_plan_skript target=source \
                     1= Backup_plan_stdout
                     2= Backup_plan_skript source \
                     3= Backup_plan_skript <<'+' source +
                     4= Backup_plan_stdout(super_raw)
                     5= Backup_plan_skript 
                        ( cat <<'+' target=source + ) >listfile
                        command -path-list listfile
                     6= -form_path_list_afio_old
                     7= -form_path_list_afio
                     8= -form_quoted_list_xorriso
                        Backup_plan_skript
                        ( cat <<+ 'target'='source' + ) | \
                        command -quoted-path-list -
                   */
 int collision_test; /* 0=do not test
                        1=report collisions 
                      */
 int collision_avoid; /* 0=leave collisions as they are
                         1=change target in case of collision
                         2=exit in case of collision
                       */
 int precheck;        /* 0=no check of file addresses
                           (may be overridden by precheck_xdev)
                         1=check file addresses
                         2=end program after check
                       */
 int precheck_expl;   /* 0=warn only
                         1=append impl_igno addresses to start list 
                         2=abort
                       */
 int precheck_expl_replace;  /*
                         0=keep problem characters in target address
                         1=replace by _XX_ (XX=hex ASCII Code)
                       */
 int precheck_unread; /*
                        -1=do not check for readability
                         0=warn only
                         1=append unreadable addresses to exclusion list
                         2=abort    
                       */
 int precheck_xdev;   /* 0= no special handling of mount points
                         1= add mount points to exclusion list
                         2= add remote mount points (e.g. NFS) to exclusion list
                       */
 int precheck_agetest; /* bit0= enable age test
                          bit5= with bit8 : mistrust directory timestamps
                          bit6= ignore timestamps
                          bit8= check for existence in old_tree
                          bit9= check type in old_tree
                         bit10= check size in old_tree
                         bit11= check crc in old_tree
                         bit12= reserved
                         bit14= ignore ctime
                         bit15= do not evaluate economy of directory splits
                         bit16= allow inode check as replacement of crc check
                         bit17= with bit16: do not compare device numbers
                         bit18= old trust (expect no directory replacement move)
                       */
 int precheck_sizetest; /* bit0= apply exclude_size test to single files 
                           bit1= eventually test size of link target 
                       */
 double exclude_size;

 int newline_adr_rule; /* How to deal with file addresses which contain
                          newline characters.
                          0= abort on sight
                          1= exclude them by -not
                          >>> 2= do not name them explicitely to formatters
                          >>> 3= escape them as in scdbackup lists
                          4= flatly use them and hope for formatter brain
                        */

 int problemlist_limit; /* maximum number of problem files before abort */

 int autom_excl_counter; /* counts the automatically excluded items */

 int volume_info;     /* 1=Pass current and total volume numbers to backup
                           script:  -v current_volume -t total_volumes 
                       */

 time_t precheck_age;
 int precheck_max_exempt;

 int prevent_doubles;  /* 1= call Backup_plan_is_planned() before appending
                             a result item */
 
 struct FilelistE *uebergrosse;           /* english : oversized */

 struct FilelistE *split_dirs; /* directories which got split by _compose() */

};


int Backup_plan_init(plan,max_items,max_size,cmd,flag)
struct Backup_plaN *plan;
int max_items;
double max_size;
char *cmd;
int flag;
{
 int hflag,ret;
 char id_seed[4096];

 strcpy(plan->backup_id,"00000000000000000000000000000000");
 plan->backup_id_ctx= NULL;
 plan->max_items= max_items;
 plan->max_items_per_tile= 100;
 plan->path_list[0]= 0;
 plan->default_max_size= max_size;
 plan->save_stack_media= 1;
 plan->max_size= max_size;
 plan->max_size_stackpt= -1;
 plan->min_size_fak= 0.8;
 plan->scattering= 0;
 plan->composition_mode= 0;
 strcpy(plan->cmd,cmd);
 plan->verify_source_filter[0]= 0;
 plan->startliste= NULL;
 plan->auslassen= NULL;
 plan->auslassen_nicht_startliste= 0;
 strcpy(plan->dir_size_cmd,"du -sk");
 plan->target_inp_trenner= Fileliste_target_inp_trenner;
 strcpy(plan->shell_to_use,"/bin/sh");
 plan->bdir[0]= 0;
 plan->result_filetypes= (1<<1)|(1<<2)|(1<<31);
#ifdef Cd_backup_planer_with_regeX
 plan->not_pattern_expr= NULL;
 plan->not_pattern_re= NULL;
 plan->not_pattern_count= 0;
 /* how to initialize this: plan->not_pattern_match[0]; */
#endif
 plan->hauptdir[0]= 0;
 plan->hauptdir_auto= 0;
 plan->auskunft_an= 1;
 plan->info_short= 0;
 plan->info_short_on_level= 0;
 plan->auskunft_name= NULL;
 plan->auskunft_groesse= 102400.0;
 plan->auskunft_include= 0;
 plan->auskunft_individuell= 1;
 plan->auskunft_skript_root= 0;
 plan->skript_status_file[0]= 0;
 plan->prompt_skript[0]= 0;
 plan->conf_dir[0]= 0;
 plan->level= -1;
 plan->configuration= 0;
 plan->list_output_format= 1|2|4|(1<<12);    /* adr:type:size:on_level_inode */
                                      /* was: 1|2|4|32;  adr:type:size:inode */
 strcpy(plan->tmp_dir,"/tmp");

#ifdef Crawler_ruleS
 plan->use_crawler= 1;
#else
 plan->use_crawler= 1; /* this may be set to 0 as it was before */
#endif

 plan->complete_list_adr[0]= 0;
 plan->append_complete_target[0]= 0;
 plan->dir_list_adr[0]= 0;
 plan->dir_list_target[0]= 0;
 plan->dir_script_adr[0]= 0;
 plan->dir_script_target[0]= 0;
 plan->content_list_adr[0]= 0;
 plan->append_content_target[0]= 0;
 plan->facl_list_adr[0]= 0;
 plan->facl_list_target[0]= 0;
 plan->fattr_list_adr[0]= 0;
 plan->fattr_list_target[0]= 0;
 plan->faclsrc_list_adr[0]= 0;
 plan->faclsrc_list_target[0]= 0;
 plan->fattrsrc_list_adr[0]= 0;
 plan->fattrsrc_list_target[0]= 0;
 plan->old_tree= NULL;
 plan->old_complete_list_adr[0]= 0;
 plan->old_vanish_list_adr[0]= 0;
 plan->vanish_list_adr[0]= 0;
 plan->vanish_list_target[0]= 0;
 plan->vanish_script_adr[0]= 0;
 plan->vanish_script_target[0]= 0;
 plan->start_time= time(0);
 plan->pre_checksum_size= 0;                    /* 1024.0*1024.0; */
 plan->checksum_pad_size= 0;
 plan->block_checksum_size= 0;
 plan->checksum_scan_mode= 0;
 plan->checksum_scan_source[0]= 0;
 plan->checksum_scan_skip= 0;
 plan->blocklist_fileadr[0]= 0;
 plan->split_large_files= 0;
 plan->split_file_dir[0]= 0;
 plan->split_file_prefix[0]= 0;
 plan->split_file_suffix[0]= 0;
 plan->split_file_size= -1.0;
 plan->split_count= 0;
 strcpy(plan->cut_file_cmd,"");
 plan->compact_num= 1;
 plan->compose_linecount= 0;
 plan->split_list_adr[0]= 0;
 plan->split_list_target[0]= 0;
 plan->volume_number_offset= 0;

 plan->anz_stuecke= 0;
 plan->stuecke= NULL;
 plan->aktstueck= plan->stuecke;
 plan->output_form= 0;
 plan->collision_test= 1;
 plan->collision_avoid= 1;
 plan->precheck= 0;
 plan->precheck_expl= 0;
 plan->precheck_expl_replace= 0;
 plan->precheck_unread= -1;
 plan->precheck_xdev= 0;
                                  /* -changetest_options type:size:inode:off */
 plan->precheck_agetest= 512|1024|(1<<16);  /* |256 = adr */
 plan->precheck_age= 0;
 plan->precheck_sizetest= 0;
 plan->exclude_size= -1.0;
 plan->newline_adr_rule= 1;
 plan->problemlist_limit= 10000;
 plan->autom_excl_counter= 0;
 plan->volume_info= 0;
 plan->precheck_max_exempt= -1;
 plan->prevent_doubles= 0;
 plan->uebergrosse= NULL;
 plan->split_dirs= NULL;

 /* try to start with a random backup id */
 /* eventually the program arguments will be hashed in, too */
 sprintf(id_seed,"%lf",Sfile_microtime(0));
 ret= Sregex_md5(&(plan->backup_id_ctx),id_seed,strlen(id_seed),NULL,1);
 if(ret<=0)
   goto failed;
 sprintf(id_seed,"%s,%d,%d",PROG_VERSION,getuid(),getpid());
 Sregex_md5(&(plan->backup_id_ctx),id_seed,strlen(id_seed),NULL,0);
 if(gethostname(id_seed,sizeof(id_seed))!=-1)
   Sregex_md5(&(plan->backup_id_ctx),id_seed,sizeof(id_seed),NULL,0);
       /* sizeof(id_seed) is not an error. See man hostname: overflow w/o 0 */

 if(Cdbp_compose_keep_sublevels>0)
   hflag= 256|512|2048|(1<<13);                         /* will need subdirs */
 else
   hflag= 2;                                         /* will need no subdirs */
 if(Fileliste_new(&(plan->startliste),0,"",hflag)<=0)
   goto failed;
 if(Fileliste_new(&(plan->auslassen),0,"",2)<=0)
   goto failed;
 if(Fileliste_new(&(plan->uebergrosse),0,"",2)<=0)
   goto failed;
 if(Fileliste_new(&(plan->split_dirs),0,"",2)<=0)
   goto failed;
 if(Fileliste_single(&(plan->auskunft_name),"",0,0.0,16)<=0)
   goto failed;

 Fileliste_set_sperrliste(plan->startliste,plan->auslassen,0);

 return(1);
failed:;
 Backup_plan_cleanup(plan,0);
 return(-1);
}


int Backup_plan_cleanup(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 int i;
 struct Backup_stuecK *akt,*next;

 akt= plan->stuecke;
 while(akt!=NULL) {
   next= akt->next;
   free((char *) akt);
   akt= next;
 }
 plan->anz_stuecke= 0;
 plan->stuecke= NULL;
 plan->aktstueck= NULL;
 Sregex_md5(&(plan->backup_id_ctx),"",0,NULL,1<<15);
 Fileliste_destroy(&(plan->uebergrosse),0);
 Fileliste_destroy(&(plan->auskunft_name),0);
 Fileliste_destroy(&(plan->auslassen),0);
 Fileliste_destroy(&(plan->startliste),0);
 Fileliste_destroy(&(plan->old_tree),0);
#ifdef Cd_backup_planer_with_regeX
 if(plan->not_pattern_expr!=NULL) {
   for(i= 0; i<plan->not_pattern_count; i++)
     free(plan->not_pattern_expr[i]);
   free(plan->not_pattern_expr);
   plan->not_pattern_expr= NULL;
 }
 if(plan->not_pattern_re!=NULL) {
   for(i= 0; i<plan->not_pattern_count; i++)
     regfree(plan->not_pattern_re[i]);
   free(plan->not_pattern_re);
   plan->not_pattern_re= NULL;
 }
 plan->not_pattern_count= 0;
#endif /* Cd_backup_planer_with_regeX */
 return(1);
}


int Backup_plan_get_split_file_dir(plan,dir,flag)
struct Backup_plaN *plan;
char **dir;
int flag;
{
 *dir= plan->split_file_dir;
 return(1);
}


int Backup_plan_get_split_file_fixes(plan,prefix,suffix,flag)
struct Backup_plaN *plan;
char **prefix;
char **suffix;
int flag;
{
 *prefix= plan->split_file_prefix;
 *suffix= plan->split_file_suffix;
 return(1);
}


int Backup_plan_get_split_dirs(plan,split_dirs,flag)
struct Backup_plaN *plan;
struct FilelistE **split_dirs;
int flag;
{
 (*split_dirs)= plan->split_dirs;
 return(1);
}


int Backup_plan_get_precheck_expl_replace(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 return(plan->precheck_expl_replace);
}


int Backup_plan_get_newline_adr_rule(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 return(plan->newline_adr_rule);
}


int Backup_plan_is_planned(plan,liste,idx,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int idx;
int flag;
{
 struct Backup_stuecK *stueck;
 struct FilelistE *sl;
 char target[4096],s_target[4096];
 int i,j;

 if(Fileliste_target(liste,idx,target,2)<=0)
   return(-1);
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   sl= stueck->liste;
   for(j=0;j<sl->anz;j++){
     if(strcmp(liste->adressen[idx],sl->adressen[j])!=0)
   continue;
     if(Fileliste_target(sl,j,s_target,2)<=0)
   continue;
     if(strcmp(target,s_target)==0) {
       if(liste->part_recs!=NULL || sl->part_recs!=NULL) {
         if(liste->part_recs==NULL || sl->part_recs==NULL)
   continue;
         if(Filelist_part_rec_is_equal(liste->part_recs[idx],sl->part_recs[j],
                                       4|8)==0)
   continue;
       }
       return(1);
     }
   }
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 return(0);
}


int Backup_plan_split(plan,liste,idx,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int idx;
int flag;
/*
 bit15= do not pacify
*/
{
 struct FilelistE *sub=NULL;
 int i,ret,subloeschen= 0,il,hflag;
 char target[4096];
 int *linecount;

 linecount= &(plan->compose_linecount);
 Print_pacifier(0,"",1|16); /* clear pacifier line */
 fprintf(stderr,"... splitting %s\n",liste->adressen[idx]);
 Fileliste_target(liste,idx,target,2);
 /* add to  plan->split_dirs */;
 if(strcmp(liste->adressen[idx],"/")!=0) {
   ret= Fileliste_append(plan->split_dirs,liste->adressen[idx],1|8|128);
   if(ret<=0) 
     return(ret);
   if(ret==1) 
     Fileliste_set_target(plan->split_dirs,0,target,1|2); /* (with auto idx) */
 }

 /* read file list from directory and process it */
 if(liste->subdirs!=NULL) {
   sub= liste->subdirs[idx];
 }
 if(sub==NULL){
   hflag= Fileliste_creation_flag(liste,0);
   if(Fileliste_new(&sub,0,liste->main_dir,hflag)<=0)
     return(-1);
   if(Fileliste_set_main_dir(sub,liste->main_dir,0)<=0)
     return(-1);
   if(Fileliste_set_target_dir(sub,target,0)<=0)
     return(-1);
   Fileliste_set_sperrliste(sub,liste->sperrliste,0);
   if(Fileliste_dir(sub,liste->adressen[idx],linecount,
                    (!(flag&(1<<15)))<<6)<=0) {
     Fileliste_destroy(&sub,0);
     return(-1);
   }
   subloeschen= 1;
 } 
 /* Do not process names which are ignored by the writer program */
 il= strlen(liste->impl_igno);
 for(i=0;i<sub->anz;i++){
   if(Fileliste_is_impl_igno(liste,sub->adressen[i],0)>0)
     Fileliste_defunct_item(sub,i,2);
 }
 
 Fileliste_inherit_target(liste,idx,sub,0);
 Fileliste_set_impl_igno(sub,liste->impl_igno,0);

 ret= Backup_plan_compose(plan,sub,1|(flag&(1<<15)));
 if(subloeschen)
   Fileliste_destroy(&sub,0);
 Fileliste_defunct_item(liste,idx,2); /* kill eventual sublist */
 if(ret<=0)
   return(ret);
 return(1);
}


int Backup_plan_stueck_header(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= only test wether a info_skript would be prepended 
*/
{
 struct FilelistE *liste= NULL;
 char auskunft_name[4096];
 int ret,idx,i,zero= 0;

 if(plan->auskunft_include && strlen(plan->auskunft_name->adressen[0]) && 
    plan->auskunft_an) {
   if(flag&1)
      {ret= 1; goto ex;}  /* info_skript would be prepended */
   idx= 0;
   if(plan->anz_stuecke >= 1) {
     /* use a number suffix for info script name */
     ret= Fileliste_copy_item(plan->auskunft_name,0,plan->auskunft_name,1,0);
     if(ret<=0)
       goto ex;
     sprintf(auskunft_name,"%s.%d",plan->auskunft_name->adressen[0],
                           plan->anz_stuecke  + plan->volume_number_offset);
     ret= Fileliste_set_address(plan->auskunft_name,1,auskunft_name,0);
     if(ret<=0)
       goto ex;
     idx= 1;
   }
   ret= Backup_stueck_append_file(plan->aktstueck,plan->auskunft_name,idx,2);
   if(ret<=0)
     goto ex;

   /*  
    '/ASKME_level_0'='$conf_dir/ASKME_level_0' \
    ...
    '/ASKME_level_${level-1}'='$conf_dir/ASKME_level_${level-1}' \
   */
   if(plan->configuration==2 && plan->level>0) {
     for(i=0;i<plan->level;i++) {
       sprintf(auskunft_name,"%s/%s_level_%d",plan->conf_dir,"ASKME",i);
       ret= Fileliste_single(&liste,auskunft_name,0,0.0,8|16);
       if(ret<=0)
         goto ex;
       ret= Fileliste_read_properties(liste,&zero,0);
       if(ret<=0)
         goto ex;
       sprintf(auskunft_name,"/%s_level_%d","ASKME",i);
       ret= Fileliste_set_target(liste,0,auskunft_name,0);
       if(ret<=0)
         goto ex;
       ret= Backup_stueck_append_file(plan->aktstueck,liste,0,2);
       if(ret<=0)
         goto ex;
       Fileliste_destroy(&liste,0);
     }
   }
 } else {
   if(flag&1)
     {ret= 0; goto ex;} /* no info_skript was prepended */
 }
 ret= 1;
ex:;
 if(liste!=NULL)
   Fileliste_destroy(&liste,0);
 return(ret);
}


int Backup_plan_is_acceptable(plan,type,flag)
struct Backup_plaN *plan;
int type;
int flag;
/*
 bit5= accept block devices (type==6)
*/
{
 if(type<=0)
   return(0);
 if(plan->result_filetypes&1)
   return(1);
 if(type==6 && flag&32)
   return(1);
 if(type==0x40000000 && (plan->result_filetypes&(1<<30)))
   return(1);
 if(type>29)
   return(0);
 if(plan->result_filetypes&(1<<type))
   return(1);
 return(0);
}


int Backup_plan_list_main_dirs(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 if(Fileliste_set_main_dir(plan->startliste,plan->hauptdir,0)<=0)
   return(-1);
 if(Fileliste_set_main_dir(plan->auslassen,plan->hauptdir,0)<=0)
   return(-1);
 if(Fileliste_set_main_dir(plan->auskunft_name,plan->hauptdir,0)<=0)
   return(-1);
 if(Fileliste_set_main_dir(plan->uebergrosse,plan->hauptdir,0)<=0)
   return(-1);
 return(1);
}


int Backup_plan_substitute_bdir(plan,text_in,text_out,name_path,flag)
struct Backup_plaN *plan;
char *text_in;
char text_out[4096]; /* is allowed to be == text_in */
char *name_path;
int flag;
/*
 bit0= this is not a level configuration setup
 bit1= do not substitute scdbackup- but only @SCDBACKUP_DIR@
 bit2= do not insist in matching name path but only insert it if missing
       Important: if name_path is a subdirectory, it has to end with '/'
 bit15= reverse substitution: replace bdir by @SCDBACKUP_DIR@
*/
{
 int bdl,npl,insert_name_path= 0,l,has_quote,quote_pos,found_len= -1;
 int version_len,version,i;
 char *cpt,*wpt,*npt,*name_pt= NULL,buf[4096],qt[2],unquoted[4096];
 char insdir[4096],prequote[4096],postquote[4096],*eff_text;

 buf[0]= qt[0]= qt[1]= 0;
 if(plan->bdir[0]==0) {
do_nothing:;
   strcpy(buf,text_in);
   strcpy(text_out,buf);
   return(2);
 }
 cpt= text_in;
 has_quote= 0;
 if(text_in[0]=='\'' || text_in[0]=='\"') {
   cpt++;
   has_quote= 1;
   qt[0]= text_in[0];
 }
 if(flag&(1<<15)) {
   l= strlen(plan->bdir);
   if(strncmp(cpt,plan->bdir,l)!=0)
     goto do_nothing;
   if(cpt[l]!='/' && cpt[l]!=0 && (has_quote==0 || cpt[l]!=qt[0]))
     goto do_nothing;
   buf[0]= 0;
   if(has_quote) {
     buf[0]= qt[0];
     buf[1]= 0;
   }
   strcat(buf,"@SCDBACKUP_DIR@");
   if(cpt[l]!=0)
     strcat(buf,cpt+l);
   strcpy(text_out,buf);
   return(1);
 }
 if((flag&1) || plan->configuration!=2)
   goto do_nothing;
 npl= strlen(name_path);
 npt= name_path;

 if(has_quote) {
   /* unquote for name_pt examination */;
   quote_pos= -1;
   wpt= unquoted;
   for(;*cpt!=0;cpt++) {
     if(quote_pos<0 && *cpt==qt[0]) {
       quote_pos= cpt-text_in;
   continue;
     }
     *(wpt++)= *cpt;
   }
   *wpt= 0;
   cpt= unquoted;
 }

 eff_text= cpt;
 if(strncmp(cpt,"@SCDBACKUP_DIR@",15)==0) {
   found_len= 15;
   if(cpt[15]!=0)
     name_pt= cpt+16;
   else
     name_pt= cpt+15;
   if(flag&4) {
     insert_name_path= 0;
     if(name_path[0]!=0) {
       if(strncmp(name_pt,name_path,npl)!=0)
         insert_name_path= 1;
     }
   } else {
     if(name_path[0]!=0)
       if(strcmp(name_pt,name_path)!=0)
         goto do_nothing;
   }
 } else if(!(flag&2)) {
   /* look for hindmost scdbackup-[0-9].[0-9].[0-9] directory */
   while(1) {
     cpt= strstr(cpt,"/scdbackup-");
     if(cpt==NULL)
   break;
     cpt+= 11;
     if(strlen(cpt)<6)
   break;
     if(cpt[0]<'0' || cpt[0]>'9')
   break;
     version= 100*(cpt[0]-'0');
     if(cpt[1]!='.')
   break;
     if(cpt[2]<'0' || cpt[2]>'9')
   break;
     version+= 10*(cpt[2]-'0');
     if(cpt[3]=='/' || cpt[3]==0) {
       version_len= 3;
     } else {
       if(cpt[3]!='.')
   break;
       if(cpt[4]<'0' || cpt[4]>'9')
   break;
       version+= (cpt[4]-'0');
       if(cpt[5]!='/' && cpt[5]!=0)
   break;
       version_len= 5;
     }
     found_len= (cpt+version_len)-eff_text;
     if(cpt[version_len]!=0)
       name_pt= cpt+version_len+1;
     else
       name_pt= cpt+version_len;
     if(name_path[0]!=0) {
       npt= name_path;
       if(version<85) { /* "0.8.5" */
         if(strncmp(name_path,"tmp/",4)!=0) {
           npt= strchr(name_path,'/');
           if(npt!=NULL)
             npt++;
           else
             npt= name_path;
         }
       }
       if(npt[0]!=0) {
         if(flag&4) {
           insert_name_path= 0;
           if(strncmp(npt,name_pt,strlen(npt))!=0)
             insert_name_path= 1;
         } else {
           l= strlen(npt);
           if(strncmp(npt,name_pt,l)!=0)
             goto do_nothing;
           if(name_pt[l]!=0 && name_pt[l]!=' ')
             goto do_nothing;
         }
       } else if(name_path[0]!=0 && (flag&4))
         insert_name_path= 1;
     }
   }
 }
 if(name_pt==NULL)
   goto do_nothing;
 bdl= strlen(plan->bdir);
 if(bdl+1+(npt-name_path)+strlen(name_pt)+2*has_quote>=sizeof(buf)) {
   fprintf(stderr,
           "--- Concatenation of -dir and file address becomes too long\n");
   return(-1);
 }
 buf[0]= 0;

 /* >>> ??? make this the standard method ? */
 if(has_quote) {
   if(quote_pos<0)
     goto do_nothing;
   insdir[0]= 0;
   if(flag&4)
     strcpy(insdir,name_path);
   else {
     if(npt>name_path) {
       strncpy(insdir,name_path,npt-name_path);
       insdir[npt-name_path]= 0;
     }
   }
   /* remove any trailing slash and eventually prepend leading slash */
   for(i= strlen(insdir)-1;i>=0;i--) {
     if(insdir[i]!='/')
   break;
     insdir[i]= 0;
   }
   if(insdir[0]!=0 && insdir[i]!='/') {
     for(i= strlen(insdir);i>=0;i--)
       insdir[i+1]= insdir[i];
     insdir[0]= '/';
   }

   prequote[0]= 0;
   if(found_len>=0 && quote_pos>found_len+1) {
     strncpy(prequote,text_in+found_len+1,quote_pos-found_len-1);
     prequote[quote_pos-found_len-1]= 0;
   }
   strcpy(postquote,text_in+quote_pos+1);

   sprintf(buf,"%s%s%s%s%s%s",qt,plan->bdir,insdir,prequote,qt,postquote);
   goto it_is_done;
 }

 strcat(buf,plan->bdir);
 if(name_pt[0]==0)
   goto it_is_done;
 l= strlen(buf);
 buf[l++]= '/';
 if(name_path[0]!=0) {
   if(flag&4) {
     if(insert_name_path) {
       strcpy(buf+l,name_path);
       l+= npl;
     }
   } else {
     if(npt>name_path) {
       strncpy(buf+l,name_path,npt-name_path);
       l+= npt-name_path;
     }
   }
 } 
 strcpy(buf+l,name_pt);
it_is_done:;
 strcpy(text_out,buf);
 return(1);
}


int Backup_plan_stacked_max_size(plan,max_size,count,flag)
struct Backup_plaN *plan;
double *max_size;
int *count;
int flag;
/*
 bit0= push max_size,count (else advance reading by one)
 bit1= reset stack to empty
 bit2= re-increment last used stack item
*/
{
 if(flag&4) { /* re-increment last used stack item */
   if(plan->max_size_stackpt>=0)
     plan->max_size_counts[plan->max_size_stackpt]++;
   return(1);
 }
 if(flag&2) { /* reset stack to empty */
   plan->max_size_stackpt= -1;
   *max_size= plan->default_max_size;
   *count= -1;
   return(1);
 }
 if(flag&1) { /* push max_size,count to stack */
   if(plan->max_size_stackpt>=0)
     if(plan->max_size_counts[plan->max_size_stackpt]<=0)
       plan->max_size_stackpt--;
   if(plan->max_size_stackpt+1>=Backup_plan_max_size_stackL) {
     fprintf(stderr,"--- Overflow of max_size stack\n");
     return(0);
   }
   plan->max_size_stackpt++;
   plan->max_size_stack[plan->max_size_stackpt]= *max_size;
   plan->max_size_counts[plan->max_size_stackpt]= *count;
   return(1);
 }
 if(plan->max_size_stackpt>=0)
   if(plan->max_size_counts[plan->max_size_stackpt]<=0)
     plan->max_size_stackpt--;
 if(plan->max_size_stackpt>=0) {
   if(plan->max_size_counts[plan->max_size_stackpt]>0) {
     plan->max_size_counts[plan->max_size_stackpt]--;
     *max_size= plan->max_size_stack[plan->max_size_stackpt];
     *count= plan->max_size_counts[plan->max_size_stackpt];
     return(1);
   }
 }
 *max_size= plan->default_max_size;
 *count= -1;
 return(1);
}


int Backup_plan_next_stueck(plan,stueck,flag)
struct Backup_plaN *plan;
struct Backup_stuecK **stueck;
int flag;
{
 int stack_count;
 char num[81];

 if(plan->save_stack_media)
   /* if suitable : revoke count as stacked media */
   if(plan->max_size>plan->default_max_size)
     if(plan->default_max_size>plan->aktstueck->bytes) {
       Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,4);
       plan->max_size= plan->default_max_size;
     }
 if(plan->compact_num)
   Sfile_scale(plan->aktstueck->bytes,num,7,1e5,1);
 else
   sprintf(num,"%.lf",plan->aktstueck->bytes);
 fprintf(stderr,"########### %d  (%d items  %s bytes)\n",
                plan->anz_stuecke + plan->volume_number_offset,
                plan->aktstueck->liste->anz,num);
 /* start new piece */
 Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,0);
 if(Backup_stueck_new(stueck,plan,plan->aktstueck,
                      plan->max_items,plan->max_size,plan->hauptdir,0)<=0)
   return(0);
 plan->aktstueck= *stueck;
 plan->anz_stuecke++;
 if(Backup_plan_stueck_header(plan,0)<=0)
   return(0); 
 return(1);
}


int Backup_plan_add_to_stueck(plan,liste,idx,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int idx;
int flag;
/*
 bit1= this is added_by_scdbackup , not a payload item
*/
{
 int ret,auslassen;

 if((liste->typen[idx]==2 
     ||(liste->typen[idx]==3 && Cd_backup_planer_follow_linkS))
    && Fileliste_search_file(plan->auslassen,liste->adressen[idx],1|4|32)>=0) {
   /* parent address of an excluded path */
   ret= Backup_plan_split(plan,liste,idx,1<<15);
   if(ret<=0)
     return(ret);
 } else {
   /* check wether already included with the same target Liste */
   if(plan->prevent_doubles)
     ret= Backup_plan_is_planned(plan,liste,idx,0);
   else
     ret= 0;
   auslassen= 0;
   if(ret<0) {
     if(verbous)
       fprintf(stderr,"---------- Backup_plan_is_planned() failed: %s\n",
                      liste->adressen[idx]);
     auslassen= 1;
   } else if(ret>0) {
     if(verbous)
       fprintf(stderr,"---------- already in backup plan: %s\n",
                      liste->adressen[idx]);
     auslassen= 1;
   }
   Backup_stueck_append_file(plan->aktstueck,liste,idx,auslassen|(flag&2));
 }
 return(1);
}


int Backup_plan_compose_by_size(plan,liste,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int flag;
/*
 bit0= recursion
 bit9= this is added_by_scdbackup , not a payload item
 bit15= do not pacify
*/
{
 struct Backup_stuecK *stueck;
 double frei,best_size,basisgroesse;
 int best,i,fileda,ret,sonderspaltung=0,pass,passes,nichts_teilbares,count;

 stueck= plan->aktstueck;
 if(stueck==NULL)  
   return(0);
 basisgroesse= plan->auskunft_groesse*plan->auskunft_include;

 do {

   nichts_teilbares= 0;
   if(stueck->bytes<=basisgroesse ||
      sonderspaltung){ /* the backup piece is still too small */
     passes= 1+!!sonderspaltung;
     sonderspaltung= 0;
     for(pass=0;pass<passes;pass++){
       /* at start : find oversized directory and split */;
       best= -1;
       best_size= 0;
       for(i=0;i<liste->anz;i++){
         if(!Backup_plan_is_acceptable(plan,(int) liste->typen[i],
                                       (int) (liste->flags[i]&32)))
       continue;
         /* With pass 0 : only take those which need to get split anyway */
         if(liste->groessen[i]<=
                          plan->max_size-basisgroesse-stueck->bytes*(!!pass))
       continue;
         if(liste->typen[i]!=2) {
           if(pass>0)
       continue;
           if(!(flag&(1<<15)))
             Print_pacifier(0,"",1|16); /* clear pacifier line */
           fprintf(stderr,"oversized file (%.lf>%.lf) : %s\n",
                         liste->groessen[i],
                         plan->max_size-basisgroesse-stueck->bytes*(!!pass),
                         liste->adressen[i]);
           Fileliste_copy_item(liste,i,plan->uebergrosse,
                               plan->uebergrosse->anz,0);
           Fileliste_defunct_item(liste,i,0);
       continue;
         }

         /* check wether the directory contains too many files and would
            make the next piece unnecessarly small */
         if(pass>0){
           count= Fileliste_count_dir_files(liste,liste->adressen[i],0);
           if(count<=0)
       continue;
           if(count>plan->max_items/2)
       continue;
         }

         if(best_size<liste->groessen[i]){ 
           best= i;
           best_size= liste->groessen[i]; 
         }
       }
       if(best>=0){
         ret= Backup_plan_split(plan,liste,best,0);
         if(ret<=0) 
           return(ret);
     break; /* pass */
       }
       if(pass==passes-1 && passes>1)
         nichts_teilbares= 1;
     }
   }

   stueck= plan->aktstueck;
   if(stueck==NULL) return(0);

   frei= plan->max_size-stueck->bytes;
   best= -1;
   best_size= -1.0;
   fileda= 0;
   for(i=0;i<liste->anz;i++){
     if(!Backup_plan_is_acceptable(plan,(int) liste->typen[i],
                                   (int) (liste->flags[i]&32)))
   continue;
     if(liste->groessen[i]<0)
   continue;
     fileda= 1;
     if(liste->groessen[i]<=frei){
       if(best_size<liste->groessen[i]){ 
         best= i;
         best_size= liste->groessen[i]; 
       }
     }
   }
   if(!fileda) 
 break;

   if(best<0) {
     /* only objects left which do not fit */

     /* check wether the current piece is still too small */
     if(stueck->bytes<plan->min_size_fak*plan->max_size) {
       if(nichts_teilbares) {
         if(plan->scattering && (flag&1) && liste!=plan->startliste) {
           /* delay the remaining objects and process others first */
           int aufgeschoben= 0;
           /* dirty hack : append all remaining objects to start list */
           for(i=0;i<liste->anz;i++){
             if(!Backup_plan_is_acceptable(plan,(int) liste->typen[i],
                                           (int) (liste->flags[i]&32)))
           continue;
             if(Fileliste_copy_item_hd(liste,i,plan->startliste,
                                       plan->startliste->anz,1)<=0)
               return(0);
             /* mark as "done"  */;
             Fileliste_defunct_item(liste,i,0);
             aufgeschoben++;
             if(!(flag&(1<<15)))
               Print_pacifier(0,"",1|16); /* clear pacifier line */
             fprintf(stderr,"deferred by -scattering -on : %s\n",
                            liste->adressen[i]);
           }
           /* leave this level of recursion */
           if(aufgeschoben>0) 
             return(1);
         }
       }else{
         /* eventually resume the old piece by a part of a large object */
         sonderspaltung= 1;
 continue;
       }
     }
     if(!(flag&(1<<15)))
       Print_pacifier(0,"",1|16); /* clear pacifier line */
     ret= Backup_plan_next_stueck(plan,&stueck,0);
     if(ret<=0)
       return(ret);
 continue;
   }

   if(stueck->liste->anz>=plan->max_items) {
     /* the current piece is full */
     if(!(flag&(1<<15)))
       Print_pacifier(0,"",1|16); /* clear pacifier line */
     ret= Backup_plan_next_stueck(plan,&stueck,0);
     if(ret<=0)
       return(ret);
   }
   if(liste->subdirs!=NULL)
     if(liste->subdirs[best]!=NULL)
       Fileliste_destroy(&(liste->subdirs[best]),2|4);
   ret= Backup_plan_add_to_stueck(plan,liste,best,(flag&512)>>8);
   if(ret<=0)
     return(ret);
 } while(1);
 return(1);
}


int Backup_plan_compose_by_sort(plan,liste,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int flag;
/*
 bit0= recursion
 bit1-8: mode for Fileliste_sort_index()
 bit9= this is added_by_scdbackup , not a payload item
 bit15= do not pacify
*/
{
 struct Backup_stuecK *stueck;
 double frei,basisgroesse;
 int i,ret,*index= NULL,index_count,idx;

 stueck= plan->aktstueck;
 if(stueck==NULL)  
   return(0);
 basisgroesse= plan->auskunft_groesse*plan->auskunft_include;

 /* sort liste alphabetically according to target addresses */
 ret= Fileliste_sort_index(liste,&index,&index_count,(flag>>1)&255);
 if(ret<=0)
   goto ex;

 for(idx=0;idx<index_count;idx++){
   i= index[idx];
   if(!Backup_plan_is_acceptable(plan,(int) liste->typen[i],
                                 (int) (liste->flags[i]&32)))
 continue;
   if(liste->groessen[i]<0)
 continue;
   if(stueck->liste->anz>=plan->max_items) {
     if(!(flag&(1<<15)))
       Print_pacifier(0,"",1|16); /* clear pacifier line */
     ret= Backup_plan_next_stueck(plan,&stueck,0);
     if(ret<=0)
       goto ex;
   }

   frei= plan->max_size-stueck->bytes;
   if(liste->groessen[i]>frei){

     /* ??? take into respect scattering ??? */

     if(liste->typen[i]==2) { /* Directory */
       /* is there still room that needs to be filled ?
          or is the directory too large anyway ? */;
       if(stueck->bytes<plan->min_size_fak*plan->max_size
          || liste->groessen[i]>plan->max_size-basisgroesse) {
         if(!(flag&(1<<15)))
           Print_pacifier(0,"",1|16); /* clear pacifier line */
         ret= Backup_plan_split(plan,liste,i,0);
         if(ret<=0) 
           goto ex;
         stueck= plan->aktstueck;
 continue;
       } else { /* directory will fit next stueck */
         if(!(flag&(1<<15)))
           Print_pacifier(0,"",1|16); /* clear pacifier line */
         ret= Backup_plan_next_stueck(plan,&stueck,0);
         if(ret<=0)
           goto ex;
       }
     } else {
       if(liste->groessen[i]>plan->max_size-basisgroesse) {
         if(!(flag&(1<<15)))
           Print_pacifier(0,"",1|16); /* clear pacifier line */
         fprintf(stderr,"oversized file (%.lf > %.lf) : %s\n",
                        liste->groessen[i],
                        plan->max_size-basisgroesse,
                        liste->adressen[i]);
         Fileliste_copy_item(liste,i,plan->uebergrosse,
                             plan->uebergrosse->anz,0);
         Fileliste_defunct_item(liste,i,2);
 continue;
       } else {
         if(!(flag&(1<<15)))
           Print_pacifier(0,"",1|16); /* clear pacifier line */
         ret= Backup_plan_next_stueck(plan,&stueck,0);
         if(ret<=0)
           goto ex;
       }
     }
   }

   stueck= plan->aktstueck;
   if(stueck==NULL)
     {ret= 0; goto ex;}
   if(liste->subdirs!=NULL)
     if(liste->subdirs[i]!=NULL)
       Fileliste_destroy(&(liste->subdirs[i]),2|4);
   ret= Backup_plan_add_to_stueck(plan,liste,i,(flag&512)>>8);
   if(ret<=0)
     goto ex;
 }
 ret= 1;
ex:;
 if(index!=NULL)
   free((char *) index);
 return(ret);
}


int Backup_plan_compose(plan,liste,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int flag;
/*
 bit0= recursion (handed over to specialized composer)
 bit1= read properties of liste
 bit2= do not print final message
 bit3= do nothing but to print final message
 bit4= print initial message
 bit9= this is added_by_scdbackup , not a payload item
 bit15= do not pacify
*/
{
 struct Backup_stuecK *stueck;
 int ret,i,nopac= 0,flpac= 64;
 char num[81];
 int *linecount;
 
 linecount= &plan->compose_linecount;
 if(!(flag&1)) {
   Cdbp_subs= 0.0;
   Cdbp_subs_lines= 0.0;
   Cdbp_subs_mem= 0.0;
   Cdbp_count_subs= 1;
 }
 if((flag&(1<<15))||Cd_backup_planer_pacifier_intervaL<0) {
   nopac= 1<<15;
   flpac= 0;
 }
 if(flag&8)
   {ret= 1; goto final_message;}
 if(Fileliste_assert_subdirs(liste,0)<=0)
   return(-1);
 if(flag&16)
   fprintf(stderr,"\nbeginning to compose backup volumes\n");
 if((flag&2)){
   if(Fileliste_read_properties(liste,linecount,
                          8|((!!plan->auslassen_nicht_startliste)<<2)|flpac)<0)
     return(0);
 }

 if(plan->stuecke==NULL) {
   ret= Backup_plan_first_piece(plan,0);
   if(ret<=0)
     return(ret);
 }
 stueck= plan->aktstueck;
 if(stueck==NULL)  
   return(0);

 if(Fileliste_set_main_dir(liste,plan->hauptdir,0)<=0)
   return(-1);

 /* leave out excluded objects */
 if((flag&1)||!plan->auslassen_nicht_startliste) {
   for(i=0;i<liste->anz;i++) {
     if(Fileliste_search_file(plan->auslassen,liste->adressen[i],32)>=0)
       Fileliste_defunct_item(liste,i,0);
   }
 }

 if(plan->composition_mode==1) {
   /* alphabetically (target) */;
   ret= Backup_plan_compose_by_sort(plan,liste,
                                    (flag&1)|(1<<1)|nopac|(flag&512));
 } else if(plan->composition_mode==2) {
   /* alphabetically (source) */;
   ret= Backup_plan_compose_by_sort(plan,liste,(flag&1)|nopac|(flag&512));
 } else if(plan->composition_mode==3) {
   /* chronologically */;
   ret= Backup_plan_compose_by_sort(plan,liste,
                                    (flag&1)|(2<<1)|nopac|(flag&512));
 } else {
   ret= Backup_plan_compose_by_size(plan,liste,(flag&1)|nopac|(flag&512));
 }
final_message:;
 if((*linecount)>0 && !(flag&(1|4|(1<<15)))) {
   Print_pacifier(*linecount,"directory tests done for composition",1|2|4|8|16);
   fprintf(stderr,"\n");;
 }
 stueck= plan->aktstueck;
 if(stueck!=NULL && ret>=0 && !(flag&(1|4))) {
   if(plan->compact_num)
     Sfile_scale(stueck->bytes,num,7,1e5,1);
   else
     sprintf(num,"%.lf",stueck->bytes);
   fprintf(stderr,"########### %d  (%d items  %s bytes)\n\n",
                  plan->anz_stuecke + plan->volume_number_offset,
                  stueck->liste->anz,num);
 }
 if(!(flag&1)) {
   Cdbp_count_subs= 0;

#ifdef Cdbackup_planer_developpeR
   if(!(flag&(1|4|(1<<15))))
     fprintf(stderr,
         "developer message:  subs= %.lf , subs_lines= %.lf , subs_mem=%.lf\n",
         Cdbp_subs,Cdbp_subs_lines,Cdbp_subs_mem);
#endif /* Cdbackup_planer_developpeR */

 }
 return(ret);
}


int Backup_plan_collision_test(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= check  plan->startliste rather  than  plan->stuecke
*/
{
 struct Backup_stuecK *stueck;
 struct FilelistE *liste= NULL;
 int i,zaehl=0,w=0,j,ret,ct_flag;

 if(plan->collision_avoid==1)      ct_flag= 2;
 else                              ct_flag= 0;
 if(plan->auslassen_nicht_startliste) ct_flag|= 4;
 else                                 ct_flag|= 8;
 if(flag&1){
   if(Fileliste_new(&liste,plan->startliste->anz+1,plan->hauptdir,2)<=0)
     return(-1);
   w= 0;
   if(plan->auskunft_include && strlen(plan->auskunft_name->adressen[0]) &&
      plan->auskunft_an) {
     Fileliste_copy_item(plan->auskunft_name,0,liste,w++,0);
   }
   for(j=0;j<plan->startliste->anz;j++){
     Fileliste_copy_item(plan->startliste,j,liste,w++,0);
   }  
 }else{
   /* determine total number */
   stueck= plan->stuecke; 
   for(i=0;i<plan->anz_stuecke;i++){
     liste= stueck->liste;
     zaehl+= stueck->liste->anz;
     stueck= stueck->next;
   }
   /* consolidate in one list */
   if(Fileliste_new(&liste,zaehl+1,plan->hauptdir,2)<=0)
     return(-1);
   stueck= plan->stuecke; 
   for(i=0;i<plan->anz_stuecke;i++){
     for(j=0;j<stueck->liste->anz;j++){
       Fileliste_copy_item(stueck->liste,j,liste,w++,0);
     }
     stueck= stueck->next;
   }
 }
 if(Fileliste_set_main_dir(liste,plan->hauptdir,0)<=0)
   return(-1);

 /* collisiontest */
 ret= Fileliste_collision_test(liste,plan->auslassen,ct_flag);
 if(plan->collision_avoid==1){
   /* copy back */
   w= 0;
   if(flag&1){
     if(plan->auskunft_include && strlen(plan->auskunft_name->adressen[0]) &&
        plan->auskunft_an) {
       Fileliste_copy_item(liste,w++,plan->auskunft_name,0,0);
     }
     for(j=0;j<plan->startliste->anz;j++){
       Fileliste_copy_item(liste,w++,plan->startliste,j,0);
     }  
   }else{
     stueck= plan->stuecke;
     for(i=0;i<plan->anz_stuecke;i++){
       for(j=0;j<stueck->liste->anz;j++){
         Fileliste_copy_item(liste,w++,stueck->liste,j,0);
       }
       stueck= stueck->next;
     }
   }
 }

 if(liste!=NULL) Fileliste_destroy(&liste,0);
 if(ret>0 && plan->collision_avoid==2) {
   ret= -2; /* abort because of collision  */
 }
 return(ret);
}


int Backup_plan_out_part_mark(plan,out,liste,idx,flag)
struct Backup_plaN *plan;
FILE *out;
struct FilelistE *liste;
int idx;
int flag;
{
 int partno,partcount;
 double start_byte,byte_count;

 if(liste->part_recs!=NULL)
   if(liste->part_recs[idx]!=NULL) {
     Filelist_part_rec_get_rec(liste->part_recs[idx],
                               &partno,&partcount,
                               &start_byte,&byte_count,0);
     fprintf(out,"**%d/%d:%.lf+%.lf**",
                 partno,partcount,start_byte,byte_count);
   }
 return(1);
}


int Backup_plan_stdout(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= stderr rather than stdout
 bit1= super raw : simple file list, empty lines separate the pieces 
*/
{
 int i,j,ftypelist_size;
 struct Backup_stuecK *stueck;
 struct FilelistE *liste;
 char ftype,ftypelist[]={"?FDLPHB"};
 FILE *out;

 ftypelist_size= strlen(ftypelist);
 out= stdout;
 if(flag&1)
   out= stderr;

 if(flag&2) goto super_raw;

 liste= plan->uebergrosse;
 for(j=0;j<liste->anz;j++){
   fprintf(out,"+++++++++++ oversized file (%.lf bytes) : %s\n",
               liste->groessen[j],liste->adressen[j]);
 }
 if(liste->anz>0) fprintf(out,"\n");

 stueck= plan->stuecke; 
 for(i=0;i<plan->anz_stuecke;i++){
   liste= stueck->liste;
   fprintf(out,"########### %d  (%d items  %.lf bytes)\n",
               i + 1 + plan->volume_number_offset, liste->anz,stueck->bytes);
   for(j=0;j<liste->anz;j++) {
     if(liste->typen[j]>=0 && liste->typen[j]<ftypelist_size)
       ftype= ftypelist[(int) liste->typen[j]];
     else
       ftype= '?';
     fprintf(out,"%c %9.lf  ",ftype,liste->groessen[j]);
     Backup_plan_out_part_mark(plan,out,liste,j,0);

     /* >>> 0520: escape newlines */;

     fprintf(out,"%s\n",liste->adressen[j]);
   }
   stueck= stueck->next;
 }
 return(1);

super_raw:;
 stueck= plan->stuecke; 
 for(i=0;i<plan->anz_stuecke;i++){
   liste= stueck->liste;
   for(j=0;j<liste->anz;j++) {
     Backup_plan_out_part_mark(plan,out,liste,j,0);

     /* >>> 0520: escape newlines */;

     fprintf(out,"%s\n",liste->adressen[j]);
   }
   fprintf(out,"\n");
   stueck= stueck->next;
 }
 return(1);
}


int Backup_plan_info_skript(plan,nr,flag)
struct Backup_plaN *plan;
int nr;
int flag;
/*
 bit0= create .pnr file with a placeholder text for piece number 
*/
{
 int ret,i,has_part_recs,saltcount,l;
 struct Lies_popeN *gelesen,gelesen_data;
 struct Backup_stuecK *stueck;
 FILE *fp;
 char line[4096],auskunft_name[4096],num[81];
 char salt_text[256],salt_md5[16];
 Sregex_md5_ctX *dummy_ctx= NULL;
 struct timeval tv; 
 struct timezone tz;

 static int salts_to_count= 8; /* >>> ??? make this a program option ? */

 gelesen= &gelesen_data;
 has_part_recs= (Backup_plan_count_part_recs(plan,0)>0);

 fp= NULL;
 strcpy(auskunft_name,plan->auskunft_name->adressen[0]);
 if(auskunft_name[0]) {
   if(flag&1) strcat(auskunft_name,".pnr");
   unlink(auskunft_name);
   fp= fopen(auskunft_name,"wb");
   if(fp==NULL){
     fprintf(stderr,"can't open info_script '%s' for writing.\n",
                    auskunft_name);
     fprintf(stderr,"reason given by OS : %s\n",strerror(errno));
   }
 } else {
   fp= stdout;
 }

 /* included ASKME is no longer supported */
 if(fp==stdout)
   return(0);

 if(fp==NULL)
   return(1);

 fprintf(fp,"#!%s\n",plan->shell_to_use);

 /* generate random salt to help encryption of payload files in archive */
 gettimeofday(&tv,&tz);
 sprintf(salt_text,"%lx%lx%lx%lx",(long) tv.tv_usec,(long) tv.tv_sec,
                                  (long) salt_text,(long) plan);
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke && strlen(salt_text)<sizeof(salt_text)-33;i++) {
   sprintf(salt_text+strlen(salt_text),"%lx",(long) stueck);
   stueck= stueck->next;
 }
 l= strlen(salt_text);
 for(i=0;i<16;i++) salt_md5[i]= 0;
 for(saltcount= 0; saltcount<salts_to_count; saltcount++) {
   if((saltcount%2)==0)
     fprintf(fp,"#");
   for(i= 0;i<l;i++)
     salt_text[i]^= salt_md5[i%16];
   Sregex_md5(&dummy_ctx,salt_text,l,salt_md5,1|2|(1<<15));
   for(i=0;i<16;i++)
     fprintf(fp,"%-2.2X",((unsigned) salt_md5[i])&255);
   if(saltcount==salts_to_count-1)
     fprintf(fp," (random)\n");
   if((saltcount%2)==1)
     fprintf(fp,"\n");
 }

 /* identification */
 fprintf(fp,"if test \"$1\" = -id -o  \"$1\" = -help -o $# = 0 -o  \"$1\" = -hilfe ; then\n");
 fprintf(fp,"\n");
 fprintf(fp,"  echo 'cd_backup_planer %s : information script'\n",PROG_VERSION);
 fprintf(fp,"  echo\n");
 /* date */
 if(Lies_popen_init("date",gelesen,0)>0) {
   if(gelesen->anz>0)
     fprintf(fp,"  echo 'backup date %s'\n",gelesen->lines[0]);
   Lies_popen_cleanup(gelesen,0);
 }
 fprintf(fp,"  echo 'id          %s  %s'\n",
            plan->backup_id,Sfile_datestr(plan->start_time,1|2|256));
 if(plan->precheck_agetest&1) {
   fprintf(fp,"  echo '");
   if(plan->level>0)
     fprintf(fp,"-level %d   ",plan->level);
   fprintf(fp,"-younger_than %s",Sfile_datestr(plan->precheck_age,1|2|256));
   fprintf(fp,"'\n");
 }
 if(flag&1){
   fprintf(fp,"  echo 'BACKUP_PLANER_PART_MESSAGE_PLACEHOLDER'\n");
 } else if(nr>0) {
   fprintf(fp,"  echo 'this is backup part no %d of %d'\n",
              nr + plan->volume_number_offset,
              plan->anz_stuecke + plan->volume_number_offset);
 }else{
   fprintf(fp,"  echo 'this backup consists of %d parts'\n",
              plan->anz_stuecke + plan->volume_number_offset);
 }
 fprintf(fp,"  echo\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");

 fprintf(fp,"if test \"$1\" = -help -o $# = 0 ; then\n");
 if(plan->info_short)
   fprintf(fp,"  echo \"usage: $0  [option]\"\n");
 else
   fprintf(fp,"  echo \"usage: $0  [option] file [file ...]\"\n");
 fprintf(fp,
 "  echo 'options:'\n");
 fprintf(fp,
 "  echo '  -help   prints this text'\n");
 fprintf(fp,
 "  echo '  -hilfe  gibt diesen Text in deutscher Sprache aus'\n");
 fprintf(fp,
 "  echo '          (prints this text in german language)'\n");
 fprintf(fp,
 "  echo '  -id     prints a message about this backup'\n");
 fprintf(fp,
 "  echo '  -count  prints the number of parts'\n");
 if(plan->info_short) {
   fprintf(fp,
    "  echo 'Note: this information script is a stub without item lists.'\n");
 } else {
   fprintf(fp,
 "  echo '  -list   prints the parts and their items'\n");
   fprintf(fp,
 "  echo '  -levelall lookup all levels (else: only youngest match)'\n");
   fprintf(fp,
 "  echo '  -suball lookup all sub items and levels (else: one per part)'\n");
   fprintf(fp,
 "  echo '  -fast   do not look for sub items'\n");
   fprintf(fp,
 "  echo '  files   (respectively directories) are looked up in the'\n");
   fprintf(fp,
 "  echo '          parts'\"'\"' item lists. If the filename or a'\n");
   fprintf(fp,
 "  echo '          directory above is found then the part number, the'\n");
   fprintf(fp,
 "  echo '          item and the filename re printed. If a filename below'\n");
   fprintf(fp,
 "  echo '          (a sub item) is found, then the action depends on'\n");
   fprintf(fp,
 "  echo '          -suball and -fast'\n");
 }
 fprintf(fp,
 "  echo\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");

 fprintf(fp,"if test \"$1\" = -hilfe ; then\n");
 if(plan->info_short) {
   fprintf(fp,"  echo \"Anwendung: $0  [option]\"\n");
 } else {
   fprintf(fp,"  echo \"Anwendung: $0  [option] file [file ...]\"\n");
 }
 fprintf(fp,
 "  echo 'Optionen :'\n");
 fprintf(fp,
 "  echo '  -hilfe  gibt diesen Text aus'\n");
 fprintf(fp,
 "  echo '  -help   prints this text in english language'\n");
 fprintf(fp,
 "  echo '          (gibt diesen Text in englischer Sprache aus)'\n");
 fprintf(fp,
 "  echo '  -id     gibt eine Meldung ueber diesen Backup aus'\n");
 fprintf(fp,
 "  echo '  -count  gibt die Zahl der Teile aus'\n");
 if(plan->info_short) {
   fprintf(fp,
    "  echo 'Hinweis: Dies ist ein Rumpfskript ohne Filelisten.'\n");
 } else {
   fprintf(fp,
 "  echo '  -list   gibt die Teile und ihre Filelisten aus'\n");
   fprintf(fp,
 "  echo '  -levelall alle -level zeigen (sonst: nur den juengsten Fund)'\n");
   fprintf(fp,
 "  echo '  -suball alle Unteradressen zeigen (sonst: eine pro Stueck)'\n");
   fprintf(fp,
 "  echo '  -fast   nicht nach Unteradressen suchen'\n");
   fprintf(fp,
 "  echo '  files   (bzw. Directories) werden in den Filelisten der'\n");
   fprintf(fp,
 "  echo '          Stuecke gesucht. Wenn ein Filename oder ein darueber-'\n");
   fprintf(fp,
 "  echo '          liegendes Directory gefunden  wird, dann werden die'\n");
   fprintf(fp,
 "  echo '          Stuecknummer, die gefundene Adresse und der Filename'\n");
   fprintf(fp,
 "  echo '          ausgegeben. Wenn eine Fileadresse gefunden wird, die'\n");
   fprintf(fp,
 "  echo '          unterhalb des Filenamens liegt (Unteradresse), dann'\n");
   fprintf(fp,
 "  echo '          haengt die Reaktion von -suball und -fast ab.'\n");
 }
 fprintf(fp,
 "  echo\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");


 fprintf(fp,"if test \"$1\" = -id -o \"$1\" = -help -o $# = 0 -o \"$1\" = -hilfe ;");
 fprintf(fp," then exit 0 ; fi\n\n");

 fprintf(fp,"if test \"$1\" = -count ; then\n");
 fprintf(fp,"  echo %d\n", plan->anz_stuecke + plan->volume_number_offset);
 fprintf(fp,"  exit 0\n");
 fprintf(fp,"fi\n\n");

 if(plan->info_short)
   goto after_lists_and_search;

 /* list funktions */
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   fprintf(fp,"cd_liste_%d() {\n", i + 1 + plan->volume_number_offset);
   ret= Backup_stueck_report(stueck,fp,"cat ",-1,-1,3|256);
   if(ret<=0)
     return(1);
   fprintf(fp,"}\n\n");
   stueck= stueck->next;
 if(stueck==NULL) break;
 }

 /* search funktions */
 fprintf(fp,"such_in_cd() {\n");
 fprintf(fp,"such_cd=$1\n");
 fprintf(fp,"for i in $(cd_liste_$such_cd)\n");
 fprintf(fp,"do\n");
 fprintf(fp,"  if grep -q '^'\"$i\"/ <<+\n");
 fprintf(fp,"$name\n");
 fprintf(fp,"+\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test \"$i\" = \"$name\"\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"done\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");
 fprintf(fp,"such_in_cd_sub() {\n");
 fprintf(fp,"such_cd=$1\n");
 fprintf(fp,"if test $sub = 1 ; then subsperr=0 ; else subsperr=1 ; fi\n");
 fprintf(fp,"for i in $(cd_liste_$such_cd)\n");
 fprintf(fp,"do\n");
 fprintf(fp,"  if grep -q '^'\"$i\"/ <<+\n");
 fprintf(fp,"$name\n");
 fprintf(fp,"+\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test \"$i\" = \"$name\"\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"    fund=1\n");
 fprintf(fp,"    namefund=1\n");
 fprintf(fp,"    return\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test $suball = 1 -o $subsperr = 0\n");
 fprintf(fp,"  then\n");
 fprintf(fp,"    if grep -q '^'\"$name\" <<+\n");
 fprintf(fp,"$i\n");
 fprintf(fp,"+\n");
 fprintf(fp,"    then\n");
 fprintf(fp,"      ok=0\n");
 fprintf(fp,"      if grep -q '^'\"$name\"/ <<+\n");
 fprintf(fp,"$i\n");
 fprintf(fp,"+\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        ok=1\n");
 fprintf(fp,"      elif grep -q '^'\"$name\"'$' <<+\n");
 fprintf(fp,"$i\n");
 fprintf(fp,"+\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        ok=1\n");
 fprintf(fp,"      fi\n");
 fprintf(fp,"      if test $ok = 1\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        echo \"$such_cd   $i   $name\"\n");
 fprintf(fp,"        fund=1\n");
 fprintf(fp,"        subsperr=1\n");
 fprintf(fp,"        if test $suball = 0 ; then return ; fi\n");
 fprintf(fp,"      fi\n");
 fprintf(fp,"    fi\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"done\n");
 fprintf(fp,"}\n");
 fprintf(fp,"\n");

after_lists_and_search:;

 /* process arguments */
 fprintf(fp,"if test \"$1\" = -list ; then\n");
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   if(plan->compact_num)
     Sfile_scale(stueck->bytes,num,7,1e5,1);
   else
     sprintf(num,"%.lf",stueck->bytes);
   fprintf(fp,"  echo '########## part %d  (%d items  estim. %s bytes)'\n",
              i + 1 + plan->volume_number_offset, stueck->liste->anz, num);
   if(!plan->info_short)
     fprintf(fp,"  cd_liste_%d\n", i + 1 + plan->volume_number_offset);
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 if(plan->info_short) {
   fprintf(fp,
    "  echo '--- No item lists available.        Keine Filelisten enthalten.        ---'\n");
   fprintf(fp,
    "  echo '--- Eventually use scdbackup_askme. Benutzen Sie ggf. scdbackup_askme. ---'\n");
 }
 fprintf(fp,"  exit 0\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");

 if(plan->info_short) {
   fprintf(fp,
    "echo '--- No item lists available.        Keine Filelisten enthalten.        ---'\n");
   fprintf(fp,
    "echo '--- Eventually use scdbackup_askme. Benutzen Sie ggf. scdbackup_askme. ---'\n");
   fprintf(fp,"exit 1\n");
   goto close_and_return; 
 }

 fprintf(fp,"level_opt=\n");
 fprintf(fp,"sub=1\n");
 fprintf(fp,"if test \"$1\" = -fast\n");
 fprintf(fp,"then\n");
 fprintf(fp," shift 1\n");
 fprintf(fp,"  sub=0\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");
 fprintf(fp,"suball=0\n");
 fprintf(fp,"if test \"$1\" = -suball\n");
 fprintf(fp,"then\n");
 fprintf(fp,"  shift 1\n");
 fprintf(fp,"  suball=1\n");
 fprintf(fp,"  level_opt=\"${level_opt} -suball\"\n");
 fprintf(fp,"  sub=1\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"levelall=0\n");
 fprintf(fp,"if test \"$1\" = -levelall\n");
 fprintf(fp,"then\n");
 fprintf(fp,"  shift 1\n");
 fprintf(fp,"  levelall=1\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"nolevel=0\n");
 fprintf(fp,"if test \"$1\" = -nolevel\n");
 fprintf(fp,"then\n");
 fprintf(fp,"  shift 1\n");
 fprintf(fp,"  nolevel=1\n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");
 fprintf(fp,"fund=0\n");
 fprintf(fp,"for name in \"$@\"\n");
 fprintf(fp,"do\n");
 fprintf(fp,"  namefund=0\n");
 fprintf(fp,"  for cd in");
 for(i=0;i<plan->anz_stuecke;i++)
   fprintf(fp," %d", i + 1 + plan->volume_number_offset);
 fprintf(fp,"\n");
 fprintf(fp,"  do\n");
 if(!has_part_recs) {
   fprintf(fp,"    if test $namefund = 0\n");
   fprintf(fp,"    then\n");
 }
 fprintf(fp,"      if test $sub = 1\n");
 fprintf(fp,"      then\n");
 fprintf(fp,"        such_in_cd_sub $cd\n");
 fprintf(fp,"      else\n");
 fprintf(fp,"        such_in_cd $cd\n");
 fprintf(fp,"      fi\n");
 if(!has_part_recs) {
   fprintf(fp,"    fi\n");
 }
 fprintf(fp,"  done\n");
 fprintf(fp,"done\n");
 fprintf(fp,"if test $fund = 1 -a $levelall = 0 ; then exit 0 ; fi\n");
 fprintf(fp,"if test $nolevel = 1 ; then exit 1 ; fi\n");

 fprintf(fp,"if test \"$level_opt\" = \"\" \n");
 fprintf(fp,"then\n");
 fprintf(fp,"  level_opt=\"-nolevel\" \n");
 fprintf(fp,"fi\n");
 fprintf(fp,"\n");
 if(plan->configuration==2 && plan->level>0) {
   fprintf(fp,"for lev in");
   for(i=plan->level-1;i>=0;i--) 
     fprintf(fp," %d",i);
   fprintf(fp,"\n");
   fprintf(fp,"do\n");
   fprintf(fp,"  if $(dirname $0)/%s_level_$lev $level_opt \"$@\" 2>&1 >/dev/null\n",
              "ASKME");
   fprintf(fp,"  then\n");
   fprintf(fp,"    fund=1\n");
   fprintf(fp,"    echo '+++ on backup level '$lev\n");
   fprintf(fp,"    $(dirname $0)/%s_level_$lev $level_opt \"$@\"\n",
              "ASKME");
   fprintf(fp,"    if test $levelall = 0\n");
   fprintf(fp,"    then\n");
   fprintf(fp,"      exit 0\n");
   fprintf(fp,"    fi\n");
   fprintf(fp,"  fi\n");
   fprintf(fp,"done\n");
 }

 fprintf(fp,"if test $fund = 1 ; then exit 0 ; fi\n");
 fprintf(fp,"exit 1\n");
 fprintf(fp,"\n");

close_and_return:; 
 if(fp!=stdout && fp!=stderr && strlen(auskunft_name)>0) {
   /* give x-permissions */
   sprintf(line,"chmod u+x %s",auskunft_name);
   system(line);
   fclose(fp);
 }
 return(0);
}


int Backup_plan_write_line_layout(plan,fp,flag)
struct Backup_plaN *plan;
FILE *fp;
int flag;
/*
 bit0-3: format
   0 = according to  plan->list_output_format&(2|4|8) (others do not reflect)
   1 = line layout of directory list
   2 = fixed line layout of split translation list
*/
{
 int format;

 format= flag&15;
 fprintf(fp,"#=# line layout : ");
 if(format==2) {
   fprintf(fp,"partprefix part_address=buffer_address \n");
   fprintf(fp,"#=# partprefix layout :");
   fprintf(fp," **partnumber/total:start_byte+byte_count** \n");
   fprintf(fp,"#=# part_address layout :");
   fprintf(fp," original_dir\"/\"marked_name\"_\"partno\"_\"total\n");
   fprintf(fp,"#=# marked_name layout : prefix\"\"original_name\"\"suffix\n");
   fprintf(fp,"#=# prefix : %s\n",plan->split_file_prefix);
   fprintf(fp,"#=# suffix : %s\n",plan->split_file_suffix);
   fprintf(fp,"#=# buffer_address layout :");
   fprintf(fp," buffer_dir\"/\"marked_name\"_\"partno\"_\"total\"-\"count\n");
   return(1);
 }
 if((plan->list_output_format&2) || format==1)
   fprintf(fp,"type ");
 if(format==1) 
   fprintf(fp,"chmod chown chgrp date ");
 if(plan->list_output_format&4)
   fprintf(fp,"size ");
 if(plan->list_output_format&8)
   fprintf(fp,"crc ");
 fprintf(fp,"adr \n");
 return(1);
}


int Backup_plan_list_header(plan,fp,title,flag)
struct Backup_plaN *plan;
FILE *fp;
char *title;
int flag;
/*
 bit0-3: format for Backup_plan_write_line_layout()
*/
{
 char hname[4096],pwd[4096];

 fprintf(fp,"#=# %s\n",title);
 if(gethostname(hname,sizeof(hname))==-1)
   strcpy(hname,"localhost"); 
 fprintf(fp,"#=# scdbackup version %s\n",PROG_VERSION);
 fprintf(fp,"#=# hostname %s\n",hname);
 fprintf(fp,"#=# id       %s\n",plan->backup_id);
 if(plan->level>=0)
   fprintf(fp,"#=# level    %d  ( -younger_than %s )\n",
               plan->level,Sfile_datestr(plan->precheck_age,1|2|256));
 fprintf(fp,"#=# date     %s\n",Sfile_datestr(plan->start_time,1|2|256));
 Backup_plan_write_line_layout(plan,fp,flag&15);
 if(getcwd(pwd,sizeof(pwd))!=NULL) {
   if(strcmp(pwd,"/")==0)
     pwd[0]= 0;
   fprintf(fp,"#=# next line tells current working directory\n%s/\n",pwd);
 }
 return(1);
}


int Backup_plan_content_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit1= unlink target only, do not create new file
*/
{
 int ret,i,j,popened,linecount= 0,hflag;
 FILE *fp= NULL;
 struct Backup_stuecK *stueck;

 if(plan->content_list_adr[0]==0)
   return(2);
 if(!(flag&2))
   fprintf(stderr,"\ngenerating -content_list_adr %s\n",
                  plan->content_list_adr);

 ret= Open_listfile(&fp,plan->content_list_adr,"w",1|(flag&2));
 if(flag&2)
   return(ret);
 popened= (ret==2);
 if(fp==NULL) {
   fprintf(stderr,"--- %s %s\n",
           "Failed to write content file list (-content_list_adr).",
           "Program aborted.");
   ret= 0; goto ex;
 }
 Backup_plan_list_header(plan,fp,"backup content list",0);
 stueck= plan->stuecke;
 hflag= ((plan->list_output_format&63)<<8)|
        ((plan->list_output_format&(128|256))<<9);
 for(i=0;i<plan->anz_stuecke;i++) {
   fprintf(fp,"#=# volume %d\n", i + 1 + plan->volume_number_offset);
   for(j=0;j<stueck->liste->anz;j++) {
     ret= Fileliste_leanfind(stueck->liste,j,
                             dev_dummy,fp,NULL,&linecount,0,2|16|hflag);
     if(ret<0)
       {ret= 0; goto ex;}
   }
   stueck= stueck->next;
   if(stueck==NULL) 
 break;
 }
 fprintf(fp,"#=# end\n");
 Print_pacifier(linecount,"file items written",1|2|4|8|16);
 fprintf(stderr,"\n");

 ret= 1;
ex:;
 if(fp!=NULL){
   if(popened)
     pclose(fp);
   else
     Sfile_fclose(fp);
 }
 return(ret);
}


int Backup_plan_split_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit1= unlink target only, do not create new file
*/
{
 int ret,i,j,popened,linecount= 0;
 FILE *fp= NULL;
 struct Backup_stuecK *stueck;
 struct FilelistE *liste;
 char target[4096],tempname[4096],src_adr[4096*2],*cpt;

 Backup_plan_count_part_recs(plan,1|2);

 if(plan->split_list_adr[0]==0)
   return(2);
 if(!(flag&2))
   fprintf(stderr,"\ngenerating -split_list_adr %s\n",
                  plan->split_list_adr);

 ret= Open_listfile(&fp,plan->split_list_adr,"w",1|(flag&2));
 if(flag&2)
   return(ret);
 popened= (ret==2);
 if(fp==NULL) {
   fprintf(stderr,"--- %s %s\n",
           "Failed to write content file list (-split_list_adr).",
           "Program aborted.");
   ret= 0; goto ex;
 }
 Backup_plan_list_header(plan,fp,"backup split file translation list",2);
 Print_pacifier(0,"  0          file items written",1|4);
 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++) {
   fprintf(fp,"#=# volume %d\n", i + 1 + plan->volume_number_offset);
   liste= stueck->liste;
   if(liste->part_recs==NULL)
     goto next_stueck;
   for(j=0;j<liste->anz;j++) {
     if(liste->typen[j]<0)
   continue;
     if(liste->part_recs[j]==NULL)
   continue;
     if(Fileliste_target(liste,j,target,0)<=0)
   continue;
     Backup_plan_out_part_mark(plan,fp,liste,j,0);

     /* >>> 0520: escape '=', '\', and newlines */

     fprintf(fp," %s=",target);
     cpt= liste->adressen[j];
     if(cpt[0]=='.'){
       if(cpt[1]=='/') cpt+= 2;
     }
     if(Filelist_part_rec_make_tempname(liste->part_recs[j],
                               plan->split_file_prefix,plan->split_file_suffix,
                               liste->adressen[j],tempname,liste->impl_igno,
                               1|2|((!!plan->precheck_expl_replace)<<2))>0) {
       Backup_plan_get_split_file_dir(plan,&cpt,0);
       sprintf(src_adr,"%s/%s",cpt,tempname);
       cpt= src_adr;
     }

     /* >>> 0520: escape newlines */

     fprintf(fp,"%s\n",cpt);
   }
next_stueck:;
   stueck= stueck->next;
   if(stueck==NULL) 
 break;
 }
 fprintf(fp,"#=# end\n");
 Print_pacifier(linecount,"",1|2|8);
 fprintf(stderr,"\n");

 ret= 1;
ex:;
 if(fp!=NULL){
   if(popened)
     pclose(fp);
   else
     Sfile_fclose(fp);
 }
 return(ret);
}


int Backup_plan_faclfattr_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= getfattr (else getfacl)
 bit1= unlink target only, do not create new file
 bit2= source addresses rather than target addresses
*/
{
 int ret,i,j,popened= 0;
 unsigned textcount= 0;
 char target[4096],cmd[4*4096],from_text[4*4096],to_text[4*4096],adr[4*4096];
 char *list_adr,*list_type,*long_type,*get_cmd,*get_cmd_tree;
 FILE *fp= NULL;
 struct Backup_stuecK *stueck;

 if(flag&1) {
   if(flag&4) {
     list_adr= plan->fattrsrc_list_adr;
     list_type= "-fattrsrc_list_adr";
     long_type= "file extended attribute list (-fattrsrc_list_adr)";
   } else {
     list_adr= plan->fattr_list_adr;
     list_type= "-fattr_list_adr";
     long_type= "file extended attribute list (-fattr_list_adr)";
   }
   get_cmd= "getfattr -P --absolute-names -d";
   get_cmd_tree= "getfattr -R -P --absolute-names -d";
 } else {
   if(flag&4) {
     list_adr= plan->faclsrc_list_adr;
     list_type= "-faclsrc_list_adr";
     long_type= "file access control list (-faclsrc_list_adr)";
   } else {
     list_adr= plan->facl_list_adr;
     list_type= "-facl_list_adr";
     long_type= "file access control list (-facl_list_adr)";
   }
   get_cmd= "getfacl -P --absolute-names";
   get_cmd_tree= "getfacl -R -P --absolute-names";
 }

 if(list_adr[0]==0)
   return(2);
 if(!(flag&2))
   fprintf(stderr,"\ngenerating %s %s\n",list_type,list_adr);

 ret= Open_listfile(&fp,list_adr,"w",1|(flag&2));
 if(flag&2)
   return(ret);
 popened= (ret==2);
 if(fp==NULL) {
   fprintf(stderr,"--- Failed to write %s .\n",list_adr);
   fprintf(stderr,"    Program aborted.\n");
   ret= 0; goto ex;
 }

 for(j=0;j<plan->split_dirs->anz;j++) {
   if(Fileliste_target(plan->split_dirs,j,target,
                       (2*!!Fileliste_target_modern))<=0)
 continue;
   sprintf(cmd,"%s %s",get_cmd,
           Text_shellsafe(plan->split_dirs->adressen[j],adr,0));
   sprintf(from_text,"# file: %s",plan->split_dirs->adressen[j]);
   if(target[0]=='/')
     sprintf(to_text,  "# file: .%s",target);
   else
     sprintf(to_text,  "# file: %s",target);
   if(to_text[strlen(to_text)-1]=='/' && from_text[strlen(from_text)-1]!='/')
     to_text[strlen(to_text)-1]= 0;
   ret= Filter_command(cmd,from_text,to_text,fp,&textcount,!!(flag&4));
   if(ret<=0)
     goto ex;

 }

 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++) {
   for(j=1;j<stueck->liste->anz;j++) {     /* j=1 : do not record ASKME file */
     if(Fileliste_target(stueck->liste,j,target,
                         (2*!!Fileliste_target_modern))<=0)
   continue;

     sprintf(cmd,"%s %s",get_cmd_tree,
                 Text_shellsafe(stueck->liste->adressen[j],adr,0));
     sprintf(from_text,"# file: %s",stueck->liste->adressen[j]);
     if(target[0]=='/')
       sprintf(to_text,  "# file: .%s",target);
     else
       sprintf(to_text,  "# file: %s",target);
     if(to_text[strlen(to_text)-1]=='/' && from_text[strlen(from_text)-1]!='/')
       to_text[strlen(to_text)-1]= 0;
     ret= Filter_command(cmd,from_text,to_text,fp,&textcount,!!(flag&4));
     if(ret<=0)
       goto ex;
 
   }
 }
 ret= 1;
ex:;
 if(!(flag&2))
   fprintf(stderr,"  %-10u file items recorded\n\n",textcount);
 
 if(fp!=NULL){
   if(popened)
     pclose(fp);
   else
     Sfile_fclose(fp);
 }
 return(ret);
}


int Backup_plan_script_header(plan,fpout,script_adr,
                              scriptdescription,rootvar,flag)
struct Backup_plaN *plan;
FILE *fpout;
char *script_adr;
char *scriptdescription;
char *rootvar;
int flag;
/*
 bit0= include a superuser stopper
 bit1= include changed types check
*/
{
 char hname[4096],*cpt;

 fprintf(fpout,"#!%s\n#\n",plan->shell_to_use);
 fprintf(fpout,"# %s\n",scriptdescription);
 if(gethostname(hname,sizeof(hname))==-1)
   strcpy(hname,"localhost"); 
 fprintf(fpout,"# hostname %s\n",hname);
 if(plan->level>=0)
   fprintf(fpout,"# level    %d  ( -younger_than %s )\n",
                 plan->level,Sfile_datestr(plan->precheck_age,1|2|256));
 fprintf(fpout,"# date     %s\n",Sfile_datestr(plan->start_time,1|2|256));
 fprintf(fpout,"# id       %s\n",plan->backup_id);
 if(flag&1) {
   fprintf(fpout,"# Use With Extreme Care !\n#\n");
 } else {
   fprintf(fpout,"# Use With Care\n#\n");
 }
 if(flag&2) {
   if(strlen(script_adr)>0) {
     cpt= strrchr(script_adr,'/');
     if(cpt==NULL)
       cpt= script_adr;
     else
       cpt++;
     fprintf(fpout,"echo \"Script : %s\" >&2\n",cpt);
   }
   fprintf(fpout,"\n");

   /* Include a prepared warning about correct sequence of vanish-install */
   /* The file will eventually be prcessed to change N into Y */
   fprintf(fpout,"warn_changed_types=N\n");

   /* >>> in script : try to check wether level was already installed */;

   fprintf(fpout,"if test \"$SCDBACKUP_RESTORE_PREP_LEVEL\" = %d\n",
                 plan->level);
   fprintf(fpout,"then\n");
   fprintf(fpout,"  dummy=dummy\n");
   fprintf(fpout,"elif test \"$warn_changed_types\" = Y\n");
   fprintf(fpout,"then\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo \"This script will remove some files which changed\"\n");
   fprintf(fpout,"  echo \"their types between backup levels. Therefore THIS\"\n");
   fprintf(fpout,"  echo \"SCRIPT MUST BE RUN FIRST and you will eventually\"\n");
   fprintf(fpout,"  echo \"have to install this backup level again.\"\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo \"Dieses Skript loescht auch einige Files, die ihren Typ\"\n");
   fprintf(fpout,"  echo \"geaendert haben. Deshalb MUSS DIESES SKRIPT ZUERST\"\n");
   fprintf(fpout,"  echo \"AUSGEFUEHRT WERDEN und Sie muessen ggf. danach die\"\n");
   fprintf(fpout,"  echo \"Installation des Backuplevels wiederholen.\"\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo \"Execute the following command and rerun this script:\"\n");
   fprintf(fpout,"  echo \"Folgendes Kommando ausfuehren und Skript neu starten:\"\n");
   fprintf(fpout,"  echo \"  export SCDBACKUP_RESTORE_PREP_LEVEL=%d\"\n",
                 plan->level);

   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo \"Aborted. | Abgebrochen. (exit=4)\" >&2\n");
   fprintf(fpout,"  exit 4\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"fi\n");
   fprintf(fpout,"\n");
   fprintf(fpout,"if test \"$SCDBACKUP_RESTORE_ALLOW_SUPERUSER\" = Y\n");
   fprintf(fpout,"then\n");
   fprintf(fpout,"  dummy=dummy\n");
   fprintf(fpout,"elif test \"$UID\" = 0\n");
   fprintf(fpout,"then\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo 'You are superuser. This is quite dangerous.'\n");
   fprintf(fpout,"  echo 'If you have to insist being superuser, execute'\n");
   fprintf(fpout,"  echo 'export SCDBACKUP_RESTORE_ALLOW_SUPERUSER=Y'\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo 'Sie sind Superuser. Das ist arg gefaehrlich.'\n");
   fprintf(fpout,
               "  echo 'Wenn Sie darauf bestehen muessen, fuehren Sie aus'\n");
   fprintf(fpout,"  echo 'export SCDBACKUP_RESTORE_ALLOW_SUPERUSER=Y'\n");
   fprintf(fpout,"  echo\n");
   fprintf(fpout,"  echo \"Aborted. | Abgebrochen. (exit=1)\" >&2\n");
   fprintf(fpout,"  exit 1\n");
   fprintf(fpout,"fi\n\n");
 }
 fprintf(fpout,"if test \"$%s\" = \"\"\n",rootvar);
 fprintf(fpout,"then\n");
 fprintf(fpout,"  echo\n");
 fprintf(fpout,"  echo 'Variable %s not set or empty'\n",rootvar);
 fprintf(fpout,"  echo 'If you are restoring your backup area under'\n");
 fprintf(fpout,"  echo 'directory /restore/home , execute :'\n");
 fprintf(fpout,"  echo '  export %s=/restore/home'\n",rootvar);
 fprintf(fpout,"  echo\n");
 fprintf(fpout,"  echo 'Variable %s ist leer. Wenn'\n",rootvar);
 fprintf(fpout,"  echo 'Sie ihren Backupbereich unterhalb von Directory'\n");
 fprintf(fpout,"  echo '/restore/home wiederherstellen, fuehren Sie aus :'\n");
 fprintf(fpout,"  echo '  export %s=/restore/home'\n",rootvar);
 fprintf(fpout,"  echo\n");
 fprintf(fpout,"  echo \"Aborted. | Abgebrochen. (exit=2)\" >&2\n");
 fprintf(fpout,"  exit 2\n");
 fprintf(fpout,"else\n");
 fprintf(fpout,"  if cd \"$%s\"\n",rootvar);
 fprintf(fpout,"  then\n");
 fprintf(fpout,"    dummy=dummy\n");
 fprintf(fpout,"  else\n");
 fprintf(fpout,"    echo '--- cannot enter directory'\n");
 fprintf(fpout,"    echo '--- kann nicht ins Directory gehen'\n");
 fprintf(fpout,"    echo '%s='$%s\n",rootvar,rootvar);
 fprintf(fpout,"    echo \"Aborted. | Abgebrochen. (exit=3)\" >&2\n");
 fprintf(fpout,"    exit 3\n");
 fprintf(fpout,"  fi\n");
 fprintf(fpout,"fi\n\n");
 return(1);
}


int Backup_plan_evaluate_precheck(plan,
    kranke_namen,kein_zugriff,mount_punkte,zu_alt,junge_ausnahmen,too_big,
    excluded_for_name,flag)
struct Backup_plaN *plan;
struct FilelistE *kranke_namen;
struct FilelistE *kein_zugriff;
struct FilelistE *mount_punkte;
struct FilelistE *zu_alt;
struct FilelistE *junge_ausnahmen;
struct FilelistE *too_big;
/* 0520 */
struct FilelistE *excluded_for_name;
int flag;
/*
return:
 <0 = error
 bit0= ill name found
 bit1= unreadable found

bit15= abort
*/
{
 int i,ret,idx,partno,partcount,exclude_anyway,will_abort= 0, hflag;
 char *wpt,*rpt,target[4096],new_target[4096];
 double size,start_byte,byte_count;
 struct Filelist_part_reC *part_rec;
 struct FilelistE *liste;

 if(plan->problemlist_limit<kranke_namen->anz ||
    plan->problemlist_limit<kein_zugriff->anz ||
    plan->problemlist_limit<too_big->anz ||
    plan->problemlist_limit<mount_punkte->anz ||
    plan->problemlist_limit<excluded_for_name->anz)
   will_abort= 1;
 for(i=0;i<kranke_namen->anz;i++){
   if(will_abort) {
     fprintf(stderr,"problematic filename : %s\n",kranke_namen->adressen[i]);
 continue;
   }
   /* check wether the name is already in start list */
   if(Fileliste_search_file(plan->startliste,kranke_namen->adressen[i],0)>=0)
 continue;
   if(plan->precheck_unread==1)
     if(Fileliste_search_file(kein_zugriff,kranke_namen->adressen[i],0)
        >=0)
 continue;
   if(Fileliste_search_file(too_big,kranke_namen->adressen[i],0)>=0)
 continue;
   if(plan->precheck_xdev>0)
     if(Fileliste_search_file(mount_punkte,kranke_namen->adressen[i],0)
        >=0)
 continue;
   if(plan->precheck_expl==1 && plan->startliste->impl_igno!=NULL){
     fprintf(stderr,"explicitely appended : %s\n",kranke_namen->adressen[i]);

     if(plan->composition_mode&(1<<15)) {
       Backup_stueck_get_list(plan->aktstueck,&liste,0);
       ret= Backup_plan_append_list_item(plan,kranke_namen,i,0);
     } else {
       liste= plan->startliste;
       ret= Fileliste_copy_item(kranke_namen,i,
                                plan->startliste,plan->startliste->anz,0);
     }
     if(ret<=0)
       goto ex;
     /* if enabled : replace evil characters in target address */
     if(plan->precheck_expl_replace){
       Fileliste_target(liste,liste->anz-1,target,1|2);
       wpt= new_target;
       for(rpt=target;*rpt!=0;rpt++)
         if(strchr(liste->impl_igno,*rpt)!=NULL){
           sprintf(wpt,"+%-2.2X+",*rpt);
           wpt+= strlen(wpt);
         } else
           *(wpt++)= *rpt;
       *wpt= 0;
       Fileliste_set_target(liste,liste->anz-1,new_target,0);
     }

   }else{
     fprintf(stderr,"+++ name will be ignored by backup command : %s\n",
             kranke_namen->adressen[i]);
   }
 }
 for(i=0;i<kein_zugriff->anz;i++){
   if(plan->precheck)
     fprintf(stderr,"+++ no permission to read %s : %s\n",
             (kein_zugriff->typen[i]==2?"directory":"file"),
             kein_zugriff->adressen[i]);
   if(plan->precheck_unread==1 && (!will_abort)
      && !(plan->composition_mode&(1<<15))) {
     fprintf(stderr,"explicitely excluded : %s\n",kein_zugriff->adressen[i]);
     Fileliste_copy_item(kein_zugriff,i,
                         plan->auslassen,plan->auslassen->anz,0);
     plan->autom_excl_counter++;
   }
 }
 for(i=0;i<too_big->anz;i++){
   /* if item is in startliste : delete it there */
   idx= Fileliste_search_file(plan->startliste,too_big->adressen[i],0);
   if(idx>=0)
     Fileliste_defunct_item(plan->startliste,idx,0);
   Fileliste_type_and_size(too_big,i,&size,
                           1|((plan->precheck_sizetest&2)<<1),0);
   if((plan->problemlist_limit<too_big->anz) ||
      ((plan->precheck_sizetest&1) && size>=plan->exclude_size)) {
     fprintf(stderr,
           "+++ single %s too large (%.lf >= -exclude_size %s%.lf) :\n   %s\n",
           (too_big->typen[i]==3?"link target":"file"),size,
           (too_big->typen[i]==3?"-":""),plan->exclude_size,
           too_big->adressen[i]);
     exclude_anyway= 1;
     if(!will_abort)
       fprintf(stderr,"explicitely excluded : %s\n",too_big->adressen[i]);
     plan->autom_excl_counter++;

     /* ??? >>> eventually attach to  plan->uebergrosse  */;

   } else {
     /* register partial files */
     partcount= size/plan->split_file_size;
     if(size-((double) partcount)*plan->split_file_size>0)
       partcount++;
     /* 0497: check wether the file is admissible at all */
     if(plan->precheck_unread==1)
       if(Fileliste_search_file(kein_zugriff,too_big->adressen[i],0)>=0)
         partcount= 0;
     if(plan->precheck_xdev>0)
       if(Fileliste_search_file(mount_punkte,too_big->adressen[i],0)>=0)
         partcount= 0;
     if(too_big->flags[i]&4) /* too old */
       partcount= 0;

     for(partno= 0; partno<partcount; partno++) {
       start_byte= ((double) partno)*plan->split_file_size;
       if(plan->composition_mode&(1<<15)) {
         ret= Backup_plan_append_list_item(plan,too_big,i,0);
         Backup_stueck_get_list(plan->aktstueck,&liste,0);
       } else {
         ret= Fileliste_copy_item(too_big,i,
                                  plan->startliste,plan->startliste->anz,0);
         liste= plan->startliste;
       }
       if(ret<=0)
         goto ex;
       /* attach part_rec */ 
       idx= liste->anz-1;
       if(start_byte+plan->split_file_size>size)
         byte_count= size-start_byte;
       else
         byte_count= plan->split_file_size;
       ret= Filelist_part_rec_new(&part_rec,
                                  partno+1,partcount,start_byte,byte_count,
                                  size, 0);
       if(ret<=0)
         {ret= -1; goto ex;}
       plan->split_count++;
       Filelist_part_rec_set_tempno(part_rec,plan->split_count,0);
       ret= Fileliste_set_part_rec(liste,idx,part_rec,0);
       if(ret<=0)
         {ret= -1; goto ex;}
       Fileliste_set_groessen(liste,idx,byte_count,0);
       Fileliste_target(liste,idx,target,2);
       hflag= (!!plan->precheck_expl_replace)<<2;
       if(strcmp(plan->split_file_dir, "-xorriso_cut_out")==0)
         hflag|= 8;
       Filelist_part_rec_make_tempname(part_rec,
                               plan->split_file_prefix,plan->split_file_suffix,
                               target,new_target,liste->impl_igno, hflag);
       Fileliste_set_target(liste,idx,new_target,0);
       Filelist_part_rec_make_tempadr(part_rec,liste->adressen[idx],
                                  plan->split_file_dir,plan->split_file_prefix,
                                  plan->split_file_suffix,liste->impl_igno,
                                  (!!plan->precheck_expl_replace)<<2);
     }
     plan->auslassen_nicht_startliste= 1;
   }
   Fileliste_copy_item(too_big,i,plan->auslassen,plan->auslassen->anz,0);
 }
 for(i=0;i<mount_punkte->anz;i++){
   if(plan->precheck_xdev>0){
     fprintf(stderr,"+++ mountpoint to filesystem of unwanted type : %s\n",
             mount_punkte->adressen[i]);
     if((!will_abort) && !(plan->composition_mode&(1<<15))) {
       fprintf(stderr,"explicitely excluded : %s\n",mount_punkte->adressen[i]);
       Fileliste_copy_item(mount_punkte,i,
                           plan->auslassen,plan->auslassen->anz,0);
       /* no plan->autom_excl_counter++ here */
     }
   }
 }

 /* 0520 */
 if(excluded_for_name->anz>0)
   fprintf(stderr,
           "\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");
 for(i=0;i<excluded_for_name->anz;i++){
   fprintf(stderr,"+++ inacceptable file address (e.g. with newline):\n%s\n",
           excluded_for_name->adressen[i]);
   idx= Fileliste_search_file(plan->startliste,
                              excluded_for_name->adressen[i],0);
   if(idx>=0)
     Fileliste_defunct_item(plan->startliste,idx,0);
   if((!will_abort) && !(plan->composition_mode&(1<<15))) {
     fprintf(stderr,"explicitely excluded : %s\n",
                    excluded_for_name->adressen[i]);
     Fileliste_copy_item(excluded_for_name,i,
                         plan->auslassen,plan->auslassen->anz,0);
     plan->autom_excl_counter++;
   }
 }
 if(excluded_for_name->anz>0)
   fprintf(stderr,
           "\n++++++++++++++++++++++++++++++++++++++++++++++++++++++++++\n");

 /* react on outcome of age test */
 if(plan->composition_mode&(1<<15)) 
   goto after_age_test_evaluation;
 if(plan->precheck_agetest&1)
   fprintf(stderr,"Options added by -younger_than :\n");
 for(i=0;i<zu_alt->anz;i++){
   fprintf(stderr,"-not %s\n",zu_alt->adressen[i]);
   while(1){ /* remove all occurences of this object from start list */
     idx= Fileliste_search_file(plan->startliste,zu_alt->adressen[i],8);
     if(idx<0)
   break;
     Fileliste_defunct_item(plan->startliste,idx,0);
   }
   if(Fileliste_search_file(plan->startliste,zu_alt->adressen[i],2)>=0){
     Fileliste_copy_item(zu_alt,i,
                         plan->auslassen,plan->auslassen->anz,0);
   }
 }
 Fileliste_garbage(plan->startliste,0);
 /* append junge_ausnahmen to start list if not already listed */;

 /* >>> ??? augment list to hashed list ? */

 if(junge_ausnahmen->anz>0 && plan->prevent_doubles)
   fprintf(stderr,"%d items to be added with -yes ...\n",junge_ausnahmen->anz);
 for(i=0;i<junge_ausnahmen->anz;i++){
   /* fprintf(stderr,"-yes %s\n",junge_ausnahmen->adressen[i]); */
   if(plan->prevent_doubles==0 ||
      Fileliste_search_file(plan->startliste,junge_ausnahmen->adressen[i],0)<0)
     Fileliste_copy_item(junge_ausnahmen,i,
                         plan->startliste,plan->startliste->anz,0);
 }
 if(junge_ausnahmen->anz>0)
   fprintf(stderr,"%d items added with -yes\n",junge_ausnahmen->anz);
 if(plan->precheck_agetest&1)
   plan->auslassen_nicht_startliste= 1;
 ret= Fileliste_special_part_recs(plan->startliste,plan->split_file_dir,
                      plan->split_file_prefix,plan->split_file_suffix,
                      &(plan->split_count),(!!plan->precheck_expl_replace)<<2);
 if(ret<=0)
   {ret= -1; goto ex;}

after_age_test_evaluation:;
 ret= (kranke_namen->anz>0)|((kein_zugriff->anz)<<1);
ex:;
 if(will_abort) {
   fprintf(stderr,"--- Problem list limit (%d) exceeded. Will abort.\n",
                  plan->problemlist_limit);

   if(plan->problemlist_limit<kranke_namen->anz)
   fprintf(stderr,
     "  - too many problematic file names (-filecheck -exclude_unreadable)\n");
   if(plan->problemlist_limit<kein_zugriff->anz)
     fprintf(stderr,
             "  - too many files without read permission (-filecheck -on)\n");
   if(plan->problemlist_limit<too_big->anz)
     fprintf(stderr,
  "  - too many files oversized files (-exclude_size , -split_large_files)\n");
   if(plan->problemlist_limit<mount_punkte->anz)
     fprintf(stderr,
             "  - too many forbidden mount points (-xdev, -localdev)\n");
   fprintf(stderr,
           "--- Above lists of problematic files are probably incomplete.\n");
   ret|= (1<<15);
 }
 return(ret);
}


#ifndef Crawler_ruleS

int Backup_plan_precheck(plan,flag)
struct Backup_plaN *plan;
int flag;
/*

return:
 <0 = error
 bit0= ill name found 
 bit1= unreadable found

 bit15= abort
*/
{
 struct FilelistE *kranke_namen,*kein_zugriff,*mount_punkte;
 struct FilelistE *zu_alt,*junge_ausnahmen,*too_big,*excluded_for_name;
 int ret,size_flag,hflag;
 double precheck_min_sbpf; /* Minimum required : saved bytes per file */
 double check_size;

 fprintf(stderr,"performing file checks\n");

 Fileliste_new(&kranke_namen,0,plan->hauptdir,2);
 Fileliste_new(&kein_zugriff,0,plan->hauptdir,2);
 Fileliste_new(&mount_punkte,0,plan->hauptdir,2);
 hflag= 1|256|1024|(1<<11)|(1<<13);
 Fileliste_new(&zu_alt,0,plan->hauptdir,hflag);
 Fileliste_new(&junge_ausnahmen,0,plan->hauptdir,hflag);
 Fileliste_new(&too_big,0,plan->hauptdir,2);
 Fileliste_new(&excluded_for_name,0,plan->hauptdir,2);
 precheck_min_sbpf= plan->max_size/((double) plan->max_items);

 if(plan->result_filetypes&(1<<31))
   Fileliste_redirect_links(plan->startliste,0);

 size_flag= plan->precheck_sizetest&(1|2);
 check_size= 9e99;
 if(plan->precheck_sizetest&1)
   check_size= plan->exclude_size;
 size_flag|= plan->split_large_files;
 if(plan->split_large_files)
   if(plan->split_file_size<check_size+1)
     check_size= plan->split_file_size+1;
 ret= Fileliste_precheck(plan->startliste,plan->old_tree,
                         kranke_namen,kein_zugriff,mount_punkte,
                         plan->precheck_age,zu_alt,junge_ausnahmen,
                         plan->precheck_max_exempt,precheck_min_sbpf,
                         check_size,too_big,excluded_for_name,
                         ((plan->precheck_xdev==1)<<1)|
                         ((plan->precheck_xdev==2)<<4)|
                         ((!plan->precheck)<<2)|
                         ((plan->precheck_agetest&1)<<3)|
                       (plan->precheck_agetest&
                        (32|64|256|512|1024|2048|16384|32768|(1<<16)|(1<<17)))|
                         (size_flag<<12)|((plan->newline_adr_rule&7)<<19)
                        );
 if(ret<0)
   goto ex;
 ret= Backup_plan_evaluate_precheck(plan,
      kranke_namen,kein_zugriff,mount_punkte,zu_alt,junge_ausnahmen,too_big,
      excluded_for_name,0);
ex:;
 Fileliste_destroy(&kranke_namen,0);
 Fileliste_destroy(&kein_zugriff,0);
 Fileliste_destroy(&mount_punkte,0);
 Fileliste_destroy(&zu_alt,0);
 Fileliste_destroy(&junge_ausnahmen,0);
 Fileliste_destroy(&too_big,0);
 Fileliste_destroy(&excluded_for_name,0);
 if(ret>0) {
   if((ret&1) && plan->precheck_expl==2)
     ret|= 1<<15;
   if((ret&2) && plan->precheck_unread==2)
     ret|= 1<<15;
 }
 return(ret);
}


int Backup_plan_complete_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= generate dirlist (with status information) 
 bit1= with bit 0 : generate dir_script (chown, chmod)
 bit2= unlink target only, do not create new file
 bit3= with bit 0 : generate both forms, script and list
*/
{
 int ret,i,popened= 0,linecount= 0,sparse_adr= 2,list_header_flag= 0;
 int contains_banned_files= 0,popened_aux= 0,find_fp_flag= 0,list_format;
 char *listdescription="",*list_adr="",*option_name="";
 char *listdescription_aux="",*list_adr_aux="",*option_name_aux="";
 FILE *fp= NULL,*fp_aux= NULL;

 if(flag&1) {
   if(flag&8) {
     listdescription= "complete backup area directory list";
     list_adr= plan->dir_list_adr;
     option_name= "-dir_list_adr";
     list_header_flag|= 1;
     listdescription_aux= "script to restore directory properties";
     list_adr_aux= plan->dir_script_adr;
     option_name_aux= "-dir_script_adr";
     find_fp_flag= 32|(1<<14);
   } else if(flag&2) {
     listdescription= "script to restore directory properties";
     list_adr= plan->dir_script_adr;
     option_name= "-dir_script_adr";
     find_fp_flag= 32|64;
   } else {
     listdescription= "complete backup area directory list";
     list_adr= plan->dir_list_adr;
     option_name= "-dir_list_adr";
     list_header_flag|= 1;
     find_fp_flag= 32;
   }
 } else {
   listdescription= "complete backup area file list";
   list_adr= plan->complete_list_adr;
   option_name= "-complete_list_adr";
   find_fp_flag= 0;
 }
 if(list_adr[0]==0)
   return(2);
 if(!(flag&4)) {
   fprintf(stderr,"generating  %s %s\n",option_name,list_adr);
   if(flag&8)
     fprintf(stderr,"       and  %s %s\n",option_name_aux,list_adr_aux);
 }

 ret= Open_listfile(&fp,list_adr,"w",1|(2*!!(flag&4)));
 if(flag&4)
   goto ex;
 popened= (ret==2);
 if(fp==NULL) {
   fprintf(stderr,"--- Failed to write %s (%s).\n",
                  listdescription,option_name);
   ret= 0; goto ex;
 }
 if(flag&8) {
   ret= Open_listfile(&fp_aux,list_adr_aux,"w",1|(2*!!(flag&4)));
   if(flag&4)
     goto ex;
   popened_aux= (ret==2);
   if(fp_aux==NULL) {
     fprintf(stderr,"--- Failed to write %s (%s).\n",
                    listdescription_aux,option_name_aux);
     ret= 0; goto ex;
   }
 }

 if((flag&1) && ((flag&2)&&!(flag&8))) {
   Backup_plan_script_header(plan,fp,list_adr,listdescription,
                             "SCDBACKUP_RESTORE_ROOT",1);
 } else {
   Backup_plan_list_header(plan,fp,listdescription,list_header_flag);
   if((flag&1)&&(flag&8)) 
     Backup_plan_script_header(plan,fp_aux,list_adr_aux,listdescription_aux,
                               "SCDBACKUP_RESTORE_ROOT",1);
 }
 Print_pacifier(0,"  0          file items written",1|4);

 if((flag&(3|8))==3)
   sparse_adr= 0;
 for(i=0;i<plan->startliste->anz;i++) {
   contains_banned_files= 0;
   if(!plan->auslassen_nicht_startliste) {
     if(Fileliste_search_file(plan->startliste->sperrliste,
                              plan->startliste->adressen[i],32)>=0)
 continue;
     if(Fileliste_search_file(plan->startliste->sperrliste,
                              plan->startliste->adressen[i],1|4|8|32)>=0)
       contains_banned_files= 128;
   }
   list_format= ((plan->list_output_format&63)<<8)|
                ((plan->list_output_format&(128|256))<<9);
   ret= Fileliste_leanfind(plan->startliste,i,
                           dev_dummy,fp,fp_aux,&linecount,0,
                           find_fp_flag|sparse_adr|16|
                           (plan->precheck_xdev<<2)|
                           contains_banned_files|list_format);
   if(ret<0)
     {ret= 0; goto ex;}
 }
 if((flag&8)||!((flag&2)&&(flag&1)))
   fprintf(fp,"#=# end\n");
 Print_pacifier(linecount,"file items written",1|2|4|8|16);
 fprintf(stderr,"\n");

 ret= 1;
ex:;
 if(fp!=NULL){
   if(popened)
     pclose(fp);
   else
     Sfile_fclose(fp);
 }
 if(fp_aux!=NULL){
   if(popened_aux)
     pclose(fp_aux);
   else
     Sfile_fclose(fp_aux);
 }
 return(ret);
}


int Backup_plan_vanished_list(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= write shell script
 bit1= write list and shell script
*/
{
 FILE *fpin= NULL,*fpout= NULL,*fpout_aux= NULL,*fp;
 char adr[4096],initial_wd[4096],*adr_arg,*option_name;
 char *adr_arg_aux= "",*option_name_aux= "";
 int ret= 0,in_popened= 0,out_popened= 0,hflag= 0,was_type_change= 0;
 int out_popened_aux= 0;

 if(flag&2) {
   adr_arg= plan->vanish_list_adr;
   adr_arg_aux= plan->vanish_script_adr;
   option_name= "-vanish_list_adr";
   option_name_aux= "-vanish_script_adr";
   hflag= 7;
 } else {
   if(flag&1) {
     adr_arg= plan->vanish_script_adr;
     option_name= "-vanish_script_adr";
     hflag= 1;
   } else {
     adr_arg= plan->vanish_list_adr;
     option_name= "-vanish_list_adr";
   }
 }
 if(plan->old_complete_list_adr[0]==0 || adr_arg[0]==0) 
   {ret= 2; goto ex;}
 if((flag&2) && adr_arg_aux[0]==0) 
   {ret= 2; goto ex;}
 fprintf(stderr,"generating %s %s %s\n",
                option_name,plan->old_complete_list_adr,adr_arg);
 if(flag&2)
   fprintf(stderr,"       and %s %s %s\n",
                  option_name_aux,plan->old_complete_list_adr,adr_arg_aux);

 if(strchr(adr_arg,'/')==NULL && plan->conf_dir[0]
    && plan->level>=0) {
   sprintf(adr,"%s/level_%d/%s",plan->conf_dir,plan->level,adr_arg);
   strcpy(adr_arg,adr);
 }
 out_popened= Open_listfile(&fpout,adr_arg,"w",1);
 if(out_popened!=2)
   out_popened= 0;
 if(fpout==NULL) {
   fprintf(stderr,"--- %s %s%s%s\n","Failed to write list of vanished files",
                  "(",option_name,"). Program aborted.");
   ret= 0; goto ex;
 }
 if(flag&2) {
   if(strchr(adr_arg,'/')==NULL && plan->conf_dir[0]
      && plan->level>=0) {
     sprintf(adr,"%s/level_%d/%s",plan->conf_dir,plan->level,adr_arg_aux);
     strcpy(adr_arg_aux,adr);
   }
   out_popened= Open_listfile(&fpout_aux,adr_arg_aux,"w",1);
   if(out_popened_aux!=2)
     out_popened_aux= 0;
   if(fpout_aux==NULL) {
     fprintf(stderr,"--- %s %s%s%s\n","Failed to write list of vanished files",
                    "(",option_name_aux,"). Program aborted.");
     ret= 0; goto ex;
   }
 }
 if((!(flag&1))||(flag&2))
   Backup_plan_list_header(plan,fpout,"vanished file objects list",0);

 if(strchr(plan->old_complete_list_adr,'/')==NULL && plan->conf_dir[0]) {
   if(plan->level>=1) {
     sprintf(adr,"%s/level_%d/%s",plan->conf_dir,plan->level-1,
                                   plan->old_complete_list_adr);
     strcpy(plan->old_complete_list_adr,adr);
   } else {
     if((flag&1)||(flag&2)) {
       if(flag&2)
         fp= fpout_aux;
       else
         fp= fpout;
       fprintf(fp,"#!%s\n#\n",plan->shell_to_use);
       fprintf(fp,
            "# -level 0 dummy of script that removes vanished files\n");
       fprintf(fp,
            "# See incremental backups of higher levels for the real thing\n");
       fprintf(fp,"exit 0\n");
     }
     ret= 2; goto ex;
   }
 }
 in_popened= Open_listfile(&fpin,plan->old_complete_list_adr,"r",1);
 if(in_popened!=2)
   in_popened= 0;
 if(fpin==NULL) {
   fprintf(stderr,"--- %s%s%s %s\n",
           "Failed to read complete file list (from ",option_name,").",
           "Program aborted.");
   ret= 0; goto ex;
 }

 if(flag&1) {
   if(flag&2)
     fp= fpout_aux;
   else
     fp= fpout;
   Backup_plan_script_header(plan,fp,adr_arg,
                    "script to remove files that vanished between two backups",
                    "SCDBACKUP_RESTORE_ROOT",1|2);
 }

 if(!(plan->precheck_agetest&512)) 
   hflag|= (1<<15);
 hflag|= ((plan->list_output_format&63)<<8);

 /* >>> ??? always enable :md5: with vanish list ?  hflag|= (8|16)<<8 */;
 /* always disable :inode: */
 hflag&= ~(1<<13);
 /* >>> ??? what about  list_output_format&(64|128): mtime, ctime ? */;

 ret= Fileliste_read_serialized_tree(NULL,fpin,plan->old_tree,fpout,fpout_aux,
                                     initial_wd,NULL,"",hflag);
 fprintf(stderr,"\n");
 if(ret<=0)
   goto ex;
 if(ret&256)
   was_type_change= 1;

 ret= 1;
ex:;
 if(fpout!=NULL && ret>0 && (flag&3)!=1)
   fprintf(fpout,"#=# end\n");
 if(in_popened) {
   if(fpin!=NULL)
     pclose(fpin);
 } else
   Sfile_fclose(fpin);
 if(out_popened) {
   if(fpout!=NULL)
     pclose(fpout);
 } else
   Sfile_fclose(fpout);
 if(out_popened_aux) {
   if(fpout_aux!=NULL)
     pclose(fpout_aux);
 } else
   Sfile_fclose(fpout_aux);
 if((flag&1) && was_type_change && ret>0) {
   /* 0520 : This fails sometime. But it shall not hamper the backup. */
   Backup_plan_enable_change_warning(plan,adr_arg,0);
 }
 return(ret);
}

#endif /* ! Crawler_ruleS */


int Backup_plan_enable_change_warning(plan,script_name,flag)
struct Backup_plaN *plan;
char *script_name;
int flag;
{
 int in_popened= 0,out_popened= 0,al,ret,errn;
 FILE *fpin= NULL,*fpout= NULL;
 char tmp_adr[4096],errmsg[3*4096];
 unsigned textcount= 0,writecount= 0;

 in_popened= Open_listfile(&fpin,script_name,"r",1);
 if(in_popened<=0)
   {ret= 0; goto ex;}
 if(in_popened!=2)
   in_popened= 0;
 strcpy(tmp_adr,script_name);
 al= strlen(tmp_adr);
 if(al>3 && ( strcmp(tmp_adr+(al-3),".gz")==0 ||
              strcmp(tmp_adr+(al-3),".GZ")==0 ))
   strcpy(tmp_adr+al-3,"_filtered.gz");
 else
   strcpy(tmp_adr+al,"_filtered");
 out_popened= Open_listfile(&fpout,tmp_adr,"w",1);
 if(out_popened<=0)
   {ret= 0; goto ex;}
 if(out_popened!=2)
   out_popened= 0;
 ret= Filter_fp(fpin,fpout,"warn_changed_types=N","warn_changed_types=Y",
                &textcount,&writecount,2);
 if(ret>0) {
   ret= Sfile_move_file(tmp_adr,script_name,&errn,errmsg,0);
   if(ret<=0) {
     fprintf(stderr,"%s\n",errmsg);
     {ret= 0; goto ex;}
   }
 }
 ret= 1;
ex:;
 if(fpin!=NULL) {
   if(in_popened)
     pclose(fpin);
   else
     Sfile_fclose(fpin);
 }
 if(fpout!=NULL) {
   if(out_popened)
     pclose(fpout);
   else
     Sfile_fclose(fpout);
 }
 return(ret);
}


int Backup_plan_open_list(plan,list_adr,option_name,list_description,listno,
                          fp,opened,flag)
struct Backup_plaN *plan;
char *list_adr;
char *option_name;
char *list_description;
int listno;  /* list type and index to bit in opened
                0=complete_list , 1=dir list , 2=dir script , 3=vanish list
                4=vanish script , 5=content list */
FILE **fp;
int *opened;
/* bit0-15 = opened , eventual with bit16-31= opened by popen */
int flag;
/*
 bit0= do not write header
 bit2= unlink target only, do not create new file
*/
{
 int ret,list_header_flag= 0,script_header_flag= 1;

 (*opened)&= ~((1<<(listno+16))|(1<<listno));
 (*fp)= NULL;
 if(list_adr[0]==0)
   return(2);
 if(!(flag&4))
   fprintf(stderr,"generating  %s %s\n",option_name,list_adr);
 ret= Open_listfile(fp,list_adr,"w",1|(2*!!(flag&4)));
 if(ret<=0)
   return(ret);
 if(flag&4)
   return(1);
 (*opened)|= ((ret==2)<<(listno+16))|(1<<listno);
 if(flag&1)
   return(1);
 /* write header */;
 if(listno==0 || listno==1 || listno==3 || listno==5) {
   if(listno==1)
     list_header_flag|= 1;
   ret= Backup_plan_list_header(plan,*fp,list_description,list_header_flag);
 } else if(listno==2 || listno==4) {
   if(listno==4)
     script_header_flag|= 2;
   ret= Backup_plan_script_header(plan,*fp,list_adr,list_description,
                                  "SCDBACKUP_RESTORE_ROOT",script_header_flag);
 }
 if(ret<=0)
   return(ret);
 return(1);
}


int Backup_plan_first_piece(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 int ret,stack_count;

 if(plan->stuecke==NULL) {
   Backup_plan_stacked_max_size(plan,&(plan->max_size),&stack_count,0);
   ret= Backup_stueck_new(&(plan->stuecke),plan,NULL,
                          plan->max_items,plan->max_size,plan->hauptdir,0);
   if(ret<=0)
     return(ret);
   plan->aktstueck= plan->stuecke;
   plan->anz_stuecke= 1;
   ret= Backup_plan_stueck_header(plan,0);
   if(ret<=0)
     return(ret);
 }
 if(plan->aktstueck==NULL)  
   return(0);
 return(1);
}


int Backup_plan_init_crawler(plan,o,flag)
struct Backup_plaN *plan;
struct CrawleR *o;
int flag;
{
 /* initialize the possibly self managed memory pointers of a crawler */
 o->boss= plan;
 o->start_list= o->old_tree= o->exclude_list= NULL;
 o->ill_names= o->no_access= o->mount_points= o->too_old= NULL;
 o->young_exempts= o->too_big= o->excluded_for_name= NULL;
 o->completelist_fp= o->dirlist_fp= o->dirscript_fp= NULL;
 o->vanishlist_fp= o->vanishscript_fp= o->contentlist_fp= NULL;
 o->completelist_count= o->dirlist_count= o->vanishlist_count= 0;
 o->type_change_count= 0;
 o->py_stack= NULL;
 o->traverse_count= o->contentlist_count= 0;
 return(1);
}


int Backup_plan_run_crawler(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
return:
 <=0 = error
 bit15= abort (due to exceeded problem limits)
*/
{
 int ret,size_flag,opened,print_end,i,has_work;
 int vanish_script_written= 0,type_changed= 0;
 struct CrawleR crawler_data,*o;
 double check_size;
 FILE **fppt;

 o= &crawler_data;
 Backup_plan_init_crawler(plan,o,0);

 /* is there something to do at all ? */
 has_work= 0;
 if(plan->old_complete_list_adr[0]!=0 && plan->old_vanish_list_adr[0]!=0)
   has_work|= 1;
 if(plan->complete_list_adr[0]!=0 ||
    plan->dir_list_adr[0]!=0 ||
    plan->dir_script_adr[0]!=0 ||
    plan->vanish_list_adr[0]!=0 ||
    plan->vanish_script_adr[0]!=0)
   has_work|= 2;
 if(plan->precheck>0 || plan->precheck_xdev || (plan->precheck_agetest&1)
    || (plan->precheck_sizetest&1) || plan->split_large_files)
   has_work|= 4;
 if(plan->composition_mode&(1<<15))
   has_work|= 8;
 if(has_work<=0)
   {ret= 2; goto ex;}

 ret= Backup_plan_load_old_tree(plan,plan->old_complete_list_adr,
                                     plan->old_vanish_list_adr,"",0);
 if(ret<=0)
   {ret= -22; goto ex;}

 if(has_work<=1)
   {ret= 2; goto ex;}

 fprintf(stderr,"tree crawler performing the following tasks:\n");
 o->old_tree= plan->old_tree;
 o->start_list= plan->startliste;
 o->exclude_list= plan->auslassen;
 o->exclude_not_startlist= plan->auslassen_nicht_startliste;
 o->age_limit= plan->precheck_age;
 o->max_exempts= plan->precheck_max_exempt;
 o->min_saved_bytes_per_file= plan->max_size/((double) plan->max_items);
 check_size= 9e99;
 if(plan->precheck_sizetest&1)
   check_size= plan->exclude_size;
 size_flag= plan->precheck_sizetest&(1|2);
 size_flag|= plan->split_large_files;
 if(plan->split_large_files)
   if(plan->split_file_size<check_size+1)
     check_size= plan->split_file_size+1;
 o->exclude_size= check_size;
 o->precheck_flag= ((plan->precheck_xdev==1)<<1)|
                   ((plan->precheck_xdev==2)<<4)|
                   (((!plan->precheck)||(plan->precheck_unread==-1))<<2)|
                   ((plan->precheck_agetest&1)<<3)|
                   (plan->precheck_agetest&
                       (32|64|256|512|1024|2048|16384|32768|
                        (1<<16)|(1<<17)|(1<<18)))|
                   (size_flag<<12);
 o->problemlist_limit= plan->problemlist_limit;
 o->linecount= 0;
 o->what_lists_flag= 0;
 if(plan->complete_list_adr[0]!=0)
   o->what_lists_flag|= 1;
 if(plan->dir_list_adr[0]!=0)
   o->what_lists_flag|= 2;
 if(plan->dir_script_adr[0]!=0)
   o->what_lists_flag|= 4;
 if(plan->vanish_list_adr[0]!=0)
   o->what_lists_flag|= 8;
 if(plan->vanish_script_adr[0]!=0)
   o->what_lists_flag|= 16;
 o->output_format_flag= ((plan->list_output_format&63)<<8)|256|
                        ((plan->list_output_format&(128|256))<<9);

 /* create classification lists */
 ret= Fileliste_new(&(o->ill_names),0,plan->hauptdir,2);
 if(ret<=0)
   goto ex;
 ret= Fileliste_new(&(o->no_access),0,plan->hauptdir,2);
 if(ret<=0)
   goto ex;
 ret= Fileliste_new(&(o->mount_points),0,plan->hauptdir,2);
 if(ret<=0)
   goto ex;
 ret= Fileliste_new(&(o->too_old),0,plan->hauptdir,2);
 if(ret<=0)
   goto ex;
 ret= Fileliste_new(&(o->young_exempts),0,plan->hauptdir,2);
 if(ret<=0)
   goto ex;
 ret= Fileliste_new(&(o->too_big),0,plan->hauptdir,2);
 if(ret<=0)
   goto ex;
 ret= Fileliste_new(&(o->excluded_for_name),0,plan->hauptdir,2);
 if(ret<=0)
   goto ex;

 /* prepare output list files */
 opened= 0;
 ret= Backup_plan_open_list(plan,plan->complete_list_adr,
            "-complete_list_adr","complete backup area file list",0,
                       &(o->completelist_fp),&opened,0);
 if(ret<=0)
   goto ex;
 ret= Backup_plan_open_list(plan,plan->dir_list_adr,
            "-dir_list_adr","complete backup area directory list",1,
                       &(o->dirlist_fp),&opened,0);
 if(ret<=0)
   goto ex;
 ret= Backup_plan_open_list(plan,plan->dir_script_adr,
            "-dir_script_adr","script to restore directory properties",2,
            &(o->dirscript_fp),&opened,0);
 if(ret<=0)
   goto ex;
 ret= Backup_plan_open_list(plan,plan->vanish_list_adr,
            "-vanish_list_adr","vanished file objects list",3,
            &(o->vanishlist_fp),&opened,0);
 if(ret<=0)
   goto ex;
 ret= Backup_plan_open_list(plan,plan->vanish_script_adr,
             "-vanish_script_adr",
             "script to remove files that vanished between two backups",4,
             &(o->vanishscript_fp),&opened,0);
 if(ret<=0)
   goto ex;
 if(o->vanishscript_fp!=NULL)
   vanish_script_written= 1;

 o->current_piece= o->pieces= NULL;
 o->py_stack= NULL;
 o->py_stack_piecemark= o->py_list_piecemark= 0;
 o->composition_mode= plan->composition_mode;
 o->max_size= plan->max_size;
 o->max_items= plan->max_items;
 o->min_size_fak= plan->min_size_fak;
 o->py_size= 0.0;
 o->py_count= 0;
 o->split_dirs= plan->split_dirs;
 o->piece_count= 0;
 o->pieces= o->current_piece= NULL;
 o->contentlist_fp= NULL;
 if(o->composition_mode&(1<<15)) {
   ret= Backup_plan_first_piece(plan,0);
   if(ret<=0)
     goto ex;
   o->piece_count= plan->anz_stuecke;
   o->pieces= plan->stuecke;
   o->current_piece= plan->aktstueck;
   ret= Fileliste_new(&(o->py_stack),10,plan->hauptdir,256|512|2048|8196);
   if(ret<=0)
     goto ex;
   ret= Backup_plan_open_list(plan,plan->content_list_adr,
                              "-content_list_adr",
                              "backup content list",5,
                              &(o->contentlist_fp),&opened,0);
   if(ret<=0)
     goto ex;
   o->contentlist_count= 0;
 }

 if(has_work&8)
   fprintf(stderr,"composing on the fly\n");
 if(has_work&4)
   fprintf(stderr,"performing file checks\n");

 ret= Crawler_do_wood(o,0);
 if(ret<=0)
   goto ex;
 if(o->type_change_count>0)
   type_changed= 1;

 if(plan->composition_mode&(1<<15)) {
   plan->anz_stuecke= o->piece_count;
   plan->stuecke= o->pieces;
   plan->aktstueck= o->current_piece;
   /* write content list of last piece */
   if(o->contentlist_fp!=NULL) {
     ret= Backup_stueck_content_list(o->current_piece,o->piece_count,
                                     o->contentlist_fp,&(o->contentlist_count),
                                     o->output_format_flag,1);
     if(ret<=0)
       goto ex;
     Print_pacifier(o->contentlist_count,"content list lines written",1|2|4|8);
   }
 }
 /* draw consequences from filelists */
 ret= Backup_plan_evaluate_precheck(plan,
          o->ill_names,o->no_access,o->mount_points,
          o->too_old,o->young_exempts,o->too_big,o->excluded_for_name,0);
 if(ret<0)
   goto ex;
 if(ret&(1<<15))
   {ret= 1<<15; goto ex;} /* abort run */
 ret= 1;
ex:;
 for(i=0;i<6;i++) {
   print_end= 0;
   if(i==0) {
     fppt= &(o->completelist_fp);
     print_end= 1;
   } else if(i==1) {
     fppt= &(o->dirlist_fp);
     print_end= 1;
   } else if(i==2) {
     fppt= &(o->dirscript_fp);
   } else if(i==3) {
     fppt= &(o->vanishlist_fp);
     print_end= 1;
   } else if(i==4) {
     fppt= &(o->vanishscript_fp);
   } else {
     fppt= &(o->contentlist_fp);
     print_end= 1;
   }
   if((*fppt)!=NULL) {
     if(ret>0 && print_end) 
       fprintf(*fppt,"#=# end\n");
     if(opened&(1<<(16+i)))
       pclose(*fppt);
     else
       Sfile_fclose(*fppt);
     (*fppt)= NULL;
   }
 }
 Fileliste_destroy(&(o->ill_names),0);
 Fileliste_destroy(&(o->no_access),0);
 Fileliste_destroy(&(o->mount_points),0);
 Fileliste_destroy(&(o->too_old),0);
 Fileliste_destroy(&(o->young_exempts),0);
 Fileliste_destroy(&(o->too_big),0);
 Fileliste_destroy(&(o->excluded_for_name),0);
 Fileliste_destroy(&(o->py_stack),0);
 if(ret>0)
   if(vanish_script_written && type_changed) {

/* 0520 : no idea why this fails sometime. But it shall not hamper the backup.
*/
     Backup_plan_enable_change_warning(plan,plan->vanish_script_adr,0);
   }
 return(ret);
}


int Backup_plan_upgrade_cfl(plan,in_listadr,out_listadr,flag)
struct Backup_plaN *plan;
char *in_listadr;
char *out_listadr;
int flag;
/* bit0= cull rather than ubgrade */
{
 int ret,opened;
 struct CrawleR crawler_data,*o;
 FILE **fppt;

 o= &crawler_data;
 Backup_plan_init_crawler(plan,o,0);

 /* load content_file_list into old_tree, memorize header lines */
   /* >>> ??? how to preserve volume markers and volume sequence ? */
 ret= Backup_plan_load_old_tree(plan,in_listadr,"_","",1|2);
 if(ret<=0)
   goto ex;
 o->old_tree= plan->old_tree;

 /* program crawler */
 o->output_format_flag= ((plan->list_output_format&8)<<8)|256|512|1024|8192|
                        ((plan->list_output_format&(128|256))<<9);
 /* >>> ??? set timestamp from header ? */
 o->age_limit= time(0);
 if(plan->precheck_age>0) 
   o->age_limit= plan->precheck_age;

 /* flag bit0=0 : enhance unchanged files by inode numbers 
             =1 : invalidate changed or vanished files
 */
 ret= Crawler_traverse_old_tree(o,o->old_tree,"","",1<<(11+2*(flag&1)));
 if(ret<=0)
   goto ex;

 /* write list lines */
 opened= 0;
 ret= Backup_plan_open_list(plan,out_listadr,
                            "-test_command upgrade_content_file_list","",0,
                            &(o->contentlist_fp),&opened,1);
 if(ret<=0)
   goto ex;
 fprintf(o->contentlist_fp,"%s",Cd_backup_planer_list_heaD);
 ret= Crawler_traverse_old_tree(o,o->old_tree,"","",4096);
 if(ret<=0)
   goto ex;

 ret= 1;
ex:
 fppt= &(o->contentlist_fp);
 if((*fppt)!=NULL) {
   if(ret>0) 
     fprintf(*fppt,"#=# end\n");
   if(opened&(1<<16))
     pclose(*fppt);
   else
     Sfile_fclose(*fppt);
   (*fppt)= NULL;
 }
 return(ret);
}


#ifdef Cdbackup_planer_iso_C
int Backup_plan_read_old_list(struct Backup_plaN *plan, char *adr,
                              char initial_wd[4096],
                              struct FilelistE **initial_wd_list, int flag)
#else /* Cdbackup_planer_iso_C */
int Backup_plan_read_old_list(plan,adr,initial_wd,initial_wd_list,flag)
struct Backup_plaN *plan;
char *adr;
char initial_wd[4096];
struct FilelistE **initial_wd_list;
int flag;
#endif /* ! Cdbackup_planer_iso_C */
/*
 bit1= load header information into Cd_backup_planer_list_heaD
 bit4= delete files from liste tree rather than to insert them
 bit5= this initial_wd and initial_wd_list are already set
 bit6= demand complete list and no content list
 bit14= choose target addresses rather than source addresses
       use text given with  initial_wd  as prefix_dir
 bit18= this is level 0 
*/
{
 FILE *fpin= NULL;
 int ret= 0,popened= 0,memorize_header= 0;

 if(adr[0]==0)
   return(0);

 popened= Open_listfile(&fpin,adr,"r",1);
 if(popened!=2)
   popened= 0;
 if(fpin==NULL) {
   fprintf(stderr,"--- %s%s%s\n",
           "Failed to read file list (from -load_old_tree : ",adr,
           "). Program aborted.");
   ret= 0; goto ex;
 }

 if(flag&2)
   memorize_header= 1<<17;
 fprintf(stderr,"List file opened :\n%s\n",adr);
 ret= Fileliste_read_serialized_tree(
          plan->old_tree,fpin,NULL,NULL,NULL,initial_wd,initial_wd_list,"",
          2|(flag&(16|32|64|(1<<14)|(1<<18)))|memorize_header);
 if(ret<=0)
   goto ex;
 if((flag&64) && ret==2) 
   goto ex;

 ret= 1;
ex:;
 if(popened) {
   if(fpin!=NULL)
     pclose(fpin);
 } else
   Sfile_fclose(fpin);
 return(ret);
}


int Backup_plan_load_old_tree(plan,old_list_adr,old_vanish_adr,tree_prefix,
                              flag)
struct Backup_plaN *plan;
char *old_list_adr;
char *old_vanish_adr;
char *tree_prefix;
int flag;
/*
 bit0= plainly load list (no layering), prepare for inodes and crc
 bit1= load header information into Cd_backup_planer_list_heaD
 bit14= choose target addresses rather than source addresses
        use tree_prefix
*/
{
 int ret= 0,demand_complete= 0,not_from_scratch= 0,level_mode= 0,i;
 int target_mode=0,no_crc;
 char adr[4096],initial_wd[4096];
 struct FilelistE *initial_wd_list= NULL;
 
 initial_wd[0]= 0;
 if(old_list_adr[0]==0 || old_vanish_adr[0]==0)
   return(2);
 strcpy(adr,old_list_adr);
 if(flag&1)
   fprintf(stderr,"loading %s\n",old_list_adr);
 else
   fprintf(stderr,"performing -load_old_tree %s %s\n",
                  old_list_adr,old_vanish_adr);

 /* first try wether this is a complete list */
 if(!(flag&1))
   if(strchr(old_list_adr,'/')==NULL && plan->conf_dir[0] && plan->level>=0) {
    if(plan->level==0)
      return(2);
     sprintf(adr,"%s/level_%d/%s",plan->conf_dir,plan->level-1,old_list_adr);
     level_mode= 1;
   }
 if(plan->old_tree==NULL) {
   no_crc= (plan->precheck_agetest^(1<<11))&(1<<11);
   if(flag&1)
     no_crc= 0;
   ret= Fileliste_new(&(plan->old_tree),0,plan->hauptdir,
                      (Fileliste_hash_sizE>0)|256|no_crc);
   if(ret<=0)
     goto ex;
   not_from_scratch= 0;
 } else
   not_from_scratch= 32;
 demand_complete= (plan->level>1 && level_mode && !(flag&1))<<6;
 target_mode= (flag&(1<<14));
 if(target_mode)
   strcpy(initial_wd,tree_prefix);
 ret= Backup_plan_read_old_list(plan,adr,initial_wd,&initial_wd_list,
                  (flag&2) | target_mode | not_from_scratch | demand_complete);
 if(ret<=0)
   goto ex;
 if(demand_complete && ret==2) { 
   fprintf(stderr,"\nGenerating old_tree from layered content/vanish lists\n");
   /* reconstruct complete list tree from content lists and vanish lists */
   for(i=0;i<plan->level;i++) {
     if(i>0) {
       /* read vanished list from level */;
       sprintf(adr,"%s/level_%d/%s",plan->conf_dir,i,old_vanish_adr);
       ret= Backup_plan_read_old_list(plan,adr,initial_wd,&initial_wd_list,
                                      16 | target_mode | not_from_scratch);
       if(ret<=0)
         goto ex;
     }
     /* read content list from level */;
     sprintf(adr,"%s/level_%d/%s",plan->conf_dir,i,old_list_adr);
     ret= Backup_plan_read_old_list(plan,adr,initial_wd,&initial_wd_list,
                                    target_mode | not_from_scratch |
                                    ((i==0)<<18));
     if(ret<=0)
       goto ex;
     not_from_scratch= 32;
   }
 }
 fprintf(stderr,"\n");

 ret= 1;
ex:;
 if(ret<=0) {
   fprintf(stderr,"--- failed : -load_old_tree %s %s\n",
                  old_list_adr,old_vanish_adr);
 }
 return(ret);
}


int Backup_plan_count_part_recs(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= set ->tempno
 bit1= set ->tempno only if 0
*/
{
 int i,split_file_count= 0,ret,count= 0;
 struct Backup_stuecK *stueck;

 stueck= plan->stuecke;
 split_file_count= 0;
 for(i=0;i<plan->anz_stuecke;i++){
   ret= Fileliste_count_part_recs(stueck->liste,&split_file_count,flag&(1|2));
   if(ret>=0)
     count+= ret;
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 return(count);
}


int Backup_plan_skript(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 int ret,i,parts_made= 0,dummy, volno, volcount;
 struct FilelistE *liste;
 struct Backup_stuecK *stueck;
 char *auskunft_name,pwd[4096],shell_safe[3*4096],num[81];
 char fname[4096],lname[4096];
 FILE *fp;

 volcount= plan->anz_stuecke + plan->volume_number_offset;
 auskunft_name= plan->auskunft_name->adressen[0];

 /* ---------------------- Information --------------------- */

 if(plan->auskunft_an) {
   ret= Backup_plan_info_skript(plan,0,0);
   if(ret) return(ret);
   if(plan->auskunft_individuell){
     ret= Backup_plan_info_skript(plan,0,1);
     if(ret) return(ret);
   }
 }
 

 /* ------------------- Production -------------------- */

 if(plan->output_form==1 || plan->output_form==4) {
   ret= Backup_plan_stdout(plan,(plan->output_form==4)<<1);
   if(ret<=0) ret= 1;
   else       ret= 0;
   return(ret);
 }

 fp= stdout;
 fprintf(fp,"#!%s\n\n",plan->shell_to_use);

 if(getcwd(pwd,sizeof(pwd))!=NULL)
   fprintf(fp,"\ncd %s\n\n",Text_shellsafe(pwd,shell_safe,0));

 fprintf(fp,"export SCDBACKUP_ACTIVE_SCRIPT\n");
 fprintf(fp,
         "export SCDBACKUP_PIECENO SCDBACKUP_PIECES SCDBACKUP_PIECE_OFFSET\n");
 fprintf(fp,"SCDBACKUP_ACTIVE_SCRIPT=\"$0\"\n");
 fprintf(fp,"SCDBACKUP_PIECES=%d\n", volcount);
 fprintf(fp,"SCDBACKUP_PIECE_OFFSET=%d\n", plan->volume_number_offset);

 fprintf(fp,"\n");
 fprintf(fp,"is_enabled() {\n");
 fprintf(fp,"  part=\"$1\"\n");
 fprintf(fp,"  if test \"$part\" = \"\" ; then return 0 ; fi\n");
 if(strlen(plan->skript_status_file)){
   fprintf(fp,"  if test \"$part\" = AUTO ; then\n");
   fprintf(fp,"    part=\n");
   fprintf(fp,"    if test -r %s ; then\n",plan->skript_status_file);
   fprintf(fp,"      part=$(cat %s)\n",plan->skript_status_file);
   fprintf(fp,"    fi\n");
   fprintf(fp,"    if test \"$part\" = \"\" ; then\n");
   fprintf(fp,"      echo 'no usable status file : '%s >&2\n",
              plan->skript_status_file);  
   fprintf(fp,"      echo 'number of start volume needed' >&2\n");
   fprintf(fp,"      echo 'example:  -resume %d' >&2\n", volcount);
   fprintf(fp,"      exit 12\n");
   fprintf(fp,"    fi\n");
   fprintf(fp,"  fi\n");
 }  
 fprintf(fp,"  if test \"$part\" -lt 0 ; then\n");
 fprintf(fp,"    if test \"$part\" -eq -\"$2\" ; then return 0 ; fi\n");
 fprintf(fp,"    if test -%d -gt \"$part\" ; then exit 11 ; fi\n", volcount);
 if(plan->volume_number_offset)
   fprintf(fp,"    if test -%d -lt \"$part\" ; then exit 11 ; fi\n",
           plan->volume_number_offset + 1);
 fprintf(fp,"    return 1\n");
 fprintf(fp,"  fi\n");
 fprintf(fp,"  if test \"$part\" -le \"$2\" ; then return 0 ; fi\n");
 fprintf(fp,"  if test %d -lt \"$part\" ; then exit 11 ; fi\n", volcount);
 fprintf(fp,"  return 1\n");
 fprintf(fp,"}\n\n");

 stueck= plan->stuecke;
 dummy= 0;
 for(i=0;i<plan->anz_stuecke;i++){
   if(Fileliste_count_part_recs(stueck->liste,&dummy,0)>0)
 break;
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 if(i<plan->anz_stuecke && stueck!=NULL)
   Fileliste__print_split_cleanup(fp,plan->split_file_dir,
                            plan->split_file_prefix,plan->split_file_suffix,1);

 fprintf(fp,"if test \"$1\" = -count ; then\n");
 fprintf(fp,"  echo %d", volcount);
 if(plan->volume_number_offset)
   fprintf(fp," - %d", plan->volume_number_offset);
 fprintf(fp,"\n");
 fprintf(fp,"  exit 0\n");
 fprintf(fp,"fi\n\n");


 liste= plan->uebergrosse;
 for(i=0;i<liste->anz;i++){
   fprintf(fp,"echo '+++++++++++ oversized file (%.lf bytes) : %s'\n",
              liste->groessen[i],liste->adressen[i]);
 }
 if(liste->anz>0) fprintf(fp,"\n");

 stueck= plan->stuecke;
 for(i=0;i<plan->anz_stuecke;i++){
   volno= i + 1 + plan->volume_number_offset;
   fprintf(fp,"if is_enabled \"$1\" %d ; then\n", volno);
   if(strlen(plan->skript_status_file))
     fprintf(fp,"echo %d >%s\n", volno, plan->skript_status_file);
   fprintf(fp,"SCDBACKUP_PIECENO=%d\n", volno);
   if(stueck->liste->part_recs!=NULL) {
     parts_made= Fileliste_make_parts(stueck->liste,fp,plan->cut_file_cmd,
                                  plan->split_file_dir,plan->split_file_prefix,
                                  plan->split_file_suffix,&plan->split_count,
                                  2|((!!plan->precheck_expl_replace)<<2));
   }
   if(plan->autom_excl_counter>0) {
      fprintf(fp,"echo\n");
      fprintf(fp,"echo '++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++'\n");
      if(plan->autom_excl_counter>1)
        fprintf(fp,"echo '+++ Note: %d files or directories were automatically excluded from this backup'\n",
                plan->autom_excl_counter);
      else
        fprintf(fp,"echo '+++ Note: 1 file or directory was automatically excluded from this backup'\n");
      fprintf(fp,"echo '++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++'\n");
      fprintf(fp,"echo\n");
   }
   if(plan->compact_num)
     Sfile_scale(stueck->bytes,num,7,1e5,1);
   else
     sprintf(num,"%.lf",stueck->bytes);
   fprintf(fp,"echo '########## part %d of %d (%d items  %s bytes)'\n",
           volno, volcount, stueck->liste->anz,num);
   if(plan->precheck_agetest&1) {
     fprintf(fp,"echo '           ");
     if(plan->level>0)
       fprintf(fp,"-level %d   ",plan->level);
     fprintf(fp,"-younger_than %s",Sfile_datestr(plan->precheck_age,1|2|256));
     fprintf(fp,"'\n");
   }
   fname[0]= lname[0]= 0;
   if(stueck->first_payload_item>=0 && stueck->last_payload_item>=0) {
     if(plan->composition_mode==1 && 
        (plan->output_form==0 || plan->output_form==5)) {
       Fileliste_target(stueck->liste,stueck->first_payload_item,fname,3);
       Fileliste_target(stueck->liste,stueck->last_payload_item,lname,3);
       if(fname[0]!=0 && lname[0]!=0 && plan->anz_stuecke>1) 
         fprintf(fp,"echo 'volume content (sorted by target address) :'\n");
     } else if(plan->composition_mode==2 || plan->composition_mode==1) {
       strcpy(fname,stueck->liste->adressen[stueck->first_payload_item]);
       strcpy(lname,stueck->liste->adressen[stueck->last_payload_item]);
       if(fname[0]!=0 && lname[0]!=0 && plan->anz_stuecke>1) {
         if(plan->output_form==0 || plan->output_form==5 ||
            plan->output_form==8)
           fprintf(fp,"echo 'volume content (sorted by source address) :'\n");
         else
           fprintf(fp,"echo 'volume content (sorted by address) :'\n");
       }
     }
   }
   if(fname[0]!=0 && lname[0]!=0 && plan->anz_stuecke>1) {
     fprintf(fp,"echo '    '%s\n",Text_shellsafe(fname,shell_safe,0));
     fprintf(fp,"echo 'to  '%s\n",Text_shellsafe(lname,shell_safe,0));
   }
   if(plan->prompt_skript[0]!=0) {
     fprintf(fp,"if test -x %s ; then\n",plan->prompt_skript);
     fprintf(fp,"  %s\n",plan->prompt_skript);
     fprintf(fp,"fi\n");
   }
   if(plan->auskunft_an && plan->auskunft_individuell &&
     strlen(auskunft_name)>0) {
     fprintf(fp,"sed -e 's/BACKUP_PLANER_PART_MESSAGE_PLACEHOLDER/");
     fprintf(fp,"this is backup part no %d of %d/' < %s.pnr > %s.%d\n",
             volno, volcount, auskunft_name, auskunft_name, volno);
     fprintf(fp,"chmod u+x %s.%d\n",auskunft_name, volno);
     fprintf(fp,"cp %s.%d %s\n",auskunft_name, volno, auskunft_name);
   }
   if(plan->output_form==5 || plan->output_form==6 || plan->output_form==7 ||
      plan->output_form==8) {
     if(plan->volume_info) {
       ret= Backup_stueck_report_pl(stueck,fp,plan->verify_source_filter,
                                    plan->cmd, volno, volcount,
                                    plan->max_items_per_tile,plan->path_list,
                                    plan->output_form&255);
     } else {
       ret= Backup_stueck_report_pl(stueck,fp,plan->verify_source_filter,
                                    plan->cmd,-1,-1,
                                    plan->max_items_per_tile,plan->path_list,
                                    plan->output_form&255);
     }
     if(ret<=0)
       return(18);
   } else {
     if(plan->volume_info) {
       ret= Backup_stueck_report(stueck,fp,plan->cmd, volno, volcount,
                                 plan->output_form&255);
     } else {
       ret= Backup_stueck_report(stueck,fp,plan->cmd,
                                 -1,-1,plan->output_form&255);
     }
     if(ret<=0)
       return(19);
   }

   if(auskunft_name[0]!=0) {
     fprintf(fp,"rm %s.%d\n",auskunft_name, volno);
   }
   if(parts_made>0)
     Fileliste__print_split_cleanup(fp,plan->split_file_dir,
                            plan->split_file_prefix,plan->split_file_suffix,2);
   fprintf(fp,"\n");

   fprintf(fp,"  if test \"$cmd_return\" -ne \"0\" ; then\n");
   fprintf(fp,"    echo aborted\n");
   fprintf(fp,"    exit \"$cmd_return\"\n");
   fprintf(fp,"  fi\n");
   fprintf(fp,"fi\n\n");
   fprintf(fp,"\n");
   stueck= stueck->next;
 if(stueck==NULL) break;
 }
 if(strlen(plan->skript_status_file))
   fprintf(fp,"rm %s\n",plan->skript_status_file);
 fprintf(fp,"SCDBACKUP_PIECENO=\n");
 if(plan->auskunft_an && plan->auskunft_individuell &&
    strlen(auskunft_name)>0) {
   fprintf(fp,"sed -e 's/BACKUP_PLANER_PART_MESSAGE_PLACEHOLDER/");
   fprintf(fp,"this backup consists of %d parts/' < %s.pnr > %s\n",
               volcount, auskunft_name, auskunft_name);
   fprintf(fp,"chmod u+x %s\n",auskunft_name);
 }
 fprintf(fp,"\n");
 fprintf(fp,"echo '########## parts in total : %d'\n", volcount);
 if(plan->precheck_agetest&1) {
   fprintf(fp,"echo '           ");
   if(plan->level>0)
     fprintf(fp,"-level %d   ",plan->level);
   fprintf(fp,"-younger_than %s",Sfile_datestr(plan->precheck_age,1|2|256));
   fprintf(fp,"'\n");
 }
 fprintf(fp,"echo 'backup script ended normally'\n");
 if(fp!=NULL && fp!=stdout && fp!=stdout) fclose(fp);

 return(0);
}


int Backup_plan_set_skript_shell(argc,argv,newshell,flag)
int argc;
char **argv;
char *newshell;
int flag;
{
 FILE *fpr= NULL,*fpw= NULL;
 char newname[4096],line[4096];
 int i;

 for(i=0;i<argc;i++) {
   fpr= fopen(argv[i],"rb");
   if(fpr==NULL) {
     fprintf(stderr,"Cannot open file for reading : %s\n   %s\n",
                    argv[i],strerror(errno));
     goto loopend;
   }
   if(Sfile_fgets(line,sizeof(line),fpr)==NULL) 
     goto loopend;
   if(strncmp(line,"#!",2)!=0)
     goto loopend;
   sprintf(newname,"%s.new",argv[i]);
   fpw= fopen(newname,"wb");
   if(fpw==NULL) {
     fprintf(stderr,"Cannot open file for writing : %s\n   %s\n",
                    newname,strerror(errno));
     goto loopend;
   }
   fprintf(fpw,"#!%s\n",newshell);
   while(Sfile_fgets(line,sizeof(line),fpr)!=NULL) {
     fprintf(fpw,"%s\n",line);
   }
   fclose(fpw);
   fpw= NULL;
   sprintf(line,"%s.old",argv[i]);
   if(rename(argv[i],line)==-1){
     fprintf(stderr,"Cannot rename %s to %s\n   %s",
                    argv[i],line,strerror(errno));
     goto loopend;
   }
   if(rename(newname,argv[i])==-1){
     fprintf(stderr,"Cannot rename %s to %s\n   %s",
                    newname,argv[i],strerror(errno));
     goto loopend;
   }
   if(unlink(line)==-1) {
     fprintf(stderr,"Cannot remove %s \n   %s",
                    line,strerror(errno));
     goto loopend;
   }
   sprintf(line,"chmod u+x '%s'",argv[i]);
   system(line);
loopend:;
   if(fpr!=NULL) fclose(fpr);
   fpr= NULL;
   if(fpw!=NULL) fclose(fpw);
   fpw= NULL;
 }
 return(1); 
}


int Backup_plan_level_create(plan,level,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= it is no fault if the directory already exists
*/
{
 int ret;
 char level_dir[4096];
 struct stat stbuf;

 if(level>0) {
   /* >>> test wether previous level exists */;
 }

 sprintf(level_dir,"%s/%s_%d",plan->conf_dir,"level",level);
 if(Cd_backup_planer_call_staT(level_dir,&stbuf)!=-1) {
   if(((stbuf.st_mode)&S_IFMT) != S_IFDIR) {
     fprintf(stderr,"-level %d : file object should be a directory\n",level);
     fprintf(stderr,"%s\n",level_dir);
     ret= 0; goto ex;
   }
   if(flag&1)
     goto after_creation;
 }

 ret= mkdir(level_dir,0x1FF);
 if(ret==-1) {
   fprintf(stderr,"-level %d : creation of directory failed :\n",level);
   perror(level_dir);
   ret= 0; goto ex;
 }
after_creation:;
 ret= Backup_plan_level_write(plan,level,0,0,"",1);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_level_write(plan,level,run_date,volumes,askme_adr,flag)
struct Backup_plaN *plan;
time_t run_date;
int volumes;
char *askme_adr;
int flag;
/*
 bit0= this is a -create_configuration: do not copy ASKME
*/
{
 int ret,errn;
 char filename[4096],level_dir[4096],datestring[80],volumetext[80];
 char errmsg[SfileadrL];

 sprintf(level_dir,"%s/%s_%d",plan->conf_dir,"level",level);

 /* format and write run date */
 sprintf(filename,"%s/%s",level_dir,"last_run_date");
 strcpy(datestring,Sfile_datestr(run_date,1|2|256));
 ret= Sfile_string_to_file(filename,datestring,1);
 if(ret<=0) {
   fprintf(stderr,"-level %d : cannot write date string\n",level);
perr:;
   perror(filename);
   ret= 0; goto ex;
 }
 /* number of volumes */
 sprintf(filename,"%s/%s",level_dir,"number_of_volumes");
 sprintf(volumetext,"%d",volumes);
 ret= Sfile_string_to_file(filename,volumetext,1);
 if(ret<=0) {
   fprintf(stderr,"-level %d : cannot write number of volumes\n",level);
   goto perr;
 }
 if(flag&(1|2))
   goto after_askme;
 sprintf(filename,"%s/%s_level_%d",plan->conf_dir,"ASKME",plan->level);
 if(askme_adr[0]!=0) {
   /* copy ASKME to conf_dir */
   ret= Sfile_copy_file(askme_adr,filename,&errn,errmsg,0);
   if(ret<=0) {
     fprintf(stderr,"-level %d : cannot copy info script\n",level);
     fprintf(stderr,"from  %s\n",askme_adr);
     fprintf(stderr,"to    %s\n",filename);
     goto perr;
   } 
 } else {
   /* create ASKME dummy which exits 1 */;
   ret= Sfile_string_to_file(filename,"exit 1\n",1);
   if(ret<=0) {
     fprintf(stderr,"-level %d : cannot write default info script\n",level);
     goto perr;
   } 
 }
 chmod(filename,S_IRUSR | S_IWUSR | S_IXUSR | 
                S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);

after_askme:;
 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_conf_dir_default(plan,progname,flag)
struct Backup_plaN *plan;
char *progname;
int flag;
{
 char *cpt;
 int l;

 cpt= strrchr(progname,'/');
 if(cpt==NULL || cpt==progname) {
   strcpy(plan->conf_dir,"..");
 } else {
   for(cpt=cpt-1;cpt>progname;cpt--)
     if((*cpt)=='/')
   break;
   if(cpt<=progname) {
     strcpy(plan->conf_dir,".");
   } else {
     l= cpt-progname;
     strncpy(plan->conf_dir,progname,l);
     plan->conf_dir[l]= 0;
   }
 }
 strcat(plan->conf_dir,"/default_conf_dir");
 return(1);
}


int Backup_plan_conf_dir_check(plan,flag)
struct Backup_plaN *plan;
int flag;
{
 int ret;
 struct stat stbuf;

 /*  test wether -conf_dir and at least level_${level-1} exists */
 if(Cd_backup_planer_call_staT(plan->conf_dir,&stbuf)==-1) {
   fprintf(stderr,"-level ... : cannot access configuration directory\n");
   perror(plan->conf_dir);
   ret= 0; goto ex;
 }
 if(((stbuf.st_mode)&S_IFMT) != S_IFDIR) {
   fprintf(stderr,"-level ... : -conf_dir is not a directory\n");
   fprintf(stderr,"%s\n",plan->conf_dir);
   ret= 0; goto ex;
 }
 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_conf_create(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
/*
 bit0= update : do not create -conf_dir directory
*/
{
 int ret,i,first_line= 1,write_text,j,k;
 char filename[4096],pwd[4096],text[80],*buf= NULL,*cpt,*wpt,subst[4096];
 static char defaultargs[][81]= {
   "@@@start safe_incremental",
   "-load_old_tree","content_file_list","vanished_file_list",
   "-changetest_options","@@@changetest_options",
   "-list_output_format","@@@list_output_format",
   "-content_list_adr",
               "/added_by_scdbackup/content_file_list.gz=content_file_list.gz",
   "-vanish_script_adr","content_file_list",
                           "/added_by_scdbackup/VANISHED.sh.gz=VANISHED.sh.gz",
   "-vanish_list_adr","content_file_list",
             "/added_by_scdbackup/vanished_file_list.gz=vanished_file_list.gz",
   "-dir_list_adr",
               "/added_by_scdbackup/complete_dir_list.gz=complete_dir_list.gz",
   "@@@end safe_incremental",
   "@end_default_args@"
 };

 /* create -conf_dir */
 if(plan->conf_dir[0]==0) {
   Backup_plan_conf_dir_default(plan,argv[0],0);
 }
 if(flag&1) {
   ret= Backup_plan_conf_dir_check(plan,0);
   if(ret<=0)
     goto ex;
   fprintf(stderr,"Updating configuration directory :\n");
 } else {
   ret= mkdir(plan->conf_dir,0x1FF);
   if(ret==-1) {
     fprintf(stderr,
          "%s -level -create_configuration : creation of directory failed :\n",
             argv[0]);
     strcpy(filename,plan->conf_dir);
perr:;
     perror(filename);
     ret= 0; goto ex;
   }
   fprintf(stderr,"Configuration directory created :\n");
 } 
 fprintf(stderr,"%s\n",plan->conf_dir);

 /* write specific_arguments */
 sprintf(filename,"%s/%s",plan->conf_dir,"specific_arguments");

 /* write defaults for handling of vanished files */
 ret= Sfile_string_to_file(filename,"-working_directory",1|4);
 if(ret<=0)
   goto write_failed;
 first_line= 0;
 getcwd(pwd,sizeof(pwd));
 ret= Sfile_string_to_file(filename,pwd,1|2|4);
 if(ret<=0)
   goto write_failed;

 for(i=0;strcmp(defaultargs[i],"@end_default_args@")!=0; i++) {

   /* >>> safe and sparse default setup */;

   if(strncmp(defaultargs[i],"@@@",3)==0) {
     write_text= 0;
     if(strcmp(defaultargs[i],"@@@changetest_options")==0) {
       strcpy(text,"type:size:");
       if(plan->precheck_agetest&(1<<16)) {
         strcat(text,"inode:");
         if(plan->precheck_agetest&(1<<17))
           strcat(text,"ignore_device:");
       } else if(plan->precheck_agetest&2048) {
         strcat(text,"crc:");
         if(plan->precheck_agetest&(1<<18))
           strcat(text,"dir_trust:");
       }
       if(plan->precheck_agetest&32768)
         strcat(text,"no_split_economy:");
       if(plan->precheck_agetest&16384)
         strcat(text,"ignore_ctime:");
       if(plan->precheck_agetest&32)
         strcat(text,"mistrust:");
       if(plan->precheck_agetest&64)
         strcat(text,"assume_old:");
       strcat(text,"adr");
       write_text= 1;
     } else if (strcmp(defaultargs[i],"@@@list_output_format")==0) {
       strcpy(text,"type:size:");
       if((plan->list_output_format&32)||(plan->precheck_agetest&(1<<16)))
         strcat(text,"inode:");
       if((plan->list_output_format&8)||
        ((plan->precheck_agetest&2048) && !(plan->precheck_agetest&(1<<16)))) {
         if(plan->list_output_format&16)
           strcat(text,"md5:");
         else
           strcat(text,"crc:");
       }
       if(plan->list_output_format&128)
         strcat(text,"mtime:");
       if(plan->list_output_format&256)
         strcat(text,"ctime:");
       strcat(text,"adr");
       write_text= 1;
     }
     if(write_text) {
       ret= Sfile_string_to_file(filename,text,1|((!first_line)<<1)|4);
       if(ret<=0)
         goto write_failed;
       first_line= 0;
     }
 continue;
   }
   ret= Sfile_string_to_file(filename,defaultargs[i],1|((!first_line)<<1)|4);
   if(ret<=0) {
write_failed:;
     fprintf(stderr,
         "%s -level -create_configuration : recording of arguments failed :\n",
             argv[0]);
     goto perr;
   }
   first_line= 0;
 }

 if(plan->info_short_on_level) {
   ret= Sfile_string_to_file(filename,"-info_script\n-short",
                             1|((!first_line)<<1)|4);
   if(ret<=0)
     goto write_failed;
   first_line= 0;
 }

 /* write arguments */
 for(i=1;i<argc;i++) {
   /* skip arguments that interfere with the configuration system */
   if(strcmp(argv[i],"-conf_dir")==0 ||
      strcmp(argv[i],"-level")==0 ||
      strcmp(argv[i],"-bdir")==0 ||
      strcmp(argv[i],"-younger_than")==0) { 
     i++;
 continue;
   }
   /* do not override 'defaultargs' (name should be: 'fixed_options') */
   for(j= 0; strcmp(defaultargs[j],"@end_default_args@")!=0; j++) {
     if(defaultargs[j][0]=='-' && strcmp(defaultargs[j],argv[i])==0) {
       /* skip arguments of option */
       for(k= j+1;strcmp(defaultargs[k],"@end_default_args@")!=0; k++) {
         if(defaultargs[k][0]=='-' || 
            strcmp(defaultargs[k],"@@@end safe_incremental")==0)
       break;
         i++;
       }
   break;    
     }
   }
   if(strcmp(defaultargs[j],"@end_default_args@")!=0)
 continue;
   /* escape NewLine */
   if(buf!=NULL)
     Smem_freE(buf);
   buf= Smem_malloC(strlen(argv[i])*2+1);
   if(buf==NULL) {
     fprintf(stderr,"%s out of memory\n", argv[0]);
     ret= -1; goto ex;
   }
   wpt= buf;
   for(cpt= argv[i];*cpt!=0;cpt++) {
     if(*cpt=='\\') {
       *(wpt++)= '\\';
       *(wpt++)= '\\';
     } else if(*cpt=='\n') {
       *(wpt++)= '\\';
       *(wpt++)= 'n';
     } else
       *(wpt++)= *cpt;
   }
   *(wpt)= 0;

   /* reverse bdir substitution */
   ret= Backup_plan_substitute_bdir(plan,buf,subst,"",1<<15);
   if(ret<=0)
     goto ex;

   ret= Sfile_string_to_file(filename,subst,1|((!first_line)<<1)|4);
   if(ret<=0)
     goto write_failed;
   first_line= 0;

   /* enforce certain file names of lists which are not fixed options of
      incremental backups but are mentioned in the documentation or are
      used by saskme */
   if(strcmp(argv[i],"-dir_script_adr")==0) {
     ret= Sfile_string_to_file(filename,
             "/added_by_scdbackup/DIR_PERMS.sh.gz=DIR_PERMS.sh.gz",
             1|((!first_line)<<1)|4);
     if(ret<=0)
       goto write_failed;
     i++;
   } else if(strcmp(argv[i],"-split_list_adr")==0) {
     ret= Sfile_string_to_file(filename,
             "/added_by_scdbackup/split_file_list.gz=split_file_list.gz",
             1|((!first_line)<<1)|4);
     if(ret<=0)
       goto write_failed;
     i++;
   }
 }

 /* create and init level directores up to plan's level : level_#  */
 if(plan->level<0)
   plan->level= 0;
 for(i=0;i<=plan->level;i++) {
   ret= Backup_plan_level_create(plan,i,(flag&1));
   if(ret<=0)
     goto ex;
 }

#ifdef NIX
 sprintf(filename,"%s/%s",plan->conf_dir,"highest_level");
 sprintf(text,"%d",plan->level);
 ret= Sfile_string_to_file(filename,text,1);
 if(ret<=0) {
   fprintf(stderr,
       "%s -level -create_configuration : cannot write highest level number\n",
           argv[0]);
   goto perr;
 }
#else

 ret= Backup_plan_conf_update(plan,1);
 if(ret<=0)
   goto ex;

#endif  /* ! NIX */

 ret= 1;
ex:;
 if(buf!=NULL)
   Smem_freE(buf);
 return(ret);
}


int Backup_plan_prepend_level_address(plan,adr,level,flag)
struct Backup_plaN *plan;
char adr[4096];
int level;
int flag;
{
 char name[4096];

 if(adr[0]) {
   if(strchr(adr,'/')==NULL) {
     if(plan->conf_dir[0] && level>=0)
       sprintf(name,"%s/level_%d/%s",plan->conf_dir,level,adr);
     else 
       sprintf(name,"%s/%s",plan->tmp_dir,adr);
     strcpy(adr,name);
   }
 }
 return(1);
}


int Backup_plan_conf_use(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
/*
 bit0= allow some command line arguments to override or extend configured 
       arguments
*/
{
 int ret,i,hargc= 0,level,configuration,add_to_level;
 int highest_level;
 char filename[4096],**hargv= NULL,level_dir[4096],date_mark[80];
 char cmd[4096],*yargv[3],conf_dir[4096],*cpt,*wpt,bdir[4096];
 char substituted[4096],backup_id[81];

 if(plan->conf_dir[0]==0) 
   Backup_plan_conf_dir_default(plan,argv[0],0);

 level= plan->level;
 configuration= plan->configuration;
 strcpy(conf_dir,plan->conf_dir);

 /*  test wether -conf_dir and at least level_${level-1} exists */
 ret= Backup_plan_conf_dir_check(plan,0);
 if(ret<=0)
   goto ex;
 sprintf(filename,"%s/highest_level",conf_dir);
 ret= Sfile_file_to_int(filename,&highest_level,1);
 if(ret<=0) {
   fprintf(stderr,"%s -level : cannot read highest level mark",argv[0]); 
perr:;
   perror(filename);
   ret= 0; goto ex;
 }
 if(level<0) { /* last used level or automatically incremented level */
   if(level==-1)
     add_to_level= 1;
   else
     add_to_level= 0;
   if(highest_level<0) {
     fprintf(stderr,"%s -level -next : illegal content %d in file\n",
                    argv[0],highest_level); 
     fprintf(stderr,"%s\n",filename);
     ret= 0; goto ex;
   }
   level= highest_level+add_to_level;
   plan->level= level;
 }

 if(highest_level<level-1) {
   fprintf(stderr,"%s -level %d is too high, at most possible : -level %d\n",
                    argv[0],level,highest_level+1); 
   ret= 0;
   goto ex;
 }

 if(level>0) {
   sprintf(level_dir,"%s/level_%d",conf_dir,level-1);
   sprintf(filename,"%s/%s",level_dir,"last_run_date");
   ret= Sfile_file_to_text(filename,date_mark,sizeof(date_mark),0);
   if(ret<=0) {
     fprintf(stderr,
           "%s -level %d : cannot read previous level's date mark\n",
           argv[0],level);
     goto perr;
   }
   sprintf(filename,"%s/backup_id",conf_dir);
   ret= Sfile_file_to_text(filename,backup_id,sizeof(backup_id),0);
   if(ret>0)
     strcpy(plan->backup_id,backup_id);
 }

 /* eventually create level directory level_$level */
 ret= Backup_plan_level_create(plan,level,1);
 if(ret<=0)
   goto ex;

 /* reprogram with arguments from specific_arguments 
    plus -younger_than from level_$level/last_run_date
    if bit0 : plus argv again 
  */
 /* read specific_arguments */
 sprintf(filename,"%s/%s",plan->conf_dir,"specific_arguments");

 ret= Sfile_read_argv(argv[0],filename,&hargc,&hargv,0); 
 if(ret<=0) {
   fprintf(stderr,
           "%s -level %d : cannot read recorded arguments\n",
           argv[0],level);
   goto perr;
 }

 for(i=1;i<hargc;i++) {
   /* unescape NewLine */;
   wpt= hargv[i];
   for(cpt= hargv[i]; (*cpt)!=0; cpt++) {
     if(*cpt=='\\') {
       if(*(cpt+1)=='\\') {
         cpt++;
       } else if(*(cpt+1)=='n') {
         *(wpt++)= '\n';
         cpt++;
   continue;
       }
     }
     *(wpt++)= *cpt;
   }
   *wpt= 0;

   ret= Backup_plan_substitute_bdir(plan,hargv[i],substituted,"",2);
   if(ret<0)
     goto ex;
   if(ret!=1)
 continue;
   if(Sregex_string(&(hargv[i]),substituted,0)<=0)
     {ret= -1; goto ex;}

 }

 /* copy some information which gets cleared during _init() */
 strcpy(cmd,plan->cmd);
 strcpy(bdir,plan->bdir);
 strcpy(backup_id,plan->backup_id);

 Backup_plan_cleanup(plan,0);
 ret= Backup_plan_init(plan,plan->max_items,plan->max_size,cmd,0);
 if(ret<=0)
   goto ex;
 plan->configuration= configuration;
 strcpy(plan->bdir,bdir);
 strcpy(plan->backup_id,backup_id);
 /* preliminary setting of conf_dir and level. will be repeated. */
 strcpy(plan->conf_dir,conf_dir);
 plan->level= level;
 ret= Backup_plan_setup(plan,hargc,hargv,1|4);
 if(ret!=0)
   {ret= 0; goto ex;}
 if(level>0) {
   yargv[0]= argv[0];
   yargv[1]= "-younger_than";
   yargv[2]= date_mark;
   ret= Backup_plan_setup(plan,3,yargv,1|4);
   if(ret!=0)
     {ret= 0; goto ex;}
 }
 if(flag&1) { /* override some settings from the configuration files */
   ret= Backup_plan_setup(plan,argc,argv,1);
   if(ret!=0)
     {ret= 0; goto ex;}
 }
 /* close the plan */
 yargv[0]= argv[0];
 ret= Backup_plan_setup(plan,1,yargv,2);
 if(ret!=0)
   {ret= 0; goto ex;}

 strcpy(plan->conf_dir,conf_dir);
 plan->level= level;
 plan->configuration= configuration;

 fprintf(stderr,"\nIncremental backup :\n");
 fprintf(stderr,"   -conf_dir %s\n",plan->conf_dir);
 fprintf(stderr,"   -level    %d\n",plan->level);
 if(plan->level>0)
   fprintf(stderr,"   ( -younger_than    %s )\n",date_mark);
 fprintf(stderr,"   ( -level_timestamp %s )\n",
                Sfile_datestr(plan->start_time,1|2|256));
 fprintf(stderr,"\n\n");

 ret= 1;
ex:;
 Sfile_destroy_argv(&hargc,&hargv,0);
 return(ret);
}


int Backup_plan_conf_update(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0= forward to Backup_plan_level_write:
       this is a -create_configuration: do not copy ASKME
*/
{
 int ret;
 char filename[4096],text[80];

 ret= Backup_plan_level_write(plan,plan->level,
                              plan->start_time,
                              plan->anz_stuecke,
                              plan->auskunft_name->adressen[0],flag&1);
 if(ret<=0)
   goto ex;
 sprintf(filename,"%s/%s",plan->conf_dir,"highest_level");
 sprintf(text,"%d",plan->level);
 ret= Sfile_string_to_file(filename,text,1);
 if(ret<=0) {
   fprintf(stderr,
           "-level %d : cannot write highest level number\n",plan->level);
   perror(filename);
   ret= 0; goto ex;
 }
 sprintf(filename,"%s/%s",plan->conf_dir,"backup_id");
 ret= Sfile_type(filename,1);
 if(plan->level==0 || ret==-1) { /* renew id on level 0 or if missing */
   strcpy(text,plan->backup_id);
   ret= Sfile_string_to_file(filename,text,1);
   if(ret<=0) {
   fprintf(stderr,
             "-level %d : cannot write backup id\n",plan->level);
     perror(filename);
     ret= 0; goto ex;
   }
 }
 ret= 1;
ex:;
 return(ret);
}


int Backup_plan_app_excl_info(plan,liste,flag)
struct Backup_plaN *plan;
struct FilelistE *liste;
int flag;
/*
 bit0= map to root directory
*/
{
char absolut[4096],relativ[4096],punktslash[4096],haupt[4096],ziel[4096];
char auskunft_name[4096],*cpt;
FILE *fp;
int ret;

 Fileliste_target(liste,0,auskunft_name,1);
 /* truncate the file */
 if((fp=fopen(auskunft_name,"wb"))!=NULL) 
   fclose(fp);
 ret= Fileliste_append(plan->auslassen,auskunft_name,1|8);
 if(ret<=0) 
   return(ret);
 cpt= strrchr(auskunft_name,'/');
 if(cpt==NULL) cpt= auskunft_name;
 else          cpt++;
 if(flag&1)
   strcpy(ziel,cpt);
 else
   strcpy(ziel,auskunft_name);
 if(ziel[0]=='/'){
   strcpy(absolut,ziel);
   strcpy(relativ,ziel+1);
   sprintf(punktslash,".%s",ziel);
 }else if(ziel[0]=='.' && ziel[1]=='/') {
   strcpy(absolut,ziel+1);
   strcpy(relativ,ziel+2);
   strcpy(punktslash,ziel);
 }else{
   sprintf(absolut,"/%s",ziel);
   strcpy(relativ,ziel);
   sprintf(punktslash,"./%s",ziel);
 }
 ret= Fileliste_append(plan->auslassen,absolut,1|4|8);
 if(ret<=0)
   return(ret);
 ret= Fileliste_append(plan->auslassen,relativ,1|4|8);
 if(ret<=0)
   return(ret);
 ret= Fileliste_append(plan->auslassen,punktslash,1|4|8);
 if(ret<=0)
   return(ret);
 if(plan->hauptdir[0] && strcmp(plan->hauptdir,"/")!=0){
   sprintf(haupt,"%s%s",plan->hauptdir,absolut);
   ret= Fileliste_append(plan->auslassen,haupt,1|4|8);
   if(ret<=0)
     return(ret);
 }
 return(1);
}


int Backup_plan_append_simple(plan,source,target,size,flag)
struct Backup_plaN *plan;
char *source;
char *target;
double *size;
int flag;
/*
 bit0= read file properties, add measured size to *size
 bit1= this is added_by_scdbackup , not a payload item
 bit0= do not add to piece, but return after reading properties
*/
{
 int ret,linecount;
 struct FilelistE *liste= NULL;

 ret= Fileliste_single(&liste,source,0,0.0,8|16);
 if(ret<=0)
   goto ex;
 ret= Fileliste_set_target(liste,0,target,0);
 if(ret<=0)
   goto ex;
 if(flag&1) {
   ret= Fileliste_read_properties_idx(liste,0,&linecount,1,0);
   if(ret<=0)
     goto ex;
   (*size)+= liste->groessen[0];
 }
 if(flag&4)
   {ret= 1; goto ex;}
 ret= Backup_plan_compose(plan,liste,2|4|((flag&2)<<8));
 if(ret<=0)
   goto ex;
 
 ret= 1;
ex:;
 Fileliste_destroy(&liste,0);
 return(ret);
}


int Backup_plan_append_list_item(plan,in_liste,idx,flag)
struct Backup_plaN *plan;
struct FilelistE *in_liste;
int idx;
int flag;
/*
 bit1= this is added_by_scdbackup , not a payload item
*/
{
 int ret;
 struct FilelistE *liste= NULL;

 ret= Fileliste_new(&liste,1,"",0);
 if(ret<=0)
   goto ex;
 ret= Fileliste_copy_item(in_liste,idx,liste,0,0);
 if(ret<=0)
   goto ex;
 ret= Backup_plan_compose(plan,liste,2|4|((flag&2)<<8));
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Fileliste_destroy(&liste,0);
 return(ret);
}


int Backup_plan_generate_lists(plan,flag)
struct Backup_plaN *plan;
int flag;
/*
 bit0-3= stage
    0= after setup of arguments
    1= after file check 
*/
{
 int ret,stage,pass,aflag,free_items,items= 0;
 struct FilelistE *liste= NULL;
 double size= 0.0,free_bytes;
 struct Backup_stuecK *dummy_volume;

 stage= (flag&15);
 if(stage==0) {

   if(plan->use_crawler) {
     ret= Backup_plan_run_crawler(plan,0);
     if(ret<=0) {
       fprintf(stderr,
           "--- File check and/or list generation failed. Program aborted.\n");
       {ret= -25; goto ex;}
     }
     if(ret&(1<<15)) {
       fprintf(stderr,
           "--- File check exceeded problem limits. Program aborted.\n");
       {ret= -15; goto ex;}
     }
     goto ex;
   }

#ifndef Crawler_ruleS
   ret= Backup_plan_load_old_tree(plan,plan->old_complete_list_adr,
                                     plan->old_vanish_list_adr,"",0);
   if(ret<=0)
     {ret= -22; goto ex;}

   if(plan->vanish_script_adr[0]!=0 && plan->vanish_list_adr[0]!=0) {
     ret= Backup_plan_vanished_list(plan,7);
     if(ret<=0)
       {ret= -19; goto ex;}
   } else {
     ret= Backup_plan_vanished_list(plan,0);
     if(ret<=0)
       {ret= -19; goto ex;}
     ret= Backup_plan_vanished_list(plan,1);
     if(ret<=0)
       {ret= -24; goto ex;}
   }

   /* complete_list_adr */
   Backup_plan_complete_list(plan,0);

   if(plan->dir_script_adr[0]!=0 && plan->dir_list_adr[0]!=0) {
     Backup_plan_complete_list(plan,1|8);
   } else {
     /* dir_list_adr */
     Backup_plan_complete_list(plan,1);
     /* dir_script_adr */
     Backup_plan_complete_list(plan,3);
   }

#endif /* ! Crawler_ruleS */

   /* remove any old result files */
   /* content_list_adr */
   Backup_plan_content_list(plan,2);
   /* facl_list_adr */
   Backup_plan_faclfattr_list(plan,2);
   /* fattr_list_adr */
   Backup_plan_faclfattr_list(plan,1|2);
   /* faclsrc_list_adr */
   Backup_plan_faclfattr_list(plan,2|4);
   /* fattrsrc_list_adr */
   Backup_plan_faclfattr_list(plan,1|2|4);
   /* split_list_adr */
   Backup_plan_split_list(plan,2);

 } else if(stage==1) {

   /* content_list_adr */
   if((plan->composition_mode&(1<<15))) {
     ret= 2;
   } else {
     ret= Backup_plan_content_list(plan,0);
   }
   if(ret<=0)
     {ret= -15; goto ex;}
   /* facl_list_adr */
   ret= Backup_plan_faclfattr_list(plan,0);
   if(ret<=0)
     {ret= -15; goto ex;}
   /* fattr_list_adr */
   ret= Backup_plan_faclfattr_list(plan,1);
   if(ret<=0)
     {ret= -15; goto ex;}
   /* faclsrc_list_adr */
   ret= Backup_plan_faclfattr_list(plan,4);
   if(ret<=0)
     {ret= -15; goto ex;}
   /* fattrsrc_list_adr */
   ret= Backup_plan_faclfattr_list(plan,1|4);
   if(ret<=0)
     {ret= -15; goto ex;}
   /* split_list_adr */
   ret= Backup_plan_split_list(plan,0);
   if(ret<=0)
     {ret= -15; goto ex;}

   for(pass=0;pass<2;pass++) {
     aflag= 2;
     /* check wether all list will fit last piece
        eventually start new piece */
     if(pass==0) {
       aflag|= 5; /* for Backup_plan_append_simple: only add count to size */
       size= 0.0;
       items= 0;
     } else {
       free_items= plan->max_items - plan->aktstueck->liste->anz;
       free_bytes= plan->max_size - plan->aktstueck->bytes;
       if((free_items<items && items<=plan->max_items) || 
           (free_bytes<size && size<=plan->max_size)) {
         Print_pacifier(0,"",1|16); /* clear pacifier line */
         ret= Backup_plan_next_stueck(plan,&dummy_volume,0);
         if(ret<=0)
           goto ex;
       }
     }
     if(plan->complete_list_adr[0]!=0 && plan->append_complete_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->complete_list_adr,
                                     plan->append_complete_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->dir_list_adr[0]!=0 && plan->dir_list_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->dir_list_adr,
                                           plan->dir_list_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->facl_list_adr[0]!=0 && plan->facl_list_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->facl_list_adr,
                                      plan->facl_list_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->fattr_list_adr[0]!=0 && plan->fattr_list_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->fattr_list_adr,
                                      plan->fattr_list_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->faclsrc_list_adr[0]!=0 && plan->faclsrc_list_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->faclsrc_list_adr,
                                      plan->faclsrc_list_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->fattrsrc_list_adr[0]!=0 && plan->fattrsrc_list_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->fattrsrc_list_adr,
                                      plan->fattrsrc_list_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->content_list_adr[0]!=0 && plan->append_content_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->content_list_adr,
                                      plan->append_content_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->vanish_list_adr[0]!=0 && plan->vanish_list_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->vanish_list_adr,
                                      plan->vanish_list_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->vanish_script_adr[0]!=0 && plan->vanish_script_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->vanish_script_adr,
                                      plan->vanish_script_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->dir_script_adr[0]!=0 && plan->dir_script_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->dir_script_adr,
                                      plan->dir_script_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
     if(plan->split_list_adr[0]!=0 && plan->split_list_target[0]!=0) {
       ret= Backup_plan_append_simple(plan,plan->split_list_adr,
                                      plan->split_list_target,&size,aflag);
       if(ret<=0)
         goto ex;
       items++;
     }
   }
 }
 ret= 1;
ex:;
 Fileliste_destroy(&liste,0);
 return(ret);
}


int Backup_plan_conf_main(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
{
 int ret,extendable= 1;

 if(plan->configuration==1 || plan->configuration==3) {
   ret= Backup_plan_conf_create(plan,argc,argv,(plan->configuration==3));
   if(ret<=0)
     return(-1);
   /* switch to -configuration use */
   extendable= 0; /* do not read arguments again */
   plan->configuration= 2;
 }

 if(plan->configuration==2) {
   Fileliste_destroy(&(plan->old_tree),0);
   ret= Backup_plan_conf_use(plan,argc,argv,extendable);
   if(ret<=0 || plan->startliste->anz<=0)
     return(-1);
 }
 /* prepend level_directory or tmp_dir */
 Backup_plan_prepend_level_address(plan,plan->complete_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->dir_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->dir_script_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->content_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->facl_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->fattr_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->faclsrc_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->fattrsrc_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->split_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->vanish_list_adr,plan->level,0);
 Backup_plan_prepend_level_address(plan,plan->vanish_script_adr,plan->level,0);
 return(1);
}


int Backup_plan_open_tag_source(plan,cmd,fpin,basevec,flag)
struct Backup_plaN *plan;
char *cmd;
FILE **fpin;
int basevec[];
int flag;
{
 int ret;

 Intvec_from_double(basevec,CdbplivS,
                  (plan->checksum_scan_skip>0?plan->checksum_scan_skip:0.0),0);
 if(plan->checksum_scan_source[0]) {
   *fpin= fopen(plan->checksum_scan_source,"r");
   if(*fpin==NULL) {
     fprintf(stderr,"\n\n%s : cannot open file : '%s'\n",
             cmd,plan->checksum_scan_source);
     fprintf(stderr,"   %s\n",
             (errno>0?strerror(errno):"-unknown error-"));
     {ret= 0; goto ex;}
   }
   if(plan->checksum_scan_skip>0) {
     ret= Fseek_double(*fpin,plan->checksum_scan_skip,0);
     if(ret) {
       fprintf(stderr,"\n\n%s : error while skip:%.f\n",
               cmd,plan->checksum_scan_skip);
       fprintf(stderr,"   %s\n",
               (errno>0?strerror(errno):"-unknown error-"));
       {ret= 0; goto ex;}
     }
   }
 } else
   *fpin= stdin;
 ret= 1;
ex:;
 return ret;
}


int Backup_plan_match_checksums(plan,cmd,listadr,
                                s_name,s_old,s_young,flag)
struct Backup_plaN *plan;
char *cmd,*listadr;
char *s_name,*s_old,*s_young;
int flag;
/*
  bit0= end after first match

  bit8= compare name
  bit9= compare date limits

 bit12= work with MD5 rather than CRC32
*/
{
 int ret,found_line= 0,valid,i,rec_count,use_handle= 512,hashtype,hflag;
 int dummy,fresh_find= 0;
 /* this -2 detects silent overflows and improves english language */ 
 int matchlist_limit= Cd_backup_planer_tag_list_maX-2;
 FILE *fp= NULL,*fpin;
 char checksum_text[256],*rec_name,size_text[80];
 unsigned char md5[16],rec_md5[16];
 unsigned long crc32,rec_crc32;
 double size_dummy,mismatches= 0.0,matches_after_mismatch= 0.0;
 double matchlist_count= 0.0;
 int read_sizevec[CdbplivS],rec_sizevec[CdbplivS];
 int sizevec[CdbplivS],match_sizevec[CdbplivS],hvec[CdbplivS];
 int scanned_vec[CdbplivS];
 time_t old_limit= 0,young_limit= 0,rec_date= 0;
 struct CheckrecorD *checklist= NULL,*c_rec;
 struct CheckrecorD *matchlist= NULL,*old_matchlist,*o_rec,*next_rec;
 char *handle= NULL;   /* should be (void *)  */
 char record_text[161],statemark[8];
 struct Checkrecord_taG *scanstate= NULL;
 struct Blksum_lisT *blocklist;

 if(!(plan->checksum_scan_mode&4)) {
   ret= Checkrecord_read_list(&checklist,listadr,&rec_count,1);
   if(ret<=0)
     goto ex;
   fprintf(stderr,"Loaded %d checksum records from %s\n",rec_count,listadr);
 }
 if(plan->checksum_scan_mode&1) {
   /* create a giant pseudo checkrecord to allow scanning up to the end */;
   ret= Checkrecord_new(&o_rec,"scan_limit",time(0),1.0e24,NULL,
                        !!(flag&(1<<12)),0,"0000000000000000",0);
   if(ret<=0)
     {ret= -1; goto ex;}
   Checkrecord_link(o_rec,checklist,2);
   if(checklist==NULL)
     checklist= o_rec;
 } 

 /* decode time limits */;
 if(strcmp(s_old,"-")==0) {
   flag&= ~256;
 } else {
   ret= String_to_time(s_old,&valid,&old_limit,0);
   if(ret<=0) {
bad_limit:;
     fprintf(stderr,
          "\n%s : NOT A USABLE TIME RANGE.  UNBRAUCHBARE ZEITANGABE :\n",cmd);
     fprintf(stderr, "%s %s\n",s_old,s_young);
     {ret= 0; goto ex;}
   }
 }
 if(strcmp(s_young,"-")==0) {
   flag&= ~512;
 } else {
   ret= String_to_time(s_young,&valid,&young_limit,0);
   if(ret<=0) 
     goto bad_limit;
 }
 if(old_limit>young_limit) {
   rec_date= old_limit;
   old_limit= young_limit;
   young_limit= rec_date;
 }

 ret= Backup_plan_open_tag_source(plan,cmd,&fpin,read_sizevec,0);
 if(ret<=0)
   goto ex;
 if(plan->checksum_scan_skip>0)
   fprintf(stderr,"\nNote: %.f bytes of input skipped. Expect no matches !\n",
           plan->checksum_scan_skip);

 for(c_rec= checklist;c_rec!=NULL;c_rec= c_rec->next) {
   /* crawl along the list */
   Checkrecord_get_name(c_rec,&rec_name,0);
   Checkrecord_get_sizevec(c_rec,rec_sizevec,&dummy,0);
   Checkrecord_get_date(c_rec,&rec_date,0);
   Checkrecord_get_checksum(c_rec,&hashtype,&rec_crc32,rec_md5,0);
   if((hashtype==1) != !!(flag&(1<<12)))
 continue;
   if(flag&(1<<8))
     if(strcmp(rec_name,s_name)!=0)
 continue;
   if(flag&(1<<9)) {
     if(rec_date<old_limit)
 continue;
     if(rec_date>young_limit)
 continue;
   }
   /* crawl along the file */
   /* how much is to be read up to the next record size ? */
   Intvec_add(sizevec,CdbplivS,rec_sizevec,read_sizevec,1);
   if(Intvec_sign(sizevec,CdbplivS,0)>0) {
     /* get this amount and the checksum */
       
     if(plan->checksum_scan_mode&1) {
       /* employ a specialized stdin/Stdout_pipe_broker checksummer */
       old_matchlist= matchlist;
       hflag= use_handle|(flag&(1<<12));
       hflag|= plan->checksum_scan_mode&2;
       hflag|= 1|(4*!!(fresh_find));
       hflag|= 8*(found_line>0 || mismatches>0);
       hflag|= 16*(mismatches>0);
       hflag|= 32;
       hflag|= (plan->checksum_scan_mode&16)<<2;
       hflag|= (plan->checksum_scan_mode&32)<<8;
       ret= Scan_checksum(&scanstate,&handle,fpin,md5,&crc32,
                          read_sizevec,sizevec,scanned_vec,
                          &matchlist,hflag);
       if(ret<=0)
         goto ex;
       fresh_find= 0;
       if(matchlist!=old_matchlist) {
         
         for(o_rec= matchlist; o_rec->prev!=NULL && o_rec!=old_matchlist;
             o_rec= o_rec->prev);
         for(;o_rec!=NULL && o_rec!=old_matchlist; o_rec= next_rec) {
           next_rec= o_rec->next;
           if(Checkrecord_is_tested_match(o_rec,0)) {
             Checkrecord_get_sizevec(o_rec,hvec,&dummy,0);
             Intvec_copy(hvec,CdbplivS,match_sizevec,0);
             if(mismatches>0)
               matches_after_mismatch++;
             found_line= 1;
           } else 
             mismatches++;

           /* eventually write scanned blocklist */
           blocklist= NULL;
           if(plan->blocklist_fileadr[0]!=0)
             Checkrecord_get_blocklist(o_rec,&blocklist,0);
           if(blocklist!=NULL) {  
             ret= Blksum_list_to_file(blocklist,plan->blocklist_fileadr,
                                      o_rec,2);
             if(ret<=0)
               goto ex;
           }
           matchlist_count++;
           if(matchlist_count>matchlist_limit) {
             Checkrecord_destroy(&o_rec,0);
         continue;
           }
         }
         if((plan->checksum_scan_mode&2) || ((flag&1)&&found_line))
 break; /* end search run */
       }
       Intvec_copy(scanned_vec,CdbplivS,sizevec,0);

     } else {
       if(flag&(1<<12))
         ret= Determine_file_checksum("-",stdin,&handle,md5,md5,
                                      &size_dummy,sizevec,NULL,
                                      1|2|4|8|use_handle|(1<<12));
       else
         ret= Determine_file_checksum("-",stdin,&handle,&crc32,&crc32,
                                      &size_dummy,sizevec,NULL,
                                      1|2|4|8|use_handle);
       if(ret<=0)
         goto ex;
     }
     use_handle|= 256;  /* mark handle as already open for next cycle */

     /* if the desired amount was not available : end search */
     Intvec_add(hvec,CdbplivS,rec_sizevec,read_sizevec,1); 
     if(Intvec_cmp(sizevec,CdbplivS,hvec,0)<0) {
       Intvec_add(read_sizevec,CdbplivS,read_sizevec,sizevec,0);
 break;
     }
     Intvec_add(read_sizevec,CdbplivS,read_sizevec,sizevec,0);
   }
   /* compare checksums */
   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       if(md5[i]!=rec_md5[i])
     break;
     if(i<16)
 continue;
   } else {
     if(crc32!=rec_crc32)
 continue;
   }

   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       sprintf(checksum_text+2*i,"%2.2x",rec_md5[i]);
   } else
     sprintf(checksum_text,"%lX",rec_crc32);
   Intvec_to_dec(rec_sizevec,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"\n\nFound matching record :\n%s %s %s %s\n\n",
                  rec_name,Sfile_datestr(rec_date,1|2|256),
                  size_text,checksum_text);
   Intvec_copy(read_sizevec,CdbplivS,match_sizevec,0);
   if(mismatches>0)
      matches_after_mismatch++;
   found_line= 1;
   fresh_find= 1;
   if(flag&1)
 break;
 }
 if(plan->checksum_scan_mode&8) {
   fprintf(stderr,
           "\n\nList of found checksum tags (+ match , - mismatch) :\n");
   o_rec= matchlist;
   if(o_rec!=NULL)
     for(; o_rec->prev!=NULL; o_rec= o_rec->prev);
   for(;o_rec!=NULL;o_rec= o_rec->next) {
     Checkrecord_to_text(o_rec,record_text,0);
     if(Checkrecord_is_tested_match(o_rec,0)) 
       strcpy(statemark,"+");
     else
       strcpy(statemark,"-");
     fprintf(stderr,"%s  %s\n",statemark,record_text);
   }
   if(matchlist_count>matchlist_limit)
     fprintf(stderr,
             "Note: At least %.lf tags were discarded to prevent overflow.\n",
              matchlist_count-matchlist_limit);
 }
 fprintf(stderr,"\n");
 if(!found_line) {
   if(mismatches>0) 
     fprintf(stderr,"Note: scanning found %.lf mismatching records.\n",
                    mismatches);
   fprintf(stderr,
       "\n\n%s : NO MATCHING RECORD FOUND. KEIN PASSENDER RECORD GEFUNDEN :\n",
        cmd);
   fprintf(stderr,"  %s %s %s\n\n",s_name,s_old,s_young);
   {ret= 0; goto ex;}
 } else {
   Intvec_add(hvec,CdbplivS,read_sizevec,match_sizevec,1);
   if(Intvec_sign(hvec,CdbplivS,0)>0) {
     Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
     fprintf(stderr,"Note: read %s surplus bytes\n",size_text);
   }
   if(matches_after_mismatch>0) 
     fprintf(stderr,"Note: scanning also found %.lf mismatching records.\n",
                    mismatches);
   if(flag&1)
     fprintf(stderr,"Note: did not read further than first match.\n");
   fprintf(stderr,
           "\n\nOK.  Matching Record found . Passender Record gefunden .\n\n");
 }

 ret= (!!found_line);
ex:;
 if(handle!=NULL)
   free((char *) handle);
 Checkrecord_tag_destroy(&scanstate,0);
 Sfile_fclose(fp,0);
 return(ret);
}


int Backup_plan_find_checksum(plan,cmd,listadr,
             s_name,s_old,s_young,s_sizevec,s_checksum,
             f_name,f_date,f_sizevec,f_checksum,f_md5,f_crc32,
             flag)
struct Backup_plaN *plan;
char *cmd,*listadr;
char *s_name,*s_old,*s_young;
int *s_sizevec;
char *s_checksum;
char f_name[256],f_date[256];
int *f_sizevec;
char f_checksum[256];
unsigned char f_md5[16];
unsigned long *f_crc32;
int flag;
/*
 bit1= only print matching record lines but do not check stdin

  bit8= compare name
  bit9= compare date limits
 bit10= compare size 
 bit11= compare checksum
*/
{
 int ret,found_line= 0,valid,i,is_md5= 0;
 FILE *fp= NULL;
 char buf[256],age_limit[14],young_limit[14];
 char read_name[256],read_date[256],read_length[256],read_sum[256];
 int read_sizevec[CdbplivS];
 time_t moment;

 fp= Sfile_fopen(listadr,"rb",0);
 if(fp==NULL)
   {ret= 0; goto ex;}

 /* decode and reformat time limits */;
 if(strcmp(s_old,"-")==0) {
   strcpy(age_limit,"-");
 } else {
   ret= String_to_time(s_old,&valid,&moment,0);
   if(ret<=0) {
bad_limit:;
     fprintf(stderr,
          "\n%s : NOT A USABLE TIME RANGE.  UNBRAUCHBARE ZEITANGABE :\n",cmd);
     fprintf(stderr, "%s %s\n",s_old,s_young);
     {ret= 0; goto ex;}
   }
   strcpy(age_limit,Sfile_datestr(moment,1|2|256));
 }
 if(strcmp(s_young,"-")==0) {
   strcpy(young_limit,"-");
 } else {
   ret= String_to_time(s_young,&valid,&moment,0);
   if(ret<=0) 
     goto bad_limit;
   strcpy(young_limit,Sfile_datestr(moment,1|2|256));
 }
 if(strcmp(age_limit,"-")!=0 && strcmp(young_limit,"-")!=0 && 
    strcmp(age_limit,young_limit)>0) {
   strcpy(buf,age_limit);
   strcpy(age_limit,young_limit);
   strcpy(young_limit,buf);
 }

 /* try to find matching line (keep last match) */;
 while(Sfile_fgets(buf,sizeof(buf)-1,fp)!=NULL) {
   read_name[0]= read_date[0]= 0;
   sscanf(buf,"%s %s %s %s",read_name,read_date,read_length,read_sum);
   ret= Intvec_from_dec(read_sizevec,CdbplivS,read_length,0);
   if(ret<=0)
     Intvec_from_double(read_sizevec,CdbplivS,-1.0,0);
   if(read_name[0]==0 || strlen(read_date)!=13)
 continue;
   if(flag&(1<<8)) {
     if(strcmp(s_name,"-")!=0 && strcmp(read_name,s_name)!=0)
 continue;
   }
   if(flag&(1<<9)) {
     if(strcmp(age_limit,"-")!=0 && strcmp(age_limit,read_date)>0)
 continue;
     if(strcmp(young_limit,"-")!=0 && strcmp(young_limit,read_date)<0)
 continue;
   }
   if(flag&(1<<10)) {
     if(Intvec_cmp(read_sizevec,CdbplivS,s_sizevec,0)!=0)
 continue;
   }
   if(flag&(1<<11)) {
     if(strcmp(s_checksum,"-")!=0 && strcmp(s_checksum,read_sum)!=0)
 continue;
   }
   if(flag&2)
     fprintf(stdout,"%s\n",buf);
   strcpy(f_name,read_name);
   strcpy(f_date,read_date);
   Intvec_copy(read_sizevec,CdbplivS,f_sizevec,0);
   strcpy(f_checksum,read_sum);

   /* determine type of checksum and set flag bit12 */;
   if(strlen(f_checksum)==32) {
     for(i=0;i<16;i++) {
       buf[0]= f_checksum[2*i];
       buf[1]= f_checksum[2*i+1];
       buf[2]= 0;
       sscanf(buf,"%x",&ret);
       f_md5[i]= ret;
     }
     is_md5= 1;
   } else {
     *f_crc32= 0;
     sscanf(f_checksum,"%lX",f_crc32);
     is_md5= 0;
   }
   found_line= 1;
 }
 if(!found_line) {
   fprintf(stderr,
        "\n%s : NO MATCHING RECORD FOUND. KEIN PASSENDER RECORD GEFUNDEN :\n",
        cmd);
   fprintf(stderr,"  %s %s %s\n",s_name,s_old,s_young);
   {ret= 0; goto ex;}
 }

 ret= (!!found_line)+is_md5;
ex:;
 Sfile_fclose(fp,0);
 return(ret);
}


int Backup_plan_filter_checksum(plan,cmd,target,
                                idx_text,age_limit_text,young_limit_text,
                                flag)
struct Backup_plaN *plan;
char *cmd,*target,*idx_text,*age_limit_text,*young_limit_text;
int flag;
/*
 bit0= read from target and compare with stdin (else write stdin sum to target)
 bit1= only print matching record lines but do not check stdin
 bit2= with bit0 : search matching record
                   either by pre_checksum or by sorted match 
 bit12= work with MD5 rather than CRC32
*/
{
 int ret,i,cut,pre_crc= 0,use_handle= 0;
 FILE *fp= NULL;
 unsigned char md5[16],old_md5[16],pre_md5[16];
 char hexmd5[33],old_date[256],idx[256];
 char old_sum[256],old_idx[256],rec_name[256],size_text[80];
 unsigned long crc32,old_crc32,pre_crc32;
/*
 double size= 0.0,old_size= 0.0;
*/
 double size_dummy;
 int sizevec[CdbplivS],old_sizevec[CdbplivS],hvec[CdbplivS],posvec[CdbplivS];
 int min_dutyvec[CdbplivS],countvec[CdbplivS];
 time_t moment;
 char *handle= NULL;   /* should be void *  */
 struct CheckrecorD *check_rec= NULL;
 char pre_idx[81],record_text[161];
 struct Blksum_lisT *blocklist= NULL;

 if(strlen(idx_text)>80) {
   fprintf(stderr,"\n%s : RECORD NAME TOO LONG (max. 80 char)\n",cmd);
   {ret= 0; goto ex;}
 }
 if(plan->pre_checksum_size>0 && strlen(idx_text)>76) {
   fprintf(stderr,
        "\n%s : RECORD NAME TOO LONG (with pre checksum max. 76 char )\n",cmd);
   {ret= 0; goto ex;}
 }
 sscanf(idx_text,"%s",idx);
 if(idx[0]==0) {
   fprintf(stderr,"\n%s : NO USABLE RECORD NAME\n",cmd);
   {ret= 0; goto ex;}
 }

 if(flag&4) {
   if(plan->pre_checksum_size<=0) {
     ret= Backup_plan_match_checksums(plan,cmd,target,
                 idx,age_limit_text,young_limit_text,
                 (flag&(1<<12))|(plan->pre_checksum_size<0));
     goto ex;
   }
   use_handle= 512; /* let checksum handle open on first call of
                        Determine_file_checksum()  */
 } else if(flag&(1|2)) {
   ret= Backup_plan_find_checksum(plan,cmd,target,
             idx,age_limit_text,young_limit_text,hvec,"-",
             old_idx,old_date,old_sizevec,old_sum,old_md5,&old_crc32,
             (flag&2)|256|512);
   if(ret<=0)
     goto ex;
   if(ret==2)
     flag|= (1<<12);
   else
     flag&= ~(1<<12);

   if(flag&2) 
     {ret= 1; goto ex;}
   Intvec_to_dec(old_sizevec,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"\nComparing with record :\n%s %s %s %s\n\n",
                  old_idx,old_date,size_text,old_sum); 
 } else {
   fp= Sfile_fopen(target,"ab",0);
   if(fp==NULL)
     {ret= 0; goto ex;}
 }

 if(flag&(1|4)) {
   cut= 4;
   if(flag&4) {
     if(plan->pre_checksum_size<0)
       Intvec_from_double(sizevec,CdbplivS,0.0,0);
     else
       Intvec_from_double(sizevec,CdbplivS,plan->pre_checksum_size,0);
     pre_crc= 8;
   } else
     Intvec_copy(old_sizevec,CdbplivS,sizevec,0);
 } else {
   cut= 0;
   if(plan->pre_checksum_size>0.0) {
     Intvec_from_double(sizevec,CdbplivS,plan->pre_checksum_size,0);
     pre_crc= 8;
   }
 }

determine_checksum:;
 if(flag&(1<<12)) {
   /* >>> ??? compute block checksums in compare mode (flag&1) ? */
   if(plan->block_checksum_size>0 && !(flag&1)) {
     ret= Blksum_list_of_stream(&blocklist,stdin,plan->block_checksum_size,
                                sizevec,md5,2);
   } else {
     ret= Determine_file_checksum("-",stdin,&handle,md5,pre_md5,
                                  &size_dummy,sizevec,NULL,
                                  1|2|cut|pre_crc|use_handle|(1<<12));
   }
 } else 
   ret= Determine_file_checksum("-",stdin,&handle,&crc32,&pre_crc32,
                                &size_dummy,sizevec,NULL,
                                1|2|cut|pre_crc|use_handle);
 if(ret<=0)
   goto ex;

 if((flag&4) && pre_crc) { /* only on first pass */
   /* to erase possible pacifier messages of program raedchen */
   fprintf(stderr,"\r                                                   \r");
   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       sprintf(hexmd5+2*i,"%2.2x",pre_md5[i]);
   } else {
     sprintf(hexmd5,"%lX",pre_crc32);
   }
   if(strcmp(idx,"-")==0)
     strcpy(rec_name,idx);
   else
     sprintf(rec_name,"%s_pre",idx);
   Intvec_from_double(hvec,CdbplivS,plan->pre_checksum_size,0);
   ret= Backup_plan_find_checksum(
             plan,cmd,target,rec_name,age_limit_text,young_limit_text,
             hvec,hexmd5 ,
             old_idx,old_date,old_sizevec,old_sum,old_md5,&old_crc32,
             (flag&2)|256|512|1024|2048);
   if(ret<0)
     goto ex;
   if(flag&2) 
     {ret= 1; goto ex;}
   if(ret==0) {

     {ret= 0; goto ex;}
   }
   Intvec_to_dec(old_sizevec,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"\nFound matching preview record :\n%s %s %s %s\n",
                  old_idx,old_date,size_text,old_sum);

   /* continue verification with record found */;
   /* obtain real record */;
   ret= Backup_plan_find_checksum(plan,cmd,target,
             idx,old_date,old_date,hvec,"-",
             old_idx,old_date,old_sizevec,old_sum,old_md5,&old_crc32,
             (flag&2)|256|512);
   if(ret<=0)
     goto ex;
   if(ret==2)
     flag|= (1<<12);
   else
     flag&= ~(1<<12);
   if(flag&2) 
     {ret= 1; goto ex;}
   Intvec_to_dec(old_sizevec,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"\nFinally comparing with record :\n%s %s %s %s\n\n",
                  old_idx,old_date,size_text,old_sum); 

   if(plan->pre_checksum_size>0) {
     Intvec_from_double(hvec,CdbplivS,plan->pre_checksum_size,0);
     Intvec_add(sizevec,CdbplivS,old_sizevec,hvec,1);
   }
   use_handle= 256;
   pre_crc= 0;
   goto determine_checksum;

 } else if(flag&1) {
   /* to erase possible pacifier messages of program raedchen */
   fprintf(stderr,"\r                                                   \r");
   /* compare length and checksum */
   if(flag&4) /* add length consumed by preview */
     if(plan->pre_checksum_size>0) {
       Intvec_from_double(hvec,CdbplivS,plan->pre_checksum_size,0);
       Intvec_add(sizevec,CdbplivS,sizevec,hvec,0);
     }
   if(Intvec_cmp(sizevec,CdbplivS,old_sizevec,0)!=0) {
     fprintf(stderr,"SIZE MISMATCH. FALSCHE GROESSE :\n");
     Intvec_to_dec(old_sizevec,CdbplivS,size_text,sizeof(size_text),0);
     fprintf(stderr,"record=%s",size_text);
     Intvec_to_dec(sizevec,CdbplivS,size_text,sizeof(size_text),0);
     fprintf(stderr,"   stdin=%s\n\n",size_text);
     ret= 0; goto ex;
   }
   if(flag&(1<<12)) {
     for(i=0;i<16;i++)
       if(md5[i]!=old_md5[i])
     break;
     if(i<16) {
       fprintf(stderr,"MD5 CHECKSUM MISMATCH. FALSCHE MD5 PRUEFSUMME :\n");
       fprintf(stderr,"record= ");
       for(i=0;i<16;i++)
         fprintf(stderr,"%2.2x",old_md5[i]);
       fprintf(stderr,"\nstdin = ");
       for(i=0;i<16;i++)
         fprintf(stderr,"%2.2x",md5[i]);
       fprintf(stderr,"\n\n");
       ret= 0; goto ex;
     }
   } else {
     if(crc32!=old_crc32) {
       fprintf(stderr,"CRC32 CHECKSUM MISMATCH. FALSCHE CRC32 PRUEFSUMME :\n");
       fprintf(stderr,"record= %lX\nstdin = %lX\n",old_crc32,crc32);
       ret= 0; goto ex;
     }
   }
   fprintf(stderr,
             "OK.  Record and data input match. Daten passen zum Record.\n\n");
 } else {
   if(Intvec_sign(sizevec,CdbplivS,0)<=0) {
     fprintf(stderr,"\n%s : EOF at stdin without any data\n",cmd);
     ret= 0; goto ex;
   }

   moment= time(0);
   Intvec_from_double(hvec,CdbplivS,plan->pre_checksum_size,0);
   if(plan->pre_checksum_size>0.0 &&
      Intvec_cmp(sizevec,CdbplivS,hvec,0)>=0) {
     sprintf(pre_idx,"%s_pre",idx);
     Checkrecord_new(&check_rec,pre_idx,moment,0.0,hvec,!!(flag&(1<<12)),
                     pre_crc32,pre_md5,0);
     if(check_rec==NULL) {
keine_checkung:;
       fprintf(stderr,"\nfailed to allocate checkrecord object\n");
       ret= -1; goto ex;
     }
     Checkrecord_to_text(check_rec,record_text,0);
     fprintf(fp,"%s\n",record_text);
     Checkrecord_destroy(&check_rec,0);
   }  
   Checkrecord_new(&check_rec,idx,moment,0.0,
                   sizevec,!!(flag&(1<<12)),crc32,md5,0);
   if(check_rec==NULL) 
     goto keine_checkung;
   Checkrecord_to_text(check_rec,record_text,0);

   if(plan->checksum_pad_size>0.0) {
     Intvec_copy(sizevec,CdbplivS,posvec,0);
     Intvec_from_double(min_dutyvec,CdbplivS,plan->checksum_pad_size,0);
     if(plan->block_checksum_size>0)
       Checkrecord_set_blocklist(check_rec,blocklist,0);
     else
       Checkrecord_set_blocklist(check_rec,NULL,0);
     ret= Checkrecord_write_tags(check_rec,stdout,posvec,min_dutyvec,countvec,
                                 0);
     if(ret<=0) {
       Intvec_to_dec(min_dutyvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,"\nfailed to write %s bytes of checksum padding\n",
                      size_text);
       fprintf(fp,"%s\n",record_text);
       ret= -1; goto ex;
     }

     if(plan->block_checksum_size>0) {
       if(plan->blocklist_fileadr[0]!=0) {
         /* write blocklist to file system */
         ret= Blksum_list_to_file(blocklist,plan->blocklist_fileadr,
                                  check_rec,2);
         if(ret<=0)
           goto ex;
       }
     }

     fprintf(stderr,"\nNote: %s added %.lf bytes of checksum padding\n",
                    "cd_backup_planer",
                    Intvec_to_double(countvec,CdbplivS,0));
   }

   fprintf(fp,"%s\n",record_text);
   Checkrecord_destroy(&check_rec,0);

   /* ??? >>> also report this to stderr */
 
 }
 ret= 1;
ex:;
 if(handle!=NULL)
   free((char *) handle);
 Checkrecord_destroy(&check_rec,0);
 Blksum_list_destroy(&blocklist,0);
 fflush(stdout);
 fflush(stderr);
 Sfile_fclose(fp,0);
 return(ret);
}


/* >>> leave this at 16 until larger random chunks get produced */
/* must be multiple of 16 */
#define Backup_plan_random_isteP 16

/* multiplication factor for fast_random */
#define Backup_plan_random_queuE 256

int Backup_plan__next_random(ctx,new_byte,result,flag)
Sregex_md5_ctX *ctx;
unsigned char *new_byte;
char result[Backup_plan_random_isteP];
int flag;
{
 int ret;
 Sregex_md5_ctX *pre_ctx= NULL;

 /* >>> produce Backup_plan_random_isteP random bytes rather than 16 */;
 if((*new_byte)==251) /* a prime number (hoping to lengthen any MD5 cycle) */
   *new_byte= 0;
 else
   (*new_byte)++;
 ret= Sregex_md5(&ctx,new_byte,1,result,0);
 if(ret<=0) {
failed:;
   fprintf(stderr,"Failed to continue MD5 sequence.\n");
   return(ret);
 }
 /* obtain intermediate value */
 ret= Sregex_md5(&pre_ctx,(unsigned char *) ctx,0,result,1|4);
 if(ret<=0)
   goto failed;
 Sregex_md5(&pre_ctx,"",0,result,2|(1<<15));

 return(1);
}


int Backup_plan_write_random(plan,target,size,startseq,flag)
struct Backup_plaN *plan;
char *target; /* fileaddress or "-" for stdout */
double size;
char *startseq; /* 0 terminated initial text. A kind of key */
int flag;
{
 int ret,iduty,no_random= 0,fast_random= 0,l,i,k,kk,start_idx;
 int dutyvec[CdbplivS],istepvec[CdbplivS];
 int indexcount= CdbplivS;
 FILE *fp= NULL;
 char num_text[8],*startpt;
 char result_queue[Backup_plan_random_queuE][Backup_plan_random_isteP];
 int istep= Backup_plan_random_isteP*Backup_plan_random_queuE;
 double duty;
 Sregex_md5_ctX *ctx= NULL;
 unsigned char new_byte;

 duty= size;

 fp= Sfile_fopen(target,"wb",0);
 if(fp==NULL) 
   {ret=0 ; goto ex;}
 if(startseq[0]=='=' && startseq[1]!=0) {
   no_random= 1;
   if(startseq[1]=='~' && startseq[2]!=0) {
     start_idx= 2;
     l= strlen(startseq+2);
     for(i= 0;i<l;i++)
       startseq[start_idx+i]= ~startseq[start_idx+i];
   } else {
     start_idx= 1;
     l= strlen(startseq+1);
   }
   if(l>16)
     l= 16;
   for(i=0;i<istep;i++)
     if(startseq[start_idx+1]!=0)
       ((char *) result_queue)[i]= startseq[start_idx+i%l];
     else
       ((char *) result_queue)[i]= startseq[start_idx];
 } else if(startseq[0]!=0) {
   startpt= startseq;
   if(startpt[0]=='-' && startpt[1]!=0) 
     startpt++;
   else
     fast_random= 1;
   ret= Sregex_md5(&ctx,startseq,strlen(startseq),result_queue[0],1);
   if(ret<=0) {
     fprintf(stderr,"Failed to initialize MD5 sequence from seed text.\n");
     goto ex;
   }
   if(fast_random) {
     for(i=0;i<Backup_plan_random_queuE;i++) {
       ret= Backup_plan__next_random(ctx,&new_byte,result_queue[i],0);
       if(ret<=0)
         goto ex;
     }
   }
 } else {
   no_random= 1;
   for(i=0;i<istep;i++)
     ((char *) result_queue)[i]= 0;
 }


 /* handle up to at least 90 bit numbers with not too much inaccuracy */
 ret= Very_large_chunk_manager(&duty,dutyvec,&indexcount,
                               ((double) -istep),istepvec,1|2|4);
 if(ret<0) {
   fprintf(stderr,"\n");
   fprintf(stderr,
       "Excuse me. You really want to produce a Xobibyte (1<<90) or more ?\n");
   fprintf(stderr,"I'm not such a kind of program, you must know.\n");
   fprintf(stderr,"\n");
   ret= 0;
   goto ex;

 }
 while(1) {
   iduty= Very_large_chunk_manager(&duty,dutyvec,&indexcount,
                                   ((double) -istep),istepvec,8);
   if(iduty<=0)
 break;
   if(no_random) {
     ret= fwrite(result_queue,iduty,1,fp);
   } else {
     if(fast_random) {
       ret= Backup_plan__next_random(ctx,&new_byte,result_queue[0],0);
       if(ret<=0)
         goto ex;
       for(k=1;k<Backup_plan_random_queuE;k++)
         for(kk=0;kk<Backup_plan_random_isteP/sizeof(int);kk++)
            ((int *) result_queue[k])[kk]^= ((int *) result_queue[k-1])[kk];
       ret= fwrite(result_queue,iduty,1,fp);
     } else {
       for(i= 0;i*Backup_plan_random_isteP<iduty;i++) {
         ret= Backup_plan__next_random(ctx,&new_byte,result_queue[i],0);
         if(ret<=0)
           goto ex;
       }  
       ret= fwrite(result_queue,iduty,1,fp);
     }
   }
   if(ret<=0) {
     fprintf(stderr,"Error when writing random data to target '%s' :\n",target);
     perror("");
     ret= 0;
     goto ex;
   }
 }

 ret= 1;
ex:;
 if(ret<=0) {
   Sfile_scale(duty,num_text,7,1e5,1);
   fprintf(stderr,
           "Random data production aborted. Left undelivered about  %s .\n",
           num_text);
 }
 Sregex_md5(&ctx,"",0,result_queue[0],1<<15);
 if(fp!=stdout && fp!=NULL)
   fclose(fp);
 return(ret);
}


int Backup_plan_verify_filelist(plan,listadr,prepend_dir,flag)
struct Backup_plaN *plan;
char *listadr;
char *prepend_dir;
int flag;
/*
 bit0-3= mode
    0= report matching files
    1= delete all verified files (DANGEROUS !!!)
    2= report absolute addresses
    3= report input lines and absolute addresses 

 bit7= expect  target=source  and process only the source part
 bit13= with bit0 : allow insecure parameters
 bit14= process target addresses rather than source addresses
*/
{
 FILE *fp;
 int ret= 0,popened= 0,pacifier,mode,frst_mode,hflag;
 char initial_wd[4096];

 mode= flag&15;
 if(mode==0)
   frst_mode= 4;
 else if(mode==1)
   frst_mode= 3;
 else if(mode==2)
   frst_mode= 5;
 else if(mode==3)
   frst_mode= 6;
 else
   return(0);
 pacifier= Cd_backup_planer_pacifier_intervaL;
 Cd_backup_planer_pacifier_intervaL= -1;
 popened= Open_listfile(&fp,listadr,"r",1);
 if(popened!=2)
   popened= 0;
 if(flag&(1<<14))
   strcpy(initial_wd,prepend_dir); /* with bit14, prepend_dir is ignored */
 hflag= (flag&((1<<7)|(1<<14)));
 if(flag&(1<<13))
   hflag|= (1<<16);
 hflag|= frst_mode;
 ret= Fileliste_read_serialized_tree(NULL,fp,NULL,NULL,NULL,initial_wd,NULL,
                                     prepend_dir,hflag);
 Cd_backup_planer_pacifier_intervaL= pacifier;
 if(popened) {
   if(fp!=NULL)
     pclose(fp);
 } else
   Sfile_fclose(fp);
 return(ret);
}


int Backup_plan_cut_file(plan,src_name,start_byte,bytes,target_name,count,flag)
struct Backup_plaN *plan;
char *src_name;
double start_byte;
double bytes;
char *target_name;
double *count;
int flag;
/*
 bit0= do not issue pacifier messages
*/
{
 int ret,read_next,position_by_read= 0,blocksize;
 int seekvec[CdbplivS],readvec[CdbplivS];
 int bsvec[CdbplivS],vlcm_flag;
 int indexcount= CdbplivS;
 int seekidxcount= CdbplivS;
 int premature_eof= 0,alignment_size= 0,aligned_blocksize,block_count= 0;
 FILE *fp_in= NULL,*fp_out= NULL;
 double bytes_to_read,bytes_read,dummy,b30= 1073741824.0;
 char buffer[65536],msg[4096+100],size_text[8],pos_text[8];
 struct stat stbuf;

 blocksize= sizeof(buffer);

 if(!(flag&1)) {
   Sfile_scale(start_byte,pos_text,7,1e5,1);
   Sfile_scale(bytes,size_text,7,1e5,1);
   sprintf(msg,"-cut_file %s %s %s %s",
               src_name,pos_text,size_text,target_name);
   Print_pacifier(0,msg,1|4|8|16|32);
 }
 *count= 0;
 if(strcmp(src_name,"-")==0) {
   fp_in= stdin;
   position_by_read= 1;
 } else
   if(Cd_backup_planer_call_staT(src_name,&stbuf)!=-1) {
     if(S_ISBLK(stbuf.st_mode)) 
       if(stbuf.st_blksize<blocksize) {
         blocksize= stbuf.st_blksize;
/*
         fprintf(stderr,"Block device detected. Block size is %d\n",
                        blocksize); 
*/
       }
     fp_in= fopen(src_name,"rb");
   }
 if(start_byte>=b30*b30*16.0 || bytes>=b30*b30*16.0)
   position_by_read= 1;
 if(fp_in==NULL) {
   fprintf(stderr,"While read-opening for -cut_file file '%s' :\n",src_name);
   perror("");
   {ret= 0; goto ex;}
 }
 if(strcmp(target_name,"-")==0) {
   fp_out= stdout;
 } else {
   ret= Type_and_size(target_name,&dummy,1);
   if(ret!=-1) {
     fprintf(stderr,"Target file for -cut_file already existing : '%s'\n",
                     target_name);
     {ret= 0; goto ex;}
   }
   fp_out= fopen(target_name,"wb");
   if(fp_in==NULL) {
     fprintf(stderr,"While write-opening for -cut_file file '%s' :\n",
                    target_name);
     perror("");
     {ret= 0; goto ex;}
   }
 }

 bytes_to_read= bytes;
 ret= Very_large_chunk_manager(&bytes_to_read,readvec,&indexcount,
                               ((double) -blocksize),bsvec,1|2|4);
 if(ret<0) {
   fprintf(stderr,"Oversized byte count : %.lf\n",bytes);
   {ret= 0; goto ex;}
 }
 alignment_size= blocksize;
 if(position_by_read) {
   bytes_to_read= start_byte;
   ret= Very_large_chunk_manager(&bytes_to_read,seekvec,&seekidxcount,
                                 ((double) -blocksize),bsvec,1|2|4);
   if(ret<0) {
     fprintf(stderr,"Oversized start address : %.lf\n",start_byte);
     {ret= 0; goto ex;}
   }
   while(1) {
     read_next= Very_large_chunk_manager(&bytes_to_read,seekvec,&seekidxcount,
                                         ((double) -blocksize),bsvec,8); 
     if(read_next==0)
   break;
     alignment_size= blocksize-(read_next%blocksize);
     ret= fread(buffer,read_next,1,fp_in);
     if(ret<=0) {
       fprintf(stderr,
             "While trying to position read pointer for -cut_file at %.lf :\n",
             start_byte);
       if(feof(fp_in))
         fprintf(stderr,"Premature EOF\n");
       else
         perror("");
       {ret= 0; goto ex;}
     }
     if(!(flag&1)) {
       if(++block_count>=256) {
         block_count= 0;
         Sfile_scale(start_byte-bytes_to_read,pos_text,7,1e5,1);
         Sfile_scale(start_byte,size_text,7,1e5,1);
         sprintf(msg,"  cut %s of %s bytes",pos_text,size_text);
         Print_pacifier(0,msg,4|16|32);
       }
     }
   }
 } else {
#ifdef Cd_backup_planer_fseekO
   ret= Cd_backup_planer_fseekO(fp_in,(off_t) start_byte,SEEK_SET);
#else
   ret= fseek(fp_in,(long) start_byte,SEEK_SET);
#endif /* Cd_backup_planer_fseekO */
   if(ret<0) {
     fprintf(stderr,
             "While trying to position read pointer for -cut_file at %.lf :\n",
             start_byte);
     perror("");
     {ret= 0; goto ex;}
   }
   alignment_size= blocksize-(((off_t) start_byte)%blocksize);
 }

 premature_eof= 0;

 /* re-align to block boundary */
 aligned_blocksize= blocksize;
 if(alignment_size>0)
   blocksize= alignment_size;
 vlcm_flag= 0;
 while(1) {
   read_next= Very_large_chunk_manager(&bytes_to_read,readvec,&indexcount,
                                       ((double) -blocksize),bsvec,vlcm_flag); 
   if(blocksize!=aligned_blocksize) {
     blocksize= aligned_blocksize;
     vlcm_flag= 0;
   } else
     vlcm_flag= 8;
   if(read_next==0)
 break;
   ret= fread(buffer,1,read_next,fp_in);
   if(ret<read_next) {
     premature_eof= 1;
     if(ret>0) {
       bytes_to_read+= (read_next-ret);
       read_next= ret;
     }
   }
   if(ret<=0) {
premature_read_eof:;
     fprintf(stderr,
             "While trying to read data for -cut_file from : '%s'\n",src_name);
     if(feof(fp_in))
       fprintf(stderr,"Premature EOF\n");
     else
       perror("System error message");
     fprintf(stderr,"after %.lf of %.lf bytes\n",(*count),bytes);
     {ret= 0; goto ex;}
   }
   bytes_read= read_next;
   ret= fwrite(buffer,bytes_read,1,fp_out);
   if(ret<1) {
     fprintf(stderr,
             "While trying to write data for -cut_file to : '%s'\n",
             target_name);
     perror("");
     {ret= 0; goto ex;}
   }
   (*count)= bytes-bytes_to_read;
   if(premature_eof)
     goto premature_read_eof;
   if(!(flag&1)) {
     if(++block_count>=256) {
       block_count= 0;
       Sfile_scale(bytes-bytes_to_read,pos_text,7,1e5,1);
       Sfile_scale(bytes,size_text,7,1e5,1);
       sprintf(msg,"  cut %s of %s bytes",pos_text,size_text);
       Print_pacifier(0,msg,4|16|32);
     }
   }
 }

ex:;
 if(!(flag&1))
   Print_pacifier(0,"",1|8|16);
 if(fp_in!=NULL && fp_in!=stdin)
   fclose(fp_in);
 if(fp_out!=NULL && fp_out!=stdout)
   fclose(fp_out);
 if(ret>0 && strcmp(src_name,"-")!=0 && strcmp(target_name,"-")!=0)
   Sfile_copy_properties(src_name,target_name,1|(2*(geteuid()==0))|4);
 return(ret);
}


int Backup_plan_extract_tag_tail(plan,limit_text,flag)
struct Backup_plaN *plan;
char *limit_text;
int flag;
/*
 bit14= accept tag at wrong position
*/
{
 int ret;
 struct Checkrecord_taG *scanstate= NULL;
 struct CheckrecorD *check_rec= NULL;
 char size_text[80];
 char *handle= NULL;
 unsigned char md5[16];
 unsigned long crc32;
 int basevec[CdbplivS],limitvec[CdbplivS],countvec[CdbplivS];
 double wert;
 FILE *fpin;

 wert= Scanf_io_size(limit_text,0);
 if(wert<=0) {
   fprintf(stderr,"-extract_tag_tail: unusable size_limit %s\n",limit_text);
   return(0);
 }
 ret= Backup_plan_open_tag_source(plan,"-extract_tag_tail",&fpin,basevec,0);
 if(ret<=0)
   goto ex;
 if(plan->checksum_scan_skip>0)
   fprintf(stderr,"\nNote: %.f bytes of input skipped.\n",
           plan->checksum_scan_skip);

 Intvec_from_double(limitvec,CdbplivS,wert,0);
 Intvec_from_double(countvec,CdbplivS,0.0,0);
 ret= Scan_checksum(&scanstate,&handle,fpin,md5,&crc32,
                    basevec,limitvec,countvec,
                    &check_rec,(1<<11)|(1<<12)|(1<<13)|(flag&(1<<14)));
 if(ret>0) {
   Intvec_to_dec(countvec,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"\nExtracted : %s bytes\n",size_text);
 } else {
   fprintf(stderr,"\nNo data extracted\n");
 }
 if(handle!=NULL)
   free((char *) handle);
 Checkrecord_destroy_all(&check_rec,0);
 Checkrecord_tag_destroy(&scanstate,0);
ex:;
 return(ret);
}


int Backup_plan_process_blocklist(plan,action,listfile,skip,source,target,flag)
struct Backup_plaN *plan;
char *action;
char *listfile;
int skip;
char *source;
char *target;
int flag;
{
 struct Checkrecord_taG *scanstate= NULL;
 char *handle= NULL;   /* should be (void *)  */
 char *dpt;
 int mode,i,ret;
 int posvec[CdbplivS],verylargevec[CdbplivS],countvec[CdbplivS];
 int startblock[CdbplivS],endblock[CdbplivS];
 struct CheckrecorD *found_rec= NULL,*rec;
 unsigned char md5[16];
 unsigned long crc32;
 FILE *fp= NULL;

 Intvec_from_double(startblock,CdbplivS,0.0,0);
 Intvec_from_double(endblock,CdbplivS,-1.0,0);
 if(strcmp(action,"report")==0) {
   mode= 0;
 } else if(strcmp(action,"collect")==0) {
   mode= 1;

 } else if(strncmp(action,"mend",4)==0) {

   mode= 2;
   if(action[4]==':') {
     /* mend[:startblock[:endblock]]  */
     /* special block number : end = -1 */;
     dpt= strchr(action+5,':');
     if(dpt!=NULL) {
       if(strcmp(dpt+1,"end")==0 || strcmp(dpt+1,"\"end\"")==0) 
         Intvec_from_double(endblock,CdbplivS,-1.0,0);
       else 
         Intvec_from_double(endblock,CdbplivS,Scanf_io_size(dpt+1,0),0);
       *dpt= 0;
     }
     dpt= action+4;
     if(strcmp(dpt+1,"end")==0 || strcmp(dpt+1,"\"end\"")==0) 
       Intvec_from_double(startblock,CdbplivS,-1.0,0);
     else 
       Intvec_from_double(startblock,CdbplivS,Scanf_io_size(dpt+1,0),0);
     mode|= 256;
   }
 } else if(strcmp(action,"merge")==0) {
   mode= 3;
 } else {
   fprintf(stderr,"-process_blocklist : unknown mode : %s\n",action);
   {ret= 0; goto ex;}
 }

 /* obtain blocklist */
 fp= Sfile_fopen(listfile,"rb",0);
 if(fp==NULL)
   {ret= 0; goto ex;}
 Intvec_from_double(posvec,CdbplivS,0.0,0);
 Intvec_from_double(verylargevec,CdbplivS,1.0e24,0);
 ret= Scan_checksum(&scanstate,&handle,fp,md5,&crc32,
                    posvec,verylargevec,countvec,
                    &found_rec,((skip==0)<<1)|64|128|1024|(1<<12)|(1<<13));
 if(ret<=0 || found_rec==NULL) {
   fprintf(stderr,"No valid block checksum list found\n");
   {ret= 0; goto ex;}
 }
 rec= found_rec;
 for(i=0;i<=skip;i++) {
   if(rec==NULL) {
     fprintf(stderr,
          "Cannot skip %d. Only %d valid block checksum lists found\n",skip,i);
     {ret= 0; goto ex;}
   }
   if(i<skip)
     rec= rec->next;
 }
 ret= Blksum_list_process(rec->blocklist,source,target,
                          startblock,endblock,mode);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(fp!=NULL)
   Sfile_fclose(fp);
 return(ret);
}


int Backup_plan_permute_blocks(plan,source,blocksize,max_blocks,mode,headsize,
                               fpo,flag)
struct Backup_plaN *plan;
char *source;
double blocksize;
double max_blocks;
int mode;
double headsize;                         /* counted in blocks, 0= automatic */
FILE *fpo;
int flag;
/*
 bit0= reverse permutation
 bit1= with bit0 : skip file header and use blocksize, max_blocks, mode
       without bit0 : do not write file header
 bit2= use max_blocks*blocksize as file_size for block devices, 
       ignore max_blocks with other file types
*/
{
 int ret,read_size,file_type,l,hblocks,nblocks,pacifier_count= 0;
 int hvec[CdbplivS];
 double file_size,complete_blocks,remainder,idx,loopcount,looplimit,todo;
 double pad_len,block_offset= 0.0,num,datastart;
 FILE *fpi= NULL;
 char buffer[64*1024],size_text[160],pacifier_msg[160+20];

 if(flag&1)
   goto unpermute;

 /* determine size of source and number of blocks */;
 file_type= Type_and_size(source,&file_size,1|4|((!(flag&4))<<3));
 if(file_type<=0 || file_type==2 || file_type==4) {
   fprintf(stderr,"Not a usable file for -permute_blocks : %s\n",source);
   {ret= 0; goto ex;}
 }
 if(file_type==6 && (flag&4))
   file_size= max_blocks*blocksize;
 if(flag&4)
   max_blocks= 0.0;
 complete_blocks= file_size/blocksize;
 /* poor man's floor() */
 ret= Intvec_from_double(hvec,CdbplivS,complete_blocks,0);
 if(ret<=0) {
   fprintf(stderr,"File size allegedly exceeds %d bit addresses.\n",
                  30*CdbplivS);
   {ret= 0; goto ex;}
 }
 complete_blocks= Intvec_to_double(hvec,CdbplivS,0);
 remainder= file_size-complete_blocks*blocksize;
 if(complete_blocks+(remainder>0)>max_blocks && max_blocks>0) {
   complete_blocks= max_blocks;
   remainder= 0.0;
   Intvec_from_double(hvec,CdbplivS,complete_blocks,0);
   Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
   fprintf(stderr,"Note: source data truncated to %s blocks\n",size_text);
 }

 fpi= Sfile_fopen(source,"rb",0);
 if(fpi==NULL)
   {ret= 0; goto ex;}

 /* write header */;
 sprintf(buffer,"%s",Cdbpl_permutation_starT);
 Intvec_from_double(hvec,CdbplivS,blocksize,0);
 Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
 sprintf(buffer+strlen(buffer),"blocksize=%s\n",size_text);
 Intvec_from_double(hvec,CdbplivS,complete_blocks,0);
 Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
 sprintf(buffer+strlen(buffer),"complete_blocks=%s\n",size_text);
 Intvec_from_double(hvec,CdbplivS,remainder,0);
 Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
 sprintf(buffer+strlen(buffer),"remainder=%s\n",size_text);
 sprintf(buffer+strlen(buffer),"permutation_key=%d\n",mode);

 l= strlen(buffer);
 hblocks= l/((int) blocksize);
 if(hblocks*blocksize!=l)
   hblocks++;
 while(1) {
   sprintf(buffer+l,"datastart=%d\n",hblocks);
   nblocks= strlen(buffer)/((int) blocksize);
   if(nblocks*blocksize!=strlen(buffer))
     nblocks++;
   if(hblocks==nblocks)
 break;
   hblocks++;
 }
 l= strlen(buffer);

 if(headsize>0) {
   if(hblocks>headsize) {
     fprintf(stderr,
             "Actual header length %d exceeds prescribed size of %.lf bytes\n",
             l,headsize*blocksize);
     {ret= 0; goto ex;}
   }
 } else {
   if(flag&2) {
     headsize= 0.0;
     goto process_blocks; /* this is not really meant as a spaghetto */    
   }
   headsize= hblocks;
 }
 
 if(headsize>1.0)
   fprintf(stderr,
         "Warning: It is better to restrict header size to a single block.\n");
 if(blocksize<256.0) 
   fprintf(stderr,
         "Warning: It is recommended not to use blocksizes below 256.\n");
 fprintf(stderr,"-permute_blocks header (occupies %.lf block%s) :\n",
                headsize,(headsize!=1.0?"s":""));
 fprintf(stderr,"%s",buffer);
 looplimit= complete_blocks+(remainder>0);
 Intvec_from_double(hvec,CdbplivS,looplimit,0);
 Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
 fprintf(stderr,"Note: blockcount for recovery will be %s\n",size_text);

 ret= Write_broker_or_file(fpo,buffer,l,0);
 if(ret<=0)
   goto ex;
 
 /* pad to full block length */
 pad_len= headsize*blocksize-l;
 ret= Pad_broker_or_file(fpo,pad_len,0);
 if(ret<=0)
   goto ex;

 goto process_blocks; /* this is not really meant as a spaghetto */    

unpermute:;

 fpi= Sfile_fopen(source,"rb",0);
 if(fpi==NULL)
   {ret= 0; goto ex;}
 
 if(flag&2) {
   complete_blocks= max_blocks;
   remainder= 0;
   block_offset= headsize;
   if(block_offset<0)
     block_offset= 0.0;
   fprintf(stderr,"Skipping file header: %.lf bytes \n",block_offset);
 } else { 
   /* read header : block_size, complete_blocks , remainder , mode */;
   fprintf(stderr,"Reading file header :\n");
   l= 0;
   if(fgets(buffer,80,fpi)==NULL) {
not_a_permuted_file:;
     fprintf(stderr,
             "This does not look like a scdbackup block permutation file.\n");
     fprintf(stderr,"(decision made after reading %d bytes)\n",l);
     {ret= 0; goto ex;}
   }
   l+= strlen(buffer);
   if(strcmp(buffer,Cdbpl_permutation_starT)!=0)
     goto not_a_permuted_file;
   fprintf(stderr,"%s",buffer);
   if(fgets(buffer,80,fpi)==NULL)
     goto not_a_permuted_file;
   l+= strlen(buffer);
   ret= Decode_numvar(buffer,"blocksize=",&blocksize,0);
   if(ret<=0)
     goto not_a_permuted_file;
   fprintf(stderr,"%s",buffer);
   if(fgets(buffer,80,fpi)==NULL)
     goto not_a_permuted_file;
   l+= strlen(buffer);
   ret= Decode_numvar(buffer,"complete_blocks=",&complete_blocks,0);
   if(ret<=0)
     goto not_a_permuted_file;
   fprintf(stderr,"%s",buffer);
   if(fgets(buffer,80,fpi)==NULL)
     goto not_a_permuted_file;
   l+= strlen(buffer);
   ret= Decode_numvar(buffer,"remainder=",&remainder,0);
   if(ret<=0)
     goto not_a_permuted_file;
   fprintf(stderr,"%s",buffer);
   if(fgets(buffer,80,fpi)==NULL)
     goto not_a_permuted_file;
   l+= strlen(buffer);
   ret= Decode_numvar(buffer,"permutation_key=",&num,0);
   if(ret<=0)
     goto not_a_permuted_file;
   if(num<0||num>7.0)
     goto not_a_permuted_file;
   mode= num;
   fprintf(stderr,"%s",buffer);
   if(fgets(buffer,80,fpi)==NULL)
     goto not_a_permuted_file;
   l+= strlen(buffer);
   ret= Decode_numvar(buffer,"datastart=",&datastart,0);
   if(ret<=0)
     goto not_a_permuted_file;
   if(datastart<=0)
     goto not_a_permuted_file;
   fprintf(stderr,"%s",buffer);

   /* read padding */;
   if(headsize>0) {
     pad_len= headsize*blocksize-l;
     block_offset= headsize;
   } else {
     pad_len= datastart*blocksize-l;
     block_offset= datastart;
   }
   while(pad_len>0) {
     if(pad_len>sizeof(buffer))
       read_size= sizeof(buffer);
     else
       read_size= pad_len;
     ret= fread(buffer,read_size,1,fpi);
     if(ret!=1)
       goto not_a_permuted_file;
     pad_len-= read_size;
   }
 }

process_blocks:; 
 /* loop over blocks */;
 looplimit= complete_blocks+(remainder>0);
 for(loopcount= 0.0;loopcount<looplimit;loopcount++) {
   /* compute source block number */;
   idx= Permutation_index(loopcount,looplimit,mode,flag&1);
   /* position at address */;
   ret= Fseek_double(fpi,(idx+block_offset)*blocksize,0);
   if(ret<0)
     {ret= 0; goto ex;}
   /* read block and write to stdout */;
   if(idx>=complete_blocks && !(flag&1))
     todo= remainder;
   else if(loopcount>=complete_blocks && (flag&1))
     todo= remainder;
   else
     todo= blocksize;
   while(todo>0) {
     if(todo>sizeof(buffer))
       read_size= sizeof(buffer);
     else
       read_size= todo;
     ret= fread(buffer,read_size,1,fpi);
     if(ret!=1) {
       Intvec_from_double(hvec,CdbplivS,idx,0);
       Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
       fprintf(stderr,"While reading block %s of file %s :\n",
                      size_text,source);
       if(errno>0)
         perror("");
       else
         fprintf(stderr,"Unexpected end of file\n");
       ret= 0; goto ex;
     }
     ret= Write_broker_or_file(fpo,buffer,read_size,1);
     if(ret<=0)
       goto ex;
     todo-= read_size;
   }
   if((!(flag&1)) && idx>=complete_blocks) {
     /* pad up last block to block size */;
     pad_len= blocksize-remainder;
     ret= Pad_broker_or_file(fpo,pad_len,0);
     if(ret<=0)
       goto ex;
   }
   pacifier_count++;
   if((pacifier_count%10)==0) {
     Intvec_from_double(hvec,CdbplivS,loopcount,0);
     Intvec_add_b30(hvec,CdbplivS,1,0);
     Intvec_to_dec(hvec,CdbplivS,size_text,sizeof(size_text),0);
     sprintf(pacifier_msg,"  %s blocks read",size_text);
     Print_pacifier(0,pacifier_msg,4|32);
     pacifier_count= 0;
   }
 }

 {ret= 1; goto ex;}

 ret= 1;
ex:;
 if(pacifier_msg[0]!=0)
   Print_pacifier(0,"",1|8|16);
 if(fpi!=NULL)
   fclose(fpi);
 if(ret>0)
   fprintf(stderr,"-permute_blocks complete.\n");
 return(ret); 
}


int Backup_plan_multiply_volume(plan,source,factor,fpo,flag)
struct Backup_plaN *plan;
char *source;
int factor;
FILE *fpo;
int flag;
{
 int ret,read_size,file_type,count;
 int out_count[CdbplivS];
 char buffer[64*1024],mark_text[160],size_text[40];
 FILE *fpi= NULL;
 double read_bytes,file_size;

 /* determine source */;
 file_type= Type_and_size(source,&file_size,1|4|8);
 if(file_type<=0 || file_type==2 || file_type==4) {
   fprintf(stderr,"Not a usable file for -multiply_volume : %s\n",source);
   {ret= 0; goto ex;}
 }

 Intvec_from_double(out_count,CdbplivS,0.0,0);
 for(count= 0; count<factor; count++) {
   fpi= Sfile_fopen(source,"rb",0);
   if(fpi==NULL)
     {ret= 0; goto ex;}
   for(read_bytes= 0.0; read_bytes<file_size;read_bytes+= read_size) {
     if(file_size-read_bytes>sizeof(buffer))
       read_size= sizeof(buffer);
     else
       read_size= file_size-read_bytes;
     ret= fread(buffer,read_size,1,fpi);
     if(ret!=1) {
       if(errno>0)
         perror("");
       else
         fprintf(stderr,"-multiply_volume : Unexpected end of file %s\n",
                        source);
       {ret= 0; goto ex;}
     }
     ret= Write_broker_or_file(fpo,buffer,read_size,0);
     if(ret<=0)
       goto ex;
     Intvec_add_b30(out_count,CdbplivS,read_size,0);
   }
   fclose(fpi);
   fpi= NULL;

   if(count<factor-1) {
     Intvec_to_dec(out_count,CdbplivS,size_text,sizeof(size_text),0);
     sprintf(mark_text,"%s%s \n",Cdbpl_multiplication_marK,size_text);
     ret= Write_broker_or_file(fpo,mark_text,strlen(mark_text),0);
     if(ret<=0)
       goto ex;
     Intvec_add_b30(out_count,CdbplivS,strlen(mark_text),0);
   }
 }
 ret= 1;
ex:;
 if(fpi!=0)
   fclose(fpi);
 return(ret);
}


int Backup_plan_extract_multiplied(plan,source,skip,fpo,flag)
struct Backup_plaN *plan;
char *source;
int skip;
FILE *fpo;
int flag;
{
 int ret,bufcount,bufpos,hflag,write_mode= 0,count,old_bufpos,skip_remainder;
 int pacifier_count= 0;
 int out_count[CdbplivS],in_count[CdbplivS],posvec[CdbplivS];
 int copy_size[CdbplivS];
 double r;
 char buffer[64*1024],size_text[40],pacifier_msg[80];
 Sregex_md5_ctX *ctx= NULL;
 unsigned long crc32;
 struct CheckrecorD *check_rec= NULL;
 struct Checkrecord_taG *scanstate= NULL;
 FILE *fpi= NULL;

 Intvec_from_double(in_count,CdbplivS,0.0,0);
 Intvec_from_double(out_count,CdbplivS,0.0,0);
 Intvec_from_double(posvec,CdbplivS,0.0,0);
 Intvec_from_double(copy_size,CdbplivS,0.0,0);
 fprintf(stderr,"-extract_multiplied : skipping %d copies\n",skip);
 fpi= Sfile_fopen(source,"rb",0);
 if(fpi==NULL)
   {ret= 0; goto ex;}
 write_mode= (skip<=0);
 bufcount= bufpos= 0;
 count= 0;
 while(1) {
   ret= fread(buffer,1,sizeof(buffer),fpi);
   if(ret<=0) {
     if(ferror(fpi) && Intvec_cmp(copy_size,CdbplivS,out_count,0)>0) {
       Print_pacifier(0,"",1|64);
       perror("");
       {ret= 0; goto ex;}
     } else if(count<skip) {
       Print_pacifier(0,"",1|64);
       fprintf(stderr,
          "-extract_multiplied : encountered end-of-file during skip phase\n");
       {ret= 0; goto ex;}
     } else
       {ret= 1; goto ex;}
   }
   bufcount= ret;
   Intvec_add_b30(in_count,CdbplivS,bufcount,0);

   pacifier_count++;
   if(pacifier_count>=16) {   
     r= Intvec_to_double(in_count,CdbplivS,0);
     Sfile_scale(r,size_text,7,1e5,1);
     if(write_mode) {
       sprintf(pacifier_msg,"  %s read",size_text);
       r= Intvec_to_double(out_count,CdbplivS,0);
       Sfile_scale(r,size_text,7,1e5,1);
       sprintf(pacifier_msg+strlen(pacifier_msg),
                            " |    %s written",size_text);
     } else {
       sprintf(pacifier_msg,"  %s skipped",size_text);
     }
     Print_pacifier(0,pacifier_msg,4|16|32);
     pacifier_count= 0;
   }
   bufpos= 0;
   while(bufpos<bufcount) {
     hflag= (scanstate==NULL);
     hflag|= 2|16;
     old_bufpos= bufpos;
     skip_remainder= 0;
     ret= Checkrecord_scan_for_tags(&check_rec,&scanstate,posvec,ctx,&crc32,
                                    buffer,bufcount,&bufpos,hflag);
     if(ret<0)
       goto ex;
     if(ret>0) {
       /* found valid multiply mark */
       count++;
       if(count==1) {
         Intvec_copy(scanstate->posvec,CdbplivS,copy_size,0);
       }
       if(count<=skip) {
         Print_pacifier(0,"",1|64);
         fprintf(stderr,"-extract_multiplied : skipped copy %d of %d\n",
                        count,skip);
       }
       if(count==skip) {
         fprintf(stderr,"-extract_multiplied : beginning to extract copy %d\n",
                        count+1);
         write_mode= 1;
         skip_remainder= bufpos-old_bufpos;
       } else if(count>skip) {
         if(bufpos>0)
           buffer[bufpos-1]= 'X'; /* make multiply mark invalid */
         write_mode= 2;
         Print_pacifier(0,"",1|64);
         fprintf(stderr,"-extract_multiplied : encountered end of copy %d\n",
                        count);
       }
     }
     if(write_mode>0) {
       ret= Write_broker_or_file(fpo,buffer+old_bufpos+skip_remainder,
                                     bufpos-old_bufpos-skip_remainder,0);
       if(ret<=0)
         goto ex;
       Intvec_add_b30(out_count,CdbplivS,bufpos-old_bufpos-skip_remainder,0);
     }
     if(write_mode==2) 
       {ret= 1; goto ex;}
     Intvec_add_b30(posvec,CdbplivS,bufpos-old_bufpos,0);
   }
 }  
ex:;
 if(fpi!=0)
   fclose(fpi);
 Sregex_md5(&ctx,NULL,0,NULL,(1<<15));
 Checkrecord_destroy(&check_rec,0);
 Checkrecord_tag_destroy(&scanstate,0);
 Print_pacifier(0,"",1|64);
 r= Intvec_to_double(in_count,CdbplivS,0);
 Sfile_scale(r,size_text,7,1e5,1);
 sprintf(pacifier_msg,"  %s read",size_text);
 r= Intvec_to_double(out_count,CdbplivS,0);
 Sfile_scale(r,size_text,7,1e5,1);
 sprintf(pacifier_msg+strlen(pacifier_msg)," |    %s written",size_text);
 fprintf(stderr,"-extract_multiplied : %s\n",pacifier_msg);
 return(ret);
}


#ifdef Cd_backup_planer_with_regeX

int Backup_plan__bourne_to_reg(bourne_expr,reg_expr,flag)
char bourne_expr[],reg_expr[];
int flag;
/* reg_expr should be twice as large as bourne_expr ( + 2 to be exact) */
/* return: 2= bourne_expr is surely a constant */
{
 char *wpt,*lpt;
 int backslash= 0,is_constant= 1,in_square_brackets= 0;
 int first_in_square_brackets=0;

 wpt= reg_expr;
 lpt= bourne_expr;

 *(wpt++)= '^';

 while(*lpt!=0){
  if(first_in_square_brackets>0)
    first_in_square_brackets--;
  if(!backslash){
    switch(*lpt){
    case '?':
      *(wpt++)= '.';
      is_constant= 0;
    break;case '*':
      *(wpt++)= '.';
      *(wpt++)= '*';
      is_constant= 0;
    break;case '.':
      *(wpt++)= '\\';
      *(wpt++)= '.';
    break;case '+':
      *(wpt++)= '\\';
      *(wpt++)= '+';
    break;case '[':
      *(wpt++)= *lpt;
      first_in_square_brackets= 2;
      in_square_brackets= 1;
      is_constant= 0;
    break;case ']':
      *(wpt++)= *lpt;
      in_square_brackets= 0;
    break;case '!':
      if(first_in_square_brackets)
        *(wpt++)= '^';
      else if(in_square_brackets)
        *(wpt++)= '!';
      else {
        *(wpt++)= '\\';
        *(wpt++)= '!';
      }
    break;case '^':
      if(in_square_brackets)
        *(wpt++)= '^';
      else
        *(wpt++)= '\\';
        *(wpt++)= '^';
    break;case '$':
      *(wpt++)= '\\';
      *(wpt++)= '$';
    break;case '\\':
      backslash= 1;
      *(wpt++)= '\\';
      is_constant= 0;
    break;default:
      *(wpt++)= *lpt;
    }
  } else {
    backslash= 0;
    *(wpt++)= *lpt;
  }
  lpt++;
 }
 *(wpt++)= '$';
 *wpt= 0;
 return(1+(is_constant>0));
}


int Backup_plan_add_not_pattern(plan, expr, flag)
struct Backup_plaN *plan;
char *expr;
int flag; /* bit0= convert from shell pattern */
{
 int i;
 regex_t **new_re_vec, *new_re;
 char **new_expr_vec, *new_expr, *expr_pt, *expr_conv_buf= NULL;

 expr_pt= expr;
 if(flag&1) {
   expr_conv_buf= malloc(strlen(expr)*2+2+1);
   if(expr_conv_buf==NULL)
     goto no_memory;
   expr_pt= expr_conv_buf;
   Backup_plan__bourne_to_reg(expr, expr_pt, 0);
 }
 new_expr_vec= TSOB_FELD(char *, (plan->not_pattern_count+1));
 new_expr= strdup(expr_pt);
 new_re_vec= TSOB_FELD(regex_t *, (plan->not_pattern_count+1));
 new_re= TSOB_FELD(regex_t, 1);
 if(new_expr_vec==NULL || new_expr==NULL || new_re_vec==NULL || new_re==NULL) {
no_memory:;
   fprintf(stderr, "Cannot allocate memory for new -not_pattern\n");
   return(-1);
 }
 if(regcomp(new_re,expr_pt,0)!=0) {
   fprintf(stderr, "Cannot compile regular expression: %s\n", expr_pt);
   free((char *) new_expr_vec);
   free((char *) new_re_vec);
   free((char *) new_expr);
   free((char *) new_re);
   if(expr_conv_buf!=NULL)
     free(expr_conv_buf);
   return(0);
 }
 for(i= 0; i<plan->not_pattern_count; i++) {
   new_expr_vec[i]= plan->not_pattern_expr[i];
   new_re_vec[i]= plan->not_pattern_re[i];
 }
 new_expr_vec[i]= new_expr;
 new_re_vec[i]= new_re;
 plan->not_pattern_count++;
 if(plan->not_pattern_expr!=NULL)
   free((char *) plan->not_pattern_expr);
 if(plan->not_pattern_re!=NULL)
   free((char *) plan->not_pattern_re);
 plan->not_pattern_expr= new_expr_vec;
 plan->not_pattern_re= new_re_vec;
 if(expr_conv_buf!=NULL)
   free(expr_conv_buf);
 return(1);
}


int Backup_plan_match_not_pattern(plan, adr, flag)
struct Backup_plaN *plan;
char *adr;
int flag;
{
 int i, ret;
 char *name,*spt;

 if(plan->not_pattern_count<=0)
   return(0);
 name= strrchr(adr,'/');
 if(name==NULL)
   name= adr;
 else
   name++;
 for(i= 0; i<plan->not_pattern_count; i++) {
   ret= regexec(plan->not_pattern_re[i], name, 1, plan->not_pattern_match, 0);
   if(ret==0)
     return(1);
 }
 return(0);
}


#endif /* Cd_backup_planer_with_regeX */


int Backup_plan_exclude(plan,adr,flag)
struct Backup_plaN *plan;
char *adr;
int flag;
{
 int ret;

 /* >>> make sure not too many items are in plan->auslassen */;

 ret= Fileliste_append(plan->auslassen,adr,1|8);
 if(ret<=0)
   return(ret);
 return(1);
}


int Backup_plan_setup(plan,argc,argv,flag)
struct Backup_plaN *plan;
int argc;
char **argv;
int flag;
/*
 bit0= do not finalize , other settings will follow
 bit1= with plan->configuration==2 allow configuration changing commands
 bit2= this is a level configuration setup
 bit8= this is the call that interprets the real program arguments
*/
{
 int i,ret,not_modus= 0,j,linecount= 0,hflag,zero= 0;
 double wert,start_byte,bytes,count;
 char line[4096],*cpt;

 if(flag&4)
   flag|= 2;

 for(i=1;i<argc;i++){
   if(strcmp(argv[i],"")==0){
     goto ignore_silently;
   }else if(strcmp(argv[i],"--")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     goto eintragen;
   }else if(strcmp(argv[i],"-absolute_filelist")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],2);
     if(ret<=0)
       exit(1);
     exit(0);

   } else if(strcmp(argv[i],"-agetest_debug")==0) {
     fprintf(stderr,"agetest_debug: enabled\n");
     Cdbackup_planer_agetest_debuG= 1;
   } else if(strcmp(argv[i],"-accept_all_filetypes")==0) {
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->result_filetypes= 1;
   } else if(strcmp(argv[i],"-alldev")==0) {
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_xdev= 0;
   }else if(strcmp(argv[i],"-backup_cmd")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->cmd,argv[i]);
     ret= Backup_plan_substitute_bdir(plan,plan->cmd,plan->cmd,"scripts/",
                                      (!(flag&4))|4);
     if(ret<=0)
       return(1);
   }else if(strcmp(argv[i],"-bd")==0){
     plan->default_max_size= 21.5*1024.0*1024.0*1024.0;
   }else if(strcmp(argv[i],"-bdir")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && (flag&4)) /* not from specific_arguments */
       goto ignore_silently;
     strcpy(plan->bdir,argv[i]);
     for(cpt= plan->bdir+strlen(plan->bdir);cpt>=plan->bdir;cpt--) {
                                                 /* remove trailing slashes */
       if(*cpt!='/')
     break;
       *cpt= 0;
     }
   }else if(strcmp(argv[i],"-bd_dl")==0 || strcmp(argv[i],"-bd-dl")==0){
     plan->default_max_size= 43.0*1024.0*1024.0*1024.0;
   }else if(strcmp(argv[i],"-blocklist_fileadr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"-")==0 || argv[i][0]==0)
       plan->blocklist_fileadr[0]= 0;
     else
       strcpy(plan->blocklist_fileadr,argv[i]);
   }else if(strcmp(argv[i],"-changetest_options")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i+= 1;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_agetest&= ~(32|64|512|1024|2048|16384|32768|(7<<16));
     plan->precheck_agetest|= 256;
     sprintf(line,":%s:",argv[i]);
     if(strstr(line,":type:")!=NULL)
       plan->precheck_agetest|= 512;
     if(strstr(line,":size:")!=NULL)
       plan->precheck_agetest|= 1024;
     if(strstr(line,":crc:")!=NULL ||
        strstr(line,":md5:")!=NULL ||
        strstr(line,":chk:")!=NULL)
       plan->precheck_agetest|= 2048;
     if(strstr(line,":inode:")!=NULL)
       plan->precheck_agetest|= (1<<16);
     if(strstr(line,":ignore_device:")!=NULL)
       plan->precheck_agetest|= (1<<17);
     if(strstr(line,":mistrust:")!=NULL)
       plan->precheck_agetest|= 32;
     if(strstr(line,":dir_trust:")!=NULL)
       plan->precheck_agetest|= (1<<18);
     if(strstr(line,":assume_old:")!=NULL)
       plan->precheck_agetest|= 64;
     if(strstr(line,":ignore_ctime:")!=NULL)
       plan->precheck_agetest|= 16384;
     if(strstr(line,":no_split_economy:")!=NULL)
       plan->precheck_agetest|= 32768;
     if(strstr(line,":off:")!=NULL)
       plan->precheck_agetest&= ~256;
     if(strstr(line,":adr:")!=NULL)
       plan->precheck_agetest|= 256;
   }else if(strcmp(argv[i],"-collision_test")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strcmp(argv[i],"-off")==0 || strcmp(argv[i],"off")==0){
       plan->collision_test= 0;
       plan->collision_avoid= 0;
     }else if(strcmp(argv[i],"-on")==0 || strcmp(argv[i],"on")==0){
       plan->collision_test= 1;
       plan->collision_avoid= 0;
     }else if(strcmp(argv[i],"-redirect")==0 || strcmp(argv[i],"redirect")==0){
       plan->collision_test= 1;
       plan->collision_avoid= 1;
     }else if(strcmp(argv[i],"-abort")==0 || strcmp(argv[i],"abort")==0){
       plan->collision_test= 1;
       plan->collision_avoid= 2;
     } else {
       fprintf(stderr,"%s : %s unknown sub option : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
   }else if(strcmp(argv[i],"-compare_checksum")==0){
     if(i+4>=argc)
       goto not_enough_arguments;
     if(plan->configuration==2 && !(flag&2)) {
       i+= 4;
       goto ignore_silently;
     }
     ret= Backup_plan_filter_checksum(plan,argv[i],argv[i+1],argv[i+2],
                                           argv[i+3],argv[i+4],1);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-complete_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->append_complete_target,
                       plan->complete_list_adr,0);

   }else if(strcmp(argv[i],"-composition_memory")==0){
     if(i+3>=argc)
       goto not_enough_arguments;
     i+= 3;
     sscanf(argv[i-2],"%d",&Cdbp_compose_keep_sublevels);
     sscanf(argv[i-1],"%d",&Cdbp_compose_sublevel_maxsize);
     if(argv[i][0]=='-') {
       Cdbp_compose_do_not_reuse= 1;
       Cdbp_compose_sublevel_maxmem= Scanf_io_size(argv[i]+1,1);
     } else {
       Cdbp_compose_do_not_reuse= 0;
       Cdbp_compose_sublevel_maxmem= Scanf_io_size(argv[i],1);
     }

   }else if(strcmp(argv[i],"-composition_mode")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"bestfit")==0 || strcmp(argv[i],"-bestfit")==0 ||
        strcmp(argv[i],"default")==0 || strcmp(argv[i],"-default")==0 ||
        strcmp(argv[i],"0")==0)
       plan->composition_mode= 0;
     else if(strcmp(argv[i],"target")==0 || strcmp(argv[i],"-target")==0 ||
             strcmp(argv[i],"alphabetical")==0 || 
             strcmp(argv[i],"-alphabetical")==0 ||
             strcmp(argv[i],"1")==0)
       plan->composition_mode= 1;
     else if(strcmp(argv[i],"source")==0 || strcmp(argv[i],"-source")==0 ||
             strcmp(argv[i],"2")==0)
       plan->composition_mode= 2;
     else if(strcmp(argv[i],"time")==0 || strcmp(argv[i],"-time")==0 ||
             strcmp(argv[i],"chronological")==0 || 
             strcmp(argv[i],"-chronological")==0 ||
             strcmp(argv[i],"3")==0)
       plan->composition_mode= 3;
     else if(strcmp(argv[i],"crawler")==0 || strcmp(argv[i],"-crawler")==0 ||
             strcmp(argv[i],"fast")==0 || strcmp(argv[i],"-fast")==0 ||
             strcmp(argv[i],"32770")==0)
       plan->composition_mode= 2|(1<<15);
     else {
       fprintf(stderr,"%s : %s unknown mode : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
       
   }else if(strcmp(argv[i],"-content_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->append_content_target,
                       plan->content_list_adr,0);
   }else if(strcmp(argv[i],"-conf_dir")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->conf_dir,argv[i]);
   }else if(strcmp(argv[i],"-cut_file")==0){
     if(i+4>=argc)
       goto not_enough_arguments;
     i+= 4;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     start_byte= Scanf_io_size(argv[i-2],1);
     bytes= Scanf_io_size(argv[i-1],1);
     if(start_byte<0.0 || bytes<0.0) {
        fprintf(stderr,
               "%s : not a usable start_byte or byte_count with -cut_file :\n",
               argv[0]);
        fprintf(stderr,"%s %s\n",argv[i-2],argv[i-1]);
        exit(0);
     }
     ret= Backup_plan_cut_file(plan,argv[i-3],start_byte,bytes,argv[i],
                               &count,0);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-debug_filelist")==0 ||
            strcmp(argv[i],"-debug_targetlist")==0) {
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     hflag= 3;
     if(strcmp(argv[i-2],"-debug_targetlist")==0)
       hflag|= (1<<14);
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],hflag);
     if(ret<=0)
       exit(1);
     exit(0);

   } else if(strcmp(argv[i],"-debug_stderr_log")==0){
     int c,was_cr= 0,blank_count= 0;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     while(1) {
       c= getchar();
       if(c==EOF)
     break;
       if(c==13) {
         if(!was_cr)
           putchar(10);
         was_cr= 1;
         blank_count= 0;
       } else if(c==' ' && was_cr){
         blank_count++; 
       } else {
         for(i=0;i<blank_count;i++)
           putchar(' ');
         blank_count= 0;
         if(c!=10 || !was_cr)
           putchar(c);
         was_cr= 0;
       }
     }
     exit(0);

   }else if(strcmp(argv[i],"-delete_filelist")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],1);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-delete_targetlist")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],1|(1<<14));
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-dir_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->dir_list_target,
                       plan->dir_list_adr,0);
   }else if(strcmp(argv[i],"-dir_script_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->dir_script_target,
                       plan->dir_script_adr,0);
   }else if(strcmp(argv[i],"-dir_size_cmd")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;

     for(j=strlen(argv[i]);j>=0;j--)
       if(argv[i][j]=='/')
     break;
     j++;
     if(strcmp(argv[i]+j,"cd_dir_isosize")==0 && !(flag&4)) {
       fprintf(stderr,"\ncd_backup_planer %s : WARNING !\n",PROG_VERSION);
       fprintf(stderr,"Usage of  cd_dir_isosize  is deprecated.\n");
       fprintf(stderr,
               "Are you using outdated scdbackup command scripts ?\n");
       fprintf(stderr,
               "Verwenden Sie veraltete scdbackup Kommandoskripte ?\n");
       fprintf(stderr,"WARNING !\n\n");
       fprintf(stderr,"(will continue in 15 seconds)\n");
       sleep(15);
     }

     for(j=0;j<strlen(argv[i]);j++) if(!isspace(argv[i][j])) break;
     if(j<strlen(argv[i])){
       Fileliste_default_dir_size_cmd(argv[i],0);
       Fileliste_set_dir_size_cmd(plan->startliste,argv[i],0);
       Fileliste_set_dir_size_cmd(plan->auslassen,argv[i],0);
       Fileliste_set_dir_size_cmd(plan->uebergrosse,argv[i],0);
     } else {
       fprintf(stderr,"%s : -dir_size_cmd '%s' = no usable command\n",
                      argv[0],argv[i]); 
       return(1);
     }
   }else if(strcmp(argv[i],"-disable_safe_leveling")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->complete_list_adr[0]= 0;
     plan->dir_list_adr[0]= 0;
     plan->dir_script_adr[0]= 0;
     plan->content_list_adr[0]= 0;
     plan->old_complete_list_adr[0]= 0;
     plan->vanish_list_adr[0]= 0;
     plan->vanish_script_adr[0]= 0;
     plan->old_vanish_list_adr[0]= 0;
   }else if(strcmp(argv[i],"-dvd")==0){
     plan->default_max_size= 4250.0*1024.0*1024.0;
   }else if(strcmp(argv[i],"-dvd_dl")==0 || strcmp(argv[i],"-dvd-dl")==0){
     plan->default_max_size= 7750.0*1024.0*1024.0;
   }else if(strcmp(argv[i],"-dvd_ram")==0 || strcmp(argv[i],"-dvd-ram")==0){
     plan->default_max_size= 4150.0*1024.0*1024.0;
   }else if(strcmp(argv[i],"-exclude_size")==0){
     int skip_char= 0;
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;

/* <<< 0497.2 Companion command  -split_large_files  is incr-variable. 
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
*/

     if(argv[i][0]=='-')
       skip_char= 1;
     wert= Scanf_io_size(argv[i]+skip_char,1);
     if(wert>=1.0) {
       plan->exclude_size= wert;
       plan->precheck_sizetest|= 1|((!!skip_char)<<1);
     } else
       plan->precheck_sizetest&= ~1;
   }else if(strcmp(argv[i],"-extract_multiplied")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     if(plan->configuration==2 && !(flag&2)) {
       i+= 2;
       goto ignore_silently;
     }
     wert= -1;
     sscanf(argv[i+2],"%lf",&wert);
     if(wert<1.0 || wert>1e9) {
       fprintf(stderr,
               "-extract_multipled : Unusable skip value %s\n",argv[i+2]);
       exit(1);
     }
     ret= Backup_plan_extract_multiplied(plan,argv[i+1],(int) wert,stdout,0);
     Stdout_pipe_broker("",0,1<<15);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-extract_tag_tail")==0 ||
            strcmp(argv[i],"-extract_tag_tail_dislocated")==0 ){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     hflag= (strcmp(argv[i-1],"-extract_tag_tail_dislocated")==0)<<14;
     ret= Backup_plan_extract_tag_tail(plan,argv[i],hflag);
     Stdout_pipe_broker("",0,1<<15);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-facl_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->facl_list_target,
                       plan->facl_list_adr,0);
   }else if(strcmp(argv[i],"-faclsrc_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->faclsrc_list_target,
                       plan->faclsrc_list_adr,0);
   }else if(strcmp(argv[i],"-fattr_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->fattr_list_target,
                       plan->fattr_list_adr,0);
   }else if(strcmp(argv[i],"-fattrsrc_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Split_fileaddress(argv[i],plan->fattrsrc_list_target,
                       plan->fattrsrc_list_adr,0);
   }else if(strcmp(argv[i],"-filecheck")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strcmp(argv[i],"-off")==0 || strcmp(argv[i],"off")==0) {
filecheck_off:;
       plan->precheck= 0;
       plan->precheck_expl= 0;
       plan->precheck_unread= -1;
     } else if(strcmp(argv[i],"-on")==0 || strcmp(argv[i],"on")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 0;
       plan->precheck_unread= 0;
     } else if(strcmp(argv[i],"-only")==0 || strcmp(argv[i],"only")==0) {
       plan->precheck= 2;
       plan->precheck_expl= 0;
       plan->precheck_unread= 0;
     } else if(strcmp(argv[i],"-append_hexcode")==0 || 
               strcmp(argv[i],"append_hexcode")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 1;
       plan->precheck_expl_replace= 1;
     } else if(strcmp(argv[i],"-append")==0 || strcmp(argv[i],"append")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 1;
       plan->precheck_expl_replace= 0;
     } else if(strcmp(argv[i],"-exclude")==0 || strcmp(argv[i],"exclude")==0) {
       /* From version 0.6 up to 0.8.5 this was not interpreted correctly.
          Thus the old mode -exclude is now the same as -off.
          Use -exclude_unreadable to get the check which was promised with
          -exclude but actually never really worked. */;
       goto filecheck_off;
     } else if(strcmp(argv[i],"-exclude_unreadable")==0 || 
               strcmp(argv[i],"exclude_unreadable")==0) {
       plan->precheck= 1;
       plan->precheck_unread= 1;
     } else if(strcmp(argv[i],"-abort")==0 || strcmp(argv[i],"abort")==0) {
       plan->precheck= 1;
       plan->precheck_expl= 2;
       plan->precheck_unread= 2;
     } else if(strcmp(argv[i],"-nl_abort")==0 ||
               strcmp(argv[i],"nl_abort")==0) {
       plan->newline_adr_rule= 0;
     } else if(strcmp(argv[i],"-nl_exclude")==0 ||
               strcmp(argv[i],"nl_exclude")==0) {
       plan->newline_adr_rule= 1;
     } else if(strcmp(argv[i],"-nl_avoid")==0 ||
               strcmp(argv[i],"nl_avoid")==0) {
       plan->newline_adr_rule= 2;
     } else if(strcmp(argv[i],"-nl_escape")==0 ||
               strcmp(argv[i],"nl_escape")==0) {
       plan->newline_adr_rule= 3;
     } else if(strcmp(argv[i],"-nl_allow")==0 ||
               strcmp(argv[i],"nl_allow")==0) {
       plan->newline_adr_rule= 4;
     } else if(strcmp(argv[i],"-nl_mkisofs_arg")==0 ||
               strcmp(argv[i],"nl_mkisofs_arg")==0) {

       /* >>> 0520: name as arguments pathspecs with newline in fileaddress */;

     } else {
       fprintf(stderr,"%s : %s unknown sub option : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
   }else if((strcmp(argv[i],"-filter_crc32")==0 || 
             strcmp(argv[i],"-filter_md5")==0)){
     int sret;

     if(i+2>=argc)
       goto not_enough_arguments;
     if(plan->configuration==2 && !(flag&2)) {
       i+= 2;
       goto ignore_silently;
     }
     ret= Backup_plan_filter_checksum(plan,argv[i],argv[i+1],argv[i+2],"","",
                                     (strcmp(argv[i],"-filter_md5")==0)<<12,0);
     sret= Stdout_pipe_broker("",0,1<<15);
     if(ret<=0 || sret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-filter_checksum_blocksize")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= Scanf_io_size(argv[i],1);
     if(wert>=0.0) {
       plan->block_checksum_size= wert;
       plan->checksum_pad_size= 1;
       plan->pre_checksum_size= 0;
     }
   }else if(strcmp(argv[i],"-filter_checksum_padsize")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= Scanf_io_size(argv[i],1);
     if(wert>=0.0) {
       plan->checksum_pad_size= wert;
       if(wert!=1.0)
         plan->block_checksum_size= 0;
     }
   }else if(strcmp(argv[i],"-filter_pipe")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Stdout_pipe_broker(argv[i],strlen(argv[i]),1);
   }else if(strcmp(argv[i],"-filter_through")==0){
     char buf[4096];
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     while(1) {
       ret=  fread(buf,1,sizeof(buf),stdin);
       if(ret<=0)
     break;
       Stdout_pipe_broker(buf,ret,0);
     }
     Stdout_pipe_broker("",0,1<<15);
     exit(0);
   }else if(strcmp(argv[i],"-find")==0){
     struct FilelistE *find_liste;

     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(Fileliste_single(&find_liste,argv[i],0,0.0,8|16)>0){
       if(Fileliste_set_main_dir(find_liste,plan->hauptdir,0)<=0) {
         Fileliste_destroy(&find_liste,0);
         return(-1);
       }
       if(Fileliste_read_properties(find_liste,&zero,2)>0)
         Fileliste_leanfind(find_liste,0,
                            dev_dummy,stdout,NULL,&linecount,0,0);
       Fileliste_destroy(&find_liste,0);
     }
     exit(0);
   }else if(strcmp(argv[i],"-find_path_list")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Fileliste_find_path_list(argv[i],0);
     exit(0);
   }else if(strcmp(argv[i],"-follow_symbolic_links")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strcmp(argv[i],"on")==0 || strcmp(argv[i],"-on")==0)
       Cd_backup_planer_follow_linkS= 1;
     else
       Cd_backup_planer_follow_linkS= 0;
   }else if(strcmp(argv[i],"-form_listing")==0){
     plan->output_form= 1;
     if(strlen(plan->auskunft_name->adressen[0])==0) plan->auskunft_an= 0;
   }else if(strcmp(argv[i],"-form_path_list")==0 ||
            strcmp(argv[i],"-form_path_list_mkisofs")==0 ||
            strcmp(argv[i],"-form_path_list_afio_old")==0 ||
            strcmp(argv[i],"-form_path_list_afio")==0 ||
            strcmp(argv[i],"-form_quoted_list_xorriso")==0
     ){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(strcmp(argv[i-2],"-form_path_list_afio_old")==0)
       plan->output_form= 6;
     else if(strcmp(argv[i-2],"-form_path_list_afio")==0)
       plan->output_form= 7;
     else if(strcmp(argv[i-2],"-form_quoted_list_xorriso")==0) {
       plan->output_form= 8;
       plan->newline_adr_rule= 4;
       Fileliste_target_modern= 1;
     } else {
       plan->output_form= 5;
       Fileliste_target_modern= 1;
     }
     if(strchr(argv[i-1],'/')!=NULL || plan->tmp_dir[0]==0 ||
        ((plan->output_form==5 || plan->output_form==8)
         && strcmp(argv[i-1], "-")==0))
       strcpy(plan->path_list,argv[i-1]);
     else
       sprintf(plan->path_list,"%s/%s",plan->tmp_dir,argv[i-1]);
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=2e9) wert= 2e9;
     if(wert>1) 
       plan->max_items_per_tile= wert;
   }else if(strcmp(argv[i],"-form_raw")==0){
     plan->output_form= 4;
     if(strlen(plan->auskunft_name->adressen[0])==0) plan->auskunft_an= 0;
   }else if(strcmp(argv[i],"-form_script")==0){
     plan->output_form= 0;
     Fileliste_target_modern= 0;
   }else if(strcmp(argv[i],"-form_script_here")==0){
     plan->output_form= 3;
   }else if(strcmp(argv[i],"-form_script_raw")==0){
     plan->output_form= 2;
   }else if(strcmp(argv[i],"-help")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(argc!=2) {
       fprintf(stderr,"%s : option -help ignored because not alone.\n",argv[0]);
     } else {
       Backup_plan_help(argv[0],0);
       return(-1);
     }
   }else if(strcmp(argv[i],"-hilfe")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(argc!=2) {
       fprintf(stderr,
               "%s : option -hilfe wurde ignoriert wegen weiterer Optionen.\n",
               argv[0]);
     } else {
       Backup_plan_hilfe(argv[0],0);
       return(-1);
     }
   }else if(strcmp(argv[i],"-implicite_ignored")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strlen(argv[i])>255){
       fprintf(stderr,"%s : -implicite_ignored '%s' = list too long \n",
                      argv[0],argv[i]); 
       return(1);
     }
     Fileliste_default_impl_igno(argv[i],0);
     Fileliste_set_impl_igno(plan->startliste,argv[i],0);
     Fileliste_set_impl_igno(plan->auslassen,argv[i],0);
     Fileliste_set_impl_igno(plan->uebergrosse,argv[i],0);
   }else if(strcmp(argv[i],"-info_exclude")==0){
     plan->auskunft_include= 0;
   }else if(strcmp(argv[i],"-info_script")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     sscanf(argv[i],"%s",line);
     if(strcmp(line,"-off")==0 || strcmp(line,"off")==0) {
       Fileliste_destroy(&plan->auskunft_name,0);
       plan->auskunft_an= 0;
       plan->auskunft_include= 0;
       if(Fileliste_single(&plan->auskunft_name,"",0,0.0,2|16)<=0)
         return(1);

     } else if(strcmp(line,"-inline")==0 || strcmp(line,"inline")==0) {

       /* <<< meanwhile this has the same effect as off */

       Fileliste_destroy(&plan->auskunft_name,0);
       plan->auskunft_an= 0;
       plan->auskunft_include= 0;
       if(Fileliste_single(&plan->auskunft_name,"",0,0.0,2|16)<=0)
         return(1);
     } else if(strcmp(line,"-short")==0) {
       plan->info_short= 1;
     } else if(strcmp(line,"-full")==0) {
       plan->info_short= 0;
     } else if(strcmp(line,"-short_on_level")==0) {
       plan->info_short_on_level= 1;
     } else {
       plan->auskunft_an= 1;
       plan->auskunft_include= 1;
       Fileliste_destroy(&plan->auskunft_name,0);
       if(Fileliste_single(&plan->auskunft_name,line,1,plan->auskunft_groesse,
                           2|16)<=0)
         return(1);
     }
   }else if(strcmp(argv[i],"-level")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     if(strcmp(argv[i],"-create_configuration")==0 || 
        strcmp(argv[i],"create_configuration")==0) {
       plan->configuration= 1;
       plan->level= 0;
     } else if(strcmp(argv[i],"-update_configuration")==0 ||
               strcmp(argv[i],"update_configuration")==0) {
       plan->configuration= 3;
       plan->level= 0;
     } else if(strcmp(argv[i],"-none")==0 || strcmp(argv[i],"-off")==0 ||
               strcmp(argv[i],"none")==0 || strcmp(argv[i],"off")==0 ) {
       plan->configuration= 0;
       plan->level= 0;
     } else if(strcmp(argv[i],"-next")==0 || strcmp(argv[i],"next")==0) {
       plan->configuration= 2;
       plan->level= -1;
     } else if(strcmp(argv[i],"-last")==0 || strcmp(argv[i],"-current")==0 ||
               strcmp(argv[i],"last")==0 || strcmp(argv[i],"current")==0) {
       plan->configuration= 2;
       plan->level= -2;
     } else {
       plan->configuration= 2;
       wert= -1;
       sscanf(argv[i],"%lf",&wert);
       if(wert>=2e9) wert= 2e9;
       if(wert>=0)
         plan->level= wert; 
       else {
         fprintf(stderr,"%s : %s illegal level number : %s\n",
                        argv[0],argv[i-1],argv[i]);
         return(1);
       }
     }
   }else if(strcmp(argv[i],"-level_timestamp")==0){
     int is_usable;
     time_t timestamp;
     char *time_text,time_text_buffer[81];

     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     time_text= argv[i];
     if(flag&4) {
       /* ignore this command when coming from the level configuration */;
     } else {
       if(argv[i][0]=='/' || argv[i][0]=='.') {
         ret= Filetimestamp(argv[i],time_text_buffer,0);
         if(ret<=0) {
           fprintf(stderr,"\n");
           fprintf(stderr,
                  "Error: -level_timestamp cannot determine date of file %s\n",
                  argv[i]);
           fprintf(stderr,"errno= %d  (%s)\n",errno,strerror(errno));
           return(0);
         }
       } 
       ret= String_to_time(time_text,&is_usable,&timestamp,0);
       if(ret>0 && (is_usable&1)) {
         if(plan->start_time>=timestamp) {
           plan->start_time= timestamp;
           if(plan->configuration==2)
             fprintf(stderr,
                    "Note: timestamp for incremental backup level set to %s\n",
                    Sfile_datestr(timestamp,1|2));
         } else {
           fprintf(stderr,"\n");
           fprintf(stderr,
 "Error: -level_timestamp for incremental backup is younger than program run\n"
                  );
           fprintf(stderr,
           "       program start time %s ,  -level_timestamp '%s'\n",
                          Sfile_datestr(plan->start_time,1|2),time_text);
           fprintf(stderr,
           "       attempted time     %s\n",Sfile_datestr(timestamp,1|2));
           return(0);
         }
       } else {
         fprintf(stderr,"\n");
         fprintf(stderr,
       "Error: -level_timestamp argument is not a known time format: '%s'\n",
                   argv[i]);
         fprintf(stderr,
       "       use either like A50818.124412 or like 081812442005.12\n");
         fprintf(stderr,
       "       or use a file address beginning with '.' or '/' as reference.\n"
                );
         return(0);
       }
     }
   }else if(strcmp(argv[i],"-list_checksum")==0){
     if(i+4>=argc)
       goto not_enough_arguments;
     i+= 4;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_filter_checksum(plan,argv[i-4],argv[i-3],argv[i-2],
                                           argv[i-1],argv[i],1|2);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-list_output_format")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i+= 1;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->list_output_format= 1;
     sprintf(line,":%s:",argv[i]);
     if(strstr(line,":type:")!=NULL)
       plan->list_output_format|= 2;
     if(strstr(line,":size:")!=NULL)
       plan->list_output_format|= 4;
     if(strstr(line,":crc:")!=NULL)
       plan->list_output_format= (plan->list_output_format&~16)|8;
     if(strstr(line,":md5:")!=NULL)
       plan->list_output_format|= 8|16;
     if(strstr(line,":inode:")!=NULL)
       plan->list_output_format|= 32;
     if(strstr(line,":mtime:")!=NULL)
       plan->list_output_format|= 128;
     if(strstr(line,":ctime:")!=NULL)
       plan->list_output_format|= 256;
     if(strstr(line,":on_level_crc:")!=NULL)
       plan->list_output_format|= (1<<10);
     if(strstr(line,":on_level_md5:")!=NULL)
       plan->list_output_format|= (1<<11);
     if(strstr(line,":on_level_inode:")!=NULL)
       plan->list_output_format|= (1<<12);
   }else if(strcmp(argv[i],"-load_old_tree")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->old_complete_list_adr,argv[i-1]);
     strcpy(plan->old_vanish_list_adr,argv[i]);
     if(plan->old_complete_list_adr[0]!=0 && plan->old_vanish_list_adr[0]!=0)
       plan->precheck_agetest|= 256|512|1024|2048;
     else
       plan->precheck_agetest&= ~256;
   } else if(strcmp(argv[i],"-localdev")==0) {
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_xdev= 2;
   }else if(strcmp(argv[i],"-main_dir")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     sscanf(argv[i],"%s",plan->hauptdir);
     plan->hauptdir_auto= 0;
     if(strcmp(plan->hauptdir,"-off")==0 || strcmp(plan->hauptdir,"off")==0) {
       plan->hauptdir[0]= 0;
     } else if(strcmp(plan->hauptdir,"-auto")==0 || 
               strcmp(plan->hauptdir,"auto")==0) {
       plan->hauptdir_auto= 1;
       plan->hauptdir[0]= 0;
     }
     if(Backup_plan_list_main_dirs(plan,0)<=0)
       return(-1);
   }else if(strcmp(argv[i],"-max_items")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=2e9) wert= 2e9;
     if(wert>1) 
       plan->max_items= wert;
   }else if(strcmp(argv[i],"-max_size")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= Scanf_io_size(argv[i],1);
     if(wert>=1.0) {
       plan->default_max_size= wert;
       plan->max_size= wert;
     }
   }else if(strcmp(argv[i],"-max_size_stack")==0){
     int count;

     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     sscanf(argv[i],"%d",&count);
     wert= Scanf_io_size(argv[i-1],1);
     if(wert>=1.0) {
       Backup_plan_stacked_max_size(plan,&wert,&count,1);
     } else if(count<=0) {
       Backup_plan_stacked_max_size(plan,&wert,&count,2);
     }
   }else if(strcmp(argv[i],"-max_size_stack_save")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"-off")==0 || strcmp(argv[i],"off")==0) {
       plan->save_stack_media= 0;
     } else if(strcmp(argv[i],"-on")==0 || strcmp(argv[i],"on")==0) {
       plan->save_stack_media= 1;
     }
   }else if(strcmp(argv[i],"-max_tree_depth")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=16)
       Cd_backup_planer_tree_max_deptH= wert;
   }else if(strcmp(argv[i],"-max_young_append")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=2e9) wert= 2e9;
     if(wert>=0)
       plan->precheck_max_exempt= wert; 
   }else if(strcmp(argv[i],"-merge_trees")==0){
     int hflag= 0,tree_flag= (1<<14);
     char tree_prefix[4096],*cpt,*dpt;
     if(i+4>=argc)
       goto not_enough_arguments;
     i+= 4;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;

     hflag= 16|32|(1<<8)|(1<<12);
     strcpy(tree_prefix,argv[i-3]);
     strcpy(line,":");
     strcat(line,argv[i]);
     strcat(line,":");
     if(strstr(line,":no_overwrite:")!=NULL)
       hflag= (hflag&~(15<<8))|(0<<8);
     if(strstr(line,":overwrite:")!=NULL)
       hflag= (hflag&~(15<<8))|(1<<8);
     if(strstr(line,":overwrite_older:")!=NULL)
       hflag= (hflag&~(15<<8))|(2<<8);
     if(strstr(line,":move:")!=NULL)
       hflag&= ~64;
     if(strstr(line,":copy:")!=NULL)
       hflag|= 64;
     if(strstr(line,":owner_plus_w:")!=NULL)
       hflag|= (1<<13);
     if(strstr(line,":chmod_original:")!=NULL)
       hflag&= ~(1<<13);
     cpt= strstr(line,":tree_prefix=");
     if(cpt!=NULL) {
       cpt+= strlen(":tree_prefix=");
       dpt= strchr(cpt,':');
       strncpy(tree_prefix,cpt,dpt-cpt);
       tree_prefix[dpt-cpt]= 0;
     }
     if(strstr(line,":tree_source:")!=NULL)
       tree_flag&= ~(1<<14);
     if(strstr(line,":tree_target:")!=NULL)
       tree_flag|= (1<<14);
     ret= Backup_plan_load_old_tree(plan,plan->old_complete_list_adr,
                                    plan->old_vanish_list_adr,tree_prefix,
                                    tree_flag);
     if(ret<=0)
       exit(1);

     ret= Fileliste_merge_trees(NULL,0,(dev_t) 0,&linecount,
                          argv[i-3],argv[i-2],argv[i-1],plan->old_tree,hflag);

     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-min_size_percent")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= -1;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=0.0 && wert<=100.0) {
        plan->min_size_fak= wert/100.0;
     }
   }else if(strcmp(argv[i],"-multiply_volume")==0) {
     if(i+2>=argc)
       goto not_enough_arguments;
     if(plan->configuration==2 && !(flag&2)) {
       i+= 2;
       goto ignore_silently;
     }
     wert= -1;
     sscanf(argv[i+2],"%lf",&wert);
     if(wert<1.0 || wert>1e9) {
       fprintf(stderr,
               "-multiply_volume : Unusable multiplication factor %s\n",
               argv[i+2]);
       exit(1);
     }
     ret= Backup_plan_multiply_volume(plan,argv[i+1],((int) wert),stdout,0);
     Stdout_pipe_broker("",0,1<<15);
     if(ret<=0)
       exit(1);
     exit(0);

   }else if(strcmp(argv[i],"-no")==0 || strcmp(argv[i],"-not")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     not_modus= 1;
   }else if(strcmp(argv[i],"-not_inode")==0){
     not_modus= 2;
     Cdbackup_planer_sperrliste_inodE= 1;

   }else if(strcmp(argv[i],"-not_pattern")==0 ||
            strcmp(argv[i],"-not_regex")==0){

#ifdef Cd_backup_planer_with_regeX

     if(i+1>=argc)
       goto not_enough_arguments;
     hflag= (strcmp(argv[i],"-not_pattern")==0);
     i++;
     ret= Backup_plan_add_not_pattern(plan, argv[i], hflag);
     if(ret<=0)
       return(1);

#else /* Cd_backup_planer_with_regeX */

    fprintf(stderr,
            "%s : Program not compiled for regular expressions and %s\n",
            argv[0], argv[i]);
    return(1);

#endif /* ! Cd_backup_planer_with_regeX */

   }else if(strcmp(argv[i],"-pacifier")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"-off")==0 || strcmp(argv[i],"off")==0) {
       Cd_backup_planer_pacifier_intervaL= -1;
       Cd_backup_planer_pacifier_show_timE= 0;
       Cd_backup_planer_compose_pacifier= 0;
     } else if(strcmp(argv[i],"-on")==0 || strcmp(argv[i],"on")==0) {
       Cd_backup_planer_pacifier_intervaL= 1;
       Cd_backup_planer_pacifier_show_timE= 0;
     } else if(argv[i][0]>='0' && argv[i][0]<='9') {
       sscanf(argv[i],"%d",&Cd_backup_planer_pacifier_intervaL);
     } else if(strcmp(argv[i],"-timestamp")==0 || 
               strcmp(argv[i],"timestamp")==0) {
       if(Cd_backup_planer_pacifier_intervaL<=0)
         Cd_backup_planer_pacifier_intervaL= 1;
       Cd_backup_planer_pacifier_show_timE= 1;
     } else if(strcmp(argv[i],"-composition_memory")==0 || 
               strcmp(argv[i],"composition_memory")==0) {
       if(Cd_backup_planer_pacifier_intervaL<=0)
         Cd_backup_planer_pacifier_intervaL= 1;
       Cd_backup_planer_compose_pacifier= 1;
     } else if(strcmp(argv[i],"-log_mode")==0 || 
               strcmp(argv[i],"log_mode")==0) {
       if(Cd_backup_planer_pacifier_intervaL<=0)
         Cd_backup_planer_pacifier_intervaL= 1;
       Cd_backup_planer_pacifier_new_linE= 1;
     }
   }else if(strcmp(argv[i],"-permute_blocks")==0){
     double blocksize,max_blocks,headsize;
     int mode= -1,hflag,key;

     if(i+5>=argc)
       goto not_enough_arguments;
     if(plan->configuration==2 && !(flag&2)) {
       i+= 5;
       goto ignore_silently;
     }
     hflag= 0;
     key= argv[i+1][0];
     if(strcmp(argv[i+1],"-restore")==0 || strcmp(argv[i+1],"restore")==0) {
       mode= 0;
       hflag|= 1;
     } else if(argv[i+1][0]=='-') {
       hflag|= 3;
       key= argv[i+1][1];
     }
     if(mode<0) {
       mode= key-'0';
       if(mode<0 || mode>7) {
         fprintf(stderr,"-permute_blocks : unusable key '%s'\n",argv[i+1]);
         exit(1);
       }
     }
     blocksize= Scanf_io_size(argv[i+3],1);
     if(blocksize<=0 && (hflag&3)!=1) {
       fprintf(stderr,"-permute_blocks : bad blocksize %s\n",argv[i+2]);
       exit(1);
     }
     if(argv[i+4][0]=='-') {
       max_blocks= 0.0;
     } else {
       max_blocks= Scanf_io_size(argv[i+4],1);
     }
     if(max_blocks<=0) {
       max_blocks= 1e24;
     } else {
       hflag|= 4;
     }
     
     headsize= Scanf_io_size(argv[i+5],1);
     if(headsize<=0)
       headsize= 0;
     ret= Backup_plan_permute_blocks(plan,argv[i+2],blocksize,max_blocks,mode,
                                     headsize,stdout,hflag);
     Stdout_pipe_broker("",0,1<<15);
     if(ret<=0)
       exit(1);
     exit(0);

   }else if(strcmp(argv[i],"-pre_checksum_size")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= Scanf_io_size(argv[i],1);
     if(wert >= -1.0) {
       plan->pre_checksum_size= wert;
       plan->block_checksum_size= 0.0;
     }
   }else if(strcmp(argv[i],"-prevent_doubles")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"-off")==0 || strcmp(argv[i],"off")==0){
       plan->prevent_doubles= 0;
     }else if(strcmp(argv[i],"-on")==0 || strcmp(argv[i],"on")==0){
       plan->prevent_doubles= 1;
     } else {
       fprintf(stderr,"%s : %s unknown sub option : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
   } else if(strcmp(argv[i],"-problemlist_limit")==0) {
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     sscanf(argv[i],"%lf",&wert);
     if(wert>=1 && wert<2e9)
       plan->problemlist_limit= wert;
     else
       fprintf(stderr,"--- -problemlist_limit %.lf : rejected (use 1 to 2e9)",
                      wert);
   }else if(strcmp(argv[i],"-process_blocklist")==0){
     int skip;

     if(i+5>=argc)
       goto not_enough_arguments;
     if(plan->configuration==2 && !(flag&2)) {
       i+= 5;
       goto ignore_silently;
     }

     skip= 0;
     skip= Scanf_io_size(argv[i+3],0);
     if(skip<0||skip>Intvec_b30)
       skip= 0;
     ret= Backup_plan_process_blocklist(plan,argv[i+1],argv[i+2],skip,
                                        argv[i+4],argv[i+5],0);
     Stdout_pipe_broker("",0,1<<15);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-prompt_script")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->prompt_skript,argv[i]);
     if(strcmp(plan->prompt_skript,"-")==0)
       plan->prompt_skript[0]= 0;

   }else if(strcmp(argv[i],"-result_filetypes")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     plan->result_filetypes= 0;
     if(strstr(argv[i],"data")!=NULL)
       plan->result_filetypes|= (1<<1);
     if(strstr(argv[i],"dir")!=NULL)
       plan->result_filetypes|= (1<<2);
     if(strstr(argv[i],"symlink")!=NULL)
       plan->result_filetypes|= (1<<3);
     if(strstr(argv[i],"block_device")!=NULL)
       plan->result_filetypes|= (1<<6);
     if(strstr(argv[i],"unclassified")!=NULL)
       plan->result_filetypes|= (1<<30);
     if(strstr(argv[i],"any")!=NULL)
       plan->result_filetypes|= 1;
     if(strstr(argv[i],"arg_redirect")!=NULL)
       plan->result_filetypes|= (1<<31);

   }else if(strcmp(argv[i],"-scan_for_checksum")==0) { 
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"-off")==0 || strcmp(argv[i],"off")==0) {
       plan->checksum_scan_mode&= ~1;
     } else {
       plan->checksum_scan_mode|= 1;
       if(strcmp(argv[i],"-on")==0 || strcmp(argv[i],"on")==0) {
         ;
       } else if(strcmp(argv[i],"-first")==0 || strcmp(argv[i],"first")==0) {
         plan->checksum_scan_mode|= 2;             /* end on first valid one */
       } else if(strcmp(argv[i],"-whole")==0 || strcmp(argv[i],"whole")==0) {
         plan->checksum_scan_mode&= ~2;         /* continue after valid ones */
       } else if(strcmp(argv[i],"-standalone")==0 || 
                 strcmp(argv[i],"standalone")==0) {
         plan->checksum_scan_mode|= 4;
       } else if(strcmp(argv[i],"-withlist")==0 || 
                 strcmp(argv[i],"withlist")==0) {
         plan->checksum_scan_mode&= ~4;
       } else if(strcmp(argv[i],"-print")==0 || 
                 strcmp(argv[i],"print")==0) {
         plan->checksum_scan_mode|= 8;
       } else if(strcmp(argv[i],"-noprint")==0 || 
                 strcmp(argv[i],"noprint")==0) {
         plan->checksum_scan_mode&= ~8;
       } else if(strcmp(argv[i],"-blocklist")==0 || 
                 strcmp(argv[i],"blocklist")==0) {
         plan->checksum_scan_mode|= 16;
       } else if(strcmp(argv[i],"-anytag")==0 || 
                 strcmp(argv[i],"anytag")==0) {
         plan->checksum_scan_mode&= ~16;
       } else if(strcmp(argv[i],"-pacifier")==0 || 
                 strcmp(argv[i],"pacifier")==0) {
         plan->checksum_scan_mode|= 32;
       } else if(strcmp(argv[i],"-nopacifier")==0 || 
                 strcmp(argv[i],"nopacifier")==0) {
         plan->checksum_scan_mode&= ~32;
       } else if(strncmp(argv[i],"-skip:",6)==0 || 
                 strncmp(argv[i],"skip:",5)==0) {
         plan->checksum_scan_skip=Scanf_io_size(argv[i]+5+(argv[i][0]=='-'),0);
       } else if(strncmp(argv[i],"-source:",8)==0 || 
                 strncmp(argv[i],"source:",7)==0) {
         strcpy(plan->checksum_scan_source,argv[i]+7+(argv[i][0]=='-'));
       }
     }
   }else if(strcmp(argv[i],"-scattering")==0) {
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"-off")==0 || strcmp(argv[i],"off")==0) {
       plan->scattering= 0;
     } else if(strcmp(argv[i],"-on")==0 || strcmp(argv[i],"on")==0) {
       plan->scattering= 1;
     }
   }else if((strcmp(argv[i],"-search_crc32")==0 ||
             strcmp(argv[i],"-search_md5")==0)){
     if(i+4>=argc)
       goto not_enough_arguments;
     i+= 4;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_filter_checksum(
                         plan,argv[i-4],argv[i-3],argv[i-2], argv[i-1],argv[i],
                         1|4|((strcmp(argv[i-4],"-search_md5")==0)<<12));
     Stdout_pipe_broker("",0,1<<15);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-set_shell_in_scdbackup_scripts")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i++;
     Backup_plan_set_skript_shell(argc-i-1,argv+i+1,argv[i],0);
     exit(0);
   }else if(strcmp(argv[i],"-shell_to_use")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->shell_to_use,argv[i]);
   }else if(strcmp(argv[i],"-special_file")==0){
     char source[4096],target[4096];

     if(i+3>=argc)
       goto not_enough_arguments;
     i+= 3;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     Fileliste_source_target(plan->startliste,argv[i-2],source,target,1);
     
     /* >>> interpret any options . Todo : colon separated list of keywords */
     if(strcmp(argv[i],"block_device")==0) {
       ret= Type_and_size(source,&wert,1|4);
       if(ret!=6) {
         fprintf(stderr,"%s -special_file : '%s' is not a block device\n",
                        argv[0],argv[i-2]); 
         return(1);
       }
     } else {
       fprintf(stderr,"%s -special_file : unknown type option '%s'\n",
                      argv[0],argv[i]);
     }
     /* add to startlist with bit5 in flags*/;
     ret= Fileliste_append(plan->startliste,argv[i-2],1|32);
     if(ret<=0)
        return(2);
     wert= Scanf_io_size(argv[i-1],0);
     Fileliste_set_groessen(plan->startliste,-1,wert,0);

   }else if(strcmp(argv[i],"-split_file_prefix")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->split_file_prefix,argv[i]);

   }else if(strcmp(argv[i],"-split_file_suffix")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->split_file_suffix,argv[i]);

   }else if(strcmp(argv[i],"-split_large_files")==0){
     if(i+4>=argc)
       goto not_enough_arguments;
     i+= 4;
     if(strcmp(argv[i-3],"-")==0 || strcmp(argv[i-3],"")==0)
       plan->split_large_files= 0;
     else {
       /* automated plan->split_file_size = plan->max_size/8*7 */;
       if(strcmp(argv[i-2],"-")==0) {
         wert= 0.0;
       } else {
         wert= Scanf_io_size(argv[i-2],1);
         if(wert<0.0) {
           fprintf(stderr,"Not a usable size with  -split_large_files : %s\n",
                          argv[i-2]);
           plan->split_large_files= 0;
           goto ignore_silently;
         }
       }
       plan->split_large_files= 1;
       strcpy(plan->split_file_dir,argv[i-3]);
       plan->split_file_size= wert;
       if(!(strcmp(argv[i-1],"-")==0 || strcmp(argv[i-1],"")==0)) {
         sprintf(plan->cut_file_cmd,"%s %s",argv[i-1],argv[i]);
         ret= Backup_plan_substitute_bdir(plan,
                        plan->cut_file_cmd,plan->cut_file_cmd,
                        "bin/cd_backup_planer",
                        !(flag&4));
         if(ret<=0)
           return(1);
       }
     }
   }else if(strcmp(argv[i],"-split_list_adr")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;

/* <<< 0497.2 Companion command  -split_large_files  is incr-variable. 
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
*/

     Split_fileaddress(argv[i],plan->split_list_target,
                       plan->split_list_adr,0);
   }else if(strcmp(argv[i],"-status_file")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->skript_status_file,argv[i]);
     if(strcmp(plan->skript_status_file,"-")==0)
       plan->skript_status_file[0]= 0;
   }else if(strcmp(argv[i],"-target_escape")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"on")==0 || strcmp(argv[i],"-on")==0)
       Fileliste_target_modern= 1;
     else
       Fileliste_target_modern= 0;
   }else if(strcmp(argv[i],"-target_separator")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->target_inp_trenner= argv[i][0];
     Fileliste_default_target_inp_trenner(argv[i],0);
     Fileliste_set_target_inp_trenner(plan->startliste,argv[i],0);
     Fileliste_set_target_inp_trenner(plan->auslassen,argv[i],0);

   }else if(strcmp(argv[i],"-test_command")==0){
     double value;
     /* A command of varying meaning */
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(strcmp(argv[i],"trust_fret64")==0) {
       Cd_backup_planer_trust_fret64= 1;
     } else if(strcmp(argv[i],"no_trust_fret64")==0) {
       Cd_backup_planer_trust_fret64= 0;
     } else if(strcmp(argv[i],"risk_d")==0) {
       Cd_backup_planer_risk_d_enabled= 1;
     } else if(strcmp(argv[i],"no_risk_d")==0) {
       Cd_backup_planer_risk_d_enabled= 0;
     } else if(strcmp(argv[i],"use_crawler")==0) {
       plan->use_crawler= 1;
     } else if(strcmp(argv[i],"do_not_use_crawler")==0) {
       plan->use_crawler= 0;
     } else if(strcmp(argv[i],"list_search_debug")==0) {
       Cdbackup_planer_listsearch_debuG= 1;
     } else if(strcmp(argv[i],"list_search_debug_extra")==0) {
       Cdbackup_planer_listsearch_debuG= 2;
     } else if(strcmp(argv[i],"tree_build_debug")==0) {
       Cdbackup_planer_treebuild_debuG= 1;
     } else if(strcmp(argv[i],"tree_build_debug_extra")==0) {
       Cdbackup_planer_treebuild_debuG= 2;
     } else if(strcmp(argv[i],"count_bytes")==0) {
       if(i+1>=argc)
         goto test_not_enough_arguments;
       i++;
       value= Count_file_bytes(argv[i],0);
       fprintf(stderr,"%lf\n",value);
       exit(0);
     } else if(strcmp(argv[i],"sizeof_ino_t")==0) {
       printf("sizeof(ino_t)= %d\n",sizeof(ino_t));
       exit(0);
     } else if(strcmp(argv[i],"dirseq_buffer_size")==0) {
       if(i+1>=argc)
         goto test_not_enough_arguments;
       i++;
       sscanf(argv[i],"%lf",&wert);
       if(wert>=2 && wert<1e6)
         Dirseq_buffer_sizE= wert;
       else
         fprintf(stderr,"--- -test_command dirseq_buffer_size %.lf : rejected (use 2 to 1e6)",wert);
     } else if(strcmp(argv[i],"upgrade_content_file_list")==0) {
       if(i+2>=argc)
         goto test_not_enough_arguments;
       i+= 2;
       ret= Backup_plan_upgrade_cfl(plan,argv[i-1],argv[i],0);
       if(ret<=0)
         exit(1);
       exit(0);
     } else if(strcmp(argv[i],"cull_content_file_list")==0) {
       if(i+2>=argc)
         goto test_not_enough_arguments;
       i+= 2;
       ret= Backup_plan_upgrade_cfl(plan,argv[i-1],argv[i],1);
       if(ret<=0)
         exit(1);
       exit(0);
     } else if(strcmp(argv[i],"tolerate_bad_dirpop")==0) {
       Cdbackup_planer_tolerate_bad_dirpoP= 1;
     } else if(strcmp(argv[i],"tolerate_bad_rmdir")==0) {
       Cdbackup_planer_tolerate_bad_rmdiR= 1;
     } else if(strcmp(argv[i],"abort_on_bad_rmdir")==0) {
       Cdbackup_planer_tolerate_bad_rmdiR= 0;
     } else if(0) {
test_not_enough_arguments:;
       fprintf(stderr,"--- with -test_command :\n");
       goto not_enough_arguments;
     } else {
       fprintf(stderr,"--- unknown sub option to -test_command : %s\n",
                      argv[i]);
       return(1);
     }

   }else if(strcmp(argv[i],"-tmp_dir")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->tmp_dir,argv[i]);
     if(strcmp(plan->tmp_dir,"-")==0)
       plan->tmp_dir[0]= 0;
     ret= Backup_plan_substitute_bdir(plan,plan->tmp_dir,plan->tmp_dir,"tmp",
                                      !(flag&4));
     if(ret<=0)
       return(1);
   }else if(strcmp(argv[i],"-tree_hashsize")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(Fileliste_hash_size_frozeN) {
       fprintf(stderr,
             "Program error : Fileliste_hash_size_frozeN on -tree_hashsize\n");
       fprintf(stderr,"will go on with old hashsize= %d\n",
                      Fileliste_hash_sizE);
     } else {
       sscanf(argv[i],"%lf",&wert);
       if(wert<2.0 || wert>1024*1024) {
         if(wert>=0.0) {
            Fileliste_hash_sizE= 0;
         } else {
           fprintf(stderr,"-tree_hashsize : Unusable value %s\n",argv[i]);
           fprintf(stderr,"will go on with old hashsize= %d\n",
                          Fileliste_hash_sizE);
         }
       } else
         Fileliste_hash_sizE= wert;
     }
   }else if(strcmp(argv[i],"-vanish_list_adr")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->old_complete_list_adr,argv[i-1]);
     Split_fileaddress(argv[i],plan->vanish_list_target,
                       plan->vanish_list_adr,0);
   }else if(strcmp(argv[i],"-vanish_script_adr")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     strcpy(plan->old_complete_list_adr,argv[i-1]);
     Split_fileaddress(argv[i],plan->vanish_script_target,
                       plan->vanish_script_adr,0);

   }else if(strcmp(argv[i],"-verify_filelist")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],0);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-verify_source_filter")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     strcpy(plan->verify_source_filter,argv[i]);
     ret= Backup_plan_substitute_bdir(plan,
                         plan->verify_source_filter,plan->verify_source_filter,
                         "bin/cd_backup_planer",!(flag&4));
     if(ret<=0)
       return(1);
   }else if(strcmp(argv[i],"-verify_sources")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i+= 1;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i],"",128);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-verify_targetlist")==0){
     if(i+2>=argc)
       goto not_enough_arguments;
     i+= 2;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= Backup_plan_verify_filelist(plan,argv[i-1],argv[i],1<<14);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-version")==0){
     printf("scdbackup program : cd_backup_planer %s\n",PROG_VERSION);
     if(i!=1 || argc!=2) {
       fprintf(stderr,"\ncd_backup_planer %s : WARNING !\n",PROG_VERSION);
       fprintf(stderr, "Obviously received option -version in error.\n");
       fprintf(stderr,"Offenbar irrtuemlicher Gebrauch von Option -version.\n");
     }
     exit(4);
   }else if(strcmp(argv[i],"-volume_info")==0){
     plan->volume_info= 1;
   }else if(strcmp(argv[i], "-volume_offset") == 0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     wert= Scanf_io_size(argv[i],0);
     if(wert < 1 || wert > 1.0e9) {
       fprintf(stderr,
              "Not a usable number for -volume_offset : '%s' is <1 or >1e9\n",
               argv[i]);
       exit(1);
     }
     plan->volume_number_offset= wert;
   }else if(strcmp(argv[i],"-working_directory")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     ret= chdir(argv[i]);
     if(ret==-1) {
       fprintf(stderr,"Cannot set -working_directory : %s\n",argv[i]);
       fprintf(stderr,"errno= %d , '%s'\n",errno,strerror(errno));
       return(21);
     }
   } else if(strcmp(argv[i],"-write_random")==0){
     if(i+3>=argc)
       goto not_enough_arguments;
     i+= 3;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     wert= Scanf_io_size(argv[i-1],0);
     if(wert<=0.0) {
       fprintf(stderr,"Not a usable size for -write_random : %s\n",argv[i-1]);
       exit(1);
     }
     ret= Backup_plan_write_random(plan,argv[i-2],wert,argv[i],0);
     if(ret<=0)
       exit(1);
     exit(0);
   }else if(strcmp(argv[i],"-write_to_file")==0){
     int new_i,k,write_pt_valid= 0;
     double byte_address, number_of_bytes;
     char buffer[4096],*target_adr= NULL,*source_adr= "",*write_pt;
     FILE *fp_out= NULL,*fp_in= NULL;

     write_pt= buffer;
     if(i+4>=argc)
       goto not_enough_arguments;
     new_i= i+3;
     target_adr= argv[i+1];
     byte_address= Scanf_io_size(argv[i+2],0);
     if(byte_address<0.0) {
       fprintf(stderr,"Not a usable byte address for -write_to_file : %s\n",
                      argv[i+2]);
       exit(1);
     }
     if(strcmp(argv[i+3],"-file")==0 || strcmp(argv[i+3],"file")==0 ||
        strcmp(argv[i+3],"\"file\"")==0)
       number_of_bytes= 0.0;
     else if(strcmp(argv[i+3],"-string")==0 || strcmp(argv[i+3],"string")==0 ||
        strcmp(argv[i+3],"\"string\"")==0)
       number_of_bytes= -1.0;
     else {
       number_of_bytes= Scanf_io_size(argv[i+3],1);
       if(number_of_bytes<=0.0) {
         fprintf(stderr,
                 "Not a usable number of bytes for -write_to_file : %s\n",
                 argv[i+3]);
         exit(1);
       }
     }
     if(number_of_bytes==-1.0) {
       number_of_bytes= strlen(argv[new_i+1]);
       write_pt= argv[new_i+1];
       write_pt_valid= 1;
       new_i++;
     } else if(number_of_bytes==0.0) {
       if(i+4<argc) {
         source_adr= argv[i+4];
         new_i= i+4;
       }
       if(strlen(source_adr)<=0) {
         fprintf(stderr,
                 "Not a usable source file for -write_to_file : %s\n",
                 source_adr);
         exit(1);
       }
     } else if(number_of_bytes>sizeof(buffer)) {
       fprintf(stderr,
               "Number of bytes too large for -write_to_file : %s\n",
               argv[i+3]);
       exit(1);
     } else {
       if(new_i+1+number_of_bytes>argc)
         goto not_enough_arguments;
       for(k=0;k<number_of_bytes;k++) {
         wert= Scanf_io_size(argv[new_i+1+k],1);
         if(wert<0 || wert>255) {
           fprintf(stderr,
                  "Byte %d for -write_to_file is outside range [0,255] : %s\n",
                  k,argv[new_i+1+k]);
           exit(1);
         }
         buffer[k]= wert;
       }
       write_pt_valid= 1;
       new_i+= k;
     }
     i= new_i;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;

     fp_out= Sfile_fopen(target_adr,"r+b",0);
     if(fp_out==NULL)
       exit(1);
     if(byte_address>0) {
       ret= Fseek_double(fp_out,byte_address,0);
       if(ret==-1) {
         fprintf(stderr,"Failed to address byte %.lf in file %s\n",
                        byte_address,target_adr);
         return(0);
       }
     }
     if(write_pt_valid) {
       ret= fwrite(write_pt,(size_t) number_of_bytes,1,fp_out);
       if(ret!=1) {
         perror("while writing : ");
         exit(1);
       }  
     } else {
       fp_in= Sfile_fopen(source_adr,"rb",0);
       if(fp_in==NULL)
         exit(1);
       while(1) {
         ret= fread(buffer,1,sizeof(buffer),fp_in);
         if(ret<=0) {
           if(feof(fp_in))
       break;
           perror("while reading : ");
           exit(1);
         }
         ret= fwrite(buffer,(size_t) ret,1,fp_out);
         if(ret!=1) {
           perror("while writing : ");
           exit(1);
         }  
       }
     }
     if(fp_in!=NULL)
       Sfile_fclose(fp_in);
     if(fp_out!=NULL)
       Sfile_fclose(fp_out);
     exit(0);
     
   }else if(strcmp(argv[i],"-xdev")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_xdev= 1;
   }else if(strcmp(argv[i],"-yes")==0){
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     not_modus= 0;
     plan->auslassen_nicht_startliste= 1;
   }else if(strcmp(argv[i],"-younger_than")==0){
     if(i+1>=argc)
       goto not_enough_arguments;
     i++;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;
     plan->precheck_agetest|= 1;
     if(String_to_time(argv[i],
                       &(plan->precheck_agetest),&(plan->precheck_age),0)<=0){
       fprintf(stderr,"%s : %s unknown time definition : %s\n",
                      argv[0],argv[i-1],argv[i]);
       return(1);
     }
     if(plan->precheck_agetest&1)
       if(plan->precheck_max_exempt<0)
         plan->precheck_max_exempt= 2000000000; /* unlimited */
   }else if(argv[i][0]=='-'){
     fprintf(stderr,"%s : unknown or incomplete option %s\n",argv[0],argv[i]);
     return(1);
   } else {
eintragen:;
     if(plan->configuration==2 && !(flag&2))
       goto ignore_silently;

     /* ??? remove trailing / ??? */
     if(not_modus)
       ret= Fileliste_append(plan->auslassen,argv[i],1|((not_modus==2)<<8));
     else
       ret= Fileliste_append(plan->startliste,argv[i],1);
     if(ret<=0) return(2);
   }
ignore_silently:;
 }

 if(!(flag&1)){
   if(plan->hauptdir_auto){
     /* determine common parent directory */ ;
     Fileliste_hauptdir(plan->hauptdir,plan->startliste,0);
   }
   if(Backup_plan_list_main_dirs(plan,0)<=0)
     return(-1);

   /* finalize setup */
   if(Fileliste_read_properties(plan->startliste,&zero,
                                2|(!!plan->auslassen_nicht_startliste)<<2)<0)
     return(2);
   if(plan->result_filetypes&(1<<31)) {
     Fileliste_redirect_links(plan->startliste,0);
   }
   if(plan->output_form==0 && Backup_plan_stueck_header(plan,1)){
     /* The info script will be included in any piece.
        Its target competitors need to be excluded */
     ret= Backup_plan_app_excl_info(plan,plan->auskunft_name,
                                         !!plan->auskunft_skript_root);
     if(ret<=0)
       return(2);
   }
   if(plan->split_large_files && plan->split_file_size<=0.0)
     plan->split_file_size = plan->max_size-1024.0*1024.0;
   if(plan->split_large_files) {
     if(plan->split_file_size<=0.0) {
       fprintf(stderr,
               "Note: -split_large_files disabled due to -max_size <= 1m\n");
       plan->split_large_files= 0;
     } else if(plan->cut_file_cmd[0]==0 &&
               strcmp(plan->split_file_dir, "-xorriso_cut_out")!=0) {
       fprintf(stderr,
            "Note: -split_large_files disabled due to empty split_command\n");
       plan->split_large_files= 0;
     }
   }

   /* check wether there are special files and assert -split_large_files */
   for(i=0;i<plan->startliste->anz;i++)
     if(plan->startliste->flags[i]&32) 
   break;
   if(i<plan->startliste->anz && plan->split_large_files==0) {
     fprintf(stderr,
             "\nProblem : Missing -split_large_files with -special_file %s\n",
             plan->startliste->adressen[i]);
     return(2);
   }
   if(plan->configuration>0) {
     if(plan->list_output_format&(1<<10))
       plan->list_output_format|= 8;
     if(plan->list_output_format&(1<<11))
       plan->list_output_format|= 8|16;
     if(plan->list_output_format&(1<<12))
       plan->list_output_format|= 32;
   }
   if((flag&8) && plan->backup_id_ctx!=NULL){
     char md5[16];
     Sregex_md5(&(plan->backup_id_ctx),"",0,md5,2);
     for(i=0;i<16;i++)
       sprintf(plan->backup_id+2*i,"%-2.2x",((unsigned) md5[i])&255);
   }
 } 
 return(0);
not_enough_arguments:;
 fprintf(stderr,"%s : Incomplete option %s\n",argv[0],argv[i]);
 if(i+1<argc) {
   fprintf(stderr,"Arguments following that option :");
   for(i++;i<argc;i++)
     fprintf(stderr," %s",argv[i]);
   fprintf(stderr,"\n");
 } else {
   fprintf(stderr,"Option not followed by any arguments at all\n");
 }
 return(1);
}


#define MELDEN(x) fprintf(fp,"%s\n",x)


int Backup_plan_hilfe(progname,flag)
char *progname;
int flag;
/*
 bit0= to stderr rather than stdout
*/
{
 static char hilfetext[][80]={
#ifdef Cdbbackup_planer_no_germaN

#ifdef Cdbbackup_planer_no_helP
"In diesem Binaerprogramm ist kein Hilfetext enthalten.",
"Lesen Sie den englischsprachigen Hilfetext doc/cd_backup_planer_help oder",
"installieren Sie scdbackup ohne es hinterher mit ./DIET abzuspecken.", 
#else
"In diesem Binaerprogramm ist keine deutschsprachige Hilfe enthalten.",
"Lesen Sie den englischsprachigen Hilfetext mit Option -help oder",
"installieren Sie scdbackup ohne es hinterher mit ./DIET abzuspecken.", 
#endif

#else 

"usage :  %s  [option | file] ...",
"",
"Gewoehnlich teilt dieses Programm eine Liste von Directories in Teile auf,",
"von denen jedes einzelne auf eine CD passt.",
"Es erzeugt dazu ein Skript mit Backupkommandos. Die Listen der Files oder",
"Directories in den einzelnen Stuecken werden anhand von Obergrenzen fuer",
"die Groesse in Bytes und die Anzahl der einzeln erwaehnten Fileadressen",
"festgelegt.",
"",
"Eine solche Ergebnisliste kann nur Directories und Datenfiles enthalten.",
"Andere Filetypen koennen in Directories enthalten sein, aber die meisten",
"ISO9660-CD-Schreibprogramme werden sie nicht verarbeiten.",
"",
"Die Ergebnislisten bestehen aus Ziel- und Quelladresse in einer Form, die",
"sich fuer das Programm  mkisofs  in seinen modernen Versionen eignet.",
"  Wenn die Quelle ein Directory ist:  /Zieladresse/=Fileadresse",
"  Wenn die Quelle ein File ist:       /Zieladresse=Fileadresse",
"Diese Listen werden an ein Backupkommando angehaengt, damit dieses",
"dann eine CD fuellt. Wenn nicht ausdruecklich angegeben und auch nicht",
"zur Vermeidung einer Kollision umgeleitet, lautet die Zieladresse",
"gleich wie die Fileadresse. Abhaengig vom -target_escape Modus wird das",
"Zeichen \"=\" entweder mit \"\\\" maskiert oder zu \"_\" gewandelt.",
"Es koennen auch andere Ausgabeformate als das hier beschriebene bestellt",
"werden.",
"",
"Weil das Arbeitsdirectory, das Rootdirectory und evtl. das -main_dir auf",
"das Rootdirectory der CD abgebildet werden, koennen Kollisionen von Ziel-",
"adressen auftreten. Ein eingebauter Kollisionstest kann diese erkennen und",
"durch Umnennen der Zieladressen vermeiden.",
"",
"Eine Pruefung kann durchgefuehrt werden, um problematische Files zu finden",
" - solche, die von mkisofs-1.12b3 ignoriert werden, wenn sie nicht",
"   ausdruecklich als Argumente gennat werden.",
" - solche, die nicht gelesen werden koennen (Files oder Directories)",
"   Das kann die Schaetzung der Ergebnisgroessen verfaelschen.",
"Die Kombination von -filecheck -append -filecheck -exclude_unreadable",
"kann helfen, diese Probleme zu vermeiden.",
"",
"Inkrementelle Backups koennen mit beliebig vielen Ebenen gemacht werden :",
" -level 0 macht einen vollen Backup ohne Zeiteinschraenkung.",
" -level N macht einen Backup der Files, die seit dem letzten -level N-1",
"  geaendert wurden. -level N-1 muss bereits existieren.",
" - Ein Backup mit -level N macht alle darueber liegenden -level ungueltig.",
"Die noetigen Daten der Ebenen werden in einem gesonderten Directory",
"verwaltet, das mit -conf_dir gewaehlt wird. Dort werden auch alle Parameter",
"des Backups wie Fileliste, Maximalgroessen usw. gespeichert.",
"",
"Inkrementelle Backups geben sich automatisch die Option -younger_than vor.",
"Durch sie werden unter Umstaenden auch aeltere Files mit aufgenommen.",
"Leider verursachen reine Namensaenderungen im allgemeinen Linux Filesystem",
"keine Aenderung an den Zeitmarken des Files. (Aber zB. chmod u+r tut es.)",
"Deshalb lassen sich oft langwierige Pruefsummenberechnungen der Fileinhalte",
"nicht vermeiden.",
"",
"Weil eine zu grosse Zahl einzelner Filenamen Probleme mit alten Versionen",
"von mkisofs bereiten wuerde, versucht das Programm hauptsaechlich, ganze",
"Directories auszuschliessen, die nur sehr wenige junge Files enthalten.",
"Diese Ausnahmen werden dann ausdruecklich in den Backup aufgenommen.",
"Es muss ausserdem eine gewisse Quote von eingesparten Bytes pro",
"hinzukommendem Filenamen erfuellt sein, damit es zum Ausschluss von",
"Directories kommt. Diese ist umso hoeher, je kleiner -max_items gewaehlt ist.",
"",
"Es kann ein gesondertes Skript erzeugt werden, das Auskunft ueber den",
"Backup gibt. Optionen dieses Skriptes sind dann:",
"-hilfe, -id, -count, -list, -levelall , -suball, -fast, -help",
"Andere Argumente werden als Filenamen in den Listen der Stuecke gesucht.",
"Wenn ein passendes Directory oder der Filename gefunden wird, dann wird",
"die Nummer des Stuecks, der gefundene und der gesuchte Name ausgegeben.",
"Wenn nichts gefunden wird, ist der exit-Wert des Skriptes 1.",
"",
"Dieses sog. Infoskript wird nur dann auf der Platte gespeichert und in",
"jedes einzelne Backupstueck eingefuegt, wenn ein Filename mit der",
"Option -info_script angegeben wurde (siehe unten).",
"Dabei wird seine Groesse ueppig mit 100kB geschaetzt. Falls Ihr Infoskript",
"wirklich groesser werden sollte, muss dass durch eine entsprechend",
"kleinere -max_size (siehe unten) ausgeglichen werden.",
"Wenn das Infoskript in die Stuecke eingefuegt wird und die Ausgabgeform",
"fuer mkisofs verwendet wird (Grundeinstellung bzw. -form_script) dann",
"werden alle Fileadressen vom Backup ausgeschlossen, die zum selben",
"Namen auf der CD fuehren wuerden.",
"",
"Wenn eine Zahl als Argument uebergeben wird, ueberspringt das Skript alle",
"Stuecke mit niedrigerer Nummer. Um zB. einen Backup fortzusetzen, der bei",
"der Abfrage fuer Stueck 3 unterbrochen wurde:",
"  skriptname 3",
"So ueberspringt das Skript die Stuecke 1 und 2.",
"Wenn das Skript mit der Option -status_file erzeugt wurde, dann kann es das",
"betroffene Stueck automatisch anhand des Statusfiles finden. Diese Automatik",
"wird durch das Argument \"AUTO\" ausgeloest.",
"",
"Argumente, die nicht mit '-' beginnen, sind Fileadressen, die entweder",
"in die Backupliste oder in die Ausschlussliste aufgenommen werden.",
"(siehe auch Optionen -yes und -no)",
"Vor den Fileadressen koennen Zieladressen stehen. Das ist nuetzlich, um",
"eine CD aus weit verstreuten Daten zusammenzustellen. Dazu kann mit der",
"Option -target_separator vorher ein Trennzeichen eingestellt werden.",
"Beispiel: -target_separator = /bilder=/home/privat/bilder ...",
"",
"Alle Fileaddressen zusammen legen das \"Backupgebiet\" fest, in dem Teil-",
"bereiche durch die Option -not ausgeblendet sein koennen. Das Backupgebiet",
"kann durch Pruefungen (zB. -younger_than) ausgeduennt werden und bildet",
"danach den \"Backupinhalt\".",
"",
"Optionen:",
"",
"Einige Optionen (wie -max_size) erwarten Groessenangaben. Das koennen",
"Dezimalzahlen sein (wie 1000000000) oder Fliesskommazahlen (wie 1.25e9).",
"Sie koennen um einen der Buchstaben {b,k,m,g,t,p,e} ergaenzt sein, wodurch",
"sie mit {1,1024,1024*1024,1024*1024*1024,...,pow(1024.0,6.0)} multipliziert",
"werden. Beispiel :",
"   640m = 640*1024*1024 = 671088640",
" ",
" -hilfe            diesen Text ausgeben und Programm beenden",
" -help             print this text and exit program",
"                   (dieser Text in englischer Sprache und Programmende)",
" -version          Versionskennung auf stdout ausgeben,",
"                   wenn andere Argumente : auf stderr Warnungstext ausgeben,",
"                   immer : Programm mit exit-Wert 4 beenden.",
" --                das naechste Argument ist ein File und keine Option",
"",
" -accept_all_filetypes",
"                   entspricht -result_filetypes any",
" -alldev           gemountete Filesysteme nicht ausschliessen.",
"                   (Grundeinstellung, siehe -xdev und -localdev)",
" -backup_cmd text  der angegebene Text wird in dem vom Programm erzeugten",
"                   Skript als Backup-Kommando verwendet. text unterliegt der",
"                   besonderen -bdir Ersetzung.",
"                   (Grundeinstellung: cd_backup_kommando)",
" -bd               stellt Groessenparamter ein fuer Single Layer Blu-ray",
"                   Medien von nominal \"25 GB\" bei mittlerer Formatierung.",
" -bdir directory   stellt das scdbackup-Installationsdirectory ein. Zzt. ist",
"                   seine einzige Verwendung die -bdir Ersetzung. Sie findet",
"                   statt, wenn File specific_arguments einer inkrementellen",
"                   Backupkonfiguration gelesen wird. Wenn solch ein Argument",
"                   mit @SCDBACKUP_DIR@ beginnt, dann wird dieses Startstueck",
"                   durch das -bdir directory ersetzt. Das gilt auch, wenn",
"                   davor noch ein Anfuehrungszeichen steht.",
"                   Bei einigen wenigen Argumenten wird das Startstueck auch",
"                   ersetzt, wenn  grep .*/scdbackup-[0-9]\\.[0-9]\\.[0-9]/.*",
"                   zutrifft. Das ist die besondere -bdir Ersetzung.",
" -bd_dl            stellt Groessenparamter ein fuer Double Layer Blu-ray",
"                   Medien von nominal \"50 GB\" bei mittlerer Formatierung.",
" -blocklist_fileadr adress_anfang",
"                   wenn adress_anfang nicht leer und nicht \"-\" ist, dann",
"                   werden aus ihm die Adressen von abzuspeichernden Block-",
"                   pruefsummenlisten erzeugt. Ob solche Pruefsummenlisten",
"                   bereitstehen, entscheiden -filter_checksum_blocksize bei",
"                   -filter_md5 und -scan_for_checksum -blocklist bei",
"                   -search_md5 . Die effektive Fileaddrese besteht dann aus",
"                   adress_anfang, dem Recordnamen, dem Datum und der Anzahl-",
"                   angabe aus dem Pruefsummenrecord. zB: wird aus \"/tmp/x_\"",
"                     /tmp/x_1_1_A41123.194219_229208064",
" -changetest_options  code:code:...",
"                   waehlt die Eigenschaften, die waehrend des Tests von",
"                   -younger_than beruecksichtigt werden sollen. Es werden",
"                   dabei die Objekte des Backupgebiets mit dem Speicherbaum",
"                   verglichen, der mit -old_tree eingelesen wurden.",
"                   Die einzelnen Codes werden mit \":\" verbunden :",
"                     type  Filetyp (zB. Directory, Datenfile, Softlink)",
"                     size  Groesse des Inhaltes in Bytes",
"                     crc   Pruefsumme des Inhalts",
"                     inode Interne ID-Nummer des Filesystemobjekts",
"                     off   Vergleich mit altem Speicherbaum wird ausgelassen",
"                     adr   Vergleich wird gemacht (Grundeinstellung)",
"                     mistrust    vermeidet eine Abkuerzung bei CRC,",
"                                 die auf Directory Zeitmarken beruht.",
"                     dir_trust   erlaubt eine schnellere Abkuerzung bei CRC,",
"                                 die nicht erwartet, dass Directories andere",
"                                 Directories durch Umnennen ersetzen.",
"                     assume_old     alle Files als zu alt werten.",
"                     ignore_ctime   nur mtime und nicht ctime beachten.",
"                     ignore_device  bei inode nicht Devicenummern pruefen.",
"                     no_split_economy  keine Pruefung auf unoekonomische",
"                                 Directoryspaltung durch kleine, alte Files.",
"                   HINWEIS: -load_old_tree mit nichtleeren Adressen setzt",
"                   -changetest_options auf \"type:size:crc:adr\"",
"                   Setzen Sie Ihre -changetest_options nach -load_old_tree",
" -collision_test modus  steuert den Kollisionstest fuer Zielnamen",
"                   modus kann sein (Grundeinstellung: -redirect): ",
"                     -off       keinen Kollisionstest machen",
"                     -on        bei Kollisionen warnen",
"                     -redirect  bei Kollisionen Zielnamen aendern",
"                     -abort     bei Kollisionen Programm abbrechen",
" -compare_checksum listenfile suchname altersgrenze jugendgrenze",
"                   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Aus listenfile",
"                   wird ein Record gelesen, der zu den anderen drei",
"                   Argumenten passt. Er sollte von -filter_md5 oder",
"                   von -filter_crc32 stammen. Damit er passt, muss der",
"                   suchname \"-\" sein oder genau der Name des Records. Wenn",
"                   altergrenze nicht \"-\" ist, dann darf er nicht aelter",
"                   sein als diese Grenze. jugendgrenze verlangt dagegen ein",
"                   gewisses Mindestalter (siehe -younger_than wegen des",
"                   Formats von Zeitangaben). Der letzte passende Record aus",
"                   listenfile wird verwendet. Die in ihm verzeichnete",
"                   Datenmenge wird von stdin gelesen und zu einer Pruefsumme",
"                   verarbeitet, die mit der des Records verglichen wird.",
"                   Liefert stdin genuegend Daten und stimmt die Pruefsumme",
"                   ueberein, dann endet das Programm mit Rueckgabewert 0,",
"                   in allen anderen Faellen mit 1.",
"                   Ueberschuessige Daten an stdin werden ignoriert !",
" -complete_list_adr fileadresse",
"                   schreibt eine komplette Fileliste des Backupgebiets in",
"                   den angegebenen File. Wenn die fileadresse kein \"/\"",
"                   enthaelt und eine -level Option angegeben ist, dann wird",
"                   der File im level_# Directory erzeugt (siehe -conf_dir).",
"                   Ohne -level ensteht ein solcher File im -tmp_dir.",
"                   Wenn fileadresse mit .gz endet, dann wird die Ausgabe",
"                   durch das programm gzip komprimiert. Wenn eine Zieladresse",
"                   angegeben ist, dann wird der Ergebnisfile zum letzten",
"                   Stueck des Backups hinzugefuegt. (zB. /compl.gz=compl.gz)",
"                   Die Liste wird in einem Format geschrieben, das Platz",
"                   spart aber immer noch von Menschen lesbar ist. Jede",
"                   Listenzeile besteht aus den mit -list_output_format",
"                   festgelegten Feldern, wobei die Fileadresse immer am",
"                   Zeilenende steht. (Felderliste siehe \"#=# line layout\")",
"                   Die ersten und letzten Zeichen einer Adresse bestimmen",
"                   ihren Typ :",
"                     /x/x/x/   absolute Directoryadresse",
"                     /x/x/x    absolute Fileadresse",
"                     ./x/      tiefergehen in Directory x",
"                     ./x       File x im aktuellen Directory",
"                     ..        ein Directory nach oben gehen",
"                     ../       zurueck zum Startdirectory",
"                     #=# xxxx  Zusatzinformationen",
"                   Die erste angegebene Directoryadresse ist absolut und",
"                   bestimmt das Startdirectory fuer relative Adressen. Mit",
"                   ./x/ und /x/x/x kann der Directorycursor aus dem",
"                   Startdirectory herausbewegt werden. Solange er im",
"                   Startdirectory ist, wird den Adressen eine Zieladresse",
"                   vorangestellt, die dann auch fuer die darunterliegenden",
"                   relativen Adressen gilt. Beide sind durch ein \"=\"-Zeichen",
"                   getrennt. Sollte die Zieladresse selbst \"=\" oder \"\\\"",
"                   enthalten, dann wird diesen jeweils \"\\\" vorangestellt.",
"                   Die Ergebnisliste eignet sich fuer -vanish_list_adr .",
" -composition_memory  tiefe breite speicherlimit",
"                   legt Groesse und Regeln des Beschleunigungspuffers fuer",
"                   die Zerlegung grosser und tiefer Baeume fest. Der Wert",
"                   tiefe bestimmt, wie viele Unterdirectoryebenen sich das",
"                   Programm bei der Groessenuntersuchung eines Directories",
"                   merkt fuer den Fall, dass die Groessen dieser Unterlinge",
"                   fuer eine Aufspaltung erneut bestimmt werden sollen. Das",
"                   vermeidet dann eine erneute Untersuchung der Filebaeume",
"                   auf der Festplatte und kann sehr viel Zeit sparen.",
"                   Der Wert breite bestimmt, wieviele Eintraege ein Unter-",
"                   directory maximal haben darf, um gepuffert zu werden.",
"                   Ihr Speicherbedarf haengt von der Laenge der Filenamen",
"                   ab. Mindestens aber 20 Bytes pro Stueck. Er wird durch",
"                   den Wert speicherlimit begrenzt. Ist dieser Zahl ein \"-\"",
"                   vorangestellt, dann wird freigegebener Pufferspeicher",
"                   nicht als wiederverwendbar gewertet.",
"                   Grundeinstellung : 3 1000 4m          (0 0 0 schaltet ab)",
" -composition_mode kriterium",
"                   legt die Reihenfolge gleichrangiger Fileobjekte bei der",
"                   Zusammenstellung der Backupstuecke fest. Bei allen Modi",
"                   wird versucht, Directorybaeume zusammenzuhalten. Beim",
"                   Anordnen der Startliste und beim Aufspalten von",
"                   Directories kann jedoch das jeweils naechste Fileobjekt",
"                   nach verschiedenen Kriterien ausgewaehlt werden :",
"                    bestfit   das jeweils groesste passende Fileobjekt",
"                    target    die alphabetisch naechsthoehere Zieladresse",
"                    source    die alphabetisch naechsthoehere Quelladresse",
"                    time      das naechstjuengere Fileobjekt",
"                    crawler   das naechste Fileobjekt das der Crawler findet",
"                    default        zZt. wie bestfit",
"                    alphabetical   zZt. wie target",
"                    chronological  zZt. wie time",
"                    fast           zZt. wie crawler",
"                   Bei den Kriterien target, source und time ist die Chance",
"                   recht hoch, dass groessere Directories auf zwei",
"                   aneinandergrenzende Stuecke aufgeteilt werden.",
"                   Die Wahrscheinlichkeit der Aufteilung wird von der Option",
"                   -min_size_percent bestimmt. Eine niedrige Prozentzahl",
"                   beguenstigt Verschwendung von Platz auf den Medien",
"                   waehrend eine hohe Prozentzahl mehr Aufteilungen bewirkt.",
"                   Verschwendung wird durch Aufteilungen nur dann verhindert,",
"                   wenn die Zahl -max_items gross genug ist, um fast alle",
"                   Teile aufzunehmen.",
"                   Das Kriterium crawler fuehrt bei normalgrossen Directories",
"                   zu aehnlichen Ergebnissen wie Kriterium source. Es spaltet",
"                   an Stueckgrenzen gnadenlos Directories auf. Es erspart",
"                   ein bis zwei Suchlaeufe ueber das Backupgebiet, die bei",
"                   grossen Directorybaeumen und wenig RAM sehr lange dauern",
"                   koennen.",
" -content_list_adr fileadresse",
"                   schreibt eine Fileliste des Backupinhalts. Diese Option",
"                   listet nur diejenigen Files, die Teil eines Backupstuecks",
"                   werden. Das Ausgabeformat ist wie bei -complete_list_adr .",
"                   Die Inhaltslisten und die Schwundlisten aller Levels",
"                   zusammen enthalten die selbe Information wie eine einzelne",
"                   Komplettliste. Eine Inhaltsliste markiert auch die Grenzen",
"                   der Backupstuecke.",
" -conf_dir directoryadresse",
"                   legt die Adresse einer Backupkonfiguration fest. Sie",
"                   enthaelt Files mit Informationen, die von inkrementellen",
"                   Backups erzeugt und genutzt werden. (siehe -level)",
"                     specific_arguments : Optionenliste des vollen Backups",
"                                          jede Zeile ist ein Argument. ",
"                     highest_level      : Hoechste gueltige Levelnummer",
"                     ASKME_level_#      : Infoskript zu -level #",
"                     level_#/           : Leveldirectory mit folgenden Files",
"                       last_run_date    : Datumskennung des letzten Backups",
"                       number_of_volumes : Anzahl der Backupstuecke",
"                       content_file_list.gz : siehe -content_list_adr",
"                                              und   -load_old_tree",
"                       vanished_file_list.gz: siehe -vanish_list_adr",
"                       VANISHED.sh.gz       : siehe -vanish_script_adr",
" -cut_file quell_adresse start_byte byte_anzahl ziel_adresse",
"                   Das Programm endet sofort wenn diese Option erledigt ist.",
"                   Sie kopiert ein Intervall von Bytes aus einem File in",
"                   einen neuen File. Zu verwenden mit -split_large_files .",
"                   Wenn quell_adresse '-' ist, dann wird von stdin gelesen",
"                   und eine Zahl von start_byte Bytes uebersprungen. Sonst",
"                   wird quell_adresse als File geoeffnet und der Lesezeiger",
"                   auf start_byte positioniert. ziel_adresse kann '-' fuer",
"                   stdout sein oder eine brauchbare Adresse fuer einen neuen",
"                   File. Die Zwischenmeldungen kontrolliert -pacifier.",
" -debug_filelist listenfile  oberdirectory",
"                   Das Programm endet sofort wenn listenfile ausgelesen ist.",
"                   Es gibt die Listenzeile und ggf. die daraus resultierende",
"                   absolute Fileadresse aus. Dabei wird oberdirectory jeder",
"                   Adresse vorangestellt. Vor Listenzeilen werden drei",
"                   Leerzeichen und die Zahl der bisher angezeigten Files",
"                   gesetzt. Sie lassen sich also mit",
"                    grep -v '^   '   wegfiltern.",
" -debug_stderr_log     Das Programm endet sofort wenn stdin ausgelesen ist.",
"                   Es wandelt CarriageReturn in Newline um und versucht",
"                   dabei auch Leerzeilen wegzulassen. Gedacht um Ergebnisse",
"                   von  scdbackup ... 2>&1 | tee -i ...  lesbarer zu machen.",
" -debug_targetlist listenfile  oberdirectory",
"                   wie -debug_filelist. Jedoch werden die Zieladressen",
"                   gezeigt, was einen Unterschied macht, wenn beim Backup",
"                   mit ziel=quelle Umleitungen bestellt wurden.",
" -delete_filelist  listenfile  oberdirectory",
"                   HINWEIS: fuer das Entfernen von Files waehrend der",
"                            Wiederherstellung eines backups eignet sich",
"                            -delete_targetlist besser.",
"                   Das Programm endet sofort wenn listenfile ausgelesen ist.",
"                   Aus listenfile (\"-\" = stdin) werden Fileadressen mit",
"                   Typ, Groesse und Pruefsumme gelesen.",
"                   Diese Angaben werden mit dem Ist-Zustand der Fileadresse",
"                   verglichen. Dabei wird oberdirectory jeder Adresse",
"                   vorangestellt. Existiert die Adresse und passen die",
"                   anderen Angaben, dann wird der File an dieser Adresse",
"                   mit oberdirectory geloescht. Directories werden geloescht",
"                   wenn sie in der Liste vorkommen und im Filesystem leer",
"                   sind. Ansonsten geht eine Fehlermeldung an stderr.",
"                   listenfile muss mit der Option -vanish_list_adr erzeugt",
"                   worden sein. Er muss Typ, Groesse, Checksumme enthalten.",
"                   War der gesamte Programmlauf fehlerfrei, dann ist der",
"                   Exitwert 0, sonst 1.",
" -delete_targetlist  listenfile  oberdirectory",
"                   wie -delete_filelist. Jedoch werden die Zieladressen",
"                   verwendet, was einen Unterschied macht, wenn beim Backup",
"                   mit ziel=quelle Umleitungen bestellt wurden.",
" -dir_list_adr fileadresse",
"                   schreibt eine komplette Directoryliste des Backupgebiets",
"                   in den angegebenen File. Das Ausgabeformat ist aehnlich",
"                   wie bei -complete_list_adr. Allerdings sind folgende",
"                   Informationen fest voreingestellt:",
"                    \"type:chmod:chown:chgrp:date:size:adr\" wobei",
"                   chmod:chown:chgrp:date Zugriffsrechte, Besitzer und Datum",
"                   angeben. Wenn bei -list_output_format MD5-Pruefsummen",
"                   bestellt sind, werden sie auch hier verwendet.",
" -dir_script_adr fileadresse",
"                   schreibt ein Shellscript zur Wiederherstellung der",
"                   Zugriffsrechte der Directories des Backupgebiets in",
"                   in den angegebenen File. Es werden die Zieladressen",
"                   relativ zum /-Directory der CDs eingetragen.",
" -dir_size_cmd text der angegebene Text wird als Kommando zur Bestimmung",
"                   der Groesse eines Directory verwendet. Das Kommando",
"                   muss die Adresse eines Directory erwarten und als",
"                   erstes Wort auf der Standardausgabe dessen Groesse in",
"                   Kilobyte angeben (wie  du -sk ). Grundeinstellung:",
"                   \"-internal\" ruft kein externes Kommando auf, sondern",
"                   summiert rekursiv die einzelnen Filegroessen und die",
"                   Laenge ihrer Adressen. Dazu kommt ein Zuschlag von 256",
"                   Bytes pro File. Dieser kann mit \"-internal+$zahl\"",
"                   anders gewaehlt werden. So rechnet \"-internal+10\" pro",
"                   File 10 Bytes hinzu.  \"-internal-$zahl\" schaltet das",
"                   Hinzuaddieren der Adressenlaenge ab.",
" -disable_safe_leveling  ist aequivalent zu",
"                   -load_old_tree \"\" \"\"",
"                   -content_list_adr \"\" -complete_list_adr \"\"",
"                   -vanish_script_adr \"\" \"\" -vanish_list_adr \"\" \"\"",
" -dvd              stellt Groessenparamter ein fuer Single Layer DVD Medien",
"                   von nominal \"4.7 GB\". Nicht geeignet fuer DVD-RAM.",
" -dvd_dl           stellt Groessenparamter ein fuer Double Layer DVD Medien",
"                   von nominal \"8.5 GB\".",
" -dvd_ram          stellt Groessenparamter ein fuer DVD-RAM Medien",
"                   von nominal \"4.7 GB\" bei mittlerer Formatierung.",
" -exclude_size groessengrenze  alle files ausschliessen, die mindestens die",
"                   angegebene Zahl von Bytes enthalten. groessengrenze 0",
"                   schaltet diesen Test ab. Wenn groessengrenze mit \"-\"", 
"                   beginnt, gilt bei Softlinks die Groesse des Linkziels.",
"                   'k' 'm' und 'g' werden erkannt wie bei -max_size .",
" -extract_multiplied quelle auslassen",
"                   Das Programm endet sofort wenn stdin schliesst. Es",
"                   liest quelle sequentiell ein und erwartet dass diese",
"                   Daten von Option -multiply_volume erzeugt wurden.",
"                   auslassen gibt die Zahl von Kopien an, die uebersprungen",
"                   werden sollen. Die naechste Kopie wird dann auf stdout",
"                   ausgegeben, eventuell gefolgt von einer entwerteten",
"                   Trennmarke.",
" -extract_tag_tail maximale_menge",
"                   Das Programm endet sofort wenn stdin schliesst. Es frisst",
"                   zunaechst alle Bytes von stdin, bis es auf den Anfang",
"                   einer moeglicherweise gueltigen Pruefsummenmarke stoesst,",
"                   die mindestens 1 Byte bewacht. Der Markenanfang wird dann",
"                   auf stdout ausgegeben ebenso die folgenden Bytes. Diese",
"                   jedoch nur bis zu maximale_menge. Man sollte eine grosse",
"                   maximale_menge angeben, um alle interessanten Daten zu",
"                   erwischen. Anders als bei -search_md5 werden hier auch",
"                   beschaedigte Tags akzeptiert.",
"                   Wurden Daten erfolgreich extrahiert, ist der Exitwert",
"                   0, sonst ist er ungleich 0.",
" -extract_tag_tail_dislocated maximale_menge",
"                   wie -extract_tag_tail mit dem Unterschied, dass schon das",
"                   erste Vorkommen des Kenntextes von Pruefsummenmarken",
"                   genuegt. Es ist nicht noetig, dass die Positionsangabe",
"                   der Marke mit der Position im Datenstrom uebereinstimmt.",
"                   Damit lassen sich zwar Marken aus Datenstroemen lesen, in",
"                   denen Bytes fehlen, man erhaelt aber auch jede Marke, die",
"                   aus irgendwelchen Gruenden Bestandteil der Backupdaten",
"                   ist. Lediglich Marken, die bei Byte 0 des Datenstroms",
"                   beginnen, werden uebersprungen.",
" -facl_list_adr  fileadresse",
"                   wendet das Programm getfacl auf das Backupgebiet an und",
"                   schreibt seine Ausgabe in die angegebene Fileadresse.",
"                   Dadurch werden die ueblichen Zugriffsrechte der Files und",
"                   ggf. auch ihre ACL-Zugriffrechte aufgezeichnet. Das Format",
"                   eignet sich als Eingabe fuer das Programm setfacl",
"                     gunzip <fileaddress.gz | setfacl --restore=-",
"                   Dabei werden die Zieladressen der Files in die entstehende",
"                   Liste eingetragen. Anwendungsbeispiel :",
"                     -facl_list_adr /extra_listen/facl_list.gz=facl_list.gz",
"                   Siehe man getfacl, setfacl, mount (Option \"acl\").",
"                   -level, -tmp_dir und gzip Kompression werden angewendet",
"                   wie bei Option -complete_list_adr.",
" -faclsrc_list_adr  fileadresse",
"                   wie -facl_list_adr. Es werden aber die Quelladdressen der",
"                   Files in die Liste eingetragen. Beide Optionen koennen",
"                   gemeinsam im selben Backup verwendet werden, wenn dabei",
"                   die fileadressen unterschiedlich sind.", 
" -fattr_list_adr  fileadresse",
"                   wendet das Programm getfattr auf das Backupgebiet an und",
"                   schreibt seine Ausgabe in die angegebene Fileadresse.",
"                   Dadurch werden eventuell vorhandene Extended File",
"                   Attributes aufgezeichnet. Das Format eignet sich als",
"                   Eingabe fuer das Programm setfattr",
"                     gunzip <fileaddress.gz | setfattr --restore=-",
"                   Dabei werden die Zieladressen der Files in die entstehende",
"                   Liste eingetragen.",
"                   Siehe man getfattr,setfattr,mount (Option \"user_xattr\")",
" -fattrsrc_list_adr  fileadresse",
"                   wie -fattr_list_adr. Es werden aber die Quelladdressen der",
"                   Files in die Liste eingetragen. Beide Optionen koennen",
"                   gemeinsam im selben Backup verwendet werden, wenn dabei",
"                   die fileadressen unterschiedlich sind.", 
" -filecheck modus  steuert die Voruntersuchung aller betroffenen Files.",
"                   modus kann sein (Grundeinstellung: -off): ",
"                    -off     keine Voruntersuchung (spart viel Zeit)",
"                    -on      warnen vor -implicite_ignored Zeichen und",
"                             bei nicht lesbaren Files",
"                    -append  wie -on. Die -implicite_ignored Namen werden",
"                             der Backupliste ausdruecklich hinzugefuegt.",
"                    -append_hexcode  wie -append. Die Problemzeichen werden",
"                             mit +XY+ kodiert, wobei XY der ASCII-Code in",
"                             hex ist.",
"                    -exclude gleich -off (wegen alteingesessenem Fehler)",
"                    -exclude_unreadable  wie -on. Schliesst aber unlesbare",
"                             Files mit -not vom Backup aus.",
"                    -abort   wie -on. Wenn eine Warnung ausgegeben wurde",
"                             bricht das Programm ab.",
"                    -only    wie -on. Programm nach Untersuchung beenden.",
"                             (exit 16 = Warnung gemeldet , exit 0 = OK)",
"                    -nl_abort  bricht das Programm ab, wenn ein File mit",
"                             einem Zeilenumbruch im Namen gefunden wird.",
"                    -nl_exclude  schliesst Files mit Zeilenumbruch im Namen",
"                             vom Backup aus.",
"                   Die Modi sind kombinierbar.",
" -filter_checksum_blocksize zahl",
"                   bestellt den zZt. umfangreichsten Anhang von Pruefsummen",
"                   bei den Optionen -filter_md5 und -filter_crc32 .",
"                   Es wird eine Pruefsummenmarke fuer den gesamten Datenstrom",
"                   und eine Liste von Pruefsummen fuer Teilbloecke dieser",
"                   Daten berechnet. Die Groesse eines einzelnen Blocks gibt",
"                   der Parameter zahl an. Eine kleine Zahl gibt viele Blocks",
"                   und daher viele 16 Byte grosse MD5-Pruefsummen.",
"                   Diese Option stellt  -filter_checksum_padsize auf 1",
"                   und -pre_checksum_size auf 0. Sie wird ggf. von diesen",
"                   beiden Optionen abgeschaltet.",
" -filter_checksum_padsize zahl",
"                   bei -filter_crc32 oder -filter_md5 wird mindestens die",
"                   hier angegebene zahl von Bytes an das Ende des Bytestroms",
"                   angehaengt. Das geschieht in Form von Pruefsummenmarken,",
"                   mit -search_crc32 bzw. -search_md5 ausgewertet werden",
"                   koennen, sofern -scan_for_checksum angeschaltet ist.",
"                   Dadurch wird ein Listenfile ueberfluessig.",
"                   Ist zahl nicht genau 1, wird -filter_checksum_blocksize",
"                   abgeschaltet.",
" -filter_crc32 listenfile recordname",
"                   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Dabei zaehlt es",
"                   die Bytes und berechnet eine CRC32 Pruefsumme. Zusammen",
"                   mit dem recordname und einer Zeitmarke wird daraus ein",
"                   Record, der an listenfile angehaengt wird. recordname darf",
"                   bis zu 80 Zeichen (ohne Leerzeichen) enthalten.",
"                   Solche Records werden von -compare_checksum verwendet.",
"                   Wenn -pre_checksum_size groesser als 0 ist, dann wird",
"                   erst ein Vorschau-Record fuer -search_crc32 geschrieben.",
"                   CRC32 hat 32 Bits und versagt bei einem von 4e9 Tests.",
"                   Das Programm gibt 0 zurueck, wenn Daten gelesen wurden.",
"                   Wenn keine Bytes aus stdin kommen, wird kein Record",
"                   angehaengt und der Exitwert ist 1.",
" -filter_md5 listenfile recordname",
"                   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Es macht das",
"                   gleiche wie -filter_crc32, jedoch mit \"RSA Data Security",
"                   MD5 Message-Digest Algorithm\". MD5 Pruefsummen haben",
"                   128 Bit und versagen bei einem von 3e38 Tests.",
" -filter_pipe shellkommando",
"                   Hiermit wird vor den Befehlen -filter_md5 , -filter_crc32",
"                   oder -filter_through ein Shellkommando festgelegt, das",
"                   die Ausgabe dieser Befehle weiterverarbeitet. Es startet",
"                   dann beim ersten Byte das auf stdin empfangen wird.",
"                   Werden keine Bytes empfangen, dann wird Shellkommando",
"                   nicht ausgefuehrt.",
" -filter_through   Das Programm endet sofort wenn stdin schliesst. Es leitet",
"                   ohne Aenderungen stdin nach stdout durch. Sinnvoll wenn",
"                   es mit -filter_pipe kombiniert wird.",
" -find directory   gibt den Namen des Directory sowie aller Files und",
"                   Directories darunter aus und beendet dann das Programm.",
"                   Wie das Programm 'find' aber ohne jede Interpretation des",
"                   Directorynamens.",
" -find_path_list filename",
"                   liest die einzelnen Zeilen aus File filename als",
"                   Fileadressen, wendet Option -find auf sie an und beendet",
"                   das Programm.",
" -follow_symbolic_links modus",
"                   Im Normalmodus \"off\" werden symbolische Links nur mit",
"                   ihrer eigenen (geringen) Filegroesse gezaehlt. Im Modus",
"                   \"on\" werden die Linkziele in die Groessenberechnung",
"                   einbezogen. Ist das Ziel ein Directory, dann wird der",
"                   ganze darunterliegende Baum gemessen und ggf. auch auf",
"                   mehrere Backupstuecke verteilt. Das Formatierungsprogramm",
"                   muss dann ebenfalls symbolischen Links folgen, damit die",
"                   eingeplanten Daten auch wirklich in den Backup aufgenommen",
"                   werden. Bei mkisofs ist Option \"-f\" noetig, bei afio",
"                   heisst die entsprechende Option \"-h\".",
"                   Bei ungesunden (zyklischen) Linksituationen kann es zu",
"                   Mehrfachkopien der betroffenen Directories im Backup",
"                   kommen. Das wird dann von Fehlermeldungen begleitet :",
"                     \"Too many levels of symbolic links\"",
" -form_listing     die Ergebniszeilen werden in folgender Form gezeigt",
"                     Typ Groesse [Intervall]Adresse",
"                   und das Backupkommando wird nicht benutzt. Ein Intervall",
"                   kann durch -split_large_files entstehen und hat die Form",
"                     **Stuecknummer/Stueckgesamtzahl:Startbyte-Endbyte**",
" -form_path_list_afio listenfilename harmlose_zahl_von_argumenten",
"                   Spezialversion von -form_script_raw. Die Fileadressen",
"                   werden keine Argumente sondern in listenfilename",
"                   geschrieben (siehe auch -form_path_list_mkisofs) und",
"                   dieser listenfilename wird mit Option -path-list an das",
"                   Schreibkommando uebergeben. Es sollte den File dann mit",
"                   -find_path_list zur vollstaendigen fileliste expandieren.",
" -form_path_list_afio_old  listenfilename harmlose_zahl_von_argumenten",
"                   wie -form_path_list_afio aber ohne dass aufgespaltene",
"                   Directories per Option -dir-path-list vom Schreibkommando",
"                   ins afio Archiv uebernommen werden.",
" -form_path_list_mkisofs listenfilename harmlose_zahl_von_argumenten",
"                   Spezialversion von -form_script . Geeignet fuer Versionen",
"                   von mkisofs, die die Option -path-list kennen.",
"                   Mit dieser Form sind sehr grosse Werte bei -max_items",
"                   zulaessig (z.B. 100000), weil die Uebergabe als Shell-",
"                   argumente vermieden wird.",
"                   listenfilename wird erzeugt, um die Liste der Fileadressen",
"                   zu speichern. Deswegen ist Schreibrecht an seinem Wirts-",
"                   directory noetig. Enthaelt listenfilename kein \"/\",",
"                   dann wird ihm das -tmp_dir vorangestellt.",
"                   harmlose_zahl_von_argumenten sollte klein genug sein,",
"                   um die Shell nicht zu ueberfordern (zB. 100).",
"                   Diese Option stellt -target_escape auf \"on\".",
" -form_quoted_list_xorriso listenfilename harmlose_zahl_von_argumenten",
"                   Spezialversion von -form_path_list_mkisofs . Geeignet",
"                   fuer Versionen von xorriso, deren -as mkisofs emulation",
"                   die Option --quoted_path_list kennt. Dieses Format kann",
"                   Filenamen mit NewLine-Zeichen verarbeiten.",
" -form_raw         die Ergebniszeilen enthalten nur die Adresse (und",
"                   eventuell ein Intervall wie bei -form_listing). Jedes",
"                   einzelne Backupteil wird durch eine Leerzeile beended.",
" -form_script      die Ergebniszeilen werden als Argumente der folgenden",
"                   Form an das Backupkommando angehaengt",
"                     /addresse[/]=addresse",
"                   dies ist die Grundeinstellung fuer das Ausgabeformat.",
"                   Diese Option stellt -target_escape auf \"off\".",
" -form_script_here die Ergebniszeilen werden als Standardeingabe der",
"                   folgenden Form an das Backupkommando angehaengt",
"                     addresse",
"                   Das geschieht durch ein here-document mit dem",
"                   Schlusszeichen \"+\".",
" -form_script_raw  die Ergebniszeilen werden als Argumente der folgenden",
"                   Form an das Backupkommando angehaengt",
"                     addresse",
" -implicite_ignored text  Abhilfe gegen eine Spezialitaet von mkisofs-1.12b3",
"                   Der angegebene Text meldet diejenigen Zeichen an, die",
"                   das Backup-Kommando veranlassen, einen File auszulassen,",
"                   sofern sein Name eines der Zeichen enthaelt und er",
"                   nicht ausdruecklich dem Backup-Kommando genannt wird.",
"                   Bei mkisofs-1.12b3 sollte der Text lauten: \"~#\"",
" -info_script name  legt den Filenamen fuer das getrennt zu speichernde",
"                   Infoskript fest. Es werden zwei Files erzeugt: einer mit",
"                   dem angegebenen Namen und einer mit dem Zusatz .pnr am",
"                   Namensende (zB. ASKME und ASKME.pnr)",
"                   Namen mit besonderer Bedeutung sind;",
"                    -off     kein Infoskript erzeugen",
"                    -short   Infoskript-Dummy ohne Filelisten",
"                    -full    voll verwendbares Infoskript",
"                    -short_on_level Bei -level -create_configuration wird",
"                             -info_script -short voreingestellt", 
"                   Alle anderen Namen werden als Speichernamen fuer das",
"                   Infoskript verwendet und in jedes einzelne Backupstueck",
"                   eingefuegt.",
"                   Anders als bei den Fileadressen ist hier als Trenn-",
"                   zeichen fuer Zieladressen immer fest \"=\" eingestellt.",
"                   (Beispiel: /ASKME=/home/cd_backup_planer_dir/ASKME)",
" -info_exclude     Das Infoskript nicht in die Backupstuecke einfuegen,",
"                   auch wenn vorher ein Filename mit -info_script fest-",
"                   gelegt wurde.",
" -level kennwort   inkrementeller Backup mit der Konfiguration, die durch",
"                   -conf_dir angewaehlt wird. Das Kennwort bestimmt ganz",
"                   wesentlich den weiteren Verlauf des Backups :",
"                     -create_configuration erzeugt eine neues -conf_dir ,",
"                      traegt dort die anderen Parameter des Laufs ein",
"                      und startet einen Level 0 Lauf.",
"                     -update_configuration traegt neue Parameter in ein",
"                      vorhandenes -conf_dir ein und startet -level 0.",
"                     -off schaltet den inkrementellen Modus ab.",
"                     -last waehlt das zuletzt benutzte Level.",
"                     -next waehlt das naechste ungenutzte Level (-last+1).",
"                     0,1,2,... direkte Anwahl des Levels durch seine Nummer.",
"                   -last, -next und Levelnummern fuehren dazu, dass die",
"                   meisten Optionen ausser -conf_dir und -level ignoriert und",
"                   durch die aus -conf_dir/specific_arguments ersetzt werden.",
"                   Einige Kommandozeilenoptionen setzen sich jedoch durch :",
"                     -backup_cmd -composition_* -dir_size_cmd",
"                     -form_* -info_* -max_* -min_*",
"                     -prompt_script -scattering -status_file",
"                     -tmp_dir -shell_to_use -verify_source_filter",
"                     (d.h. das Backupgebiet kann so nicht geaendert werden.)",
"                   Wenn der File specific_arguments von diesem Programm",
"                   erzeugt oder geaendert wird, beginnt er mit Zeilen die",
"                   folgende Optionen voreinstellen :",
"                     -working_directory ...Adresse...",
"                     -load_old_tree content_file_list",
"                                    vanished_file_list",
"                     -changetest_options type:size:inode:adr",
"                     -list_output_format type:size:inode:adr",
"                     -content_list_adr content_file_list.gz",
"                     -vanish_script_adr content_file_list",
"                                  /tmp/VANISHED.sh.gz=VANISHED.sh.gz",
"                     -vanish_list_adr content_file_list",
"                                      vanished_file_list.gz",
"                   Wenn Sie keine Inhaltslisten, Schwundlisten und",
"                   Loeschskripte bei inkrementellen Backups wollen, dann kann",
"                   das mit Option -disable_safe_leveling bestellt werden.",
"                   -changetest_options und -list_output_format werden",
"                   entsprechend der Startargumente eingestellt. Obiges",
"                   Beispiel geht von der Grundeinstellung aus.",
" -level_timestamp datumstext_oder_fileadresse",
"                   legt das eigene Datum eines inkrementellen Backuplaufs",
"                   auf einen frueheren Zeitpunkt als den Programmstart.",
"                   Dieses Datum wird spaeter vom naechsthoeheren Level als",
"                   Stichdatum fuer -younger_than verwendet.",
"                   Beginnt datumstext_oder_fileadresse mit '.' oder '/',",
"                   dann ist es eine Fileadresse, von deren mtime oder ctime",
"                   die juengere verwendet wird. Ansonsten ist es eine",
"                   Zeitangabe wie bei -younger_than beschrieben.",
"                   Normalerweise ist diese Einstellung nicht noetig.",
"                   Wenn man jedoch den inkrementellen Backup von einem",
"                   Filesystemsnapshot macht,dann MUSS das DATUM DES SNAPSHOTS",
"                   mit -level_timestamp gesetzt werden.",
"                   Es ist ungefaehrlich, wenn dieses Datum etwas zu alt ist.",
" -list_checksum listenfile suchname altersgrenze jugendgrenze",
"                   Das Programm endet sofort nachdem es diejenigen Records",
"                   aus listenfile ausgegeben hat, auf die die anderen drei",
"                   Argumente passen. Der letzte dieser Records wuerde von",
"                   -compare_checksum (siehe dort) verwendet werden.",
"                   Wird mindestens ein Record gefunden, dann endet das",
"                   Programm mit Exitwert 0, sonst gibt es 1 zurueck.",
" -list_output_format  code:code:...",
"                   legt fest, welche Informationen in die Listenzeilen",
"                   geschrieben werden sollen, die durch -*_list_adr Optionen",
"                   entstehen.",
"                   Die einzelnen Codes werden mit \":\" verbunden :",
"                     type  erster Buchstabe zeigt Filetyp an :",
"                           d = Directory , l = Softlink , b = Blockdevice",
"                           - = Datenfile , p = Pipe , 0 = unbekannt",
"                           Eine per Komma getrennte Liste von Zusatz-",
"                           informationen kann angefuegt sein:",
"                            cZahl ... ctime in Sekunden seit 1970",
"                            dZahl ... Devicenummer",
"                            iZahl ... inode-Nummer",
"                            tZahl ... mtime in Sekunden seit 1970",
"                            pStart:Vgr ... zeigt Splitfile Stueckadresse an",
"                                      nennt Startbyte und volle Groesse",
"                            PStart:Sgr ... zeigt Splitfile Originaladresse an",
"                                      nennt Startbyte und Stueckgroesse",
"                     size  Groesse des Inhaltes in Bytes (-1 = unbekannt)",
"                     crc   CRC32 Pruefsumme des Inhalts (-1 = unknown)",
"                     md5   MD5 (128 Bit) statt CRC32 (32 Bit)",
"                     inode inode-Nummer wird an type angehaengt, bei",
"                           Directories oder absoluten Adressen auch die",
"                           Devicenummer. zB.: -,i10201  oder  d,d772,i4013",
"                     mtime mtime an type anhaengen. zB.: -,t1151999973",
"                     ctime ctime an type anhaengen. zB.: -,c1151999973",
"                     adr   Fileadresse (immer an, Code wird ignoriert)",
"                     on_level_crc wie crc aber nur wenn -level nicht -off ist",
"                     on_level_md5 wie md5 wenn -level nicht -off ist",
"                     on_level_inode wie inode wenn -level nicht -off ist",
"                   Die Reihenfolge ist immer  type size crc adr . zB:",
"                     d 91286 A929B47C /home/thomas/",
" -load_old_tree  altlisten_adresse  schwund_listenadresse",
"                   laedt eine Komplettliste oder die Inhaltslisten und",
"                   Schwundlisten eines Backups mit mehreren Leveln um einen",
"                   Speicherbaum aufzubauen, der alten den Zustand des",
"                   Backupgebietes wiedergibt, wie er beim Backup des",
"                   naechsttieferen Levels gewesen ist. Der Baum wird von den",
"                   Optionen -younger_than und -vanish_list_adr benutzt, um",
"                   Files zu finden, die seitdem geloescht oder umbenannt",
"                   wurden. Wenn eine -level Option vorhanden ist und die",
"                   Adressen kein \"/\" enthalten, dann werden die Files aus",
"                   den tieferen level_# Directories gelesen. Daher geschieht",
"                   bei Level 0 durch diese Option garnichts.",
"                   Wenn altlisten_adresse eine Komplettliste ist, dann muss",
"                   sonst nichts gelesen werden und schwund_listenadresse darf",
"                   irgendein nichtleerer Text sein. Wenn altlisten_adresse",
"                   nicht existiert, aber mit Zusatz .gz gefunden wird, dann",
"                   wird dieser .gz File verwendet. Wenn ein File mit .gz",
"                   gelesen wird, geschieht das durch das Programm gunzip.",
"                   WARNUNG: So ein Baum kann viel Speicher verbrauchen.",
"                   HINWEIS: altlisten_adresse von dieser Option ist identisch",
"                   mit altlisten_adresse von -vanish_list_adr. Wenn beide",
"                   Optionen angegeben sind, setzt sich die altlisten_adresse",
"                   der letzten durch durch.",
"                   HINWEIS: -load_old_tree mit nichtleeren Adressen setzt",
"                   -changetest_options auf \"type:size:crc:adr\"",
" -localdev         (betrifft zZt. nur NFS auf SGI-IRIX Systemen)",
"                   alle Filesysteme auslassen, die per Netzwerk angeschlossen",
"                   sind. Gegenteil: -alldev , siehe auch -xdev .",
" -main_dir text    Name eines Directory, das auf das root-Directory",
"                   der CD abgebildet werden soll. Besonders nuetzlich",
"                   ist der Text  -auto . Dann wird das gemeinsame Ober-",
"                   directory der Backupliste verwendet. Der Text  -off",
"                   schaltet diese Abbildung aus (Grundeinstellung).",
" -max_items zahl   die Maximalzahl von einzeln erwaehnten Files und",
"                   Directories pro Backupstueck wird festgelegt.",
"                   (Grundeinstellung: 400)",
" -max_size zahl    die Maximale Groesse eines Backupstuecks in Byte",
"                   wird festgelegt. (Grundeinstellung: 600m)",
" -max_size_stack groesse anzahl",
"                   meldet eine Anzahl von Backupstuecken (d.h. Datentraeger)",
"                   mit abweichender Kapazitaet an. Das kann mehrere Male",
"                   geschehen, wobei die zuletzt angemeldete Gruppe mit",
"                   hoechster Prioritaet verplant wird. Beachten Sie dann die",
"                   Groessenanmeldung der einzelnen Stuecke, um die richtigen",
"                   Medien auszuwaehlen. Zum Beispiel: 25 CDs zu 700 MB, der",
"                   Rest hat 650 MB :  -max_size_stack 685m 25 -max_size 640m",
" -max_size_stack_save -on|-off",
"                   Es kann vorkommen, dass groessere Stuecke nur soweit",
"                   ausgelastet sind, dass auch -max_size ausreichen wuerde.",
"                   Normalerweise (\"-on\") gilt das -max_size_stack Stueck",
"                   dann als nicht verbraucht. Das kann dazu fuehren, dass",
"                   ein grosser Datentraeger erst unerwartet spaet angefordert",
"                   wird. Im Modus \"-off\" werden die -max_size_stack Stuecke",
"                   stur der Reihe nach verplant.",
" -max_tree_depth zahl",
"                   Wenn extrem viele Directories ineinander verschachtelt",
"                   sind, kann es bei ihrer Untersuchung zum Ueberlauf des",
"                   Stack kommen. Dieser Ueberlauf ist gefaehrlich und wird",
"                   daher durch eine Tiefenbeschraenkung und einen",
"                   kontrollierten Programmabbruch verhindert. Die Grund-",
"                   einstellung von 128 kann auf 16 gesenkt oder beliebig",
"                   erhoeht werden.",
"                   Unter Linux wird die Groesse des Stack mit ulimit -s",
"                   eingestellt. Sie sollte mindestens 20 kB * zahl betragen.",
" -merge_trees quelle ziel papierkorb optionen",
"                   Das Programm endet sofort wenn diese Option ausgefuehrt",
"                   worden ist. Sie bewegt quelle und alle Fileobjekte",
"                   darunter auf die neue Adresse ziel und analoge Adressen",
"                   darunter. Die Bewegung darf die Grenzen von Filesystemen",
"                   ueberschreiten.",
"                   ziel und eventuell schon darunter vorhandene Directories",
"                   uebernehmen die Zugangsrechte von quelle und seinen",
"                   Unterdirectories. Files unter ziel, die ueberschrieben",
"                   werden, koennen in einer analogen Adressstruktur unter",
"                   Directory papierkorb aufgefangen werden. Ist papierkorb",
"                   ein einzelner Bindestrich \"-\" so werden vorhandene Files",
"                   unter ziel beim Ueberschreiben geloescht.",
"                   Ist vor dieser Option mit -load_old_tree die Adresse einer",
"                   Directoryliste (erzeugt mit -dir_list_adr) angegeben, dann",
"                   werden die dort verzeichneten Eigenschaften ggf. auf die",
"                   Directories unter ziel uebertragen. Ansonsten werden die",
"                   Eigenschaften von quelle abkopiert.",
"                   Folgende Optionen werden erkannt :",
"                     overwrite        vorhandene Files immer ueberschreiben",
"                     no_overwrite     vorhandene Files nie ueberschreiben",
"                                      Rechte existierender Directories nicht",
"                                      aendern.",
"                     overwrite_older  vorhandene Files ueberschreiben, wenn",
"                                      in ziel aelter als in quell",
"                     move             Files bewegen (in quelle loeschen)",
"                     copy             Files kopieren (in quelle erhalten)",
"                     owner_plus_w     immer w-Recht fuer Besitzer erteilen",
"                     chmod_original   Rechte ungeaendert uebernehmen",
"                     tree_prefix=<path>  voranzustellendes Directory beim",
"                                      Aufbau von -load_old_tree. (Nur noetig",
"                                      wenn <path> ungleich quelle ist.)",
"                     tree_source      bei -load_old_tree die Quelladressen",
"                                      einlesen.",
"                     tree_target      bei -load_old_tree die Zieladressen",
"                                      einlesen.",
"                   Optionen koennen mit Doppelpunkt verbunden werden. Die",
"                   leere Optionsliste \"-\" ist gleich :",
"                     \"overwrite:move:chmod_original:tree_target\" .",
" -min_size_percent zahl  Prozentanteil von -max_size, der mindestens ge-",
"                   fuellt sein muss, bevor ein neues Stueck eroeffnet wird.",
"                   Wenn dieser Anteil nicht gefuellt ist, kann ein",
"                   groesseres Directory aufgespalten werden.",
"                   (Grundeinstellung:  80.0)",
" -multiply_volume quelle wiederhol_faktor",
"                   Das Programm endet sofort wenn stdin schliesst. Es",
"                   liest quelle mehrfach und gibt diese Kopien hintereinander",
"                   auf stdout aus. Zwischen die Kopien werden Markierungen",
"                   gesetzt, die es erlauben, eine einzelne Kopie mit der",
"                   Option -extract_multiplied herauszuziehen.",
" -no               alle folgenden Fileadressen sollen nicht mit in",
"                   den Backup aufgenommen werden. Diese Option kann",
"                   zu einer langen Ergebnisliste fuehren, weil das",
"                   Auslassen eines Directories die Erwaehnung all",
"                   seiner Nachbarn erfordert.",
" -not              wirkt genau wie -no",
"",
" -not_inode        wirkt wie -no, versucht aber nicht nur anhand der Adressen",
"                   sondern auch anhand von Device- und Inode-Nummer die",
"                   auszuschliessenden Fileobjekte zu erkennen. Dieser Test",
"                   ist aufwendiger und erfordert u.U. mehr Plattenzugriffe.",
" -not_pattern shellpattern  fuegt das Argument shellpattern zur Liste der",
"                   Suchausdruecke hinzu, die auf Filenamen angewendet werden",
"                   und bei einem Treffer zur Abweisung des Files fuehren.",
"                   shellpattern kann die ueblichen Platzhalter *,?,[...]",
"                   enthalten.",
" -not_regex regex  fuegt das Argument regex zur Liste der Suchausdruecke",
"                   hinzu, die auf Filenamen angewendet werden und bei einem",
"                   Treffer zur Abweisung des Files fuehren.",
"                   regex ist eine regular expression, wie sie in den",
"                   Programmen grep, sed oder vi ueblich ist.",
" -pacifier modus|rythmus  einige Optionen geben Meldungen ueber ihren",
"                   Fortschritt auf stderr aus. Wenn Sie stderr in einen",
"                   File leiten, kann das laestig sein.",
"                     -off       schaltet die Meldungen aus",
"                     -on        schaltet einen Rythmus von 1 Meldung/s ein",
"                     Eine Zahl waehlt einen Rythmus in Sekunden.",
"                     -timestamp schreibt die Zahl der verstrichenen Sekunden",
"                                an den Zeilenanfang.",
"                     -composition_memory zeigt den Pufferzustand der",
"                                gleichnamigen Option waehrend der Komposition",
"                     -log_mode  ungekuerzte Meldungen in eigenen Zeilen",
"                   Bis auf -off koennen alle diese Modi kombiniert werden :",
"                     -pacifier 3 -pacifier -timestamp -pacifier -log_mode", 
" -permute_blocks key quelle blockgroesse blockzahl datenstart",
"                   Das Programm endet sofort wenn stdin schliesst. Es",
"                   liest File quelle mit wahlfreiem Zugriff und schreibt",
"                   sein Ergebnis auf stdout. Dabei werden die Daten",
"                   blockweise auf andere Positionen im Ausgabestrom versetzt.",
"                   Vor die vertauschten Blocks wird ein Header geschrieben,",
"                   der angibt wie die Daten veraendert wurden :",
"                     scdbackup block permutation header v0.1",
"                     blocksize=...        Zahl der Bytes pro Block",
"                     complete_blocks=...  Zahl der kompletten Blocks",
"                     remainder=...        Bytes im letzten Block",
"                     permutation_key=...  Vertauschungsregel (0 bis 7)",
"                     datastart=...        Nummer des ersten Datenblocks",
"                   key gibt die Vertauschungsregel an :",
"                     0= keine Vertauschung",
"                     1= Filehaelften vertauschen",
"                     2= Blockweise spiegeln",
"                     3= 1. und 2. Viertel tauschen, 3. und 4. ebenso",
"                     plus 4 = zusaetzlich Reissverschluss der Haelften",
"                     (zB. 7 = 3+4 = Viertel tauschen, dann Reissverschluss)",
"                   key \"restore\" veranlasst Ruecktauschen in den Original-",
"                   zustand anhand des Fileheaders von quelle.",
"                   Steht vor key ein \"-\", dann wird ebenfalls die Ver-",
"                   tauschung wieder rueckgaengig gemacht. Hierbei wird aber",
"                   dem Fileheader nicht getraut, sondern die Parameter",
"                   key, blockgroesse, blockzahl und datenstart verwendet.",
"                   Beim Vertauschen kann blockzahl fuer Blockdevices",
"                   verwendet werden, um die zu lesenden Blocks festzulegen.",
"                   Sonst wird diese Zahl aus der Filegroesse errechnet, die",
"                   fuer Blockdevices durch Bytezaehlen ermittelt wird.",
"                   Normalerweise sollten blockzahl und datenstart beim",
"                   Vertauschen und beim Ruecktauschen aber 0 sein.",
" -pre_checksum_size zahl  legt die Groesse des Anfangsstuecks im Datenstrom",
"                   fest, das fuer die Vorschaupruefsumme verwendet wird.",
"                   Wenn diese Zahl groesser als 0 ist, schreiben -filter_md5",
"                   und -filter_crc32 zwei Records. Der erste mit dem Anhang",
"                   \"_pre\" am Namen kann von den Optionen -search_md5 und",
"                   -search_crc32 genutzt werden, um den zweiten, vollen",
"                   Record zu finden. Erlaubt sind Zahlen groesser oder",
"                   gleich -1. Grundeinstellung ist 0.",
"                   Ist zahl groesser als 0, wird -filter_checksum_blocksize",
"                   abgeschaltet.",
" -prevent_doubles on|off  bei Einstellung \"on\" wird bei jedem neuen Eintrag",
"                   geprueft, ob er schon im entstehenden Backupplan enthalten",
"                   ist. Ggf. wird er dann ausgelassen. Dieser Test kann",
"                   nennenswert Zeit benoetigen und ist daher normalerweise",
"                   abgeschaltet.", 
" -problemlist_limit zahl",
"                   Legt die Maximalzahl von Eintraegen in Problemlisten",
"                   fest, die von -filecheck, -exclude_size, -xdev veranlasst",
"                   sein koennen. Wenn zahl bei einer dieser Listen ueber-",
"                   schritten wird, bricht die Planung schliesslich ab.",
"                   Grundeinstellung : 10000",
" -process_blocklist aktion listenfile auslassen quelle ziel",
"                   gibt eine Uebersicht ueber eventuelle Datenschaeden und",
"                   bietet auch Reparaturmoeglichkeiten. Nach Ausfuehrung",
"                   dieser Option endet das Programm.",
"                   Die Fehlererkennung beruht auf einer Blockpruefsummenliste",
"                   (siehe -filter_checksum_blocksize). Die Liste wird aus",
"                   listenfile gelesen, wo sie an einer gueltigen Pruefsummen-",
"                   marke haengen muss. Als listenfile kommen u.a. in Frage:",
"                   Backupstuecke, die mit -filter_checksum_blocksize ",
"                   geschrieben wurden oder Ergebnisse von -blocklist_fileadr.",
"                   auslassen gibt die Zahl der Marken an, die uebersprungen",
"                   werden sollen, bis schliesslich eine eingelesen wird.",
"                   Im Normalfall ist der Wert 0 richtig.",
"                   quelle ist die Fileadresse (\"-\"=stdin) eines eventuell",
"                   beschaedigten Backupstuecks.",
"                   ziel ist eine Fileadresse, wo unbeschaedigte Datenbloecke",
"                   zusammengesetzt werden (\"-\"=stdin/stdout).",
"                   Je nach aktion ist bei diesen beiden wahlfreier Zugriff", 
"                   noetig oder es reicht sequentieller (in einem Durchlauf).",
"                   Folgende Kennwoerter fuer aktion werden erkannt :",
"                   \"report\" liest quelle sequentiell und druckt in",
"                      getrennten Zeilen die Nummern schadhafter Bloecke auf",
"                      stdout. Das Ende dieser Fehlerliste wird durch die",
"                      Zeile \"@\" markiert.",
"                      ziel wird bei \"report\" nicht benutzt.",
"                   \"collect\" liest quelle sequentiell und schreibt die",
"                      unbeschaedigten Blocks mit wahlfreiem Zugriff nach",
"                      ziel.",
"                   \"mend\" liest ziel mit wahlfreiem Zugriff und greift bei",
"                      beschaedigten Blocks wahlfrei auf quelle zu, um von",
"                      dort eine unbeschaedigte Kopie zu holen. Gelingt das,",
"                      dann wird dieser Block nach ziel geschrieben.",
"                      mend erlaubt die optionale Angabe einer Startnummer",
"                      und auch einer Endnummer. Die Blocknummerierung beginnt",
"                      bei 0. Der Text \"end\" bezeichnet den letzten Block.",
"                      Ist die Startnummer groesser als die Endnummer, dann",
"                      werden die Blocks entsprechend rueckwaerts bearbeitet.",
"                      Beispiele:  mend  mend:1200  mend:10k:20k  mend:end:0",
"                   \"merge\" liest quelle und ziel sequentiell und schreibt",
"                      unbeschaedigte Blocks oder beschaedigte aus ziel oder",
"                      aber entsprechend viele 0-Bytes auf stdout.",
"                   Bei schreibender aktion wird die Blockpruefsummenliste an",
"                   ziel angehaengt.",
"                   Eine kurze Beschreibung der Blocklistenparameter und eine",
"                   Zusammenfassung des Ergebnisses wird auf stderr gedruckt.",
"                   Wenn man stderr mit einem Programm auswertet, sollte",
"                   man vor die Option -process_blocklist die Option",
"                   -pacifier -off setzen.",
"                   Der Exitwert ist 0, wenn die Ergebnisdaten als komplett",
"                   und unbeschaedigt bewertet werden. Sonst ist er nicht 0.",
" -result_filetypes typenliste",
"                   legt die Filetypen fest, deren Adressen in der Ergebnis-",
"                   liste bzw. -skript stehen duerfen. typenliste besteht aus",
"                   einem oder mehreren dieser Woerter, getrennt durch \":\".",
"                     data          gewoehnliche Datenfiles",
"                     dir           Directories",
"                     symlink       symbolische Links",
"                     block_device  Geraetefiles fuer Festplatten,CD, u.a.",
"                     unclassified  alles andere",
"                     any           alles zusammen",
"                     arg_redirect  ersetze symlink Argumente durch Linkziel",
"                   Filetypen, die nicht in der typenliste erwaehnt sind,",
"                   werden verworfen, wenn sie in die Liste eines",
"                   Backupstuecks eingetragen werden sollen.",
"                   Wenn Pseudotyp  arg_redirect  gegeben ist, werden die",
"                   Argumente dieses Programms auf symbolische Links geprueft",
"                   und diese durch ihre Linkziele (Datenquellen) ersetzt.",
" -scan_for_checksum modus",
"                   steuert fuer -search_crc32 und -search_md5 die Suche nach",
"                   angehaengten Pruefsummenmarken (-filter_checksum_padsize).",
"                   Folgende Modi werden erkannt :",
"                    -on     Suche angeschaltet. Zusatzoptionen unveraendert.",
"                    -off    Suche abgeschaltet. Zusatzoptionen unveraendert.",
"                    -first  Suche an. Nach erster gueltiger Marke enden.",
"                    -whole  Suche an. Bis Ende des Bytestroms lesen.",
"                    -standalone  Suche an. listenfile nicht lesen.",
"                    -withlist    Suche an. Mit listenfile.",
"                    -print       Suche an. Gueltige Marken am Ende auflisten.",
"                    -noprint     Suche an. Nicht auflisten.",
"                    -anytag      Alle Arten von Marken akzeptieren.",
"                    -blocklist   Nur Marken mit Blockpruefsummenliste.",
"                    -pacifier    Den Fortschritt des Lesens melden (stderr).",
"                    -nopacifier  Keine Fortschrittsmeldungen ausgeben.",
"                    -source:text Fileadresse anstelle der Standardeingabe",
"                    -skip:zahl   Anzahl der Bytes, die bei der Eingabe",
"                                 ausgelassen wurden. Bei -source: werden",
"                                 auch entprechend viele uebersprungen.",
"                   Mehrfachnennungen sind erlaubt. Grundeinstellung :",
"                    -scan_for_checksum -withlist -scan_for_checksum -off",
"                    -scan_for_checksum -anytag -scan_for_checksum -nopacifier",
" -scattering -on|-off  Modus -on erlaubt das Verlagern grosser Files auf",
"                   andere CDs als die, die ihre Directorynachbarn aufnehmen.",
"                   Das hilft -min_size_percent zu erfuellen, kann aber zu",
"                   weit verstreuten Directoryinhalten fuehren. Diese Option",
"                   wirkt nur zusammen mit  -composition_mode default .",
" -search_crc32 listenfile suchname altersgrenze jugendgrenze",
"                   wie -compare_checksum aber mit automatischer Erkennung des",
"                   Datenstroms und nur fuer Records mit CRC32 Pruefsummen.",
"                   Die Suchmethode haengt von -pre_checksum_size ab. Wenn",
"                   sie 0 ist, dann werden alle Records in den Arbeitsspeicher",
"                   geladen, nach Groesse sortiert und mit Zwischenpruefsummen",
"                   des Datenstroms verglichen. Der Datenstrom wird bis zu",
"                   seinem Ende gelesen. Mehr als eine Uebereinstimmung kann",
"                   gefunden werden (besonders mit Vorschaurecords).",
"                   Ist -pre_checksum_size -1, dann wird das lesen nach der",
"                   ersten Uebereinstimmung beendet.",
"                   Wenn -pre_checksum_size >0 ist, dann wird zunaechst ein",
"                   passender Vorschaurecord gesucht, um den zugehoerigen",
"                   Record mit der vollen Pruefsumme zu finden. Die Records",
"                   fuer die Vorschau werden von -filter_md5 und -filter_crc32",
"                   erzeugt, wenn -pre_checksum_size nicht 0 ist. Diese",
"                   Vorschaugroesse muss bei filter_ und search_ gleich sein.",
" -search_md5 listenfile suchname altersgrenze jugendgrenze",
"                   wie -search_crc32 aber mit denjenigen Records, die MD5",
"                   Pruefsummen haben.",
" -set_shell_in_scdbackup_scripts shellname filename ... filename",
"                   eine besondere Option, die nur fuer die Konfiguration",
"                   verwendet wird. Die gegebenen Files werden geprueft,",
"                   ob sie mit \"#!\" beginnen. Wenn ja, wird der Rest der",
"                   Zeile durch shellname ersetzt. Es duerfen keine Optionen",
"                   mehr folgen, weil der Rest der Argumente als Liste von",
"                   Filenamen verstanden wird.",
" -shell_to_use progname  der progname wird nach \"#!\" in die erste Zeile",
"                   von Skripten gesetzt, die von diesem Programm erzeugt",
"                   werden. Grundeinstellung: /bin/sh",
" -special_file filenadresse groesse typ",
"                   meldet einen File als besondere Datenquelle an. Zur Zeit",
"                   ist nur der typ \"block_device\" zugelassen. Dazu muss",
"                   noch mit -split_large_files ein split_directory und ein",
"                   split_kommando angegeben werden. Der Dateninhalt eines so",
"                   angemeldeten Blockdevice wird dann mit der angegebenen",
"                   Groesse in Form von einem oder mehreren Teilstuecken in",
"                   den Backupplan aufgenommen. Ist groesse <=0, dann wird",
"                   sie durch Auslesen des Blockdevice automatisch bestimmt.",
"                   fileadresse darf die Form ziel=quelle haben.",
" -split_file_prefix  text",
"                   legt einen Text fest, der vor die Namen aller Stuecke",
"                   gesetzt wird, die durch -split_large_files entstehen.",
" -split_file_suffix  text",
"                   wie -split_file_prefix. Der Text wird aber zwischen Namen",
"                   und Zaehlnummern gesetzt. Siehe auch -split_list_adr .",
" -split_large_files split_directory split_groesse split_kommando sk_option",
"                   legt die Maximalgroesse fuer Datenfiles fest. Ebenso die",
"                   Adresse eines Pufferdirectories und eines Shellkommandos,",
"                   das ein Intervall aus einem File extrahieren kann. Dieses",
"                   Programm selbst kann mit sk_option -cut_file als solch",
"                   ein Shellkommando dienen. split_kommando unterliegt",
"                   der besonderen -bdir Ersetzung.",
"                   Files, die groesser sind als split_groesse, werden als",
"                   mehrere getrennte Stuecke behandelt. Shellskripte, die",
"                   von diesem programm ausgegeben werden, benutzen dann das",
"                   split_kommando, um die Stuecke im split_directory bereit",
"                   zu stellen. Alle Filestuecke eines Backupteils (einer CD)",
"                   muessen im split_directory Platz haben. Dort muss es einen",
"                   File namens scdbackup_split_file_directory_mark geben.",
"                   split_directory '-' schaltet diese Funktionalitaet ab.",
"                   split_directory '-xorriso_cut_out' nutzt Option -cut_out",
"                   des ISO 9660 Formatierprogramms xorriso. Dann ist kein",
"                   Directory noetig, split_kommando und sk_option werden",
"                   dann ignoriert.",
"                   split_groesse '-' bestellt die effektive -max_size Groesse",
"                   minus 1m. Hinweis: Die Option -exclude_size kann Files",
"                   bereits vor der Aufteilung in Intervalle ausschliessen.",
"                   split_kommando '-' laesst das vorher eingestellte Kommando",
"                   unveraendert. sk_option sollte dann auch '-' sein.",
" -split_list_adr fileaddresse",
"                   schreibt eine Liste der Filestuecke, die durch Option",
"                   -split_large_files entstehen. Die Information in diesem",
"                   File kann hilfreich sein, wenn ein Ausgabeformat wie",
"                   -form_path_list_afio gewaehlt wurde. Beim Wiederherstellen",
"                   solch eines Backups erscheinen die Stuecke im",
"                   split_directory und muessen an ihre Originalpositionen",
"                   gebracht werden.",
"                   Aufnahme in den Backup sind wie bei -complete_list_adr.",
"                   Jede Datenzeile beschreibt ein Filestueck in der Form",
"                     intervall stueck_adresse=puffer_adresse",
"                   intervall hat die Form",
"                     **stuecknummer/stueckgesamtzahl:start_byte+byte_zahl**",
"                   stueck_adresse besteht aus",
"                     orignal_dir\"/\"fix_name\"_\"stuecknummer\"_\"gesamtzahl",
"                   fix_name besteht aus:",
"                     prefix_text\"\"original_name\"\"suffix_text",
"                   puffer_adresse ist zusammengesetzt aus :",
"                     split_dir\"/\"basename_von_stueck_adresse\"-\"hexnummer",
"                   wobei hexnummer im gesamten Backup einmalig ist.",
" -status_file filename  Wenn der Filename nicht leer und nicht \"-\" ist,",
"                   dann wird das Ergebnisskript die aktuelle Stuecknummer in",
"                   diesen File schreiben, bevor es das -backup_cmd aufruft.",
"                   Es wird dann in der Lage sein, einen Backup am Anfang des",
"                   unterbrochenen Stueckes wieder aufzunehmen. Das geschieht,",
"                   wenn das Argument keine Stuecknummer sondern \"AUTO\" ist.",
" -target_escape modus  entscheidet, wie \"=\"-Zeichen in Zieladressen",
"                   behandelt werden. modus \"on\" bewirkt, dass \"=\" und",
"                   \"\\\" durch je ein \"\\\"-Zeichen gekennzeichnet werden.",
"                   Diese Zieladressen werden von modernen mkisofs-Versionen",
"                   verstanden, die Option -graft-points kennen.",
"                   modus \"off\" sorgt dafuer, dass in Zieladressen \"=\"",
"                   durch \"_\" ersetzt wird. \"\\\" bleibt ungeaendert.",
"                   Dies eignet sich fuer mkisofs 1.12b3.",
"                   Hinweis: Option -form_path_list setzt modus auf \"on\"",
"                            und -form_script setzt ihn auf \"off\". Danach",
"                            kann man ihn mit -target_escape ueberstimmen.",
" -target_separator text  das erste Zeichen des Textes wird als Trenn-",
"                   zeichen zwischen Zieladressen und den Fileadressen",
"                   eingestellt. Das wirkt nur auf Eingaben und nicht auf",
"                   Ausgaben des Programms, wo ggf. \"=\" verwendet wird.",
"                   Ein leerer Text schaltet die Zieladressen-Erkennung",
"                   ab. Grundeinstellung ist \"=\". In der Zieladresse",
"                   vorkommende Trennzeichen und \"\\\"-Zeichen muessen durch",
"                   ein vorangestelltes \"\\\" gekennzeichnet sein.",
"                   Beispiel: Bilde /home/buero auf das /-Directory der CD ab",
"                    -target_separator ':' /:/home/buero -target_separator ''",
"                   Beispiel: Bilde /home/test/=\\ ab auf /=\\",
"                     -target_separator '=' '/\\=\\\\'='/home/test/=\\'",
" -tmp_dir directory_adresse",
"                   bestimmt ein Directory, in dem temporaere Files abgelegt",
"                   werden koennen. U.a. wird es fuer Listenfileadressen",
"                   ohne '/' verwendet, die eigentlich in ein Leveldirectory",
"                   sollen, wenn kein -level gesetzt ist (-complete_list_adr).",
"                   directory_adresse unterliegt der besonderen -bdir",
"                   Ersetzung.",
" -tree_hashsize zahl   Option zur Abwaegung des Tempos der Filesuche in den",
"                   Directories von old_tree gegenueber dem Speicherbedarf",
"                   des old_tree. Jedes Directory in old_tree verbraucht",
"                   zahl*sizeof(int) Bytes fuer den Hashvektor. Die Suche",
"                   wird etwa um den Faktor zahl beschleunigt.",
"                   Die Grundeinstellung 20 ist gut fuer normalgrosse",
"                   Directories mit durchschnittlich 20 bis 1000 Files.",
"                   Directories mit zigtausend Files koennen aber sehr",
"                   bremsen und Baeume mit sehr vielen Directories koennen",
"                   den Arbeitsspeicher zu sehr belasten.",
" -vanish_list_adr  altlisten_adresse  resultat_adresse",
"                   erzeugt eine Schwundliste durch Vergleich mit den",
"                   Aufzeichnungen im File altlisten_adresse. Objekte, die",
"                   nicht mehr im Filesystem vorkommen, werden im File",
"                   resultat_adresse aufgelistet. Wenn eine -level Option",
"                   vorhanden ist und resultat_adresse kein \"/\" enthaelt,",
"                   dann wird der File im level_# Directory erzeugt (siehe",
"                   conf_dir). Wenn resultat_adresse mit .gz endet, dann wird",
"                   die Ausgabe durch das Programm gzip geleitet.",
"                   Wenn eine Zieladresse angegeben ist, dann wird der",
"                   Ergebnisfile zum letzten Stueck des Backups hinzugefuegt.",
"                   (zB: /schwund.gz=schwund.gz)",
"                   altlisten_adresse sollte ein File sein, der von",
"                   -complete_list_adr erzeugt wurde. Wenn mit -load_old_tree",
"                   ein Baum geladen wird, dann darf es auch eine Inhaltsliste",
"                   sein. Der Baum ersetzt dann die Komplettliste.",
"                   HINWEIS: altlisten_adresse von dieser Option ist identisch",
"                   mit altlisten_adresse von -load_old_tree. Wenn beide",
"                   Optionen angegeben sind, setzt sich die altlisten_adresse",
"                   der letzten durch.",
"                   WARNUNG: Ohne geladenen Baum oder bei -changetest_options",
"                   \"off\" kann die Schwundliste alte Namen von umbenannten",
"                   Files enthalten, obwohl deren neue Namen *nicht* im",
"                   inkrementellen Backup enthalten sind. Am besten ist es,",
"                   bei -changetest_options auch CRCs zu bestellen.",
" -vanish_script_adr  altlisten_adresse  result_address",
"                   WARNUNG: Das Ergebnisskript ist nicht ungefaehrlich !",
"                            Vor allem, wenn es ohne CRC-Test erzeugt wird.",
"                   Wie -vanish_list_adr , aber mit einem Shellskript als",
"                   Ergebnis, das Fileobjekte aus der Schwundliste loescht.",
"                   Solch eine Loeschung ist noetig, wenn man ein hoeheres",
"                   Level bei der Wiederherstellung des Backupgebiets ueber",
"                   das vorherige Level kopiert. Leider ist dieses Skript zu",
"                   hirnlos, um seine Opfer vorher genau zu pruefen.",
"                   HINWEIS: das Skript enthaelt nur die Zieladressen.",
" -verify_filelist  listenfile  oberdirectory",
"                   HINWEIS: Zum Pruefen von Files auf CD oder waehrend des",
"                            Wiederherstellens eignet sich -verify_targetlist",
"                            besser.",
"                   Das Programm endet sofort wenn listenfile ausgelesen ist.",
"                   Aus listenfile (\"-\" = stdin) werden Fileadressen und",
"                   ggf. auch Typ, Groesse und Pruefsumme gelesen.",
"                   Diese Angaben werden mit dem Ist-Zustand der Fileadresse",
"                   verglichen. Dabei wird oberdirectory jeder Adresse",
"                   vorangestellt. Existiert die Adresse und passen die",
"                   anderen Angaben, dann wird die Fileadresse einschliesslich",
"                   oberdirectory auf stdout ausgegeben.",
"                   Ansonsten geht eine Fehlermeldung an stderr.",
"                   War der gesamte Programmlauf fehlerfrei, dann ist der",
"                   Exitwert 0, sonst 1.",
" -verify_sources  listenfile",
"                   Das Programm endet sofort wenn listenfile ausgelesen ist.",
"                   Diese Option arbeitet fast wie -verify_filelist. Es werden",
"                   aber Adressen der Form  ziel=quelle  erwartet und nur die",
"                   Quelle wird ueberprueft. Ein Oberdirectory wird nicht",
"                   vorangestellt.",
" -verify_source_filter  shell_filter",
"                   Im Ausgabeformat -form_path_list_mkisofs koennen aus der",
"                   Fileliste jeder CD unmittelbar vor der der Aufforderung,",
"                   die CD einzulegen, solche Files entfernt werden, die",
"                   der shell_filter nicht durchlaesst. Will man zB. alle",
"                   Files weglassen, die seit der Planung von der Platte",
"                   verschwunden sind, ist ein geeigneter Filter :",
"                     ..../cd_backup_planer -verify_sources - ",
"                   Ein Leertext als shell_filter schaltet diese Pruefung ab.",
"                   shell_filter unterliegt der besonderen -bdir Ersetzung.",
" -verify_targetlist  listenfile  oberdirectory",
"                   wie -verify_filelist. Jedoch werden die Zieladressen",
"                   verwendet, was einen Unterschied macht, wenn beim Backup",
"                   mit ziel=quelle Umleitungen bestellt wurden.",
" -volume_info      Uebergibt an das Backup-Kommando vor den Fileadressen",
"                   zusaetzliche Argumente:  -v Stuecknummer -t Stueckzahl",
" -volume_offset  zahl",
"                   Addiert die Zahl auf alle Ausgaben von Backupstueck-",
"                   Nummern und zieht sie von Eingaben ab.",
" -working_directory  directory_adresse",
"                   stellt das aktuelle Directory auf die angegebene Adresse.",
"                   Diese Option sollte (wenn ueberhaupt) nur einmal vor der",
"                   ersten Fileadresse angegeben werden.",
" -write_random  ziel menge zufallssaat",
"                   Das Programm endet sofort wenn diese Option erledigt ist.",
"                   Sie schreibt in den File ziel (\"-\"=stdout) die bestellte",
"                   menge von Pseudozufallszahlen oder aber ein periodisches",
"                   Muster von bis zu 16 Bytes. Beginnt zufallssaat mit einem",
"                   \"=\" dann ist der weitere Text das periodische Muster.",
"                   Folgt auf \"=\" das Zeichen \"~\", dann wird das binaere",
"                   Komplement des Textes als periodisches Muster verwendet.",
"                   Ansonsten wird zufallssaat als Schluessel fuer einen auf",
"                   MD5 und EXOR basierenden Zufallsgenerator benutzt.",
"                   Beginnt dieser Schluessel mit \"-\" dann ist der Rest",
"                   der Schluessel fuer ein langsames, EXOR-freies Verfahren.",
"                   Die Zahlenfolge jeder zufallssaat ist reproduzierbar.",
"                   Bei menge sind grosse Werte wie 100g oder 2t zulaessig,",
"                   sofern das Betriebssystem damit umgehen kann.",
" -write_to_file zielfile byteadresse bytezahl bytecode [... bytecode]",
" -write_to_file zielfile byteadresse \"string\" textstueck",
" -write_to_file zielfile byteadresse \"file\" quellfile",
"                   Das Programm endet sofort wenn dieser Befehl abgearbeitet",
"                   ist. Es oeffnet den existierenden File zielfile zum",
"                   Schreiben und adressiert Byte byteadresse (Start = 0).",
"                   Ist das dritte Argument eine Zahl zwischen 1 und 4096,",
"                   dann ist das die Zahl der folgenden dezimalen Bytecodes.",
"                   Erlaubt sind also Bytewerte zwischen 0 und 255.",
"                   Bei \"string\" wird das vierte Argument in den zielfile",
"                   uebertragen. Bei \"file\", ist das vierte Argument die",
"                   Adresse eines existierenden Files, der dann ganz in",
"                   zielfile kopiert wird.",
"                   zielfile \"-\" bedeutet stdout und ist nur mit",
"                   byteadresse 0 erlaubt. quellfile \"-\" ist stdin.",
" -xdev             alle Filesysteme auslassen, die unterhalb der Directories",
"                   in der Backupliste gemountet sind. Gegenteil: -alldev .",
" -yes              Die folgenden Fileadressen werden in die Backupliste",
"                   aufgenommen. Sie werden also noetigenfalls zerlegt und",
"                   dem Backupkommando uebergeben. Diese Option beendet",
"                   die Wirkung von \"-no\". Sie ist die Grundeinstellung.",
"                   Wenn die Option -yes mindestens einmal vorkommt, dann",
"                   werden die Fileadressen der Backupliste auf jeden Fall",
"                   ins Ergebnis uebernommen, auch wenn eine gleichlautende",
"                   in der Ausschlussliste vorkommt.",
"                    ( -yes file -not file  bringt file in den Backup",
"                      file -not file       laesst file aus) ",
" -younger_than zeitangabe  Es wird versucht, moeglichst nur Files in den",
"                   Backup aufzunehmem, deren Inhalt oder inode nach dem",
"                   angegebenen Zeitpunkt geaendert wurde. Bewegungen im",
"                   selben Filesystem bleiben leider unberuecksichtigt !",
"                   Die Zeitangaben koennen folgende Form haben :",
"                   Absolute Zeitpunkte koennen durch das Eingabeformat des",
"                   des UNIX-Befehls 'date' MMDDhhmm[[CC]YY][.ss]] festgelegt",
"                   werden. Zum Beispiel :",
"                     020616122000 heisst '6. Februar 2000 16:12:00 Ortszeit'",
"                   Ausserdem wird das Format YYMMDD[.hhmm[ss]] verstanden:",
"                     2.Jan 2000 = A00102 , 11.Mai 2001 23:45 = A10511.2345",
"                     1.Mrz 2010 = B00301 , 14.Jul 1998 = 980714",
"                   Relativangaben beginnen mit '-' oder '+'. Sie werden zum",
"                   aktuellen Zeitpunkt addiert. '-' zeigt also in die",
"                   Vergangenheit, '+' in die Zukunft. Die Grundeinheit ist",
"                   die Sekunde. Sie ist durch einen Buchstaben am Ende der",
"                   Zahl aenderbar:",
"                     Sekunden 1s=1 , Stunden 1h=3600s , Tage 1d=24h ,",
"                     Wochen 1w=7d , Monate 1m=31d, Jahre y : Xy=X*365.25d+1d",
"                   Bitte beachten: 'm' und 'y' sind oft etwas groesser als",
"                   der genaue Kalenderzeitraum.",
"                   Beispiel:",
"                     -3w  heisst 'vor 3 Wochen' = -1814400s .",
"",
"Beispiele:",
"",
"Den Inhalt des Arbeitsdirectory in Stuecke zu 625 MB und maximal 200",
"einzelne Adressen zerteilen. Das Infoskript ASKME im Arbeitsdirectory",
"erzeugen und in jedes einzelne Stueck des Backup aufnehmen.",
"",
"  %s \\",
"    -max_items 200 -max_size 625m -info_script ASKME . >bskript",
"  chmod u+x bskript",
"",
"Um den Backup dann wirklich durchzufuehren :",
"  bskript",
"",
"",
"Mit der folgenden Option wird das Infoskript an einer separaten Stelle",
"erzeugt und in das Rootdirectory der CD abgebildet.",
"  -info_script /ASKME=/home/backup_planer_dir/ASKME",
"",
"",
"Zerteile die Directories  \"buero\" und \"bilder\" in Stuecke von",
"Standardgroesse (600 MB / 400 Adressen). Nicht die Unterdirectories",
"\"buero/privat\" und \"bilder/freundinnen\" aufnehmen.",
"Anstelle von \"bilder/freundinnen\" eine zweite Kopie von \"bilder/katzen\"",
"einfuegen.",
"",
"  %s \\",
"    buero bilder -target_separator '=' /bilder/freundinnen=bilder/katzen \\",
"    -not buero/privat bilder/freundinnen \\",
"  >bskript",
"",
"",
"DC-6250 Baender mit komprimierten tar-Archiven fuellen.",
"",
"  %s \\",
"    -backup_cmd 'read dummy ; tar cvzf /dev/rmt0' \\",
"    -form_script_raw -max_size 245m -max_items 100 \\",
"    -info_script ASKME \\",
"    . \\",
"  >bskript",
"",
"",
"Eine Konfiguration fuer inkrementelle Backups erzeugen und den ersten Backup",
"-level 0 planen lassen. Durch -level -create_configuration wird ein neues",
"Directory angelegt, das (wegen -conf_dir) $HOME/my_backup_conf heisst.",
"",
"  %s \\",
"    -conf_dir $HOME/my_backup_conf -level -create_configuration \\",
"    ... alle Argumente, die den vollen Backup definieren ... \\",
"  >bskript",
"",
"Diese Konfiguration kann man nutzen, indem man einfach das Directory und die",
"gewuenschte Ebene angibt. Die anderen Parameter des Backups werden aus dem",
"-conf_dir gelesen. (File specific_arguments : eine Zeile = ein Argument)",
"",
"Man sollte sich ein festes Schema fuer Backups auferlegen und durchhalten.",
"Angenommen, wir haetten am Sonntag der 0. Woche den Backup konfiguriert",
"und also auch den ersten -level 0 Backup gemacht :",
"",
"Ein Schema, das von Montag bis Samstag nur die Aenderungen des Tages",
"aufzeichnet und am Sonntag einen vollen Backup vorsieht.",
"  Montag:   -conf_dir $HOME/my_backup_conf -level -next",
"  ...",
"  Samstag:  -conf_dir $HOME/my_backup_conf -level -next",
"  Sonntag:  -conf_dir $HOME/my_backup_conf -level 0",
"Nun koennen die Datentraeger von Montag bis Samstag recycelt werden.",
"",
"Ein Schema, das etwa vierteljaehrlich volle Backups macht, sowie nach vier",
"Wochen und nach jeder einzelnen Woche auffrischt. (-conf_dir wie oben)",
"Dieses Schema macht Aenderungen an Files ueber laengere Zeit nachvollziehbar.",
"  1. Woche :  -level 1",
"  2. Woche :  -level 2",
"  3. Woche :  -level 2",
"  4. Woche :  -level 2",
"  5. Woche :  -level 1",
"  6. Woche :  -level 2",
"  7. Woche :  -level 2",
"  8. Woche :  -level 2",
"  9. Woche :  -level 1",
" 10. Woche :  -level 2",
" 11. Woche :  -level 2",
" 12. Woche :  -level 2",
" 13. Woche :  -level 0",
"Danach beginnt man wieder bei Woche 1.",
"Der Backup von Woche 12 sichert auch die Wochen 11 und 10 ab.",
"Der Backup von Woche 9 sichert Wochen 8 bis 1 ab.",
"Die Levels 0 und 1 sollte man laenger aufheben.",
"",
"Um beide gezeigte Schemata zu kombinieren, sollte man im ersten Schema den",
"Sonntagslauf durch den woechentlichen Lauf des zweiten Schemas ersetzen.",
"",
"Wenn Sie je einen inkrementellen Backup mit mehreren Levels wiederherstellen",
"muessen, beginnen Sie mit Level 0. Nachdem das auf der Platte ist, muessen",
"die Files geloescht werden, die in der Woche bis zum Level 1 verschwunden",
"sind.",
"Zu diesem Zweck finden Sie im letzten Backupstueck jedes Levels einen File",
"namens /tmp/VANISHED.sh.gz . Er hat keine x-Rechte, um ueble Unfaelle zu",
"vermeiden. Man kann ihm nur trauen, wenn er durch Vergleich mit einem",
"aufgezeichneten Speicherbaum unter Verwendung von CRCs oder inode-Nummern",
"erstellt wurde. Das ist zwar Grundeinstellung, aber die kann geaendert werden",
"mit Optionen wie: -content_list_adr , -list_output_format , -load_old_tree ,",
"-changetest_options , -vanish_script_adr ",
"Besonders misstrauische Datenbesitzer verwenden bei -list_output_format die",
"Option \"md5\" statt \"crc\".",
"",
"Man kann das Skript auf die Platte kopieren, es mit gunzip auspacken und",
"ihm mit  chmod u+x  Ausfuehrungserlaubnis erteilen und es dann starten.",
"Man kann es aber auch mit dieser Shellpipe direkt ausfuehren :",
"  gunzip </cdrom/tmp/VANISHED.sh.gz | /bin/sh",
"Man sollte aber vorher auf jeden Fall einen Blick auf seinen Inhalt werfen",
"(mit \"less\" statt \"/bin/sh\"). Letzte Chance diese Files zu begnadigen.",
"",
"Das VANISHED.sh Skript von Level 1 kann irgendwann nach der Installation von",
"Level 0 und vor der Installation von Level 2 ausgefuehrt werden. Wenn",
"Plattenplatz knapp ist, macht man es besser vor der Installation von Level 1.",
"Allerdings ist es etwas sicherer, das Skript auszufuehren, nachdem Level 1",
"auf die Platte kopiert wurde. Man hat dann eher zuviele als zuwenig Files,",
"wenn man unterbrochen wird.",
"Wenn man dann die Levels 2,3,4,... installiert, darf man nicht vergessen,",
"auch jeweils deren VANISHED.sh Skripte rechtzeitig auszufuehren. Eventuell",
"geraten Sie auch beim Installieren auf die Platte und sollten dort besser",
"wieder geloescht werden.",
"",
"",

#endif /* ! Cdbbackup_planer_no_germaN */

"",
"@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@",
"                     dirscript_size  Groessen als Kommentartexte ins Ergebnis",
"                           von -dir_script_adr schreiben. zB.: #size 4096",
""
 };
 char line[4096];
 FILE *fp;
 int i;
 
 fp= stdout;
 if(flag&1)
   fp= stderr;
 MELDEN("");
 sprintf(line,"cd_backup_planer Version %s",PROG_VERSION);
 MELDEN(line);
 MELDEN("");
 for(i=0;strcmp(hilfetext[i],
                "@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@"
                )!=0;i++){
   sprintf(line,hilfetext[i],progname);
   MELDEN(line);
 }
 return(1);
}


int Backup_plan_help(progname,flag)
char *progname;
int flag;
/*
 bit0= to stderr rather than stdout
*/
{
 static char hilfetext[][80]={

#ifdef Cdbbackup_planer_no_helP

"This binary program does not contain a help text.",
"Read doc/cd_backup_planer_help if available. Else install scdbackup",
"without executing ./DIET afterwards.",

#else 

"usage :  %s  [option | file] ...",
"",
"Usually this command splits a list of directories into volumes",
"which fit on CD-R media.",
"It generates a script with backup commands. The lists of files",
"or directories are determined according to maximum capacities",
"in terms of total size and number of items.",
"",
"A result list can contain only directories and datafiles.",
"Other file types may be included in directories but most CD",
"writer programs will not process them.",
"",
"The result lists consist of target and source addresses suitable",
"for the program  mkisofs  in its more modern versions:",
"  If fileaddress is a directory:  /targetaddress/=fileaddress",
"  If it is a data file:           /targetaddress=fileaddress",
"These lists are appended to the backup command in order to fill",
"one CD medium. If not given explicitely and not redirected by a",
"collision, the targetaddress is the same as the fileaddress.",
"Depending on the -target_escape mode the character \"=\" either",
"is escaped or changed to \"_\".",
"Other output forms of the result lists are available too.",
"",
"Since the working directory, the root directory and possibly",
"the -main_dir all are mapped to the root directory collisions",
"of targetadresses may occur. A builtin collision test can detect",
"and avoid them by renaming one of the affected targetadresses.",
"",
"A filecheck can be made to find problematic files:",
" - those which are ignored by mkisofs-1.12b3 if not given",
"   explicitely as arguments.",
" - those which cannot be read (files or directories)",
"   this can mislead the estimation of resulting directory sizes.",
"The combined options -filecheck -append -filecheck -exclude_unreadable",
"can help to work around these problems.",
"",
"Incremental backups are possible with an arbitrary number of levels",
" -level 0 does a full backup with no time constraints.",
" -level N backups those files that have changed since the last run",
"  of -level N-1 . -level N-1 has to exist already.",
" - A backup with -level N makes invalid all levels above N.",
"The data necessary for keeping track of levels are kept in a",
"dedicated directory which is addressed by -conf_dir. It also hosts",
"all parameters of the backup like filelist or maximum sizes.",
"",
"Incremental backups use the option -younger_than which possibly",
"also includes some older files in the backup. Regrettably, mere",
"name changes do not change the file's time stamps in the general",
"Linux filesystem model (other than e.g.:  chmod u+r ). Therefore",
"lengthy checksum computations often become unavoidable.",
"",
"Since a large number of single filenames could make problems with",
"older versions of mkisofs, the incremental backup mainly tries to",
"exclude directories which contain only very few young files and to",
"mention only these few exemption files. Also a certain effectivity",
"is required in terms of byte gain versus number of filenames. This",
"required quote is high if -max_items is set to a small value.",
"",
"An additional script can be generated which is able to print",
"several informations about the backup.",
"Special options are:",
"-hilfe, -id, -count, -list, -levelall , -suball, -fast, -help",
"Other arguments are looked up as file names in the lists.",
"If a name or a directory above this name is found, then the part",
"number, the item and the file name are printed. If none of the",
"files can be found this way, then the script's exit value is 1.",
"",
"This script is stored on the disk and included in each part of",
"backup if a name is given with option -info_script (see below).",
"It is estimated to have 100k size (in case that your scripts",
"actually exceed this size reduce -max_size accordingly).",
"If the info script gets included and the output form for mkisofs",
"(default or -form_script) is used then all possible fileadresses",
"which may lead to the same name on the CD are excluded.",
"",
"A number as optional argument tells the backup script to skip all",
"volumes with lower numbers. To continue a backup which has been",
"interrupted while volume 3 was ready to be written:",
"  scriptname 3",
"Then the script skips volumes 1 and 2.",
"If the script has been generated with -status_file then it can find",
"the appropriate volume automatically by reading the status file",
"This automatic is triggered with the argument \"AUTO\".",
"",
"Arguments which do not begin with '-' are file addresses.",
"Fileadresses may be prefixed by target adresses. This is useful",
"for composing a CD from various scattered sources. The usual",
"separating character is \"=\" but another one may be chosen",
"with the option -target_separator.",
"Example: -target_separator '=' /images=/home/private/images ...",
"",
"All fileaddresses together define the \"backup area\" from which some",
"sub trees may be excluded by options -no, -not, -xdev, -localdev .",
"This set of file objects may get examined and reduced further",
"(e.g. by -younger_than) and so becomes the \"backup content\".",
"",
"Options:",
"",
"Several options (like -max_size) expect size arguments. This may be integer",
"numbers (like 1000000000) or floating point numbers (like 1.25e9). They may",
"have appended a binary unit letter out of {b,k,m,g,t,p,e} which multiplies",
"the number by {1,1024,1024*1024,1024*1024*1024,...,pow(1024.0,6.0)}.",
"Example: 640m = 640*1024*1024 = 671088640",
" ",
" -help             print this text and exit program",
" -hilfe            diesen Text ausgeben und Programm beenden",
"                   (like -help , use german language)",
" -version          print version identification on stdout,",
"                   if not only argument: print warning message on stderr,",
"                   in any case : end program with exit value 4.",
" --                the next argument is a file, not an option.",
"",
" -accept_all_filetypes",
"                   Is equivalent to :  -result_filetypes any",
" -alldev           do not exclude mounted filesystems. (default)",
" -backup_cmd text  the given text is used as command in the script generated",
"                   by this program. text is subject to special -bdir",
"                   substitution. (default: cd_backup_kommando)",
" -bd               set size paramters suitable for single layer Blu-ray",
"                   media of nominal \"25 GB\" with medium size formatting.",
" -bdir directory   sets the scdbackup installation directory. Currently the",
"                   only use of this information is -bdir substitution. It",
"                   is performed when reading file specific_arguments of an",
"                   incremental backup configuration. If such an argument",
"                   starts with @SCDBACKUP_DIR@ then this start piece gets",
"                   replaced by the -bdir directory. This works too if a",
"                   quotation mark is prepended.",
"                   With some very few arguments the start piece is replaced",
"                   also if it matches grep expression",
"                     .*/scdbackup-[0-9]\\.[0-9]\\.[0-9]/.* ",
"                   This is called special -bdir substitution.",
" -bd_dl            set size paramters suitable for double layer Blu-ray",
"                   media of nominal \"50 GB\" with medium size formatting.",
" -blocklist_fileadr address_prefix",
"                   if address_prefix is not empty and not \"-\" then it will",
"                   be used to generate addresses of block checksum lists",
"                   which are eventually to be stored. Wether there are such",
"                   lists depends on:",
"                     -filter_checksum_blocksize with -filter_md5",
"                     -scan_for_checksum -blocklist with -search_md5",
"                   The effective file address will consist of address_prefix,",
"                   record name, date stamp and size value of the checksum",
"                   record. For example : \"/tmp/x_\" may lead to:",
"                     /tmp/x_1_1_A41123.194219_229208064",
" -changetest_options  code:code:...",
"                   chooses properties to be compared during check",
"                   -younger_than. Compared are the file objects on",
"                   the disk with those obtained by -load_old_tree.",
"                   Multiple codes are separated by \":\". ",
"                     type  e.g. directory, data file, softlink",
"                     size  bytecount",
"                     crc   CRC-32 or MD5 content checksum",
"                     inode internal ID number of filesystem object",
"                     off   disable comparison with old tree",
"                     adr   enable comparison with old tree (default)",
"                     mistrust    avoid a shortcut with CRC depending on",
"                                 directory timestamps.",
"                     dir_trust   allows an even faster shortcut with CRC but",
"                                 is not safe if directories get renamed",
"                                 in order to replace existing directories.",
"                     assume_old  assume all timestamps as old enough",
"                     ignore_ctime  inspect only mtime but not ctime.",
"                     ignore_device with inode: do not compare device numbers",
"                     no_split_economy  no check for uneconomical directory",
"                                       splittings by small, old files.",
"                   NOTE: -load_old_tree with nonempty addresses",
"                   sets -changetest_options to \"type:size:crc:adr\"",
"                   Set own -changetest_options after -load_old_tree",
"                   If both :inode: and :crc: are chosen, and both information",
"                   is available, then :inode: will be used for decision.",
" -collision_test mode  controls the collision test for target names",
"                   Modes can be (default is -redirect ) :",
"                     -off       do not perform collision test",
"                     -on        warn about collisions",
"                     -redirect  change target names if necessary",
"                     -abort     abort if a collision is detected",
" -compare_checksum recordlist searchname old_limit young_limit",
"                   the program ends immediately when stdin closes.",
"                   It passes stdin to stdout without changes.",
"                   Reads from file recordlist a record matching the",
"                   other three arguments. The record may have been",
"                   written by -filter_crc32 or -filter_md5. In order",
"                   to match, searchname has to be \"-\" or exactly",
"                   the same as the name of the record. If old_limit",
"                   is not \"-\" all older records do not match. If",
"                   young_limit is not \"-\" then all younger ones do",
"                   not match. (see -younger_than for time formats)",
"                   The last matching record in the list is read and",
"                   its recorded data amount is read from stdin to",
"                   compute a checksum which is compared with the",
"                   recorded checksum. If stdin delivers enough",
"                   bytes and the checksums are equal, then the",
"                   return value of the program is 0. In case of",
"                   not enough input or wrong checksum exit is 1.",
"                   Any surplus bytes at stdin are ignored.",
" -complete_list_adr fileaddress",
"                   writes a complete file list of the backup area",
"                   to the given fileaddress. If fileaddress contains no \"/\"",
"                   and a -level is given then the file is created in the",
"                   level_# directory. (See below : -conf_dir) If no -level",
"                   is given then such a file goes to -tmp_dir.",
"                   If the given fileaddress ends with .gz then the",
"                   output gets piped through program gzip.",
"                   If a target is given in fileaddress then the",
"                   resulting file will be appended to the last",
"                   backup volume. (e.g. /complete.gz=complete.gz)",
"                   The list is written in a special format to save",
"                   disk space while still being human readable. The",
"                   first and the last characters of an address determine",
"                   its type",
"                     /x/x/x/   absolute directory address",
"                     /x/x/x    absolute file address",
"                     ./x/      descent into directory x",
"                     ./x       file x in current directory",
"                     ..        one directory up",
"                     ../       to start directory",
"                     #=# xxxx  additional information",
"                   The first directory address given is absolute and",
"                   sets the start directory for relative addresses.",
"                   By ./x/ and /x/x/x the directory cursor can be",
"                   moved away from the start directory. As long as",
"                   the cursor is in the start directory, addresses",
"                   get prepended a target address which also is",
"                   valid for all relative addresses below.",
"                   Both parts are separated by a \"=\"-character. If the",
"                   target address itself contains \"=\" or \"\\\" then they",
"                   are prepended by a \"\\\" each.",
"                   The resulting list is usable with -vanish_list_adr",
" -composition_memory  depth width memory_limit",
"                   defines size and rules for an acceleration buffer which",
"                   may help with dividing very large and deep directory",
"                   trees into digestible pieces. Parameter  depth  controls",
"                   how many levels of sub directories shall be recorded for",
"                   the case that their sizes is requested again after a",
"                   directory split. If they are recorded then there is no",
"                   to examine the subtrees on disk again. Thus the buffer",
"                   can save a lot of time but can also need lots of memory.",
"                   Parameter  width  controls how many files a directory may",
"                   contain to be recorded in the buffer.",
"                   The memory consumption per file depends mainly on the",
"                   lenght of the filenames. Plus about 20 Bytes. Total",
"                   memory consumption is curbed by parameter memory_limit.",
"                   If memory_limit is preceeded by \"-\" then freed buffer",
"                   is not regarded as reusable.",
"                   Default : 3 1000 4m     (0 0 0 would disable the buffer)",
" -composition_mode criterion",
"                   defines the sequence of equal ranked file objects when",
"                   backup volumes are composed. All modes try to keep",
"                   directory trees together. When arranging objects from the",
"                   start list or objects within split directories there",
"                   may be used one of several criteria to choose the",
"                   next file object :",
"                    bestfit   the largest object that will fit",
"                    target    the next higher target address (alphabetically)",
"                    source    the next higher source address",
"                    time      the next younger file object",
"                    crawler   the next one found by tree crawler",
"                    default        currently like bestfit",
"                    alphabetical   currently like target",
"                    chronological  currently like time",
"                    fast           currently like crawler",
"                   With criteria other than bestfit there is an increased",
"                   chance to get larger directories split over two adjacent",
"                   backup volumes. This probability is mainly controled by",
"                   -min_size_percent where a low percentage causes waste of",
"                   media capacity and high percentage causes more directory",
"                   splits. Waste will be prevented by splitting only if the",
"                   number of -max_items is large enough to accept nearly all",
"                   Fileobjects in the split directory.",
"                   Criterion crawler yields results like criterion source",
"                   as long as the directories have a moderate size. At volume",
"                   borders it always splits directories. It avoids one or",
"                   two passes over the backup area which may need substantial",
"                   time if RAM cannot buffer the whole backup area tree",
"                   information.",
" -content_list_adr fileaddress",
"                   writes a file list of the backup content. With",
"                   the same rules and formats as -complete_list_adr",
"                   this option lists only those file objects which",
"                   are part of a backup volume. The content lists",
"                   and vanish lists of all levels together contain",
"                   the same information as a single complete list.",
"                   Additionally a content list tells the volume",
"                   numbers by marking volume borders.",
" -conf_dir directoryaddress",
"                   sets the location of a backup configuration. It",
"                   contains files which are created and used by",
"                   incremental backups. (see -level)",
"                     specific_arguments : full backup's options list",
"                                          each line is one argument.",
"                     highest_level      : highest valid level number",
"                     ASKME_level_#      : info script of -level #",
"                     level_#/           : level directory, contains",
"                       last_run_date    : time stamp of last run",
"                       number_of_volumes : as it says",
"                       content_file_list.gz : see -content_list_adr",
"                                              and -load_old_tree",
"                       vanished_file_list.gz: see -vanish_list_adr",
"                       VANISHED.sh.gz       : see -vanish_script_adr",
" -cut_file source_address start_byte byte_count target_address",
"                   The program ends immediately after this option is",
"                   done. It copies an interval of bytes out of a file into",
"                   a new one. Intended to be used by -split_large_files .",
"                   If source_address is '-' then data are read from stdin",
"                   and a number of start_byte bytes is skipped, else the",
"                   source_address is opened as file and the read pointer is",
"                   positioned at start_byte. target_address may be '-' for",
"                   stdout or a suitable address for a new file. The",
"                   intermediate messages are controlled by -pacifier .",
" -debug_filelist list_adr  directory_prefix",
"                   The program ends immediately after this option is done.",
"                   It prints the list lines together with eventual resulting",
"                   absolute file addresses. directory_prefix is prepended to",
"                   any of these addresses. List lines get prepended three",
"                   blanks and the count of already listed file items.",
"                   So they may get filtered away by  grep -v '^   '",
" -debug_stderr_log     The program ends immediately after stdin closes.",
"                   It changes carriage returns into newlines and tries to",
"                   spare empty lines. Intended for making more readable",
"                   the results of    scdbackup ... 2>&1 | tee -i ...",
" -debug_targetlist list_adr  directory_prefix",
"                   like -debug_filelist but showing the target addresses",
"                   of the files.",
" -delete_filelist  list_adr  directory_prefix",
"                   NOTE: for removing obsolete files during the restore of",
"                         a backup you should better use -delete_targetlist.",
"                   The program ends immediately after this option is",
"                   done. It reads file addresses from list_adr",
"                   together with type, size, checksum properties.",
"                   These informations get compared with the current",
"                   state of the file whose address is composed from",
"                   directory_prefix and the address read from the",
"                   list.",
"                   The file list has to be generated with ",
"                    -vanish_list_adr and must contain type, size",
"                   and checksum informations.",
"                   If the addressed file exists and if the other",
"                   properties match, then the file gets deleted.",
"                   Directories get deleted if they are empty.",
"                   If everything existed and matched, the program",
"                   exits with value 0. Else this value is 1.",
" -delete_targetlist  list_adr  directory_prefix",
"                   like -delete_filelist but processing the target addresses",
"                   of the vanished files. If there have been target=source",
"                   redirections when a backup was made, then this can make",
"                   a big difference.",
" -dir_list_adr fileaddress",
"                   writes a complete directory list of the backup area to",
"                   the given fileaddress. The output format is similar to",
"                   -complete_list_adr but the following fields are fixely",
"                   enabled: \"type:chmod:chown:chgrp:date:size:crc:adr\" with",
"                   chmod:chown:chgrp:date defining access permissions,",
"                   ownership and time stamp. If MD5 checksums have been",
"                   enabled by -list_output_format then they are used here",
"                   rather than CRC32 checksums.",
" -dir_script_adr fileaddress",
"                   writes a shell script for restoring the access permissions",
"                   of the directories of the backup area. The script will use",
"                   target addresses relative to the CDs' root directory.",
" -dir_size_cmd text the given text is used as command for determination of",
"                   a directory's size. The command has to expect a single",
"                   argument (the adress of the directory) and give back as",
"                   first word on standard output the size in terms of",
"                   kilobytes (like  du -sk  does).",
"                   Default:  \"-internal\" does not run an external command",
"                   but recursivly sums up the sizes of file objects and the",
"                   length of their addresses. Added is an overhead of 256",
"                   bytes per file. It may be changed by \"-internal+$num\" .",
"                   For example \"-internal+10\" adds 10 extra bytes.",
"                   \"-internal-$num\" disables adding of address length.",
" -disable_safe_leveling  is equivalent to ",
"                   -load_old_tree \"\" \"\"",
"                   -content_list_adr \"\" -complete_list_adr \"\"",
"                   -vanish_script_adr \"\" \"\" -vanish_list_adr \"\" \"\"",
" -dvd              set size paramters suitable for single layer DVD media",
"                   of nominal \"4.7 GB\". Not suitable for DVD-RAM.",
" -dvd_dl           set size paramters suitable for double layer DVD media",
"                   of nominal \"8.5 GB\".",
" -dvd_ram          set size paramters suitable for DVD-RAM  media",
"                   of nominal \"4.7 GB\" with medium size formatting.",
" -exclude_size sizelimit  exclude any single file which contains at least",
"                   sizelimit bytes. A sizelimit of 0 disables this test.",
"                   If sizelimit begins with \"-\" then with symbolic links",
"                   the size of the link target gets tested.",
"                   'k' 'm' and 'g' are recognized like with -max_size .",
" -extract_multiplied source skip",
"                   the program ends immediately when stdin closes. It",
"                   reads source sequentially and expects its data to stem",
"                   from option -multiply_volumes. The given number of copies",
"                   get skipped and the next one gets written to stdout.",
"                   Eventually a defunct marker tag is appended.",
" -extract_tag_tail maximum_size",
"                   the program ends immediately when stdin closes. It",
"                   consumes stdin until the header of a potentially valid",
"                   checksum tag is encountered, which is preceeded by at",
"                   least one byte. Beginning with this tag header, the rest",
"                   of the stdin stream is passed to stdout. Different from",
"                   -search_md5 even damaged tags will be accepted.",
"                   The passed amount of data will consist of the header plus",
"                   at most the amount of data given by maximum_size. Use a",
"                   generous maximum_size to catch all interesting data.",
"                   The exit value is 0 if data have been extracted. It is",
"                   non 0 else.", 
" -extract_tag_tail_dislocated maximum_size",
"                   like -extract_tag_tail but rather accepting the first",
"                   occurence of a checksum tag identifier text. It is not",
"                   necessary that the position information of the tag",
"                   matches the position of the identifier text within the",
"                   data stream. Therefore it is possible to retrieve tags",
"                   from data streams where bytes are missing. On the other",
"                   hand this command accepts checksum tags which are part of",
"                   the payload data (which is not too uncommon). Only tags",
"                   which start at position 0 in the data stream get skipped.",
" -facl_list_adr  fileaddress",
"                   applies program getfacl to the backup area and writes",
"                   its output to the given fileaddress. This will record",
"                   standard file permissions as well as eventual ACL",
"                   settings. The format is suitable for program setfacl",
"                     gunzip <facl_list.gz | setfacl --restore=-",
"                   The adresses recorded within the facl file are changed",
"                   to the target adresses of the backup items. Example :",
"                     -facl_list_adr /extra_lists/facl_list.gz=facl_list.gz",
"                   See man getfacl, setfacl, mount (option \"acl\").",
"                   -level, -tmp_dir and gzip compression apply like",
"                   described with -complete_list_adr.",
" -faclsrc_list_adr  fileaddress",
"                   same as -facl_list_adr but recording the original source",
"                   addresses as on disk. Both options may be used (with",
"                   different fileaddresses) to produce both lists in the",
"                   same backup.",
" -fattr_list_adr  fileaddress",
"                   applies program getfattr to the backup area and writes",
"                   its output to the given fileaddress. This will record",
"                   XFS style extended file attributes as far as they are",
"                   enabled and present. Recorded addresses get changed to",
"                   the according target addresses.",
"                   See man getfattr, setfattr, mount (option \"user_xattr\")",
"                   -level, -tmp_dir and gzip compression apply like",
"                   described with -complete_list_adr.",
" -fattrsrc_list_adr  fileaddress",
"                   same as -fattr_list_adr but recording the original source",
"                   addresses as on disk. Both options may be used (with",
"                   different fileaddresses) to produce both lists in the",
"                   same backup.",
" -filecheck mode   controls the examination of all files.",
"                   Modes can be (default is -off ) :",
"                    -off     no examination (saves much time)",
"                    -on      warn about -implicite_ignored files",
"                             and unreadable files",
"                    -append  like -on. Append -implicite_ignored",
"                             files explicitely to the backup list.",
"                    -append_hexcode  like -append.",
"                             The problem causing characters are",
"                             replaced by +XY+ (XY=ASCII code in hex)",
"                    -exclude same as -off (due to a long existent bug)",
"                    -exclude_unreadable  like -on. Exclude unreadable",
"                             files explicitely from the backup list.",
"                    -abort   like -on. Abort program if warnings",
"                             have been issued.",
"                    -only    like -on. End program after check.",
"                             (exit 16 = warning issued , 0 = OK)",
"                   The modes may be combined.",
" -filter_checksum_blocksize number",
"                   sets the currently completest attachment of checksum",
"                   informations with options -filter_md5 and -filter_crc32 .",
"                   A checksum tag for the whole stream is appended as well",
"                   as a list of block checksums. The size of a block is set",
"                   by the given number. A small number will result in many",
"                   blocks and therefore will attach a large number of 16 byte",
"                   MD5 checksums.",
"                   This option sets -filter_checksum_padsize to 1 and",
"                   -pre_checksum_size to 0. It will eventually get disabled",
"                   by any of these two.",
" -filter_checksum_padsize number",
"                   with -filter_crc32 or -filter_md5 add at least the given",
"                   number of bytes in form of checksum tags at the end of",
"                   the filtered byte stream. Such tags may be found by",
"                   -search_crc32 resp. -search_md5 if -scan_for_checksum is",
"                   enabled. Thus avoiding the need of a record list.",
"                   If number is not 1 then -filter_checksum_blocksize gets",
"                   disabled.",
" -filter_crc32 recordlist recordname",
"                   the program ends immediately when stdin closes.",
"                   Passes stdin to stdout without changes. Computes",
"                   a CRC32 checksum and counts the bytes. Together",
"                   with the recordname and a time mark they get",
"                   appended as a record to file recordlist.",
"                   The recordname maybe a text without spaces with",
"                   up to 80 characters.",
"                   Such records are used by -compare_checksum .",
"                   If -pre_checksum_size is greater than 0 there will also",
"                   be written a preview record for -search_crc32 .",
"                   CRC32 got 32 bits and may fail in 1 of 4e9 tests.",
"                   The program will return 0 if any data were read.",
"                   If no bytes were seen at stdin, no record is",
"                   written and exit value is 1.",
" -filter_md5 recordlist recordname",
"                   the program ends immediately when stdin closes.",
"                   Like -filter_crc32 but with \"RSA Data Security,",
"                   MD5 Message-Digest Algorithm\" checksum.",
"                   MD5 got 128 bits and may fail in 1 of 3e38 tests.",
" -filter_pipe shellkommando",
"                   Given before -filter_md5 , -filter_crc32 or",
"                   -filter_through this defines a shell command",
"                   which gets started as soon as the first input",
"                   byte arrives on stdin. The filter's output is",
"                   piped into that command. If no bytes are received",
"                   then the shell command is not started at all.",
" -filter_through   the program ends immediately when stdin closes.",
"                   Passes stdin to stdout without changes. Useful",
"                   together with -filter_pipe .",
" -find directory   prints the directory's name and the names of all files",
"                   and directories below. Afterwards the program ends.",
"                   Like program 'find' but without any interpretation of",
"                   the given name.",
" -find_path_list filename",
"                   reads fileaddresses line by line from given file, applies",
"                   option -find to each of it and finally ends the program.",
" -follow_symbolic_links mode",
"                   In normal mode \"off\" symbolic links are counted only",
"                   with their own (small) file size. In mode \"on\" they will",
"                   be accounted with the size of their link target. In case",
"                   of directories this means that the whole tree below",
"                   the target is measured and eventually gets split over",
"                   several volumes. The formatter program will then have",
"                   to follow symbolic links in order to put the target data",
"                   into the backup. With mkisofs this is achieved by option",
"                   \"-f\" and with afio the appropriate option is \"-h\".",
"                   In case of ill (cyclical) link situations there may be",
"                   multiple copies of the affected directories included in",
"                   the backup. Watch for error messages like :",
"                     \"Too many levels of symbolic links\"",
" -form_listing     the result list lines are shown in the form",
"                     Type Size [Partprefix]Address",
"                   the backup_cmd is not used. Partprefix may result from",
"                   -split_large_files and eventually has the form",
"                      **partnumber/total:start_byte+byte_count**",
" -form_path_list_afio listfilename safe_number_of_arguments",
"                   a special version of -form_script_raw. The file",
"                   addresses do not become arguments but are written",
"                   to listfilename (see -form_path_list_mkisofs) and",
"                   this listfilename is handed by option -path-list",
"                   to the writing command. It may then use option",
"                   -find_path_list to get a full file list for afio.",
" -form_path_list_afio_old listfilename safe_number_of_arguments",
"                   like -form_path_list_afio but without writing splitted",
"                   directories via option -dir-path-list of the writing",
"                   command.",
" -form_path_list_mkisofs listfilename safe_number_of_arguments",
"                   a special version of -form_script suitable for",
"                   versions of mkisofs which know option -path-list",
"                   With this form a very large setting of -max_times",
"                   is permittable (like 100000) since it avoids to",
"                   use them as arguments of shell commands.",
"                   listfilename will be created to store the list of",
"                   items. Therefore write permission is needed for",
"                   its hosting directory. If listfilename does not contain",
"                   a \"/\" then it gets prepended -tmp_dir.",
"                   safe_number_of_arguments should be small enough",
"                   to comfort your shell (e.g. 100).",
"                   This option sets -target_escape to \"on\".",
" -form_quoted_list_xorriso listfilename safe_number_of_arguments",
"                   a special version of -form_path_list_mkisofs suitable for",
"                   versions of xorriso which in its -as mkisofs emulation",
"                   knows option --quoted_path_list. This format can process",
"                   filenames which contain NewLine-characters.",
" -form_raw         the result list lines only show the addresses",
"                   (with eventual partprefix like with -form_listing).",
"                   The volume ends are marked by empty lines.",
" -form_script      the result list lines are appended as arguments",
"                   to the command in the form",
"                     /address[/]=address",
"                   this is the default output format.",
"                   This option sets -target_escape to \"off\".",
" -form_script_here the result list lines are appended as standard",
"                   input lines to the command in the form",
"                     address",
"                   this is done inline by a here-document with",
"                   delimiter \"+\".",
" -form_script_raw  the result list lines are appended as arguments",
"                   to the command in the form",
"                     address",
" -implicite_ignored text  addresses a problem with mkisofs-1.12b3",
"                   The text tells which characters cause the",
"                   backup command to ignore filenames that are",
"                   not explicitely given as arguments.",
"                   For mkisofs-1.12b3 text should be: \"~#\"",
" -info_script name  set the fileaddress for the separate information",
"                   script. Be aware that two files are created.",
"                   One with the name given and one with name",
"                   extension .pnr (e.g. ASKME and ASKME.pnr).",
"                   Special names are:",
"                    -off     don't generate info script",
"                    -short   info script dummy without item lists",
"                    -full    fully usable info script",
"                    -short_on_level  With -level -create_configuration",
"                             default becomes -info_script -short", 
"                   other names are used to store the script and",
"                   included in each part of the backup.",
"                   Different from file addresses the separating",
"                   character for the -info_script name is fixed",
"                   \"=\".",
"                   Example:  /ASKME=/home/cd_backup_planer_dir/ASKME",
" -info_exclude     don't include the info script in all backup",
"                   parts automatically despite a file name was set",
"                   with -info_script previously.",
" -level code       incremental backup with the configuration that",
"                   was selected by -conf_dir. The code determines",
"                   the further action of the backup :",
"                     -create_configuration creates -conf_dir ,",
"                      records the other parameters of the backup",
"                      and finally starts a -level 0 run.",
"                     -update_configuration records new parameters",
"                      in an existing -conf_dir and starts -level 0.",
"                     -off disables the incremental mode.",
"                     -last chooses the level of the most recent run.",
"                     -next chooses the next unused level (-last+1).",
"                     0,1,2,... direct level choice by number.",
"                   By -last, -next or level numbers, many options other",
"                   than -conf_dir and -level are ignored and replaced",
"                   by those from -conf_dir/specific_arguments. A few",
"                   command line options will be able to override the",
"                   recorded ones :",
"                     -backup_cmd -composition_* -dir_size_cmd",
"                     -form_* -info_* -max_* -min_*",
"                     -prompt_script -scattering -status_file",
"                     -tmp_dir -shell_to_use -verify_source_filter",
"                     (I.e. you cannot change the backup area that way)",
"                   If created or updated by this program, specific_arguments",
"                   begins with lines equivalent to these options:",
"                     -working_directory ...address...",
"                     -load_old_tree content_file_list",
"                                    vanished_file_list",
"                     -changetest_options type:size:inode:adr",
"                     -list_output_format type:size:inode:adr",
"                     -content_list_adr content_file_list.gz",
"                     -vanish_script_adr content_file_list",
"                                  /tmp/VANISHED.sh.gz=VANISHED.sh.gz",
"                     -vanish_list_adr content_file_list",
"                                      vanished_file_list.gz",
"                   If you do not want a vanish script in incremental",
"                   backups then you may disable this feature by :",
"                   option -disable_safe_leveling within your own",
"                   options to this program.",
"                   -changetest_options and -list_output_format are set",
"                   according to the start arguments. Above example assumes",
"                   the default settings.",
" -level_timestamp datestring_or_fileaddress",
"                   sets the own timestamp of an incremental backup run to",
"                   a point in time which is older than the program's start.",
"                   This timestamp will be later used by the next higher",
"                   level as criterion for -younger_than .",
"                   If datestring_or_fileaddress begins with '.' or '/' then",
"                   it is a file address of which the younger one of ctime",
"                   or mtime will be used. Else it is a time_definition as",
"                   expected by -younger_than .",
"                   Usually this setting is not needed. But if the incremental",
"                   backup is done from a filesystem snapshot then the",
"                   timestamp MUST be set to the TIMEPOINT OF THE SNAPSHOT.",
"                   It is harmless, if the timestamp is a bit too old.",
" -list_checksum recordlist searchname old_limit young_limit",
"                   the program ends immediately after listing those",
"                   records from recordlist which match the other",
"                   three arguments. The last one listed would be the",
"                   one used by -compare_checksum (see there).",
"                   The program will return 0 if any records were",
"                   listed. Otherwise its exit value is 1.",
" -list_output_format  code:code:...",
"                   defines which information shall be written to the",
"                   lines resulting from -*_list_adr options.",
"                   Multiple codes are separated by \":\".",
"                     type  the first letter indicates the file type :",
"                           d = directory , l = softlink , b = block device",
"                           - = datafile , p = pipe , 0 = unknown",
"                           Extra information may be appended as a comma",
"                           separated list:",
"                            cZahl ... ctime in seconds since 1970",
"                            dZahl ... device number",
"                            iZahl ... inode number",
"                            tZahl ... mtime in seconds since 1970",
"                            pStart:Fsiz ... indicates splitfile part address",
"                                      tells start byte and full size",
"                            PStart:Ssiz ... indicates original address",
"                                      tells start byte and split part size",
"                     size  bytecount (-1 = unknown)",
"                     crc   CRC32 content checksum (-1 = unknown)",
"                     md5   MD5 (128 bit) rather than CRC32 (32 bit)",
"                     inode append inode-number to type. With directories",
"                           or absolute addresses the device number is",
"                           appended, too. E.g.:  -,i10201  or  d,d772,i4013",
"                     mtime append mtime to type. E.g.: -,t1151999973",
"                     ctime append ctime to type. E.g.: -,c1151999973",
"                     adr   file address (always on, code ignored)",
"                     on_level_crc  like crc but only if -level is not -off",
"                     on_level_md5  like md5 if -level is not -off",
"                     on_level_inode  like inode if -level is not -off",
"                   The sequence is always type size crc adr . Eg.:",
"                     d 91286 A929B47C /home/thomas/",
" -load_old_tree  old_list_adr vanish_list_adr",
"                   loads a complete file list or a layered set of",
"                   content and vanish lists in order to form a tree",
"                   of file objects. This tree represents an old",
"                   state of the backup area (typically from the time",
"                   when the next lower level's backup was made).",
"                   This tree is used by -younger_than and by",
"                   -vanish_list_adr to determine files which have",
"                   been moved or deleted since then.",
"                   If a -level is given and the addresses contain",
"                   no \"/\" then the files are read from the lower",
"                   level_# directories. Thus no action at level 0.",
"                   If old_list_adr points to a complete list, then",
"                   no vanish lists need to be read and therefore",
"                   vanish_list_adr may be any nonempty text.",
"                   If a list file itself is missing but exists",
"                   with name extension .gz then that file used. If",
"                   the effective filename ends with .gz it gets read",
"                   via a pipe through program gunzip.",
"                   CAUTION: a loaded tree might need much memory.",
"                   NOTE : old_list_adr from this option is identical",
"                   with old_list_adr from -vanish_list_adr. If both",
"                   options are given, the last one will finally set",
"                   this address.",
"                   NOTE: -load_old_tree with nonempty addresses",
"                   sets -changetest_options to \"type:size:crc:adr\"",
" -localdev         (in effect only with NFS on SGI-IRIX systems)",
"                   exclude all remote filesystems.",
"                   Opposite of -alldev. See also -xdev",
" -main_dir text    name of a directory which shall be mapped to the",
"                   root directory of the CD. Especially useful is",
"                   the text  -auto . With this, the common directory",
"                   of the backup list (if any) is used. Text  -off",
"                   disables this mapping (default).",
" -max_items number the maximum number of items in a result list",
"                   of directories and files appended to the",
"                   backup command. (default: 400)",
" -max_size number  the maximum number of bytes per volume.",
"                   (default: 600m)",
" -max_size_stack size count",
"                   registers a number of volumes (i.e. media) with a",
"                   capacity different from -max_size. This may happen several",
"                   times. The volume group registered last will be used for",
"                   planning first. Watch the size announcement of the",
"                   volume to choose the appropriate media size. Example :",
"                   If you got 25 CDs with 700 MB and the rest stores 650 MB :",
"                     -max_size_stack 685m 25 -max_size 640m",
" -max_size_stack_save -on|off",
"                   It may happen that larger volumes cannot be filled with",
"                   more than -max_size bytes. In that case in default mode",
"                   \"-on\" the -max_size_stack volume is not counted as",
"                   consumed and will be retried with the next backup volume.",
"                   This can cause large media to be required surprisingly",
"                   late. In mode \"-off\" -max_size_stack volumes are",
"                   consumed one by one unconditionally.",
" -max_tree_depth number",
"                   If very many directories are nested then a stack overflow",
"                   can occur. An overflow is dangerous and thus is prevented",
"                   by a depth limit and a controlled program abort. The",
"                   default number is 128. It can be reduced to 16 or raised",
"                   to an arbitrary high number.",
"                   The stack size can be set under Linux by ulimit -s.",
"                   It should be at least 20 kB for each depth step.",
" -merge_trees source target dump options",
"                   The program ends immediately after this option has been",
"                   executed. It moves source and all file objects underneath",
"                   to target and analog addresses underneath. The move may",
"                   cross filesystem borders. target and the existing",
"                   directories underneath get the access permissions of",
"                   source and its subdirectories.",
"                   Files which get overwritten may be saved in a directory",
"                   tree underneath  dump . If dump is a single \"-\" then",
"                   eventually existing files underneath of target get lost",
"                   when overwritten.",
"                   If prior to this option a directory list is given with",
"                   option -load_old_tree then the properties recorded there",
"                   will be copied to the target directories. (Such a list",
"                   may be generated via -dir_list_adr.) Without the list",
"                   directory properties will be copied from source.",
"                   There are the following options :",
"                     overwrite        overwrite existing files",
"                     no_overwrite     do not overwrite such files",
"                                      do not change existing",
"                                      directory permission",
"                     overwrite_older  overwrite only if target file",
"                                      is older than source file",
"                     move             move files (delete source)",
"                     copy             copy files (preserve source)",
"                     owner_plus_w     give w-permission to owner in any case",
"                     chmod_original   copy permissions unchanged",
"                     tree_prefix=<path>  the prefix directory for loading",
"                                      a directory list with -load_old_tree.",
"                                      Only needed if <path> is not source.",
"                     tree_source      with -load_old_tree read source part",
"                                      of addresses.",
"                     tree_target      with -load_old_tree read target part.",
"                   Options may be concatenated by a colon. The empty option",
"                   list \"-\" is the same as :",
"                     \"overwrite:move:chmod_original:tree_target\" .",
" -min_size_percent number  percentage of -max_size which has to be",
"                   used of a volume before a new one may be opened.",
"                   If this percentage isn't reached then a large",
"                   directory may be split to fill the volume.",
"                   (default is 80.0)",
" -multiply_volume source repeat_faktor",
"                   the program ends immediately when stdin closes. It",
"                   reads source several times and puts out the copies to",
"                   stdout. Markers are inserted between the copies in order",
"                   to enable option -extract_multiplied to fetch single",
"                   copies out of the resulting data.",
" -no               all following fileaddresses are to be omitted",
"                   from the result lists. This option can cause",
"                   the result lists to fragment since omitting a",
"                   address within a directory requires to put to",
"                   the list all other addresses within this",
"                   directory.",
" -not              same as -no",
"",
" -not_inode        works like -no, but additionally tries to recognize the",
"                   excluded file objects by their device and inode numbers.",
"                   This test is more expensive and may put extra workload",
"                   on the disk.",
" -not_pattern shellpattern  adds shellpattern to the list of search",
"                   expressions which get tested against file leafnames and in",
"                   case of a match cause the exclusion of the file.",
"                   shellpattern may use the wildcards *,?,[...].",
" -not_regex regex  adds the regular expression regex to the list of search", 
"                   expressions which get tested against file leafnames and in",
"                   case of a match cause the exclusion of the file.",
"                   regex is an expression as known from programs grep,",
"                   sed or vi.",
" -pacifier mode|rythm  some options print intermediate messages about",
"                   their progress to stderr. If you are logging stderr",
"                   this may be annoying.",
"                     -off       disables these message",
"                     -on        enables a rythm of 1 message per second",
"                     A number sets the rythm to the a value in seconds",
"                     -timestamp prepends the messages by the number",
"                                of elapsed seconds.",
"                     -composition_memory displays the bufferstate of option",
"                                -composition_memory",
"                     -log_mode  non-truncated messages in distinct lines",
"                   All modes besides -off are combinable. E.g.:",
"                     -pacifier 3 -pacifier -timestamp -pacifier -log_mode", 
" -permute_blocks key source blocksize blockcount datastart",
"                   the program ends immediately when stdin closes. It",
"                   reads file source via random access and writes the",
"                   result to stdout. The output data are permuted blockwise.",
"                   The output data are preceded by a header part which tells",
"                   how the data have been swapped and moved :",
"                     scdbackup block permutation header v0.1",
"                     blocksize=...        Number of bytes per block",
"                     complete_blocks=...  Number of complete blocks",
"                     remainder=...        Number of bytes in last block",
"                     permutation_key=...  Permutation scheme (0 to 7)",
"                     datastart=...        Number of first data block",
"                   key defines the permutation scheme :",
"                     0= blocks stay in sequence",
"                     1= swap both halves of the file",
"                     2= mirror sequence of blocks",
"                     3= swap 1st and 2nd quarter, sawp 3rd and 4th",
"                     plus 4 = additionally interlace 1st and 2nd half",
"                     (e.g. 7 = 3+4 = swap quarters, then interlace halves)",
"                   key \"restore\" will revert the permution and put out",
"                   the original file state.",
"                   If key is preceded by \"-\" then reversion is performed,",
"                   too. But in this case, the fileheader is skipped an the",
"                   parameters are taken from arguments : key, blocksize,",
"                   blockcount, and datastart.",
"                   When converting to permuted state, blockcount can be",
"                   used to set the size of a block device. Else the size is",
"                   taken from file size or by counting block device bytes.",
"                   Under normal circumstances, blockcount and datastart",
"                   should be 0, though.",
" -pre_checksum_size number  sets the amount of data from the start",
"                   of the stream which is used to compute a preview",
"                   checksum. If this number greater than 0 then -filter_md5",
"                   and -filter_crc32 write two records. The first",
"                   with name appendix \"_pre\" may be used by option",
"                   -search_md5 and -search_crc32 to locate the",
"                   second one which describes the full data stream.",
"                   Number must be greater or equal -1. Default is 0.",
"                   If number is larger than 0 then",
"                   -filter_checksum_blocksize gets disabled.",
" -prevent_doubles on|off  if set to \"on\" the emerging backup plan is",
"                   checked wether a newly appended item already is",
"                   and eventually the new item is discarded. This test may",
"                   need lots of time and is therefore disabled by default.", 
" -problemlist_limit number",
"                   Set the maximum item number for eventual lists of",
"                   problem files from -filecheck, -exclude_size, -xdev.",
"                   If the number is exceeded with on of these lists then",
"                   the planning run will finally abort. (Default: 10000)",
" -process_blocklist action listfile skip source target",
"                   provides an overview over eventual data damages and also",
"                   recovery facilities. The program ends after this option",
"                   is done. It depends on block checksum lists.",
"                   See -filter_checksum_blocksize .",
"                   The block checksum list is read from listfile where it",
"                   has to be appended to a valid checksum tag. That may be",
"                   a backup volume written with -filter_checksum_blocksize",
"                   or a -blocklist_fileadr result.",
"                   The skip value is the number of tags to be skipped until",
"                   finally one gets read and used for processing. Normally",
"                   0 is ok.",
"                   source is a file address (\"-\"=stdin) of a potentially",
"                   damaged backup volume.",
"                   target will be a file address (\"-\"=stdin/stdout)",
"                   where undamaged blocks get assembled.", 
"                   Depending on action both may have to be random access",
"                   files or may be allowed to be single pass byte streams.",
"                   The following keywords for action are defined:",
"                   \"report\"  reads a stream from source and prints",
"                      -one per line- a list of block numbers to stdout. The",
"                      end of this bad block list is marked by the line \"@\".",
"                      target is not used.",
"                   \"collect\" reads a stream from source and writes the",
"                      undamaged blocks to target via random access.",
"                   \"mend\" reads source with random access and eventually",
"                      reads copies of damaged target blocks from source",
"                      by random access. If this succeeds then the undamaged",
"                      block copy gets written into target.",
"                      mend supports an optional start block number and an",
"                      even more optional end block number. Numbering starts",
"                      with 0. Keyword \"end\" depicts the last block.",
"                      If start block number is larger than end block number",
"                      then the blocks are processed in that reversed order.",
"                      Examples:  mend  mend:1200  mend:10k:20k  mend:end:0",
"                   \"merge\" reads source and target sequentially an writes",
"                      either an undamaged block, or the damaged one from",
"                      target or even a block of zeros to stdout.",
"                   actions which write to target do also append the block",
"                   checksum list to target's payload data.",
"                   A short description of the list parameters and a",
"                   summary of the result are printed to stderr.",
"                   If stderr is to be interpreted by a program then you",
"                   should set option  -pacifier -off  before option",
"                   -process_blocklist .",
"                   The exit value is 0 if the resulting data are supposed to",
"                   be completely undamaged. It is non 0 else.",
" -result_filetypes typelist",
"                   defines what file types may appear in the result list",
"                   resp result script. typelist consists of one or more of",
"                   the following words separated by \":\".",
"                     data          ordinary data files",
"                     dir           directories",
"                     symlink       symbolic links",
"                     block_device  device files of hard disks, CD, ...",
"                     unclassified  anything different from above list",
"                     any           everything together",
"                     arg_redirect  redirect explicitely given symlinks",
"                   File types which are not mentioned in typelist get",
"                   discarded when they are to be added to the list of a",
"                   backup piece.",
"                   If pseudo-type  arg_redirect  is given, then the arguments",
"                   of this program get checked for symbolic links. Those are",
"                   replaced by their link targets (data sources).",
" -scan_for_checksum mode",
"                   controls -search_crc32 and -search_md5 scanning for",
"                   checksum tags (appended by -filter_checksum_padsize).",
"                   Following modes are defined :",
"                    -on     Search enabled. Search properties unchanged.",
"                    -off    Search disabled. Search properties unchanged.",
"                    -first  Search on. End after first valid tag.",
"                    -whole  Search on. Search up to end of byte stream.",
"                    -standalone  Search on. Do not read recordlist.",
"                    -withlist    Search on. Read and use recordlist.",
"                    -print  Search on. List valid tags at end of run.",
"                    -noprint     Search on. Do not list.",
"                    -anytag      Accept any valid checksum tag.",
"                    -blocklist   Accept only tags with block checksum list.",
"                    -pacifier    Print progress messages to stderr.",
"                    -nopacifier  Do not print progress messages.",
"                    -source:text Use file address standard input",
"                    -skip:zahl   Number of bytes which got skipped at input.",
"                                 With -source: this skipping is done after",
"                                 opening the file address.",
"                   Combined modes are permitted. Default :",
"                    -scan_for_checksum -withlist -scan_for_checksum -off",
"                    -scan_for_checksum -anytag -scan_for_checksum -nopacifier",
" -scattering -on|-off  mode -on allows moving large files to CDs",
"                   different than those which host their directory",
"                   siblings. This helps to fulfill -min_size_percent",
"                   but can lead to scattered directory contents. This option",
"                   is in effect only with  -composition_mode default .",
" -search_crc32 recordlist searchname old_limit young_limit",
"                   like -compare_checksum but with automatic volume",
"                   recognition and only for records with CRC32",
"                   checksums. The search method depends on the",
"                   setting of  -pre_checksum_size  . If it is 0 then",
"                   all records are read, sorted by size and compared",
"                   with intermediate checksums of the data stream.",
"                   The stream is read up to its end and more than",
"                   one match may occur if streams with identical",
"                   data were recorded. With -pre_checksum_size -1",
"                   reading ends at the first match.",
"                   If -pre_checksum_size is >0 then at first a",
"                   matching preview record is searched to determine",
"                   the record with the full stream's checksum.",
"                   The preview checksums are computed by -filter_md5",
"                   and -filter_crc32 if pre_checksum_size was not",
"                   set to 0. The sizes of record and search have",
"                   to match in this case.",
" -search_md5 recordlist searchname old_limit young_limit",
"                   like -search_crc32 but with all records which",
"                   have MD5 checksums.",
" -set_shell_in_scdbackup_scripts shellname filename ... filename",
"                   a very special option which is used only for",
"                   configuration. The given files get tested wether",
"                   they start with \"#!\". If true, the rest of the",
"                   first line gets replaced by shellname. No other",
"                   option may follow since the rest of arguments is",
"                   used as a list of filenames.",
" -shell_to_use progname  the progname is written after \"#!\"",
"                   into the first line of scripts generated by",
"                   this program. Default: /bin/sh",
" -special_file fileaddress size type",
"                   defines a file as a special data source. Currently only",
"                   type \"block_device\" is permitted. Additionally option",
"                   -split_large_files has to be given with a split_directory",
"                   and a split_command. The content of the block device will",
"                   be included in the backup in form of one or more parts.",
"                   If size is <=0, then the device is read up to EOF and",
"                   the resulting count is used. Fileaddress may be of form",
"                   target=source.",
" -split_file_prefix  text",
"                   defines a text which is put in front of the names of file",
"                   pieces which emerge by -split_large_files.",
" -split_file_suffix  text",
"                   like -split_file_prefix . The text is put between name",
"                   and counting numbers. See also -split_list_adr.",
" -split_large_files split_directory split_size split_command sc_option",
"                   sets the size limit for regular files together with a",
"                   directory address and a shell command which is able",
"                   to cut a piece out of a file. This program itself can",
"                   act as such a command with sc_option -cut_file . Argument",
"                   split_command is subject to special -bdir substitution.",
"                   Files which are larger than split_size are processed as",
"                   several separate parts. Shell scripts resulting from this",
"                   program will use split_command to temporarily store the",
"                   parts in split_directory. All partial files of a volume",
"                   have to fit into this directory. The directory has to",
"                   contain a file named scdbackup_split_file_directory_mark .",
"                   If split_directory is '-' then this feature is disabled.",
"                   split_directory '-xorriso_cut_out' uses option -cut_out",
"                   of ISO 9660 formatter program xorriso. In this case no",
"                   directory is needed, split_command and sc_option get",
"                   ignored then.",
"                   If split_size is '-' then the effective -max_size - 1m",
"                   is used. Keep in mind that -exclude_size can prevent",
"                   large files from being split and processed at all.",
"                   If split_command is '-' then the previously set command",
"                   stays in effect. Use '-' as sc_option in this case.",
" -split_list_adr fileaddress",
"                   writes a list of file parts caused by -split_large_files.",
"                   The information in this file may be helpful if a script",
"                   output format is chosen which does not contain target",
"                   address (e.g. -form_path_list_afio). When restoring such",
"                   a backup, the file parts will appear in split_directory",
"                   and need to be put to their original positions.",
"                   General list format, working directory, compression and",
"                   backup inclusion are as described at -complete_list_adr.",
"                   Each data line describes a file part by the form",
"                     partprefix part_address=buffer_address",
"                   partprefix has the form",
"                     **partnumber/total:start_byte+byte_count**",
"                   part_address consists of",
"                     original_dir\"/\"marked_name\"_\"partno\"_\"total", 
"                   marked_name consists of",
"                     prefix_text\"\"original_name\"\"suffix_text",
"                   buffer_address is composed of",
"                     split_dir\"/\"basename_of_part_address\"-\"hexcount",
"                   where hexcount is unique for the whole backup.",
" -status_file filename  If the filename is not empty and not \"-\"",
"                   then the result script will write the current",
"                   volume number into this file before it calls",
"                   the -backup_cmd . Also the script will be able",
"                   to resume at the start of an interrupted volume",
"                   when given the argument \"AUTO\" rather than",
"                   a volume number.",
" -tmp_dir directory_address",
"                   sets a directory for storing temporary files. One of its",
"                   uses is with list file addresses without '/' which",
"                   would go to the level directories if there was any -level",
"                   set (see -complete_list_adr). directory_address is",
"                   subject to special -bdir substitution.",
" -target_escape mode  controls the handling of \"=\"-characters in target",
"                   addresses. mode \"on\" causes \"=\" and \"\\\" to be",
"                   preceeded by an additional \"\\\"-character. Such target",
"                   addresses are recognized by modern versions of mkisofs",
"                   which know option -graft-points.",
"                   mode \"off\" exchanges in target addresses each \"=\"",
"                   by \"_\". \"\\\" is not changed. Use with mkisofs 1.12b3.",
"                   Note: Option -form_path_list sets mode to \"on\"",
"                         -form_script sets it to \"off\". Afterwards it may",
"                         be set explicitely by -target_escape.",
" -target_separator text    the first character in text is used as",
"                   separating character between target address",
"                   and file address. This does only apply to input but",
"                   not to output where eventually \"=\" is used.",
"                   An empty text disables explicit target addresses.",
"                   Default setting is \"=\". Within the target address",
"                   eventual \"=\" and \"\\\" characters have to be escaped",
"                   by preceeding \"\\\" characters.", 
"                   Example: map /home/office to the / directory of the CD:",
"                    -target_separator ':' /:/home/office -target_separator ''",
"                   Example: map /home/test/=\\ to /=\\",
"                     -target_separator '=' '/\\=\\\\'='/home/test/=\\'",
" -tree_hashsize number   option for balancing speed of file search in the",
"                   directories of old_tree against the memory consumption of",
"                   -old_tree. Each directory in old_tree occupies by its",
"                   hashvector  number*sizeof(int) bytes. The search in large",
"                   directories is accelerated by at about the factor number.",
"                   The default of 20 is well suited for normal sized",
"                   directories of 20 to 1000 files. Directories with several",
"                   ten thousand files may slow down the program substantially",
"                   whereas trees with very many directories may consume too",
"                   much memory.",
" -vanish_list_adr  old_list_adr  result_address",
"                   checks the file objects which have been recorded",
"                   some time ago in old_list_adr. Those objects",
"                   which have vanished since then are reported to",
"                   file result_address. If a -level is given and",
"                   result_address contains no \"/\" then the file is",
"                   created in the level_# directory. (See -conf_dir)",
"                   If result_address ends with .gz then the output",
"                   gets piped through program gzip.",
"                   If a target is given in result_address then the",
"                   resulting file will be appended to the last",
"                   backup volume. (e.g. /vanished.gz=vanished.gz)",
"                   old_list_adr may be output of -complete_list_adr",
"                   Alternatively a plain list of file addresses may",
"                   be used.",
"                   If a tree gets loaded by -load_old_tree then it",
"                   is used for comparison. If there is no old tree",
"                   then the old_list_adr gets read directly but it",
"                   has to be a complete list in a single file.",
"                   NOTE : old_list_adr from this option is identical",
"                   with old_list_adr from -load_old_tree. If both",
"                   options are given, the last one will finally set",
"                   the effective address.",
"                   CAUTION: If -load_old_tree adresses are empty or",
"                   -changetest_options is \"off\" then the result",
"                   list may contain old file adresses that have been",
"                   moved but the backup plan will not contain the",
"                   new addresses. Best is to also test for CRCs",
"                   (see -changetest_options)",
" -vanish_script_adr  old_content_list_adr  result_address",
"                   CAUTION: Running the result script is dangerous !",
"                            Especially if generated without CRC test",
"                   Same as -vanish_list_adr but reports shell",
"                   commands to form a script that removes all",
"                   vanished objects. Such a removal is necessary",
"                   before applying an incremental level during",
"                   re-install of a multilevel backup. But this",
"                   script does not check against cruel accidents.",
"                   NOTE: the script contains target addresses only.",
" -verify_filelist  list_adr  directory_prefix",
"                   NOTE: If you want to verify the backuped files on CD or",
"                         during restore then you should better use",
"                         -verify_targetlist.",
"                   The program ends immediately after this option is",
"                   done. It reads file addresses from list_adr",
"                   together with eventual informations about type,",
"                   size and checksum. The informations get compared",
"                   with the current state of the file. Its address",
"                   is composed from directory_prefix and the address",
"                   read from the list.",
"                   If the addressed file exists and if the other",
"                   properties match, then the file's address is",
"                   printed on stdout. Else there will be an error",
"                   message at stderr.",
"                   If everything existed and matched, the program",
"                   exits with value 0. Else this value is 1.",
"                   Currently this is mainly a harmless version of",
"                    -delete_filelist . But that might change.",
" -verify_sources  listfile",
"                   the program ends immediately after this option is",
"                   done. This option is similar to -verify_filelist",
"                   but it expects addresses of form target=source",
"                   and only checks the source parts. No directory",
"                   prefix gets applied.",
" -verify_source_filter  shell_filter",
"                   With output format -form_path_list_mkisofs the",
"                   filelist can be cleaned from undesired addresses",
"                   immediately before the prompt to insert a CD.",
"                   The decision is made by shell_filter which has to",
"                   put out any valid file to stdout and omit any",
"                   invalid file. Use this filter in order to omit",
"                   all files which have been removed since the",
"                   backup was planned :",
"                     ..../cd_backup_planer -verify_sources -",
"                   An empty text as shell_filter disables filtering.",
"                   shell_filter is subject to special -bdir substitution.",
" -verify_targetlist  list_adr  directory_prefix",
"                   like -verify_filelist but processing the target addresses",
"                   of the listed files. If there have been target=source",
"                   redirections when a backup was made, then this can make",
"                   a big difference.",
" -volume_info      pass additional arguments to the backup command",
"                   (before file names) : -v volume -t total_volumes",
" -volume_offset  number",
"                   Adds number to all output of volume numbers and subtracts",
"                   it from input of volume numbers.",
" -working_directory  directory_address",
"                   set the working directory to the given address.",
"                   This option should be used only once and only",
"                   before any fileaddress is given.",
" -write_random  target amount randomseed",
"                   the program ends immediately after this option is done.",
"                   It writes into the target file (\"-\" = stdout) the given",
"                   amount of random bytes or a periodical byte pattern of",
"                   up to 16 bytes. If randomseed begins with \"=\" then the",
"                   rest of it is the periodical pattern. If the \"=\" is",
"                   followed by \"~\" then that pattern is inverted into its",
"                   binary complement. Without \"=\", randomseed is the key",
"                   for a pseudo random number generator based on MD5 and",
"                   EXOR. If this key starts with \"-\" then it is used for a",
"                   slow method which is free of EXOR. The number sequence of",
"                   a particular randomseed is reproducible. Large amounts",
"                   like 100g or 2t are permitted if the OS can handle them.",
" -write_to_file targetfile byteaddress bytecount bytecode [... bytecode]",
" -write_to_file targetfile byteaddress \"string\" piece_of_text",
" -write_to_file targetfile byteaddress \"file\" sourcefile",
"                   the program ends immediately after this option is done.",
"                   It opens the existing targetfile for writing and positions",
"                   the write pointer at byteaddress (file start = 0).",
"                   If the third argument is a number between 1 and 4096 then",
"                   it is the count of the following bytecode arguments.",
"                   Permissible are decimal bytecodes between 0 and 255.",
"                   With \"string\" the forth argument is written into",
"                   targetfile. With \"file\" the forth argument gives the",
"                   path of a file of which the content will be copied into",
"                   targetfile.",
"                   targetfile \"-\" means stdout. It is only permitted with",
"                   byteaddress 0. sourcefile \"-\" means stdin.",
" -xdev             exclude all filesystems mounted below the items",
"                   of the backuplist. Opposite of -alldev .",
" -yes              all following fileaddresses are appended to the",
"                   backuplist (i.e. these files are to be backuped).",
"                   This option ends the range of option \"-no\".",
"                   It is the default mode for fileaddresses.",
"                   If the option -yes is given at least once, then",
"                   the option -no does not apply to the backuplist.",
"                   ( -yes x -not x   : file x is in backup",
"                     x -not x        : file x is not in backup)",
" -younger_than time_definition  try to backup only files where the",
"                   content or inode changed since the given time.",
"                   Moves of a file within the same file system",
"                   cannot be detected, regrettably.",
"                   Time definitions can have the following forms:",
"                   Absolute timepoints can be defined by the input",
"                   format MMDDhhmm[[CC]YY][.ss]] like with UNIX",
"                   command 'date'. For example:",
"                     020616122000 is 6 Feb 2000 16:12:00 local time.",
"                   Also the format YYMMDD[.hhmm[ss]] is recognized:",
"                     11 May 2001 23:45 = A10511.2345",
"                     1 Mar 2010 = B00301 , 14 Jul 1998 = 980714",
"                   Relative times are preceded by '-' or '+'. They",
"                   are added to the current time. '-' therefore",
"                   points to the past, '+' points to the future. The",
"                   basic unit is the second. It can be modified by",
"                   a letter at the end of the number:",
"                     seconds 1s=1 , hours 1h=3600s , days 1d=24h ,",
"                     weeks 1w=7d , months 1m=31d ,",
"                     years y : Xy=X*365.25d+1d",
"                   Please note : 'm' and 'y' are possibly slightly",
"                   larger than the actual calendar time range.",
"                   Example:",
"                     -3w  is 3 weeks in the past = -1814400s .",
"",
"Examples:",
"",
"Divide the content of the working directory into volumes of",
"625 MB and maximum 200 independend items. Create info script ASKME",
"in working directory and include it in each backup part.",
"",
"  %s \\",
"     -max_items 200 -max_size 625m -info_script ASKME .  >bscript",
"  chmod u+x bscript",
"",
"To actually perfom the backup, execute :",
"  bscript",
"",
"",
"With this option the ASKME script is generated at a separate place",
"and mapped to the root directory :",
"",
"  %s \\",
"    -info_script /ASKME=/home/backup_planer_dir/ASKME",
"",
"",
"Divide the directores \"office\" and \"images\" into volumes of",
"default size. Do not include subdirectories \"office/private\" and",
"\"images/girlfriends\"",
"Replace the omitted \"images/girlfriends\" by \"images/cats\"",
"",
"  %s \\",
"    office images -target_separator = /images/girlfriends=images/cats \\",
"    -not office/private images/girlfriends \\",
"  >bskript",
"",
"",
"Fill DC-6250 tapes with compressed tar archives.",
"",
"  %s \\",
"    -backup_cmd 'read dummy ; tar cvzf /dev/rmt0' \\",
"    -form_script_raw -max_size 245m -max_items 100 \\",
"    -info_script ASKME \\",
"    . \\",
"  >bskript",
"",
"",

"Create a configuration for incremental backups and perform planning",
"of the first -level 0 run. By -level -create_configuration a new",
"directory gets created which will be called $HOME/my_backup_conf .",
"",
"  %s \\",
"    -conf_dir $HOME/my_backup_conf -level -create_configuration \\",
"    ... all arguments defining the full backup ... \\",
"  >bskript",
"",
"The configuration may be used by simply giving this directory and",
"the desired level. All other parameters of the backup will be read",
"from $HOME/my_backup_conf/specific_arguments (1 line = 1 argument).",
"",
"One should use a fixed backup scheme. Let us assume that we made",
"the configuration and the first -level 0 backup at sunday of week 0.",
"",
"A backup scheme which from monday to saturday records only the",
"daily changes and makes a full backup each sunday.",
"  Monday  :  -conf_dir $HOME/my_backup_conf -level -next",
"  ...",
"  Saturday:  -conf_dir $HOME/my_backup_conf -level -next",
"  Sunday  :  -conf_dir $HOME/my_backup_conf -level 0",
"Now the media of monday to saturday may be recycled.",
"",
"A scheme which makes quarterly (13 weeks) full backups, four-weekly",
"and weekly updates (-conf_dir same as above). This scheme lets you",
"retrieve old file versions for quite a long time.",
"  Week  1 :  -level 1",
"  Week  2 :  -level 2",
"  Week  3 :  -level 2",
"  Week  4 :  -level 2",
"  Week  5 :  -level 1",
"  Week  6 :  -level 2",
"  Week  7 :  -level 2",
"  Week  8 :  -level 2",
"  Week  9 :  -level 1",
"  Week 10 :  -level 2",
"  Week 11 :  -level 2",
"  Week 12 :  -level 2",
"  Week 13 :  -level 0",
"For week 14 hop to week 1.",
"The backup of week 12 also covers weeks 11 and 10.",
"The backup of week 9 also covers week 8 to 1.",
"It is advised to keep levels 0 and 1 for a longer time.",
"",
"To combine both schemes shown, simply replace the sunday run of",
"scheme 1 by the weekly run of scheme 2.",
"",
"If you ever need to re-install an incremental backup with multiple",
"levels you will begin with copying level 0 first. Then you need to",
"remove the files which have vanished in the week between the level 0",
"backup and the next level.",
"For that purpose you will find on the last backup volume of level 1",
"a file named /tmp/VANISHED.sh.gz . It does not have x-permission to",
"avoid ugly accidents. You can only trust it if it has been produced",
"reading a content list with CRCs or inode numbers. This is true by",
"default but may be changed with options like -content_list_adr ,",
"-list_output_format , -load_old_tree , -changetest_options ,",
"-vanish_script_adr .",
"If you want to be very cautious, you may use option \"md5\" rather",
"than \"crc\" with -list_output_format.",
"",
"You will have to apply  chmod u+x  and  gunzip  after copying or you",
"will have to execute it on-the-fly via a pipe like :",
"   gunzip </cdrom/tmp/VANISHED.sh.gz | /bin/sh",
"Better have a look at its content to decide wether you think the",
"condemned files are really obsolete. Last chance to save old trash.",
"",
"The vanish script of level 1 may be executed at any time after",
"level 0 was installed and before level 2 gets installed. None of",
"the condemned files is contained in the level 1 backup.",
"If disk space is short, better execute it before installing level 1.",
"Otherwise it is more convenient to execute /tmp/VANISHED.sh after",
"it was copied to the disk.",
"When installing the following levels 2,3,4,... do not forget to",
"execute their vanish scripts in time and to remove the last one",
"from the /tmp directory after all work is done.",
"",
"",

#endif /* ! Cdbbackup_planer_no_helP */

"",
"@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@",
"                     dirscript_size  add size comments to commands in",
"                           result of -dir_script_adr. E.g.: #size 4096",
""
 };
 char line[4096];
 FILE *fp;
 int i;
 
 fp= stdout;
 if(flag&1)
   fp= stderr;

 MELDEN("");
 sprintf(line,"cd_backup_planer version %s",PROG_VERSION);
 MELDEN(line);
 MELDEN("");
 for(i=0;strcmp(hilfetext[i],
                "@ENDE_DES_HILFETEXTS_(HOFFENTLICH_EINMALIG_DOOFER_TEXT)@"
                )!=0;i++){
   sprintf(line,hilfetext[i],progname);
   MELDEN(line);
 }

 return(1);
}


/* ----------------- Backup_plan - end ---------- */


static char cmd[]= {"cd_backup_kommando"};
int max_items= 400;
double max_size= 600*1024*1024;

int main(argc,argv)
int argc;
char **argv;
{
 int ret;
 struct Backup_plaN *plan,plan_data;

 Cd_backup_planer_start_timE= time(0);
 getcwd(Cd_backup_planer_cwD,sizeof(Cd_backup_planer_cwD)-1);

 plan= &plan_data;
 ret= Backup_plan_init(plan,max_items,max_size,cmd,0);
 if(ret<=0) return(2);

 ret= Backup_plan_setup(plan,argc,argv,2|8);
 if(ret<0)
   exit(0); /*    handsome end (e.g. -help) */
 if(ret>0)
   exit(ret); /*  troublesome end */

 fprintf(stderr,"\ncd_backup_planer %s : planning backup\n\n",PROG_VERSION);
 if(plan->startliste->anz<=0 && plan->configuration!=2) {
   fprintf(stderr,"\n--- No files given as arguments !\n\n");

#ifdef Cdbackup_planer_developpeR
   { int i;
     for(i=0;i<argc;i++)
       fprintf(stderr,"%d: %s\n",i,argv[i]);
     fprintf(stderr,"\n\n");
   }
#endif /* Cdbackup_planer_developpeR */
   
   exit(1);
 }

 
 ret= Backup_plan_conf_main(plan,argc,argv,0);
 if(ret<=0)
   exit(1);

 ret= Backup_plan_list_main_dirs(plan,0);
 if(ret<=0)
   exit(1);

 ret= Backup_plan_generate_lists(plan,0);
 if(ret<=0)
   exit(1-ret);

#ifndef Crawler_ruleS

 if(!plan->use_crawler)
  if(plan->precheck>0 || plan->precheck_xdev || (plan->precheck_agetest&1)
    || (plan->precheck_sizetest&1) || plan->split_large_files) {
   ret= Backup_plan_precheck(plan,1);
   if(ret<0 || (ret&(1<<15))) {
     fprintf(stderr,"--- File check failed. Program aborted.\n");
     exit(16);
   }
 }

#endif /* ! Crawler_ruleS */

 if(plan->precheck==2)
   exit(0);

 Fileliste_destroy(&(plan->old_tree),0);

 if(plan->collision_test==1) 
   if(Backup_plan_collision_test(plan,1)<0) {
     fprintf(stderr,"--- Collision test failed. Program aborted.\n");
     exit(16);
   }

 if(plan->auskunft_name->adressen[0][0]) 
   unlink(plan->auskunft_name->adressen[0]);

 if((plan->composition_mode&(1<<15))) { 
   /* composition was done in crawler */;
 } else {
   ret= Backup_plan_compose(plan,plan->startliste,2|4|16);
   if(ret<=0)
     exit(1-ret);
 }

 /* eventually omit content list if it was done in crawler */
 ret= Backup_plan_generate_lists(plan,1);
 if(ret<=0)
   exit(1-ret);

 Backup_plan_compose(plan,NULL,8); /* just print final message */

 if(plan->collision_test==2 ) { /* || plan->collision_test==1) */
   fprintf(stderr,"performing collision test\n");
   if(Backup_plan_collision_test(plan,0)<0) {
     fprintf(stderr,"--- Collision test failed. Program aborted.\n");
     exit(16);
   }
 }

 fprintf(stderr,"writing result script\n");
 ret= Backup_plan_skript(plan,0);
 if(ret!=0)
   exit(ret);

 if(plan->configuration==2) {
   fprintf(stderr,"updating backup configuration\n");
   ret= Backup_plan_conf_update(plan,0);
   if(ret<=0)
     exit(1);
 }
 fprintf(stderr,"cd_backup_planer %s : planning complete",PROG_VERSION);
 if(Cd_backup_planer_pacifier_show_timE)
   fprintf(stderr," after %d seconds",
           (int) time(0)-Cd_backup_planer_start_timE);
 fprintf(stderr,"\n");

#ifdef Cdbackup_planer_developpeR
 if(!plan->use_crawler) {
   if(!Cd_backup_planer_trust_fret64)
     fprintf(stderr,
             "developer message:  false new= %d , false old= %d\n",
             Cd_backup_planer_false_young,Cd_backup_planer_false_zu_alt);
   if(Cd_backup_planer_risk_d_enabled)
     fprintf(stderr,
             "developer message:  double on del= %d , not empty on del= %d\n",
         Cd_backup_planer_double_deleted,Cd_backup_planer_not_empty_on_delete);
 }
 if(Cd_backup_planer_double_deleted>0 || 
    Cd_backup_planer_not_empty_on_delete>0) {
   fprintf(stderr,
         "FATAL: The outcome of an experimental setting made this\n");
   fprintf(stderr,
         "       backup run questionable.\n");
   fprintf(stderr,
         "       Rerun it without this option:\n"); 
   fprintf(stderr,
         "          -test_command risk_d\n"); 
   exit(1);
 } else if(!plan->use_crawler)
   fprintf(stderr,"     user message:  Do Not Worry\n");
#endif /* Cdbackup_planer_developpeR */

 exit(0);
}
